----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    rod_bit_gen.vhd
-- Description:
--    ROD Bit Stream Generator
--       SCT Version
--
-- Reads a test data file and produces a data stream to help simulate
-- the serial data coming from the 96 front end links. The data file
-- contains "1" and "0" characters. This program sequentially scans
-- through the data file and outputs the serial data from MSB to LSB at
-- 96 bits per clock cycle. The data file can have spaces and
-- carriages returns in it as white space and can have the # sign to
-- comment out the rest of line. Note that tabs are not recognized as
-- white space and other characters will cause a value 'Z' to be output.
-- The data file can have up to 20,000,000 characters in it. This
-- module will also produce the 40 Mhz system clock and will issue a
-- powerup reset pulse over the first 4 clock cycles.
--
-- The code works by first reading in the data file and storing the
-- entire string of bits into a large array of characters. As the
-- data file is being read, whitespace and comments are filtered out
-- before being written into the array.
--
-- After the data from the file is written in, the bits are then read
-- back out 96 bits at a time (MSB to LSB) to simulate the raw SCT
-- serial data.
--
----------------------------------------------------------------------------

library IEEE ;
use IEEE.STD_LOGIC_1164.all ;
use IEEE.std_logic_arith.all;

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

entity dbg_bits_source is
  port (
      clk_out           : out  STD_LOGIC ; -- 40 Mhz clock to be distributed
      rst_out           : out  STD_LOGIC ; -- Powerup Reset
      -- 96 link front end data
      link_output       : out  STD_LOGIC_VECTOR(95 downto 0); 
      vcc_out           : out  STD_LOGIC ; -- connects to VCC to drive it.
      gnd_out           : out  STD_LOGIC   -- connects to GND to drive it.
      ) ;
END  dbg_bits_source;

architecture rtl of dbg_bits_source is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  signal tsf_ctr: integer; -- index into data bit stream character array
  signal rst : std_logic := '1'; -- internal version of powerup reset signal
  signal clk : std_logic := '0'; -- internal version of clock signal
  signal data_input : STD_LOGIC_VECTOR(95 downto 0); -- internal
     -- version of 96 link output signals
  constant clk_period : time := 25 ns; -- set system clock period here
  constant max_file_length : integer := 60000000; -- max file size.
     -- Change here for bigger files (keep in mind it eats up memory)
 
----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------
  
BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

  -- since you can't read an output signal, I used internal versions
  -- that get reassigned here before going to the output.
  clk_out <= clk;
  rst_out <= rst;
  link_output <= data_input;

------------------------------------------------------------------------------
-- Test Stand: File to Bitstream  
------------------------------------------------------------------------------
  vcc_out <= '1';
  gnd_out <= '0';

  -- Generate reset pulse and clock signals
  rst <= '0' after ((clk_period * 4) + 1 ns);
  clk <= not clk after (clk_period/2);
    
  rodbitstr: process  (clk, rst)
    type load_file_type is file of character;
    -- data file path/filename
  
  file load_file: load_file_type is "/home/edafiles/atlas/rod/fpga_code/formatter_sct/simulation_input_files/96link.rod";

    type storage_array is array(0 to max_file_length) of character;
    -- character array containing the data bitstream
    variable storage: storage_array;
    variable index: integer;
    -- location of last character
    variable max_tsf_ctr: integer;
    variable offset: integer;
    variable tsf_ctr: integer;
    variable i: integer;
    variable outpnibble : std_logic_vector (95 downto 0) := (others =>'0');
  begin
    if (rst = '1') then
      -- Main loop to read data file, filter comments/white space and
      -- write data to character array
      tsf_ctr := 0;
      index:=0;
      data_input <= (others => '0');
      while not endfile(load_file) loop
	read(load_file, storage(index));
	if (storage(index) /= lf and storage(index) /= ' ') then
          assert (index < max_file_length)
            report "input file is too long"
            severity failure;
          if (storage(index) = '#') then
            while (not endfile(load_file) and storage(index) /= lf) loop
              if (not endfile(load_file)) then
                 read(load_file, storage(index));
              end if;
            end loop;
            index:= index - 1;
          end if;  
	  index := index+1;
        else
          storage(index):='0';
	end if;
        max_tsf_ctr := index;
      end loop;
    elsif (clk'event and clk = '1') then
      -- Simple clocked loop that read character array and outputs 96
      -- bit data words. Any character not a '1' or '0' gets 'Z' value
      for i in 95 downto 0 loop        
        if (storage(tsf_ctr) = '0') then
          outpnibble(i) := '0';
        elsif (storage(tsf_ctr) = '1') then
          outpnibble(i) := '1';
        else       
          outpnibble(i) := 'Z';
        end if;
      data_input <= outpnibble after 5 ns;
      -- Stops when the last character is read in and outputs all
      -- zeroes after that. The last word in the data file should
      -- contain dummy data of all zeroes to make sure it doesn't cut off
      --  or lose the last few data bits in the file.
      if (tsf_ctr < max_tsf_ctr-1) then
        tsf_ctr := tsf_ctr + 1;
      else
        data_input <= (others => '0') after 5 ns;
      end if;
      end loop;
    end if;
end process;

END rtl;

