----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    dsp_sbsram.vhd
-- Description:
--    ROD dsp sbsram simulator
--       SCT Version
--
-- Produces a data stream to help simulate the DSP sbsram interface
-- 
-- 
----------------------------------------------------------------------------

library IEEE ;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;

library UNISIM;
use unisim.all;

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

entity dsp_sbsram is
port (
    clk40_in : in  std_logic; -- Distributed 40 Mhz clock
    clk80_in : in  std_logic; -- Distributed 80 Mhz clock
    rst_n_in : in  std_logic; -- Powerup Reset

-- DSP signals
--    dsp0_sbsramdata_in   : in  std_logic_vector(31 downto 0);
    dsp0_int4_in : in  std_logic;
    dsp0_sbram_me	 : out std_logic;
    dsp0_sbsramclk       : out std_logic;
    dsp0_send_data       : out std_logic;
    dsp0_halt_out	 : out std_logic;
--    dsp1_sbsramdata_in   : in  std_logic_vector(31 downto 0);
    dsp1_int4_in : in  std_logic;
    dsp1_sbram_me        : out std_logic;
    dsp1_sbsramclk       : out std_logic;
    dsp1_send_data       : out std_logic;
    dsp1_halt_out	 : out std_logic;
--    dsp2_sbsramdata_in   : in  std_logic_vector(31 downto 0);
    dsp2_int4_in : in  std_logic;
    dsp2_sbram_me        : out std_logic;
    dsp2_sbsramclk       : out std_logic;
    dsp2_send_data       : out std_logic;
    dsp2_halt_out	 : out std_logic;
--    dsp3_sbsramdata_in   : in  std_logic_vector(31 downto 0);
    dsp3_int4_in : in  std_logic;
    dsp3_sbram_me	 : out std_logic;
    dsp3_sbsramclk       : out std_logic;
    dsp3_send_data       : out std_logic;
    dsp3_halt_out	 : out std_logic 
    );
end  dsp_sbsram;

architecture rtl of dsp_sbsram is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

signal tsf_ctr : integer; -- index into data bit stream character array
signal rst_n : std_logic; -- internal version of powerup reset signal
signal clk40 : std_logic; -- internal version of clock signal
signal gnd   : std_logic; 

signal dsp0_send_data_n_i : std_logic;
signal dsp1_send_data_n_i : std_logic;
signal dsp2_send_data_n_i : std_logic;
signal dsp3_send_data_n_i : std_logic;

signal dsp0_sbram_me_i : std_logic;
signal dsp1_sbram_me_i : std_logic;
signal dsp2_sbram_me_i : std_logic;
signal dsp3_sbram_me_i : std_logic;

signal dsp0_haltoutput_i : std_logic;
signal dsp1_haltoutput_i : std_logic;
signal dsp2_haltoutput_i : std_logic;
signal dsp3_haltoutput_i : std_logic;

signal dsp0_count_i : integer;
signal dsp1_count_i : integer;
signal dsp2_count_i : integer;
signal dsp3_count_i : integer;

----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------
  

begin  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------
  
rst_n <= rst_n_in;
gnd <= '0';

dsp0_sbsramclk <= clk80_in;
dsp1_sbsramclk <= clk80_in;
dsp2_sbsramclk <= clk80_in;
dsp3_sbsramclk <= clk80_in; 
 
dsp0_send_data <= dsp0_send_data_n_i;
dsp1_send_data <= dsp1_send_data_n_i;
dsp2_send_data <= dsp2_send_data_n_i;
dsp3_send_data <= dsp3_send_data_n_i;

dsp0_sbram_me <= dsp0_sbram_me_i;
dsp1_sbram_me <= dsp1_sbram_me_i;
dsp2_sbram_me <= dsp1_sbram_me_i;
dsp3_sbram_me <= dsp1_sbram_me_i;

dsp0_halt_out <= dsp0_haltoutput_i;
dsp1_halt_out <= dsp1_haltoutput_i;
dsp2_halt_out <= dsp2_haltoutput_i;
dsp3_halt_out <= dsp3_haltoutput_i;


------------------------------------------------------------------------------
-- Test Stand: File to Bitstream  
------------------------------------------------------------------------------

dsp_send_data_n : process (clk40, 
                           rst_n,
                           dsp0_int4_in,
                           dsp1_int4_in,
                           dsp2_int4_in,
                           dsp3_int4_in
                           )

begin
    if (rst_n = '0') then
        dsp0_send_data_n_i <= '1';
        dsp0_sbram_me_i <= '1';
        dsp0_haltoutput_i <= '0';
        dsp1_send_data_n_i <= '1';
        dsp1_sbram_me_i <= '1';
        dsp1_haltoutput_i <= '0';
        dsp2_send_data_n_i <= '1';
        dsp2_sbram_me_i <= '1';
        dsp2_haltoutput_i <= '0';
        dsp3_send_data_n_i <= '1';
        dsp3_sbram_me_i <= '1';
        dsp3_haltoutput_i <= '0';
        dsp0_count_i <= 0;
        dsp1_count_i <= 0;
        dsp2_count_i <= 0;
        dsp3_count_i <= 0;
    elsif (clk40'event and clk40 = '1') then
 
       if ((dsp0_int4_in = '0' and dsp0_count_i < 32) or (dsp0_int4_in = '1' and dsp0_count_i > 0 and dsp0_count_i < 32)) then
            dsp0_send_data_n_i <= '0';
            dsp0_sbram_me_i <= '0';
            dsp0_count_i <= dsp0_count_i + 1;
        elsif (dsp0_count_i = 32) then
            dsp0_send_data_n_i <= '1';
            dsp0_sbram_me_i <= '1';
            dsp0_count_i <= 0;
        end if;

       if ((dsp1_int4_in = '0' and dsp1_count_i < 32) or (dsp1_int4_in = '1' and dsp1_count_i > 0 and dsp1_count_i < 32)) then
            dsp1_send_data_n_i <= '0';
            dsp1_sbram_me_i <= '0';
            dsp1_count_i <= dsp1_count_i + 1;
        elsif (dsp1_int4_in = '1' and dsp1_count_i = 32) then
            dsp1_send_data_n_i <= '1';
            dsp1_sbram_me_i <= '1';
            dsp1_count_i <= 0;
        end if;

       if ((dsp2_int4_in = '0' and dsp2_count_i < 32) or (dsp2_int4_in = '1' and dsp2_count_i > 0 and dsp2_count_i < 32)) then
            dsp2_send_data_n_i <= '0';
            dsp2_sbram_me_i <= '0';
            dsp2_count_i <= dsp2_count_i + 1;
        elsif (dsp2_int4_in = '1' and dsp2_count_i = 32) then
            dsp2_send_data_n_i <= '1';
            dsp2_sbram_me_i <= '1';
            dsp2_count_i <= 0;
        end if;

       if ((dsp3_int4_in = '0' and dsp3_count_i < 32) or (dsp3_int4_in = '1' and dsp3_count_i > 0 and dsp3_count_i < 32)) then
            dsp3_send_data_n_i <= '0';
            dsp3_sbram_me_i <= '0';
            dsp3_count_i <= dsp3_count_i + 1;
        elsif (dsp3_int4_in = '1' and dsp3_count_i = 32) then
            dsp3_send_data_n_i <= '1';
            dsp3_sbram_me_i <= '1';
            dsp3_count_i <= 0;
        end if;
    end if;
end process dsp_send_data_n;

end rtl;

