----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    efb_input_source.vhd
-- Description:
--    ROD EFB input simulator
--       SCT Version
--
-- Reads a test data file and produces a data stream to help simulate
-- the EFB output into the Router. 
--
-- The code works by first reading in the data file and storing the
-- entire string of bits into a large array of characters. As the
-- data file is being read, whitespace and comments are filtered out
-- before being written into the array.
--
----------------------------------------------------------------------------

library IEEE ;
use IEEE.STD_LOGIC_1164.all ;
use IEEE.std_logic_arith.all;

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

entity efb_input_source is
port (
    clk40_in : in  STD_LOGIC ; -- Distributed 40 Mhz clock 
    rst_n_in : in  STD_LOGIC ; -- Powerup Reset

    -- Event Data Bus
    event_data_out1      : out  STD_LOGIC_VECTOR(39 downto 0);
    formatter_enabled_n1 : out  STD_LOGIC_VECTOR(3 downto 0);
    data_valid_out1      : out  STD_LOGIC;
    event_data_out2      : out  STD_LOGIC_VECTOR(39 downto 0);
    formatter_enabled_n2 : out  STD_LOGIC_VECTOR(3 downto 0);
    data_valid_out2      : out  STD_LOGIC
    ) ;
end  efb_input_source;

architecture rtl of efb_input_source is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  signal tsf_ctr: integer; -- index into data bit stream character array
  signal rst_n : std_logic; -- internal version of powerup reset signal
  signal clk40 : std_logic; -- internal version of clock signal

  signal data_output: STD_LOGIC_VECTOR(89 downto 0); -- efb output signals

  constant max_file_length : integer := 60000000; -- max file size.
     -- Change here for bigger files (keep in mind it eats up memory)
 
----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------
  
BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

    clk40 <= clk40_in;
    rst_n <= rst_n_in;

  -- since you can't read an output signal, I used internal versions
  -- that get reassigned here before going to the output.
    event_data_out1 <= data_output(39 downto 0);
    formatter_enabled_n1 <= data_output(43 downto 40);
    data_valid_out1 <= data_output(44);
    event_data_out2 <= data_output(84 downto 45);
    formatter_enabled_n2 <= data_output(88 downto 85);
    data_valid_out2 <= data_output(89);

------------------------------------------------------------------------------
-- Test Stand: File to Bitstream  
------------------------------------------------------------------------------
    
  eventdata: process  (clk40, rst_n)
    type load_file_type is file of character;
    -- data file path/filename
  
  file load_file: load_file_type is "/home/edafiles/atlas/rod/fpga_code/router/simulation_input_files/efb_input.rod";

    type storage_array is array(0 to max_file_length) of character;
    -- character array containing the data bitstream
    variable storage: storage_array;
    variable index: integer;
    -- location of last character
    variable max_tsf_ctr: integer;
    variable offset: integer;
    variable tsf_ctr: integer;
    variable i: integer;
    variable outpnibble : std_logic_vector (89 downto 0) := (others =>'0');
  begin
    if (rst_n = '0') then
      -- Main loop to read data file, filter comments/white space and
      -- write data to character array
      tsf_ctr := 0;
      index:=0;
      data_output <= (others => '0');
      while not endfile(load_file) loop
	read(load_file, storage(index));
	if (storage(index) /= lf and storage(index) /= ' ') then
          assert (index < max_file_length)
            report "input file is too long"
            severity failure;
          if (storage(index) = '#') then
            while (not endfile(load_file) and storage(index) /= lf) loop
              if (not endfile(load_file)) then
                 read(load_file, storage(index));
              end if;
            end loop;
            index:= index - 1;
          end if;  
	  index := index+1;
        else
          storage(index):='0';
	end if;
        max_tsf_ctr := index;
      end loop;
    elsif (clk40'event and clk40 = '1') then      -- Simple clocked loop that read character array and outputs 50
      -- bit data words. Any character not a '1' or '0' gets 'Z' value
        for i in 89 downto 0 loop                
	      if (storage(tsf_ctr) = '0') then
                outpnibble(i) := '0';
            elsif (storage(tsf_ctr) = '1') then
                outpnibble(i) := '1';
            else       
                outpnibble(i) := 'Z';
            end if;
            data_output <= outpnibble after 5 ns;
            -- Stops when the last character is read in and outputs all
            -- zeroes after that. The last word in the data file should
            -- contain dummy data of all zeroes to make sure it doesn't cut off
            --  or lose the last few data bits in the file.
            if (tsf_ctr < max_tsf_ctr-1) then
                tsf_ctr := tsf_ctr + 1;
            else
            data_output <= (others => '0') after 5 ns;
            end if;
        end loop;
    end if;
end process;

END rtl;

