----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    efb_source.vhd
-- Description:
--    ROD EFB output simulator
--       SCT Version
--
-- Reads a test data file and produces a data stream to help simulate
-- the EFB output into the Router. 
--
-- The code works by first reading in the data file and storing the
-- entire string of bits into a large array of characters. As the
-- data file is being read, whitespace and comments are filtered out
-- before being written into the array.
--
----------------------------------------------------------------------------

library IEEE ;
use IEEE.std_logic_1164.all ;
use IEEE.std_logic_arith.all;

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

entity router_bus_source is
port (
    clk_in  : in  std_logic ; -- Distributed 40 Mhz clock 
    rst_n_in  : in  std_logic; -- Powerup Reset
	
	-- Router Bus signals
	router_bus_data_out	  : out  std_logic_vector(15 downto 0) ;
	router_bus_addr_out	  : out  std_logic_vector(7 downto 0) ;
	router_bus_strb_n_out : out  std_logic;
	router_bus_rnw_out	  : out  std_logic;
    router_bus_ack_in	  : in 	 std_logic;
    router_bus_done_out   : out  std_logic
	) ;
end  router_bus_source;

architecture rtl of router_bus_source is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  signal tsf_ctr : integer; -- index into data bit stream character array
  signal rst : std_logic:= '0'; -- internal version of powerup reset signal
  signal clk : std_logic:= '0'; -- internal version of clock signal
  signal data_output : std_logic_vector(26 downto 0); -- internal
     -- version of efb output signals
  constant max_file_length : integer := 60000000; -- max file size.
     -- Change here for bigger files (keep in mind it eats up memory)
 
----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------
  
BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

  -- since you can't read an output signal, I used internal versions
  -- that get reassigned here before going to the output.
  clk <= clk_in;
  rst <= rst_n_in;
  router_bus_data_out <= data_output(15 downto 0);
  router_bus_addr_out <= data_output(23 downto 16);
  router_bus_strb_n_out <= data_output(24);
  router_bus_rnw_out <= data_output(25);
  router_bus_done_out <= data_output(26);

------------------------------------------------------------------------------
-- Test Stand: File to Bitstream  
------------------------------------------------------------------------------

  eventdata: process  (clk, rst)
    type load_file_type is file of character;
    -- data file path/filename
  
  file load_file: load_file_type is "/mnt/edafiles/atlas/rod/fpga_code/router/simulation_input_files/router_bus_out.rod";

    type storage_array is array(0 to max_file_length) of character;
    -- character array containing the data bitstream
    variable storage: storage_array;
    variable index: integer;
    -- location of last character
    variable max_tsf_ctr: integer;
    variable offset: integer;
    variable tsf_ctr: integer;
    variable i: integer;
    variable outpnibble : std_logic_vector (26 downto 0) := (others =>'0');
  begin
    if (rst = '0') then
      -- Main loop to read data file, filter comments/white space and
      -- write data to character array
      tsf_ctr := 0;
      index:=0;
      data_output <= (others => '0');
      while not endfile(load_file) loop
	    read(load_file, storage(index));
	    if (storage(index) /= lf and storage(index) /= ' ') then
          assert (index < max_file_length)
            report "input file is too long"
            severity failure;
          if (storage(index) = '#') then
            while (not endfile(load_file) and storage(index) /= lf) loop
              if (not endfile(load_file)) then
                 read(load_file, storage(index));
              end if;
            end loop;
            index:= index - 1;
          end if;  
	  index := index+1;
        else
          storage(index):='0';
	end if;
        max_tsf_ctr := index;
      end loop;
    elsif (clk'event and clk = '1') then      -- Simple clocked loop that read character array and outputs 96
      -- bit data words. Any character not a '1' or '0' gets 'Z' value
      for i in 26 downto 0 loop                
	if (storage(tsf_ctr) = '0') then
          outpnibble(i) := '0';
        elsif (storage(tsf_ctr) = '1') then
          outpnibble(i) := '1';
        else       
          outpnibble(i) := 'Z';
        end if;
      data_output <= outpnibble after 5 ns;
      -- Stops when the last character is read in and outputs all
      -- zeroes after that. The last word in the data file should
      -- contain dummy data of all zeroes to make sure it doesn't cut off
      --  or lose the last few data bits in the file.
      if (tsf_ctr < max_tsf_ctr-1) then
        tsf_ctr := tsf_ctr + 1;
      else
        data_output <= (others => '1') after 5 ns;
      end if;
      end loop;
    end if;
end process;

END rtl;

