----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    efb_source.vhd
-- Description:
--    ROD EFB output simulator
--       SCT Version
--
-- Reads a test data file and produces a data stream to help simulate
-- the EFB output into the Router. 
--
-- The code works by first reading in the data file and storing the
-- entire string of bits into a large array of characters. As the
-- data file is being read, whitespace and comments are filtered out
-- before being written into the array.
--
----------------------------------------------------------------------------
library IEEE ;
use IEEE.STD_LOGIC_1164.all ;
use IEEE.std_logic_arith.all;
----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

entity rtr_trap_source is
port (
    clk40_in : in  STD_LOGIC ; -- Distributed 40 Mhz clock 
    rst_in   : in  STD_LOGIC ; -- Powerup Reset
	
  -- Event Data Bus
    event_data_out  : out  STD_LOGIC_VECTOR(31 downto 0);
    data_valid_out  : out  STD_LOGIC;
    tristate_n_out  : out  STD_LOGIC_VECTOR(2 downto 0);
  -- tristate_n(0) -> efb_a output
  -- tristate_n(1) -> efb_b output
  -- tristate_n(2) -> header/trailer output

    hw_capture_out  : out  STD_LOGIC;
    mask_dfr_out    : out  STD_LOGIC;
    trap_enable_out : out  STD_LOGIC;
    trap_reset_out  : out  STD_LOGIC;
    trap0_load_out  : out  STD_LOGIC;
    trap1_load_out  : out  STD_LOGIC;

    rod_tt_out      : out std_logic_vector(7 downto 0);
    tim_tt_out      : out std_logic_vector(1 downto 0);
    atlas_tt_out    : out std_logic_vector(7 downto 0);

    frame_xfer_size : out std_logic_vector(9 downto 0);
--    int_mask_delay  : out std_logic_vector(15 downto 0);
    int_mask_delay  : out std_logic_vector(5 downto 0);
    prescale0       : out std_logic_vector(15 downto 0);
    prescale1       : out std_logic_vector(15 downto 0);
    trap_match0     : out std_logic_vector(7 downto 0);
    trap_match1     : out std_logic_vector(7 downto 0);
    cmd_reg0        : out std_logic_vector(6 downto 0);
    cmd_reg1        : out std_logic_vector(4 downto 0)

    rden_out        : out std_logic;
    ) ;
end  rtr_trap_source;

architecture rtl of rtr_trap_source is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  signal tsf_ctr: integer; -- index into data bit stream character array
  signal data_output : STD_LOGIC_VECTOR(49 downto 0); -- internal
  signal count_i : unsigned(9 downto 0);
  -- version of efb output signals
  constant max_file_length : integer := 60000000; -- max file size.
  -- Change here for bigger files (keep in mind it eats up memory)
 
----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------
  
BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

  -- since you can't read an output signal, I used internal versions
  -- that get reassigned here before going to the output.
  --event_data_out <= data_output(45 downto 0);
  event_data_out <= data_output(31 downto 0);
  tristate_n_out <= data_output(48 downto 46);
  data_valid_out <= data_output(49);

------------------------------------------------------------------------------
-- Test Stand: File to Bitstream  
------------------------------------------------------------------------------

  eventdata: process  (clk40_in, rst_in)
    type load_file_type is file of character;
    -- data file path/filename
  
  file load_file: load_file_type is "/mnt/edafiles/atlas/rod/fpga_code/router/simulation_input_files/efb_out.rod";

    type storage_array is array(0 to max_file_length) of character;
    -- character array containing the data bitstream
    variable storage: storage_array;
    variable index: integer;
    -- location of last character
    variable max_tsf_ctr: integer;
    variable offset: integer;
    variable tsf_ctr: integer;
    variable i: integer;
    variable outpnibble : std_logic_vector (49 downto 0) := (others =>'0');
  begin
    if (rst_in = '0') then
      -- Main loop to read data file, filter comments/white space and
      -- write data to character array
      tsf_ctr := 0;
      index:=0;
      data_output <= (others => '0');
      while not endfile(load_file) loop
	read(load_file, storage(index));
	if (storage(index) /= lf and storage(index) /= ' ') then
          assert (index < max_file_length)
            report "input file is too long"
            severity failure;
          if (storage(index) = '#') then
            while (not endfile(load_file) and storage(index) /= lf) loop
              if (not endfile(load_file)) then
                 read(load_file, storage(index));
              end if;
            end loop;
            index:= index - 1;
          end if;  
	  index := index+1;
        else
          storage(index):='0';
	end if;
        max_tsf_ctr := index;
      end loop;
    elsif (clk40_in'event and clk40_in = '1') then      -- Simple clocked loop that read character array and outputs 50
      -- bit data words. Any character not a '1' or '0' gets 'Z' value
      for i in 49 downto 0 loop                
        if (storage(tsf_ctr) = '0') then
          outpnibble(i) := '0';
        elsif (storage(tsf_ctr) = '1') then
          outpnibble(i) := '1';
        else       
          outpnibble(i) := 'Z';
        end if;
        data_output <= outpnibble after 5 ns;
   -- Stops when the last character is read in and outputs all
   -- zeroes after that. The last word in the data file should
   -- contain dummy data of all zeroes to make sure it doesn't cut off
   --  or lose the last few data bits in the file.
        if (tsf_ctr < max_tsf_ctr-1) then
          tsf_ctr := tsf_ctr + 1;
        else
          data_output(48 downto 0) <= (others => '1');
          data_output(49) <= '0';
        end if;
      end loop;
    end if;
end process;

control_data: process  (clk40_in, rst_in)
begin
  if (rst_in = '0') then
    hw_capture_out  <= '0'; 
    mask_dfr_out    <= '0'; 
    trap_enable_out <= '0'; 
    trap_reset_out  <= '1'; 
    trap0_load_out  <= '0'; 
    trap1_load_out  <= '0'; 
    rod_tt_out      <= (others => '0'); 
    tim_tt_out      <= (others => '0'); 
    atlas_tt_out    <= (others => '0'); 
    frame_xfer_size <= "0100000000"; 
    int_mask_delay  <= (others => '0'); 
    prescale0       <= (others => '0'); 
    prescale1       <= (others => '0'); 
    trap_match0     <= (others => '0'); 
    trap_match1     <= (others => '0'); 
    cmd_reg0        <= (others => '0'); 
    cmd_reg1        <= (others => '0');
    count_i         <= (others => '0');
  elsif (clk40_in'event and clk40_in = '1') then
    if count_i < 1023 then
       hw_capture_out  <= '1'; 
       mask_dfr_out    <= '0'; 
       trap_enable_out <= '1'; 
       trap_reset_out  <= '0'; 
       trap0_load_out  <= '0'; 
       trap1_load_out  <= '0'; 
       rod_tt_out      <= "00000000"; 
       tim_tt_out      <= "00"; 
       atlas_tt_out    <= "00000000"; 
       frame_xfer_size <= "0100000000"; 
       int_mask_delay  <= "100000"; 
       prescale0       <= "0000000000000001"; 
       prescale1       <= "0000000000000001"; 
       trap_match0     <= "00000000"; 
       trap_match1     <= "00000001"; 
       cmd_reg0        <= "0001000"; 
       cmd_reg1        <= "00000";
       count_i <= count_i + 1;
    else
      mask_dfr_out    <= '1';
      count_i <= (others => '0');
    end if;
    
  end if;
end process;


END rtl;



