--------------------------------------------------------------------------
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 1999
-- ReadOutDriver Electronics
--
--------------------------------------------------------------------------
-- Filename: register_block.vhd
-- Description:
--      Register file for Router register bus. Examine the comments
--	    specifically to discover the exact register definition.
--
--------------------------------------------------------------------------
-- Structure:
--------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk40
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
--------------------------------------------------------------------------
-- Author: Johm M. Joseph
-- Board Engineer: John M. Joseph
-- History:
--          15 March 2000 JMJ Adapted control registers from 
--          ROD resources: register_block.vhd 
--          
--          07 March 2001 JMJ Added new registers to allow more function
--          16 April 2002 JMJ Revised per new DSP Algorithms
--
--------------------------------------------------------------------------
-- Notes:
--------------------------------------------------------------------------
--                 Router Register Map
--  Address
-- 00000000 (RW) = DSP 0 Command Register0
--       Bit 0 : Trap Atlas Specific Event Type
--       Bit 1 : Trap TIM Specific Event Type
--       Bit 2 : Trap ROD Specific Event Type
--       Bit 3 : Trap Events in Error Format
--       Bit 4 : Trap Events in SLink Format with Error Codes in the Trailer
--       Bit 5 : Trap All Events using S-Link Data
--       Bit 6 : Set Trap into Data Driven Mode
--
-- 00000001 (RW) = DSP 0 Command Register1
--       Bit 0 : Trap Atlas Specific Event Type
--       Bit 1 : Trap TIM Specific Event Type
--       Bit 2 : Trap ROD Specific Event Type
--       Bit 3 : Trap Error Events
--       Bit 4 : Exclusion Flag (Trap all except the Type Match)
--
-- 00000010 (RW) = DSP 0 Load/Reset Trap Register
--       Bit 0 : Reset Trap
--       Bit 1 : Load New Trap0 Value
--       Bit 2 : Load New Trap1 Value
--       Bit 3 : Flush Trap FIFO
--
-- 00000011 (RW) = DSP 0 Status Register
--       Bit 0 : Trap Enable Control from DSP
--       Bit 1 : Interrupt Mask Control from DSP
--       Bit 2 : Trapping FIFO Empty
--       Bit 3 : Trapping FIFO Full
--       Bit 4 : Trap0 - Trapping Data
--       Bit 5 : Trap1 - Trapping Data
--       Bit 6 : Trap0 Ready (Clears on WR to Cmd Reg0, Set on Load New Trap0)
--       Bit 7 : Trap1 Ready (Clears on WR to Cmd Reg1, Set on Load New Trap1)
--
-- 00000100 (RW) = DSP 0 Trap Match Register0
-- 00000101 (RW) = DSP 0 PreScale Value0
-- 00000110 (RW) = DSP 0 Trap Match Register1
-- 00000111 (RW) = DSP 0 PreScale Value1
-- 00001000 (RW) = DSP 0 Data Frame Transfer Size Register
-- 00001001 (R ) = DSP 0 FIFO Wr Pointer Register
-- 00001010 (RW) = DSP 0 FIFO Rd Pointer Register
-- 00001100 (RW) = DSP 0 Interrupt Mask Delay Count (5 downto 0)
-- 00001101 (RW) = DSP 0 Interrupt Period (Debug Register)
--
-- 00010000 (RW) = DSP 1 Command Register0
-- 00010001 (RW) = DSP 1 Command Register1
-- 00010010 (RW) = DSP 1 Load/Reset Trap Register
-- 00010011 (RW) = DSP 1 Status Register
-- 00010100 (RW) = DSP 1 Trap Match Register0
-- 00010101 (RW) = DSP 1 PreScale Value0
-- 00010110 (RW) = DSP 1 Trap Match Register1
-- 00010111 (RW) = DSP 1 PreScale Value1
-- 00011000 (RW) = Not Used --DSP 1 Data Frame Transfer Size Register
-- 00011001 (R ) = DSP 1 FIFO Word Count Register
-- 00011010 (RW) = DSP 1 FIFO Rd Pointer Register
-- 00011100 (RW) = DSP 1 Interrupt Mask Delay Count
-- 00011101 (RW) = DSP 1 Interrupt Period (Debug Register)
--
-- 00100000 (RW) = DSP 2 Command Register0
-- 00100001 (RW) = DSP 2 Command Register1
-- 00100010 (RW) = DSP 2 Load/Reset Trap Register
-- 00100011 (RW) = DSP 2 Status Register
-- 00100100 (RW) = DSP 2 Trap Match Register0
-- 00100101 (RW) = DSP 2 PreScale Value0
-- 00100110 (RW) = DSP 2 Trap Match Register1
-- 00100111 (RW) = DSP 2 PreScale Value1
-- 00101000 (RW) = Not Used --DSP 2 Data Frame Transfer Size Register
-- 00101001 (R ) = DSP 2 FIFO Word Count Register
-- 00101010 (RW) = DSP 2 FIFO Rd Pointer Register
-- 00101100 (RW) = DSP 2 Interrupt Mask Delay Count
-- 00101101 (RW) = DSP 2 Interrupt Period (Debug Register)
--
-- 00110000 (RW) = DSP 3 Command Register0
-- 00110001 (RW) = DSP 3 Command Register1
-- 00110010 (RW) = DSP 3 Load/Reset Trap Register
-- 00110011 (RW) = DSP 3 Status Register
-- 00110100 (RW) = DSP 3 Trap Match Register0
-- 00110101 (RW) = DSP 3 PreScale Value0
-- 00110110 (RW) = DSP 3 Trap Match Register1
-- 00110111 (RW) = DSP 3 PreScale Value1
-- 00111000 (RW) = Not Used --DSP 3 Data Frame Transfer Size Register
-- 00111001 (R ) = DSP 3 FIFO Word Count Register
-- 00111010 (RW) = DSP 3 FIFO Rd Pointer Register
-- 00111100 (RW) = DSP 3 Interrupt Mask Delay Count
-- 00111101 (RW) = DSP 3 Interrupt Period (Debug Register)
--
-- 01000000 (RW) = Router Command/Status Register
--       Bit  0: Dump Atlas Specific Event Type
--       Bit  1: Dump TIM Specific Event Type
--       Bit  2: Dump ROD Specific Event Type
--       Bit  3: Inhibit S-Link Write Enable
--       Bit  4: Reset S-Link
--       Bit  5: Set S-Link Test Bit
--       Bit  6: Enable Calibration Back Pressure to EFB
--       Bit  7: Mask S-Link Down
--       Bit  8: S-Link Xoff Status (Read Only)
--       Bit  9: S-Link Bad Link Status (Read Only)
--       Bit 10: Stop Output (Back pressure to the EFB)
--
-- 01000001 (RW) = S-Link Atlas/TIM Dump Match Register
--       Bits [7:0]: Atlas Specific Event Type
--       Bits [9:8]: TIM Specific Event Type
--
-- 01000010 (RW) = S-Link ROD Dump Match Register
-- 01000011 (R)  = Code Version
--
--  Here's a more precise list of what I'd love counted 
--  (anything from top to bottom of the list that can make it into a version, makes me happy)
--  CLK Counter (Already there)
--  Data Word Count (Already there)
--  Fragment count (Count up whenever a new fragment passes through, to have an inherent trigger rate count for normalising data)
--  Hit count (Count up on each hitword that passes through)
--  Error counts (According to flagged errors in table 31 of the user manual):
--- Timeout (Bit 2)
--- EOCOverflow (Bit 22) <-- This one here would be particularly interesting to me.
--- Afull/Truncation (Bit 16)
--- Row Column error (Bit 20)
--------------------------------------------------------------------------
-- LIBRARY INCLUDES
--------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations

--------------------------------------------------------------------------
-- PORT DECLARATION
--------------------------------------------------------------------------

entity register_block is
  port (
    clk40_in                     : in     std_logic; -- clk40 input
    rst_n_in                     : in     std_logic; -- async global reset
    router_bus_strb_n_in         : in     std_logic;
    router_bus_rnw_in	           : in     std_logic;
    router_bus_adr_in            : in     std_logic_vector( 7 downto 0);
    router_bus_data_in           : in     std_logic_vector(15 downto 0);
    router_bus_data_out          : out    std_logic_vector(15 downto 0);
    router_bus_ack_out           : out    std_logic;
    global_cmd_reg_out           : out    std_logic_vector( 7 downto 0);
    slink_status_reg_in          : in     std_logic_vector( 2 downto 0);
    slink_AT_dump_match_reg_out  : out    std_logic_vector( 9 downto 0);
    slink_ROD_dump_match_reg_out : out    std_logic_vector( 5 downto 0);
    dsp0_cmd_reg0_out            : buffer std_logic_vector( 8 downto 0);
    dsp0_cmd_reg1_out            : buffer std_logic_vector( 4 downto 0);
    dsp0_fifo_flush_out          : out    std_logic;
    dsp0_trap_reset_out          : out    std_logic;
    dsp0_trap0_load_out          : out    std_logic;
    dsp0_trap1_load_out          : out    std_logic;
    dsp0_status_reg_in           : in     std_logic_vector( 5 downto 0);
    dsp0_trap_match0_out         : buffer std_logic_vector( 7 downto 0); 
    dsp0_prescale0_out           : buffer std_logic_vector(15 downto 0); 
    dsp0_trap_match1_out         : buffer std_logic_vector( 7 downto 0); 
    dsp0_prescale1_out           : buffer std_logic_vector(15 downto 0); 
    dsp0_fifo_word_count_reg_in  : in     std_logic_vector(12 downto 0);
    dsp0_int_mask_delay_out      : out    std_logic_vector( 5 downto 0); 
    dsp1_cmd_reg0_out            : buffer std_logic_vector( 8 downto 0);
    dsp1_cmd_reg1_out            : buffer std_logic_vector( 4 downto 0);
    dsp1_fifo_flush_out          : out    std_logic;
    dsp1_trap_reset_out          : out    std_logic;
    dsp1_trap0_load_out          : out    std_logic;
    dsp1_trap1_load_out          : out    std_logic;
    dsp1_status_reg_in           : in     std_logic_vector( 5 downto 0);
    dsp1_trap_match0_out         : buffer std_logic_vector( 7 downto 0); 
    dsp1_prescale0_out           : buffer std_logic_vector(15 downto 0); 
    dsp1_trap_match1_out         : buffer std_logic_vector( 7 downto 0); 
    dsp1_prescale1_out           : buffer std_logic_vector(15 downto 0); 
    dsp1_fifo_word_count_reg_in  : in     std_logic_vector(12 downto 0);
    dsp1_int_mask_delay_out      : out    std_logic_vector( 5 downto 0); 
    dsp2_cmd_reg0_out            : buffer std_logic_vector( 8 downto 0);
    dsp2_cmd_reg1_out            : buffer std_logic_vector( 4 downto 0);
    dsp2_fifo_flush_out          : out    std_logic;
    dsp2_trap_reset_out          : out    std_logic;
    dsp2_trap0_load_out          : out    std_logic;
    dsp2_trap1_load_out          : out    std_logic;
    dsp2_status_reg_in           : in     std_logic_vector( 5 downto 0);
    dsp2_trap_match0_out         : buffer std_logic_vector( 7 downto 0); 
    dsp2_prescale0_out           : buffer std_logic_vector(15 downto 0); 
    dsp2_trap_match1_out         : buffer std_logic_vector( 7 downto 0); 
    dsp2_prescale1_out           : buffer std_logic_vector(15 downto 0); 
    dsp2_fifo_word_count_reg_in  : in     std_logic_vector(12 downto 0);
    dsp2_int_mask_delay_out      : out    std_logic_vector( 5 downto 0); 
    dsp3_cmd_reg0_out            : buffer std_logic_vector( 8 downto 0);
    dsp3_cmd_reg1_out            : buffer std_logic_vector( 4 downto 0);
    dsp3_fifo_flush_out          : out    std_logic;
    dsp3_trap_reset_out          : out    std_logic;
    dsp3_trap0_load_out          : out    std_logic;
    dsp3_trap1_load_out          : out    std_logic;
    dsp3_status_reg_in           : in     std_logic_vector( 5 downto 0);
    dsp3_trap_match0_out         : buffer std_logic_vector( 7 downto 0); 
    dsp3_prescale0_out           : buffer std_logic_vector(15 downto 0); 
    dsp3_trap_match1_out         : buffer std_logic_vector( 7 downto 0); 
    dsp3_prescale1_out           : buffer std_logic_vector(15 downto 0); 
    dsp3_fifo_word_count_reg_in  : in     std_logic_vector(12 downto 0);
    dsp3_int_mask_delay_out      : out    std_logic_vector( 5 downto 0);
    dll_status_in                : in     std_logic_vector( 1 downto 0);
    rtr_dsp_spares_in            : in     std_logic_vector(13 downto 0);
    rod_type_in                  : in     std_logic;
    test_mux_out                 : out    std_logic_vector( 3 downto 0);
    dsp0_fifo_rd_pointer_in      : in     std_logic_vector( 9 downto 0);
    dsp1_fifo_rd_pointer_in      : in     std_logic_vector( 9 downto 0);
    dsp2_fifo_rd_pointer_in      : in     std_logic_vector( 9 downto 0);
    dsp3_fifo_rd_pointer_in      : in     std_logic_vector( 9 downto 0);
    dsp0_dfrdy_stat_in           : in     std_logic_vector( 3 downto 0);
    dsp1_dfrdy_stat_in           : in     std_logic_vector( 3 downto 0);
    dsp2_dfrdy_stat_in           : in     std_logic_vector( 3 downto 0);
    dsp3_dfrdy_stat_in           : in     std_logic_vector( 3 downto 0);
    clr_sl_reset_in              : in     std_logic;
    dsp0_int4_out                : out    std_logic;
    dsp1_int4_out                : out    std_logic;
    dsp2_int4_out                : out    std_logic;
    dsp3_int4_out                : out    std_logic;
    mask_fifo_int4_out           : out    std_logic_vector( 3 downto 0);
    endian_select_out            : out    std_logic;
    force_xoff_out               : out    std_logic;
    sl_fifo_stat_in              : in     std_logic_vector(2 downto 0);
    hold_next_delay_out          : out    std_logic_vector(15 downto 0);
    slink_data_valid_in          : in     std_logic;
    fragment_count_in            : in     std_logic_vector(31 downto 0);
    hit_count_in                 : in     std_logic_vector(31 downto 0);
    to_error_count_in            : in     std_logic_vector(31 downto 0);
    eoc_over_count_in            : in     std_logic_vector(31 downto 0);
    afull_count_in               : in     std_logic_vector(31 downto 0);
    rc_error_count_in            : in     std_logic_vector(31 downto 0);
    sync_err_count_in            : in     std_logic_vector(31 downto 0);
    bcid_chkfail_count_in        : in     std_logic_vector(31 downto 0);
    l1_chkfail_count_in          : in     std_logic_vector(31 downto 0);
    hitoverflow_count_in         : in     std_logic_vector(31 downto 0);
    regparity_count_in           : in     std_logic_vector(31 downto 0);
    bitflip_count_in             : in     std_logic_vector(31 downto 0);
    hammingcode_count_in         : in     std_logic_vector(31 downto 0);
    total_error_count_in         : in     std_logic_vector(31 downto 0);
    eot_count_in                 : in     std_logic_vector(31 downto 0);
    status_counter_crtl_out      : out    std_logic_vector( 1 downto 0)
    );   
end register_block;


architecture rtl of register_block is
--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------

type states is (
  idle, 
  acknowledge
  );
signal acknowledge_cycle_state : states;

signal dsp0_cmd_reg0_i    : std_logic_vector( 8 downto 0);
signal dsp0_cmd_reg1_i    : std_logic_vector( 4 downto 0);
signal dsp0_fifo_flush_i  : std_logic;
signal dsp0_trap_reset_i  : std_logic;
signal dsp0_trap0_load_i  : std_logic;
signal dsp0_trap1_load_i  : std_logic;
signal dsp0_trap_match0_i : std_logic_vector( 7 downto 0); 
signal dsp0_prescale0_i   : std_logic_vector(15 downto 0); 
signal dsp0_trap_match1_i : std_logic_vector( 7 downto 0); 
signal dsp0_prescale1_i   : std_logic_vector(15 downto 0); 
signal dsp0_int_dly_cnt_i : std_logic_vector( 5 downto 0); 
signal dsp0_int_period_i  : std_logic_vector(15 downto 0); 
signal dsp1_cmd_reg0_i    : std_logic_vector( 8 downto 0);
signal dsp1_cmd_reg1_i    : std_logic_vector( 4 downto 0);
signal dsp1_fifo_flush_i  : std_logic;
signal dsp1_trap_reset_i  : std_logic;
signal dsp1_trap0_load_i  : std_logic;
signal dsp1_trap1_load_i  : std_logic;
signal dsp1_trap_match0_i : std_logic_vector( 7 downto 0); 
signal dsp1_prescale0_i   : std_logic_vector(15 downto 0); 
signal dsp1_trap_match1_i : std_logic_vector( 7 downto 0); 
signal dsp1_prescale1_i   : std_logic_vector(15 downto 0); 
signal dsp1_int_dly_cnt_i : std_logic_vector( 5 downto 0); 
signal dsp1_int_period_i  : std_logic_vector(15 downto 0); 
signal dsp2_cmd_reg0_i    : std_logic_vector( 8 downto 0);
signal dsp2_cmd_reg1_i    : std_logic_vector( 4 downto 0);
signal dsp2_fifo_flush_i  : std_logic;
signal dsp2_trap_reset_i  : std_logic;
signal dsp2_trap0_load_i  : std_logic;
signal dsp2_trap1_load_i  : std_logic;
signal dsp2_trap_match0_i : std_logic_vector( 7 downto 0); 
signal dsp2_prescale0_i   : std_logic_vector(15 downto 0); 
signal dsp2_trap_match1_i : std_logic_vector( 7 downto 0); 
signal dsp2_prescale1_i   : std_logic_vector(15 downto 0); 
signal dsp2_int_dly_cnt_i : std_logic_vector( 5 downto 0); 
signal dsp2_int_period_i  : std_logic_vector(15 downto 0); 
signal dsp3_cmd_reg0_i    : std_logic_vector( 8 downto 0);
signal dsp3_cmd_reg1_i    : std_logic_vector( 4 downto 0);
signal dsp3_fifo_flush_i  : std_logic;
signal dsp3_trap_reset_i  : std_logic;
signal dsp3_trap0_load_i  : std_logic;
signal dsp3_trap1_load_i  : std_logic;
signal dsp3_trap_match0_i : std_logic_vector( 7 downto 0); 
signal dsp3_prescale0_i   : std_logic_vector(15 downto 0); 
signal dsp3_trap_match1_i : std_logic_vector( 7 downto 0); 
signal dsp3_prescale1_i   : std_logic_vector(15 downto 0); 
signal dsp3_int_dly_cnt_i : std_logic_vector( 5 downto 0); 
signal dsp3_int_period_i  : std_logic_vector(15 downto 0); 

signal global_cmd_reg_i       : std_logic_vector( 7 downto 0); 
signal slink_AT_dump_match_i  : std_logic_vector( 9 downto 0);
signal slink_ROD_dump_match_i : std_logic_vector( 5 downto 0);
signal test_mux_i             : std_logic_vector( 3 downto 0);

signal dsp0_int_count : unsigned(17 downto 0);
signal dsp1_int_count : unsigned(17 downto 0);
signal dsp2_int_count : unsigned(17 downto 0);
signal dsp3_int_count : unsigned(17 downto 0);

signal xoff_count_i        : unsigned(15 downto 0);
signal clr_xoff_count_i    : std_logic;
signal clr_latched_xoff_i  : std_logic;
signal clr_xoff_count_ii   : std_logic;
signal clr_latched_xoff_ii : std_logic;
signal latched_xoff_i      : std_logic;
signal xoff_d1             : std_logic;
signal endian_select_i     : std_logic;
signal force_xoff_i        : std_logic;
signal rb_strb_n_i         : std_logic;
signal hold_next_delay_i   : std_logic_vector(15 downto 0);

type rb_access_states is (
  idle,
  write,
  read,
  done
  );
signal rb_access : rb_access_states;

signal dsp0_trap_fifo_full_i : std_logic;
signal dsp0_clear_status_i   : std_logic;
signal dsp1_trap_fifo_full_i : std_logic;
signal dsp1_clear_status_i   : std_logic;
signal dsp2_trap_fifo_full_i : std_logic;
signal dsp2_clear_status_i   : std_logic;
signal dsp3_trap_fifo_full_i : std_logic;
signal dsp3_clear_status_i   : std_logic;

signal slink_counter_control_i : std_logic_vector(1 downto 0);
signal slink_clk_count  : unsigned(31 downto 0);
signal slink_word_count : unsigned(31 downto 0);

begin
--------------------------------------------------------------------------
-- PROCESS DECLARATION
--------------------------------------------------------------------------

dsp0_trap_reset_out <= dsp0_trap_reset_i;
dsp1_trap_reset_out <= dsp1_trap_reset_i;
dsp2_trap_reset_out <= dsp2_trap_reset_i;
dsp3_trap_reset_out <= dsp3_trap_reset_i;

dsp0_fifo_flush_out <= dsp0_fifo_flush_i;
dsp1_fifo_flush_out <= dsp1_fifo_flush_i;
dsp2_fifo_flush_out <= dsp2_fifo_flush_i;
dsp3_fifo_flush_out <= dsp3_fifo_flush_i;

dsp0_trap0_load_out <= dsp0_trap0_load_i;
dsp0_trap1_load_out <= dsp0_trap1_load_i;
dsp1_trap0_load_out <= dsp1_trap0_load_i;
dsp1_trap1_load_out <= dsp1_trap1_load_i;
dsp2_trap0_load_out <= dsp2_trap0_load_i;
dsp2_trap1_load_out <= dsp2_trap1_load_i;
dsp3_trap0_load_out <= dsp3_trap0_load_i;
dsp3_trap1_load_out <= dsp3_trap1_load_i;

hold_next_delay_out <= hold_next_delay_i;
status_counter_crtl_out <= slink_counter_control_i;
force_xoff_out <= force_xoff_i;

--------------------------------------------------------------------------
-- Process to Enable Read/Write to Router Registers
--------------------------------------------------------------------------
rw_registers : process (
  clk40_in, 
  rst_n_in,
  router_bus_data_in, 
  router_bus_adr_in,
  router_bus_rnw_in, 
  router_bus_strb_n_in,
  slink_status_reg_in,
  dsp0_fifo_word_count_reg_in,
  dsp1_fifo_word_count_reg_in,
  dsp2_fifo_word_count_reg_in,
  dsp3_fifo_word_count_reg_in,
  dsp0_status_reg_in, 
  dsp1_status_reg_in,
  dsp2_status_reg_in, 
  dsp3_status_reg_in
  )
subtype select_type is std_logic_vector(7 downto 0);
variable dsp0_trap0_ready : std_logic;
variable dsp0_trap1_ready : std_logic;
variable dsp1_trap0_ready : std_logic;
variable dsp1_trap1_ready : std_logic;
variable dsp2_trap0_ready : std_logic;
variable dsp2_trap1_ready : std_logic;
variable dsp3_trap0_ready : std_logic;
variable dsp3_trap1_ready : std_logic;
constant code_version     : std_logic_vector(7 downto 0) := X"61"; --v6.1
constant board_rev        : std_logic_vector(6 downto 0) := "0001111";  --RevF
begin
  if (rst_n_in = '0') then
    dsp0_cmd_reg0_i        <= (others => '0');
    dsp0_cmd_reg1_i        <= (others => '0');
    dsp0_fifo_flush_i      <= '0';
    dsp0_trap_reset_i      <= '0';
    dsp0_trap0_load_i      <= '1';
    dsp0_trap1_load_i      <= '0';
    dsp0_trap_match0_i     <= (others => '0');
    dsp0_prescale0_i       <= (others => '0');
    dsp0_trap_match1_i     <= (others => '0');
    dsp0_prescale1_i       <= (others => '0');
    dsp0_int_dly_cnt_i     <= (others => '0');
    dsp0_int_period_i      <= (others => '0');
    dsp1_cmd_reg0_i        <= (others => '0');
    dsp1_cmd_reg1_i        <= (others => '0');
    dsp1_fifo_flush_i      <= '0';
    dsp1_trap_reset_i      <= '0';
    dsp1_trap0_load_i      <= '1';
    dsp1_trap1_load_i      <= '0';
    dsp1_trap_match0_i     <= (others => '0');
    dsp1_prescale0_i       <= (others => '0');
    dsp1_trap_match1_i     <= (others => '0');
    dsp1_prescale1_i       <= (others => '0');
    dsp1_int_dly_cnt_i     <= (others => '0');
    dsp1_int_period_i      <= (others => '0');
    dsp2_cmd_reg0_i        <= (others => '0');
    dsp2_cmd_reg1_i        <= (others => '0');
    dsp2_fifo_flush_i      <= '0';
    dsp2_trap_reset_i      <= '0';
    dsp2_trap0_load_i      <= '1';
    dsp2_trap1_load_i      <= '0';
    dsp2_trap_match0_i     <= (others => '0');
    dsp2_prescale0_i       <= (others => '0');
    dsp2_trap_match1_i     <= (others => '0');
    dsp2_prescale1_i       <= (others => '0');
    dsp2_int_dly_cnt_i     <= (others => '0');
    dsp2_int_period_i      <= (others => '0');
    dsp3_cmd_reg0_i        <= (others => '0');
    dsp3_cmd_reg1_i        <= (others => '0');
    dsp3_fifo_flush_i      <= '0';
    dsp3_trap_reset_i      <= '0';
    dsp3_trap0_load_i      <= '1';
    dsp3_trap1_load_i      <= '0';
    dsp3_trap_match0_i     <= (others => '0');
    dsp3_prescale0_i       <= (others => '0');
    dsp3_trap_match1_i     <= (others => '0');
    dsp3_prescale1_i       <= (others => '0');
    dsp3_int_dly_cnt_i     <= (others => '0');
    dsp3_int_period_i      <= (others => '0');
    global_cmd_reg_i       <= (others => '0');
    slink_AT_dump_match_i  <= (others => '0');
    slink_ROD_dump_match_i <= (others => '0');
    slink_counter_control_i <= (others => '0');
    test_mux_i             <= (others => '0');
    hold_next_delay_i      <= X"000A";  -- Default = 250ns
    clr_xoff_count_i       <= '0';
    clr_latched_xoff_i     <= '0';
    endian_select_i        <= '1';   -- default to LE
    force_xoff_i           <= '0';
    dsp0_trap_fifo_full_i  <= '0';
    dsp0_clear_status_i    <= '0';
    dsp1_trap_fifo_full_i  <= '0';
    dsp1_clear_status_i    <= '0';
    dsp2_trap_fifo_full_i  <= '0';
    dsp2_clear_status_i    <= '0';
    dsp3_trap_fifo_full_i  <= '0';
    dsp3_clear_status_i    <= '0';
    dsp0_trap0_ready       := '0';
    dsp0_trap1_ready       := '0';
    dsp1_trap0_ready       := '0';
    dsp1_trap1_ready       := '0';
    dsp2_trap0_ready       := '0';
    dsp2_trap1_ready       := '0';
    dsp3_trap0_ready       := '0';
    dsp3_trap1_ready       := '0';
    rb_strb_n_i            <= '0';
  elsif (clk40_in'event AND clk40_in = '1') then
    rb_strb_n_i <= router_bus_strb_n_in;
    case rb_access is
      when idle =>
        router_bus_data_out <= (others => '0');
        clr_xoff_count_i   <= '0';
        clr_latched_xoff_i <= '0';
        dsp0_trap_reset_i <= '0';
        dsp0_fifo_flush_i <= '0';
        dsp0_trap0_load_i <= '0';
        dsp0_trap1_load_i <= '0';
        dsp1_trap_reset_i <= '0';
        dsp1_fifo_flush_i <= '0';
        dsp1_trap0_load_i <= '0';
        dsp1_trap1_load_i <= '0';
        dsp2_trap_reset_i <= '0';
        dsp2_fifo_flush_i <= '0';
        dsp2_trap0_load_i <= '0';
        dsp2_trap1_load_i <= '0';
        dsp3_trap_reset_i <= '0';
        dsp3_fifo_flush_i <= '0';
        dsp3_trap0_load_i <= '0';
        dsp3_trap1_load_i <= '0';
        if (rb_strb_n_i ='1' AND router_bus_strb_n_in = '0') then
          if (router_bus_rnw_in = '1') then
            rb_access <= read;
          else
            rb_access <= write;
          end if;
        end if;

      when read =>
        case select_type(router_bus_adr_in) is
-- DSP0 REGS          
          when X"00" => router_bus_data_out( 8 downto  0) <= dsp0_cmd_reg0_out;
          when X"01" => router_bus_data_out( 4 downto  0) <= dsp0_cmd_reg1_out;
          when X"03" => router_bus_data_out( 2 downto  0) <= dsp0_status_reg_in(2 downto 0);
                        router_bus_data_out( 3)           <= dsp0_trap_fifo_full_i;
                        router_bus_data_out( 5 downto  4) <= dsp0_status_reg_in(5 downto 4);
                        router_bus_data_out( 6)           <= dsp0_trap0_ready;
                        router_bus_data_out( 7)           <= dsp0_trap1_ready;
                        router_bus_data_out( 8)           <= '1';
                        dsp0_clear_status_i <= '1';
          when X"04" => router_bus_data_out( 7 downto  0) <= dsp0_trap_match0_out;
          when X"05" => router_bus_data_out               <= dsp0_prescale0_out;
          when X"06" => router_bus_data_out( 7 downto  0) <= dsp0_trap_match1_out;
          when X"07" => router_bus_data_out               <= dsp0_prescale1_out;
          when X"09" => router_bus_data_out(15 downto 12) <= dsp0_dfrdy_stat_in; --dsp0_fifo_word_count_reg_in(12 downto 10);
                        router_bus_data_out( 9 downto  0) <= dsp0_fifo_word_count_reg_in( 9 downto  0);
--        when X"0A" => router_bus_data_out(15 downto 12) <= dsp0_dfrdy_stat_in;
--                      router_bus_data_out( 9 downto  0) <= dsp0_fifo_rd_pointer_in;
          when X"0C" => router_bus_data_out( 5 downto  0) <= dsp0_int_dly_cnt_i;
          when X"0D" => router_bus_data_out               <= dsp0_int_period_i;
-- DSP1 REGS          
          when X"10" => router_bus_data_out( 8 downto  0) <= dsp1_cmd_reg0_out;
          when X"11" => router_bus_data_out( 4 downto  0) <= dsp1_cmd_reg1_out;
          when X"13" => router_bus_data_out( 2 downto  0) <= dsp1_status_reg_in(2 downto 0);
                        router_bus_data_out( 3)           <= dsp1_trap_fifo_full_i;
                        router_bus_data_out( 5 downto  4) <= dsp1_status_reg_in(5 downto 4);
                        router_bus_data_out( 6)           <= dsp1_trap0_ready;
                        router_bus_data_out( 7)           <= dsp1_trap1_ready;
                        router_bus_data_out( 8)           <= '1';
                        dsp1_clear_status_i <= '1';
          when X"14" => router_bus_data_out( 7 downto  0) <= dsp1_trap_match0_out;
          when X"15" => router_bus_data_out               <= dsp1_prescale0_out;
          when X"16" => router_bus_data_out( 7 downto  0) <= dsp1_trap_match1_out;
          when X"17" => router_bus_data_out               <= dsp1_prescale1_out;
          when X"19" => router_bus_data_out(15 downto 12) <= dsp1_dfrdy_stat_in;  --dsp1_fifo_word_count_reg_in(12 downto 10);
                        router_bus_data_out( 9 downto  0) <= dsp1_fifo_word_count_reg_in( 9 downto  0);
--        when X"1A" => router_bus_data_out(15 downto 12) <= dsp1_dfrdy_stat_in;
--                      router_bus_data_out( 9 downto  0) <= dsp1_fifo_rd_pointer_in;
          when X"1C" => router_bus_data_out( 5 downto  0) <= dsp1_int_dly_cnt_i;
          when X"1D" => router_bus_data_out               <= dsp1_int_period_i;
-- DSP2 REGS          
          when X"20" => router_bus_data_out( 8 downto  0) <= dsp2_cmd_reg0_out;
          when X"21" => router_bus_data_out( 4 downto  0) <= dsp2_cmd_reg1_out;
          when X"23" => router_bus_data_out( 2 downto  0) <= dsp2_status_reg_in(2 downto 0);
                        router_bus_data_out( 3)           <= dsp2_trap_fifo_full_i;
                        router_bus_data_out( 5 downto  4) <= dsp2_status_reg_in(5 downto 4);
                        router_bus_data_out( 6)           <= dsp2_trap0_ready;
                        router_bus_data_out( 7)           <= dsp2_trap1_ready;
                        router_bus_data_out( 8)           <= '1';
                        dsp2_clear_status_i <= '1';
          when X"24" => router_bus_data_out( 7 downto  0) <= dsp2_trap_match0_out;
          when X"25" => router_bus_data_out               <= dsp2_prescale0_out;
          when X"26" => router_bus_data_out( 7 downto  0) <= dsp2_trap_match1_out;
          when X"27" => router_bus_data_out               <= dsp2_prescale1_out;
          when X"29" => router_bus_data_out(15 downto 12) <= dsp2_dfrdy_stat_in;  --dsp2_fifo_word_count_reg_in(12 downto 10);
                        router_bus_data_out( 9 downto  0) <= dsp2_fifo_word_count_reg_in( 9 downto  0);
--        when X"2A" => router_bus_data_out(15 downto 12) <= dsp2_dfrdy_stat_in;
--                      router_bus_data_out( 9 downto  0) <= dsp2_fifo_rd_pointer_in;
          when X"2C" => router_bus_data_out( 5 downto  0) <= dsp2_int_dly_cnt_i;
          when X"2D" => router_bus_data_out               <= dsp2_int_period_i;
-- DSP3 REGS          
          when X"30" => router_bus_data_out( 8 downto  0) <= dsp3_cmd_reg0_out;
          when X"31" => router_bus_data_out( 4 downto  0) <= dsp3_cmd_reg1_out;
          when X"33" => router_bus_data_out( 2 downto  0) <= dsp3_status_reg_in(2 downto 0);
                        router_bus_data_out( 3)           <= dsp3_trap_fifo_full_i;
                        router_bus_data_out( 5 downto  4) <= dsp3_status_reg_in(5 downto 4);
                        router_bus_data_out( 6)           <= dsp3_trap0_ready;
                        router_bus_data_out( 7)           <= dsp3_trap1_ready;
                        router_bus_data_out( 8)           <= '1';
                        dsp3_clear_status_i <= '1';
          when X"34" => router_bus_data_out( 7 downto  0) <= dsp3_trap_match0_out;
          when X"35" => router_bus_data_out               <= dsp3_prescale0_out;
          when X"36" => router_bus_data_out( 7 downto  0) <= dsp3_trap_match1_out;
          when X"37" => router_bus_data_out               <= dsp3_prescale1_out;
          when X"39" => router_bus_data_out(15 downto 12) <= dsp3_dfrdy_stat_in; --dsp3_fifo_word_count_reg_in(12 downto 10);
                        router_bus_data_out( 9 downto  0) <= dsp3_fifo_word_count_reg_in( 9 downto  0);
--        when X"3A" => router_bus_data_out(15 downto 12) <= dsp3_dfrdy_stat_in;
--                      router_bus_data_out( 9 downto  0) <= dsp3_fifo_rd_pointer_in;
          when X"3C" => router_bus_data_out( 5 downto  0) <= dsp3_int_dly_cnt_i;
          when X"3D" => router_bus_data_out               <= dsp3_int_period_i;
-- MAIN CONTROL/STATUS
          when X"40" => router_bus_data_out(15)           <= force_xoff_i;
                        router_bus_data_out(14)           <= endian_select_i;
                        router_bus_data_out(13)           <= latched_xoff_i;
                        router_bus_data_out(12 downto 11) <= dll_status_in;
                        router_bus_data_out(10 downto  8) <= slink_status_reg_in;
                        router_bus_data_out( 7 downto  0) <= global_cmd_reg_i;
                        clr_latched_xoff_i <= '1';
          when X"41" => router_bus_data_out( 9 downto  0) <= slink_AT_dump_match_i;
          when X"42" => router_bus_data_out( 5 downto  0) <= slink_ROD_dump_match_i;
          when X"43" => router_bus_data_out               <= rod_type_in &
                                                             board_rev   & 
                                                             code_version;
          when X"44" => router_bus_data_out( 5 downto  4) <= slink_counter_control_i;
                        router_bus_data_out( 3 downto  0) <= test_mux_i;
          when X"45" => router_bus_data_out               <= std_logic_vector(xoff_count_i);
                        clr_xoff_count_i                  <= '1';
          when X"46" => router_bus_data_out(15 downto 10) <= rtr_dsp_spares_in(13 downto 8);
                        router_bus_data_out( 2 downto  0) <= sl_fifo_stat_in;
          when X"47" => router_bus_data_out               <= hold_next_delay_i;
          when X"48" => router_bus_data_out               <= std_logic_vector(slink_clk_count(31 downto 16));
          when X"49" => router_bus_data_out               <= std_logic_vector(slink_clk_count(15 downto  0));
          when X"4A" => router_bus_data_out               <= std_logic_vector(slink_word_count(31 downto 16));
          when X"4B" => router_bus_data_out               <= std_logic_vector(slink_word_count(15 downto  0));
          when X"4C" => router_bus_data_out               <= fragment_count_in(31 downto 16);  -- Read Only (Fragment Count)
          when X"4D" => router_bus_data_out               <= fragment_count_in(15 downto  0);
          when X"4E" => router_bus_data_out               <= hit_count_in(31 downto 16);       -- Read Only (Hit Count)
          when X"4F" => router_bus_data_out               <= hit_count_in(15 downto  0);
          when X"50" => router_bus_data_out               <= to_error_count_in(31 downto 16);  -- Read Only (TimeOut Error Count - Bit 2)
          when X"51" => router_bus_data_out               <= to_error_count_in(15 downto  0);  
          when X"52" => router_bus_data_out               <= eoc_over_count_in(31 downto 16);  -- Read Only (EOC Overflow Error Count - Bit 22)
          when X"53" => router_bus_data_out               <= eoc_over_count_in(15 downto  0);
          when X"54" => router_bus_data_out               <= afull_count_in(31 downto 16);     -- Read Only (Afull Error - Bit 16)
          when X"55" => router_bus_data_out               <= afull_count_in(15 downto  0);
          when X"56" => router_bus_data_out               <= rc_error_count_in(31 downto 16);  -- Read Only (Row/Column Error - Bit 20)
          when X"57" => router_bus_data_out               <= rc_error_count_in(15 downto  0);
          when X"58" => router_bus_data_out               <= sync_err_count_in(31 downto 16);
          when X"59" => router_bus_data_out               <= sync_err_count_in(15 downto  0);
          when X"5A" => router_bus_data_out               <= bcid_chkfail_count_in(31 downto 16);
          when X"5B" => router_bus_data_out               <= bcid_chkfail_count_in(15 downto  0);
          when X"5C" => router_bus_data_out               <= l1_chkfail_count_in(31 downto 16);
          when X"5D" => router_bus_data_out               <= l1_chkfail_count_in(15 downto  0);
          when X"5E" => router_bus_data_out               <= hitoverflow_count_in(31 downto 16);
          when X"5F" => router_bus_data_out               <= hitoverflow_count_in(15 downto  0);
          when X"60" => router_bus_data_out               <= regparity_count_in(31 downto 16);
          when X"61" => router_bus_data_out               <= regparity_count_in(15 downto  0);
          when X"62" => router_bus_data_out               <= bitflip_count_in(31 downto 16);
          when X"63" => router_bus_data_out               <= bitflip_count_in(15 downto  0);
          when X"64" => router_bus_data_out               <= hammingcode_count_in(31 downto 16);
          when X"65" => router_bus_data_out               <= hammingcode_count_in(15 downto  0);
          when X"66" => router_bus_data_out               <= total_error_count_in(31 downto 16);
          when X"67" => router_bus_data_out               <= total_error_count_in(15 downto  0);
          when X"68" => router_bus_data_out               <= eot_count_in(31 downto 16);
          when X"69" => router_bus_data_out               <= eot_count_in(15 downto  0);
          
          when others => router_bus_data_out              <= (others => '0');
        end case;
        rb_access <= done;
--
      when write =>
        case select_type(router_bus_adr_in) is
-- DSP0 REGS          
          when X"00" => dsp0_cmd_reg0_i    <= router_bus_data_in(8 downto 0);
                        dsp0_trap0_ready   := '0';
          when X"01" => dsp0_cmd_reg1_i    <= router_bus_data_in(4 downto 0);
                        dsp0_trap1_ready   := '0';
          when X"02" => dsp0_trap_reset_i  <= router_bus_data_in(0);
                        dsp0_trap0_load_i  <= router_bus_data_in(1);
                        dsp0_trap1_load_i  <= router_bus_data_in(2);
                        dsp0_fifo_flush_i  <= router_bus_data_in(3);
                        if (router_bus_data_in(1) = '1') then
                          dsp0_trap0_ready := '1';
                        end if;
                        if (router_bus_data_in(2) = '1') then
                          dsp0_trap1_ready := '1';
                        end if;
                        if (router_bus_data_in(0) = '1') then
                          dsp0_trap0_ready := '0';
                          dsp0_trap1_ready := '0';
                        end if;
--        when X"03" => -- NOP dsp0_status_reg_in;
          when X"04" => dsp0_trap_match0_i <= router_bus_data_in(7 downto 0);
          when X"05" => dsp0_prescale0_i   <= router_bus_data_in;
          when X"06" => dsp0_trap_match1_i <= router_bus_data_in(7 downto 0);
          when X"07" => dsp0_prescale1_i   <= router_bus_data_in;
--        when X"09" => -- NOP dsp0_fifo_word_count_reg_in;
--        when X"0A" => -- NOP 
          when X"0C" => dsp0_int_dly_cnt_i <= router_bus_data_in(5 downto 0);
          when X"0D" => dsp0_int_period_i  <= router_bus_data_in;
-- DSP1 REGS          
          when X"10" => dsp1_cmd_reg0_i    <= router_bus_data_in(8 downto 0);
                        dsp1_trap0_ready   := '0';
          when X"11" => dsp1_cmd_reg1_i    <= router_bus_data_in(4 downto 0);
                        dsp1_trap1_ready   := '0';
          when X"12" => dsp1_trap_reset_i  <= router_bus_data_in(0);
                        dsp1_trap0_load_i  <= router_bus_data_in(1);
                        dsp1_trap1_load_i  <= router_bus_data_in(2);
                        dsp1_fifo_flush_i  <= router_bus_data_in(3);
                        if (router_bus_data_in(1) = '1') then
                          dsp1_trap0_ready := '1';
                        end if;
                        if (router_bus_data_in(2) = '1') then
                          dsp1_trap1_ready := '1';
                        end if;
                        if (router_bus_data_in(0) = '1') then
                          dsp1_trap0_ready := '0';
                          dsp1_trap1_ready := '0';
                        end if;
--        when X"13" => -- NOP dsp1_status_reg_in;
          when X"14" => dsp1_trap_match0_i <= router_bus_data_in(7 downto 0);
          when X"15" => dsp1_prescale0_i   <= router_bus_data_in;
          when X"16" => dsp1_trap_match1_i <= router_bus_data_in(7 downto 0);
          when X"17" => dsp1_prescale1_i   <= router_bus_data_in;
--        when X"19" => -- NOP dsp1_fifo_word_count_reg_in;
--        when X"1A" => -- NOP
          when X"1C" => dsp1_int_dly_cnt_i <= router_bus_data_in(5 downto 0);
          when X"1D" => dsp1_int_period_i  <= router_bus_data_in;
-- DSP2 REGS          
          when X"20" => dsp2_cmd_reg0_i    <= router_bus_data_in(8 downto 0);
                        dsp2_trap0_ready   := '0';
          when X"21" => dsp2_cmd_reg1_i    <= router_bus_data_in(4 downto 0);
                        dsp2_trap1_ready   := '0';
          when X"22" => dsp2_trap_reset_i  <= router_bus_data_in(0);
                        dsp2_trap0_load_i  <= router_bus_data_in(1);
                        dsp2_trap1_load_i  <= router_bus_data_in(2);
                        dsp2_fifo_flush_i  <= router_bus_data_in(3);
                        if (router_bus_data_in(1) = '1') then
                          dsp2_trap0_ready := '1';
                        end if;
                        if (router_bus_data_in(2) = '1') then
                          dsp2_trap1_ready := '1';
                        end if;
                        if (router_bus_data_in(0) = '1') then
                          dsp2_trap0_ready := '0';
                          dsp2_trap1_ready := '0';
                        end if;
--        when X"23" => -- NOP dsp2_status_reg_in;
          when X"24" => dsp2_trap_match0_i <= router_bus_data_in(7 downto 0);
          when X"25" => dsp2_prescale0_i   <= router_bus_data_in;
          when X"26" => dsp2_trap_match1_i <= router_bus_data_in(7 downto 0);
          when X"27" => dsp2_prescale1_i   <= router_bus_data_in;
--        when X"29" => -- NOP dsp2_fifo_word_count_reg_in;
--        when X"2A" => -- NOP
          when X"2C" => dsp2_int_dly_cnt_i <= router_bus_data_in(5 downto 0);
          when X"2D" => dsp2_int_period_i  <= router_bus_data_in;
-- DSP3 REGS          
          when X"30" => dsp3_cmd_reg0_i    <= router_bus_data_in(8 downto 0);
                        dsp3_trap0_ready   := '0';
          when X"31" => dsp3_cmd_reg1_i    <= router_bus_data_in(4 downto 0);
                        dsp3_trap1_ready   := '0';
          when X"32" => dsp3_trap_reset_i  <= router_bus_data_in(0);
                        dsp3_trap0_load_i  <= router_bus_data_in(1);
                        dsp3_trap1_load_i  <= router_bus_data_in(2);
                        dsp3_fifo_flush_i  <= router_bus_data_in(3);
                        if (router_bus_data_in(1) = '1') then
                          dsp3_trap0_ready := '1';
                        end if;
                        if (router_bus_data_in(2) = '1') then
                          dsp3_trap1_ready := '1';
                        end if;
                        if (router_bus_data_in(0) = '1') then
                          dsp3_trap0_ready := '0';
                          dsp3_trap1_ready := '0';
                        end if;
--        when X"33" => -- NOP dsp3_status_reg_in;
          when X"34" => dsp3_trap_match0_i <= router_bus_data_in(7 downto 0);
          when X"35" => dsp3_prescale0_i   <= router_bus_data_in;
          when X"36" => dsp3_trap_match1_i <= router_bus_data_in(7 downto 0);
          when X"37" => dsp3_prescale1_i   <= router_bus_data_in;
--        when X"39" => -- NOP dsp3_fifo_word_count_reg_in;
--        when X"3A" => -- NOP
          when X"3C" => dsp3_int_dly_cnt_i     <= router_bus_data_in( 5 downto 0);
          when X"3D" => dsp3_int_period_i      <= router_bus_data_in;
-- MAIN CONTROL/STATUS
          when X"40" => global_cmd_reg_i       <= router_bus_data_in( 7 downto 0);
                        endian_select_i        <= router_bus_data_in(14);
                        force_xoff_i           <= router_bus_data_in(15);
          when X"41" => slink_AT_dump_match_i  <= router_bus_data_in( 9 downto 0);
          when X"42" => slink_ROD_dump_match_i <= router_bus_data_in( 5 downto 0);
          when X"44" => slink_counter_control_i <= router_bus_data_in(5 downto 4);
                        test_mux_i              <= router_bus_data_in(3 downto 0);
--        when X"45" => Read Only
--        when X"46" => Read Only
          when X"47" => hold_next_delay_i      <= router_bus_data_in;
--        when X"48" => Read Only
--        when X"49" => Read Only
--        when X"4A" => Read Only
--        when X"4B" => Read Only
--        when X"4C" => Read Only (Fragment Count)
--        when X"4D" => Read Only 
--        when X"4E" => Read Only (Hit Count)
--        when X"4F" => Read Only 
--        when X"50" => Read Only (TimeOut Error Count - Bit 2)
--        when X"51" => Read Only 
--        when X"52" => Read Only (EOC Overflow Error Count - Bit 22)
--        when X"53" => Read Only 
--        when X"54" => Read Only (Afull Error - Bit 16)
--        when X"55" => Read Only 
--        when X"56" => Read Only (Row/Column Error - Bit 20)
--        when X"57" => Read Only 
--   To 
--        when X"69" => Read Only 
--   
          when others => -- NOP
        end case;
        rb_access <= done;

      when done =>
        if (acknowledge_cycle_state <= idle) then
          rb_access <= idle;
        end if;
          
      when others =>
        rb_access <= idle;
    end case;

    dsp0_int_mask_delay_out <= dsp0_int_dly_cnt_i;
    dsp1_int_mask_delay_out <= dsp1_int_dly_cnt_i;
    dsp2_int_mask_delay_out <= dsp2_int_dly_cnt_i;
    dsp3_int_mask_delay_out <= dsp3_int_dly_cnt_i;

    global_cmd_reg_out           <= global_cmd_reg_i;
    slink_AT_dump_match_reg_out  <= slink_AT_dump_match_i;
    slink_ROD_dump_match_reg_out <= slink_ROD_dump_match_i;
    test_mux_out                 <= test_mux_i;
    endian_select_out            <= endian_select_i;

    if (clr_sl_reset_in = '1') then
      global_cmd_reg_i(4) <= '0';
    end if;

    if (dsp0_clear_status_i = '1') then
      dsp0_trap_fifo_full_i <= dsp0_status_reg_in(3);
      dsp0_clear_status_i <= '0';
    elsif (dsp0_status_reg_in(3) = '1') then
      dsp0_trap_fifo_full_i <= '1';
    end if;
    if (dsp1_clear_status_i = '1') then
      dsp1_trap_fifo_full_i <= dsp1_status_reg_in(3);
      dsp1_clear_status_i <= '0';
    elsif (dsp1_status_reg_in(3) = '1') then
      dsp1_trap_fifo_full_i <= '1';
    end if;
    if (dsp2_clear_status_i = '1') then
      dsp2_trap_fifo_full_i <= dsp2_status_reg_in(3);
      dsp2_clear_status_i <= '0';
    elsif (dsp2_status_reg_in(3) = '1') then
      dsp2_trap_fifo_full_i <= '1';
    end if;
    if (dsp3_clear_status_i = '1') then
      dsp3_trap_fifo_full_i <= dsp3_status_reg_in(3);
      dsp3_clear_status_i <= '0';
    elsif (dsp3_status_reg_in(3) = '1') then
      dsp3_trap_fifo_full_i <= '1';
    end if;

    if (dsp0_trap_reset_i = '1') then
      dsp0_cmd_reg0_i    <= (others => '0');
      dsp0_cmd_reg1_i    <= (others => '0');
      dsp0_trap_match0_i <= (others => '0');
      dsp0_trap_match1_i <= (others => '0');
      dsp0_prescale0_i   <= (others => '0');
      dsp0_prescale1_i   <= (others => '0');
    end if;
    if (dsp1_trap_reset_i = '1') then
      dsp1_cmd_reg0_i    <= (others => '0');
      dsp1_cmd_reg1_i    <= (others => '0');
      dsp1_trap_match0_i <= (others => '0');
      dsp1_trap_match1_i <= (others => '0');
      dsp1_prescale0_i   <= (others => '0');
      dsp1_prescale1_i   <= (others => '0');
    end if;
    if (dsp2_trap_reset_i = '1') then
      dsp2_cmd_reg0_i    <= (others => '0');
      dsp2_cmd_reg1_i    <= (others => '0');
      dsp2_trap_match0_i <= (others => '0');
      dsp2_trap_match1_i <= (others => '0');
      dsp2_prescale0_i   <= (others => '0');
      dsp2_prescale1_i   <= (others => '0');
    end if;
    if (dsp3_trap_reset_i = '1') then
      dsp3_cmd_reg0_i    <= (others => '0');
      dsp3_cmd_reg1_i    <= (others => '0');
      dsp3_trap_match0_i <= (others => '0');
      dsp3_trap_match1_i <= (others => '0');
      dsp3_prescale0_i   <= (others => '0');
      dsp3_prescale1_i   <= (others => '0');
    end if;
  end if;
end process rw_registers;

--------------------------------------------------------------------------
-- Process to Count Xoff 
--------------------------------------------------------------------------
xoff_counter : process (
  clk40_in, 
  rst_n_in
  )
begin
  if (rst_n_in = '0') then
    xoff_count_i   <= (others => '0');
    latched_xoff_i <= '0';
  elsif (clk40_in'event AND clk40_in = '1') then
    xoff_d1 <= slink_status_reg_in(0);
    if (clr_xoff_count_ii = '1') then
      xoff_count_i <= (others => '0');
    elsif (slink_status_reg_in(0) = '1' AND xoff_d1 = '0' AND xoff_count_i < 65535) then
      xoff_count_i <= xoff_count_i + 1 ;
    end if;

    if (clr_latched_xoff_ii = '1') then
      latched_xoff_i <= slink_status_reg_in(0);
    else
      latched_xoff_i <= slink_status_reg_in(0) OR latched_xoff_i;
    end if;
  end if;
end process xoff_counter;

--------------------------------------------------------------------------
-- Process to Count Slink Words 
--------------------------------------------------------------------------
sl_word_counter : process (
  clk40_in, 
  rst_n_in
  )
begin
  if (rst_n_in = '0') then
    slink_clk_count  <= (others => '0');
    slink_word_count <= (others => '0');
  elsif (clk40_in'event AND clk40_in = '1') then
    if (slink_counter_control_i(1) = '1') then
      slink_clk_count  <= (others => '0');
      slink_word_count <= (others => '0');
    elsif (slink_counter_control_i(0) = '1') then
      slink_clk_count  <= slink_clk_count + '1';
      slink_word_count <= slink_word_count + slink_data_valid_in;
    end if;
  end if;
end process sl_word_counter;

--------------------------------------------------------------------------
-- Process to Load New Trap Parameters
--------------------------------------------------------------------------
load_trap_parameters : process (
  clk40_in, 
  rst_n_in
  )
begin
  if (rst_n_in = '0') then
    dsp0_cmd_reg0_out    <= (others => '0');
    dsp0_cmd_reg1_out    <= (others => '0');
    dsp0_trap_match0_out <= (others => '0');
    dsp0_trap_match1_out <= (others => '0');
    dsp0_prescale0_out   <= (others => '0');
    dsp0_prescale1_out   <= (others => '0');
    dsp1_cmd_reg0_out    <= (others => '0');
    dsp1_cmd_reg1_out    <= (others => '0');
    dsp1_trap_match0_out <= (others => '0');
    dsp1_trap_match1_out <= (others => '0');
    dsp1_prescale0_out   <= (others => '0');
    dsp1_prescale1_out   <= (others => '0');
    dsp2_cmd_reg0_out    <= (others => '0');
    dsp2_cmd_reg1_out    <= (others => '0');
    dsp2_trap_match0_out <= (others => '0');
    dsp2_trap_match1_out <= (others => '0');
    dsp2_prescale0_out   <= (others => '0');
    dsp2_prescale1_out   <= (others => '0');
    dsp3_cmd_reg0_out    <= (others => '0');
    dsp3_cmd_reg1_out    <= (others => '0');
    dsp3_trap_match0_out <= (others => '0');
    dsp3_trap_match1_out <= (others => '0');
    dsp3_prescale0_out   <= (others => '0');
    dsp3_prescale1_out   <= (others => '0');
  elsif (clk40_in'event AND clk40_in = '1') then
    if (dsp0_trap_reset_i = '1') then
      dsp0_cmd_reg0_out    <= (others => '0');
      dsp0_cmd_reg1_out    <= (others => '0');
      dsp0_trap_match0_out <= (others => '0');
      dsp0_trap_match1_out <= (others => '0');
      dsp0_prescale0_out   <= (others => '0');
      dsp0_prescale1_out   <= (others => '0');
    else
      dsp0_cmd_reg0_out    <= dsp0_cmd_reg0_i;
      dsp0_cmd_reg1_out    <= dsp0_cmd_reg1_i;
      dsp0_trap_match0_out <= dsp0_trap_match0_i;
      dsp0_trap_match1_out <= dsp0_trap_match1_i;
      dsp0_prescale0_out   <= dsp0_prescale0_i;
      dsp0_prescale1_out   <= dsp0_prescale1_i;
    end if;
 
    if (dsp1_trap_reset_i = '1') then
      dsp1_cmd_reg0_out    <= (others => '0');
      dsp1_cmd_reg1_out    <= (others => '0');
      dsp1_trap_match0_out <= (others => '0');
      dsp1_trap_match1_out <= (others => '0');
      dsp1_prescale0_out   <= (others => '0');
      dsp1_prescale1_out   <= (others => '0');
    else
      dsp1_cmd_reg0_out    <= dsp1_cmd_reg0_i;
      dsp1_cmd_reg1_out    <= dsp1_cmd_reg1_i;
      dsp1_trap_match0_out <= dsp1_trap_match0_i;
      dsp1_trap_match1_out <= dsp1_trap_match1_i;
      dsp1_prescale0_out   <= dsp1_prescale0_i;
      dsp1_prescale1_out   <= dsp1_prescale1_i;
    end if;

    if (dsp2_trap_reset_i = '1') then
      dsp2_cmd_reg0_out    <= (others => '0');
      dsp2_cmd_reg1_out    <= (others => '0');
      dsp2_trap_match0_out <= (others => '0');
      dsp2_trap_match1_out <= (others => '0');
      dsp2_prescale0_out   <= (others => '0');
      dsp2_prescale1_out   <= (others => '0');
    else
      dsp2_cmd_reg0_out    <= dsp2_cmd_reg0_i;
      dsp2_cmd_reg1_out    <= dsp2_cmd_reg1_i;
      dsp2_trap_match0_out <= dsp2_trap_match0_i;
      dsp2_trap_match1_out <= dsp2_trap_match1_i;
      dsp2_prescale0_out   <= dsp2_prescale0_i;
      dsp2_prescale1_out   <= dsp2_prescale1_i;
    end if;

    if (dsp3_trap_reset_i = '1') then
      dsp3_cmd_reg0_out    <= (others => '0');
      dsp3_cmd_reg1_out    <= (others => '0');
      dsp3_trap_match0_out <= (others => '0');
      dsp3_trap_match1_out <= (others => '0');
      dsp3_prescale0_out   <= (others => '0');
      dsp3_prescale1_out   <= (others => '0');
    else
      dsp3_cmd_reg0_out    <= dsp3_cmd_reg0_i;
      dsp3_cmd_reg1_out    <= dsp3_cmd_reg1_i;
      dsp3_trap_match0_out <= dsp3_trap_match0_i;
      dsp3_trap_match1_out <= dsp3_trap_match1_i;
      dsp3_prescale0_out   <= dsp3_prescale0_i;
      dsp3_prescale1_out   <= dsp3_prescale1_i;
    end if;
  end if;
end process load_trap_parameters;

--------------------------------------------------------------------------
-- Process to Send Interrupts to the Slave DSP for Diagnostics
--------------------------------------------------------------------------
send_int4 : process (
  clk40_in, 
  rst_n_in
  )
begin       
  if (rst_n_in = '0') then
    dsp0_int4_out <= '0';
    dsp1_int4_out <= '0';
    dsp2_int4_out <= '0';
    dsp3_int4_out <= '0';
    dsp0_int_count <= (others => '0');
    dsp1_int_count <= (others => '0');
    dsp2_int_count <= (others => '0');
    dsp3_int_count <= (others => '0');
    mask_fifo_int4_out <= (others => '0');
  elsif (clk40_in'event AND clk40_in = '1') then
    if (dsp0_int_period_i = X"0000") then
      dsp0_int4_out <= '0';
      mask_fifo_int4_out(0) <= '0';
      dsp0_int_count <= (others => '0');
    elsif (unsigned(dsp0_int_period_i) = 1) then
      mask_fifo_int4_out(0) <= '1';
      if (dsp0_int_count = 8) then
        dsp0_int4_out <= '1';
        dsp0_int_count <= dsp0_int_count + 1 ;
      elsif (dsp0_int_count = 28) then
        dsp0_int4_out <= '0';
      else
        dsp0_int_count <= dsp0_int_count + 1 ;
      end if;
    else
      dsp0_int4_out <= '0';
      mask_fifo_int4_out(0) <= '1';
      if (unsigned(dsp0_int_period_i) = dsp0_int_count(17 downto 2)) then
        dsp0_int4_out <= '1';
        dsp0_int_count <= (others => '0');
      else
        dsp0_int_count <= dsp0_int_count + 1 ;
      end if;
    end if;

    if (dsp1_int_period_i = X"0000") then
      dsp1_int4_out <= '0';
      mask_fifo_int4_out(1) <= '0';
      dsp1_int_count <= (others => '0');
    elsif (unsigned(dsp1_int_period_i) = 1 ) then
      mask_fifo_int4_out(1) <= '1';
      if (dsp1_int_count = 8) then
        dsp1_int4_out <= '1';
        dsp1_int_count <= dsp1_int_count + 1 ;
      elsif (dsp1_int_count = 28) then
        dsp1_int4_out <= '0';
      else
        dsp1_int_count <= dsp1_int_count + 1 ;
      end if;
    else
      dsp1_int4_out <= '0';
      mask_fifo_int4_out(1) <= '1';
      if (unsigned(dsp1_int_period_i) = dsp1_int_count(17 downto 2)) then
        dsp1_int4_out <= '1';
        dsp1_int_count <= (others => '0');
      else
        dsp1_int_count <= dsp1_int_count + 1;
      end if;
    end if;

    if (dsp2_int_period_i = X"0000") then
      dsp2_int4_out <= '0';
      mask_fifo_int4_out(2) <= '0';
      dsp2_int_count <= (others => '0');
    elsif (unsigned(dsp2_int_period_i) = 1 ) then
      mask_fifo_int4_out(2) <= '1';
      if (dsp2_int_count = 8) then
        dsp2_int4_out <= '1';
        dsp2_int_count <= dsp2_int_count + 1 ;
      elsif (dsp2_int_count = 28) then
        dsp2_int4_out <= '0';
      else
        dsp2_int_count <= dsp2_int_count + 1 ;
      end if;
    else
      dsp2_int4_out <= '0';
      mask_fifo_int4_out(2) <= '1';
      if (unsigned(dsp2_int_period_i) = dsp2_int_count(17 downto 2)) then
        dsp2_int4_out <= '1';
        dsp2_int_count <= (others => '0');
      else
        dsp2_int_count <= dsp2_int_count + 1;
      end if;
    end if;

    if (dsp3_int_period_i = X"0000") then
      dsp3_int4_out <= '0';
      mask_fifo_int4_out(3) <= '0';
      dsp3_int_count <= (others => '0');
    elsif (unsigned(dsp3_int_period_i) = 1 ) then
      mask_fifo_int4_out(3) <= '1';
      if (dsp3_int_count = 8) then
        dsp3_int4_out <= '1';
        dsp3_int_count <= dsp3_int_count + 1 ;
      elsif (dsp3_int_count = 28) then
        dsp3_int4_out <= '0';
      else
        dsp3_int_count <= dsp3_int_count + 1 ;
      end if;
    else
      dsp3_int4_out <= '0';
      mask_fifo_int4_out(3) <= '1';
      if (unsigned(dsp3_int_period_i) = dsp3_int_count(17 downto 2)) then
        dsp3_int4_out <= '1';
        dsp3_int_count <= (others => '0');
      else
        dsp3_int_count <= dsp3_int_count + 1;
      end if;
    end if;
  end if;
end process;           
 
-------------------------------------------------------------------------------
-- Logic to generate Access Acknowledge
-------------------------------------------------------------------------------
access_acknowledge : process (
  rst_n_in, 
  clk40_in
  )
begin                      
  if (rst_n_in = '0') then
    router_bus_ack_out <= '0';
    acknowledge_cycle_state <= idle;
    clr_xoff_count_ii   <= '0';
    clr_latched_xoff_ii <= '0';
  elsif (clk40_in'event and clk40_in = '1') then
    case acknowledge_cycle_state is
      when idle => 
        router_bus_ack_out <= '0';
        clr_xoff_count_ii   <= '0';
        clr_latched_xoff_ii <= '0';
        if (router_bus_strb_n_in = '0') then 
          acknowledge_cycle_state <= acknowledge;
        end if;
        
      when acknowledge => 
        router_bus_ack_out <= '1';
        if (router_bus_strb_n_in = '1') then 
          acknowledge_cycle_state <= idle;
          clr_xoff_count_ii   <= clr_xoff_count_i;
          clr_latched_xoff_ii <= clr_latched_xoff_i;
        end if;

      when others =>
        router_bus_ack_out <= '0';
        acknowledge_cycle_state <= idle;
    end case;
  end if;  
end process access_acknowledge;     

end rtl; -- code of register_block
