----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    sl_fifo.vhd
-- Description:
--    Paramatrizable Width FIFO
--    Depth of FIFO is 32 words
--    uses 4 bit wide x 5 bits deep dual port mems so widths must be
--    in multiples of 4
--
--    After you issue a read request, it takes 2 cycles for you to
--    get the data back. The almost full, full, and empty flags are
--    updated on the next clock cycle after a ren or wen signal is sent
----------------------------------------------------------------------------

library IEEE ;
use IEEE.STD_LOGIC_1164.all ;
use IEEE.std_logic_arith.all;

-- pragma translate_off
library UNISIM;
use unisim.all;
-- pragma translate_on

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

ENTITY fifo IS
  generic (
    width : integer := 1
    );
  port (
    clk         : in  std_logic;
    rst         : in  std_logic;
    data_in     : in  std_logic_vector(width-1 DOWNTO 0);
    data_out    : out std_logic_vector(width-1 DOWNTO 0);
    ren         : in  std_logic;
    wen         : in  std_logic;
    full        : out std_logic;
    empty       : out std_logic;
    almost_full : out std_logic 
    );
END FIFO ;

architecture rtl of FIFO is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  signal mem_waddr     : STD_LOGIC_VECTOR (3 downto 0) ;
  signal waddr         : UNSIGNED (3 downto 0) ;
  signal mem_raddr     : STD_LOGIC_VECTOR (3 downto 0) ;
  signal raddr         : UNSIGNED (3 downto 0) ;
  signal num_words     : UNSIGNED (3 downto 0) ;
  signal mem_din       : STD_LOGIC_VECTOR (width-1 downto 0) ;
  signal mem_wren      : STD_LOGIC ;
  signal mem_rdout     : STD_LOGIC_VECTOR (width-1 downto 0) ;
  signal full_i        : STD_LOGIC ;
  signal empty_i       : STD_LOGIC ;
  signal almost_full_i : STD_LOGIC ;

----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------

component RAM16X1D
  port (
    DPO   : out std_logic;
    SPO   : out std_logic;
    A0    : in  std_logic;
    A1    : in  std_logic;
    A2    : in  std_logic;
    A3    : in  std_logic;
    D     : in  std_logic;
    DPRA0 : in  std_logic;
    DPRA1 : in  std_logic;
    DPRA2 : in  std_logic;
    DPRA3 : in  std_logic;
    WCLK  : in  std_logic;
    WE    : in  std_logic
    );
end component;

BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

gen_mem_block:
for i in 0 to (width-1) generate
  mem_block1: RAM16X1D
    port map (
      DPO   => mem_rdout(i),
      SPO   => open,
      A0    => mem_waddr(0),
      A1    => mem_waddr(1),
      A2    => mem_waddr(2),
      A3    => mem_waddr(3),
      D     => mem_din(i),
      DPRA0 => mem_raddr(0),
      DPRA1 => mem_raddr(1),
      DPRA2 => mem_raddr(2),
      DPRA3 => mem_raddr(3),
      WCLK  => clk,
      WE    => mem_wren
      );
end generate;

mem_waddr   <= std_logic_vector(waddr);
mem_raddr   <= std_logic_vector(raddr);
full        <= full_i;
empty       <= empty_i;
almost_full <= almost_full_i;
mem_din     <= data_in;
mem_wren    <= wen AND (NOT full_i);

data_out_mux : process (
  clk, 
  rst
  )
begin
  if (rst = '1') then
    data_out <= (others => '0');
  elsif (clk'event and clk = '1') then
    for i in 0 to (width-1) loop
      if (empty_i = '0' AND ren = '1') then
        data_out(i) <= mem_rdout(i);
      else
        data_out(i) <= '0';
      end if;
    end loop;
  end if;
end process data_out_mux;

fifo_logic : process (
  clk, 
  rst
  )
begin -- fifo_logic
  if (rst = '1') then
    almost_full_i <= '0';
    empty_i       <= '1';
    full_i        <= '0';
    num_words     <= (others => '0');
    raddr         <= (others => '0');
    waddr         <= (others => '0');
  elsif (clk'event and clk = '1') then
    if (num_words(3) = '1') then
      almost_full_i <= '1';
    else
      almost_full_i <= '0';
    end if;
    if (wen = '1' and full_i = '0') then
      waddr <= waddr + 1;
      if ((not (ren = '1' and empty_i = '0'))) then
        num_words <= num_words + 1;
        empty_i <= '0';
        if (std_logic_vector(num_words) = "1110") then
          full_i <= '1';
        end if;
      end if;
    end if;    
    if (ren = '1' and empty_i = '0') then
      raddr <= raddr + 1;
      if ((not (wen = '1' and full_i = '0'))) then
        num_words <= num_words - 1;
        full_i <= '0';
        if (std_logic_vector(num_words) = "0001") then
          empty_i <= '1';
        end if;
      end if;
    end if;
  end if;
end process fifo_logic;

END rtl;
