--------------------------------------------------------------------------
-- Particle Physics Detector Electronics Support
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 2000
-- ReadOutDriver Electronics
--
--------------------------------------------------------------------------
-- Filename: trap data event.vhd
-- Description:
--  Searches the event headers for specific event types, and traps data
--  for the backend dsp.  The algorithms for trapping are as follows:
--
--  The following signals are required by all trapping algorithms.
--  DSPn Command Register0
--       Bit 0 : Trap Atlas Specific Event Type
--       Bit 1 : Trap TIM Specific Event Type
--       Bit 2 : Trap ROD Specific Event Type
--       Bit 3 : Data Format - Error / Slink
--       Bit 4 : Exclusion Flag (Trap all except the Type Match)
--       Bit 5 : Trap All Events
--       Bit 6 : Set Trap into Data Driven Mode
--
--  DSPn Reset/Load Trap Register
--       Bit 0 : Reset Trap (clears addr pointers only)
--       Bit 1 : Load Trap0
--       Bit 2 : 
--       Bit 3 : Flush FIFO
--
--  DSPn Trap Match Register0
--    The dsp_trap_match_reg defines the type of event to trap.
--        A.  Atlas Specific -> 8 bits
--        B.  TIM Specific   -> 2 bits
--        C.  ROD Specific   -> 6 bits

--  DSPn PreScale Value0
--    The PreScale register is used as follows:
--      If the value is "rrrrrrrr00000000", only 1 Event will be trapped.
--    The trap will remain in reset until Load New Mask is set again.

--      If the value is "rrrrrrrr00000001", all events will be trapped until
--    the DSP clears the Trap Enable Bit to the Router

--      For all other combinations, the register prescales according to the 
--    programmed value.  The Prescale Value, Bits(7:0), set the period of the
--    trap.  Bits(15:8) act as a Remainder Value.

--  DSPn FIFO Word Count Register
--    The dsp_fifo_word_count_reg register stores the actual occupancy of the
--  Router to DSP EMIF FIFO
--
--  DSPn Interrupt Mask Delay Count
--    The Interrupt Mask Delay Count Register stores the value used to increase
--  the mask time of the DataFrameRdy interrupt at the end of the Slave DSP ISR.
--
--  The following are the connections between a slave DSP and the router
--  FPGA which are involved in event fetching.  Note that the protocol makes
--  use of the external interrupts.
--
--   From Router                   To DSP
--     DSPn_DATABLOCKREADY_N         EXT_INT4
--     DSPn_EVENT_DATA_OUT[31:0]     ED31-ED0   (also tied to SDRAM data)
--                                   (address lines not connected to router)
--
--   From DSP                      To Router
--     CLKX0                         DSPn_TRAP_EN       (Enable Event Trapping)
--     DX1                           DSPn_MASK_DFR_IN   (Interrupt OK)
--     SSCLK                         DSPn_SBSRAMCLK_IN  (80 MHz SBSRAM Clock)
--     CE1                           DSPn_SBSRAM_ME_IN  (Chip Enable to FIFO)
--     SSOE                          DSPn_SBSRAM_SEND_DATA_IN (Output Enable)
--
-- In the router FPGA there are 4 FIFOs, one for each of the slave DSPs.
-- These FIFOs are 1024 x 32 words deep.
--
-- End of Frame Word:
-- 0x400F => Trap 0, No Errors
-- 0x401F => Trap 1, No Errors
-- 0x400E => Trap 0, Errors
-- 0x401E => Trap 1, Errors
-- 0x400D => Trap 0, No Errors, Data Loss
-- 0x401D => Trap 1, No Errors, Data Loss
-- 0x400C => Trap 0, Errors, Data Loss
-- 0x401C => Trap 1, Errors, Data Loss

-- There are two cases.
--
----------------------------------------------------------------------------
--
-- Case 1: Specific Event Type Trapping
--
--
----------------------------------------------------------------------------
--
-- Case 2: Error Event Trapping
--
--
--------------------------------------------------------------------------
-- Structure:
--------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk40
--    *The timing structure of all engine code will be:
--
--------------------------------------------------------------------------
-- Author:         John Joseph
-- Board Engineer:
-- History:
--    3/17/00 JMJ First version
--
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- LIBRARY INCLUDES
--------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations
--------------------------------------------------------------------------
-- SIMULATION LIBRARY INCLUDES
--------------------------------------------------------------------------
entity trap_event_data is
  port (                
    clk40_in            : in  std_logic;
    clk80_in            : in  std_logic;  -- sbsram clk 80MHz
    rst_n_in            : in  std_logic;
    fifo_flush_in       : in  std_logic;
    trap_reset_in       : in  std_logic;
    trap0_load_in       : in  std_logic;
    trap1_load_in       : in  std_logic;
    trap_enable_in      : in  std_logic;
    mask_dfr_in         : in  std_logic;
    hw_capture_in       : in  std_logic;  -- flag to indicate that the header words have been captured
    dsp_cmd_reg0_in     : in  std_logic_vector( 8 downto 0);
    dsp_cmd_reg1_in     : in  std_logic_vector( 4 downto 0);
    dsp_trap_match0_in  : in  std_logic_vector( 7 downto 0);
    dsp_prescale0_in    : in  std_logic_vector(15 downto 0);
    dsp_trap_match1_in  : in  std_logic_vector( 7 downto 0);
    dsp_prescale1_in    : in  std_logic_vector(15 downto 0);
    int_mask_delay_in   : in  std_logic_vector( 5 downto 0);
    atlas_tt_in         : in  std_logic_vector( 7 downto 0);
    tim_tt_in           : in  std_logic_vector( 1 downto 0);
    rod_tt_in           : in  std_logic_vector( 5 downto 0);
    err_evt_in          : in  std_logic;
    eot_evt_in          : in  std_logic;
    fifo_cen_in         : in  std_logic; -- CE from DSP
    fifo_rden_in        : in  std_logic; -- OE from DSP
    fifo_wren_in        : in  std_logic; -- enable writes to FIFO
    data_in             : in  std_logic_vector(31 downto 0);
    data_out            : out std_logic_vector(31 downto 0);
    word_count_out      : out std_logic_vector(12 downto 0); 
    full_flag_out       : out std_logic; --  fifo full flag
    afull_flag_out      : out std_logic; --  fifo almost full flag
    empty_flag_out      : out std_logic; --  fifo empty flag
    data_frame_rdy_out  : out std_logic; --  data frame ready
    trap0_status_out    : out std_logic; --  Status of Trap0 1=Trap On
    trap1_status_out    : out std_logic; --  Status of Trap1 1=Trap On
    event_ok_in         : in  std_logic;
    fifo_rd_pointer_out : out std_logic_vector( 9 downto 0);
    data_frame_rdy_stat : out std_logic_vector( 3 downto 0); 
    mask_fifo_int4_in   : in  std_logic 
    );
end trap_event_data;

architecture rtl of trap_event_data is
--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------

type trap0_states is (trap0_on, trap0_off);
signal trap0_state : trap0_states;

type trap1_states is (trap1_on, trap1_off);
signal trap1_state : trap1_states;

type event_trap_states is (idle, trap_data);
signal event_trap_state : event_trap_states;

signal reset_trap_i          : std_logic;
signal init_trap0_i          : std_logic;
signal init_trap1_i          : std_logic;
signal hw_capture_i          : std_logic;
signal hw_capture_d          : std_logic;
signal rst_occupancy_i       : std_logic;
signal rst_occupancy_ii      : std_logic;
signal rst_fifo_n_i          : std_logic;
signal trap_event_i          : std_logic;
signal eof_location_i        : unsigned(15 downto 0);
signal event_data_i          : std_logic_vector(31 downto 0);
signal wr_event_data_i       : std_logic;
signal last_word_i           : std_logic;
signal trap0_done_i          : std_logic;
signal trap1_done_i          : std_logic;
signal trap0_match_i         : std_logic;
signal trap1_match_i         : std_logic;
signal trap0_num_match_i     : unsigned(7 downto 0);
signal trap1_num_match_i     : unsigned(7 downto 0);
signal inc_trap0_num_match_i : std_logic;
signal inc_trap1_num_match_i : std_logic;
signal trap_sel_i            : std_logic;
signal trap_sel_l            : std_logic;
signal trap0_status_i        : std_logic;
signal trap1_status_i        : std_logic;
signal no_errors_i           : std_logic;
signal lost_data_i           : std_logic;
signal occ_count_out_i       : std_logic_vector(12 downto 0);
signal eof_word_i            : std_logic;
signal wr_count_to_fifo_i    : std_logic;
signal fifo_full_i           : std_logic;
signal fifo_afull_i          : std_logic;
signal data_frame_rdy_i      : std_logic;
signal mask_fifo_afull_i     : std_logic;


-------------------------------------------------------------------------------
-- COMPONENT DECLARATION
-------------------------------------------------------------------------------

component dsp_data_fifo
  port (
    clk40_in              : in  std_logic; -- clk40 input
    clk80_in              : in  std_logic; -- sbsram clk 80MHz
    rst_n_in              : in  std_logic; -- async global reset
    mask_dfr_in           : in  std_logic;
    int_mask_delay_in     : in  std_logic_vector( 5 downto 0);
    cen_in                : in  std_logic; -- CE from DSP
    rden_in               : in  std_logic; -- OE from DSP
    wren_in               : in  std_logic; -- enable writes
    wr_evt_word_count_in  : in  std_logic; -- write event word count
    data_in               : in  std_logic_vector(31 downto 0);
    data_out              : out std_logic_vector(31 downto 0);
    occupancy_count_out   : out std_logic_vector(12 downto 0); 
    afull_flag_out        : out std_logic; --  fifo almost full flag
    full_flag_out         : out std_logic; --  fifo full flag
    empty_flag_out        : out std_logic; --  fifo empty flag
    data_frame_rdy_out    : out std_logic; --  data frame ready
    rd_pointer_out        : out std_logic_vector( 9 downto 0);
    data_frame_rdy_stat   : out std_logic_vector( 3 downto 0);
    fifo_data_fill_mode   : in  std_logic; 
    mask_fifo_int4_in     : in  std_logic 
    );
end component;

begin
-------------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
-------------------------------------------------------------------------------
U1 : dsp_data_fifo
  port map (
    clk40_in              => clk40_in,
    clk80_in              => clk80_in,
    rst_n_in              => rst_fifo_n_i,
    mask_dfr_in           => mask_dfr_in,
    int_mask_delay_in     => int_mask_delay_in,
    cen_in                => fifo_cen_in,
    rden_in               => fifo_rden_in,
    wren_in               => wr_event_data_i,
    wr_evt_word_count_in  => last_word_i,
    data_in               => event_data_i,
    data_out              => data_out,
    occupancy_count_out   => occ_count_out_i,
    afull_flag_out        => fifo_afull_i,
    full_flag_out         => fifo_full_i,
    empty_flag_out        => empty_flag_out,
    data_frame_rdy_out    => data_frame_rdy_i,
    rd_pointer_out        => fifo_rd_pointer_out,
    data_frame_rdy_stat   => data_frame_rdy_stat,
    fifo_data_fill_mode   => dsp_cmd_reg0_in(6),  -- 1 = data driven
    mask_fifo_int4_in     => mask_fifo_int4_in
    );
    
-------------------------------------------------------------------------------
-- PROCESS DECLARATION
-------------------------------------------------------------------------------
word_count_out   <= "00" & 
                     occ_count_out_i(12 downto 10) &
                     occ_count_out_i( 7 downto  0);

trap0_status_out <= trap0_status_i;
trap1_status_out <= trap1_status_i;
full_flag_out    <= fifo_full_i;   -- OR NOT trap_enable_in;
afull_flag_out   <= fifo_afull_i;  -- AND NOT mask_fifo_afull_i; 

data_frame_rdy_out <= data_frame_rdy_i AND trap_enable_in;

-------------------------------------------------------------------------------
-- Register Input Signal as Required
-------------------------------------------------------------------------------
signal_register : process (
  rst_n_in, 
  clk40_in, 
  hw_capture_in
  )
begin
  if (rst_n_in = '0') then
    hw_capture_i <= '0';
    hw_capture_d <= '0';
    no_errors_i  <= '1';
    init_trap0_i <= '1';
    init_trap1_i <= '1';
    rst_fifo_n_i <= '0';
    mask_fifo_afull_i <= '0';
    rst_occupancy_ii <= '0';
  elsif (clk40_in'event and clk40_in = '1') then
    hw_capture_i <= hw_capture_in;
    hw_capture_d <= hw_capture_i;
    no_errors_i  <= event_ok_in;
    init_trap0_i <= trap_reset_in OR trap0_load_in OR NOT trap_enable_in;
    init_trap1_i <= trap_reset_in OR trap1_load_in OR NOT trap_enable_in OR dsp_cmd_reg0_in(5);
    rst_occupancy_ii <= rst_occupancy_i;
    rst_fifo_n_i <= rst_n_in AND NOT (rst_occupancy_i AND NOT rst_occupancy_ii);
    if (trap0_state = trap0_off AND trap1_state = trap1_off) then
      mask_fifo_afull_i <= '1';
    else
      mask_fifo_afull_i <= '0';
    end if;
  end if;
end process;

-------------------------------------------------------------------------------
-- Logic to count Trap0 Matchs
-------------------------------------------------------------------------------
trap0_match : process (
  rst_n_in, 
  clk40_in, 
  trap_reset_in, 
  trap0_load_in, 
  trap_enable_in, 
  hw_capture_i, 
  atlas_tt_in, 
  Tim_tt_in, 
  rod_tt_in
  )
begin
  if (rst_n_in = '0') then
    trap0_match_i         <= '0';
    inc_trap0_num_match_i <= '0';
  elsif (clk40_in'event and clk40_in = '1') then
    if (init_trap0_i = '1') then
      trap0_match_i         <= '0';
      inc_trap0_num_match_i <= '0';
    elsif (lost_data_i = '1' OR fifo_full_i = '1') then
      trap0_match_i         <= '0';
      inc_trap0_num_match_i <= hw_capture_i;
    else
      if (dsp_cmd_reg0_in(5) = '1') then -- Trap All
        trap0_match_i <= '1';
        inc_trap0_num_match_i <= hw_capture_i;
      elsif (dsp_cmd_reg0_in(7) = '1') then  -- Trap Error Events
        if (err_evt_in = '1') then
          trap0_match_i <= '1';
          inc_trap0_num_match_i <= hw_capture_i;
        else
          trap0_match_i <= '0';
        end if;
      elsif (dsp_cmd_reg0_in(8) = '1') then  -- Trap Error Events
        if (eot_evt_in = '1') then
          trap0_match_i <= '1';
          inc_trap0_num_match_i <= hw_capture_i;
        else
          trap0_match_i <= '0';
        end if;
      elsif (dsp_cmd_reg0_in(2 downto 0) = "001") then  -- Atlas Trigger Type
-- After the header words have been captured, test for a match.  Set trap0_match_i
-- and increment trap0_num_match if a match is detected
        if (dsp_trap_match0_in(7 downto 0) = atlas_tt_in) then
          trap0_match_i <= '1';
          inc_trap0_num_match_i <= hw_capture_i;
        else
          trap0_match_i <= '0';
        end if;
      elsif (dsp_cmd_reg0_in(2 downto 0) = "010") then  -- Tim Trigger Type
        if (dsp_trap_match0_in(1 downto 0) = tim_tt_in) then
          trap0_match_i <= '1';
          inc_trap0_num_match_i <= hw_capture_i;
        else
          trap0_match_i <= '0';
        end if;
      elsif (dsp_cmd_reg0_in(2 downto 0) = "100") then  -- ROD Specific Event Type
        if (dsp_trap_match0_in(5 downto 0) = rod_tt_in) then
          trap0_match_i <= '1';
          inc_trap0_num_match_i <= hw_capture_i;
        else
          trap0_match_i <= '0';
        end if;          
      end if;
    end if;
  end if;
end process trap0_match;   

trap0_match_count : process (
  rst_n_in, 
  clk40_in, 
  trap_reset_in, 
  trap0_load_in, 
  trap_enable_in, 
  dsp_prescale0_in, 
  inc_trap0_num_match_i
  )
variable trap0_prescale : unsigned(7 downto 0);                              
begin
  if (rst_n_in = '0') then
    trap0_num_match_i <= (others => '0');
    trap0_prescale    := (others => '0');
  elsif (clk40_in'event and clk40_in = '1') then
    if (trap_reset_in = '1') then
      trap0_num_match_i <= (others => '0');
      trap0_prescale    := (others => '0');
    elsif (trap0_load_in = '1') then
      trap0_num_match_i <= (others => '0');
      trap0_prescale := unsigned(dsp_prescale0_in(7 downto 0)) - 1;
    elsif (inc_trap0_num_match_i = '1') then
      if (trap0_num_match_i = trap0_prescale) then
        trap0_num_match_i <= (others => '0');
      else
        trap0_num_match_i <= trap0_num_match_i + 1;
      end if;
    end if;
  end if;
end process trap0_match_count;   

-------------------------------------------------------------------------------
-- Logic to count Trap1 Matchs
-------------------------------------------------------------------------------
trap1_match : process (
  rst_n_in, 
  clk40_in, 
  trap_reset_in, 
  trap1_load_in, 
  trap_enable_in, 
  hw_capture_i, 
  atlas_tt_in, 
  tim_tt_in, 
  rod_tt_in
  )
begin
  if (rst_n_in = '0') then
    trap1_match_i         <= '0';
    inc_trap1_num_match_i <= '0';
  elsif (clk40_in'event and clk40_in = '1') then
    if (init_trap1_i = '1') then
      trap1_match_i         <= '0';
      inc_trap1_num_match_i <= '0';
    elsif (lost_data_i = '1' OR fifo_full_i = '1') then
      trap1_match_i         <= '0';
      inc_trap1_num_match_i <= hw_capture_i;
    else
      if (dsp_cmd_reg1_in(2 downto 0) = "001") then  -- Atlas Trigger Type
-- After the header words have been captured, test for a match.  Set trap1_match_i
-- and increment trap0_num_match if a match is detected
        if (dsp_trap_match1_in(7 downto 0) = atlas_tt_in) then
          trap1_match_i <= '1';
          inc_trap1_num_match_i <= hw_capture_i;
        else
          trap1_match_i <= '0';
        end if;
      elsif (dsp_cmd_reg1_in(2 downto 0) = "010") then  -- Tim Trigger Type
        if (dsp_trap_match1_in(1 downto 0) = tim_tt_in) then
          trap1_match_i <= '1';
          inc_trap1_num_match_i <= hw_capture_i;
        else
          trap1_match_i <= '0';
        end if;
      elsif (dsp_cmd_reg1_in(2 downto 0) = "100") then  -- ROD Specific Event Type
        if (dsp_trap_match1_in(5 downto 0) = rod_tt_in) then
          trap1_match_i <= '1';
          inc_trap1_num_match_i <= hw_capture_i;
        else
          trap1_match_i <= '0';
        end if;          
      end if;
    end if;
  end if;
end process trap1_match;   

trap1_match_count : process (
  rst_n_in, 
  clk40_in, 
  trap_reset_in, 
  trap1_load_in, 
  trap_enable_in, 
  dsp_prescale1_in, 
  inc_trap1_num_match_i
  )
variable trap1_prescale : unsigned(7 downto 0);                              
begin
  if (rst_n_in = '0') then
    trap1_num_match_i <= (others => '0');
    trap1_prescale    := (others => '0');
  elsif (clk40_in'event and clk40_in = '1') then
    if (trap_reset_in = '1') then
      trap1_num_match_i <= (others => '0');
      trap1_prescale    := (others => '0');
    elsif (trap1_load_in = '1') then
      trap1_num_match_i <= (others => '0');
      trap1_prescale := unsigned(dsp_prescale1_in(7 downto 0)) - 1;
    elsif (inc_trap1_num_match_i = '1') then
      if (trap1_num_match_i = trap1_prescale) then
        trap1_num_match_i <= (others => '0');
      else
        trap1_num_match_i <= trap1_num_match_i + 1;
      end if;
    end if;
  end if;
end process trap1_match_count;   
  
-------------------------------------------------------------------------------
-- Logic to enable trapping of Event Data
-------------------------------------------------------------------------------
trap_event_data : process (
  rst_n_in, 
  clk40_in, 
  trap_reset_in, 
  trap_enable_in, 
  trap0_load_in
  )
begin
  if (rst_n_in = '0') then
    reset_trap_i     <= '1';
    trap0_done_i     <= '0';
    trap0_status_i   <= '0';
    trap0_state      <= trap0_off;
    trap1_done_i     <= '0';
    trap1_status_i   <= '0';
    trap1_state      <= trap1_off;
    event_trap_state <= idle;
  elsif (clk40_in'event and clk40_in = '1') then
    if (trap_reset_in = '1' OR trap_enable_in = '0')then
      reset_trap_i     <= '1';
      trap0_done_i     <= '0';
      trap0_status_i   <= '0';
      trap0_state      <= trap0_off;
      trap1_done_i     <= '0';
      trap1_status_i   <= '0';
      trap1_state      <= trap1_off;
      event_trap_state <= idle;
    elsif (trap_enable_in = '1') then
      reset_trap_i <= '0';
      case event_trap_state is
        when idle =>
          trap0_status_i   <= '0';
          trap0_state      <= trap0_off;
          trap1_status_i   <= '0';
          trap1_state      <= trap1_off;
          if (hw_capture_d = '1' AND fifo_afull_i = '0') then
            event_trap_state <= trap_data;
          end if;
        when trap_data =>
    -- If a trap match has been detected, trap according to the values set in the prescale register
          if (trap0_match_i = '1') then
            trap1_state <= trap1_off;
    -- Trap 1 event only
            if (dsp_prescale0_in(7 downto 0) = "00000000" AND trap0_done_i = '0') then
              trap0_state <= trap0_on;
              if (eof_word_i = '1') then
                trap0_done_i <= '1';
              end if;
    -- Trap all events
            elsif (dsp_prescale0_in(7 downto 0) = "00000001") then
              trap0_state <= trap0_on;
    -- Trap according to the value of the remainder
            elsif (dsp_prescale0_in(7 downto 0) > "00000001") then
              if (std_logic_vector(trap0_num_match_i) = dsp_prescale0_in(15 downto 8)) then
                trap0_state  <= trap0_on;
              else
                trap0_state  <= trap0_off;
              end if;
            else
              trap0_state <= trap0_off;
            end if;
          elsif (trap1_match_i = '1') then
            trap0_state <= trap0_off;
    -- Trap 1 event only
            if (dsp_prescale1_in(7 downto 0) = "00000000" AND trap1_done_i = '0') then
              trap1_state <= trap1_on;
              if (eof_word_i = '1') then
                trap1_done_i <= '1';
              end if;
    -- Trap all events
            elsif (dsp_prescale1_in(7 downto 0) = "00000001") then
              trap1_state <= trap1_on;
    -- Trap according to the value of the remainder
            elsif (dsp_prescale1_in(7 downto 0) > "00000001") then
              if (std_logic_vector(trap1_num_match_i) = dsp_prescale1_in(15 downto 8)) then
                trap1_state  <= trap1_on;
              else
                trap1_state  <= trap1_off;
              end if;
            else
              trap1_state <= trap1_off;
            end if;
          else
            trap0_state <= trap0_off;
            trap1_state <= trap1_off;
          end if;
          if (fifo_full_i = '1') then
            trap0_state <= trap0_off;
            trap1_state <= trap1_off;
            event_trap_state <= idle;
          end if;
        when others =>
          trap0_state <= trap0_off;
          trap1_state <= trap1_off;
      end case;
    else
      trap0_state <= trap0_off;
      trap1_state <= trap1_off;
    end if;
-- Trap status bits sent to the Register Block
    if (trap0_state = trap0_on) then
      trap0_status_i <= '1';
    else
      trap0_status_i <= '0';
    end if;
    if (trap1_state = trap1_on) then
      trap1_status_i <= '1';
    else
      trap1_status_i <= '0';
    end if;
  end if;
end process; -- 

-------------------------------------------------------------------------------
-- Trap Enable Logic
-------------------------------------------------------------------------------
trap_enable : process (
  rst_n_in, 
  trap0_state, 
  trap1_state, 
  reset_trap_i,
  fifo_wren_in, 
  fifo_flush_in
  )
begin
  if (rst_n_in = '0') then
    rst_occupancy_i <= '1';
    trap_event_i    <= '0';
  elsif (reset_trap_i = '1') then
    rst_occupancy_i <= '1';
    trap_event_i    <= '0';
  elsif (trap0_state = trap0_on OR trap1_state = trap1_on) then
    rst_occupancy_i <= fifo_flush_in;
    trap_event_i    <= fifo_wren_in;
  elsif (trap0_state = trap0_off AND trap1_state = trap1_off) then
    rst_occupancy_i <= fifo_flush_in;
    trap_event_i    <= '0';
  end if;
end process; -- 

-------------------------------------------------------------------------------
-- Write In Data Register
-------------------------------------------------------------------------------
event_data_register : process (
  rst_n_in, 
  clk40_in
  )
begin
  if (rst_n_in = '0') then
    event_data_i    <= (others => '0');
    wr_event_data_i <= '0';
    last_word_i     <= '0';
  elsif (clk40_in'event and clk40_in = '1') then
    if (wr_count_to_fifo_i = '1') then
      event_data_i    <= X"40" &     -- 0x40
                         "000" &     trap_sel_l  &  -- Trap 0 or Trap 1
                         "11"  & NOT lost_data_i & no_errors_i &  -- 
                         std_logic_vector(eof_location_i);  -- 0xnnnn
      wr_event_data_i <= '1';
      last_word_i     <= '1';  
    else
      event_data_i    <= data_in;
      wr_event_data_i <= trap_event_i;
      last_word_i     <= '0';
    end if;
  end if;
end process;

-------------------------------------------------------------------------------
-- End of Event Location Counter (16 bits)
-------------------------------------------------------------------------------
eof_location_counter : process (
  rst_n_in, 
  clk40_in, 
  data_in, 
  fifo_wren_in
  )
begin
  if (rst_n_in = '0') then
    wr_count_to_fifo_i <= '0';
    eof_location_i <= (others => '0');
    eof_word_i <= '0';
    lost_data_i <= '0';
  elsif (clk40_in'event AND clk40_in = '1') then
    wr_count_to_fifo_i <= '0';
    eof_word_i <= '0';
    if (hw_capture_i = '1' AND fifo_full_i = '0') then
      lost_data_i <= '0';
    end if;
    if (trap0_state = trap0_on OR trap1_state = trap1_on) then
      if (rst_occupancy_i = '1') then
        eof_location_i <= (others => '0');
      elsif (fifo_wren_in = '1' AND data_in = X"B0F00000") then
        eof_location_i <= X"0001";
      elsif (fifo_wren_in = '1' AND data_in = X"E0F00000") then
        eof_location_i <= eof_location_i + 1;
        wr_count_to_fifo_i <= NOT dsp_cmd_reg0_in(6);
        eof_word_i <= '1';
      elsif (fifo_wren_in = '1' AND -- mask_dfr_in = '1' AND
             occ_count_out_i(12 downto 10) = "011"    AND
             occ_count_out_i(7 downto 0) >= X"FC") then  -- 252
        eof_location_i <= eof_location_i + 1;
        wr_count_to_fifo_i <= NOT dsp_cmd_reg0_in(6);
        eof_word_i  <= '1';
        lost_data_i <= '1';
      elsif (fifo_wren_in = '1') then
        eof_location_i <= eof_location_i + 1;
      end if;
    end if;
  end if;
end process eof_location_counter;
 
-------------------------------------------------------------------------------
-- Trap ID Bit logic
-------------------------------------------------------------------------------
trap_selected : process (
  clk40_in, 
  rst_n_in
  )
begin  -- process trap_type
  if (rst_n_in = '0') then
    trap_sel_i <= '0';
    trap_sel_l <= '0';
  elsif (clk40_in'event and clk40_in = '1') then
    if (trap1_match_i = '1') then
      trap_sel_i <= '1';
    else
      trap_sel_i <= '0';
    end if;
    if (wr_event_data_i = '0') then
      trap_sel_l <= trap_sel_i;
    end if;
  end if;  
end process trap_selected;

end rtl; -- code for trap event data