--------------------------------------------------------------------------
-- Particle Physics Detector Electronics Support
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 2000
-- ReadOutDriver Electronics
--
--------------------------------------------------------------------------
-- Filename: write_over_link_header.vhd
-- Description:
--  Searches the event data for link headers, and replaces the Level 1 ID
--  data bits with Link, Formatter, and EFB information
--
--------------------------------------------------------------------------
-- Structure:
--------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk40
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
--
--   
--------------------------------------------------------------------------
-- Author:         John Joseph
-- Board Engineer:
-- History:
--    3/10/00 JMJ First version
--    
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- LIBRARY INCLUDES
--------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations

--------------------------------------------------------------------------
-- SIMULATION LIBRARY INCLUDES
--------------------------------------------------------------------------

entity write_over_link_header is
  port (
    clk40_in           : in  std_logic; -- 40MHz clock input
    rst_n_in           : in  std_logic; -- active low reset
    event_data_in      : in  std_logic_vector(45 downto 0);
    data_source_n_in   : in  std_logic_vector( 2 downto 0);
         -- "011" -> Header/Trailer
         -- "110" -> EFB A
         -- "101" -> EFB B
    data_valid_in          : in  std_logic; -- data valid from EFB
    slink_wen_out          : out std_logic; --
    slink_data_out         : out std_logic_vector(31 downto 0);
    slink_uctrl_n_out      : out std_logic;
    slink_dword_out        : out std_logic;
    dsps_data_out          : out std_logic_vector(31 downto 0);
    dsps_wen_out           : out std_logic; --
    dspe_data_out          : out std_logic_vector(31 downto 0);
    dspe_wen_out           : out std_logic; --
    L1TT_hw_out            : out std_logic_vector( 9 downto 0);
    DET_hw_out             : out std_logic_vector( 5 downto 0);
    ERR_hw_out             : out std_logic;
    EOT_hw_out             : out std_logic;
    hw_capture_out         : out std_logic;
    event_ok               : out std_logic;
    slink_xoff_in          : in  std_logic;
    hold_efb_out           : out std_logic;
    sl_fifo_stat_out       : out std_logic_vector(2 downto 0);
    rol_test_en_in         : in  std_logic;
    fragment_count_out     : out std_logic_vector(31 downto 0);
    hit_count_out          : out std_logic_vector(31 downto 0);
    to_error_count_out     : out std_logic_vector(31 downto 0);
    eoc_over_count_out     : out std_logic_vector(31 downto 0);
    afull_count_out        : out std_logic_vector(31 downto 0);
    rc_error_count_out     : out std_logic_vector(31 downto 0);
    sync_err_count_out     : out std_logic_vector(31 downto 0); -- (Bit 19 in the Error Flags Status Word 0)
    bcid_chkfail_count_out : out std_logic_vector(31 downto 0); -- (Bit 30)
    l1_chkfail_count_out   : out std_logic_vector(31 downto 0); -- (Bit 31)
    hitoverflow_count_out  : out std_logic_vector(31 downto 0); -- (Bit 27)
    regparity_count_out    : out std_logic_vector(31 downto 0); -- (Bit 24)
    bitflip_count_out      : out std_logic_vector(31 downto 0); -- (Bit 26)
    hammingcode_count_out  : out std_logic_vector(31 downto 0); -- (Bit 23)
    total_error_count_out  : out std_logic_vector(31 downto 0); -- (Bit 23)
    eot_count_out          : out std_logic_vector(31 downto 0); -- (Bit 18)
    stat_count_crtl_in     : in  std_logic_vector( 1 downto 0)
    );
end write_over_link_header;

architecture rtl of write_over_link_header is
--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------

signal efb_number_i     : std_logic;
signal form_number_i    : std_logic_vector(1 downto 0);
signal link_number_i    : std_logic_vector(3 downto 0);
signal timeout_error_i  : std_logic;
signal condensed_mode_i : std_logic;
signal l1id_error_i     : std_logic;
signal bcid_error_i     : std_logic;
signal link_masked_i    : std_logic;
signal spare_i          : std_logic_vector( 1 downto 0);
signal event_data_i     : std_logic_vector(31 downto 0);
signal event_data_ii    : std_logic_vector(31 downto 0);
signal dspe_data_i      : std_logic_vector(31 downto 0);
signal dsps_data_i      : std_logic_vector(31 downto 0);
signal slink_uctrl_n_i  : std_logic;
signal data_valid_i     : std_logic;
signal data_valid_ii    : std_logic;
signal data_source_n_i  : std_logic_vector(2 downto 0);

signal data_pipeline_bypass_i  : std_logic;
signal link_error_status_reg_i : std_logic_vector(191 downto 0);

signal uctrl_header_flag_i : std_logic; -- flag to tag the Event Header
signal en_trailer_count_flag_i : std_logic;
                      -- flag to enable the Event Trailer word counter
signal header_count    : unsigned(3 downto 0);
signal header_count_d1 : unsigned(3 downto 0);
                      -- Event Header word counter
signal trailer_count    : unsigned(2 downto 0);
signal trailer_count_d1 : unsigned(2 downto 0);
                          -- Event Trailer word counter

signal data_word_i    : std_logic;
signal sl_pipe_in_i   : std_logic_vector(34 downto 0);
signal sl_pipe_out_i  : std_logic_vector(34 downto 0);

signal dspe_hold_reg0 : std_logic_vector(32 downto 0);
signal dspe_hold_reg1 : std_logic_vector(32 downto 0);
signal dspe_hold_reg2 : std_logic_vector(32 downto 0);
signal dspe_hold_reg3 : std_logic_vector(32 downto 0);

signal dsps_hold_reg0 : std_logic_vector(32 downto 0);
signal dsps_hold_reg1 : std_logic_vector(32 downto 0);
signal dsps_hold_reg2 : std_logic_vector(32 downto 0);
signal dsps_hold_reg3 : std_logic_vector(32 downto 0);

signal sl_pipe_ren    : std_logic;
signal sl_pipe_wen    : std_logic;
signal sl_pipe_full   : std_logic;
signal sl_pipe_empty  : std_logic;
signal sl_pipe_afull  : std_logic;

signal rst_i : std_logic;

type sl_xoff_states is (
  idle,             -- 0x01
  store_pipeline,   -- 0x02
  wait_for_xoff,    -- 0x04
  xmit_pipeline,    -- 0x08
  done              -- 0x10
  );
signal sl_xoff_state : sl_xoff_states;
signal xoff_pipeline : std_logic_vector(7 downto 0);

signal fragment_count_i     : unsigned(31 downto 0);
signal hit_count_i          : unsigned(31 downto 0);
signal to_error_count_i     : unsigned(31 downto 0);
signal eoc_over_count_i     : unsigned(31 downto 0);
signal afull_count_i        : unsigned(31 downto 0);
signal rc_error_count_i     : unsigned(31 downto 0);
signal sync_err_count_i     : unsigned(31 downto 0); -- (Bit 19 in the Error Flags Status Word 0)
signal bcid_chkfail_count_i : unsigned(31 downto 0); -- (Bit 30)
signal l1_chkfail_count_i   : unsigned(31 downto 0); -- (Bit 31)
signal hitoverflow_count_i  : unsigned(31 downto 0); -- (Bit 27)
signal regparity_count_i    : unsigned(31 downto 0); -- (Bit 24)
signal bitflip_count_i      : unsigned(31 downto 0); -- (Bit 26)
signal hammingcode_count_i  : unsigned(31 downto 0); -- (Bit 23)
signal total_error_count_i  : unsigned(31 downto 0); -- (Bit 23)
signal eot_count_i          : unsigned(31 downto 0); -- (Bit 18)

--------------------------------------------------------------------------
-- COMPONENT DECLARATION
--------------------------------------------------------------------------
component fifo
  generic (
    width : integer := 1
    );
  port (
    clk         : in  std_logic;
    rst         : in  std_logic;
    data_in     : in  std_logic_vector(width-1 DOWNTO 0);
    data_out    : out std_logic_vector(width-1 DOWNTO 0);
    ren         : in  std_logic;
    wen         : in  std_logic;
    full        : out std_logic;
    empty       : out std_logic;
    almost_full : out std_logic 
    );
END component;

begin
--------------------------------------------------------------------------
-- PROCESS DECLARATION
--------------------------------------------------------------------------

-- Process to register all incoming data and control signals

register_data : process (
  rst_n_in, 
  clk40_in, 
  event_data_in,
  data_valid_in, 
  data_source_n_in
  )
begin
  if (rst_n_in = '0') then
    spare_i          <= (others => '0');
    link_masked_i    <= '0';
    bcid_error_i     <= '0';
    l1id_error_i     <= '0';
    condensed_mode_i <= '0';
    timeout_error_i  <= '0';
    efb_number_i     <= '0';
    form_number_i    <= (others => '0');
    link_number_i    <= (others => '0');
    event_data_i     <= (others => '0');
    data_valid_i     <= '0';
    data_source_n_i  <= (others => '1');
  elsif (clk40_in'event and clk40_in = '1') then
    spare_i          <= event_data_in(45 downto 44);
    link_masked_i    <= event_data_in(43);
    bcid_error_i     <= event_data_in(42);
    l1id_error_i     <= event_data_in(41);
    condensed_mode_i <= event_data_in(40);
    timeout_error_i  <= event_data_in(39);
    efb_number_i     <= event_data_in(38);
    form_number_i    <= event_data_in(37 downto 36);
    link_number_i    <= event_data_in(35 downto 32);
    event_data_i     <= event_data_in(31 downto  0);
    data_valid_i     <= data_valid_in;
    data_source_n_i  <= data_source_n_in;
  end if;
end process; -- register_data;


fragment_count_out     <= std_logic_vector(fragment_count_i);
hit_count_out          <= std_logic_vector(hit_count_i);
to_error_count_out     <= std_logic_vector(to_error_count_i);
eoc_over_count_out     <= std_logic_vector(eoc_over_count_i);
afull_count_out        <= std_logic_vector(afull_count_i);
rc_error_count_out     <= std_logic_vector(rc_error_count_i);
sync_err_count_out     <= std_logic_vector(sync_err_count_i);
bcid_chkfail_count_out <= std_logic_vector(bcid_chkfail_count_i);
l1_chkfail_count_out   <= std_logic_vector(l1_chkfail_count_i);
hitoverflow_count_out  <= std_logic_vector(hitoverflow_count_i);
regparity_count_out    <= std_logic_vector(regparity_count_i);
bitflip_count_out      <= std_logic_vector(bitflip_count_i);
hammingcode_count_out  <= std_logic_vector(hammingcode_count_i);
total_error_count_out  <= std_logic_vector(total_error_count_i);
eot_count_out          <= std_logic_vector(eot_count_i);

write_over_link_header: process (
  rst_n_in, 
  clk40_in, 
  event_data_in, 
  event_data_i, 
  data_source_n_i
  )
variable upper_data_id : unsigned(2 downto 0); -- link header ID for upper data word
variable lower_data_id : unsigned(2 downto 0); -- link header ID for lower data word
variable link_error_reg_loc : unsigned(6 downto 0);
variable link_error_reg_pointer : natural RANGE 0 TO 123;
begin
-- load id variables with correct data bits
  upper_data_id := unsigned(event_data_i(31 downto 29)); --
  lower_data_id := unsigned(event_data_i(15 downto 13)); --

  if (rst_n_in = '0') then
    event_data_ii           <= (others => '0');  -- clear internal event data array
    dspe_data_i             <= (others => '0');  -- clear internal event data array
    dsps_data_i             <= (others => '0');  -- clear internal event data array
    data_valid_ii           <= '0';
    slink_uctrl_n_i         <= '1';              -- clear uctrl n bit
    uctrl_header_flag_i     <= '0';              -- clear uctrl header flag
    en_trailer_count_flag_i <= '0';              -- disable trailer counter
    trailer_count           <= (others => '0');  -- clear trailer counter
    trailer_count_d1        <= (others => '0');  -- clear trailer counter
    header_count            <= (others => '0');  -- clear header counter
    header_count_d1         <= (others => '0');  -- clear header counter
    link_error_status_reg_i <= (others => '0');
    link_error_reg_pointer  := 0;
    link_error_reg_loc      := (others => '0');
    hw_capture_out          <= '0';
    data_pipeline_bypass_i  <= '0';
    L1TT_hw_out             <= (others => '0');
    DET_hw_out              <= (others => '0');
    ERR_hw_out              <= '0';
    EOT_hw_out              <= '0';
    event_ok                <= '1';
    data_word_i             <= '0';
    fragment_count_i        <= (others => '0');
    hit_count_i             <= (others => '0');
    to_error_count_i        <= (others => '0');
    eoc_over_count_i        <= (others => '0');
    afull_count_i           <= (others => '0');
    rc_error_count_i        <= (others => '0');
    sync_err_count_i        <= (others => '0');
    bcid_chkfail_count_i    <= (others => '0');
    l1_chkfail_count_i      <= (others => '0');
    hitoverflow_count_i     <= (others => '0');
    regparity_count_i       <= (others => '0');
    bitflip_count_i         <= (others => '0');
    hammingcode_count_i     <= (others => '0');
    total_error_count_i     <= (others => '0');
    eot_count_i             <= (others => '0');
  
  elsif (clk40_in'event and clk40_in = '1') then
    if (stat_count_crtl_in(1) = '1') then
      fragment_count_i     <= (others => '0');
      to_error_count_i     <= (others => '0');
      eoc_over_count_i     <= (others => '0');
      afull_count_i        <= (others => '0');
      rc_error_count_i     <= (others => '0');
      hit_count_i          <= (others => '0');
      sync_err_count_i     <= (others => '0');
      bcid_chkfail_count_i <= (others => '0');
      l1_chkfail_count_i   <= (others => '0');
      hitoverflow_count_i  <= (others => '0');
      regparity_count_i    <= (others => '0');
      bitflip_count_i      <= (others => '0');
      hammingcode_count_i  <= (others => '0');
      total_error_count_i  <= (others => '0');
      eot_count_i          <= (others => '0');
    end if;

    if (rol_test_en_in = '1') then
      data_word_i <= '0';
      data_pipeline_bypass_i <= '0'; -- no bypass: wait until H4 is in the pipeline
      uctrl_header_flag_i    <= '0'; -- clear uctrl header flag
      hw_capture_out         <= '0';
      if (event_data_i(31 downto 19) = "1011000011110") then         -- check for first word of header
        event_data_ii          <= event_data_i(31 downto 19) & "0000000000000000000"; -- pass data to slink 
        dspe_data_i            <= event_data_i(31 downto 19) & "0000000000000000000"; -- pass data to dsp
        dsps_data_i            <= event_data_i(31 downto 19) & "0000000000000000000"; -- pass data to dsp
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '0';            -- set slink_uctrl_n
        L1TT_hw_out            <= event_data_i( 9 downto  0); -- Load Atlas/TIM match reg
        DET_hw_out             <= event_data_i(15 downto 10); -- Load ROD Det match reg
        ERR_hw_out             <= event_data_i(16);           -- Detect Error Event
        EOT_hw_out             <= event_data_i(18);           -- Detect Event Over Threshold
        eot_count_i            <= eot_count_i + event_data_i(18); -- Count the number of Events over Threshold
        hw_capture_out         <= '1';
        event_ok               <= '1';
      elsif (event_data_i(31 downto 16) = X"E0F0") then
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= event_data_i;
        dsps_data_i            <= event_data_i;
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '0';            -- set slink_uctrl_n
      else
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= event_data_i;
        dsps_data_i            <= event_data_i;
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '1';            -- clear slink_uctrl_n
      end if;
    elsif (data_source_n_i = "011") then     -- check for Header/Trailer source 
      data_word_i <= '0';
      if (uctrl_header_flag_i = '0') then         -- check for first word of header
        event_data_ii          <= event_data_i(31 downto 19) & "0000000000000000000"; -- pass data to slink 
        dspe_data_i            <= event_data_i(31 downto 19) & "0000000000000000000"; -- pass data to dsp
        dsps_data_i            <= event_data_i(31 downto 19) & "0000000000000000000"; -- pass data to dsp
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '0';            -- set slink_uctrl_n
        uctrl_header_flag_i    <= '1';            -- set uctrl header flag
        L1TT_hw_out            <= event_data_i( 9 downto  0); -- Load Atlas/TIM match reg
        DET_hw_out             <= event_data_i(15 downto 10); -- Load ROD Det match reg
        ERR_hw_out             <= event_data_i(16);           -- Detect Error Event
        EOT_hw_out             <= event_data_i(18);           -- Detect Event Over Threshold
        eot_count_i            <= eot_count_i + event_data_i(18); -- Count the number of Events over Threshold
        header_count           <= header_count + '1';
        hw_capture_out         <= '1';
        data_pipeline_bypass_i <= '0';            -- no bypass: wait until H4 is in the pipeline
        event_ok               <= '1';
--      event_ok               <= NOT event_data_i(16);
-- Fragment Count
        if (stat_count_crtl_in(0) = '1') then
          if (data_valid_i = '1') then
            fragment_count_i <= fragment_count_i + '1';
          end if;
        end if;

      elsif (std_logic_vector(header_count) = X"1" AND uctrl_header_flag_i = '1') then 
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= event_data_i;   -- pass data to dsp
        dsps_data_i            <= event_data_i;   -- pass data to dsp
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '1';            -- clear slink_uctrl_n
        uctrl_header_flag_i    <= '1';            -- set uctrl header flag
        header_count           <= header_count + '1';
        hw_capture_out         <= '0';
        data_pipeline_bypass_i <= '0';            -- no bypass: wait until H5 is in the pipeline
      elsif (std_logic_vector(header_count) = X"2" AND uctrl_header_flag_i = '1') then 
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= "00000000000000000000000000000110"; -- 0x6h Num Header Words Error Format
        dsps_data_i            <= "00000000000000000000000000000110"; -- 0x6h Num Header Words Error Format
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '1';            -- clear slink_uctrl_n
        uctrl_header_flag_i    <= '1';            -- set uctrl header flag
        header_count           <= header_count + '1';
        hw_capture_out         <= '0';
        data_pipeline_bypass_i <= '0';            -- no bypass: wait until H5 is in the pipeline
      elsif (    std_logic_vector(header_count) >= X"3" 
             AND std_logic_vector(header_count) <= X"5"
             AND uctrl_header_flag_i = '1') then 
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= event_data_i;   -- pass data to dsp
        dsps_data_i            <= event_data_i;   -- pass data to dsp
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '1';            -- clear slink_uctrl_n
        uctrl_header_flag_i    <= '1';            -- set uctrl header flag
        header_count           <= header_count + '1';
        hw_capture_out         <= '0';
        data_pipeline_bypass_i <= '0';            -- no bypass: wait until H5 is in the pipeline
      elsif (std_logic_vector(header_count) = X"6" AND uctrl_header_flag_i = '1') then
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= event_data_i;   -- pass data to dsp
        dsps_data_i            <= event_data_i;   -- pass data to dsp
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '1';            -- clear slink_uctrl_n
        uctrl_header_flag_i    <= '1';            -- set uctrl header flag
        header_count           <= header_count + '1';
        hw_capture_out         <= '0';
        data_pipeline_bypass_i <= '1';            -- move this word to the pipeline cell 
                                                  -- currently containing H2
      elsif (std_logic_vector(header_count) = X"7" AND uctrl_header_flag_i = '1') then 
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= event_data_i;   -- pass data to dsp
        dsps_data_i            <= event_data_i;   -- pass data to dsp
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '1';            -- clear slink_uctrl_n
        uctrl_header_flag_i    <= '1';            -- set uctrl header flag
        header_count           <= header_count + '1';
        hw_capture_out         <= '0';
        data_pipeline_bypass_i <= '1';            -- move this word to the pipeline cell 
                                                  -- currently containing H3
      elsif (std_logic_vector(header_count) = X"8" AND uctrl_header_flag_i = '1') then
        event_data_ii           <= event_data_i;   -- pass data to slink
        dspe_data_i             <= event_data_i;   -- pass data to dsp
        dsps_data_i             <= event_data_i;   -- pass data to dsp
        data_valid_ii           <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i         <= '1';            -- clear slink_uctrl_n
        uctrl_header_flag_i     <= '1';            -- set uctrl header flag
        header_count            <= header_count + '1';
        hw_capture_out          <= '0';
        link_error_status_reg_i <= (others => '0');
        link_error_reg_pointer  := 0;
        link_error_reg_loc      := (others => '0');
        data_pipeline_bypass_i  <= '1';            -- move this word to the pipeline cell 
                                                   -- currently containing H4
      elsif (std_logic_vector(header_count) = X"9" AND uctrl_header_flag_i = '1') then
        event_data_ii           <= event_data_i;   -- pass data to slink
        dspe_data_i             <= event_data_i;   -- pass data to dsp
        dsps_data_i             <= event_data_i;   -- pass data to dsp
        data_valid_ii           <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i         <= '1';            -- clear slink_uctrl_n
        uctrl_header_flag_i     <= '1';            -- set uctrl header flag
        header_count            <= header_count + '1';
        trailer_count           <= (others => '0');  -- clear trailer counter
        en_trailer_count_flag_i <= '1';  -- enable trailer word counter
        hw_capture_out          <= '0';
        data_pipeline_bypass_i  <= '1';            -- move this word to the pipeline cell 
                                                   -- currently containing H5
      elsif (std_logic_vector(trailer_count) = "000" AND en_trailer_count_flag_i = '1') then
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= event_data_i;   -- pass data to dsp
        dsps_data_i            <= event_data_i;   -- pass data to dsp
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '1';            -- clear slink_uctrl_n
        trailer_count          <= trailer_count + '1'; -- count trailer words
        data_pipeline_bypass_i <= '1';

        if (event_data_i /= "00000000000000000000000000000000" AND data_valid_i = '1') then  -- checks the error count
          event_ok <= '0';
        end if;
-- TimeOut Error Count - Bit 2
        if (stat_count_crtl_in(0) = '1' AND data_valid_i = '1') then
          if (event_data_i(2) = '1') then  -- checks the error count
            to_error_count_i <= to_error_count_i + '1';
          end if;
-- Afull Error - Bit 16
          if (event_data_i(16) = '1') then  -- checks the error count
            afull_count_i <= afull_count_i + '1';
          end if;

          if (event_data_i(19) = '1') then   -- (Bit 19 in the Error Flags Status Word 0)
            sync_err_count_i <= sync_err_count_i + '1';
          end if;

-- PXL >> Row/Column Error, SCT >> Flagged Error Count - Bit 20
          if (event_data_i(20) = '1') then  -- checks the error count
            rc_error_count_i <= rc_error_count_i + '1';
          end if;

-- PXL >> EOC Overflow Error Count, SCT >> Non-Seq Hit - Bit 22
          if (event_data_i(22) = '1') then  -- checks the error count
            eoc_over_count_i <= eoc_over_count_i + '1';
          end if;

          if (event_data_i(23) = '1') then  -- (Bit 23)
            hammingcode_count_i <= hammingcode_count_i + '1';
          end if;

          if (event_data_i(24) = '1') then  -- (Bit 24)
            regparity_count_i <= regparity_count_i + '1';
          end if;

            if (event_data_i(25) = '1') then
              bitflip_count_i <= bitflip_count_i + '1';
            end if;

          if (event_data_i(27) = '1') then  -- (Bit 27)
            hitoverflow_count_i <= hitoverflow_count_i + '1';
          end if;

          if (event_data_i(30) = '1') then  -- (Bit 30)
            bcid_chkfail_count_i <= bcid_chkfail_count_i + '1';
          end if;

          if (event_data_i(31) = '1') then  -- (Bit 31)
            l1_chkfail_count_i <= l1_chkfail_count_i + '1';
          end if;

        end if;

      elsif (std_logic_vector(trailer_count) = "001" AND en_trailer_count_flag_i = '1') then
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= event_data_i;   -- pass data to dsp
        dsps_data_i            <= event_data_i;   -- pass data to dsp
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '1';            -- clear slink_uctrl_n
        trailer_count          <= trailer_count + '1'; -- count trailer words
        data_pipeline_bypass_i <= '1';

        if (stat_count_crtl_in(0) = '1' AND data_valid_i = '1') then
          total_error_count_i <= total_error_count_i + unsigned(event_data_i(15 downto 0));
        end if;

      elsif (std_logic_vector(trailer_count) = "010" AND en_trailer_count_flag_i = '1') then
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= link_error_status_reg_i(31 downto 0);
        dsps_data_i            <= link_error_status_reg_i(31 downto 0);
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '1';            -- clear slink_uctrl_n
        trailer_count          <= trailer_count + '1'; -- count trailer words
        data_pipeline_bypass_i <= '1';

      elsif (std_logic_vector(trailer_count) = "011" AND en_trailer_count_flag_i = '1') then
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= link_error_status_reg_i(63 downto 32);
        dsps_data_i            <= link_error_status_reg_i(63 downto 32);
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '1';            -- clear slink_uctrl_n
        trailer_count          <= trailer_count + '1'; -- count trailer words
        data_pipeline_bypass_i <= '1';

      elsif (std_logic_vector(trailer_count) = "100" AND en_trailer_count_flag_i = '1') then
        event_data_ii          <= event_data_i;   -- pass data to slink
        dspe_data_i            <= link_error_status_reg_i(95 downto 64);
        dsps_data_i            <= link_error_status_reg_i(95 downto 64);
        data_valid_ii          <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i        <= '1';            -- clear slink_uctrl_n
        trailer_count          <= trailer_count + '1'; -- count trailer words
        data_pipeline_bypass_i <= '1';

      elsif (std_logic_vector(trailer_count) = "101" AND en_trailer_count_flag_i = '1') then 
        event_data_ii           <= event_data_i;   -- pass data to slink  
        dspe_data_i             <= event_data_i;   -- pass data to dsp
        dsps_data_i             <= event_data_i;   -- pass data to dsp
        data_valid_ii           <= data_valid_i;   -- pass data valid
        slink_uctrl_n_i         <= '0';            -- set slink_uctrl_n
        uctrl_header_flag_i     <= '0';      -- clear uctrl header flag
        en_trailer_count_flag_i <= '0'; -- disable trailer word counter
        data_pipeline_bypass_i  <= '0'; -- no bypass : this word is placed
                                        -- at the beginning of the pipeline
        header_count            <= (others => '0');  -- clear header counter
        trailer_count           <= (others => '0');  -- clear trailer counter
      end if;                
    elsif (data_source_n_i = "110" OR data_source_n_i = "101") then
      hw_capture_out         <= '0';
      data_pipeline_bypass_i <= '1';
      data_word_i            <= '1';
      case upper_data_id is     -- compare link header id array to event data
        when "001"  => 
          event_data_ii(31 downto 16) <= event_data_i(31 downto 28) & 
                                         timeout_error_i            &
                                         l1id_error_i               &
                                         bcid_error_i               &
                                         condensed_mode_i           &
                                         link_masked_i              &
                                         efb_number_i               &
                                         form_number_i              &
                                         link_number_i;
          dsps_data_i(31 downto 16)   <= event_data_i(31 downto 28) &
                                         timeout_error_i            &
                                         l1id_error_i               &
                                         bcid_error_i               &
                                         condensed_mode_i           &
                                         link_masked_i	            &
                                         efb_number_i	              &
                                         form_number_i              &
                                         link_number_i;
          dspe_data_i(31 downto 16)   <= event_data_i(31 downto 16);
          data_valid_ii               <= data_valid_i;  -- pass data valid
          slink_uctrl_n_i             <= '1';           -- clear slink_uctrl_n

          if (timeout_error_i = '1' OR link_masked_i = '1') then   -- "11"
            link_error_status_reg_i(2*(link_error_reg_pointer))    <= '1';
            link_error_status_reg_i(2*(link_error_reg_pointer)+ 1) <= '1';
          elsif (bcid_error_i = '1' OR l1id_error_i = '1') then   -- "10"
            link_error_status_reg_i(2*(link_error_reg_pointer))    <= '0';
            link_error_status_reg_i(2*(link_error_reg_pointer)+ 1) <= '1';
          else  -- Good Data  "01"
            link_error_status_reg_i(2*(link_error_reg_pointer))   <= '1';
            link_error_status_reg_i(2*(link_error_reg_pointer)+1) <= '0';
          end if;  -- No Data does not have a test because the register is always
                     -- Cleared at the beginning of each fragment

         when others => 
           event_data_ii(31 downto 16) <= event_data_i(31 downto 16);
           dspe_data_i(31 downto 16)   <= event_data_i(31 downto 16);
           dsps_data_i(31 downto 16)   <= event_data_i(31 downto 16);
           data_valid_ii               <= data_valid_i;  -- pass data valid
           slink_uctrl_n_i             <= '1';           -- clear slink_uctrl_n
      end case;

      case lower_data_id is     -- compare link header id array to event data
        when "001"  => 
          event_data_ii(15 downto 0) <= event_data_i(15 downto 12) &
                                        timeout_error_i            &
                                        l1id_error_i               &
                                        bcid_error_i               &
                                        condensed_mode_i           &
                                        link_masked_i              &
                                        efb_number_i               &
                                        form_number_i              &
                                        link_number_i;
          dsps_data_i(15 downto 0)   <= event_data_i(31 downto 28) &
                                        timeout_error_i            &
                                        l1id_error_i               &
                                        bcid_error_i               &
                                        condensed_mode_i           &
                                        link_masked_i              &
                                        efb_number_i               &
                                        form_number_i              &
                                        link_number_i;
          dspe_data_i(15 downto 0)   <= event_data_i(15 downto 0);
          data_valid_ii              <= data_valid_i;   -- pass data valid
          slink_uctrl_n_i            <= '1';            -- clear slink_uctrl_n
          if (timeout_error_i = '1' OR link_masked_i = '1') then   -- "11"
            link_error_status_reg_i(2*(link_error_reg_pointer))   <= '1';
            link_error_status_reg_i(2*(link_error_reg_pointer)+1) <= '1';
          elsif (bcid_error_i = '1' OR l1id_error_i = '1') then   -- "10"
            link_error_status_reg_i(2*(link_error_reg_pointer))   <= '0';
            link_error_status_reg_i(2*(link_error_reg_pointer)+1) <= '1';
          else  -- Good Data  "01"
            link_error_status_reg_i(2*(link_error_reg_pointer))   <= '1';
            link_error_status_reg_i(2*(link_error_reg_pointer)+1) <= '0';
          end if;  -- No Data does not have a test because the register is always
                   -- Cleared at the beginning of each fragment

        when others => 
          event_data_ii(15 downto 0) <= event_data_i(15 downto 0);
          dspe_data_i(15 downto 0)   <= event_data_i(15 downto 0);
          dsps_data_i(15 downto 0)   <= event_data_i(15 downto 0);
          data_valid_ii              <= data_valid_i;   -- pass data valid
          slink_uctrl_n_i            <= '1';            -- clear slink_uctrl_n
      end case;
    
    -- Hit Count
      if (stat_count_crtl_in(0) = '1' AND data_valid_i = '1') then
        if (event_data_i(31 downto 29) = "100" AND event_data_i(31 downto 29) = "100") then
          hit_count_i <= hit_count_i + "10";
        elsif (event_data_i(31 downto 29) = "100" AND event_data_i(31 downto 29) /= "100") then
          hit_count_i <= hit_count_i + '1';
        elsif (event_data_i(31 downto 29) /= "100" AND event_data_i(31 downto 29) = "100") then
          hit_count_i <= hit_count_i + '1';
        end if;
      end if;


  -- Hit Count
      if (stat_count_crtl_in(0) = '1' AND data_valid_i = '1') then
        if (event_data_i(31 downto 29) = "100") then
          hit_count_i <= hit_count_i + '1';
        end if;
      end if;

    else
      event_data_ii   <= event_data_i;
      dspe_data_i     <= event_data_i;
      dsps_data_i     <= event_data_i;
      data_valid_ii   <= data_valid_i;
      slink_uctrl_n_i <= '1';
      data_word_i     <= '0';
      hw_capture_out  <= '0';

    end if;

--  Convert Link, Formatter, and EFB ID info to unsigned
--  event_data_in(38)    - EFB #
--  event_data_in(37:36) - Formatter #
--  event_data_in(35:32) - FE Link # -> Values 0 to 11
--  This function must modify the value of the Link info to pack 16 bits down to 12

    if (event_data_in(31 downto 28) = "0010" OR event_data_in(15 downto 12) = "0010") then
      if (event_data_in(38 downto 36) = "000") then
        link_error_reg_loc := unsigned(event_data_in(38 downto 32));
        link_error_reg_pointer := CONV_INTEGER(link_error_reg_loc);  --  Function to convert Binary to Decimal
      elsif (event_data_in(38 downto 36) = "001") then
        link_error_reg_loc := unsigned(event_data_in(38 downto 32));
        link_error_reg_pointer := CONV_INTEGER(link_error_reg_loc) - 4;
      elsif (event_data_in(38 downto 36) = "010") then
        link_error_reg_loc := unsigned(event_data_in(38 downto 32));
        link_error_reg_pointer := CONV_INTEGER(link_error_reg_loc) - 8;
      elsif (event_data_in(38 downto 36) = "011") then
        link_error_reg_loc := unsigned(event_data_in(38 downto 32));
        link_error_reg_pointer := CONV_INTEGER(link_error_reg_loc) - 12;
      elsif (event_data_in(38 downto 36) = "100") then
        link_error_reg_loc := unsigned(event_data_in(38 downto 32));
        link_error_reg_pointer := CONV_INTEGER(link_error_reg_loc) - 16;
      elsif (event_data_in(38 downto 36) = "101") then
        link_error_reg_loc := unsigned(event_data_in(38 downto 32));
        link_error_reg_pointer := CONV_INTEGER(link_error_reg_loc) - 20;
      elsif (event_data_in(38 downto 36) = "110") then
        link_error_reg_loc := unsigned(event_data_in(38 downto 32));
        link_error_reg_pointer := CONV_INTEGER(link_error_reg_loc) - 24;
      elsif (event_data_in(38 downto 36) = "111") then
        link_error_reg_loc := unsigned(event_data_in(38 downto 32));
        link_error_reg_pointer := CONV_INTEGER(link_error_reg_loc) - 28;
      end if;
    end if;

--  Counter delays to line up the DSP pipeline data
    trailer_count_d1 <= trailer_count;
    header_count_d1  <= header_count;
  end if;
end process; -- write_over_link_header


--------------------------------------------------------------------------
-- SLINK DATA PIPELINE
--------------------------------------------------------------------------
--
--
--                         +----------------------+   
--                         |                      |
--   event_data_ii   ------| data_in     data_out |------ slink_data_out 
--   data_valid_ii   ------|                      |------ slink_wen_out
--   slink_uctrl_n_i ------|                      |------ slink_uctrl_n_out
--                         |                      |
--   data_valid_ii   ------| wen              ren |------ NOT slink_xoff_in
--                         |                      |
--             rst_i ------| rst             full |------ sl_pipe_full
--          clk40_in ------|>clk            empty |------ sl_pipe_empty
--                         |                afull |------ sl_pipe_afull
--                         |                      |
--                         +----------------------+
--

slink_pipeline : fifo
  generic map(
    width => 35
    )
  port map (
    clk         => clk40_in,
    rst         => rst_i,
    data_in     => sl_pipe_in_i,
    data_out    => sl_pipe_out_i,
    ren         => sl_pipe_ren,
    wen         => sl_pipe_wen,
    full        => sl_pipe_full,
    empty       => sl_pipe_empty,
    almost_full => sl_pipe_afull
    );  

rst_i       <= NOT rst_n_in;
sl_pipe_ren <= NOT slink_xoff_in;

sl_fifo_stat_out <= sl_pipe_full & sl_pipe_afull & sl_pipe_empty;


slink_data_pipeline : process (
  rst_n_in, 
  clk40_in
  ) 
begin
  if (rst_n_in = '0') then
    sl_pipe_in_i      <= (others => '0');
    sl_pipe_wen       <= '0';
    hold_efb_out      <= '0';
    slink_data_out    <= (others => '0');
    slink_wen_out     <= '0';
    slink_uctrl_n_out <= '1';
    sl_xoff_state     <= idle;
    xoff_pipeline     <= (others => '0');
  elsif (clk40_in'event and clk40_in = '1') then
    sl_pipe_in_i      <= data_word_i & NOT slink_uctrl_n_i & data_valid_ii & event_data_ii; 
    sl_pipe_wen       <= data_valid_ii;

    slink_data_out    <=     sl_pipe_out_i(31 downto 0);
    slink_wen_out     <=     sl_pipe_out_i(32);
    slink_uctrl_n_out <= NOT sl_pipe_out_i(33);
    slink_dword_out   <=     sl_pipe_out_i(34);

    xoff_pipeline <= xoff_pipeline(6 downto 0) & slink_xoff_in;
    case sl_xoff_state is
      when idle =>  -- 0x01
        hold_efb_out <= '0';
--  When the Slink asserts Xoff, apply back pressure to the EFB    
        if (xoff_pipeline /= X"00") then    -- if (slink_xoff_in = '1') then
          sl_xoff_state <= store_pipeline;
        end if;

      when store_pipeline =>  -- 0x02  
        hold_efb_out  <= '1';
        sl_xoff_state <= wait_for_xoff;

--  Now wait for Xoff to clear
      when wait_for_xoff =>  -- 0x04
        if (xoff_pipeline = X"00") then     -- if (slink_xoff_in = '0') then
          sl_xoff_state <= xmit_pipeline;
        end if;

--  Empty the FIFO then release back pressure to EFB
      when xmit_pipeline =>  -- 0x08
        if (xoff_pipeline /= X"00") then    -- if (slink_xoff_in = '1') then
          sl_xoff_state <= wait_for_xoff;
        elsif (sl_pipe_empty = '1') then
          sl_xoff_state <= idle;  --done;
        end if;

--  Return to beginning of the state machine
      when done =>  -- 0x10
        hold_efb_out  <= '0';
        sl_xoff_state <= idle;

      when others =>
        hold_efb_out  <= '0';
        sl_xoff_state <= idle;
    end case;
  end if;
end process; -- slink_data_pipeline


dsp_error_format_pipeline : process (
  rst_n_in, 
  clk40_in
  ) 
begin
  if (rst_n_in = '0') then
    dspe_hold_reg0 <= (others => '0');
    dspe_hold_reg1 <= (others => '0');
    dspe_hold_reg2 <= (others => '0');
    dspe_hold_reg3 <= (others => '0');
    dspe_data_out  <= (others => '0');
    dspe_wen_out   <= '0';
  elsif (clk40_in'event and clk40_in = '1') then
    if (data_pipeline_bypass_i = '1') then
      dspe_hold_reg3 <= data_valid_ii & dspe_data_i;
    elsif (data_pipeline_bypass_i = '0' AND 
           std_logic_vector(header_count_d1) < "0111") then
      dspe_hold_reg0 <= data_valid_ii & dspe_data_i;
      dspe_hold_reg1 <= dspe_hold_reg0;
      dspe_hold_reg2 <= dspe_hold_reg1;
      dspe_hold_reg3 <= dspe_hold_reg2;
    elsif (data_pipeline_bypass_i = '0' AND
           std_logic_vector(trailer_count_d1) = "101") then
      dspe_hold_reg0 <= data_valid_ii & dspe_data_i;
      dspe_hold_reg1 <= data_valid_ii & link_error_status_reg_i(191 downto 160);
      dspe_hold_reg2 <= data_valid_ii & link_error_status_reg_i(159 downto 128);
      dspe_hold_reg3 <= data_valid_ii & link_error_status_reg_i(127 downto  96);
    end if;
    dspe_data_out <= dspe_hold_reg3(31 downto 0);
    dspe_wen_out  <= dspe_hold_reg3(32);
  end if;
end process; -- dsp_data_pipeline


dsp_slink_format_pipeline : process (
  rst_n_in, 
  clk40_in
  ) 
begin
  if (rst_n_in = '0') then
    dsps_hold_reg0 <= (others => '0');
    dsps_hold_reg1 <= (others => '0');
    dsps_hold_reg2 <= (others => '0');
    dsps_hold_reg3 <= (others => '0');
    dsps_data_out  <= (others => '0');
    dsps_wen_out   <= '0';
  elsif (clk40_in'event and clk40_in = '1') then
    if (data_pipeline_bypass_i = '1') then
      dsps_hold_reg3 <= data_valid_ii & dsps_data_i;
    elsif (data_pipeline_bypass_i = '0' AND
           std_logic_vector(header_count_d1) < "0111") then
      dsps_hold_reg0 <= data_valid_ii & dsps_data_i;
      dsps_hold_reg1 <= dsps_hold_reg0;
      dsps_hold_reg2 <= dsps_hold_reg1;
      dsps_hold_reg3 <= dsps_hold_reg2;
    elsif (data_pipeline_bypass_i = '0' AND
           std_logic_vector(trailer_count_d1) = "101") then
      dsps_hold_reg0 <= data_valid_ii & dsps_data_i;
      dsps_hold_reg1 <= data_valid_ii & link_error_status_reg_i(191 downto 160);
      dsps_hold_reg2 <= data_valid_ii & link_error_status_reg_i(159 downto 128);
      dsps_hold_reg3 <= data_valid_ii & link_error_status_reg_i(127 downto  96);
    end if;
    dsps_data_out <= dsps_hold_reg3(31 downto 0);
    dsps_wen_out  <= dsps_hold_reg3(32);
  end if;
end process; -- dsp_data_pipeline

end rtl; -- code for write_over_link_header
