--------------------------------------------------------------------------
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 1999
-- ReadOutDriver Electronics
--  
--------------------------------------------------------------------------
-- Filename: 256x32spram.vhd
-- Description:
--  FifoBuffer for event header trailer debug
--  Current implementation is XILINX specific and uses built in 
--  block rams for fifo memories. 
--
--  This block is completely Xilinx specific code,
--  and comes directly from the "xapp130.pdf" file available from 
--  the Xilinx appnote web page.
--  Although there is copious documentation there, it is not all very 
--  clear. Specifically, there are no timing diagrams or examples of
--  how the memories need to be accessed.
--  
--------------------------------------------------------------------------
-- Structure:
--------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk40
--    *The timing structure of all engine code will be:
--      1.  Latch inputs on positive edge of clk40
--      2.  Perform all logic operations on positive edge of clk40
--      3.  Perform all memory operations on positive edge of clk40
--------------------------------------------------------------------------
-- Author:         Mark L. Nagel
-- Board Engineer: 
-- History:
--    08/09/99 MLN First version
--    Add occupancy counter to set write in address
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- LIBRARY INCLUDES
--------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations

--------------------------------------------------------------------------
-- SIMULATION LIBRARY INCLUDES
--------------------------------------------------------------------------

-- pragma translate_off
library UNISIM;
use unisim.all;
-- pragma translate_on

--------------------------------------------------------------------------
-- PORT DECLARATION
--------------------------------------------------------------------------

entity spram_256x32 is
  port (
    clk40_in        : in  std_logic; -- clk40 input
    rst_fifo_in     : in  std_logic; -- fifo reset
    writein_bus_in  : in  std_logic_vector(31 downto 0);
    write_strobe_in : in  std_logic; --
    read_strobe_in  : in  std_logic; --
    readout_bus_out : out std_logic_vector(31 downto 0);
    fifo_empty_out  : out std_logic; --
    fifo_full_out   : out std_logic
    );
end spram_256x32;

--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------
architecture behave of spram_256x32 is

signal fifo_writein_bus_buf : std_logic_vector(31 downto 0);
signal fifo_readout_bus_buf : std_logic_vector(31 downto 0);
                               
signal fifo_address_i : std_logic_vector(7 downto 0);
signal fifo_wr_addr_i : unsigned(7 downto 0);
signal fifo_rd_addr_i : unsigned(7 downto 0);

signal fifo_en_i : std_logic;
signal fifo_wr_i : std_logic;

signal write_strobe_i : std_logic; --
signal read_strobe_i  : std_logic; --

--------------------------------------------------------------------------
-- COMPONENT DECLARATION
--------------------------------------------------------------------------

component RAMB4_S16
  port (
    DI    : in  STD_LOGIC_VECTOR (15 downto 0);
    EN    : in  STD_ULOGIC;
    WE    : in  STD_ULOGIC;                
    RST   : in  STD_ULOGIC;
    CLK   : in  STD_ULOGIC;
    ADDR  : in  STD_LOGIC_VECTOR (7 downto 0);
    DO    : out STD_LOGIC_VECTOR (15 downto 0)
    ); 
end component;

attribute INIT_00: string;
attribute INIT_01: string;
attribute INIT_02: string;
attribute INIT_03: string;
attribute INIT_04: string;
attribute INIT_05: string;
attribute INIT_06: string;
attribute INIT_07: string;
attribute INIT_08: string;
attribute INIT_09: string;
attribute INIT_0A: string;
attribute INIT_0B: string;
attribute INIT_0C: string;
attribute INIT_0D: string;
attribute INIT_0E: string;
attribute INIT_0F: string;

attribute INIT_00 of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_01 of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_02 of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_03 of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_04 of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_05 of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_06 of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_07 of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_08 of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_09 of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0A of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0B of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0C of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0D of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0E of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0F of MB0: label is
"0000000000000000000000000000000000000000000000000000000000000000";

attribute INIT_00 of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_01 of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_02 of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_03 of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_04 of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_05 of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_06 of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_07 of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_08 of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_09 of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_0A of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_0B of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_0C of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_0D of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_0E of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000"; 
attribute INIT_0F of MB1: label is
"0000000000000000000000000000000000000000000000000000000000000000";

--------------------------------------------------------------------------
--  Truth Table for RAMB4_Sn
--
--  --------------------------------------------------------------------
--  |         Inputs                    ||      Outputs                |
--  --------------------------------------------------------------------
--  | EN | RST | WE | CLK | ADDR |  DI  ||    DO     |  RAM Contents   |
--  --------------------------------------------------------------------
--  |  0 |  X  |  X |  X  |  X   |  X   ||   NoChg   |     NoChng      |
--  --------------------------------------------------------------------
--  |  1 |  1  |  0 |  ^  |  X   |  X   ||    0      |     NoChng      |
--  --------------------------------------------------------------------
--  |  1 |  1  |  1 |  ^  | addr | data ||    0      | RAM(addr)<=data |
--  --------------------------------------------------------------------
--  |  1 |  0  |  0 |  ^  | addr |  X   || RAM(addr) |     NoChng      |
--  --------------------------------------------------------------------
--  |  1 |  0  |  1 |  ^  | addr | data ||   data    | RAM(addr)<=data |
--  --------------------------------------------------------------------
--
--------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
--------------------------------------------------------------------------
begin

MB1 : RAMB4_S16
  port map(
    DI    => fifo_writein_bus_buf(31 downto 16),
    EN    => fifo_en_i,
    WE    => fifo_wr_i,
    RST   => fifo_wr_i,
    CLK   => clk40_in,
    ADDR  => fifo_address_i,
    DO    => fifo_readout_bus_buf(31 downto 16)
    );

MB0 : RAMB4_S16
  port map(
    DI    => fifo_writein_bus_buf(15 downto 0),
    EN    => fifo_en_i,
    WE    => fifo_wr_i,
    RST   => fifo_wr_i,
    CLK   => clk40_in,
    ADDR  => fifo_address_i,
    DO    => fifo_readout_bus_buf(15 downto 0)
    );

-------------------------------------------------------------------------------
-- PROCESS DECLARATION
-------------------------------------------------------------------------------

fifo_writein_bus_buf <= writein_bus_in;
readout_bus_out      <= fifo_readout_bus_buf;

fifo_controller : process (
  rst_fifo_in, 
  clk40_in, 
  write_strobe_in, 
  read_strobe_in
  )
variable fifo_occupancy_i : unsigned(7 downto 0);
variable fifo_empty_i : std_logic;
variable fifo_full_i  : std_logic;
begin
  if (rst_fifo_in = '1') then
    fifo_wr_addr_i   <= (others => '0');
    fifo_rd_addr_i   <= (others => '0');
    fifo_occupancy_i := (others => '0');
    fifo_empty_i     := '1';
    fifo_full_i      := '0';
    fifo_en_i        <= '0';
    fifo_wr_i        <= '0';
    read_strobe_i    <= '0';
    write_strobe_i   <= '0';
  elsif (clk40_in'event and clk40_in = '1') then
    write_strobe_i <= write_strobe_in AND NOT fifo_full_i;
    read_strobe_i  <= read_strobe_in  AND NOT fifo_empty_i;
    fifo_en_i      <= (write_strobe_in AND NOT fifo_full_i) OR (read_strobe_in AND NOT fifo_empty_i);
    fifo_wr_i      <=  write_strobe_in;

    if (write_strobe_i = '1' AND std_logic_vector(fifo_wr_addr_i) < "11111111") then
      fifo_wr_addr_i <= fifo_wr_addr_i + '1';
    elsif (write_strobe_i = '1' AND std_logic_vector(fifo_wr_addr_i) = "11111111") then
      fifo_wr_addr_i <= (others => '0');
    end if;

    if (read_strobe_i = '1' AND std_logic_vector(fifo_rd_addr_i) < "11111111") then
      fifo_rd_addr_i <= fifo_rd_addr_i + '1';
    elsif (read_strobe_i = '1' AND std_logic_vector(fifo_rd_addr_i) = "11111111") then
      fifo_rd_addr_i <= (others => '0');
    end if;

    if (write_strobe_in = '1' AND read_strobe_in = '0' AND fifo_full_i = '0') then
      fifo_occupancy_i := fifo_occupancy_i + '1';
    elsif (write_strobe_in = '0' AND read_strobe_in = '1' AND fifo_empty_i = '0') then
      fifo_occupancy_i := fifo_occupancy_i - '1';
    end if;

    if (std_logic_vector(fifo_occupancy_i) = "00000000") then
      fifo_empty_i := '1';
      fifo_full_i  := '0';
    elsif (std_logic_vector(fifo_occupancy_i) >= "00000001"  AND
           std_logic_vector(fifo_occupancy_i) <  "11111111") then
      fifo_empty_i := '0';
      fifo_full_i  := '0';
    elsif (std_logic_vector(fifo_occupancy_i) >= "11111111") then
      fifo_empty_i := '0';
      fifo_full_i  := '1';
    end if;
    fifo_empty_out <= fifo_empty_i;
    fifo_full_out  <= fifo_full_i;
  end if;
end process fifo_controller;

address_mux : process (
  rst_fifo_in, 
  read_strobe_in, 
  read_strobe_i,
  fifo_wr_addr_i, 
  fifo_rd_addr_i, 
  fifo_address_i,
  fifo_wr_i, 
  write_strobe_i, 
  write_strobe_in
  )
begin
  if (rst_fifo_in = '1') then
    fifo_address_i <= (others => '0');
  elsif (write_strobe_in = '1' OR write_strobe_i = '1' OR fifo_wr_i = '1') then
    fifo_address_i <= std_logic_vector(fifo_wr_addr_i);
  elsif (read_strobe_in = '1' OR read_strobe_i = '1' OR fifo_wr_i = '0') then
    fifo_address_i <= std_logic_vector(fifo_rd_addr_i);
  end if;
end process address_mux;

end behave; -- Code for 256x32spram
