----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    bc_l1_check.vhd
-- Description:
--    ROD Event Fragment Builder
--    SCT Version
--
--    Checks that the BCID and L1 IDs are correct. Flags any errors in the BC
--    or L1 ID. Data from the event fifo keeps track of an L1 offset value for
--    each link. This offset allows the EFB to compensate for missing or extra
--    triggers on a given link.
--
--    Checks the Hit data to make sure that the FE Modules are sending data 
--    from the chips in ascending order
--
--    INPUT DATA FORMAT: data_in[39:0]
--
--      data_in[31:0]  detector hit data
--      data_in[35:32] link ch address (0-11) (set in Formatter)
--      data_in[36]    timeout error          (set in Formatter)
--      data_in[37]    condensed mode         (set in Formatter)
--      data_in[38]    Link masked by MDSP    (set in Formatter)
--      data_in[39]    formatter/efb spare1
--
--    OUTPUT DATA FORMAT: data_out[45:0]
--
--      data_out[31:0] detector data
--      data_out[32:38] link address
--                      [38]    efb engine (0-1)
--                      [37:36] formatter (0-3)
--                      [35:32] link (0-11)
--      data_out[39] timeout error
--      data_out[40] condensed mode
--      data_out[41] L1ID error
--      data_out[42] BCID error
--      data_out[43] Link masked by MDSP
--      data_out[44] formatter/efb spare1
--      data_out[45] end of event fragment
--      data_out[46] Non Sequential Chip Error
--
--    EVENT FIFO DATA FORMAT:
--
--      Level 1 ID        <= ev_fifo_data (11 downto 8)
--      BC ID             <= ev_fifo_data (7 downto 0)
--      L1 Offset Link 0  <= ev_fifo_data (13 downto 12)
--      L1 Offset Link 1  <= ev_fifo_data (15 downto 14)
--      L1 Offset Link 2  <= ev_fifo_data (17 downto 16)
--      L1 Offset Link 3  <= ev_fifo_data (19 downto 18)
--       .
--       .
--       .
--      L1 Offset Link 44 <= ev_fifo_data (101 downto 100)
--      L1 Offset Link 45 <= ev_fifo_data (103 downto 102)
--      L1 Offset Link 46 <= ev_fifo_data (105 downto 104)
--      L1 Offset Link 47 <= ev_fifo_data (107 downto 106)
--
--      Link Offset = "00" <= No Change to L1 ID
--      Link Offset = "01" <= Increment L1 ID
--      Link Offset = "10" <= Decrement L1 ID
--
--    The end of a fragment is signalled when form_enabled_n[3] goes from
--    active low to inactive high. A minimum of three cycles must follow in
--    which no data is transferred. There doesn't have to be a gap between
--    the data words from different formatters but there must be a gap between
--    events
--
--
----------------------------------------------------------------------------

library IEEE ;
use IEEE.STD_LOGIC_1164.all ;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

-- pragma translate_off
library UNISIM;
use unisim.all;
-- pragma translate_on

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

ENTITY bc_l1_check IS
  port (
      clk                 : in  STD_LOGIC ;
      rst                 : in  STD_LOGIC ;
      data_in             : in  STD_LOGIC_VECTOR(39 DOWNTO 0) ;
      data_valid_in_n     : in  STD_LOGIC_VECTOR( 3 DOWNTO 0) ;
      form_enabled_n      : in  STD_LOGIC_VECTOR( 3 DOWNTO 0) ;
      data_out            : out STD_LOGIC_VECTOR(46 DOWNTO 0) ;
      data_valid_out      : out STD_LOGIC ;

      ev_fifo_ren         : out STD_LOGIC ;
      ev_fifo_empty       : in  STD_LOGIC ;
      ev_fifo_empty_error : out STD_LOGIC ;
      ev_fifo_data        : in  STD_LOGIC_VECTOR(111 DOWNTO 0) ;

      gatherer_id         : in  STD_LOGIC ;
      bcid_offset         : in  STD_LOGIC_VECTOR( 7 DOWNTO 0);
      grp_evt_count_out   : out STD_LOGIC_VECTOR(31 downto 0) ;
      grp_evt_enable_in   : in  STD_LOGIC ;
      l1_id_fifo_data     : out STD_LOGIC_VECTOR(15 downto 0) ;
      l1_id_fifo_addr     : in  STD_LOGIC_VECTOR( 7 DOWNTO 0) ;
      enable_l1_trap_in   : in  STD_LOGIC ;
      l1_trap_full_out    : out STD_LOGIC ;
      latched_l1id        : out STD_LOGIC_VECTOR(3 DOWNTO 0) ;
      latched_bcid        : out STD_LOGIC_VECTOR(7 DOWNTO 0)
      );
END bc_l1_check ;

architecture rtl of bc_l1_check is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  signal data_in_d1             : STD_LOGIC_VECTOR(39 DOWNTO 0) ;
  signal data_in_d2             : STD_LOGIC_VECTOR(39 DOWNTO 0) ;
  signal data_in_d3             : STD_LOGIC_VECTOR(39 DOWNTO 0) ;

  signal data_valid_in_d1       : STD_LOGIC_VECTOR(3 DOWNTO 0) ;
  signal data_valid_in_d2       : STD_LOGIC_VECTOR(3 DOWNTO 0) ;
  signal data_valid_in_d3       : STD_LOGIC_VECTOR(3 DOWNTO 0) ;

  signal form_enabled_d0        : STD_LOGIC_VECTOR(3 DOWNTO 0) ;
  signal form_enabled_d1        : STD_LOGIC_VECTOR(3 DOWNTO 0) ;
  signal form_enabled_d2        : STD_LOGIC_VECTOR(3 DOWNTO 0) ;
  signal form_enabled_d3        : STD_LOGIC_VECTOR(3 DOWNTO 0) ;

  signal last_word_out          : STD_LOGIC ;
  signal first_word_in          : STD_LOGIC ;
  signal next_to_last_word_out  : STD_LOGIC ;
  signal ev_fifo_ren_i          : STD_LOGIC ;  
  signal ev_fifo_ren_dly1       : STD_LOGIC ;  
  signal get_event_flag         : STD_LOGIC ;
  signal processing_event       : STD_LOGIC ;

  type   offset_array_type is ARRAY (0 to 47) of UNSIGNED (3 DOWNTO 0); 
  signal l1_offset              : offset_array_type;
  signal l1_id                  : UNSIGNED (3 DOWNTO 0);
  signal bc_id                  : STD_LOGIC_VECTOR (7 DOWNTO 0);  
  signal grp_id                 : STD_LOGIC_VECTOR (3 DOWNTO 0);  
  signal new_ev_data_flag       : STD_LOGIC ;
  signal expected_l1            : UNSIGNED (3 DOWNTO 0);
  signal expected_bc            : STD_LOGIC_VECTOR (7 DOWNTO 0);
  signal num_accept_bc          : UNSIGNED (7 DOWNTO 0);

  signal link_number            : STD_LOGIC_VECTOR (5 DOWNTO 0);  
  signal grp_evt_count_i        : UNSIGNED (31 DOWNTO 0);
  signal flip_b3                : std_logic ;

  signal read_addr  : STD_LOGIC_VECTOR ( 7 DOWNTO 0);
  signal write_addr : STD_LOGIC_VECTOR ( 7 DOWNTO 0);
  signal read_data  : STD_LOGIC_VECTOR (15 DOWNTO 0);
  signal write_data : STD_LOGIC_VECTOR (15 DOWNTO 0);
  signal gnd_bus    : STD_LOGIC_VECTOR (15 DOWNTO 0);
  signal gnd        : STD_LOGIC;
  signal write_en   : STD_LOGIC;
  signal pwr        : STD_LOGIC;
  
  type state_typedef is (
    idle, 
    trap_event,
    hold_for_readout
    );
  signal trap_event_id_state : state_typedef;

-------------------------------------------------------------------------------
-- COMPONENT DECLARATION
-------------------------------------------------------------------------------
component RAMB4_S16_S16
  port (
    ADDRA: IN  std_logic_vector( 7 downto 0);
    ADDRB: IN  std_logic_vector( 7 downto 0);
    DIA:   IN  std_logic_vector(15 downto 0);
    DIB:   IN  std_logic_vector(15 downto 0);
    WEA:   IN  std_logic;
    WEB:   IN  std_logic;
    CLKA:  IN  std_logic;
    CLKB:  IN  std_logic;
    RSTA:  IN  std_logic;
    RSTB:  IN  std_logic;
    ENA:   IN  std_logic;
    ENB:   IN  std_logic;
    DOA:   OUT std_logic_vector(15 downto 0);
    DOB:   OUT std_logic_vector(15 downto 0)
    );
END component;

BEGIN  --  Main Body of VHDL code

-------------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
-------------------------------------------------------------------------------
U1: RAMB4_S16_S16
  port map (
    ADDRA => read_addr,
    ADDRB => write_addr, 
    DIB   => write_data,
    DIA   => gnd_bus,
    WEA   => gnd,
    WEB   => write_en,
    CLKA  => clk,
    CLKB  => clk, 
    RSTA  => gnd,
    RSTB  => gnd,
    ENA   => pwr,
    ENB   => pwr, 
    DOA   => read_data,
    DOB   => open
    );

l1_id_fifo_data <= read_data;
read_addr <= l1_id_fifo_addr;
gnd_bus <= (others => '0');
gnd <= '0';
pwr <= '1';

bram_data : process (
  clk, 
  rst
  )
begin
  if (rst = '1') then
    write_data <= (others => '0');
    write_addr <= (others => '0');
    write_en <= '0';
    trap_event_id_state <= idle;
    l1_trap_full_out <= '0';
  elsif (clk'event and clk = '1') then
    write_data(15 downto 8) <= data_in_d3(23 downto 16);  --    write_data(11 downto  8) <= std_logic_vector(l1_id);
    write_data( 7 downto 4) <= std_logic_vector(expected_l1);
    write_data( 3 downto 0) <= data_in_d3(27 downto 24);
    case trap_event_id_state is
      when idle => 
        l1_trap_full_out <= '0';
        write_en <= '0';
        if (enable_l1_trap_in = '1' AND first_word_in = '1') then
          trap_event_id_state <= trap_event;
          latched_l1id <= std_logic_vector(l1_id);
          latched_bcid <= std_logic_vector(bc_id);
        end if;
        
      when trap_event =>
        write_en <= '0';
        if (data_valid_in_d3 = "1111") then
          if (data_in_d3(31 downto 29) = "001") then
            write_en <= '1';
          end if;
        end if;
        if (last_word_out = '1') then
          l1_trap_full_out <= '1';
          trap_event_id_state <= hold_for_readout;
        end if;

      when others =>
        write_en <= '0';
        if (enable_l1_trap_in = '0') then
          trap_event_id_state <= idle;
        end if;
    end case;      

    if    (link_number(5 downto 4) = "00") then
      if (link_number(3 downto 0) < X"C") then
        write_addr <= "00" & link_number;
      else
        write_addr <= X"FF";
      end if;
    elsif (link_number(5 downto 4) = "01") then
      case link_number(3 downto 0) is
        when X"0" => write_addr <= X"0C";
        when X"1" => write_addr <= X"0D";
        when X"2" => write_addr <= X"0E";
        when X"3" => write_addr <= X"0F";
        when X"4" => write_addr <= X"10";
        when X"5" => write_addr <= X"11";
        when X"6" => write_addr <= X"12";
        when X"7" => write_addr <= X"13";
        when X"8" => write_addr <= X"14";
        when X"9" => write_addr <= X"15";
        when X"A" => write_addr <= X"16";
        when X"B" => write_addr <= X"17";
        when others => write_addr <= X"FF";
      end case;
    elsif (link_number(5 downto 4) = "10") then
      case link_number(3 downto 0) is
        when X"0" => write_addr <= X"18";
        when X"1" => write_addr <= X"19";
        when X"2" => write_addr <= X"1A";
        when X"3" => write_addr <= X"1B";
        when X"4" => write_addr <= X"1C";
        when X"5" => write_addr <= X"1D";
        when X"6" => write_addr <= X"1E";
        when X"7" => write_addr <= X"1F";
        when X"8" => write_addr <= X"20";
        when X"9" => write_addr <= X"21";
        when X"A" => write_addr <= X"22";
        when X"B" => write_addr <= X"23";
        when others => write_addr <= X"FF";
      end case;
    elsif (link_number(5 downto 4) = "11") then
      case link_number(3 downto 0) is
        when X"0" => write_addr <= X"24";
        when X"1" => write_addr <= X"25";
        when X"2" => write_addr <= X"26";
        when X"3" => write_addr <= X"27";
        when X"4" => write_addr <= X"28";
        when X"5" => write_addr <= X"29";
        when X"6" => write_addr <= X"2A";
        when X"7" => write_addr <= X"2B";
        when X"8" => write_addr <= X"2C";
        when X"9" => write_addr <= X"2D";
        when X"A" => write_addr <= X"2E";
        when X"B" => write_addr <= X"2F";
        when others => write_addr <= X"FF";
      end case;
    end if;
  end if;
end process;

-------------------------------------------------------------------------------
-- 
-------------------------------------------------------------------------------

form_enabled_d0 <= (NOT form_enabled_n(3)) & 
                   (NOT form_enabled_n(2)) &
                   (NOT form_enabled_n(1)) &
                   (NOT form_enabled_n(0));

register_data : process (
  clk, 
  rst
  )
begin
  if (rst = '1') then
    data_in_d1       <= (others => '0');
    data_in_d2       <= (others => '0');
    data_in_d3       <= (others => '0');
    data_valid_in_d1 <= (others => '0');
    data_valid_in_d2 <= (others => '0');
    data_valid_in_d3 <= (others => '0');
    form_enabled_d1  <= (others => '0');
    form_enabled_d2  <= (others => '0');
    form_enabled_d3  <= (others => '0');
    data_valid_out   <= '0';
  elsif (clk'event and clk = '1') then
    -- register inputs
    data_in_d1 <= data_in;
    data_in_d2 <= data_in_d1;
    data_in_d3 <= data_in_d2;
 
    data_valid_in_d1 <= (NOT data_valid_in_n(3)) & (NOT data_valid_in_n(2)) &
                        (NOT data_valid_in_n(1)) & (NOT data_valid_in_n(0));
    data_valid_in_d2 <= data_valid_in_d1;
    data_valid_in_d3 <= data_valid_in_d2;

    form_enabled_d1  <= form_enabled_d0;
    form_enabled_d2  <= form_enabled_d1;
    form_enabled_d3  <= form_enabled_d2;
      -- register outputs
    data_valid_out <= ((data_valid_in_d3(0) AND form_enabled_d3(0)) OR
                       (data_valid_in_d3(1) AND form_enabled_d3(1)) OR
                       (data_valid_in_d3(2) AND form_enabled_d3(2)) OR
                       (data_valid_in_d3(3) AND form_enabled_d3(3)));
  end if;
end process register_data;


detect_event_start_end: process (
  clk, 
  rst
  )
begin
  if (rst = '1') then
    ev_fifo_empty_error <= '0';
    get_event_flag      <= '1';
    processing_event    <= '0';
  elsif (clk'event and clk = '1') then
    -- event fifo read enable
    if (next_to_last_word_out = '1') then
      get_event_flag   <= '1';
      processing_event <= '0';
    elsif (ev_fifo_ren_i = '1') then
      get_event_flag   <= '0';
      processing_event <= '1';
    end if;
    -- check if data from new event has arrived but event fifo is empty
    -- this is a fatal sync error that cannot be recovered, reset needed
    if (first_word_in = '1' AND processing_event = '0') then
      ev_fifo_empty_error <= '1';
    end if;
  end if;
end process detect_event_start_end;

ev_fifo_ren_i <= (NOT ev_fifo_empty) and get_event_flag;
ev_fifo_ren   <= ev_fifo_ren_i;

first_word_in         <= (form_enabled_d2(0) AND NOT form_enabled_d3(0));
next_to_last_word_out <= (form_enabled_d1(3) AND NOT form_enabled_d0(3));
last_word_out         <= (form_enabled_d3(3) AND NOT form_enabled_d2(3));

get_l1_bc_id : process (
  clk, 
  rst
  )
begin
  if (rst = '1') then
    new_ev_data_flag <= '0';
    ev_fifo_ren_dly1 <= '0';
    l1_id  <= (others => '0');
    bc_id  <= (others => '0');
    grp_id <= (others => '0');
    for i in 0 to 47 loop
      l1_offset(i) <= (others => '0');
    end loop;
  elsif (clk'event and clk = '1') then
    ev_fifo_ren_dly1 <= ev_fifo_ren_i;    -- add 1 clock of delay to access RAMB
    new_ev_data_flag <= ev_fifo_ren_dly1;
    if (new_ev_data_flag = '1') then
      l1_id  <= unsigned(ev_fifo_data(11 downto 8));  -- unsigned(ev_fifo_data (15 downto 8)
      bc_id  <= ev_fifo_data(7 downto 0);
      grp_id <= ev_fifo_data(15 downto 12);
      for i in 0 to 47 loop
        if    (ev_fifo_data((17+(i*2)) downto (16+(i*2))) = "01") then
          l1_offset(i) <= l1_offset(i) + 1;
        elsif (ev_fifo_data((17+(i*2)) downto (16+(i*2))) = "10") then
          l1_offset(i) <= l1_offset(i) - 1;
        end if;
      end loop;
    end if;
  end if;
end process get_l1_bc_id;


compute_expected_l1 : process (
  link_number, 
  l1_id, 
  l1_offset
  )
begin
  case link_number is
    when "000000" => expected_l1 <= l1_id + l1_offset( 0) + 1 ;  -- the +1 offset per ATC-TD-EC-0002
    when "000001" => expected_l1 <= l1_id + l1_offset( 1) + 1 ;
    when "000010" => expected_l1 <= l1_id + l1_offset( 2) + 1 ;
    when "000011" => expected_l1 <= l1_id + l1_offset( 3) + 1 ;
    when "000100" => expected_l1 <= l1_id + l1_offset( 4) + 1 ;
    when "000101" => expected_l1 <= l1_id + l1_offset( 5) + 1 ;
    when "000110" => expected_l1 <= l1_id + l1_offset( 6) + 1 ;
    when "000111" => expected_l1 <= l1_id + l1_offset( 7) + 1 ;
    when "001000" => expected_l1 <= l1_id + l1_offset( 8) + 1 ;
    when "001001" => expected_l1 <= l1_id + l1_offset( 9) + 1 ;
    when "001010" => expected_l1 <= l1_id + l1_offset(10) + 1 ;
    when "001011" => expected_l1 <= l1_id + l1_offset(11) + 1 ;

    when "010000" => expected_l1 <= l1_id + l1_offset(12) + 1 ;
    when "010001" => expected_l1 <= l1_id + l1_offset(13) + 1 ;
    when "010010" => expected_l1 <= l1_id + l1_offset(14) + 1 ;
    when "010011" => expected_l1 <= l1_id + l1_offset(15) + 1 ;
    when "010100" => expected_l1 <= l1_id + l1_offset(16) + 1 ;
    when "010101" => expected_l1 <= l1_id + l1_offset(17) + 1 ;
    when "010110" => expected_l1 <= l1_id + l1_offset(18) + 1 ;
    when "010111" => expected_l1 <= l1_id + l1_offset(19) + 1 ;
    when "011000" => expected_l1 <= l1_id + l1_offset(20) + 1 ;
    when "011001" => expected_l1 <= l1_id + l1_offset(21) + 1 ;
    when "011010" => expected_l1 <= l1_id + l1_offset(22) + 1 ;
    when "011011" => expected_l1 <= l1_id + l1_offset(23) + 1 ;

    when "100000" => expected_l1 <= l1_id + l1_offset(24) + 1 ;
    when "100001" => expected_l1 <= l1_id + l1_offset(25) + 1 ;
    when "100010" => expected_l1 <= l1_id + l1_offset(26) + 1 ;
    when "100011" => expected_l1 <= l1_id + l1_offset(27) + 1 ;
    when "100100" => expected_l1 <= l1_id + l1_offset(28) + 1 ;
    when "100101" => expected_l1 <= l1_id + l1_offset(29) + 1 ;
    when "100110" => expected_l1 <= l1_id + l1_offset(30) + 1 ;
    when "100111" => expected_l1 <= l1_id + l1_offset(31) + 1 ;
    when "101000" => expected_l1 <= l1_id + l1_offset(32) + 1 ;
    when "101001" => expected_l1 <= l1_id + l1_offset(33) + 1 ;
    when "101010" => expected_l1 <= l1_id + l1_offset(34) + 1 ;
    when "101011" => expected_l1 <= l1_id + l1_offset(35) + 1 ;

    when "110000" => expected_l1 <= l1_id + l1_offset(36) + 1 ;
    when "110001" => expected_l1 <= l1_id + l1_offset(37) + 1 ;
    when "110010" => expected_l1 <= l1_id + l1_offset(38) + 1 ;
    when "110011" => expected_l1 <= l1_id + l1_offset(39) + 1 ;
    when "110100" => expected_l1 <= l1_id + l1_offset(40) + 1 ;
    when "110101" => expected_l1 <= l1_id + l1_offset(41) + 1 ;
    when "110110" => expected_l1 <= l1_id + l1_offset(42) + 1 ;
    when "110111" => expected_l1 <= l1_id + l1_offset(43) + 1 ;
    when "111000" => expected_l1 <= l1_id + l1_offset(44) + 1 ;
    when "111001" => expected_l1 <= l1_id + l1_offset(45) + 1 ;
    when "111010" => expected_l1 <= l1_id + l1_offset(46) + 1 ;
    when "111011" => expected_l1 <= l1_id + l1_offset(47) + 1 ;
    when others   => expected_l1 <= l1_id;
  end case;
end process compute_expected_l1;
  
--expected_bc <= bc_id + bcid_offset;  -- The offset is added for synchronization
expected_bc <= bc_id;  -- The offset is added for in the Evt ID FIFO Stage

decode_link_number: process (
  form_enabled_d3, 
  data_in_d3
  )
begin
  link_number(3 downto 0) <= data_in_d3 (35 downto 32);
  if (form_enabled_d3 = "1000") then
    link_number(5 downto 4) <= "11";
  elsif (form_enabled_d3 = "0100") then
    link_number(5 downto 4) <= "10";
  elsif (form_enabled_d3 = "0010") then
    link_number(5 downto 4) <= "01";
  else
    link_number(5 downto 4) <= "00";
  end if;
end process decode_link_number;
  

check_id_tags : process (
  clk, 
  rst
  )
begin
  if (rst = '1') then
    data_out <= (others => '0');
    flip_b3  <= '0';
    num_accept_bc <= (others => '0');
  elsif (clk'event and clk = '1') then
    data_out(31 downto  0) <= data_in_d3(31 downto 0);
    data_out(37 downto 32) <= link_number;
    data_out(38)           <= gatherer_id;
    data_out(40 downto 39) <= data_in_d3(37 downto 36);
    data_out(44 downto 43) <= data_in_d3(39 downto 38);
          
-- Check L1ID, mask for headers with timeout errors >> V2.5F and Enabled Link not Masked v2.DF
    if (data_in_d3(31 downto 29) = "001" AND  -- Header
        data_in_d3(36) = '0'             AND  -- No TO Error
        data_in_d3(38) = '0'             AND  -- Enabled Link Active   -- Changed in v2F
        data_in_d3(27 downto 24) /= std_logic_vector(expected_l1)) then 
      data_out(41) <= '1';
    end if;

-- Check BCID, mask for headers with timeout errors >> v2.5F and Enabled Link not Masked v2.DF
    if (data_in_d3(31 downto 29) = "001" AND  -- Header
        data_in_d3(36) = '0'             AND  -- No TO Error
        data_in_d3(38) = '0'             AND  -- Enabled Link Active   -- Changed in v2F
        data_in_d3(23 downto 16) /= expected_bc) then
      data_out(42) <= '1';
    else
      data_out(42) <= '0';
    end if;

  -- Check for non-sequential, non-monotonic Chip Numbers in the Data
      -- Sequential Chip Test
        -- o Latch the first chip in the link data.
           --  First chip of even links is always between 0 and 5 for normal or redundant output mode
           --  First chip of odd link is always between 8 and 13 for normal output mode 
           --  First chip of odd link can be between 0 and 13 for redundant output mode with 
             --  the first set of chip numbers usually between 8 and 13.  The order can be any subset
             --  of patterns within the order 8-13-0-5, but 0-5 will not preceed 8-13 in any known case.
        -- o If bit3 = 0, proceed with the test as before - the chip/channel number
        --   should be monotonically increasing for both the possible cases.
        -- o If bit3 = 1, flip bit3 for the purposes of the test for all chips.
        --  This will make the test pass for the 8-13,0-5 case as it makes it appear
        --  to be 0-5,8-13.

      -- The data is not packed at this point in the EFB so the link headers are always in the 
      -- upper half word.  Trailers can be in either half word.
        --    INPUT DATA FORMAT: data_in[39:0]
        --      data_in[31:0]  detector hit data
        --      data_in[35:32] link ch address (0-11) (set in Formatter)
        --      data_in[36]    timeout error          (set in Formatter)
        --      data_in[37]    condensed mode         (set in Formatter)

      --  37 36 35  32 31              16 15              0
      --   1     0000   1FFFFCCCCCCCnnnn   0100000000000000
      --   1     0000   1FFFFCCCCCCCnnnn   1FFFFCCCCCCCnnnn
      --   1     0000   1FFFFCCCCCCCnnnn   1FFFFCCCCCCCnnnn
      --   1     0000   0010xxxxxxxxxxxx   1FFFFCCCCCCCnnnn
      --   1     0000   1FFFFCCCCCCCnnnn   0100000000000000
      --   1     0000   1FFFFCCCCCCCnnnn   1FFFFCCCCCCCnnnn
      --   1     0000   1FFFFCCCCCCCnnnn   1FFFFCCCCCCCnnnn
      --   1     0000   0010xxxxxxxxxxxx   1FFFFCCCCCCCnnnn

--  First Chip Test must occur 1 clock before performing test on hit data.
    if (
        data_valid_in_d2 /= "0000"       AND  -- Check for a data valid
        data_in_d2(37) = '1'             AND  -- Condensed Mode Bit
        data_in_d2(31 downto 29) = "001" AND  -- Link Header in upper half word
        data_in_d2(15) = '1'                  -- Hit data in lower half word
        ) then
      flip_b3 <= data_in_d2(14); -- set flip_bit if 1st chip is between 8 and 13
    end if;          

    if (
        (
         data_valid_in_d3 /= "0000" AND  -- Check for a data valid
         data_in_d3(37) = '1'       AND  -- Condensed Mode Bit
         data_in_d3(31) = '1'       AND  -- Upper Half Word == hit data
         data_in_d3(15) = '1'       AND  -- Lower Half Word == hit data
         (
          (
           flip_b3 = '1' AND  -- if first chip = 8 to 13, flip bit 3 and test
           (
            NOT data_in_d3(14) & data_in_d3(13 downto 4) < NOT data_in_d3(30) & data_in_d3(29 downto 20) -- Seq chip Test
            )
           )
         OR
          (
           flip_b3 = '0' AND  -- if first chip = 0 to 5, test
           (
            data_in_d3(14 downto 4) < data_in_d3(30 downto 20)  -- Seq chip Test
            )
           )
          )
         )
       OR
        (
         data_valid_in_d2 /= "0000" AND  -- Check for a data valid
         data_valid_in_d3 /= "0000" AND  -- Check for a data valid             
         data_in_d2(37) = '1'       AND  -- Condensed Mode Bit
         data_in_d3(37) = '1'       AND  -- Condensed Mode Bit
         data_in_d2(31) = '1'       AND  -- Upper Half Word == hit data
         data_in_d3(15) = '1'       AND  -- Lower Half Word == hit data
         (
          (
           flip_b3 = '1' AND  -- if first chip = 8 to 13, flip bit 3 and test
           (
            NOT data_in_d2(30) & data_in_d2(29 downto 20) < NOT data_in_d3(14) & data_in_d3(13 downto 4) -- Seq chip Test
            )
           )
         OR
          (
           flip_b3 = '0' AND  -- if first chip = 0 to 5, test
           (
            data_in_d2(30 downto 20) < data_in_d3(14 downto 4) -- Seq chip Test
            )
           )
          )
         )
        ) then
      data_out(46) <= '1';  -- used by the Error Detect Block to record error
    else
      data_out(46) <= '0';
    end if;
--    data_out(44) <= data_in_d3(39) OR data_in_d3(38);
    data_out(45) <= last_word_out;
  end if;
end process check_id_tags;

group_event_count : process (
  clk, 
  rst
  )
begin
  if (rst = '1') then
    grp_evt_count_i <= (others => '0');
  elsif (clk'event and clk = '1') then
    if (grp_evt_enable_in = '0') then 
      grp_evt_count_i <= (others => '0');
    else
      if (last_word_out = '1') then
        case grp_id is
          when "0000" => grp_evt_count_i( 3 downto  0) <= grp_evt_count_i( 3 downto  0) + 1;
          when "0001" => grp_evt_count_i( 7 downto  4) <= grp_evt_count_i( 7 downto  4) + 1;
          when "0010" => grp_evt_count_i(11 downto  8) <= grp_evt_count_i(11 downto  8) + 1;
          when "0011" => grp_evt_count_i(15 downto 12) <= grp_evt_count_i(15 downto 12) + 1;
          when "0100" => grp_evt_count_i(19 downto 16) <= grp_evt_count_i(19 downto 16) + 1;
          when "0101" => grp_evt_count_i(23 downto 20) <= grp_evt_count_i(23 downto 20) + 1;
          when "0110" => grp_evt_count_i(27 downto 24) <= grp_evt_count_i(27 downto 24) + 1;
          when "0111" => grp_evt_count_i(31 downto 28) <= grp_evt_count_i(31 downto 28) + 1;
          when others => null;
        end case;
      end if;
    end if;
  end if;
end process group_event_count;
grp_evt_count_out <= STD_LOGIC_VECTOR(grp_evt_count_i);

END rtl;
