-------------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
-------------------------------------------------------------------------------
-- Filename:
--    register_block_interface.vhd
-- Description:
--    ROD Dual Gatherer
--    SCT/Pixel Version
--
--  Notes:
--
-------------------------------------------------------------------------------
--
--
-- Error Mask Registers    
-- Address 00000000 -> 00001100 (RW) = Error mask for EFB 1, Link  1-12
-- Address 00010000 -> 00011100 (RW) = Error mask for EFB 1, Link 13-24
-- Address 00100000 -> 00101100 (RW) = Error mask for EFB 1, Link 25-36
-- Address 00110000 -> 00111100 (RW) = Error mask for EFB 1, Link 37-48
-- Address 01000000 -> 01001100 (RW) = Error mask for EFB 2, Link 49-60
-- Address 01010000 -> 01011100 (RW) = Error mask for EFB 2, Link 61-72
-- Address 01100000 -> 01101100 (RW) = Error mask for EFB 2, Link 73-84
-- Address 01110000 -> 01111100 (RW) = Error mask for EFB 2, Link 85-96

-- Format Version Register
-- Address 10000000 (RW) = Format Version

-- Source ID Register
-- Address 10000001 (RW) = Source ID

-- Run Number Register
-- Address 10000010 (RW) = Run Number

-- Send Empty Events Control Register
-- Address 10000100 (RW) = Bit    0 : Send Empty Events
--                         Bit    1 : Mask BCID Error
--                         Bit    2 : Group Event Counter Enable
--                         Bits[6:4]: Input Link Test Out Mux

-- Formatter Mask Control/Status Register
-- Address 10000101  (R)   Bit 0: NOT form_enabled_n1_i(0)
--                   (R)   Bit 1: NOT form_enabled_n1_i(1)
--                   (R)   Bit 2: NOT form_enabled_n1_i(2)
--                   (R)   Bit 3: NOT form_enabled_n1_i(3)
--                   (R)   Bit 4: NOT form_enabled_n2_i(0)
--                   (R)   Bit 5: NOT form_enabled_n2_i(1)
--                   (R)   Bit 6: NOT form_enabled_n2_i(2)
--                   (R)   Bit 7: NOT form_enabled_n2_i(3)

-- EFB RunTime Status Register
-- Address 10000110 (R)  = Bit  0: FIFO1  NOT "almost_full_n" Status Flag
--                         Bit  1: err_sumry_fifo1_almost_full
--                         Bit  2: ev_id_fifo_empty_error1
--                         Bit  3: fifo 1 pause to Formatter
--                         Bit  4: FIFO2  NOT "almost_full_n" Status Flag
--                         Bit  5: err_sumry_fifo2_almost_full
--                         Bit  6: ev_id_fifo_empty_error2
--                         Bit  7: fifo 2 pause to Formatter
--                         Bit  8: halt output from router
--                         Bit  9: rod_event_type_empty
--                         Bit 10: rod_event_type_afull
--                         Bit 11: rod_event_type_full
--    

-- EFB Header ID Status Registers
-- Address 10001000 (R)  = ev_header_data;
-- Address 10001001 (R)  = ev_fifo_data1(15 downto 0);
-- Address 10001010 (R)  = ev_fifo_data2(15 downto 0);
-- Address 10001100 (R)  = Group Event Counters
--                         Bits[ 3: 0] - Event Count ROD Type 0
--                         Bits[ 7: 4] - Event Count ROD Type 1
--                         Bits[11: 8] - Event Count ROD Type 2
--                         Bits[15:12] - Event Count ROD Type 3
--                         Bits[19:16] - Event Count ROD Type 4
--                         Bits[23:20] - Event Count ROD Type 5
--                         Bits[27:24] - Event Count ROD Type 6
--                         Bits[31:28] - Event Count ROD Type 7
-- Address 10001101 (R)  = Bandwidth Counter
--                         Bits[15: 0] - Data Count
--                         Bits[31:16] - Time Count
-- EFB Diagnostic Mode Control Register
-- Address 10010000 (RW) = Bit 0: FIFO Diag Mode Enable EFB Control
--                                Write to Eventmem using the ROD Bus
--                                EFB Path, play out to the Router
--                       = Bit 1: FIFO Diag Mode Enable EventMem Control
--                                Write to Eventmem using the ROD Bus
--                                EFB Path, read back over the Eventmem bus
--                       = Bit 2: FIFO Diag Mode write to  EventMem 
--                                from normal EFB, read back over
--                                the Eventmem bus
--         (Bits 0 and 2 cannot be turned on simultaneously)

-- Address 10010001 (RW) = Bit  0: FIFO 1 Select / Enable Output
--                         Bit  1: FIFO 2 Select / Enable Output
--                         Bit  2: HT FIFO Select / Enable Output
--         (Bits 0 to 2 cannot be turned on simultaneously)
--                         Bit  3: Output 1 Event Fragment
--                         Bit  4: Flag to indicate FIFO1 Bus is Enabled
--                         Bit  5: Flag to indicate FIFO2 Bus is Enabled
--                         Bit  6: Flag to indicate Event Header/Trailer
--                                Bus is enabled
--         (At most only one bit among Bits 4 to 6 can be set TRUE)
--                         Bit  7: Not Used

-- EFB FIFO Reset Control Register
-- Address 10010010 (W)  = Bit 0: FIFO Reset. Also cancels an ongoing
--                                play command
-- EFB FIFO Status Register
-- Address 10010011 (R)  = Bit  0: FIFO1  NOT "empty_n" Status Flag
--                         Bit  1: FIFO1  NOT "almost_empty_n" Status Flag
--                         Bit  2: FIFO1  NOT "full_n" Status Flag
--                         Bit  3: FIFO1  NOT "almost_full_n" Status Flag
--                         Bit  4: FIFO2  NOT "empty_n" Status Flag
--                         Bit  5: FIFO2  NOT "almost_empty_n" Status Flag
--                         Bit  6: FIFO2  NOT "full_n" Status Flag
--                         Bit  7: FIFO2  NOT "almost_full_n" Status Flag
--                         Bit  8: HTFIFO "empty" Status Flag
--                         Bit  9: HTFIFO "full" Status Flag
--                         Bit 10: ev_id_empty1
--                         Bit 11: ev_id_almost_full1
--                         Bit 12: ev_id_empty2
--                         Bit 13: ev_id_almost_full2
--                         Bit 14: header_ev_id_empty
--                         Bit 15: header_ev_id_almost_full

-- Diag Mode FIFO Word Input Register
-- Address 10010100 (RW) = FIFO Input Data Word [15:0]
-- Address 10010101 (RW) = FIFO Input Data Word [31:0]
-- Address 10010110 (RW) = Bit [13:0]: FIFO Input Data Word [45:32]

-- Diag Mode FIFO Write Control Register 
-- Address 10010111 (W)  = Bit 0: Write Input Data Word to selected FIFO

-- Diag Mode FIFO Word Count Control Register s
-- Address 10011000 (RW) = Bit [13:0]: Number of Words to Play from FIFO A
-- Address 10011001 (RW) = Bit [13:0]: Number of Words to Play from FIFO B

-- Diag Mode FIFO Play Control Register 
-- Address 10011010 (RW) = Bit 0: Write a '1' to Play the specified number
--                                of words out of the selected FIFO. A '1'
--                                is read if the fifo is still being played
--                                A '0' is read when the play command has
--                                completed.
--
--efb_misc_stts_reg_i <=   count_fifo1_full              -- (30)
--                       & count_fifo2_full              -- (29)
--                       & err_sumry_fifo1_full          -- (28)
--                       & err_sumry_fifo2_full          -- (27)
--                       & ev_data_empty_o               -- (26)
--                       & ev_id_full1                   -- (25)
--                       & ev_id_full2                   -- (24)
--                       & header_ev_id_full             -- (23)
--                       & ev_data_almost_full_n_o       -- (22)
--                       & header_ev_id_almost_full      -- (21)
--                       & ev_id_almost_full2            -- (20)
--                       & ev_id_almost_full1            -- (19)
--                       & header_ev_id_data_out(35)     -- (18)
--                       & form_enabled_n2_i             -- (17:14)
--                       & form_enabled_n1_i             -- (13:10)
--                       & gatherer_halt_output_i        -- (9)
--                       & fifo1_pause_o                 -- (8)
--                       & fifo2_pause_o                 -- (7)
--                       & header_ev_id_empty            -- (6)
--                       & ev_id_empty2                  -- (5)
--                       & ev_id_empty1                  -- (4)
--                       & ev_id_fifo_empty_error2       -- (3)
--                       & ev_id_fifo_empty_error1       -- (2)
--                       & err_sumry_fifo2_almost_full   -- (1)
--                       & err_sumry_fifo1_almost_full;  -- (0)
--
-------------------------------------------------------------------------------

library IEEE ;
use IEEE.STD_LOGIC_1164.all ;
use IEEE.std_logic_arith.all;

-------------------------------------------------------------------------------
--PORT DECLARATION
-------------------------------------------------------------------------------

ENTITY efb_bus_interface IS
  port (
    clk                      : in  STD_LOGIC;
    rst                      : in  STD_LOGIC;
    err_mask_wen1            : out STD_LOGIC;
    err_mask_din1            : out STD_LOGIC_VECTOR(31 DOWNTO 0);
    err_mask_dout1           : in  STD_LOGIC_VECTOR(31 DOWNTO 0);
    err_mask_sel1            : out STD_LOGIC_VECTOR( 5 DOWNTO 0);      
    err_mask_wen2            : out STD_LOGIC;
    err_mask_din2            : out STD_LOGIC_VECTOR(31 DOWNTO 0);
    err_mask_dout2           : in  STD_LOGIC_VECTOR(31 DOWNTO 0);
    err_mask_sel2            : out STD_LOGIC_VECTOR( 5 DOWNTO 0);      
    source_id                : out STD_LOGIC_VECTOR(31 DOWNTO 0);
    format_version           : out STD_LOGIC_VECTOR(31 DOWNTO 0);
    run_number               : out STD_LOGIC_VECTOR(31 DOWNTO 0);
    send_empty_events        : out STD_LOGIC;
    efb_bus_sel_n            : in  STD_LOGIC;
    efb_bus_rnw              : in  STD_LOGIC;
    efb_bus_hwsel            : in  STD_LOGIC;
    efb_bus_addr             : in  STD_LOGIC_VECTOR( 7 DOWNTO 0);
    efb_bus_data_in          : in  STD_LOGIC_VECTOR(15 DOWNTO 0);
    efb_bus_data_out         : out STD_LOGIC_VECTOR(15 DOWNTO 0);
    efb_bus_data_oe_n        : out STD_LOGIC;
    efb_bus_ack              : out STD_LOGIC;
    out_fifo_diag_mode       : out STD_LOGIC_VECTOR(2 DOWNTO 0);
    out_fifo1_diag_en        : out STD_LOGIC;
    out_fifo2_diag_en        : out STD_LOGIC;
    ht_fifo_diag_en          : out STD_LOGIC;
    out_fifo_rst             : out STD_LOGIC;
    out_fifo1_empty_n        : in  STD_LOGIC;
    out_fifo1_aempty_n       : in  STD_LOGIC;
    out_fifo1_full_n         : in  STD_LOGIC;
    out_fifo1_afull_n        : in  STD_LOGIC;
    out_fifo1_play_done      : in  STD_LOGIC;
    out_fifo2_empty_n        : in  STD_LOGIC;
    out_fifo2_aempty_n       : in  STD_LOGIC;
    out_fifo2_full_n         : in  STD_LOGIC;
    out_fifo2_afull_n        : in  STD_LOGIC;
    out_fifo2_play_done      : in  STD_LOGIC;
    ht_fifo_empty_n          : in  STD_LOGIC;
    ht_fifo_full_n           : in  STD_LOGIC;
    out_fifo_data_word       : out STD_LOGIC_VECTOR(45 DOWNTO 0);
    out_fifo_write_data      : out STD_LOGIC;
    out_fifo1_num_words_play : out STD_LOGIC_VECTOR(13 DOWNTO 0);
    out_fifo2_num_words_play : out STD_LOGIC_VECTOR(13 DOWNTO 0);
    out_fifo_play_data       : out STD_LOGIC;
    diag_tri_state_out_fifo1 : in  STD_LOGIC;
    diag_tri_state_out_fifo2 : in  STD_LOGIC;
    diag_tri_state_ev_header : in  STD_LOGIC;
    eventmem_status_reg      : in  STD_LOGIC_VECTOR( 7 DOWNTO 0);
    efb_misc_stts_reg        : in  STD_LOGIC_VECTOR(30 DOWNTO 0);
    ev_header_data           : in  STD_LOGIC_VECTOR(15 DOWNTO 0);
    ev_fifo_data1            : in  STD_LOGIC_VECTOR(15 DOWNTO 0);
    ev_fifo_data2            : in  STD_LOGIC_VECTOR(15 DOWNTO 0);
    mask_bcid_error          : out STD_LOGIC;
    mask_l1id_error          : out STD_LOGIC;
    mask_empty_error         : out STD_LOGIC;
    bcid_offset              : out STD_LOGIC_VECTOR( 7 DOWNTO 0);
    rod_type                 : in  STD_LOGIC;
    link_input_test_sel      : out STD_LOGIC_VECTOR( 3 DOWNTO 0);
    grp_evt_enable_out       : out STD_LOGIC;
    event_count_in           : in  STD_LOGIC_VECTOR(31 downto 0);
    bandwidth_status_in      : in  STD_LOGIC_VECTOR(31 downto 0);
    spare_pins_in            : in  STD_LOGIC_VECTOR(23 DOWNTO 0);
    dll_locked_in            : in  STD_LOGIC;
    mask_boc_clock_err_out   : out STD_LOGIC;
    mask_tim_clock_err_out   : out STD_LOGIC;
    sweeper_event_in         : in  STD_LOGIC;
    sync_event_error_in      : in  STD_LOGIC;
    l1id_fifo_data           : in  STD_LOGIC_VECTOR(31 downto 0);
    l1id_fifo_addr           : out STD_LOGIC_VECTOR( 7 downto 0);
    enable_l1_trap_out       : out STD_LOGIC;
    l1_trap_full_in          : in  STD_LOGIC;
    latched_l1id0            : in  STD_LOGIC_VECTOR(3 DOWNTO 0);
    latched_l1id1            : in  STD_LOGIC_VECTOR(3 DOWNTO 0);
    latched_bcid0            : in  STD_LOGIC_VECTOR(7 DOWNTO 0);
    latched_bcid1            : in  STD_LOGIC_VECTOR(7 DOWNTO 0);
    tim_bcid_in_rol_out      : out STD_LOGIC;
    bcid_rollover_select     : out STD_LOGIC;
    hv_flag_set_out          : out STD_LOGIC_VECTOR( 3 DOWNTO 0);
    event_threshold_limit    : out STD_LOGIC_VECTOR(15 downto 0);
    evid_fifo_count_in       : in  STD_LOGIC_VECTOR( 7 downto 0);
    evtmem_fifo_wc_stts_in   : in  STD_LOGIC_VECTOR(21 downto 0);
    diag_fifo_control      : out std_logic_vector(31 downto 0);
    diag_fifo_status       : in  std_logic_vector(31 downto 0);
    diag_fifo_trig_mask    : out std_logic_vector(31 downto 0); 
    diag_fifo_data_in      : in  std_logic_vector(31 downto 0);
    announce_ecr_enable      : out std_logic;
    l1id_err_data           : in  STD_LOGIC_VECTOR(31 downto 0);
    bcid_err_data           : in  STD_LOGIC_VECTOR(31 downto 0);
    desync_cnter_addr       : out STD_LOGIC_VECTOR( 3 downto 0);
    desync_cnter_rst        : out STD_LOGIC
    );
END efb_bus_interface ;

architecture rtl of efb_bus_interface is
  
----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

signal efb_bus_sel_n_d1           : STD_LOGIC ;
signal efb_rnw_d1                 : STD_LOGIC ;
signal efb_hwsel_d1               : STD_LOGIC ;
signal efb_hwsel_d2               : STD_LOGIC ;
signal efb_hwsel_d3               : STD_LOGIC ;
signal efb_bus_addr_d1            : STD_LOGIC_VECTOR( 7 DOWNTO 0) ;
signal efb_id_d1                  : STD_LOGIC_VECTOR(15 DOWNTO 0) ;

type state_typedef IS (
  idle, 
  strobe, 
  read, 
  write, 
  ack
  );
signal efb_access_state           : state_typedef;

signal run_number_i               : unsigned(31 DOWNTO 0) ;
signal source_id_i                : STD_LOGIC_VECTOR(31 DOWNTO 0) ;
signal format_version_i           : STD_LOGIC_VECTOR(31 DOWNTO 0) ;
signal pulse_strobe               : STD_LOGIC ;
signal xoff                       : STD_LOGIC ;

signal fifo_diag_mode_en_i        : STD_LOGIC_VECTOR(2 DOWNTO 0) ;
signal out_fifo1_sel_i            : STD_LOGIC ;
signal out_fifo2_sel_i            : STD_LOGIC ;
signal ht_fifo_sel_i              : STD_LOGIC ;
signal out_1_ev_frag_i            : STD_LOGIC ;
signal out_fifo_rst_i             : STD_LOGIC ;
signal out_fifo_data_word_i       : STD_LOGIC_VECTOR(45 DOWNTO 0) ;
signal out_fifo1_num_words_play_i : STD_LOGIC_VECTOR(13 DOWNTO 0) ;
signal out_fifo2_num_words_play_i : STD_LOGIC_VECTOR(13 DOWNTO 0) ;
signal out_fifo_play_pending      : STD_LOGIC ;
signal clr_play_pending           : STD_LOGIC ;
signal send_empty_events_i        : STD_LOGIC ;
signal bcid_offset_i              : STD_LOGIC_VECTOR( 7 DOWNTO 0) ;
signal mask_bcid_error_i          : STD_LOGIC ;
signal mask_l1id_error_i          : STD_LOGIC ;
signal mask_empty_evt_err_i       : STD_LOGIC ;
signal link_input_test_sel_i      : STD_LOGIC_VECTOR( 3 DOWNTO 0) ;
signal grp_evt_enable_i           : STD_LOGIC ;

signal err_mask_clr_cntr          : UNSIGNED(5 DOWNTO 0) ;
signal err_mask_clr_flag          : STD_LOGIC ;

signal evh_count_i                : UNSIGNED( 3 downto 0) ;
signal evt_count_i                : UNSIGNED( 2 downto 0) ;
signal out_fifo1_word_count_i     : UNSIGNED(13 DOWNTO 0) ;
signal out_fifo2_word_count_i     : UNSIGNED(13 DOWNTO 0) ;

type   state_playtype IS (
  idle, 
  header, 
  fifo1, 
  fifo2, 
  wait_for_outmem_delay, 
  trailer, 
  clr_play
  );
signal output_play_state : state_playtype;

type array31of8bitsunsigned is array (0 to 31) of unsigned(7 downto 0);
signal count_error_bit : array31of8bitsunsigned;

signal mask_boc_clock_err_i : STD_LOGIC;
signal mask_tim_clock_err_i : STD_LOGIC;
--signal sync_event_error_i   : STD_LOGIC;
signal mask_sweeper_err_i   : STD_LOGIC;
signal enable_l1_trap_i     : STD_LOGIC;
signal tim_bcid_in_rol_i    : STD_LOGIC;
signal bcid_rollover_sel_i  : STD_LOGIC;
signal hv_flag_set_i        : STD_LOGIC_VECTOR(3 DOWNTO 0);
signal event_threshold_limit_i : STD_LOGIC_VECTOR(15 DOWNTO 0);

signal diag_fifo_trig_mask_i  : std_logic_vector(31 downto 0); 
signal diag_fifo_trig_mask_o  : std_logic_vector(31 downto 0); 
signal diag_fifo_control_i    : std_logic_vector(31 downto 0);
signal diag_fifo_control_o    : std_logic_vector(31 downto 0);
signal pulse_width            : unsigned(1 downto 0);

signal announce_ecr_enable_i  : std_logic;

----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------

BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

register_data : process (
  clk, 
  rst, 
  efb_bus_sel_n, 
  efb_bus_rnw, 
  efb_bus_addr,
  efb_bus_data_in
  )
begin
  if (rst = '1') then
    efb_bus_sel_n_d1 <= '1';
    efb_hwsel_d1     <= '0';
    efb_hwsel_d2     <= '0';
    efb_hwsel_d3     <= '0';
    efb_rnw_d1       <= '1';
    efb_bus_addr_d1  <= (others => '0');
    efb_id_d1        <= (others => '0');
  elsif (clk'event and clk = '1') then
    efb_bus_sel_n_d1 <= efb_bus_sel_n;
    efb_hwsel_d1     <= efb_bus_hwsel;
    efb_hwsel_d2     <= efb_hwsel_d1;
    efb_hwsel_d3     <= efb_hwsel_d2;
    efb_rnw_d1       <= efb_bus_rnw;
    efb_bus_addr_d1  <= efb_bus_addr;
    efb_id_d1        <= efb_bus_data_in;
  end if;
end process register_data;

efb_access_fsm : process (
  clk, 
  rst, 
  efb_misc_stts_reg
  )
variable link_number  : unsigned(5 downto 0);
variable adder        : natural RANGE 0 TO 15;
constant code_version : std_logic_vector(7 downto 0) := X"60";  -- v6.0
constant board_rev    : std_logic_vector(6 downto 0) := "0001111";  -- RevF
begin
  if (rst = '1') then
    source_id_i <= (others => '0');
    run_number_i <= (others => '0');
    format_version_i <= X"03010000";  -- per ATL-D-ES-0019 v3.1
--  format_version_i <= X"04000000";  -- per ATL-D-ES-0019 v4.0 
    err_mask_sel1 <= (others => '0');
    err_mask_sel2 <= (others => '0');
    err_mask_din1 <= (others => '0');
    err_mask_din2 <= (others => '0');
    err_mask_wen1 <= '0';
    err_mask_wen2 <= '0';
    send_empty_events_i <= '0';
    mask_bcid_error_i <= '0';
    mask_l1id_error_i <= '0';
    mask_empty_evt_err_i <= '0';
    bcid_offset_i <= (others => '0');
    efb_access_state <= idle;
    efb_bus_ack <= '0';
    efb_bus_data_oe_n <= '1';
    efb_bus_data_out <= (others => '0');
    pulse_strobe <= '0';
    fifo_diag_mode_en_i <= (others => '0'); 
    out_1_ev_frag_i <= '0';
    out_fifo1_sel_i <= '0';
    out_fifo2_sel_i <= '0';
    ht_fifo_sel_i <= '0';
    out_fifo_rst_i <= '1';
    out_fifo_data_word_i <= (others => '0');
    out_fifo1_num_words_play_i <= (others => '0');
    out_fifo2_num_words_play_i <= (others => '0');
    out_fifo_play_pending <= '0';
    out_fifo_write_data <= '0';
    err_mask_clr_cntr <= "000000";
    err_mask_clr_flag <= '1';
    link_input_test_sel_i <= (others => '1');
    grp_evt_enable_i <= '0';
--    sync_event_error_i <= '0';
    mask_sweeper_err_i <= '0';
    l1id_fifo_addr <= (others => '0');
    enable_l1_trap_i <= '0';  -- '1'; -- sim
    tim_bcid_in_rol_i <= '0';
    bcid_rollover_sel_i <= '0';  --'1'  sim
    hv_flag_set_i <= (others =>'0');
    event_threshold_limit_i <= (others =>'0');
    diag_fifo_control_i    <= x"00102002"; -- trig_pos = 0x010, stretch_count = 0x20, wr_en = true, sw_trig_en = false, stretch_en = false
    diag_fifo_control_o    <= x"00102002"; -- trig_pos = 0x010, stretch_count = 0x20, wr_en = true, sw_trig_en = false, stretch_en = false
    diag_fifo_trig_mask_i  <= x"00000000";
    diag_fifo_trig_mask_o  <= x"00000000";
    pulse_width            <= "00";
    announce_ecr_enable_i  <= '0';
  elsif (clk'event and clk = '1') then
    --  Function to convert Binary to Decimal
    link_number := unsigned(efb_bus_addr_d1(5 downto 0));
    adder       := CONV_INTEGER(link_number(3 downto 0));
    if (err_mask_clr_flag = '1') then -- Following a reset, clear Error Mask LUT
      err_mask_sel1 <= std_logic_vector(err_mask_clr_cntr); 
      err_mask_sel2 <= std_logic_vector(err_mask_clr_cntr); 
      err_mask_din1 <= (others => '0');
      err_mask_din2 <= (others => '0');
      err_mask_wen1 <= '1';
      err_mask_wen2 <= '1';
      err_mask_clr_cntr <= err_mask_clr_cntr + 1;
      if (std_logic_vector(err_mask_clr_cntr) = "111111") then
        err_mask_clr_flag <= '0';
      end if;
      out_fifo_rst_i <= '1'; -- After a reset, flush Event FIFO.
    else 
      err_mask_sel1 <= efb_bus_addr_d1(5 downto 0);
      err_mask_sel2 <= efb_bus_addr_d1(5 downto 0);
--    err_mask_din1 <= efb_id_d1(13 downto 0);
--    err_mask_din2 <= efb_id_d1(13 downto 0);
      err_mask_wen1 <= '0';
      err_mask_wen2 <= '0';
    end if;

    mask_empty_evt_err_i <= '0';
    send_empty_events_i <= '0';
    out_fifo_rst_i <= '0';
    out_fifo_write_data <= '0';
    
    diag_fifo_control_o(7) <= '0';
--    diag_fifo_control_o(0) <= '0'; -- set back flush
--    diag_fifo_control_o(4) <= '0'; -- set back sw trig
    
    if (clr_play_pending = '1' OR out_fifo_rst_i = '1') then
      out_fifo_play_pending <= '0';
    end if;

    if (out_fifo_rst_i = '1') then
      out_fifo_data_word_i <= (others => '0');
      out_fifo1_num_words_play_i <= (others => '0');
      out_fifo2_num_words_play_i <= (others => '0');
    end if;
    
--  if (sweeper_event_in = '1' AND mask_sweeper_err_i = '0') then
--    run_number_i(30 downto 0) <= run_number_i(30 downto 0) + 1;  -- V1Cf JMJ
--    run_number_i(23 downto 0) <= run_number_i(23 downto 0) + 1;
--    end if;
    
--    if (sync_event_error_in = '1') then
--      sync_event_error_i <= '1';
--    end if;

    l1id_fifo_addr <= "00" & NOT efb_bus_addr_d1(5) & efb_bus_addr_d1(4 downto 0);
    desync_cnter_addr <= efb_bus_addr_d1(3 downto 0);
    desync_cnter_rst <= '0';

    case efb_access_state is
      when idle   =>
        diag_fifo_control_i(0) <= '0'; -- set back flush
        diag_fifo_control_i(4) <= '0'; -- set back sw trig
    
        diag_fifo_trig_mask_o <= diag_fifo_trig_mask_i;
--        diag_fifo_control_o   <= diag_fifo_control_i;
        diag_fifo_control_o(31 downto 8) <= diag_fifo_control_i(31 downto 8);
--        diag_fifo_control_o(7) <= '0';
        diag_fifo_control_o(6 downto 0) <= diag_fifo_control_i(6 downto 0);
        
        efb_bus_ack <= '0';
        efb_bus_data_oe_n <= '1';
        if (efb_bus_sel_n_d1 = '0') then 
          efb_access_state <= strobe;
        end if;

      when strobe =>
        efb_bus_ack <= '0';
        if (efb_rnw_d1 = '0') then
          efb_access_state <= write;
          pulse_strobe <= '1';
          efb_bus_data_oe_n <= '1';
        else
          efb_access_state <= read;
          pulse_strobe <= '0';
          efb_bus_data_oe_n <= '0';
        end if;

      when read   =>
        efb_bus_data_out  <= (others => '0');
        efb_bus_data_oe_n <= '0';
        if (efb_bus_addr_d1(7 downto 6) = "00" AND adder < 12) then
          if (efb_hwsel_d1 = '0') then
            efb_bus_data_out <= err_mask_dout1(15 downto 0);
          else
            efb_bus_data_out <= err_mask_dout1(31 downto 16);
          end if;
        elsif (efb_bus_addr_d1(7 downto 6) = "01" AND adder < 12) then
          if (efb_hwsel_d1 = '0') then
            efb_bus_data_out <= err_mask_dout2(15 downto 0);
          else
            efb_bus_data_out <= err_mask_dout2(31 downto 16);
          end if;
        elsif (efb_bus_addr_d1 >= X"A0" AND efb_bus_addr_d1 < X"D0") then
          if (efb_hwsel_d1 = '0') then
            efb_bus_data_out <= l1id_fifo_data(15 downto 0);
          else
            efb_bus_data_out <= l1id_fifo_data(31 downto 16);
          end if;
          -- L1ID error counters (addresses were reused from L1Trap range !!)
        elsif efb_bus_addr_d1(7 downto 4) = X"D" then
          if (efb_hwsel_d1 = '0') then
            efb_bus_data_out <= l1id_err_data(15 downto 0);
          else
            efb_bus_data_out <= l1id_err_data(31 downto 16);
          end if;
          -- BCID error counters
        elsif efb_bus_addr_d1(7 downto 4) = X"E" then
          if (efb_hwsel_d1 = '0') then
            efb_bus_data_out <= bcid_err_data(15 downto 0);
          else
            efb_bus_data_out <= bcid_err_data(31 downto 16);
          end if;
        else
          case efb_bus_addr_d1 is
            when X"80" =>  -- 0x00402200 Format Version Register  --
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= format_version_i(15 downto 0);
              else
                efb_bus_data_out <= format_version_i(31 downto 16);
              end if;
            when X"81" =>   -- 0x00402204 Source ID Register  --
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= source_id_i(15 downto 0);
              else
                efb_bus_data_out <= source_id_i(31 downto 16);
              end if;
            when X"82" =>   -- 0x00402208 Run Number Register  --
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= std_logic_vector(run_number_i(15 downto 0));
              else
                efb_bus_data_out <= std_logic_vector(run_number_i(31 downto 16));
              end if;
--          when X"83" =>  -- 0x0040220C TBD Register  --
--            if (efb_hwsel_d1 = '0') then
--            else
--            end if;
            when X"84" => -- 0x00402210
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out( 0)           <= mask_empty_evt_err_i;
                efb_bus_data_out( 1)           <= mask_bcid_error_i;
                efb_bus_data_out( 2)           <= grp_evt_enable_i;
                efb_bus_data_out( 3)           <= mask_l1id_error_i;
                efb_bus_data_out( 7 downto  4) <= link_input_test_sel_i;
                efb_bus_data_out( 8)           <= mask_tim_clock_err_i;
                efb_bus_data_out( 9)           <= mask_boc_clock_err_i;
                efb_bus_data_out(10)           <= mask_sweeper_err_i;
                efb_bus_data_out(11)           <= enable_l1_trap_i;
                efb_bus_data_out(12)           <= tim_bcid_in_rol_i;
                efb_bus_data_out(13)           <= bcid_rollover_sel_i;
                efb_bus_data_out(14)           <= '0';
                efb_bus_data_out(15)           <= announce_ecr_enable_i;
--                efb_bus_data_out(15 downto 14) <= (others => '0');
              else
                efb_bus_data_out( 3 downto  0) <= (others => '0');
                efb_bus_data_out( 7 downto  4) <= hv_flag_set_i;  -- (23:20)
                efb_bus_data_out(15 downto  8) <= bcid_offset_i;  -- (31:24)
              end if;
            when X"85" => -- 0x00402214
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out(15 downto 0) <= event_threshold_limit_i;
              else
                efb_bus_data_out <= (others => '0');
              end if;
            when X"86" => -- 0x00402218
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out( 0) <= NOT out_fifo1_afull_n;
                efb_bus_data_out( 1) <= efb_misc_stts_reg(0);  -- err_sumry_fifo1_almost_full
                efb_bus_data_out( 2) <= efb_misc_stts_reg(2);  -- ev_id_fifo_empty_error1
                efb_bus_data_out( 3) <= efb_misc_stts_reg(8);  -- fifo 1 pause to Formatter
                efb_bus_data_out( 4) <= NOT out_fifo2_afull_n;
                efb_bus_data_out( 5) <= efb_misc_stts_reg(1);  -- err_sumry_fifo2_almost_full
                efb_bus_data_out( 6) <= efb_misc_stts_reg(3);  -- ev_id_fifo_empty_error2
                efb_bus_data_out( 7) <= efb_misc_stts_reg(7);  -- fifo 2 pause to Formatter
                efb_bus_data_out( 8) <= efb_misc_stts_reg(9);  -- halt output from router
                efb_bus_data_out( 9) <= NOT efb_misc_stts_reg(22); -- ev_data_almost_full_n_o          
                efb_bus_data_out(10) <= efb_misc_stts_reg(18); -- header_ev_id_data(34)
                efb_bus_data_out(11) <= l1_trap_full_in;  --sync_event_error_i;
                efb_bus_data_out(12) <= '0';
                efb_bus_data_out(13) <= '0';
                efb_bus_data_out(14) <= '0';
                efb_bus_data_out(15) <= dll_locked_in;
              else
                efb_bus_data_out <= (others => '0');
                -- Moved from 0x00402214 5/26/10 JMJ
                efb_bus_data_out(0) <= NOT efb_misc_stts_reg(10); -- NOT form_enabled_n1
                efb_bus_data_out(1) <= NOT efb_misc_stts_reg(11);
                efb_bus_data_out(2) <= NOT efb_misc_stts_reg(12);
                efb_bus_data_out(3) <= NOT efb_misc_stts_reg(13);
                efb_bus_data_out(4) <= NOT efb_misc_stts_reg(14); -- NOT form_enabled_n2
                efb_bus_data_out(5) <= NOT efb_misc_stts_reg(15);
                efb_bus_data_out(6) <= NOT efb_misc_stts_reg(16);
                efb_bus_data_out(7) <= NOT efb_misc_stts_reg(17);
--                sync_event_error_i <= sync_event_error_in; -- Set new value
              end if;
            when X"87" => -- 0x0040221C
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= rod_type & board_rev & code_version;
              else
                efb_bus_data_out <= (others => '0');
              end if;
            when X"88" => -- 0x00402220
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= ev_header_data;
              else
                efb_bus_data_out <= (others => '0');
              end if;
            when X"89" => -- 0x00402224
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= ev_fifo_data1(15 downto  0);
              else
                efb_bus_data_out <= (others => '0');
              end if;
            when X"8A" => -- 0x00402228
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= ev_fifo_data2(15 downto  0);
              else
                efb_bus_data_out <= (others => '0');
              end if;
--          when X"8B" => -- 0x0040222C
            when X"8C" => -- 0x00402230 Group Event Count Register  --
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= event_count_in(15 downto 0);
              else
                efb_bus_data_out <= event_count_in(31 downto 16);
              end if;
            when X"8D" => -- 0x00402234
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= bandwidth_status_in(15 downto 0);
              else
                efb_bus_data_out <= bandwidth_status_in(31 downto 16);
              end if;
--            when X"8E" => -- 0x00402238
            when X"8F" => -- 0x0040223C
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out( 7 downto 0) <= evtmem_fifo_wc_stts_in(7 downto 0);
                efb_bus_data_out(15 downto 8) <= "00000" & evtmem_fifo_wc_stts_in(10 downto 8);
              else
                efb_bus_data_out( 7 downto 0) <= evtmem_fifo_wc_stts_in(18 downto 11);
                efb_bus_data_out(15 downto 8) <= "00000" & evtmem_fifo_wc_stts_in(21 downto 19);
              end if;
            when X"90" => -- 0x00402240
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out(0) <= fifo_diag_mode_en_i(0);  -- EFB Control
                efb_bus_data_out(1) <= fifo_diag_mode_en_i(1);  -- Eventmem Control
                efb_bus_data_out(2) <= fifo_diag_mode_en_i(2);  -- Normal Writes to FIFOs/Eventmem Reads
              else
                efb_bus_data_out <= (others => '0');
              end if;
            when X"91" => -- 0x00402244
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out( 0) <= out_fifo1_sel_i;
                efb_bus_data_out( 1) <= out_fifo2_sel_i;
                efb_bus_data_out( 2) <= ht_fifo_sel_i;
                efb_bus_data_out( 3) <= out_1_ev_frag_i;
                efb_bus_data_out( 4) <= diag_tri_state_out_fifo1;
                efb_bus_data_out( 5) <= diag_tri_state_out_fifo2;
                efb_bus_data_out( 6) <= diag_tri_state_ev_header;
              else
                efb_bus_data_out <= (others => '0');
              end if;
--          when X"92" =>  -- 0x00402248 Used for FIF Reset and Control
            when X"93" => -- 0x0040224C
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out( 0) <= NOT out_fifo1_empty_n;
                efb_bus_data_out( 1) <= NOT out_fifo1_aempty_n;
                efb_bus_data_out( 2) <= NOT out_fifo1_afull_n;
                efb_bus_data_out( 3) <= NOT out_fifo1_full_n;
                efb_bus_data_out( 4) <= NOT out_fifo2_empty_n;
                efb_bus_data_out( 5) <= NOT out_fifo2_aempty_n;
                efb_bus_data_out( 6) <= NOT out_fifo2_afull_n;
                efb_bus_data_out( 7) <= NOT out_fifo2_full_n;
                efb_bus_data_out( 8) <= ht_fifo_empty_n;
                efb_bus_data_out( 9) <= ht_fifo_full_n;
                efb_bus_data_out(10) <= efb_misc_stts_reg( 4); -- ev_id_empty1
                efb_bus_data_out(11) <= efb_misc_stts_reg(19); -- ev_id_almost_full1
                efb_bus_data_out(12) <= efb_misc_stts_reg( 5); -- ev_id_empty2
                efb_bus_data_out(13) <= efb_misc_stts_reg(20); -- ev_id_almost_full2  
                efb_bus_data_out(14) <= efb_misc_stts_reg( 6); -- header_ev_id_empty
                efb_bus_data_out(15) <= efb_misc_stts_reg(21); -- header_ev_id_almost_full
              else
                efb_bus_data_out( 0) <= efb_misc_stts_reg(23);  -- header_ev_id_full
                efb_bus_data_out( 1) <= efb_misc_stts_reg(26);  -- ev_data_empty_o
                efb_bus_data_out( 2) <= efb_misc_stts_reg(25);  -- ev_id_full1
                efb_bus_data_out( 3) <= efb_misc_stts_reg(24);  -- ev_id_full2
                efb_bus_data_out( 4) <= efb_misc_stts_reg(30);  -- count_fifo1_full
                efb_bus_data_out( 5) <= efb_misc_stts_reg(29);  -- count_fifo2_full
                efb_bus_data_out( 6) <= efb_misc_stts_reg(28);  -- err_sumry_fifo1_full
                efb_bus_data_out( 7) <= efb_misc_stts_reg(27);  -- err_sumry_fifo2_full
                efb_bus_data_out(15 downto 8) <= evid_fifo_count_in;
              end if;
            when X"94" =>  -- 0x00402250
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= out_fifo_data_word_i(15 downto 0); 
              else
                efb_bus_data_out <= (others => '0');
              end if;
            when X"95" =>  -- 0x00402254
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= out_fifo_data_word_i(31 downto 16); 
              else
                efb_bus_data_out <= (others => '0');
              end if;
            when X"96" =>  -- 0x00402258
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out(13 downto 0) <= out_fifo_data_word_i(45 downto 32);
              else
                efb_bus_data_out <= (others => '0');
              end if;
            when X"98" =>  -- 0x00402260
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out(13 downto 0) <= out_fifo1_num_words_play_i;
              else
                efb_bus_data_out <= (others => '0');
              end if;
            when X"99" =>  -- 0x00402264
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out(13 downto 0) <= out_fifo2_num_words_play_i;
              else
                efb_bus_data_out <= (others => '0');
              end if;
            when X"9A" => -- 0x00402268
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out(0)           <= out_fifo_play_pending;
                efb_bus_data_out(15 downto 8) <= NOT eventmem_status_reg;
              else
                efb_bus_data_out <= (others => '0');
              end if;
            when X"9B" =>  -- 0x0040226C
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= spare_pins_in(15 downto  0);
              else
                efb_bus_data_out <= "00000000" & spare_pins_in(23 downto 16);
              end if;
--            when X"9C" => send_empty_events_i; -- 0x00402270
            when X"9F" =>  -- 0x0040227C
              if (efb_hwsel_d1 = '0') then
                efb_bus_data_out( 3 downto 0) <= latched_l1id0;
                efb_bus_data_out( 7 downto 4) <= "0000";
                efb_bus_data_out(15 downto 8) <= latched_bcid0;
              else
                efb_bus_data_out( 3 downto 0) <= latched_l1id1;
                efb_bus_data_out( 7 downto 4) <= "0000";
                efb_bus_data_out(15 downto 8) <= latched_bcid1;
              end if;

--            when X"A0" =>  --0x00402280
--            when X"A1" =>  --0x00402284
--            when X"A2" =>  --0x00402288
--            when X"A3" =>  --0x0040228C

            when X"F0" => 
	            if (efb_hwsel_d1 = '0') then
                efb_bus_data_out(15 downto 8) <= diag_fifo_control_i(15 downto 8);
                efb_bus_data_out(7) <= '0';
                efb_bus_data_out(6 downto 0) <= diag_fifo_control_i(6 downto 0);
	            elsif (efb_hwsel_d1 = '1') then
                efb_bus_data_out <= diag_fifo_control_i(31 downto 16);
              end if;
            when X"F1" => 
	            if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= diag_fifo_status(15 downto 0);
	            elsif (efb_hwsel_d1 = '1') then
                efb_bus_data_out <= diag_fifo_status(31 downto 16);
              end if;
            when X"F2" => 
              if (pulse_width = 1) then
                diag_fifo_control_o(7) <= '1';
                pulse_width <= pulse_width + 1 ;
              elsif (pulse_width = 3) then
                diag_fifo_control_o(7) <= '0';
              else
                diag_fifo_control_o(7) <= '0';
                pulse_width <= pulse_width + 1 ;
              end if;               
	            if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= diag_fifo_data_in(15 downto 0);
	            elsif (efb_hwsel_d1 = '1') then
                efb_bus_data_out <= diag_fifo_data_in(31 downto 16);
                pulse_width      <= "00";
              end if;
            when X"F3" => 
	            if (efb_hwsel_d1 = '0') then
                efb_bus_data_out <= diag_fifo_trig_mask_i(15 downto 0);
	            elsif (efb_hwsel_d1 = '1') then
                efb_bus_data_out <= diag_fifo_trig_mask_i(31 downto 16);
              end if;
              
            when others => efb_bus_data_out <= (others => '0');
          end case;        
        end if;
        efb_bus_ack <= '0';
        if (efb_hwsel_d2 = '1') then
          efb_access_state <= ack;
        end if;
      when write  =>
        pulse_strobe <= '0';
        efb_bus_data_oe_n <= '1';
--      err_mask_wen1 <= (NOT efb_bus_addr_d1(7)) AND (NOT efb_bus_addr_d1(6)) AND NOT efb_hwsel_d1;
--      err_mask_wen2 <= (NOT efb_bus_addr_d1(7)) AND (    efb_bus_addr_d1(6)) AND NOT efb_hwsel_d1;          
        if (efb_bus_addr_d1(7) = '0' AND efb_bus_addr_d1(6) = '0' AND adder < 12) then
          if (efb_hwsel_d1 = '0') then
            err_mask_din1(15 downto  0) <= efb_id_d1;
          else
            err_mask_din1(31 downto 16) <= efb_id_d1;
          end if;
        elsif (efb_bus_addr_d1(7) = '0' AND efb_bus_addr_d1(6) = '1' AND adder < 12) then
          if (efb_hwsel_d1 = '0') then
            err_mask_din2(15 downto  0) <= efb_id_d1;
          else
            err_mask_din2(31 downto 16) <= efb_id_d1;
          end if;
        elsif efb_bus_addr_d1(7 downto 4) = X"D" or efb_bus_addr_d1(7 downto 4) = X"E" then
          desync_cnter_rst <= '1';
        else
          case efb_bus_addr_d1 is
            when X"80" =>  -- 0x00402200  Format Version Register
            --format_version_i (15 downto  0) <= efb_id_d1;
              if (efb_hwsel_d1 = '0') then
                format_version_i(15 downto 0) <= efb_id_d1;
  --  Upper Half Word of the Format Version is Read Only
  --          else
  --            format_version_i(31 downto 16) <= efb_id_d1;
              end if;
            when X"81" =>  -- 0x00402204  Source ID Register
              if (efb_hwsel_d1 = '0') then
                source_id_i(15 downto  0) <= efb_id_d1;  -- Module ID  (from Host)
              else
                source_id_i(19 downto 16) <= efb_id_d1( 3 downto 0); -- SubDetector ID
                source_id_i(31 downto 24) <= efb_id_d1(15 downto 8); -- User field
              end if;
            when X"82" =>  -- 0x00402208  Run Number Register  --
              if (efb_hwsel_d1 = '0') then
                run_number_i(15 downto 0)  <= unsigned(efb_id_d1);
              else
                run_number_i(30 downto 16) <= unsigned(efb_id_d1(14 downto 0));
              end if;
  --        when X"83" =>  NOT USED
            when X"84" => -- 0x00402210
              if (efb_hwsel_d1 = '0') then
                mask_empty_evt_err_i  <= efb_id_d1( 0);
                mask_bcid_error_i     <= efb_id_d1( 1);
                grp_evt_enable_i      <= efb_id_d1( 2);
                mask_l1id_error_i     <= efb_id_d1( 3);
                link_input_test_sel_i <= efb_id_d1( 7 downto 4);
                mask_tim_clock_err_i  <= efb_id_d1( 8);
                mask_boc_clock_err_i  <= efb_id_d1( 9);
                mask_sweeper_err_i    <= efb_id_d1(10);
                enable_l1_trap_i      <= efb_id_d1(11);
                tim_bcid_in_rol_i     <= efb_id_d1(12);
                bcid_rollover_sel_i   <= efb_id_d1(13);
                announce_ecr_enable_i <= efb_id_d1(15);
              else
                hv_flag_set_i         <= efb_id_d1( 7 downto 4);  -- (23:20)
                bcid_offset_i         <= efb_id_d1(15 downto 8);  -- (31:24)
              end if;
            when X"85" => -- 0x00402214
              if (efb_hwsel_d1 = '0') then
                event_threshold_limit_i <= efb_id_d1;
              end if;
            when X"90" => -- 0x00402240
              if (efb_hwsel_d1 = '0' AND (
                  (not (efb_id_d1(0) = '1' and efb_id_d1(1) = '1')) and
                  (not (efb_id_d1(1) = '1' and efb_id_d1(2) = '1')) and
                  (not (efb_id_d1(0) = '1' and efb_id_d1(2) = '1')))) then
                fifo_diag_mode_en_i(0) <= efb_id_d1(0);
                fifo_diag_mode_en_i(1) <= efb_id_d1(1);
                fifo_diag_mode_en_i(2) <= efb_id_d1(2);
              end if;
            when X"91" => -- 0x00402244
              if (efb_hwsel_d1 = '0') then
                if ((not (efb_id_d1(0) = '1' and efb_id_d1(1) = '1')) and
                    (not (efb_id_d1(1) = '1' and efb_id_d1(2) = '1')) and
                    (not (efb_id_d1(0) = '1' and efb_id_d1(2) = '1'))) then
                  out_fifo1_sel_i <= efb_id_d1(0);
                  out_fifo2_sel_i <= efb_id_d1(1);
                  ht_fifo_sel_i   <= efb_id_d1(2);
                end if;
                out_1_ev_frag_i <= efb_id_d1(3);
              end if;
            when X"92" => -- 0x00402248
              if (efb_hwsel_d1 = '0') then
                out_fifo_rst_i      <= pulse_strobe AND efb_id_d1(0);
              else
                send_empty_events_i <= pulse_strobe AND efb_id_d1(0);
              end if;
            when X"94" => -- 0x00402250
              if (efb_hwsel_d1 = '0') then
                out_fifo_data_word_i(15 downto  0) <= efb_id_d1; 
              end if;
            when X"95" => -- 0x00402254
              if (efb_hwsel_d1 = '0') then
                out_fifo_data_word_i(31 downto 16) <= efb_id_d1; 
              end if;
            when X"96" => -- 0x00402258
              if (efb_hwsel_d1 = '0') then
                out_fifo_data_word_i(45 downto 32) <= efb_id_d1(13 downto 0);
              end if;
            when X"97" => -- 0x0040225C
              if (efb_hwsel_d1 = '0') then
                out_fifo_write_data <= pulse_strobe;
              end if;
            when X"98" => -- 0x00402260
              if (efb_hwsel_d1 = '0') then
                out_fifo1_num_words_play_i <= efb_id_d1(13 downto 0);
              end if;
            when X"99" => -- 0x00402264
              if (efb_hwsel_d1 = '0') then
                out_fifo2_num_words_play_i <= efb_id_d1(13 downto 0);
              end if;
            when X"9A" => -- 0x00402268
              if (efb_hwsel_d1 = '0') then
                out_fifo_play_pending <= efb_id_d1(0);
              end if;
--            when X"9C" => -- 0x00402270"
        
            when X"F0" => 
	            if (efb_hwsel_d1 = '0') then
                diag_fifo_control_i(6 downto 0) <= efb_id_d1(6 downto 0);
                diag_fifo_control_i(15 downto 8) <= efb_id_d1(15 downto 8);
              elsif (efb_hwsel_d1 = '1') then
                diag_fifo_control_i(31 downto 16) <= efb_id_d1;
              end if;
    --        when X"F1" =>  -- Diag FIFO Status 
    --        when X"F2" =>  -- FIFO Data Mail Box
            when X"F3" => 
	            if (efb_hwsel_d1 = '0' ) then
                diag_fifo_trig_mask_i(15 downto 0) <= efb_id_d1;
              elsif (efb_hwsel_d1 = '1' ) then
                diag_fifo_trig_mask_i(31 downto 16) <= efb_id_d1;
              end if;
            when others =>
          end case;
        end if;
        efb_bus_ack <= '0';
        if (efb_hwsel_d3 = '1') then
          efb_access_state <= ack;
          err_mask_wen1 <= (NOT efb_bus_addr_d1(7)) AND (NOT efb_bus_addr_d1(6));
          err_mask_wen2 <= (NOT efb_bus_addr_d1(7)) AND (    efb_bus_addr_d1(6));          
        end if;

      when ack =>
        efb_bus_ack <= '1';
        if (efb_bus_sel_n_d1 = '1') then
          efb_bus_data_oe_n <= '1';
          efb_bus_ack <= '0';
          efb_access_state <= idle;
        end if;
      when others =>
        efb_access_state <= idle;
         efb_bus_ack <= '1';
    end case;
-- Source ID ROD Specific Value
    source_id_i(23 downto 20) <= X"2";  -- SCT   0x2x
--  source_id_i(31 downto 24) <= (others => '0');  -- Reserved User Field in Ver4.0
  end if;
end process efb_access_fsm;
  
diag_fifo_trig_mask <= diag_fifo_trig_mask_o;
diag_fifo_control   <= diag_fifo_control_o;

source_id      <= source_id_i;
format_version <= format_version_i;
run_number     <= std_logic_vector(run_number_i);

out_fifo_diag_mode       <= fifo_diag_mode_en_i;
out_fifo_data_word       <= out_fifo_data_word_i;
out_fifo1_num_words_play <= out_fifo1_num_words_play_i;
out_fifo2_num_words_play <= out_fifo2_num_words_play_i;

out_fifo_rst        <= out_fifo_rst_i;
send_empty_events   <= send_empty_events_i;
bcid_offset         <= bcid_offset_i;
mask_bcid_error     <= mask_bcid_error_i;
mask_l1id_error     <= mask_l1id_error_i;
mask_empty_error    <= mask_empty_evt_err_i;
link_input_test_sel <= link_input_test_sel_i;
grp_evt_enable_out  <= grp_evt_enable_i;

mask_boc_clock_err_out <= mask_boc_clock_err_i;
mask_tim_clock_err_out <= mask_tim_clock_err_i;

enable_l1_trap_out <= enable_l1_trap_i;
tim_bcid_in_rol_out <= tim_bcid_in_rol_i;
bcid_rollover_select <= bcid_rollover_sel_i;
hv_flag_set_out <= hv_flag_set_i;
announce_ecr_enable <= announce_ecr_enable_i;

event_threshold_limit <= event_threshold_limit_i;

diag_play_mode_mux : process (
  rst, 
  clk,
  fifo_diag_mode_en_i,
  out_fifo1_play_done,
  out_fifo2_play_done,
  out_fifo_play_pending,
  out_fifo1_num_words_play_i,
  out_fifo2_num_words_play_i,
  output_play_state
  )
begin
  if (rst = '1') then
    xoff                   <= '0';
    out_fifo1_diag_en      <= '0';
    out_fifo2_diag_en      <= '0';
    ht_fifo_diag_en        <= '0';
    out_fifo_play_data     <= '0';
    clr_play_pending       <= '1';
    evh_count_i            <= (others => '0');
    evt_count_i            <= (others => '0');
    out_fifo1_word_count_i <= (others => '0');
    out_fifo2_word_count_i <= (others => '0');
    output_play_state      <= idle;

  elsif (clk'event AND clk ='1') then
    xoff <= efb_misc_stts_reg(9);
    case output_play_state is
      when idle =>
-- Logic required to mux data to all EFB controlled FIFOs for write operations.
-- Write to the FIFOs is allowed in idle state only.
        out_fifo1_diag_en <= (fifo_diag_mode_en_i(0) OR fifo_diag_mode_en_i(1)) 
                              AND out_fifo1_sel_i;
        out_fifo2_diag_en <= (fifo_diag_mode_en_i(0) OR fifo_diag_mode_en_i(1)) 
                              AND out_fifo2_sel_i;
        ht_fifo_diag_en   <= (fifo_diag_mode_en_i(0) OR fifo_diag_mode_en_i(1)) 
                              AND ht_fifo_sel_i;
-- No play allowed here
        out_fifo_play_data <= '0';
        clr_play_pending   <= '0';
-- Clear all Play counters
        evh_count_i            <= (others => '0');
        evt_count_i            <= (others => '0');
        out_fifo1_word_count_i <= (others => '0');
        out_fifo2_word_count_i <= (others => '0');      
-- If the play bit is set in diag mode 1, start playing the data.
        if (fifo_diag_mode_en_i = "001" AND out_fifo_play_pending = '1' AND xoff = '0') then
          output_play_state <= header;
        end if;
        
--  Start Header-Trailer FIFO, and count out the first 9 words (H0 -> H8)
      when header =>
       if (xoff = '0') then
        ht_fifo_diag_en        <= '1';
        out_fifo1_diag_en      <= '0';
        out_fifo2_diag_en      <= '0';
        out_fifo_play_data     <= '1';
        clr_play_pending       <= '0';
        evh_count_i            <= evh_count_i + '1';
        evt_count_i            <= evt_count_i;
        out_fifo1_word_count_i <= out_fifo1_word_count_i;
        out_fifo2_word_count_i <= out_fifo2_word_count_i;
        if (std_logic_vector(evh_count_i) = "1001") then
          ht_fifo_diag_en    <= '0';
          out_fifo1_diag_en  <= '0';
          out_fifo2_diag_en  <= '0';
          out_fifo_play_data <= '0';
          clr_play_pending   <= '0';
          output_play_state  <= fifo1;
        end if;
       else
        out_fifo_play_data <= '0';
       end if;

--  Start playing Eventmem A
      when fifo1 =>
       if (xoff = '0') then
        ht_fifo_diag_en        <= '0';
        out_fifo1_diag_en      <= '1';
        out_fifo2_diag_en      <= '0';
        out_fifo_play_data     <= '1';
        clr_play_pending       <= '0';
        evh_count_i            <= evh_count_i;
        evt_count_i            <= evt_count_i;
        out_fifo1_word_count_i <= out_fifo1_word_count_i + 1;
        out_fifo2_word_count_i <= out_fifo2_word_count_i;
        if (std_logic_vector(out_fifo1_word_count_i) = out_fifo1_num_words_play_i) then 
          ht_fifo_diag_en    <= '0';
          out_fifo1_diag_en  <= '0';
          out_fifo2_diag_en  <= '0';
          out_fifo_play_data <= '0';
          clr_play_pending   <= '0';
          output_play_state  <= fifo2;
        end if;
       else
        out_fifo_play_data <= '0';
       end if;

--  Start playing Eventmem B
      when fifo2 =>
       if (xoff = '0') then
        ht_fifo_diag_en        <= '0';
        out_fifo1_diag_en      <= '0';
        out_fifo2_diag_en      <= '1';
        out_fifo_play_data     <= '1';
        clr_play_pending       <= '0';
        evh_count_i            <= evh_count_i;
        evt_count_i            <= evt_count_i;
        out_fifo1_word_count_i <= out_fifo1_word_count_i;
        out_fifo2_word_count_i <= out_fifo2_word_count_i + 1;
        if (std_logic_vector(out_fifo2_word_count_i) = out_fifo2_num_words_play_i) then 
          ht_fifo_diag_en    <= '0';
          out_fifo1_diag_en  <= '0';
          out_fifo2_diag_en  <= '0';
          out_fifo_play_data <= '0';
          clr_play_pending   <= '0';
          output_play_state  <= wait_for_outmem_delay;
        end if;
       else
        out_fifo_play_data <= '0';
       end if;

      when wait_for_outmem_delay =>
        ht_fifo_diag_en        <= '0';
        out_fifo1_diag_en      <= '0';
        out_fifo2_diag_en      <= '0';
        out_fifo_play_data     <= '0';
        clr_play_pending       <= '0';
        evh_count_i            <= evh_count_i;
        evt_count_i            <= evt_count_i;
        out_fifo1_word_count_i <= out_fifo1_word_count_i;
        out_fifo2_word_count_i <= out_fifo2_word_count_i;
        if (out_fifo2_play_done = '1') then
          output_play_state <= trailer;
        end if;
        
--  Start Header-Trailer FIFO, and play out the trailer words (T0 -> T5)
      when trailer =>
       if (xoff = '0') then
        ht_fifo_diag_en        <= '1';
        out_fifo1_diag_en      <= '0';
        out_fifo2_diag_en      <= '0';
        out_fifo_play_data     <= '1';
        clr_play_pending       <= '0';
        evh_count_i            <= evh_count_i;
        evt_count_i            <= evt_count_i + 1;
        out_fifo1_word_count_i <= out_fifo1_word_count_i;
        out_fifo2_word_count_i <= out_fifo2_word_count_i;
        if (std_logic_vector(evt_count_i) = "110") then
          ht_fifo_diag_en    <= '0';
          out_fifo1_diag_en  <= '0';
          out_fifo2_diag_en  <= '0';
          out_fifo_play_data <= '0';
          clr_play_pending   <= '1';
          output_play_state  <= clr_play;
        end if;
       else
        out_fifo_play_data <= '0';
       end if;

--  Clear the Play pending bit
      when clr_play =>
        ht_fifo_diag_en        <= '0';
        out_fifo1_diag_en      <= '0';
        out_fifo2_diag_en      <= '0';
        out_fifo_play_data     <= '0';
        clr_play_pending       <= '1';
        evh_count_i            <= (others => '0');
        evt_count_i            <= (others => '0');
        out_fifo1_word_count_i <= (others => '0');
        out_fifo2_word_count_i <= (others => '0');
        output_play_state      <= idle;
    end case;
  end if;
end process diag_play_mode_mux;

END rtl;

