----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    err_detect.vhd
-- Description:
--    ROD Dual Gatherer
--    Scans for errors in the data stream, contains maskable error
--      counters.
--
----------------------------------------------------------------------------
--  History
--    07-06-2005 : Organized Error Map to match Format 3.0 Document
--
----------------------------------------------------------------------------
--
--  Error Flags used in the Generic Field of the first status element
--  Common to SCT and Pixel Data formats
--  Bit 0 = 0x0001 BC ID Error
--  Bit 1 = 0x0002 EL1 ID Error
--  Bit 2 = 0x0004 FE Module Timeout
--  Bit 3 = 0x0008 Data may be incorrect -> see Bits[31:16]
--  Bit 4 = 0x0010 Internal Buffer Overflow -> See Bits[17:16]
--
--  SCT Version Error Bits
--  Bit 16 = 0x0001 Almost full error, data truncated (HT Limit)
--  Bit 17 = 0x0002 Data overflow (ROD Busy) 
--  Bit 18 = 0x0004 Header bit error
--  Bit 19 = 0x0008 Sync error
--  Bit 20 = 0x0010 FE Flagged error
--  Bit 21 = 0x0020 Condensed mode hit pattern error
--  Bit 22 = 0x0040 Non-sequential FE chip number
--  Bit 23 = 0x0080 Invalid FE chip number (6,7,14,15)
--  Bit 24 = 0x0100 Trailer bit error
--
-- err_sumry_fifo_data(14: 0) <= err_sumry_count(14:0)
-- err_sumry_fifo_data(47:16) <= err_sumry_flags(31:0)
--
----------------------------------------------------------------------------

library IEEE ;
use IEEE.STD_LOGIC_1164.all ;
use IEEE.std_logic_arith.all;

-- pragma translate_off
library UNISIM;
use unisim.all;
-- pragma translate_on

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

ENTITY err_detect IS
  port (
    clk                 : in  STD_LOGIC ;
    rst                 : in  STD_LOGIC ;
    data_in             : in  STD_LOGIC_VECTOR(46 DOWNTO 0) ;
    data_valid_in       : in  STD_LOGIC ;
    data_out            : out STD_LOGIC_VECTOR(45 DOWNTO 0) ;
    data_valid_out      : out STD_LOGIC ;
    err_sumry_fifo_wen  : out STD_LOGIC ;
    err_sumry_fifo_data : out STD_LOGIC_VECTOR(47 DOWNTO 0) ;
    err_mask_wen        : in  STD_LOGIC ;
    err_mask_din        : in  STD_LOGIC_VECTOR(31 DOWNTO 0) ;
    err_mask_dout       : out STD_LOGIC_VECTOR(31 DOWNTO 0) ;
    err_mask_sel        : in  STD_LOGIC_VECTOR( 5 DOWNTO 0) ;
    mask_bcid_error     : in  STD_LOGIC ;
    mask_l1id_error     : in  STD_LOGIC
    );
END err_detect ;

architecture rtl of err_detect is
  
----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

signal pwr  : STD_LOGIC ;
signal gnd  : STD_LOGIC ;
signal gndv : STD_LOGIC_VECTOR(15 downto 0);

signal mem_rdout : STD_LOGIC_VECTOR(31 downto 0);
signal mem_wdout : STD_LOGIC_VECTOR(31 downto 0);
signal mem_raddr : STD_LOGIC_VECTOR( 7 downto 0);
signal mem_waddr : STD_LOGIC_VECTOR( 7 downto 0);
signal mem_wdin  : STD_LOGIC_VECTOR(31 downto 0);
signal mem_wren  : STD_LOGIC;

signal err_sumry_flags : STD_LOGIC_VECTOR(31 downto 0);
signal err_sumry_count : UNSIGNED(14 downto 0);
signal inc_err_count0  : STD_LOGIC;
signal inc_err_count1  : STD_LOGIC;
signal link_num        : STD_LOGIC_VECTOR(5 downto 0);

signal err_mask_dout_i      : STD_LOGIC_VECTOR(31 DOWNTO 0);
signal err_sumry_mask       : STD_LOGIC_VECTOR(31 DOWNTO 0);
signal err_sumry_word       : STD_LOGIC_VECTOR(31 DOWNTO 0);
signal last_word            : STD_LOGIC;
signal last_word_d1         : STD_LOGIC;
signal err_sumry_word_valid : STD_LOGIC;

----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------

component RAMB4_S16_S16
  port (
    DOA    : out STD_LOGIC_VECTOR(15 downto 0);
    DOB    : out STD_LOGIC_VECTOR(15 downto 0);
    ADDRA  : in  STD_LOGIC_VECTOR( 7 downto 0);
    DIA    : in  STD_LOGIC_VECTOR(15 downto 0);
    ENA    : in  STD_ULOGIC;
    CLKA   : in  STD_ULOGIC;
    WEA    : in  STD_ULOGIC;                
    RSTA   : in  STD_ULOGIC;
    ADDRB  : in  STD_LOGIC_VECTOR( 7 downto 0);
    DIB    : in  STD_LOGIC_VECTOR(15 downto 0);
    ENB    : in  STD_ULOGIC;
    CLKB   : in  STD_ULOGIC;
    WEB    : in  STD_ULOGIC;
    RSTB   : in  STD_ULOGIC
    ); 
end component;


BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

-- generate 256 x 32 dual port memory to store error masks
MB0 : RAMB4_S16_S16 -- Block[15..0] of 16bits
  port map (
    DOA   => mem_wdout(15 downto 0),
    DOB   => mem_rdout(15 downto 0),
    ADDRA => mem_waddr,
    DIA   => mem_wdin(15 downto 0),
    ENA   => pwr,
    CLKA  => clk,
    WEA   => mem_wren,
    RSTA  => rst,
    ADDRB => mem_raddr,
    DIB   => gndv,
    ENB   => pwr,
    CLKB  => clk,
    WEB   => gnd,
    RSTB  => rst
    ); 

MB1 : RAMB4_S16_S16 -- Block[31..16] of 16bits
  port map (
    DOA   => mem_wdout(31 downto 16),
    DOB   => mem_rdout(31 downto 16),
    ADDRA => mem_waddr,
    DIA   => mem_wdin(31 downto 16),
    ENA   => pwr,
    CLKA  => clk,
    WEA   => mem_wren,
    RSTA  => rst,
    ADDRB => mem_raddr,
    DIB   => gndv,
    ENB   => pwr,
    CLKB  => clk,
    WEB   => gnd,
    RSTB  => rst
    ); 

pwr  <= '1';
gnd  <= '0';
gndv <= (others => '0');

mem_raddr <= "00" & link_num(5 downto 0);
mem_waddr <= "00" & err_mask_sel(5 downto 0);

mem_wdin        <= err_mask_din;
mem_wren        <= err_mask_wen;
err_mask_dout   <= mem_wdout;
err_mask_dout_i <= mem_rdout(31 downto 0);

-- Begin main body of error processing logic

link_num <= data_in (37 downto 32);

register_data : process (clk, rst)
  begin -- register
    if (rst = '1') then
      data_out             <= (others => '0');
      data_valid_out       <= '0';
      last_word            <= '0';
      last_word_d1         <= '0';
      err_sumry_word_valid <= '0';
      err_sumry_mask       <= (others => '0');
    elsif (clk'event and clk = '1') then
      data_out             <= data_in(45 downto 0);
      data_valid_out       <= data_valid_in;
      last_word            <= data_in(45);
      last_word_d1         <= last_word;
      err_sumry_word_valid <= data_valid_in;
      err_sumry_mask       <= err_mask_dout_i;
    end if;
  end process register_data;

decode_errors : process (clk, rst)
begin -- register
  if (rst = '1') then
    err_sumry_word <= (others => '0');
    inc_err_count0 <= '0';
    inc_err_count1 <= '0';
  elsif (clk'event and clk = '1') then
    err_sumry_word <= (others => '0');
    inc_err_count0 <= '0';
    inc_err_count1 <= '0';

    -- Bit[0] : check BC ID
    if (data_in(42) = '1' AND mask_bcid_error = '0') then
      err_sumry_word(0) <= '1';
      inc_err_count0 <= '1';
    end if;

    -- Bit[1] : check L1 ID
    if (data_in(41) = '1' AND mask_l1id_error = '0') then
      err_sumry_word(1) <= '1';
      inc_err_count0 <= '1';
    end if;

    -- Bit[2] : check timeout
    if (data_in(39) = '1') then
      err_sumry_word(2) <= '1';
      inc_err_count0 <= '1';
    end if;

    -- Bit[3] : Data may contain errors, Check bits[16:31]

    -- Bit[4]  : Internal Buffer overflow error, Check bits[16:17]
    -- Bit[16] : HT Limit (Data Truncated)
	  -- Bit[17] : ROD Busy Limit (Data Overflow)
    if (data_in(31 downto 29) = "010") then   -- SCT/Pixel
    --  check almost full error, data truncated
      err_sumry_word(16) <= data_in(27);
    -- check overflow
      err_sumry_word(17) <= data_in(26);
      inc_err_count0 <= data_in(27) OR data_in(26);  --'1';
    elsif (data_in(15 downto 13) = "010") then  -- SCT
    --  check almost full error, data truncated
      err_sumry_word(16) <= data_in(11);                    
      -- check overflow
      err_sumry_word(17) <= data_in(10);
      inc_err_count1 <= data_in(11) OR data_in(10);  --'1';
    end if;

    -- Bit[18] : check header bit error
    if (data_in(31 downto 28) = "0011") then  -- SCT/Pixel
      err_sumry_word(18) <= '1';
      inc_err_count0 <= '1';
    end if;

    -- Bit[19] : check sync error
    if (data_in(31 downto 29) = "011") then  -- SCT/Pixel
      err_sumry_word(19) <= '1';
      inc_err_count0 <= '1';
    end if;
    if (data_in(31 downto 29) /= "010"  AND 
        data_in(15 downto 13)  = "011") then        
      err_sumry_word(19) <= '1';
      inc_err_count1 <= '1';
    end if;

    -- Bit[20] : SCT == check FE flagged error
	  --          PXL == invalid Row or Column
    if (data_in(31 downto 29)  = "000" AND 
        data_in(18 downto 16) /= "000") then
      err_sumry_word(20) <= '1';
      inc_err_count0 <= '1';
    end if;
    if (data_in(31 downto 29) /= "010" AND 
        data_in(15 downto 13)  = "000" AND 
        data_in( 2 downto  0) /= "000") then        
      err_sumry_word(20) <= '1';
      inc_err_count1 <= '1';
    end if;

    -- Bit[21] : SCT == check hit pattern error
	 --           PXL == check for skipped events
    if (data_in(40) = '1' AND data_in(31) = '1') then
      err_sumry_word(21) <= data_in(18) OR (data_in(16) AND data_in(19));
      inc_err_count0 <= data_in(18) OR (data_in(16) AND data_in(19));  --'1';
    end if;
    if (data_in(31 downto 29) /= "010" AND data_in(40)='1' AND data_in(15)='1') then
      err_sumry_word(21) <= data_in(2) OR (data_in(0) AND data_in(3));
      inc_err_count1 <= data_in(2) OR (data_in(0) AND data_in(3)); --'1';
    end if;

    -- Bit[22] : SCT == Non-Sequential FE chip number
    if (data_in(46) = '1') then -- Detection engine in bc_l1_check.vhd
      err_sumry_word(22) <= '1';
      inc_err_count0 <= '1';
    end if;

    -- Bit[23] : SCT == Invalid SCT FE Chips	(6,7,14,15)
    if (data_in(31) = '1' AND data_in(29 downto 28) = "11") then
      err_sumry_word(23) <= '1';
       inc_err_count0 <= '1';
    end if;

    -- Bit[24] : SCT == Trailer Bit Error
    if (data_in(31 downto 29) = "010") then   -- SCT/Pixel
      err_sumry_word(24) <= data_in(28);
      inc_err_count0 <= data_in(28);  --'1';
    end if;

    -- Bit[25] : SCT == MDSP Link Mask Enable
    if (data_in(43) = '1' AND data_in(31 downto 29) = "001") then
      err_sumry_word(25) <= '1';
      inc_err_count0 <= '1';
    end if;
  end if;
end process decode_errors;


process_errors : process (clk, rst)
begin -- register
  if (rst = '1') then
    err_sumry_flags <= (others => '0');
    err_sumry_count <= (others => '0');
    err_sumry_fifo_wen <= '0';
  elsif (clk'event and clk = '1') then  
    err_sumry_fifo_wen <= last_word;
    if (last_word_d1 = '1') then
      err_sumry_flags <= (others => '0');
      err_sumry_count <= (others => '0');
    elsif (err_sumry_word_valid = '1') then
      for i in 0 to 2 loop
        err_sumry_flags(i) <= err_sumry_flags(i) OR (err_sumry_word(i) AND NOT err_sumry_mask(i));
      end loop;
      err_sumry_flags(3) <= err_sumry_flags(3) OR (((err_sumry_word(18) AND NOT err_sumry_mask(18))  OR 
                                                    (err_sumry_word(19) AND NOT err_sumry_mask(19))  OR
                                                    (err_sumry_word(20) AND NOT err_sumry_mask(20))  OR
                                                    (err_sumry_word(21) AND NOT err_sumry_mask(21))  OR
                                                    (err_sumry_word(22) AND NOT err_sumry_mask(22))  OR
                                                    (err_sumry_word(23) AND NOT err_sumry_mask(23))  OR
                                                    (err_sumry_word(24) AND NOT err_sumry_mask(24))  OR
                                                    (err_sumry_word(25) AND NOT err_sumry_mask(25))  OR
                                                    (err_sumry_word(26) AND NOT err_sumry_mask(26))  OR
                                                    (err_sumry_word(27) AND NOT err_sumry_mask(27))  OR
                                                    (err_sumry_word(28) AND NOT err_sumry_mask(28))  OR
                                                    (err_sumry_word(29) AND NOT err_sumry_mask(29))  OR
                                                    (err_sumry_word(30) AND NOT err_sumry_mask(30))  OR
                                                    (err_sumry_word(31) AND NOT err_sumry_mask(31))) AND 
                                                     NOT err_sumry_mask(3));
      err_sumry_flags(4) <= err_sumry_flags(4) OR (((err_sumry_word(16) AND NOT err_sumry_mask(16))  OR 
                                                    (err_sumry_word(17) AND NOT err_sumry_mask(17))) AND 
                                                     NOT err_sumry_mask(4));
      for i in 16 to 31 loop
        err_sumry_flags(i) <= err_sumry_flags(i) OR (err_sumry_word(i) AND NOT err_sumry_mask(i));
      end loop;
      if (((err_sumry_word( 2 downto  0) AND NOT err_sumry_mask( 2 downto  0)) /= "000") OR
          ((err_sumry_word(31 downto 16) AND NOT err_sumry_mask(31 downto 16)) /= X"0000")) then 
        if (std_logic_vector(err_sumry_count) < "111111111111111") then
          err_sumry_count <= err_sumry_count + inc_err_count0 + inc_err_count1;
        end if;
      end if;
    end if;
  end if;
end process process_errors;

err_sumry_fifo_data <= err_sumry_flags & '0' & std_logic_vector(err_sumry_count);

END rtl;

