----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    ev_data_decode.vhd
-- Description:
--    ROD Dual Gatherer
--    SCT Version
--    Reads Event FIFO data from the external FIFO and writes the 8 bits
--    BCID and 4 bit L1 ID data into the Event ID FIFOs of each gatherer.
--    The full L1, BC ID, and Trig type info are sent to the logic that
--    generates the event fragment header/trailer.
--
--    A full event of data must be read out over 17 words from the
--    external FIFO since there are so many bits of data.
--    The 4K FIFO can buffer the data from 240 Events
--
--  word 0  : L1 ID[15:0]                            ev_data_sh_reg( 15: 0)
--  word 1  : ECR ID[7:0] & L1 ID[23:16]             ev_data_sh_reg( 31:16)
--  word 2  : BC ID[11:0] & RoL & BOC OK & TIM OK    ev_data_sh_reg( 47:32)
--  word 3  : TT - ROD[5:0] & TIM[1:0] & ATLAS[7:0]  ev_data_sh_reg( 63:48)
--  word 4  : Dynamic Mask for Links [ 7: 0]         ev_data_sh_reg(159:64)
--  word 5  : Dynamic Mask for Links [15: 8]
--  word 6  : Dynamic Mask for Links [23:16]
--  word 7  : Dynamic Mask for Links [31:24]
--  word 8  : Dynamic Mask for Links [39:32]
--  word 9  : Dynamic Mask for Links [47:40]
--  word 10 : Dynamic Mask for Links [55:48]         ev_data_sh_reg(255:160)
--  word 11 : Dynamic Mask for Links [63:56]
--  word 12 : Dynamic Mask for Links [71:64]
--  word 13 : Dynamic Mask for Links [79:72]
--  word 14 : Dynamic Mask for Links [87:80]
--  word 15 : Dynamic Mask for Links [95:88]
--
--header_ev_id_data(79 DOWNTO 0) <= ev_data_sh_reg (63 DOWNTO 0); --Header ID Info
--
--ev_id_data1(111 DOWNTO 0) <=  ev_data_sh_reg(159 DOWNTO 64)   -- DynMask0
--                            & ev_data_sh_reg( 61 DOWNTO 58)   -- ROD Trigger Type (3:0)
--                            & ev_data_sh_reg(  3 DOWNTO  0)   -- L1ID(3:0)
--                            & ev_data_sh_reg( 43 DOWNTO 36);  -- BCID
--
--ev_id_data2(111 DOWNTO 0) <=  ev_data_sh_reg(255 DOWNTO 159)  -- DynMask0
--                            & ev_data_sh_reg( 61 DOWNTO  58)  -- ROD Trigger Type (3:0)
--                            & ev_data_sh_reg(  3 DOWNTO   0)  -- L1ID(3:0)
--                            & ev_data_sh_reg( 43 DOWNTO  36); -- BCID
--
----------------------------------------------------------------------------

library IEEE ;
use IEEE.STD_LOGIC_1164.all ;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

ENTITY ev_data_decode IS
  port (
    clk                   : in  STD_LOGIC ;
    rst                   : in  STD_LOGIC ;
    ev_data_in            : in  STD_LOGIC_VECTOR(15 DOWNTO 0) ;
    ev_data_wen_n         : in  STD_LOGIC ;
    ev_data_almost_full_n : out STD_LOGIC ;
    ev_data_empty         : out STD_LOGIC ;
    header_ev_id_data     : out STD_LOGIC_VECTOR(63 DOWNTO 0) ;
    header_ev_id_wen      : out STD_LOGIC ;
    ev_id_data1           : out STD_LOGIC_VECTOR(111 DOWNTO 0) ;
    ev_id_wen1            : out STD_LOGIC ;
    ev_id_data2           : out STD_LOGIC_VECTOR(111 DOWNTO 0) ;
    ev_id_wen2            : out STD_LOGIC ;      
    ev_fifo1_af           : in  STD_LOGIC ;
    ev_fifo2_af           : in  STD_LOGIC ;
    ev_fifo3_af           : in  STD_LOGIC ;
    bcid_offset           : in  STD_LOGIC_VECTOR(7 DOWNTO 0) ;
    bcid_rollover_select  : in  STD_LOGIC
    );
END ev_data_decode ;

architecture rtl of ev_data_decode is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  type ev_id_state_typedef IS (
    idle, 
    read, 
    check_space, 
    write_data
    );
  signal ev_id_state         : EV_ID_STATE_TYPEDEF ;

  signal ev_data_empty_i     : STD_LOGIC ;
  signal repeat_counter      : UNSIGNED(4 DOWNTO 0) ; 
  signal ev_data_sh_reg      : STD_LOGIC_VECTOR(255 DOWNTO 0) ;
  signal ev_data_in_d1       : STD_LOGIC_VECTOR( 15 DOWNTO 0) ;
  signal ev_data_out         : STD_LOGIC_VECTOR( 15 DOWNTO 0) ;
  signal ev_data_ren         : STD_LOGIC ;
  signal fifocount_out       : STD_LOGIC_VECTOR(  9 DOWNTO 0) ;
  signal ev_data_available   : STD_LOGIC ;
  signal ev_data_almost_full : STD_LOGIC ;
  signal ev_data_wen         : STD_LOGIC ;
  signal bcid_value          : STD_LOGIC_VECTOR(11 DOWNTO 0) ;
  signal bcid_comp_value     : STD_LOGIC_VECTOR(11 DOWNTO 0) ;
  signal bcid_offset_i       : STD_LOGIC_VECTOR(11 DOWNTO 0) ;
  
  
----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------

component fifo1024x16
  port (
    clock_in        : IN  std_logic;
    read_enable_in  : IN  std_logic;
    write_enable_in : IN  std_logic;
    write_data_in   : IN  std_logic_vector(15 downto 0);
    fifo_gsr_in     : IN  std_logic;
    read_data_out   : OUT std_logic_vector(15 downto 0);
    full_out        : OUT std_logic;
    empty_out       : OUT std_logic;
    fifocount_out   : OUT std_logic_vector( 9 downto 0)
    );
END component;

component fifo_rambs16
  generic (
    width : integer := 1 
    );
  port (
    clk_in      : in  STD_LOGIC;
    rst_in      : in  STD_LOGIC;
    data_in     : in  STD_LOGIC_VECTOR(16*(width-1)+15 DOWNTO 0);
    data_out    : out STD_LOGIC_VECTOR(16*(width-1)+15 DOWNTO 0);
    ren         : in  STD_LOGIC;
    wen         : in  STD_LOGIC;
    full        : out STD_LOGIC;
    empty       : out STD_LOGIC;
    almost_full : out STD_LOGIC 
    );
END component ;

BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

event_fifo_comp: fifo1024x16 
  port map (
    clock_in        => clk,
    read_enable_in  => ev_data_ren,
    write_enable_in => ev_data_wen,
    write_data_in   => ev_data_in_d1,
    fifo_gsr_in     => rst,
    read_data_out   => ev_data_out,
    full_out        => open,
    empty_out       => ev_data_empty_i,
    fifocount_out   => fifocount_out
    );

check_for_event : process (fifocount_out)
begin
  if (fifocount_out > "0000001111") then  -- at least one event (16W) in the FIFO
    ev_data_available <= '1';
  else
    ev_data_available <= '0';
  end if;
end process check_for_event;

register_data : process (clk, rst)
begin
  if (rst = '1') then
    ev_data_wen <= '0';
    ev_data_almost_full <= '0';
    ev_data_almost_full_n <= '1';
    ev_data_in_d1 <= (others => '0');
    ev_data_empty <= '1';
  elsif (clk'event AND clk = '1') then
                                            -- 64 events at 16 w/evt = 1024 words/almost full
    if (fifocount_out > "1110000000") then  -- 56 events at 16 w/evt = 896 words/almost full
      ev_data_almost_full <= '1';
    else
      ev_data_almost_full <= '0';
    end if;
    ev_data_almost_full_n <= NOT ev_data_almost_full;
    ev_data_in_d1 <= ev_data_in;
    ev_data_wen <= NOT ev_data_wen_n;
    ev_data_empty <= ev_data_empty_i;
  end if;
end process register_data;
  
read_ev_data : process (clk, rst)
begin -- read_ev_data
  if (rst = '1') then
    ev_id_state <= idle;
    ev_data_ren <= '0';
    header_ev_id_wen <= '0';
    ev_id_wen1 <= '0';
    ev_id_wen2 <= '0';
    repeat_counter <= (others => '0');
    ev_data_sh_reg <= (others => '0');
    bcid_value <= (others => '1');
--    bcid_comp_value <= (others => '1');
  elsif (clk'event AND clk = '1') then
    case ev_id_state is
      when idle =>
        ev_data_ren <= '0';
        header_ev_id_wen <= '0';
        ev_id_wen1 <= '0';
        ev_id_wen2 <= '0';
        repeat_counter <= (others => '0');
        if (ev_data_available = '1') then
          ev_data_ren <= '1';
          ev_id_state <= read;
        end if;
      when read =>
        ev_data_sh_reg(239 downto 0) <= ev_data_out & ev_data_sh_reg(239 downto 16);
        if (repeat_counter = 3) then
          bcid_value <= ev_data_out(15 downto 4);
        end if;
        if (repeat_counter = 15) then
          ev_id_state <= check_space;
          ev_data_ren <= '0';
        end if;
        repeat_counter <= repeat_counter + 1;
      when check_space =>
        ev_data_sh_reg(255 downto 0) <= ev_data_out & ev_data_sh_reg(239 downto 0);
        if (ev_fifo1_af = '0' AND ev_fifo2_af = '0' AND ev_fifo3_af = '0') then
          ev_id_state <= write_data;
          header_ev_id_wen <= '1';
          ev_id_wen1  <= '1';
          ev_id_wen2  <= '1';
        end if;
      when write_data =>
        header_ev_id_wen <= '0';
        ev_id_wen1  <= '0';
        ev_id_wen2  <= '0';
        ev_id_state <= idle;
    end case;
  end if;
end process read_ev_data;

calc_bcid_comp_value : process (clk, rst)
begin
  if (rst = '1') then
    bcid_comp_value <= (others => '0');
    bcid_offset_i <= (others => '0');
  elsif (clk'event AND clk = '1') then
    if (bcid_offset(7) = '1') then
      bcid_offset_i <= X"F" & bcid_offset;
    else
      bcid_offset_i <= X"0" & bcid_offset;
    end if;
    if (bcid_rollover_select = '1' AND (bcid_value + bcid_offset_i > X"DEB")) then
      bcid_comp_value <= bcid_value + bcid_offset_i - X"DEC"; 
    else
      bcid_comp_value <= bcid_value + bcid_offset_i;
    end if;
  end if;
end process calc_bcid_comp_value;
 
header_ev_id_data(63 DOWNTO 0) <= ev_data_sh_reg (63 DOWNTO 0); --Header ID Info

ev_id_data1(111 DOWNTO 0) <=  ev_data_sh_reg (159 DOWNTO 64)   -- DynMask0
                            & ev_data_sh_reg ( 61 DOWNTO 58)   -- ROD Trigger Type (3:0)
                            & ev_data_sh_reg (  3 DOWNTO  0)   -- L1ID(3:0)
                            & bcid_comp_value(  7 DOWNTO  0);  --& ev_data_sh_reg( 43 DOWNTO 36);    -- BCID

ev_id_data2(111 DOWNTO 0) <=  ev_data_sh_reg (255 DOWNTO 160)  -- DynMask1
                            & ev_data_sh_reg ( 61 DOWNTO  58)  -- ROD Trigger Type (3:0)
                            & ev_data_sh_reg (  3 DOWNTO   0)  -- L1ID(3:0)
                            & bcid_comp_value(  7 DOWNTO   0); --& ev_data_sh_reg( 43 DOWNTO  36);   -- BCID

--                          
END rtl;