----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    fifo.vhd
-- Description:
--    Paramatrizable Width FIFO
--    Depth of FIFO is 32 words
--    uses 4 bit wide x 5 bits deep dual port mems so widths must be
--    in multiples of 4
--
--    After you issue a read request, it takes 2 cycles for you to
--    get the data back. The almost full, full, and empty flags are
--    updated on the next clock cycle after a ren or wen signal is sent
----------------------------------------------------------------------------

library IEEE ;
use IEEE.STD_LOGIC_1164.all ;
use IEEE.std_logic_arith.all;

--library UNISIM;
--use unisim.RAM16X1D;

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

ENTITY FIFO IS
  generic (
    width : integer := 1 
    );

  port (
    clk         : in  STD_LOGIC ;
    rst         : in  STD_LOGIC ;
    data_in     : in  STD_LOGIC_VECTOR(width-1 DOWNTO 0) ;
    data_out    : out STD_LOGIC_VECTOR(width-1 DOWNTO 0) ;
    ren         : in  STD_LOGIC ;
    wen         : in  STD_LOGIC ;
    full        : out STD_LOGIC ;
    empty       : out STD_LOGIC ;
    almost_full : out STD_LOGIC 
    );
END FIFO ;

architecture rtl of FIFO is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  signal mem_waddr     : STD_LOGIC_VECTOR (4 downto 0) ;
  signal waddr         : UNSIGNED (4 downto 0) ;
  signal mem_raddr     : STD_LOGIC_VECTOR (4 downto 0) ;
  signal raddr         : UNSIGNED (4 downto 0) ;
  signal num_words     : UNSIGNED (4 downto 0) ;
  signal mem_din       : STD_LOGIC_VECTOR (width-1 downto 0) ;
  signal mem_wren1     : STD_LOGIC ;
  signal mem_wren2     : STD_LOGIC ;
  signal mem_rdout1    : STD_LOGIC_VECTOR (width-1 downto 0) ;
  signal mem_rdout2    : STD_LOGIC_VECTOR (width-1 downto 0) ;
  signal full_i        : STD_LOGIC ;
  signal empty_i       : STD_LOGIC ;
  signal almost_full_i : STD_LOGIC ;
  

----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------

component RAM16X1D
  port (
    DPO   : out std_logic;
    SPO   : out std_logic;
    A0    : in  std_logic;
    A1    : in  std_logic;
    A2    : in  std_logic;
    A3    : in  std_logic;
    D     : in  std_logic;
    DPRA0 : in  std_logic;
    DPRA1 : in  std_logic;
    DPRA2 : in  std_logic;
    DPRA3 : in  std_logic;
    WCLK  : in  std_logic;
    WE    : in  std_logic
    );
end component;

BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

gen_mem_block:
for i in 0 to (width-1) generate
  mem_block1: RAM16X1D
    port map (
      DPO   => mem_rdout1(i),
      SPO   => open,
      A0    => mem_waddr(0),
      A1    => mem_waddr(1),
      A2    => mem_waddr(2),
      A3    => mem_waddr(3),
      D     => mem_din(i),
      DPRA0 => mem_raddr(0),
      DPRA1 => mem_raddr(1),
      DPRA2 => mem_raddr(2),
      DPRA3 => mem_raddr(3),
      WCLK  => clk,
      WE    => mem_wren1
      );

  mem_block2: RAM16X1D
    port map (
      DPO   => mem_rdout2(i),
      SPO   => open,
      A0    => mem_waddr(0),
      A1    => mem_waddr(1),
      A2    => mem_waddr(2),
      A3    => mem_waddr(3),
      D     => mem_din(i),
      DPRA0 => mem_raddr(0),
      DPRA1 => mem_raddr(1),
      DPRA2 => mem_raddr(2),
      DPRA3 => mem_raddr(3),
      WCLK  => clk,
      WE    => mem_wren2
      );
end generate;

mem_waddr   <= std_logic_vector(waddr);
mem_raddr   <= std_logic_vector(raddr);
full        <= full_i;
empty       <= empty_i;
almost_full <= almost_full_i;
mem_din     <= data_in;
mem_wren1   <= wen and (not mem_waddr(4)) and (not full_i);
mem_wren2   <= wen and (mem_waddr(4)) and (not full_i);

data_out_mux : process (mem_rdout1, mem_rdout2, mem_raddr)
begin
  for i in 0 to (width-1) loop
    if (mem_raddr(4) = '0') then
      data_out(i)    <= mem_rdout1(i);
    else
      data_out(i)    <= mem_rdout2(i);    
    end if;
  end loop;
end process data_out_mux;

fifo_logic : process (clk, rst)
begin -- fifo_logic
  if (rst = '1') then
    almost_full_i <= '0';
    empty_i       <= '1';
    full_i        <= '0';
    num_words     <= "00000";
    raddr         <= "11111";
    waddr         <= "00000";
  elsif (clk'event and clk = '1') then
    if (std_logic_vector(num_words(4 downto 2)) = "111") then
      almost_full_i <= '1';
    else
      almost_full_i <= '0';
    end if;
    if (wen = '1' and full_i = '0') then
      waddr <= waddr + 1;
      if ((not (ren = '1' and empty_i = '0'))) then
        num_words <= num_words + 1;
        empty_i <= '0';
        if (std_logic_vector(num_words) = "11110") then
          full_i <= '1';
        end if;
      end if;
    end if;    
    if (ren = '1' and empty_i = '0') then
      raddr <= raddr + 1;
      if ((not (wen = '1' and full_i = '0'))) then
        num_words <= num_words - 1;
        full_i <= '0';
        if (std_logic_vector(num_words) = "00001") then
          empty_i <= '1';
        end if;
      end if;
    end if;
  end if;
end process fifo_logic;

END rtl;

