-----------------------------------------------------------------------------
--                                                                         --
--  Module      : fifo1024x16.vhd                   Last Update: 3/11/03   --
--                                                                         --
--  Description : FIFO controller top level.                               --
--                Implements 1024x16 FIFO with common read/write clocks.   --
--                                                                         --
--  The following VHDL code implements a 1024x16 FIFO in a Virtex          --
--  device.  The inputs are a Clock, a Read Enable, a Write Enable,        --
--  Write Data, and a FIFO_gsr signal as an initial reset.  The outputs    --
--  are Read Data, Full, Empty, and the FIFOcount outputs, which           --
--  indicate how full the FIFO is.                                         --
--                                                                         --
-----------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;

-- pragma translate_off
library UNISIM;
use unisim.all;
-- pragma translate_on

entity fifo1024x16 is
  port (
    clock_in        : IN  std_logic;
    read_enable_in  : IN  std_logic;
    write_enable_in : IN  std_logic;
    write_data_in   : IN  std_logic_vector(15 downto 0);
    fifo_gsr_in     : IN  std_logic;
    read_data_out   : OUT std_logic_vector(15 downto 0);
    full_out        : OUT std_logic;
    empty_out       : OUT std_logic;
    fifocount_out   : OUT std_logic_vector(9 downto 0)
    );
END fifo1024x16;

architecture rtl of fifo1024x16 is
  
-------------------------------------------------------------------------------
-- SIGNAL DECLARATION
-------------------------------------------------------------------------------

  signal read_enable  : std_logic;
  signal write_enable : std_logic;
  signal fifo_gsr     : std_logic;
  signal read_data    : std_logic_vector(15 downto 0);
  signal write_data   : std_logic_vector(15 downto 0);
  signal full         : std_logic;
  signal empty        : std_logic;
  signal read_addr    : std_logic_vector(9 downto 0);
  signal write_addr   : std_logic_vector(9 downto 0);
  signal fcounter     : std_logic_vector(9 downto 0);
  signal read_allow   : std_logic;
  signal write_allow  : std_logic;
  signal gnd          : std_logic;
  signal gnd_bus      : std_logic_vector(3 downto 0);
  signal pwr          : std_logic;

-------------------------------------------------------------------------------
-- COMPONENT DECLARATION
-------------------------------------------------------------------------------
component RAMB4_S4_S4
  port (
    ADDRA: IN std_logic_vector(9 downto 0);
    ADDRB: IN std_logic_vector(9 downto 0);
    DIA:   IN std_logic_vector(3 downto 0);
    DIB:   IN std_logic_vector(3 downto 0);
    WEA:   IN std_logic;
    WEB:   IN std_logic;
    CLKA:  IN std_logic;
    CLKB:  IN std_logic;
    RSTA:  IN std_logic;
    RSTB:  IN std_logic;
    ENA:   IN std_logic;
    ENB:   IN std_logic;
    DOA:   OUT std_logic_vector(3 downto 0);
    DOB:   OUT std_logic_vector(3 downto 0)
    );
END component;

BEGIN
-------------------------------------------------------------------------------
-- SIGNAL ASSIGNMENT
-------------------------------------------------------------------------------

  read_enable   <= read_enable_in;
  write_enable  <= write_enable_in;
  fifo_gsr      <= fifo_gsr_in;
  write_data    <= write_data_in;
  read_data_out <= read_data;
  full_out      <= full;
  empty_out     <= empty;
  read_allow    <= read_enable  AND NOT empty;
  write_allow   <= write_enable AND NOT full;

  fifocount_out <= fcounter;

  gnd_bus <= (others => '0');
  gnd <= '0';
  pwr <= '1';
   
--------------------------------------------------------------------------
--                                                                      --
-- BLOCK RAM instantiation for FIFO.  Module is 1024x16, of which one   -- 
-- address location is sacrificed for the overall speed of the design.  --
--                                                                      --
--------------------------------------------------------------------------

gen_mem_block:
  for i in 0 to 3 generate
    bram1: RAMB4_S4_S4
      port map (
        ADDRA => read_addr,
        ADDRB => write_addr, 
        DIB   => write_data((i*4)+3 downto i*4),
        DIA   => gnd_bus,
        WEA   => gnd,
        WEB   => write_allow,
        CLKA  => clock_in,
        CLKB  => clock_in, 
        RSTA  => gnd,
        RSTB  => gnd,
        ENA   => read_allow,
        ENB   => pwr, 
        DOA   => read_data((i*4)+3 downto i*4),
        DOB   => open
        );
end generate;

-------------------------------------------------------------------------------
--                                                                           --
-- Empty flag is set on fifo_gsr (initial), or when on the next clock cycle, --
-- Write Enable is low, and either the FIFOcount is equal to 0, or it is     --
-- equal to 1 and Read Enable is high (about to go Empty).                   --
--                                                                           --
-------------------------------------------------------------------------------

proc1: PROCESS (clock_in, fifo_gsr)
BEGIN
  IF (fifo_gsr = '1') THEN
    empty <= '1';
  ELSIF (clock_in'EVENT AND clock_in = '1') THEN
    IF (    (fcounter(9 downto 1) = "000000000")
        AND (write_enable = '0')
        AND (fcounter(0) = '0' OR read_enable = '1')) THEN
      empty <= '1';
    ELSE
      empty <= '0';
    END IF;
  END IF;
END PROCESS proc1;

-------------------------------------------------------------------------------
--                                                                           --
-- Full flag is set on fifo_gsr (but it is cleared on the first valid clock  --
-- edge after fifo_gsr is removed), or when on the next clock cycle, Read    --
-- Enable is low, and either the FIFOcount is equal to 1FF (hex), or it is   --
-- equal to 1FE and the Write Enable is high (about to go Full).             --
--                                                                           --
-------------------------------------------------------------------------------

proc2: PROCESS (clock_in, fifo_gsr)
BEGIN
  IF (fifo_gsr = '1') THEN
    full <= '0';
  ELSIF (clock_in'EVENT AND clock_in = '1') THEN
    IF (    (fcounter(9 downto 1) = "111111111")
        AND (read_enable = '0')
        AND (fcounter(0) = '1' OR write_enable = '1')) THEN
      full <= '1';
    ELSE
      full <= '0';
    END IF;
  END IF;
END PROCESS proc2;

proc3: PROCESS (clock_in, fifo_gsr)
BEGIN
  IF (fifo_gsr = '1') THEN
    read_addr <= (others => '0');
  ELSIF (clock_in'EVENT AND clock_in = '1') THEN
    IF (read_allow = '1') THEN
      read_addr <= unsigned(read_addr) + 1;
    END IF;
  END IF;
END PROCESS proc3;

proc4: PROCESS (clock_in, fifo_gsr)
BEGIN
  IF (fifo_gsr = '1') THEN
    write_addr <= (others => '0');
  ELSIF (clock_in'EVENT AND clock_in = '1') THEN
    IF (write_allow = '1') THEN
      write_addr <= unsigned(write_addr) + 1;
    END IF;
  END IF;
END PROCESS proc4;

-----------------------------------------------------------------------------
--                                                                         --
--  Generation of FIFOcount outputs.  Used to determine how full FIFO is,  --
--  based on a counter that keeps track of how many words are in the FIFO. --
--  Also used to generate Full and Empty flags.  Only the upper four bits  --
--  of the counter are sent outside the module                             --
--                                                                         --
-----------------------------------------------------------------------------

proc5: PROCESS (clock_in, fifo_gsr)
BEGIN
  IF (fifo_gsr = '1') THEN
    fcounter <= (others => '0');
  ELSIF (clock_in'EVENT AND clock_in = '1') THEN
    IF    (read_allow = '1' AND write_allow = '0') THEN
      fcounter <= unsigned(fcounter) - '1';
    ELSIF (read_allow = '0' AND write_allow = '1') THEN
      fcounter <= unsigned(fcounter) + '1';
    END IF;
  END IF;
END PROCESS proc5;

END rtl;

