----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    fifo_rambs16.vhd
-- Description:
--    Paramatrizable Width FIFO
--    Depth of FIFO is 256 words
--    uses 16 bit wide x 256 bits deep dual port mems so widths must be
--    in multiples of 16
--
--    After you issue a read request, it takes 2 cycles for you to
--    get the data back. The almost full, full, and empty flags are
--    updated on the next clock cycle after a ren or wen signal is sent
----------------------------------------------------------------------------

library IEEE ;
use IEEE.STD_LOGIC_1164.all;
use IEEE.std_logic_arith.all;

-- pragma translate_off
library UNISIM;
use unisim.all;
-- pragma translate_on

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

ENTITY fifo_rambs16 IS
  generic (
    width : integer := 1 
    );
  port (
    clk_in         : in  STD_LOGIC;
    rst_in         : in  STD_LOGIC;
    data_in        : in  STD_LOGIC_VECTOR(16*(width-1)+15 DOWNTO 0);
    data_out       : out STD_LOGIC_VECTOR(16*(width-1)+15 DOWNTO 0);
    ren            : in  STD_LOGIC;
    wen            : in  STD_LOGIC;
    full           : out STD_LOGIC;
    empty          : out STD_LOGIC;
    almost_full    : out STD_LOGIC;
    fifo_count_out : out STD_LOGIC_VECTOR( 7 downto 0) 
    );
END fifo_rambs16 ;

architecture rtl of fifo_rambs16 is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  signal pwr           : STD_LOGIC;
  signal gnd           : STD_LOGIC;
  signal mem_waddr     : STD_LOGIC_VECTOR(7 downto 0);
  signal waddr         : UNSIGNED(7 downto 0);
  signal mem_raddr     : STD_LOGIC_VECTOR (7 downto 0);
  signal raddr         : UNSIGNED(7 downto 0);
  signal num_words     : UNSIGNED(7 downto 0);
  signal mem_din       : STD_LOGIC_VECTOR(16*(width-1)+15 downto 0);
  signal mem_wren      : STD_LOGIC;
  signal mem_rdout     : STD_LOGIC_VECTOR(16*(width-1)+15 downto 0);
  signal full_i        : STD_LOGIC;
  signal empty_i       : STD_LOGIC;
  signal almost_full_i : STD_LOGIC;
  signal logic0vector  : STD_LOGIC_VECTOR(15 downto 0);
  signal clr_fifo      : STD_LOGIC;

----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------

component RAMB4_S16_S16
  port (
    DOA    : out STD_LOGIC_VECTOR(15 downto 0);
    DOB    : out STD_LOGIC_VECTOR(15 downto 0);
    ADDRA  : in  STD_LOGIC_VECTOR( 7 downto 0);
    DIA    : in  STD_LOGIC_VECTOR(15 downto 0);
    ENA    : in  STD_ULOGIC;
    CLKA   : in  STD_ULOGIC;
    WEA    : in  STD_ULOGIC;                
    RSTA   : in  STD_ULOGIC;
    ADDRB  : in  STD_LOGIC_VECTOR( 7 downto 0);
    DIB    : in  STD_LOGIC_VECTOR(15 downto 0);
    ENB    : in  STD_ULOGIC;
    CLKB   : in  STD_ULOGIC;
    WEB    : in  STD_ULOGIC;
    RSTB   : in  STD_ULOGIC
    ); 
end component;


BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

gen_ramb_fifo : for i in 0 to (width-1) generate
  MB : RAMB4_S16_S16 -- Block[15..0] of 16bits
  port map (
    DOA   => open,
    DOB   => mem_rdout((16*i)+15 downto 16*i),
    ADDRA => mem_waddr,
    DIA   => mem_din((16*i)+15 downto 16*i),
    ENA   => pwr,
    CLKA  => clk_in,
    WEA   => mem_wren,
    RSTA  => rst_in,
    ADDRB => mem_raddr,
    DIB   => logic0vector,
    ENB   => pwr,
    CLKB  => clk_in,
    WEB   => gnd,
    RSTB  => rst_in
    ); 
end generate;

pwr          <= '1';
gnd          <= '0';
logic0vector <= (others => '0');
mem_waddr    <= std_logic_vector(waddr);
mem_raddr    <= std_logic_vector(raddr);
full         <= full_i;
empty        <= empty_i;
almost_full  <= almost_full_i;
mem_wren     <= (wen AND NOT full_i) OR clr_fifo;

data_out     <= mem_rdout;

fifo_data_mux : process (
  clr_fifo,
  data_in
  )
begin
  if (clr_fifo = '1') then
    mem_din  <= (others => '0');
--    data_out <= (others => '0');
  else
    mem_din  <= data_in;
--    data_out <= mem_rdout;
  end if;
end process;

fifo_logic : process (
  clk_in, 
  rst_in
  )
begin -- fifo_logic
  if (rst_in = '1') then
    almost_full_i  <= '0';
    empty_i        <= '1';
    full_i         <= '0';
    num_words      <= (others => '0');
    raddr          <= (others => '1');
    waddr          <= "11111100";  -- Set WADDR below 0
    clr_fifo       <= '1';
    fifo_count_out <= (others => '0');
  elsif (clk_in'event AND clk_in = '1') then
    fifo_count_out <= std_logic_vector(num_words);
    if (clr_fifo = '1') then
      almost_full_i <= '0';
      empty_i       <= '1';
      full_i        <= '0';
      num_words     <= (others => '0');
      raddr         <= (others => '1');
      if (waddr = 8) then  -- Only clear values around 0
        clr_fifo <= '0';
        waddr    <= (others => '0');
      else
        waddr <= waddr + 1;
      end if;
    else
      if (std_logic_vector(num_words(7 downto 2)) = "111111") then
        almost_full_i <= '1';
      else
        almost_full_i <= '0';
      end if;

      if (wen = '1' AND full_i = '0') then
        waddr <= waddr + 1;
        if ((NOT (ren = '1' AND empty_i = '0'))) then
          num_words <= num_words + 1;
          empty_i <= '0';
          if (std_logic_vector(num_words) >= "11111110") then  -- 20151026  Added >=
            full_i <= '1';
          end if;
        end if;
      end if;    

      if (ren = '1' AND empty_i = '0') then
        raddr <= raddr + 1;
        if ((NOT (wen = '1' AND full_i = '0'))) then
          num_words <= num_words - 1;
          full_i <= '0';
          if (std_logic_vector(num_words) <= "00000001") then  -- 20151026  Added <=
            empty_i <= '1';
          end if;
        end if;
      end if;
    end if;
  end if;
end process fifo_logic;

END rtl;