----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    format_data.vhd
-- Description:
--    ROD Dual Gatherer
--    SCT Version
--    Performs final data formatting, removes events with no data,
--    and counts how many words are in the output memories for the
--    given event fragment.
----------------------------------------------------------------------------

library IEEE ;
use IEEE.STD_LOGIC_1164.all ;
use IEEE.std_logic_arith.all;

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

ENTITY format_data IS
  port (
    clk              : in  STD_LOGIC ;
    rst              : in  STD_LOGIC ;
    data_in          : in  STD_LOGIC_VECTOR(45 DOWNTO 0) ;
    data_valid_in    : in  STD_LOGIC ;
    data_out         : out STD_LOGIC_VECTOR(45 DOWNTO 0) ;
    data_valid_out_n : out STD_LOGIC ;
    count_fifo_wen   : out STD_LOGIC ;
    count_fifo_data  : out STD_LOGIC_VECTOR(13 DOWNTO 0)
    );
END format_data ;

architecture rtl of format_data is


----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

signal data_in_d1              : STD_LOGIC_VECTOR(45 DOWNTO 0); 
signal data_valid_in_d1        : STD_LOGIC ;
signal last_word               : STD_LOGIC ;
signal lsw_valid               : STD_LOGIC ;
signal msw_valid               : STD_LOGIC ;
signal extra_data              : STD_LOGIC_VECTOR(15 downto 0);
signal extra_halfword          : STD_LOGIC ;
signal data_out_p2             : STD_LOGIC_VECTOR(45 downto 0);
signal data_valid_out_p2       : STD_LOGIC ;
signal last_word_flush         : STD_LOGIC ;
signal data_out_p1             : STD_LOGIC_VECTOR(45 downto 0);
signal data_valid_out_p1       : STD_LOGIC ;
signal count_fifo_wen_i        : STD_LOGIC ;
signal word_count              : UNSIGNED (13 DOWNTO 0); 

BEGIN  --  Main Body of VHDL code

register_data : process (clk, rst)
  begin
    if (rst = '1') then
      data_in_d1 <= (others => '0');
      data_valid_in_d1 <= '0';
      last_word <= '0';
    elsif (clk'event and clk = '1') then
      data_in_d1 <= data_in;
      data_valid_in_d1 <= data_valid_in;
      last_word <= data_in(45);
    end if;
  end process register_data;

mark_valid_halfwords : process (clk, rst)
  begin
    if (rst = '1') then
      msw_valid <= '0';
      lsw_valid <= '0';
    elsif (clk'event and clk = '1') then
      msw_valid <= '0';
      lsw_valid <= '0';
      if (data_valid_in = '1') then
        -- not an empty event with no errors
        if (not(data_in(31 downto 28) = "0010" and
                data_in(15 downto  0) = "0100000000000000" and
--                data_in(15 downto 10) = "010000" and
                data_in(39) = '0' and data_in(41) = '0' and
                data_in(42) = '0' and data_in(43) = '0')) then
          -- msw is valid if it isn't a trailer with no errors
          if (data_in(31 downto 16) /= "0100000000000000") then
--          if (data_in(31 downto 26) /= "010000") then
            msw_valid <= '1';
          end if;
          -- lsw is valid if there is no trailer in msw and
          -- lsw is not a trailer with no error
          if (data_in(31 downto 29) /= "010" and
              data_in(15 downto  0) /= "0100000000000000") then
--              data_in(15 downto 10) /= "010000") then
            lsw_valid <= '1';
          end if;
          -- if msw is header and there is a preamble error or there
          -- is an L1/BC ID error or a timeout error lsw must be taken too
          -- even if not needed.
          if (data_in(31 downto 29) = "001" and
               (data_in(28) = '1' or data_in(39) = '1' or
                data_in(41) = '1' or data_in(42) = '1' or data_in(43) = '1')) then
            lsw_valid <= '1';
          end if;
          -- if this is the last data  word from the formatter and there is no
          -- extra half word, then both msw and lsw are valid
          if (data_in(45) = '1' and
              data_in(31) = '1' and
              data_in(15 downto 10) = "010000" and
              extra_halfword = '0') then
            lsw_valid <= '1';
          end if; 
          if (data_in(45) = '1' and
              data_in(31) = '1' and
              data_in(15 downto 10) = "010000" and
              extra_halfword = '1') then
            lsw_valid <= '0';
          end if; 
        end if;
      end if;
    end if;
  end process mark_valid_halfwords;

repack_data : process (clk, rst)
  begin
    if (rst = '1') then
      extra_halfword     <= '0';
      extra_data         <= (others => '0');
      data_out_p2        <= (others => '0');
      data_valid_out_p2  <= '0';
      last_word_flush    <= '0';
    elsif (clk'event and clk = '1') then
      data_valid_out_p2  <= '0';
      last_word_flush    <= last_word;
--  ********************  NEW: added in Ver19F *****************
--    if (last_word_flush = '1' and extra_halfword = '1') then  -- Remove if dropping last trailer
--      extra_halfword <= '0';                                  --
--      data_valid_out_p2 <= '1';                               --
--      data_out_p2(45 downto 0) <= "0000000"                   --
--                                & data_out_p2(38 downto 32)   --
--                                & extra_data                  --
--                                & "0100000000000000";         -- Remove if dropping last trailer
      if (last_word = '1' and extra_halfword = '1') then  -- Use when dropping last trailer
--    elsif (last_word = '1' and extra_halfword = '1') then     -- Remove if dropping last trailer
        if (msw_valid = '1') then
          if (lsw_valid = '1') then -- Usually H/T word on an odd boundary
--          extra_halfword <= '1';  -- Remove if dropping last trailer
            data_out_p2(45 downto 0) <= '0'
                                      & data_in_d1(44 downto 32) 
                                      & extra_data
                                      & data_in_d1(31 downto 16);
          elsif (lsw_valid = '0') then -- Usually D/T word on an odd boundary
--          extra_halfword <= '0';  -- Remove if dropping last trailer
            data_out_p2(45 downto 0) <= "0000000"
                                      & data_out_p2(38 downto 32)
                                      & extra_data
                                      & data_in_d1(31 downto 16);
          end if;
        elsif (msw_valid = '0') then -- Extra D word on an odd boundary
          if (lsw_valid = '0') then
--          extra_halfword <= '0';  -- Remove if dropping last trailer
            data_out_p2(45 downto 0) <= "0000000"
                                      & data_out_p2(38 downto 32)
                                      & extra_data
                                      & "0100000000000000";
          end if;
        end if;
        data_valid_out_p2 <= '1';
        extra_halfword <= '0'; -- Use when dropping last trailer
--  ^^^^^^^^^^^^^^^^^^^^  NEW: added in Ver19F  ^^^^^^^^^^^^^^^^
      elsif (data_valid_in_d1 = '1') then
        if (msw_valid = '1' and lsw_valid = '0' and extra_halfword = '0') then
          extra_halfword <= '1';
          extra_data(15 downto 0) <= data_in_d1(31 downto 16);
        elsif (msw_valid = '1' and lsw_valid = '0' and extra_halfword = '1') then
          extra_halfword <= '0';
          data_out_p2(45 downto 0) <= '0' 
                                    & data_in_d1(44 downto 32) 
                                    & extra_data 
                                    & data_in_d1(31 downto 16);
          data_valid_out_p2 <= '1';
        elsif (msw_valid = '1' and lsw_valid = '1' and extra_halfword = '0') then
          extra_halfword <= '0';
          data_out_p2(45 downto 0) <= '0' & data_in_d1(44 downto 0);
          data_valid_out_p2 <= '1';
        elsif (msw_valid='1' and lsw_valid='1' and extra_halfword='1') then
          extra_halfword <= '1';
          extra_data(15 downto 0) <= data_in_d1(15 downto 0);
          data_out_p2(45 downto 0) <= '0' 
                                    & data_in_d1(44 downto 32) 
                                    & extra_data 
                                    & data_in_d1(31 downto 16);
          data_valid_out_p2 <= '1';
        end if;
      end if;
    end if;
  end process repack_data;
  
send_out_and_count : process (clk, rst)
  begin
    if (rst = '1') then
      data_out_p1       <= (others => '0');
      data_out          <= (others => '0');
      data_valid_out_p1 <= '0';
      data_valid_out_n  <= '1';
      count_fifo_wen_i  <= '0';
      word_count        <= (others => '0');
    elsif (clk'event and clk = '1') then
      data_out_p1       <= data_out_p2;
      data_out          <= data_out_p1;
      data_valid_out_p1 <= data_valid_out_p2;
      data_valid_out_n  <= NOT data_valid_out_p1;

      count_fifo_wen_i <= last_word_flush;
      if (count_fifo_wen_i = '1' AND data_valid_out_p2 = '1') then
        word_count <= "00000000000001";
      elsif (count_fifo_wen_i = '1' AND data_valid_out_p2 = '0') then
        word_count <= "00000000000000";
      elsif (count_fifo_wen_i = '0' AND data_valid_out_p2 = '1') then
        word_count <= word_count + 1;
      end if;        
    end if;
  end process send_out_and_count;
    
count_fifo_wen <= count_fifo_wen_i;
count_fifo_data <= std_logic_vector (word_count);

END rtl;

