--------------------------------------------------------------------------
-- LIBRARY INCLUDES
--------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.numeric_std.all; -- needed for +/- operations

--------------------------------------------------------------------------
-- PORT DECLARATION
--------------------------------------------------------------------------
entity ringbuf is
  generic(
    g_bramCount     : integer := 4;
    g_widthPerBram  : integer := 1
  );
  port(
    IN_clk          : in    std_logic;
    IN_rst          : in    std_logic;
    IN_control      : in    std_logic_vector(31 downto 0);
    OUT_status      : out   std_logic_vector(31 downto 0);
    IN_trig_mask    : in    std_logic_vector(31 downto 0);
    IN_triggers     : in    std_logic_vector(31 downto 0);
    IN_data         : in    std_logic_vector(31 downto 0);
    IN_valid        : in    std_logic;
    OUT_data        : out   std_logic_vector(31 downto 0);
    OUT_valid       : out   std_logic
  );
end ringbuf;

architecture rtl of ringbuf is

function addressWidth(widthPerBram : integer) return integer is
  variable v_ret : integer;
begin
  v_ret := -1;
  
  for i in 0 to 4 loop
    if widthPerBram = 2**i then
      v_ret := 12-i;
    end if;
  end loop;
  
  return v_ret;
end addressWidth;

constant c_addrHigh : integer := addressWidth(g_widthPerBram) - 1;

--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------

signal s_data_out        : std_logic_vector(31 downto 0);
--signal r_data_out        : std_logic_vector(31 downto 0);
signal r_wr_adr          : std_logic_vector(11 downto 0);
signal r_rd_adr          : std_logic_vector(11 downto 0);
signal r_valid           : std_logic;
signal r_got_trigger     : std_logic;
signal r_got_all_data    : std_logic;
signal r_words_left      : std_logic_vector(11 downto 0);
signal s_wren            : std_logic;
signal s_wren_i          : std_logic;
signal r_wren            : std_logic;
signal s_status          : std_logic_vector(31 downto 0);
signal s_trig_pos        : std_logic_vector(11 downto 0);
signal s_strech_count    : std_logic_vector(7 downto 0);
signal s_rd_pulse        : std_logic;
signal s_sw_trig         : std_logic;
signal s_sw_trig_en      : std_logic;
signal s_stretch         : std_logic;
signal s_flush           : std_logic;
signal s_trig            : std_logic;
signal r_wr_counter      : std_logic_vector(7 downto 0);

--------------------------------------------------------------------------
-- COMPONENT DECLARATION
--------------------------------------------------------------------------

component bram
  generic (
    g_data_width : integer := 1
  );
  port (
    IN_clk    : in  std_logic; -- clk40 input
    IN_rst    : in  std_logic; -- asynchronous global reset
    IN_data   : in  std_logic_vector(g_data_width-1 downto 0);
    IN_wr_adr : in  std_logic_vector(11 downto  0); -- 
    IN_rd_adr : in  std_logic_vector(11 downto  0); --
    IN_wren   : in  std_logic; --
    OUT_data  : out std_logic_vector(g_data_width-1 downto 0)
  );
end component;

--------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
--------------------------------------------------------------------------
begin

MB : for i in 0 to g_bramCount-1 generate
  
  buf : bram
  generic map (
    g_data_width => g_widthPerBram
  )
  port map (
    IN_clk      => IN_clk,
    IN_rst      => IN_rst,
    IN_data     => IN_data(g_widthPerBram*(i+1)-1 downto g_widthPerBram*i),
    IN_wr_adr   => r_wr_adr,
    IN_rd_adr   => r_rd_adr,
    IN_wren     => s_wren_i,
    OUT_data    => s_data_out(g_widthPerBram*(i+1)-1 downto g_widthPerBram*i)
  );
end generate;


OUT_status <= s_status;
OUT_data <= s_data_out;
OUT_valid <= r_valid;

s_data_out(31 downto g_widthPerBram*g_bramCount) <= (others => '0');

s_status(31 downto 18) <= (others => '0');
s_status(17) <= r_got_all_data; -- full
s_status(16) <= not r_got_trigger; -- empty
s_status(15 downto c_addrHigh + 1) <= (others => '0');
s_status(c_addrHigh downto 0) <= std_logic_vector( unsigned(r_wr_adr(c_addrHigh downto 0)) - unsigned(r_rd_adr(c_addrHigh downto 0))); -- occ count

s_trig_pos     <= IN_control(27 downto 16);
s_strech_count <= IN_control(15 downto 8);
s_rd_pulse     <= IN_control(7);
s_sw_trig      <= IN_control(4);
s_sw_trig_en   <= IN_control(3);
s_stretch      <= IN_control(2);
s_wren         <= IN_control(1) and IN_valid;
s_flush        <= IN_control(0);

s_wren_i <= (s_wren and not r_got_all_data) when s_stretch = '0' else
            (s_wren and not r_got_all_data) when unsigned(r_wr_counter) = 0 else
            '0';

s_trig <= s_sw_trig WHEN s_sw_trig_en = '1' ELSE
          '1' WHEN (IN_trig_mask and IN_triggers) /= x"00000000" ELSE
          '0' ;


--------------------------------------------------------------------------
-- PROCESS DECLARATION
--------------------------------------------------------------------------

p_strech_fifo: process(IN_clk)
begin
  if rising_edge(IN_clk) then
    if (IN_rst = '1' or s_flush = '1') then
      r_wr_counter <= (others => '0');
    else
      if unsigned(r_wr_counter) = 0 then
        r_wr_counter <= std_logic_vector(unsigned(s_strech_count) - 1);
      else
        r_wr_counter <= std_logic_vector(unsigned(r_wr_counter) - 1);
      end if;
    end if;
  end if;
end process;

p_ringbuf : process (IN_clk)
begin
  if rising_edge(IN_clk) then
    if (IN_rst = '1' or s_flush = '1') then
      r_wren <= '1';
      r_valid <= '0';
      r_got_trigger <= '0';
      r_got_all_data <= '0';
      r_wr_adr <= (others => '0');
      r_rd_adr <= (others => '0');
      r_words_left <= (others => '1');
    else
      r_valid <= '0';
--      r_data_out <= s_data_out;
      
      if s_wren_i = '1' then
        
        if unsigned(r_words_left(c_addrHigh downto 0)) = 1 then
          r_got_all_data <= '1';
          r_words_left <= (others => '1');
        else
          r_wr_adr(c_addrHigh downto 0) <= std_logic_vector( unsigned(r_wr_adr(c_addrHigh downto 0)) + 1);
          
          if unsigned(r_rd_adr(c_addrHigh downto 0)) = unsigned(r_wr_adr(c_addrHigh downto 0))+1 then
            r_rd_adr(c_addrHigh downto 0) <= std_logic_vector( unsigned(r_rd_adr(c_addrHigh downto 0)) + 1);
          end if;
          
        end if;
        
        if r_got_trigger = '1' then
          r_words_left(c_addrHigh downto 0) <= std_logic_vector( unsigned(r_words_left(c_addrHigh downto 0)) - 1);
        end if;
      end if;
      
      if s_trig = '1' and r_got_trigger = '0' then
        r_got_trigger <= '1';
        r_words_left(c_addrHigh downto 0) <= std_logic_vector( 2**(c_addrHigh+1) - 1 - unsigned(s_trig_pos(c_addrHigh downto 0)));
      end if;
      
      if s_rd_pulse = '1' and r_got_all_data = '1' then
        if unsigned(r_rd_adr(c_addrHigh downto 0)) = unsigned(r_wr_adr(c_addrHigh downto 0)) then
          r_valid <= '0';
          r_got_trigger <= '0';
          r_got_all_data <= '0';
        else
          r_rd_adr(c_addrHigh downto 0) <= std_logic_vector( unsigned(r_rd_adr(c_addrHigh downto 0)) + 1);
          r_valid <= '1';
        end if;
      end if;
     
    end if;
  end if;
end process p_ringbuf;

end rtl;
