-------------------------------------------------------------------------------
--
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 2003
-- ReadOutDriver Electronics
--
-------------------------------------------------------------------------------
--
-- Filename: dual256x32fifo.vhd                  
-- Description:
--  Input registered and Output registered fifo.     
--
-------------------------------------------------------------------------------
-- Structure: 
--  Multiple state machines in a flat topology/hierarchy which communicate
--   with single wire signals.  
-------------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
-------------------------------------------------------------------------------
-- History:
--    23 February 2000 - JMJ first version : 
-- Notes: 
--
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- LIBRARY INCLUDES
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations

-------------------------------------------------------------------------------
-- PORT DECLARATION
-------------------------------------------------------------------------------
entity dual_512x32fifo is
  port(
    clk_in            : in  std_logic;
    rclk_in           : in  std_logic;
    rst_n_in          : in  std_logic;
    link_enable_in    : in  std_logic_vector( 1 downto 0);
    wren_in           : in  std_logic_vector( 1 downto 0);
    rden_in           : in  std_logic_vector( 1 downto 0);
    fifo0_in          : in  std_logic_vector(31 downto 0); --
    fifo1_in          : in  std_logic_vector(31 downto 0); --
    fifo_out          : out std_logic_vector(31 downto 0); --
    data_valid_out    : out std_logic; --
    fifo_ef_out       : out std_logic_vector( 1 downto 0);
    fifo_ff_out       : out std_logic_vector( 1 downto 0);
    fifo0_trailer_out : out std_logic_vector( 5 downto 0);
    fifo1_trailer_out : out std_logic_vector( 5 downto 0);
    occupancy0_out    : out unsigned(8 downto 0);  --
    occupancy1_out    : out unsigned(8 downto 0)   --
    );
end dual_512x32fifo;

architecture rtl of dual_512x32fifo is
-------------------------------------------------------------------------------
-- SIGNAL DECLARATION
-------------------------------------------------------------------------------

signal fifo0_din_i      : std_logic_vector(31 downto 0);
signal fifo0_dout_i     : std_logic_vector(31 downto 0);
signal fifo1_din_i      : std_logic_vector(31 downto 0);
signal fifo1_dout_i     : std_logic_vector(31 downto 0);

signal wren0_i          : std_logic;
signal fifo0_wen_i      : std_logic;
signal fifo0_clr_i      : std_logic;
signal fifo0_wr_i       : std_logic;
signal inc_wr_addr0_i   : std_logic;
signal fifo0_ren_i      : std_logic;
signal fifo0_wr_addr    : unsigned(8 downto 0);
signal fifo0_rd_addr    : unsigned(8 downto 0);
signal occupancy0_count : unsigned(8 downto 0);
signal fifo0_full_i     : std_logic;
signal fifo0_empty_i    : std_logic;
signal wren1_i          : std_logic;
signal fifo1_wen_i      : std_logic;
signal fifo1_clr_i      : std_logic;
signal fifo1_wr_i       : std_logic;
signal inc_wr_addr1_i   : std_logic;
signal fifo1_ren_i      : std_logic;
signal fifo1_wr_addr    : unsigned(8 downto 0);
signal fifo1_rd_addr    : unsigned(8 downto 0);
signal occupancy1_count : unsigned(8 downto 0);
signal fifo1_full_i     : std_logic;
signal fifo1_empty_i    : std_logic;
signal fifo0_out_sel_i  : std_logic;
signal fifo1_out_sel_i  : std_logic;
signal fifo0_trailer_i  : std_logic;
signal fifo1_trailer_i  : std_logic;

signal clr_next_mem_loc0_i : std_logic;
signal clr_next_mem_loc1_i : std_logic;

signal data_valid0_i : std_logic;
signal data_valid1_i : std_logic;

signal logic0 : std_logic;
signal logic1 : std_logic;

type states is (
  idle,
  play,
  play_one_word,
  done
  );
  
signal fifo0_readout : states;
signal fifo1_readout : states;

signal rden_i        : std_logic_vector( 1 downto 0);

-------------------------------------------------------------------------------
-- COMPONENT DECLARATION
-------------------------------------------------------------------------------

component dpram_512x8
  port(
    clk_in           : in  std_logic;
    rst_n_in         : in  std_logic;
    dprm_writein_bus : in  std_logic_vector(7 downto 0);
    writein_addr     : in  unsigned(8 downto 0);
    readout_addr     : in  unsigned(8 downto 0);
    write_strobe     : in  std_logic;
    read_strobe      : in  std_logic;
    dprm_readout_bus : out std_logic_vector(7 downto 0)
    );
end component;

begin
-------------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
-------------------------------------------------------------------------------
U1 : for i in 0 to 3 generate 
  MB1 : dpram_512x8
    port map (
      clk_in           => clk_in,
      rst_n_in         => rst_n_in,
      dprm_writein_bus => fifo0_din_i((i*8)+7 downto (i*8)),
      writein_addr     => fifo0_wr_addr(8 downto 0),
      readout_addr     => fifo0_rd_addr(8 downto 0),
      write_strobe     => fifo0_wr_i,
      read_strobe      => logic1,
      dprm_readout_bus => fifo0_dout_i((i*8)+7 downto (i*8))
      );
end generate;

U2 : for i in 0 to 3 generate 
  MB2 : dpram_512x8
    port map (
      clk_in           => clk_in,
      rst_n_in         => rst_n_in,
      dprm_writein_bus => fifo1_din_i((i*8)+7 downto (i*8)),
      writein_addr     => fifo1_wr_addr(8 downto 0),
      readout_addr     => fifo1_rd_addr(8 downto 0),
      write_strobe     => fifo1_wr_i,
      read_strobe      => logic1,
      dprm_readout_bus => fifo1_dout_i((i*8)+7 downto (i*8))
      );
end generate;
  

-------------------------------------------------------------------------------
-- SIGNALS
-------------------------------------------------------------------------------
logic0 <= '0';
logic1 <= '1';

fifo0_wr_i <= fifo0_wen_i OR clr_next_mem_loc0_i;
fifo1_wr_i <= fifo1_wen_i OR clr_next_mem_loc1_i;

fifo_ef_out <= fifo1_empty_i & fifo0_empty_i;
fifo_ff_out <= fifo1_full_i & fifo0_full_i;

fifo0_trailer_out <= fifo0_dout_i(31 downto 29) & fifo0_dout_i(15 downto 13);
fifo1_trailer_out <= fifo1_dout_i(31 downto 29) & fifo1_dout_i(15 downto 13);

occupancy0_out <= occupancy0_count;
occupancy1_out <= occupancy1_count;

data_valid_out <= data_valid0_i OR data_valid1_i;

-------------------------------------------------------------------------------
-- PROCESS DECLARATION
-------------------------------------------------------------------------------
-- delay wren to ram block by 1 clk
-- write enable delay register and mask if the FIFO is full
wen_delay : process (
  rst_n_in, 
  clk_in, 
  wren_in
  )
begin
  if (rst_n_in = '0') then
    wren0_i <= '0';
    wren1_i <= '0';
  elsif (clk_in'event AND clk_in = '1') then
    wren0_i <= wren_in(0) AND NOT fifo0_full_i;
    wren1_i <= wren_in(1) AND NOT fifo1_full_i;
  end if;
end process wen_delay;

-------------------------------------------------------------------------------
-- Select the Write Enable used by the RAM Block.
-- If both links are active, each FIFO is 256x32
-- If one link is masked, the FIFO is 512x32
--  Following the FIFO write of event data, the WR Address is incremented by 1,
--  and the memory location is cleared of old data.  This is required so that
--  an old trailer will not be detected by the Readout Controller.
--  
wen_select : process (
  rst_n_in, 
  clk_in, 
  link_enable_in,
  wren0_i,
  wren1_i,
  fifo0_wr_addr,
  fifo1_wr_addr
  )
begin
  if (rst_n_in = '0') then
    fifo0_wen_i         <= '0';
    fifo1_wen_i         <= '0';
    inc_wr_addr0_i      <= '0';
    inc_wr_addr1_i      <= '0';
    clr_next_mem_loc0_i <= '0';
    clr_next_mem_loc1_i <= '0';
    fifo0_clr_i         <= '0';
    fifo1_clr_i         <= '0';
  elsif (clk_in'event AND clk_in = '1') then
    if (link_enable_in(0) = '1') then
      fifo0_wen_i         <= wren0_i;
      inc_wr_addr0_i      <= fifo0_wen_i;
      clr_next_mem_loc0_i <= inc_wr_addr0_i AND NOT fifo0_full_i;
    else     
      fifo0_wen_i         <= '0';
      inc_wr_addr0_i      <= '0';
      clr_next_mem_loc0_i <= '0';
    end if;    
    
    if (link_enable_in(1) = '1') then
      fifo1_wen_i         <= wren1_i;
      inc_wr_addr1_i      <= fifo1_wen_i;
      clr_next_mem_loc1_i <= inc_wr_addr1_i AND NOT fifo1_full_i;
    else     
      fifo1_wen_i         <= '0';
      inc_wr_addr1_i      <= '0';
      clr_next_mem_loc1_i <= '0';
    end if;    
  end if;
end process wen_select;

-------------------------------------------------------------------------------
-- input address counter
input_counters : process (
  rst_n_in, 
  clk_in,
  link_enable_in
  )
begin
  if (rst_n_in = '0') then
    fifo0_wr_addr <= (others => '0');
    fifo1_wr_addr <= (others => '0');
  elsif (clk_in'event AND clk_in = '1') then
    if (link_enable_in(0) = '1') then
      if (inc_wr_addr0_i = '1') then
        fifo0_wr_addr <= fifo0_wr_addr + 1;
      end if;
    else     
      fifo0_wr_addr <= (others => '0');
    end if;    
    
    if (link_enable_in(1) = '1') then
      if (inc_wr_addr1_i = '1') then
        fifo1_wr_addr <= fifo1_wr_addr + 1;
      end if;
    else     
      fifo1_wr_addr <= (others => '0');
    end if;    
  end if;
end process input_counters;

-------------------------------------------------------------------------------
input_mux : process (
  rst_n_in, 
  clk_in,
  fifo0_in,
  fifo1_in
  )
begin
  if (rst_n_in = '0') then
    fifo0_din_i <= (others => '0');
    fifo1_din_i <= (others => '0');
  elsif (clk_in'event AND clk_in = '1') then
    if (wren_in(0) = '1') then 
      fifo0_din_i <= fifo0_in;
    end if;
    if (wren_in(1) = '1') then 
      fifo1_din_i <= fifo1_in;
    end if;

--  After the Event Data is written to the FIFO, clear the data register for
--  the Next Location Clear operation    
    if (fifo0_wen_i = '1') then
      fifo0_din_i <= (others => '0');
    end if;
    if (fifo1_wen_i = '1') then
      fifo1_din_i <= (others => '0');
    end if;
  end if;
end process;
      
-------------------------------------------------------------------------------
-- Select the Read Enable used by the RAM Block.
-- If both links are active, each FIFO is 256x32
-- If one link is masked, the FIFO is 512x32
ren_register : process (
  rst_n_in, 
  clk_in
  )
begin
  if (rst_n_in = '0') then
    fifo0_ren_i <= '0';
    fifo1_ren_i <= '0';
  elsif (clk_in'event AND clk_in = '1') then
    fifo0_ren_i <= rden_i(0);
    fifo1_ren_i <= rden_i(1);
  end if;
end process ren_register;

-------------------------------------------------------------------------------
-- Select the Read Enable used by the RAM Block.
-- If both links are active, each FIFO is 256x32
-- If one link is masked, the FIFO is 512x32
ren_select : process (
  rst_n_in, 
  link_enable_in,
  rden_in,
  fifo0_rd_addr,
  fifo1_rd_addr
  )
begin
  if (rst_n_in = '0') then
    rden_i <= (others => '0');
  else
    if (link_enable_in(0) = '0') then
      rden_i(0) <= '0';
    else
      rden_i(0) <= rden_in(0);
    end if;

    if (link_enable_in(1) = '0' ) then
      rden_i(1) <= '0';
    else
      rden_i(1) <= rden_in(1);
    end if;
  end if;
end process ren_select;

-------------------------------------------------------------------------------
output_mux : process (
  rst_n_in,
  clk_in,
  link_enable_in,
  fifo0_dout_i,
  fifo1_dout_i,
  fifo0_ren_i,
  fifo1_ren_i 
  )
begin
  if (rst_n_in = '0') then
    fifo_out <= (others => '0');
  elsif (clk_in'event AND clk_in = '1') then
    if (link_enable_in = "00") then 
      fifo_out <= (others => '0');
    else 
      if (fifo0_ren_i = '1') then
        fifo_out <= fifo0_dout_i;
      elsif (fifo1_ren_i = '1') then
        fifo_out <= fifo1_dout_i;
      end if;
    end if;
  end if;
end process output_mux;
      
-------------------------------------------------------------------------------
-- output address counter
output_counter : process (
  rst_n_in, 
  clk_in,
  link_enable_in
  )
begin
  if (rst_n_in = '0') then
    fifo0_rd_addr <= (others => '0');
    fifo1_rd_addr <= (others => '0');
    data_valid0_i <= '0';
    data_valid1_i <= '0';
    fifo0_readout <= idle;
    fifo1_readout <= idle;
  elsif (clk_in'event and clk_in = '1') then
    if (link_enable_in(0) = '0') then
      fifo0_rd_addr <= (others => '0');
    else
      case fifo0_readout is
        when idle => 
          data_valid0_i <= '0';
          if (fifo0_empty_i = '0') then
            if (fifo0_dout_i(15 downto 13) = "010" OR fifo0_dout_i(31 downto 29) = "010") then
              fifo0_readout <= play_one_word;
            elsif (rden_i(0) = '1') then
              fifo0_readout <= play;
              fifo0_rd_addr <= fifo0_rd_addr + 1;
            end if;
          end if;

        when play =>
          if (fifo0_empty_i = '0') then
            if ((rden_i(0) = '1' AND (fifo0_dout_i(15 downto 13) = "010" OR fifo0_dout_i(31 downto 29) = "010")) OR
                (rden_i(0) = '1' AND  fifo0_rd_addr = fifo0_wr_addr)) then
              data_valid0_i <= '1';
              fifo0_readout <= done;
            elsif (rden_i(0) = '1') then
              fifo0_rd_addr <= fifo0_rd_addr + 1;
              data_valid0_i <= '1';
            else
              data_valid0_i <= '0';
            end if;
          else
            data_valid0_i <= '0';
          end if;
              
        when play_one_word =>
          if (fifo0_ren_i = '1') then
            data_valid0_i <= '1';
            fifo0_rd_addr <= fifo0_rd_addr + 1;
            fifo0_readout <= done;
          end if;

        when done =>
          data_valid0_i <= '0';
          if (rden_i(0) = '0') then
            fifo0_readout <= idle;
          end if;
      end case;          
    end if;
            
    if (link_enable_in(1) = '0') then
      fifo1_rd_addr <= (others => '0');
    else
      case fifo1_readout is
        when idle => 
          data_valid1_i <= '0';
          if (fifo1_empty_i = '0') then
            if (fifo1_dout_i(15 downto 13) = "010" OR fifo1_dout_i(31 downto 29) = "010") then
              fifo1_readout <= play_one_word;
            elsif (rden_i(1) = '1') then
              fifo1_readout <= play;
              fifo1_rd_addr <= fifo1_rd_addr + 1;
            end if;
          end if;
            
        when play =>
          if (fifo1_empty_i = '0') then
            if ((rden_i(1) = '1' AND (fifo1_dout_i(15 downto 13) = "010" OR fifo1_dout_i(31 downto 29) = "010")) OR
                (rden_i(1) = '1' AND  fifo1_rd_addr = fifo1_wr_addr)) then
              data_valid1_i <= '1';
              fifo1_readout <= done;
            elsif (rden_i(1) = '1') then
              fifo1_rd_addr <= fifo1_rd_addr + 1;
              data_valid1_i <= '1';
            else
              data_valid1_i <= '0';
            end if;
          else
            data_valid1_i <= '0';
          end if;
              
        when play_one_word =>
          if (fifo1_ren_i = '1') then
            data_valid1_i <= '1';
            fifo1_rd_addr <= fifo1_rd_addr + 1;
            fifo1_readout <= done;
          end if;

        when done =>
          data_valid1_i <= '0';
          if (rden_i(1) = '0') then
            fifo1_readout <= idle;
          end if;

        when others =>
          data_valid1_i <= '0';
          fifo1_readout <= idle;
      end case;          
    end if;
  end if;
end process output_counter;

-------------------------------------------------------------------------------
-- fifo word occupancy counter
occupancy_counter : process (
  rst_n_in, 
  clk_in
  )
begin
  if (rst_n_in = '0') then
    occupancy0_count <= (others => '0');
    occupancy1_count <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
    if (link_enable_in(0) = '1') then
      occupancy0_count <= fifo0_wr_addr - fifo0_rd_addr;
    else
      occupancy0_count <= (others => '0');
    end if;
          
    if (link_enable_in(1) = '1') then
      occupancy1_count <= fifo1_wr_addr - fifo1_rd_addr;
    else
      occupancy1_count <= (others => '0');
    end if;
  end if;
end process occupancy_counter;

fifo0_flags : process (
  rst_n_in, 
  occupancy0_count,
  link_enable_in
  )
begin
  if (rst_n_in = '0') then
    fifo0_empty_i <= '1';
    fifo0_full_i  <= '0';
  else
    if (occupancy0_count = 0) then
      fifo0_empty_i <= '1';
    else
      fifo0_empty_i <= '0';
    end if;     
      
    if (link_enable_in(0) = '1' AND occupancy0_count = 510) then
      fifo0_full_i <= '1';
    else
      fifo0_full_i <= '0';
    end if;     
  end if;
end process;

fifo1_flags : process (
  rst_n_in, 
  occupancy1_count,
  link_enable_in
  )
begin
  if (rst_n_in = '0') then
    fifo1_empty_i <= '1';
    fifo1_full_i  <= '0';
  else
    if (occupancy1_count = 0) then
      fifo1_empty_i <= '1';
    else
      fifo1_empty_i <= '0';
    end if;     
      
    if (link_enable_in(1) = '1' AND occupancy1_count = 510) then
      fifo1_full_i <= '1';
    else
      fifo1_full_i <= '0';
    end if;     
  end if;
end process;

end rtl;
