----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    dual_link_formatter.vhd
-- Description:
--    ROD Front End Decoder
--       Dual Channel Decoder wired to the Dual 256x32 FIFO
--       SCT Version
--
----------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.std_logic_arith.all;

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

entity dual_link_formatter is
  port(
    clk_in                 : in  std_logic; -- 40 MHz clock
    rclk_in                : in  std_logic; -- 80 MHz clock
    rst_n_in               : in  std_logic; -- Powerup global reset
    ecr_in                 : in  std_logic;

 -- enable link formatter activity
    link_enable_in         : in  std_logic_vector( 1 DOWNTO 0);
    link_mask_en_in        : in  std_logic_vector( 1 DOWNTO 0);

    expanded_mode_in       : in  std_logic_vector( 1 DOWNTO 0);
 -- 1 = uncondensed data mode,
 -- 0 = condensed data mode

    config_mode_in         : in  std_logic_vector( 1 DOWNTO 0);
 -- 1 = config mode, 0 = normal

    compressed_mode_in     : in  std_logic_vector( 1 DOWNTO 0);
 -- 1 = compressed mode, 0 = normal

    edge_mode_in           : in  std_logic_vector( 1 DOWNTO 0);
 -- 1 = edge detect "01X" for hit pattern
 -- 0 = level detect "X1X" for hit pattern

 -- front end serial data
    link_data_in           : in  std_logic_vector( 1 DOWNTO 0);

 -- Header/Trailer Limit value
    ht_limit_in            : in  unsigned(8 DOWNTO 0);

 -- ROD Busy Limit value
    rod_busy_limit_in      : in  unsigned(8 DOWNTO 0);

 -- fifo REN
    fifo_ren_in            : in  std_logic_vector( 1 DOWNTO 0);

 -- decoder is in 1 = normal data mode, 0 = raw data mode 
    normal_not_raw_out     : out std_logic_vector( 1 DOWNTO 0);

 -- trailer being written
    decoder_wr_trailer_out : out std_logic_vector( 1 DOWNTO 0);

    ht_limit_out           : out std_logic_vector( 1 DOWNTO 0);
    rod_busy_limit_out     : out std_logic_vector( 1 DOWNTO 0);
    fifo_ef_out            : out std_logic_vector( 1 DOWNTO 0);
    fifo_ff_out            : out std_logic_vector( 1 DOWNTO 0);
    trailer_detect_out     : out std_logic_vector(11 DOWNTO 0);
    occ_count0_out         : out std_logic_vector( 8 DOWNTO 0);
    occ_count1_out         : out std_logic_vector( 8 DOWNTO 0);
 -- Output data word
    fifo_data_out          : out std_logic_vector(31 DOWNTO 0);
    data_valid_out         : out std_logic; --
 -- Pipeline Counter Values
--    l1a1_pipeline_in       : in  std_logic_vector( 3 DOWNTO 0);
--    l1a2_pipeline_in       : in  std_logic_vector( 3 DOWNTO 0);
--    l1a3_pipeline_in       : in  std_logic_vector( 3 DOWNTO 0);
--    l1a4_pipeline_in       : in  std_logic_vector( 3 DOWNTO 0);
--    l1a5_pipeline_in       : in  std_logic_vector( 3 DOWNTO 0);
--    l1a6_pipeline_in       : in  std_logic_vector( 3 DOWNTO 0);
--    l1a7_pipeline_in       : in  std_logic_vector( 3 DOWNTO 0);
--    l1a8_pipeline_in       : in  std_logic_vector( 3 DOWNTO 0);
-- FE Pipleine Mask Values
--    link0_pipeline_mask    : in  std_logic_vector(31 DOWNTO 0);
--    link1_pipeline_mask    : in  std_logic_vector(31 DOWNTO 0);
    lnk0_masked_chip_id_in : in  std_logic_vector( 7 downto 0);
    lnk1_masked_chip_id_in : in  std_logic_vector( 7 downto 0);
    data_overflow_limit_in : in  unsigned(11 downto 0);
    hit_count_rst_in       : in  std_logic_vector( 1 DOWNTO 0);
    hit_count_out          : out STD_LOGIC_VECTOR(23 DOWNTO 0); 
    decoder_active_out     : out std_logic_vector( 1 DOWNTO 0);
	  dovr_error_out         : out std_logic_vector( 1 DOWNTO 0);
    link_data_tx_out       : out std_logic_vector( 1 DOWNTO 0)
    );
end dual_link_formatter ; 

architecture rtl of dual_link_formatter is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

signal ht_limit_i   : STD_LOGIC_VECTOR( 1 DOWNTO 0);
signal rb_limit_i   : STD_LOGIC_VECTOR( 1 DOWNTO 0);
signal fifo_wen_i   : STD_LOGIC_VECTOR( 1 DOWNTO 0);
signal fifo_data0_i : STD_LOGIC_VECTOR(31 DOWNTO 0);
signal fifo_data1_i : STD_LOGIC_VECTOR(31 DOWNTO 0);
signal occ_count0_i : UNSIGNED(8 DOWNTO 0);
signal occ_count1_i : UNSIGNED(8 DOWNTO 0);
signal link_en_i    : STD_LOGIC_VECTOR( 1 DOWNTO 0);
signal link_data_i  : std_logic_vector( 1 DOWNTO 0);
signal fifo_ef_o    : std_logic_vector( 1 DOWNTO 0);

signal decoder_active_o : STD_LOGIC_VECTOR( 1 DOWNTO 0);
signal link_data_tx_o   : STD_LOGIC_VECTOR( 1 DOWNTO 0);
signal hit_count_o      : STD_LOGIC_VECTOR(23 DOWNTO 0); 

 
----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------
component link_formatter
  port(
    clk_in             : in  STD_LOGIC;
    rst_n_in           : in  STD_LOGIC;
    ecr_in             : in  STD_LOGIC;
    almost_full        : in  STD_LOGIC;
    enable             : in  STD_LOGIC;
    expanded_mode      : in  STD_LOGIC;
    config_mode        : in  STD_LOGIC;
    compressed_mode    : in  STD_LOGIC;
    edge_mode          : in  STD_LOGIC;
    link_in            : in  STD_LOGIC;
    normal_not_raw     : out STD_LOGIC;
    data_strobe        : out STD_LOGIC;
    decoder_wr_trailer : out STD_LOGIC;
    decoded_data       : out STD_LOGIC_VECTOR(31 DOWNTO 0);
--    l1a1_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a2_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a3_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a4_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a5_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a6_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a7_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a8_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    link_pipeline_mask : in  std_logic_vector(31 DOWNTO 0);
    masked_chip_id_in      : in  std_logic_vector( 7 downto 0);
    data_overflow_limit_in : in  unsigned(11 downto 0); --
    hit_count_rst_in       : in std_logic;
    hit_count_out          : out std_logic_vector(11 DOWNTO 0);
    decoder_active_out     : out std_logic;
	 dovr_error_out         : out std_logic;
    link_data_tx_out       : out std_logic  -- 
    );
end component ; 

component dual_512x32fifo
  port(
    clk_in            : in  std_logic;
    rclk_in           : in  std_logic;
    rst_n_in          : in  std_logic;
    link_enable_in    : in  std_logic_vector( 1 downto 0);
    wren_in           : in  std_logic_vector( 1 downto 0);
    rden_in           : in  std_logic_vector( 1 downto 0);
    fifo0_in          : in  std_logic_vector(31 downto 0); --
    fifo1_in          : in  std_logic_vector(31 downto 0); --
    fifo_out          : out std_logic_vector(31 downto 0); --
    data_valid_out    : out std_logic; --
    fifo_ef_out       : out std_logic_vector( 1 downto 0);
    fifo_ff_out       : out std_logic_vector( 1 downto 0);
    fifo0_trailer_out : out std_logic_vector( 5 downto 0);
    fifo1_trailer_out : out std_logic_vector( 5 downto 0);
    occupancy0_out    : out unsigned(8 downto 0);  --
    occupancy1_out    : out unsigned(8 downto 0)   --
    );
end component ; 
  

BEGIN  --  Main Body of VHDL code
----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

decoder_active_out <= decoder_active_o; 
link_data_tx_out   <= link_data_tx_o;
hit_count_out      <= hit_count_o;

U1 : link_formatter
  port map(
    clk_in                 => clk_in,
    rst_n_in               => rst_n_in,
    ecr_in                 => ecr_in,
    almost_full            => ht_limit_i(0),
    enable                 => link_en_i(0),   --link_enable_in(0),
    expanded_mode          => expanded_mode_in(0),
    config_mode            => config_mode_in(0),
    compressed_mode        => compressed_mode_in(0),
    edge_mode              => edge_mode_in(0),
    link_in                => link_data_i(0),  --link_data_in(0),
    normal_not_raw         => normal_not_raw_out(0),
    data_strobe            => fifo_wen_i(0),
    decoder_wr_trailer     => decoder_wr_trailer_out(0),
    decoded_data           => fifo_data0_i,
--    l1a1_pipeline_in       => l1a1_pipeline_in,
--    l1a2_pipeline_in       => l1a2_pipeline_in,
--    l1a3_pipeline_in       => l1a3_pipeline_in,
--    l1a4_pipeline_in       => l1a4_pipeline_in,
--    l1a5_pipeline_in       => l1a5_pipeline_in,
--    l1a6_pipeline_in       => l1a6_pipeline_in,
--    l1a7_pipeline_in       => l1a7_pipeline_in,
--    l1a8_pipeline_in       => l1a8_pipeline_in,
--    link_pipeline_mask     => link0_pipeline_mask,
    masked_chip_id_in      => lnk0_masked_chip_id_in,
    data_overflow_limit_in => data_overflow_limit_in,
    hit_count_rst_in       => hit_count_rst_in(0),
    hit_count_out          => hit_count_o(11 downto 0),
    decoder_active_out     => decoder_active_o(0),
	  dovr_error_out         => dovr_error_out(0),
    link_data_tx_out       => link_data_tx_o(0)
    );

U2 : link_formatter
  port map(
    clk_in                 => clk_in,
    rst_n_in               => rst_n_in,
    ecr_in                 => ecr_in,
    almost_full            => ht_limit_i(1),
    enable                 => link_en_i(1),   --link_enable_in(1),
    expanded_mode          => expanded_mode_in(1),
    config_mode            => config_mode_in(1),
    compressed_mode        => compressed_mode_in(1),
    edge_mode              => edge_mode_in(1),
    link_in                => link_data_i(1),  --link_data_in(1),
    normal_not_raw         => normal_not_raw_out(1),
    data_strobe            => fifo_wen_i(1),
    decoder_wr_trailer     => decoder_wr_trailer_out(1),
    decoded_data           => fifo_data1_i,
--    l1a1_pipeline_in       => l1a1_pipeline_in,
--    l1a2_pipeline_in       => l1a2_pipeline_in,
--    l1a3_pipeline_in       => l1a3_pipeline_in,
--    l1a4_pipeline_in       => l1a4_pipeline_in,
--    l1a5_pipeline_in       => l1a5_pipeline_in,
--    l1a6_pipeline_in       => l1a6_pipeline_in,
--    l1a7_pipeline_in       => l1a7_pipeline_in,
--    l1a8_pipeline_in       => l1a8_pipeline_in,
--    link_pipeline_mask     => link1_pipeline_mask,
    masked_chip_id_in      => lnk1_masked_chip_id_in,
    data_overflow_limit_in => data_overflow_limit_in,
    hit_count_rst_in       => hit_count_rst_in(1),
    hit_count_out          => hit_count_o(23 downto 12),
    decoder_active_out     => decoder_active_o(1),
    dovr_error_out         => dovr_error_out(1),
    link_data_tx_out       => link_data_tx_o(1)
    );

U3 : dual_512x32fifo
  port map(
    clk_in            => clk_in,
    rclk_in           => rclk_in,
    rst_n_in          => rst_n_in,
    link_enable_in    => link_en_i, --link_enable_in,
    wren_in           => fifo_wen_i,
    rden_in           => fifo_ren_in,
    fifo0_in          => fifo_data0_i,
    fifo1_in          => fifo_data1_i,
    fifo_out          => fifo_data_out,
    data_valid_out    => data_valid_out,
    fifo_ef_out       => fifo_ef_o,
    fifo_ff_out       => fifo_ff_out,
    fifo0_trailer_out => trailer_detect_out( 5 downto 0),
    fifo1_trailer_out => trailer_detect_out(11 downto 6),
    occupancy0_out    => occ_count0_i,
    occupancy1_out    => occ_count1_i
    );

--------------------------------------------------------------------------
-- SIGNALS
--------------------------------------------------------------------------

occ_count0_out <= std_logic_vector(occ_count0_i);
occ_count1_out <= std_logic_vector(occ_count1_i);

ht_limit_out       <= ht_limit_i;
rod_busy_limit_out <= rb_limit_i;

fifo_ef_out <= fifo_ef_o;

--------------------------------------------------------------------------
-- PROCESS DECLARATION
--------------------------------------------------------------------------
register_signal : process (
  rst_n_in,
  clk_in
  )
begin
  if (rst_n_in = '0') then
    link_en_i   <= (others => '0');
	 link_data_i <= (others => '0');
  elsif (clk_in'event AND clk_in = '1') then
    link_en_i   <= link_enable_in AND NOT link_mask_en_in;
    link_data_i <= link_data_in   AND NOT link_mask_en_in;
  end if;
end process register_signal;

header_trailer_limit_comparator : process (
  rst_n_in,
  clk_in
  )
begin
  if (rst_n_in = '0') then
    ht_limit_i <= (others => '0');
  elsif (clk_in'event AND clk_in = '1') then
    if (link_enable_in(0) = '1') then
      if (occ_count0_i >= ht_limit_in) then
        ht_limit_i(0) <= '1';
      else
        ht_limit_i(0) <= '0';
      end if;      
    else
      ht_limit_i(0) <= '0';
    end if;

    if (link_enable_in(1) = '1') then
      if (occ_count1_i >= ht_limit_in) then
        ht_limit_i(1) <= '1';
      else
        ht_limit_i(1) <= '0';
      end if;      
    else
      ht_limit_i(1) <= '0';
    end if;
  end if;
end process header_trailer_limit_comparator;
     
rod_busy_limit_comparator : process (
  rst_n_in,
  clk_in
  )
begin
  if (rst_n_in = '0') then
    rb_limit_i <= (others => '0');
  elsif (clk_in'event AND clk_in = '1') then
    if (link_enable_in(0) = '1') then
      if (occ_count0_i >= rod_busy_limit_in) then
        rb_limit_i(0) <= '1';
      else
        rb_limit_i(0) <= '0';
      end if;      
    else
      rb_limit_i(0) <= '0';
    end if;

    if (link_enable_in(1) = '1') then
      if (occ_count1_i >= rod_busy_limit_in) then
        rb_limit_i(1) <= '1';
      else
        rb_limit_i(1) <= '0';
      end if;      
    else
      rb_limit_i(1) <= '0';
    end if;
  end if;
end process rod_busy_limit_comparator;
     
end rtl;

