-------------------------------------------------------------------------------
--
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 2000
-- ReadOutDriver Electronics
-------------------------------------------------------------------------------
--
-- Filename: fe_pipeline_counter.vhd
-- Description:
--   MOD 12 counter and 8x4 array to store pipeline numbers correlated to L1A
--   Used to mask stuck bits in the SCT FE Module Time Bin Pipelines
--
-------------------------------------------------------------------------------
-- Structure: 
--  Multiple state machines in a flat topology/hierarchy which communicate
--   with single wire signals.  
-------------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
-------------------------------------------------------------------------------
-- Author: John M. Joseph
-- Board Engineer: John M. Joseph
-- History:
--    24 Sep 2009 -  JMJ first version
-------------------------------------------------------------------------------
-- LIBRARY INCLUDES
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations
use IEEE.std_logic_unsigned.all; 
-------------------------------------------------------------------------------
-- PORT DECLARATION
-------------------------------------------------------------------------------
entity fe_pipeline_counter is
  port (
    clk_in    : in  std_logic; -- clk40 input
    rst_n_in  : in  std_logic; -- asynchronous global reset
    ecr_in    : in  std_logic; 
    l1a_in    : in  std_logic; 
    pipe0_out : out std_logic_vector(3 downto 0);
    pipe1_out : out std_logic_vector(3 downto 0);
    pipe2_out : out std_logic_vector(3 downto 0);
    pipe3_out : out std_logic_vector(3 downto 0);
    pipe4_out : out std_logic_vector(3 downto 0);
    pipe5_out : out std_logic_vector(3 downto 0);
    pipe6_out : out std_logic_vector(3 downto 0);
    pipe7_out : out std_logic_vector(3 downto 0)
    );
end fe_pipeline_counter;

architecture rtl of fe_pipeline_counter is
-------------------------------------------------------------------------------
-- SIGNAL DECLARATION
-------------------------------------------------------------------------------

signal l1a_count : unsigned(2 downto 0);
signal pipeline_count : unsigned(3 downto 0);

-------------------------------------------------------------------------------
-- PROCESS DECLARATION
-------------------------------------------------------------------------------
begin
-------------------------------------------------------------------------------
-- Pipeline Counter
-------------------------------------------------------------------------------
pipeline_counter : process (
  clk_in, 
  rst_n_in
  )
begin
  if (rst_n_in = '0') then
    l1a_count <= (others => '0');
    pipeline_count <= (others => '0');
    pipe0_out <= (others => '0');
    pipe1_out <= (others => '0');
    pipe2_out <= (others => '0');
    pipe3_out <= (others => '0');
    pipe4_out <= (others => '0');
    pipe5_out <= (others => '0');
    pipe6_out <= (others => '0');
    pipe7_out <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
    if (ecr_in = '1') then
      l1a_count <= (others => '0');
      pipeline_count <= (others => '0');
    elsif (pipeline_count = 11) then
      pipeline_count <= (others => '0');
    else
      pipeline_count <= pipeline_count + '1';
    end if;
    
    if (l1a_in = '1') then
      case l1a_count is
        when "000"  => pipe0_out <= std_logic_vector(pipeline_count);
        when "001"  => pipe1_out <= std_logic_vector(pipeline_count);
        when "010"  => pipe2_out <= std_logic_vector(pipeline_count);
        when "011"  => pipe3_out <= std_logic_vector(pipeline_count);
        when "100"  => pipe4_out <= std_logic_vector(pipeline_count);
        when "101"  => pipe5_out <= std_logic_vector(pipeline_count);
        when "110"  => pipe6_out <= std_logic_vector(pipeline_count);
        when others => pipe7_out <= std_logic_vector(pipeline_count);
      end case;
      l1a_count <= l1a_count + '1';
    end if;
  end if;  
end process pipeline_counter;

end rtl;
