------------------------------------------------------------------------------
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 2003
-- ReadOutDriver Electronics
------------------------------------------------------------------------------
-- Filename: fifo_readout_controller.vhd
-- Description:
--  Fifo ReadOut Controller
--    This logic block is used to control the read out of all 12 link fifos
--    in the SCT Formatter.
--
--  Inputs:
--    From Globals
--      Clock and Reset..(2)
--
--    From the Dual Link Formatters
--      Link FIFO Data.........(32 x 6)
--      FIFO Empty Status......( 2 x 6)
--      Trailer Detect Status..(12 x 6)
--
--    From the Mode Bits Decoder
--      Dynamic ModeBits......(24)
--      Mode Bits DataReady...( 1)
--      Mode Bits FIFO Empty..( 1)
--
--    From the Register Block
--      Link Enable..........................(12)
--      Condensed Mode Bit...................(12)
--      Token to FIFO Readout Timeout Limit..( 8) 
--      Data Overflow Limit..................( 9)
--
--    From the Trigger Counter / Token Generator
--      Token In..(1)
--
--    From the EFB
--      Hold Output..(1)
--
--  Outputs:
--    To the Dual Link Formatters
--      FIFO Read Enable..(12)
--      FIFO Flush........(12)
--
--    To the Mode Bits Decoder
--      MB FIFO Read Enable..(1)
--
--    To the Register Block
--      Time Out Error Out.......(12)
--      Data Overflow Error Out..(12)
--
--    To the Trigger Counter / Token Generator
--      Token Out..(1)
--
--    To the EFB
--      Formatter FIFO Data Out..(32)
--      Active Link Number.......( 4)
--      Time Out Error Out.......( 1)
--      Condensed Mode Bit.......( 1)
--      Data Valid...............( 1)
--      Chip Has Token...........( 1)
--
--
-- Read Out Mode Definitions: 
--  +----+----+------------------------------------------------------------+
--  |Mode| MB |                  Description                               |
--  +----+----+------------------------------------------------------------+
--  | M0 |[00]| Play 1 event from the Link FIFO, normal data flow from the |
--  |    |    | Formatter to the EFB.                                      |
--  +----+----+------------------------------------------------------------+
--  | M1 |[01]| Mask this link for 1 trigger.  This action dumps one event |
--  |    |    | then increments to the next link FIFO.                     |
--  +----+----+------------------------------------------------------------+
--  | M2 |[10]| Skip this link for 1 trigger.  This action does not change |
--  |    |    | any of the counters or pointers used by the Link FIFO.     |
--  +----+----+------------------------------------------------------------+
--  | M3 |[11]| Dump (play data, suppress Data Valid) 1st event in FIFO,   |
--  |    |    | play 2nd event to the EFB.                                 |
--  +----+----+------------------------------------------------------------+
--
--  Readout States:
--    Idle (IDLE) : 
--      Wait for Mode Bits from ROD Controller FPGA or ModeBits FIFO (512x24).
--      Mode bits for all links are loaded at the same time.  This simplifies
--      and minimizes FPGA utilization, and does not affect the performance
--      of the Formatter because the data is always played out in link order
--      for each event.
--
--    Fetch Mode Bits (FMB) :
--      Send a 1 Clock pulse Read Enable Pulse to the Mode Bits FIFO to fetch
--      Read Out Mode bits for the current Trigger
--
--    Wait For Token (WFT) :
--      After Mode Bits are loaded, Formatter is ready for the token to arrive.
--      In this state, the link enable signal is checked to determine the
--      active links in the chip, the occupancy status of each Link FIFO is 
--      checked to see if any data has been decoded, and the Mode Bits are
--      monitored to determine the proper action to perform.  While waiting
--      for the token, the link counter can be incremented to the first
--      active link (enable_link_in).  When the token arrives, the state 
--      machine will branch to a different state depending on the value
--      of the Mode Bits and the Link Enable Signal.
--
--    Read Out Link FIFO (ROLF) :
--      Begin readout of a link FIFO.  If a trailer is present when first
--      entering this state, the FIFO has a Header/Trailer Word only for 
--      this trigger, and can begin readout of the next link on the next 
--      clock.  
--
--    Read Out Link FIFO / Set Data Valid (ROLF_SDV) :
--
--    Wait for Data (WFD) :
--      If readout is true for a specific link and the link fifo is empty,
--      wait in this state until the empty flag is clear, go to ROLF
--
--    Dump One Event (DOE) :
--      Read one event from the Link FIFO, no Data Valid signal
--
--    Next Link (NLINK) :
--      Increment to the next active link
--
--    Done (DONE):
--      Pass the Token to the next Formatter and return to Idle
--
--    Pause Output (PAUSE):
--      If hold_output_in = '1' then stop data output and hold token.  This 
--      state condition is tested outside of the main case statement.
--
------------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk40
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
------------------------------------------------------------------------------
-- Author: John Joseph
-- Board Engineer: 
-- History:
------------------------------------------------------------------------------
-- LIBRARY INCLUDES
------------------------------------------------------------------------------
--
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations
--
-------------------------------------------------------------------------------
-- PORT DECLARATION
-------------------------------------------------------------------------------
entity fifo_readout_controller is
  port(
-- Inputs:    
    clk_in                  : in  std_logic;
    rst_n_in                : in  std_logic;
    enable_link_in          : in  std_logic_vector(11 downto 0);
    link_mask_en_in         : in  std_logic_vector(11 downto 0);
    link_mask_en_out        : out std_logic_vector(11 downto 0);
    condensed_mode_bit_in   : in  std_logic_vector(11 downto 0);
    token_in                : in  std_logic;
    hold_output_in          : in  std_logic;
    fifo_timeout_limit_in   : in  unsigned( 7 downto 0); -- 
--    data_overflow_limit_in  : in  unsigned(11 downto 0); -- 
    link01_fifo_data_in     : in  std_logic_vector(31 downto 0);
    link23_fifo_data_in     : in  std_logic_vector(31 downto 0);
    link45_fifo_data_in     : in  std_logic_vector(31 downto 0);
    link67_fifo_data_in     : in  std_logic_vector(31 downto 0);
    link89_fifo_data_in     : in  std_logic_vector(31 downto 0);
    linkAB_fifo_data_in     : in  std_logic_vector(31 downto 0);
    data_valid_in           : in  std_logic_vector( 5 downto 0);
    link_fifo_empty_in      : in  std_logic_vector(11 downto 0);
    link_fifo_full_in       : in  std_logic_vector(11 downto 0);
    link01_trailer_det_in   : in  std_logic_vector(11 downto 0);
    link23_trailer_det_in   : in  std_logic_vector(11 downto 0);
    link45_trailer_det_in   : in  std_logic_vector(11 downto 0);
    link67_trailer_det_in   : in  std_logic_vector(11 downto 0);
    link89_trailer_det_in   : in  std_logic_vector(11 downto 0);
    linkAB_trailer_det_in   : in  std_logic_vector(11 downto 0);
    decoder_wr_trailer_in   : in  std_logic_vector(11 downto 0);
    mb_ready_in             : in  std_logic;
    mb_fifo_empty_in        : in  std_logic;
    mode_bits_in            : in  std_logic_vector(23 downto 0); 
-- Outputs:    
    formatter_fifo_data_out : out std_logic_vector(31 downto 0);
    link_num_has_token_out  : out std_logic_vector( 3 downto 0);
    timeout_error_out       : out std_logic;
    condensed_mode_out      : out std_logic;
    link_masked_out         : out std_logic;
    data_valid_out          : out std_logic;
    chip_has_token_out      : out std_logic;
    timeout_error_status    : out std_logic_vector(11 downto 0);
--    data_overflow_status    : out std_logic_vector(11 downto 0);
    mb_ren_out              : out std_logic;
    fifo_ren_out            : out std_logic_vector(11 downto 0);
    token_out               : out std_logic;
    readout_state_out       : out std_logic_vector( 3 downto 0);
    roc_ready_out           : out std_logic;
    cal_mode_sel_in         : in  std_logic;
    decoder_active_in       : in  std_logic_vector(11 DOWNTO 0) -- 
    );
end fifo_readout_controller;

architecture rtl of fifo_readout_controller is
--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------

type ro_states is (
  reset,       --                                        
  idle,        -- wait for mode                          
  fmb,         -- fetch current Mode Bits                
  wmb,         -- wait 1 clk for current Mode Bits                
  lmb,         -- latch current Mode Bits                
  wft,         -- wait for token                         
  rolf,        -- read out link fifo                     
  wfd,         -- wait for data                          
  wfd_td,      -- trailer detected in wait for data      
  doe,         -- dump one event from the link fifo
  roe,         -- read one event from the link fifo
  nlnk,        -- start readout of next link
  err_det,     -- Timeout, Data Overflow or Link Masked detected
  send_lmhtw,  -- Send Link Masked Header/Trailer word
  send_lmtw,   -- Send Link Masked Header/Trailer word
--  trto,        -- token to readout timeout error
--  dovr,        -- data overflow condition error
  pause,       -- back pressure from EFB
  wrlw,        -- write last word to hold fifo
  done,        -- done with link event, pass token
  pass_token   -- pass token
  );
signal readout_state : ro_states;

type t_states is (
  idle,        -- wait for mode                          
  chip_has_token,
  ready_to_pass_token
  );
signal token_state : t_states;

type   trailer_detect_array is array (0 to 23) of std_logic_vector(2 downto 0); --
signal trailer_detect_i   : trailer_detect_array; --

signal to_count           : unsigned(7 downto 0);
signal to_bit_count       : unsigned(4 downto 0);
signal to_error_i         : std_logic;
signal global_to_error_i  : std_logic;
signal en_error_i         : std_logic;
signal clr_to_counter_i   : std_logic;
signal en_to_counter_i    : std_logic;
signal dovr_count         : unsigned(11 downto 0);
--signal dovr_error_i       : std_logic;
signal clr_dovr_counter_i : std_logic;
signal inc_dovr_counter_i : std_logic;
signal link_count         : unsigned(3 downto 0);
signal output_link_count  : unsigned(3 downto 0);
signal clr_link_counter_i : std_logic;
signal mask_data_valid_i  : std_logic;
signal trailer_in_true_i  : std_logic;
signal chip_has_token_i   : std_logic;
signal link_mb_i          : std_logic_vector(1 downto 0);
signal fifo_ren_i         : std_logic_vector(11 downto 0);
signal link_fifo_empty_i  : std_logic;
signal condensed_mode_i   : std_logic;
signal to_status_i        : std_logic_vector(11 downto 0);
--signal dovr_status_i      : std_logic_vector(11 downto 0);
signal link               : natural RANGE 0 to 12;       
signal link_num           : natural RANGE 0 to 12;

signal hold_din_i   : std_logic_vector(39 downto 0);
signal hold_dout_i  : std_logic_vector(39 downto 0);
signal hold_wr_addr : unsigned(8 downto 0);
signal hold_rd_addr : unsigned(8 downto 0);
signal hold_wr_i    : std_logic;
signal hold_rd_i    : std_logic;

signal send_masked_htw : std_logic;
signal send_masked_tw  : std_logic;
signal link_masked_i   : std_logic_vector(1 downto 0);
signal mask_cht_i      : std_logic;
signal link_mask_en_i  : std_logic_vector(11 downto 0);
signal rolf_flag_i     : std_logic;

--------------------------------------------------------------------------
-- COMPONENT DECLARATION
--------------------------------------------------------------------------

component dpram_512x8
  port(
    clk_in           : in  std_logic;
    rst_n_in         : in  std_logic;
    dprm_writein_bus : in  std_logic_vector(7 downto 0);
    writein_addr     : in  unsigned(8 downto 0);
    readout_addr     : in  unsigned(8 downto 0);
    write_strobe     : in  std_logic;
    read_strobe      : in  std_logic;
    dprm_readout_bus : out std_logic_vector(7 downto 0)
    );
end component;

begin
-------------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
-------------------------------------------------------------------------------
U1 : for i in 0 to 4 generate 
  MB1 : dpram_512x8
    port map (
      clk_in           => clk_in,
      rst_n_in         => rst_n_in,
      dprm_writein_bus => hold_din_i((i*8)+7 downto (i*8)),
      writein_addr     => hold_wr_addr(8 downto 0),
      readout_addr     => hold_rd_addr(8 downto 0),
      write_strobe     => hold_wr_i,
      read_strobe      => hold_rd_i,
      dprm_readout_bus => hold_dout_i((i*8)+7 downto (i*8))
      );
end generate;

--------------------------------------------------------------------------
-- SIGNALS
--------------------------------------------------------------------------

formatter_fifo_data_out <= hold_dout_i(31 downto  0);
link_num_has_token_out  <= hold_dout_i(35 downto 32);
timeout_error_out       <= hold_dout_i(36);
condensed_mode_out      <= hold_dout_i(37);
link_masked_out         <= hold_dout_i(38);
--link_mask_en_out        <= link_mask_en_i;

hold_wr_i <= inc_dovr_counter_i;
hold_rd_i <= NOT hold_output_in;

fifo_ren_out         <= fifo_ren_i;
timeout_error_status <= to_status_i;
--data_overflow_status <= dovr_status_i;

trailer_detect_i( 0) <= link01_trailer_det_in( 2 downto 0);
trailer_detect_i( 1) <= link01_trailer_det_in( 5 downto 3);
trailer_detect_i( 2) <= link01_trailer_det_in( 8 downto 6);
trailer_detect_i( 3) <= link01_trailer_det_in(11 downto 9);
trailer_detect_i( 4) <= link23_trailer_det_in( 2 downto 0);
trailer_detect_i( 5) <= link23_trailer_det_in( 5 downto 3);
trailer_detect_i( 6) <= link23_trailer_det_in( 8 downto 6);
trailer_detect_i( 7) <= link23_trailer_det_in(11 downto 9);
trailer_detect_i( 8) <= link45_trailer_det_in( 2 downto 0);
trailer_detect_i( 9) <= link45_trailer_det_in( 5 downto 3);
trailer_detect_i(10) <= link45_trailer_det_in( 8 downto 6);
trailer_detect_i(11) <= link45_trailer_det_in(11 downto 9);
trailer_detect_i(12) <= link67_trailer_det_in( 2 downto 0);
trailer_detect_i(13) <= link67_trailer_det_in( 5 downto 3);
trailer_detect_i(14) <= link67_trailer_det_in( 8 downto 6);
trailer_detect_i(15) <= link67_trailer_det_in(11 downto 9);
trailer_detect_i(16) <= link89_trailer_det_in( 2 downto 0);
trailer_detect_i(17) <= link89_trailer_det_in( 5 downto 3);
trailer_detect_i(18) <= link89_trailer_det_in( 8 downto 6);
trailer_detect_i(19) <= link89_trailer_det_in(11 downto 9);
trailer_detect_i(20) <= linkAB_trailer_det_in( 2 downto 0);
trailer_detect_i(21) <= linkAB_trailer_det_in( 5 downto 3);
trailer_detect_i(22) <= linkAB_trailer_det_in( 8 downto 6);
trailer_detect_i(23) <= linkAB_trailer_det_in(11 downto 9);

-------------------------------------------------------------------------
-- PROCESS DECLARATION
--------------------------------------------------------------------------
-- Hold FIFO Counters
hold_fifo_counters : process (
  clk_in,
  rst_n_in
  )
begin
  if (rst_n_in = '0') then
    hold_wr_addr <= (others =>'0');
    hold_rd_addr <= (others =>'0');
    data_valid_out <= '0';
  elsif (clk_in'event AND clk_in = '1') then
--  if (hold_output_in = '1') then
      if (inc_dovr_counter_i = '1') then
        hold_wr_addr <= hold_wr_addr + 1 ;
      end if;
--  end if;

    if (hold_output_in = '0') then
      if (hold_wr_addr /= hold_rd_addr) then
        hold_rd_addr <= hold_rd_addr + 1 ;
        data_valid_out <= '1';
      else
        data_valid_out <= '0';
      end if;
    else
      data_valid_out <= '0';
    end if;
  end if;
end process;

-- State machine next state determination logic 
ro_state_logic : process (
  clk_in,
  rst_n_in
  )
begin
  if (rst_n_in = '0') then
    chip_has_token_i   <= '0';
    mb_ren_out         <= '0';
    fifo_ren_i         <= (others => '0');
    token_out          <= '0';
    link_mb_i          <= "00";
    link_count         <= (others => '0');
    link_num           <=  0 ;
    en_to_counter_i    <= '0';
    clr_to_counter_i   <= '1';
    clr_dovr_counter_i <= '1';
    clr_link_counter_i <= '1';
    output_link_count  <= (others => '1');
    readout_state      <= reset;
    token_state        <= idle;
    mask_data_valid_i  <= '0';
    trailer_in_true_i  <= '0';
    readout_state_out  <= (others => '0');
    en_error_i         <= '0';
    roc_ready_out      <= '0';
    send_masked_htw    <= '0';
    send_masked_tw     <= '0';
    link_masked_i      <= "00";
    mask_cht_i         <= '1';
    link_mask_en_i     <= (others => '0');
    link_mask_en_out   <= (others => '0');
    rolf_flag_i        <= '0';
  elsif (clk_in'event AND clk_in = '1') then
    link_mask_en_i <= link_mask_en_in;
	 
    if (clr_link_counter_i = '1') then
      link_num <=  0 ;
    end if;      

    link_count <= CONV_UNSIGNED(link_num,4);
    link       <= link_num;

    if (enable_link_in = "000000000000") then
      readout_state      <= idle;
      chip_has_token_i   <= token_in;
      mask_cht_i         <= NOT token_in;
      mb_ren_out         <= mb_ready_in;
      fifo_ren_i         <= (others => '0');
      token_out          <= token_in OR (chip_has_token_i AND mask_cht_i);
      clr_to_counter_i   <= '1';
      clr_dovr_counter_i <= '1';
      clr_link_counter_i <= '1';
      en_error_i         <= '0';
      readout_state_out  <= X"F";
      output_link_count  <= (others => '1');
      roc_ready_out      <= '1';
      send_masked_htw    <= '0';
      send_masked_tw     <= '0';
      link_masked_i      <= "00";
      rolf_flag_i        <= '0';
      token_state        <= idle;
    else
      output_link_count <= link_count;
      en_error_i        <= '1';

      case token_state is
        when idle =>
          chip_has_token_i <= '0';
          token_out        <= '0';
          if (token_in = '1') then
            token_state <= chip_has_token;
          end if;
          
        when chip_has_token =>
          chip_has_token_i <= '1';
          if (readout_state = done) then
            token_state <= ready_to_pass_token;
          end if;

        when ready_to_pass_token =>
          chip_has_token_i <= '0';
          token_out        <= '1';
          token_state <= idle;
          
        when others => 
          chip_has_token_i <= '0';
          token_out        <= '0';
          token_state <= idle;
      end case;

      case readout_state is
        when reset =>
          readout_state_out  <= X"1";
--          chip_has_token_i   <= '0';
--          token_out          <= '0';
          mb_ren_out         <= '0';
          fifo_ren_i         <= (others =>'0');
          en_to_counter_i    <= '0';
          clr_to_counter_i   <= '1';
          clr_dovr_counter_i <= '1';
          clr_link_counter_i <= '1';
          mask_data_valid_i  <= '0';
          trailer_in_true_i  <= '0';
          roc_ready_out      <= '0';
          send_masked_htw    <= '0';
          send_masked_tw     <= '0';
          link_masked_i      <= "00";
          rolf_flag_i        <= '0';
          readout_state      <= idle;

        when idle =>
          readout_state_out  <= X"2";
--          chip_has_token_i   <= token_in;   --'0';
--          token_out          <= token_in;   --'0'; If no MBs, but a token arrives, pass it
          mb_ren_out         <= '0';
          fifo_ren_i         <= (others =>'0');
          en_to_counter_i    <= '0';
          clr_to_counter_i   <= '1';
          clr_dovr_counter_i <= '0';
          clr_link_counter_i <= '0';
          mask_data_valid_i  <= '0';
          trailer_in_true_i  <= '0';
          roc_ready_out      <= '0';
          send_masked_htw    <= '0';
          send_masked_tw     <= '0';
          link_masked_i      <= "00";
          rolf_flag_i        <= '0';
          link_mask_en_out   <= link_mask_en_i;

          if (mb_ready_in = '1' OR mb_fifo_empty_in = '0') then
            mb_ren_out    <= '1';
            readout_state <= fmb;
          end if;

        when fmb =>  -- Fetch Current Mode Bits
          mb_ren_out    <= '0';
          link_masked_i <= "00";
          readout_state <= wmb;

        when wmb =>  -- Wait 1 clk for ModeBits to arrive
          mb_ren_out    <= '0';
          readout_state <= lmb;

        when lmb =>  -- Wait for current Modebits to arrive
          readout_state_out <= X"3";
          if (enable_link_in(link_num) = '0') then
            link_num  <= link_num + 1 ;
          elsif (enable_link_in(link_num) = '1') then
            link_mb_i     <= mode_bits_in((2*link_num)+1) & mode_bits_in(2*link_num);
            readout_state <= wft;
            roc_ready_out <= '1';
          end if;
            
        when wft =>  -- Wait for the Token to arrive/ Find the first active link
          readout_state_out <= X"4";
          if (token_in = '1') then
--            chip_has_token_i  <= '1';
            roc_ready_out     <= '0';
            if (mode_bits_in = X"AAAAAA") then
              readout_state <= done;
            elsif (link_mask_en_i(link_num) = '1') then
              readout_state <= err_det;
              link_masked_i <= "01";
            else
              case link_mb_i is
--  ** Do Not Delete these Comments **  --
--              when "00"   =>
--                readout_state        <= rolf;
--                fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);
--              when "01"   =>
--                readout_state        <= doe;
--                fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);
                when "10"   =>   -- skip link
                  readout_state <= nlnk;
                  link_num      <= link_num + 1 ;
                  link_mb_i     <= mode_bits_in((2*(link_num+1))+1) & mode_bits_in(2*(link_num+1));
                when "11"   =>
                  readout_state        <= doe;
                  fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);
                when others =>
                  fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);
                  if (hold_output_in = '0') then
                    if (link_fifo_empty_in(link_num) = '1') then
                      readout_state <= wfd;
                    else
                      readout_state <= rolf;
                      rolf_flag_i   <= '1';
--                    fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);
                    end if;
                  else
                    readout_state <= pause;
                  end if;
              end case;
            end if;
          end if;  
          
        when rolf => 
          readout_state_out    <= X"5";
          clr_dovr_counter_i   <= '0';
          en_to_counter_i      <= '0';
          clr_to_counter_i     <= '1'; --cal_mode_sel_in;
          fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);  --'1';
          roc_ready_out        <= '0';

          if (link_mask_en_i(link_num) = '1') then
            readout_state <= err_det;
            link_masked_i <= "10";
          elsif (link_fifo_empty_in(link_num) = '1') then
            readout_state <= wfd;
          elsif (link_count < 11 AND (trailer_detect_i((2*link_num)+1) = "010" OR 
                                      trailer_detect_i((2*link_num)  ) = "010" OR 
                                      trailer_in_true_i = '1')                   ) then
            link_num          <= link_num + 1 ;
            link_mb_i         <= mode_bits_in((2*(link_num+1))+1) & mode_bits_in(2*(link_num+1));
            trailer_in_true_i <= '0';
            readout_state     <= nlnk;
          elsif (link_count = 11 AND (trailer_detect_i((2*link_num)+1) = "010" OR 
                                      trailer_detect_i((2*link_num)  ) = "010" OR 
                                      trailer_in_true_i = '1')                   ) then
            trailer_in_true_i <= '0';
            readout_state     <= wrlw;
          elsif (link_count > 11) then
            readout_state <= done;
          end if;

          if (link_mb_i = "01") then
            mask_data_valid_i <= '1';
          else
            mask_data_valid_i <= '0';
          end if;
          
-- If needed, recover "wfd" from s3Ff (but has the Start/Stop CLK/2 issue)
-- s43f implemented 1 change only with original TO counter (4.2f uses new,
-- smaller incremet TO Counter)
        when wfd =>
          readout_state_out    <= X"6";
          fifo_ren_i(link_num) <= '0';
          clr_to_counter_i     <= '0';
          en_to_counter_i      <= link_fifo_empty_in(link_num) AND NOT decoder_active_in(link_num);
          if (link_mask_en_i(link_num) = '1') then
            readout_state <= err_det;
            link_masked_i <= "10";
          elsif (to_error_i = '1' OR global_to_error_i = '1') then
            readout_state <= err_det;
            link_masked_i <= "00";
          elsif (link_fifo_empty_in(link_num) = '0') then
            if (trailer_detect_i((2*link_num)+1) = "010" OR trailer_detect_i(2*link_num) = "010") then
              readout_state <= wfd_td;
              trailer_in_true_i <= '1';
            else
              readout_state <= rolf;
            end if;
          end if;

        when wfd_td =>
          readout_state        <= rolf;
          fifo_ren_i(link_num) <= '1';

        when doe =>
          readout_state_out    <= X"7";
          fifo_ren_i(link_num) <= '1';
          mask_data_valid_i    <= '1';
          roc_ready_out        <= '0';
          if (trailer_detect_i((2*link_num)+1) = "010" OR 
              trailer_detect_i((2*link_num)  ) = "010" OR
              link_fifo_empty_in(link_num) = '1'         ) then
            readout_state        <= roe;
            fifo_ren_i(link_num) <= '0';
          end if;

        when roe =>
         if (hold_output_in = '0' AND hold_wr_addr = hold_rd_addr) then
            readout_state        <= rolf;
            fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);
          end if;

        when nlnk =>
          readout_state_out      <= X"8";
          fifo_ren_i(link_num-1) <= '0';
          trailer_in_true_i      <= '0';
          clr_dovr_counter_i     <= '1';
          clr_to_counter_i       <= '1'; --cal_mode_sel_in;
          roc_ready_out          <= '0';
          send_masked_htw        <= '0';
          send_masked_tw         <= '0';
          rolf_flag_i            <= '0';
          if (link_num > 11) then
            readout_state <= done;
          elsif (enable_link_in(link_num) = '1') then
            if (link_mask_en_i(link_num) = '1') then
              readout_state <= err_det;
              link_masked_i <= "01";
            else
              case link_mb_i is
--              when "00"   =>
--                readout_state <= rolf;
--                fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);
--              when "01"   => 
--                readout_state        <= doe;
--                fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);
                when "10"   =>   -- skip link
                  readout_state <= nlnk;
                  link_num      <= link_num + 1 ;
                  link_mb_i     <= mode_bits_in((2*(link_num+1))+1) & mode_bits_in(2*(link_num+1));
                when "11"   =>
                  readout_state        <= doe;
                  fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);
                when others =>
                  if (hold_output_in = '0') then
                    readout_state        <= rolf;
                    fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);
                    rolf_flag_i          <= NOT link_fifo_empty_in(link_num);
                  else
                    readout_state <= pause;
                  end if;
              end case;
            end if;
          else
            link_num  <= link_num + 1 ;
            link_mb_i <= mode_bits_in((2*(link_num+1))+1) & mode_bits_in(2*(link_num+1));
          end if;

        when err_det =>
          readout_state_out  <= X"9";
          en_to_counter_i    <= '0';
          clr_to_counter_i   <= '1'; --cal_mode_sel_in;
          clr_dovr_counter_i <= '1';
          if (link_masked_i = "01") then
            readout_state <= send_lmhtw;
          elsif (link_masked_i = "10") then
            readout_state <= send_lmtw;
          elsif (link_num < 11) then
            readout_state <= nlnk;
            link_num      <= link_num + 1 ;
            link_mb_i     <= mode_bits_in((2*(link_num+1))+1) & mode_bits_in(2*(link_num+1));
          else
            readout_state <= done;
          end if;     
          link_mask_en_out <= link_mask_en_i;

        when send_lmhtw =>
          link_masked_i <= "00";
          send_masked_htw <= '1';
          send_masked_tw  <= '0';
          if (link_num < 11) then
            readout_state <= nlnk;
            link_num      <= link_num + 1 ;
            link_mb_i     <= mode_bits_in((2*(link_num+1))+1) & mode_bits_in(2*(link_num+1));
          else
            readout_state <= done;
          end if;            

        when send_lmtw =>
          link_masked_i <= "00";
          send_masked_htw <= '0';
          send_masked_tw  <= '1';
          if (link_num < 11) then
            readout_state <= nlnk;
            link_num      <= link_num + 1 ;
            link_mb_i     <= mode_bits_in((2*(link_num+1))+1) & mode_bits_in(2*(link_num+1));
          else
            readout_state <= done;
          end if;            

        when pause => 
          readout_state_out <= X"B";
          fifo_ren_i(link_num) <= '0';
          if (hold_output_in = '0' AND hold_wr_addr = hold_rd_addr) then
            readout_state        <= rolf;
            fifo_ren_i(link_num) <= NOT link_fifo_empty_in(link_num);
          end if;
            
        when wrlw =>
          readout_state <= done;

        when done =>
          readout_state_out <= X"C";
          if (hold_wr_addr = hold_rd_addr) then
--            token_out     <= '1';
            readout_state <= pass_token;
          end if;
          send_masked_htw <= '0';
          send_masked_tw  <= '0';
          fifo_ren_i      <= (others => '0');

        when pass_token =>
          readout_state_out  <= X"C";
--          token_out          <= '0';
          en_to_counter_i    <= '0';
          clr_to_counter_i   <= '1'; --cal_mode_sel_in;
          clr_dovr_counter_i <= '1';
          clr_link_counter_i <= '1';
          readout_state      <= idle;
          fifo_ren_i         <= (others => '0');
           
        when others =>
          readout_state_out <= X"E";
          if (link_num < 11) then
            readout_state <= nlnk;
            link_num      <= link_num + 1 ;
            link_mb_i     <= mode_bits_in((2*(link_num+1))+1) & mode_bits_in(2*(link_num+1));
          else
            readout_state <= done;
          end if;            
      end case;
    end if;
  end if;
end process ro_state_logic;

output_data_select : process (
  rst_n_in,
  clk_in
  )
begin
  if (rst_n_in = '0') then
    hold_din_i(31 downto 0) <= (others => '0');
    inc_dovr_counter_i      <= '0';
  elsif (clk_in'event AND clk_in = '1') then
    if (send_masked_htw = '1') then
      hold_din_i(31 downto 0) <= X"20004000";
      inc_dovr_counter_i      <= send_masked_htw;
    elsif (send_masked_tw = '1') then
      hold_din_i(31 downto 0) <= X"44000000";
      inc_dovr_counter_i      <= send_masked_tw;
    elsif (to_error_i = '1' AND en_error_i = '1') then
      if (rolf_flag_i = '0') then
        hold_din_i(31 downto 0) <= X"20004000";
      else
        hold_din_i(31 downto 0) <= X"44000000";
      end if;
      inc_dovr_counter_i      <= to_error_i;
    elsif (readout_state = doe) then
      hold_din_i(31 downto 0) <= (others => '0');
      inc_dovr_counter_i      <= '0';
    else
      case output_link_count(3 downto 1) is 
        when "000"  => hold_din_i(31 downto 0) <= link01_fifo_data_in;
                       inc_dovr_counter_i      <= data_valid_in(0) AND NOT mask_data_valid_i;
        when "001"  => hold_din_i(31 downto 0) <= link23_fifo_data_in;
                       inc_dovr_counter_i      <= data_valid_in(1) AND NOT mask_data_valid_i;
        when "010"  => hold_din_i(31 downto 0) <= link45_fifo_data_in;
                       inc_dovr_counter_i      <= data_valid_in(2) AND NOT mask_data_valid_i;
        when "011"  => hold_din_i(31 downto 0) <= link67_fifo_data_in;
                       inc_dovr_counter_i      <= data_valid_in(3) AND NOT mask_data_valid_i;
        when "100"  => hold_din_i(31 downto 0) <= link89_fifo_data_in;
                       inc_dovr_counter_i      <= data_valid_in(4) AND NOT mask_data_valid_i;
        when "101"  => hold_din_i(31 downto 0) <= linkAB_fifo_data_in;
                       inc_dovr_counter_i      <= data_valid_in(5) AND NOT mask_data_valid_i;
        when others => hold_din_i(31 downto 0) <= (others => '0');
                       inc_dovr_counter_i      <= '0';
      end case;
    end if;
  end if;
end process output_data_select;

output_control_reg : process (
  rst_n_in,
  clk_in
  )
begin
  if (rst_n_in = '0') then
    chip_has_token_out       <= '0';
    hold_din_i(39 downto 32) <= (others => '0');
  elsif (clk_in'event AND clk_in = '1') then
    chip_has_token_out       <= chip_has_token_i;
    hold_din_i(38)           <= send_masked_htw;  --link_mask_en_in(link);
    hold_din_i(37)           <= NOT condensed_mode_bit_in(link);
    hold_din_i(36)           <= to_error_i;
    hold_din_i(35 downto 32) <= std_logic_vector(output_link_count);
  end if;
end process output_control_reg;

error_status : process (
  rst_n_in,
  clk_in
  )
begin
  if (rst_n_in = '0') then
    to_status_i   <= (others => '0');
--    dovr_status_i <= (others => '0');
  elsif (clk_in'event AND clk_in = '1') then
--    dovr_status_i <= (others => '0');
    case output_link_count is 
      when "0000" => 
        to_status_i   <= "00000000000" & to_error_i;
      when "0001" =>
        to_status_i   <= "0000000000" & to_error_i   & '0';
      when "0010" =>
        to_status_i   <= "000000000" & to_error_i   & "00";
      when "0011" =>
        to_status_i   <= "00000000" & to_error_i   & "000";
      when "0100" =>
        to_status_i   <= "0000000" & to_error_i   & "0000";
      when "0101" =>
        to_status_i   <= "000000" & to_error_i   & "00000";
      when "0110" =>
        to_status_i   <= "00000" & to_error_i   & "000000";
      when "0111" => 
        to_status_i   <= "0000" & to_error_i   & "0000000";
      when "1000" =>
        to_status_i   <= "000" & to_error_i   & "00000000";
      when "1001" =>
        to_status_i   <= "00" & to_error_i   & "000000000";
      when "1010" =>
        to_status_i   <= '0' & to_error_i   & "0000000000";
      when "1011" =>
        to_status_i   <= to_error_i   & "00000000000";
      when others =>
        to_status_i   <= (others => '0');
    end case;
  end if;
end process error_status;

timeout_counter : process (
  clk_in,
  rst_n_in
  )
begin
  if (rst_n_in = '0') then
    to_count          <= (others => '0');
    to_bit_count      <= (others => '0');
    to_error_i        <= '0';
    global_to_error_i <= '0';
  elsif (clk_in'event AND clk_in = '1') then
    if (clr_to_counter_i = '1') then
      to_count     <= (others => '0');
      to_bit_count <= (others => '0');
    elsif (en_to_counter_i = '1') then
      if (to_bit_count = 31) then
        to_count <= to_count + 1;
      end if;
      to_bit_count <= to_bit_count + 1;
    else
      to_bit_count <= (others => '0');
    end if;
    
    if ((to_count >= fifo_timeout_limit_in AND to_bit_count = 23) OR (global_to_error_i = '1' AND readout_state = err_det)) then
      to_error_i <= en_error_i AND link_fifo_empty_in(link_num);
    else
      to_error_i <= '0';
    end if;
    
    if (to_error_i = '1' AND readout_state = err_det) then
      global_to_error_i <= '1';
    end if;
    
    if (readout_state = pass_token) then
      global_to_error_i <= '0';
    end if;
  end if;
end process timeout_counter;


end rtl; -- end code for fifo_readout_controller 