----------------------------------------------------------------------------
-- CERN
-- Author: Tom Barber
----------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

entity sct_sim_control is
  port(
    clk_in         : in  std_logic; -- 40 MHz clock
    rst_n_in       : in  std_logic; -- Powerup global reset
    inlinks        : in  std_logic_vector(11 DOWNTO 0); -- Real Input links
    config_in      : in  std_logic_vector(31 DOWNTO 0); -- configuration word for the sim_module
    level1_trigger : in  std_logic; -- Level 1 Trigger Input
    ecr_in         : in  std_logic; -- ECR Input
    bcr_in         : in  std_logic; -- Bcr Input
    outlinks       : out std_logic_vector(11 DOWNTO 0) -- output for real or encoded data
    );
end sct_sim_control; 

architecture rtl of sct_sim_control is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

signal outlinks_i : std_logic_vector (3 downto 0); 

type   rnd_seed_array is array (0 to 5) of std_logic_vector(3 downto 0); --
signal rnd_seed : rnd_seed_array;

signal l1_trigger_count : std_logic_vector (3 downto 0); -- count the number of L1  
signal bc_count         : std_logic_vector (7 downto 0); -- count the number of BC

----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------

component sct_sim_module is
  port(
    clk_in           : in  std_logic; -- 40 MHz clock
    rst_n_in         : in  std_logic; -- Powerup global reset
    rnd_seed         : in  std_logic_vector( 3 DOWNTO 0);
    config_in        : in  std_logic_vector(31 DOWNTO 0); -- configuration word for the sim_module
    level1_trigger   : in  std_logic; -- Level 1 Trigger Input
    l1_trigger_count : in  std_logic_vector (3 downto 0); -- count the number of L1
    bc_count   	     : in  std_logic_vector (7 downto 0); -- count the number of BC
    -- Single engine
--    inlinks          : in  std_logic_vector(1 downto 0);
    outlinks         : out std_logic_vector(1 downto 0)
    );
end component ; 

component level1_stretcher is
  port( 
    clk_in         : in  STD_LOGIC;                       -- 40 MHz clock
    rst_n_in       : in  STD_LOGIC;
    level1_trigger : in  STD_LOGIC;
    level1_accept  : in  STD_LOGIC_VECTOR (3 downto 0);
    BCR            : in  STD_LOGIC;
    ECR            : in  STD_LOGIC;
    write_event_id : out STD_LOGIC;
    BCID           : out STD_LOGIC_VECTOR (7 downto 0);
    LVL1ID         : out STD_LOGIC_VECTOR (3 downto 0)
  );
end component level1_stretcher;

BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

rnd_seed(0) <= X"2";
rnd_seed(1) <= X"4";
rnd_seed(2) <= X"7";
rnd_seed(3) <= X"9";
rnd_seed(4) <= X"C";
rnd_seed(5) <= X"E";
  
-- For a single engine simulator
-- This has a slice utilization of 67%
  U1 : sct_sim_module
    port map(
      clk_in           => clk_in,
      rst_n_in         => rst_n_in,
      rnd_seed         => rnd_seed(0),
      config_in        => config_in,
      level1_trigger   => level1_trigger,
      l1_trigger_count => l1_trigger_count,
      bc_count         => bc_count,
--      inlinks          => open,
      outlinks         => outlinks_i(1 downto 0)
    );

  U2 : sct_sim_module
    port map(
      clk_in           => clk_in,
      rst_n_in         => rst_n_in,
      rnd_seed         => rnd_seed(3),
      config_in        => config_in,
      level1_trigger   => level1_trigger,
      l1_trigger_count => l1_trigger_count,
      bc_count         => bc_count,
--      inlinks          => open,
      outlinks         => outlinks_i(3 downto 2)
    );
---------------------------------

level1_handler : level1_stretcher
  port map(
    clk_in         => clk_in,
	  rst_n_in       => rst_n_in,
    level1_trigger => level1_trigger,
	  level1_accept  => (others => '0'),  -- set to zero for now...
	  BCR            => bcr_in,
	  ECR            => ecr_in,
    write_event_id => open,
    BCID           => bc_count,
    LVL1ID         => l1_trigger_count
  );

--------------------------------------------------------------------------
-- SIGNALS
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- PROCESS DECLARATION
--------------------------------------------------------------------------

--some_process : process (
--  rst_n_in,
--  clk_in
--  )
--begin
--  if (rst_n_in = '0' or reset_cnt = '1') then 
--  elsif (clk_in'event AND clk_in = '1') then
--  end if;
--end process;

output_mux : process (
  config_in,
  inlinks,
  outlinks_i
)
begin
--Check whether to activate the simulator based on the config register
  if (config_in(0) = '0') then
    outlinks <= inlinks;
  else
-- 2 links per module, 6 modules per formatter = 12 links
-- low and high indicate which chips to send out
-- Just keep these two lines for a multiple engine simulator
-- Uncomment the following lines for a multi-engine simulator
    outlinks( 0) <= outlinks_i(0);
    outlinks( 1) <= outlinks_i(1);
    outlinks( 2) <= outlinks_i(2);
    outlinks( 3) <= outlinks_i(3);
    outlinks( 4) <= outlinks_i(0);
    outlinks( 5) <= outlinks_i(1);
    outlinks( 6) <= outlinks_i(2);
    outlinks( 7) <= outlinks_i(3);
    outlinks( 8) <= outlinks_i(0);
    outlinks( 9) <= outlinks_i(1);
    outlinks(10) <= outlinks_i(2);
    outlinks(11) <= outlinks_i(3);
  end if;
end process;

end rtl;
