----------------------------------------------------------------------------
-- CERN
-- Author: Jens Dopke, Tom Barber
-- Credits: All the credits go to B. Demirkoz - It's her fault !
----------------------------------------------------------------------------
-- Filename:
--    sct_sim_module.vhd
-- Description:
--    Formatter embedded simulation giving module-like data for the rest of
--       readout chain. The number of hits is defineable and the strobe_out
--       of data is given by a level 1 trigger input.          
----------------------------------------------------------------------------
--
--    What is to be done for simulation of SCT (Pixel later)
--
--  A total of five processe should be needed (could maybe be four):
--                         --- simulator_state ---
--   - Processes running the State machine
--       Should be done, counting might be wrong though (by one bit each)
--       Needs a controlling instance to check whether I fucked it up...
--                         --- state_progress  ---
--   - Process progressing the present to the next state with each clock
--       Hands on new settings for chipnumber and hits
--                         --- config_transfer ---
--   - A process depending on the state machine to transfer configuration
--     data into the internal config, which has got to be hold while
--     transferring a full simulated event into the rest of the formatter
--                         --- state_output    ---
--   - A process using the states to determine the data, which is to be
--     routed out of the simu. This sets parallel data at the beginning of
--     each state and routes it out through serial shifting
--                         --- output_mux      ---
--   - By looking at the simulation enable flag, this process selects
--     either simulated or real data to be forwarded to the rest of the
--     formatter for processing
--
--    Question one's gotta think about:
--
--   - How to receive a correct BCID (should be straight forward, but...)
--   - Give capabilities of compressed events ?
--   - Switchable randomization ?
--   - Hit location should be increasing during an event, shouldn't it ?
--       (This is gonna be a tough one, when thinkin of randomization...)
--
--  Configuration bits
--  ------------------
--  
--  0: Activation (0 - route input to output, 1 - ignore input, generate random events)
--  1: Send only clock/2
--  2: Send config readback
--  3-7: custom header
--  8-9: hit prob
-- 10-11:err prob                                   
-- 12-15:chipnum
-- 16-22: hitnum
-- 23-25: hitmap
-- 26-27: cluster size
-- 28-29: Link 0 error probability
-- 30-31: Link 1 error probability

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

entity sct_sim_module is
  port(
    clk_in           : in  std_logic; -- 40 MHz clock
    rst_n_in         : in  std_logic; -- Powerup global reset
    rnd_seed         : in  std_logic_vector( 3 DOWNTO 0);
    config_in        : in  std_logic_vector(31 DOWNTO 0); -- configuration word for the sim_module
    level1_trigger   : in  std_logic; -- Level 1 Trigger Input
    l1_trigger_count : in std_logic_vector (3 downto 0); -- count the number of L1
    bc_count   	     : in std_logic_vector (7 downto 0); -- count the number of BC
--    inlinks          : in  std_logic_vector(1 downto 0);
    outlinks         : out std_logic_vector(1 downto 0)
    );
end sct_sim_module ; 

architecture rtl of sct_sim_module is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  type   sct_simu_state_typedef IS (
    idle,
    send_header, 
    send_event_id,
    send_empty_event,
    send_error,
    send_new_cluster,
    send_hit,
    send_trailer,
    send_config_readback,
    send_clock_by_two
    );
  
  signal nhitsincluster       : UNSIGNED (1 DOWNTO 0); -- number of hits in cluster (strobed from the config word)
  signal next_nhitsincluster  : UNSIGNED (1 DOWNTO 0);
  signal nhits                : UNSIGNED (6 DOWNTO 0); -- number of hits (strobed from the config word)
  signal next_nhits           : UNSIGNED (6 DOWNTO 0);
  signal chipnum              : UNSIGNED (3 downto 0); -- number of chip link 0
  signal chipnum_high         : UNSIGNED (3 downto 0); -- number of chip link 1 
  signal next_chipnum         : UNSIGNED (3 downto 0);
  signal config_i             : std_logic_vector(31 DOWNTO 0); -- internal configuration word for the simulator
  signal bit_count            : UNSIGNED (5 downto 0); -- counter for where we are in the output
  signal next_bit_count       : UNSIGNED (5 downto 0);
  signal serial_data_out_high : STD_LOGIC_VECTOR (31 downto 0); -- upper link out
  signal serial_data_out_low  : STD_LOGIC_VECTOR (31 downto 0); -- lower link out
  signal next_state           : sct_simu_state_typedef; -- SCT Decoder FSM states
  signal pres_state           : sct_simu_state_typedef; -- SCT Decoder FSM states 
  signal serial_data_in       : STD_LOGIC_VECTOR (18 downto 0); -- from the random generator

----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------

component lfsr is
  port(
    clk_in      : in  std_logic; -- 40 MHz clock
    rst_n_in    : in  std_logic; -- Powerup global reset
	  rnd_seed_in : in  std_logic_vector( 3 DOWNTO 0); -- Random seed
    rnd_num     : out std_logic_vector( 18 DOWNTO 0) -- random number output
    );
end component ; 

BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

random_generator : lfsr
  port map (
    clk_in      => clk_in,
    rst_n_in    => rst_n_in,
    rnd_seed_in => rnd_seed,
    rnd_num     => serial_data_in
    );

--------------------------------------------------------------------------
-- SIGNALS
--------------------------------------------------------------------------

outlinks(0) <= serial_data_out_low(0);
outlinks(1) <= serial_data_out_high(0);

--------------------------------------------------------------------------
-- PROCESS DECLARATION
--------------------------------------------------------------------------

--some_process : process (
--  rst_n_in,
--  clk_in
--  )
--begin
--  if (rst_n_in = '0' or reset_cnt = '1') then 
--  elsif (clk_in'event AND clk_in = '1') then
--  end if;
--end process;

simulator_state : process (
  pres_state,
  next_state,
  bit_count,
  next_bit_count,
  level1_trigger,
  config_i,
  chipnum,
  chipnum_high,
  next_chipnum,
  nhits,
  next_nhits,
  nhitsincluster,
  next_nhitsincluster,
  l1_trigger_count,
  bc_count
  )
begin
    -- default FSM values
    next_state <= pres_state;
    next_bit_count <= bit_count + "000001";
    next_chipnum <= chipnum;
    next_nhitsincluster <= nhitsincluster;
    next_nhits <= nhits;
    -- check for state transition conditions
    case pres_state is
--------------------------------------------------------------------------------
-- idle should move to send_header, in case a level 1 trigger is received
--------------------------------------------------------------------------------
      when idle =>
         if (level1_trigger = '1') then
-- Config register bit 1 determines clock/2
				if (config_i(1) = '0') then
					next_state <= send_header;
				else	
					next_state <= send_clock_by_two;
				end if;				
--				next_state <= send_header;
         end if;
         next_bit_count <= "000000";
         next_chipnum <= "0000";
--------------------------------------------------------------------------------
-- send_header moves to send_event_id for now - no fancy error introduction...
--------------------------------------------------------------------------------
       when send_header =>
		    next_chipnum <= "0000";
          if (std_logic_vector(bit_count) = "000100") then
	   		 next_state <= send_event_id;
             next_bit_count <= "000000";
          end if;
--------------------------------------------------------------------------------
-- send_event_id should move to new_cluster for standard events, to raw data
-- sending in case of activated flags, and to trailers in case a trailer error
-- flag is to be raised for events.
--------------------------------------------------------------------------------
      when send_event_id =>
-- Initialise value of nhits
		next_nhits <= UNSIGNED(config_i(22 downto 16));
	if (std_logic_vector(bit_count) = "001101") then
-- Decide here what to do: hits, errors, empty
-- frequency of error events, empty events depends on random numbers
-- Config register 2 determines config readback mode
-- if the following is uncommented
-- at the moment it is used for debug mode
--		if (config_i(2) = '0') then -- send an event
-- Should we send a full event or empty one?
			case (config_i(9 downto 8)) is
				when "00" => -- always send a hit
					next_state <= send_new_cluster;
				when "01" => -- half hit / half empty
					if (serial_data_in(0) = '0') then
						next_state <= send_new_cluster;
					else
						next_state <= send_empty_event;
					end if;
				when "10" => -- mostly empty, some hits (per strip every 128 events)
					if (serial_data_in(7 downto 0) = "00000000") then
						next_state <= send_new_cluster;
					else
						next_state <= send_empty_event;
					end if;			
				when "11" => -- always an empty event
					next_state <= send_empty_event;
				when others =>
					next_state <= send_new_cluster;
			end case;
			
-- Also add probability that we send an error instead			
			case (config_i(11 downto 10)) is
				when "00" => -- always send hit/empty, so do nothing									
				when "01" => -- mostly OK, but some errors
					if (serial_data_in(15 downto 8) = "00000000") then
						next_state <= send_new_cluster;
					else
						next_state <= send_empty_event;
					end if;			
				when "10" => -- half OK / half error
					if (serial_data_in(8) = '0') then
						next_state <= send_error;
					end if;
				when "11" => -- always an error
					next_state <= send_error;
				when others => -- stay OK					
			end case;

--		else --  send config readback
--			next_state <= send_config_readback;
--		end if;
--		next_state <= send_new_cluster;
		next_bit_count <= "000000";
	end if;
		  
--------------------------------------------------------------------------------
-- send_event_id moves to send_empty event if not hits are required, then
-- sends a trailer.
--------------------------------------------------------------------------------
      when send_empty_event =>
        if (std_logic_vector(bit_count) = "000010") then
          next_state <= send_trailer;
          next_bit_count <= "000000";
        end if;
		  
       when send_error =>
          if (std_logic_vector(bit_count) = "001010") then
            if (std_logic_vector(chipnum) = 
               std_logic_vector(config_i(15 downto 12))) then
                  next_state <= send_trailer;
            else
                  next_state <= send_error;
                  next_chipnum  <= chipnum + "0001";
            end if;
            next_bit_count <= "000000";
          end if;
			 
		when send_config_readback =>
          if (std_logic_vector(bit_count) = "011011") then
            if (std_logic_vector(chipnum) = "0101") then
                  next_state <= send_trailer;
            else
                  next_state <= send_config_readback;
                  next_chipnum  <= chipnum + "0001";
            end if;
                  next_bit_count <= "000000";
          end if;
			 
		when send_clock_by_two =>
			if (std_logic_vector(bit_count) = "000011") then
				next_state <= send_clock_by_two;
				next_bit_count <= "000000";
			end if;
--------------------------------------------------------------------------------
-- send_event_id should move to new_cluster for standard events, to raw data
-- sending in case of activated flags, and to trailers in case a trailer error
-- flag is to be raised for events.
--------------------------------------------------------------------------------
      when send_new_cluster =>
        next_nhitsincluster <= UNSIGNED(config_i(27 downto 26));
--        next_nhitsincluster <= "0000000";
        if (std_logic_vector(bit_count) = "001100") then
          next_state <= send_hit;
          next_bit_count <= "000000";
        end if;

--------------------------------------------------------------------------------
-- send_hit cycles through the number of wanted hits and sends them out
-- moves on to send_trailer after all chips were cycled through
-- otherwise goes back to send_new_cluster
--------------------------------------------------------------------------------
      when send_hit =>
        if (std_logic_vector(bit_count) = "000011") then
          if (nhitsincluster = 0) then
--          Up to a maximum of 6 chips
--			   if (std_logic_vector(chipnum) = "0101") then
--          now just loop over 1 chip to reduce occupancy
--				if (std_logic_vector(chipnum) = "0000") then
--          Or make the chip number configurable...
            if (nhits = 0) then
              if (std_logic_vector(chipnum) = std_logic_vector(config_i(15 downto 12))) then
                  next_state <= send_trailer;
              else
				      next_state <= send_new_cluster;
			         next_chipnum  <= chipnum + "0001";
				      next_nhits <= UNSIGNED(config_i(22 downto 16));
				  end if;
            else
				  next_state <= send_new_cluster;
				  next_nhits <= nhits - "0000001";
				end if;
			 else
			   next_state <= send_hit;
			   next_nhitsincluster <= nhitsincluster - "01";
			 end if;
		    next_bit_count <= "000000";
		  end if;
--------------------------------------------------------------------------------
-- send_trailer sends out a trailer to finalize the event transfer
--------------------------------------------------------------------------------
      when send_trailer =>
        if (std_logic_vector(bit_count) = "001111") then
--        Increment the level one internal trigger counter
--        This is handled now by the external level1_handler
--			 if (std_logic_vector(l1_trigger_count) = "1111") then
--			    next_l1_trigger_count <= "0000";
--			 else
--			    next_l1_trigger_count <= l1_trigger_count + "0001";
--			end if;
			next_state <= idle;
		  end if;
		  
      when others =>
		  next_state <= idle;
		  next_chipnum <= "0000";
		  next_bit_count <= "000000";
    end case;
end process;

state_progress : process (
  rst_n_in,
  clk_in
  )
begin
  if (rst_n_in = '0') then 
    pres_state <= idle;
	 bit_count <= "000000";
	 chipnum <= "0000";
	 chipnum_high <= "1101";
--	 l1_trigger_count <= "0000";
--	 bc_count <= "00000000";
  elsif (clk_in'event AND clk_in = '1') then
    pres_state <= next_state;
	 bit_count <= next_bit_count;
	 chipnum <= next_chipnum;
--  Ensure that the chip on side 1 is opposite the one on side 0
--  also need to preserve the chip sequence on both sides.
	 chipnum_high <= ("1101" - UNSIGNED(config_i(15 downto 12))) + next_chipnum;
	 nhitsincluster <= next_nhitsincluster;
	 nhits <= next_nhits;
--	 l1_trigger_count <= next_l1_trigger_count;
--	 bc_count <= next_bc_count;
  end if;
end process;

state_output : process (
  clk_in,
  rst_n_in,
  config_in
)
begin
  if (rst_n_in = '0') then
    serial_data_out_low <= (others => '0');
    serial_data_out_high <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
 -- move the data stream on by one bit
    serial_data_out_low  <= "0" & serial_data_out_low(31 downto 1);	
    serial_data_out_high <= "0" & serial_data_out_high(31 downto 1);
	 
--------------------------------------------------------------------------------
-- FOLLOWING should be the defining data_out state machinery
--------------------------------------------------------------------------------
    case pres_state is
--------------------------------------------------------------------------------
-- idle should move to send_header, in case a level 1 trigger is received
--------------------------------------------------------------------------------
      when idle =>
        serial_data_out_low  <= (others => '0');
        serial_data_out_high <= (others => '0');
--------------------------------------------------------------------------------
--  send_header should move to different stages, after the header is sent,
--  depending on configuration data:
--  If failures shall be generated without data, then rise an error-flag
--
--  NB - all of the serial_data_out are read from right to left!
--------------------------------------------------------------------------
		when send_header =>
			if (std_logic_vector(bit_count) = "000000") then
-- Check if we want a custom header
				if (std_logic_vector(config_i(7 downto 3)) = "00000") then		
--                                                                          preamble
					serial_data_out_low(31 downto 0) <=  "0000000000000" & "0000000000000010111";
					serial_data_out_high(31 downto 0) <= "0000000000000" & "0000000000000010111";
				else
					serial_data_out_low(31 downto 0) <= "0000000000000" & "00000000000000"  &
													std_logic(config_i(3)) & std_logic(config_i(4)) &
													std_logic(config_i(5)) & std_logic(config_i(6)) &
													std_logic(config_i(7));
					serial_data_out_high(31 downto 0) <= "0000000000000" & "00000000000000"  &
													std_logic(config_i(3)) & std_logic(config_i(4)) &
													std_logic(config_i(5)) & std_logic(config_i(6)) &
													std_logic(config_i(7));
				end if;
		  end if;
--------------------------------------------------------------------------------
-- send_event_id should move to new_cluster for standard events, to raw data
-- sending in case of activated flags, and to trailers in case a trailer error
-- flag is to be raised for events.
--------------------------------------------------------------------------------
      when send_event_id =>
        if (std_logic_vector(bit_count) = "000000") then
--        Check config register to use internal or pre-defined lvl1id
--        General format for event headers:
--                                                        bcid         lvl1id   data type (always 0)
--		    serial_data_out_low(31 downto 0)  <= "0000000000000" &"000001" & "00001101" & "1101" & "0";
--			 serial_data_out_high(31 downto 0) <= "0000000000000" &"000001" & "00001101" & "1101" & "0";
--
--		    if (std_logic_vector(config_i(23 downto 19)) = "00000") then
			serial_data_out_low(31 downto 0)  <= "0000000000000" &"000001" &
				std_logic(bc_count(0)) & std_logic(bc_count(1)) &
				std_logic(bc_count(2)) & std_logic(bc_count(3)) &
				std_logic(bc_count(4)) & std_logic(bc_count(5)) &
				std_logic(bc_count(6)) & std_logic(bc_count(7)) &			
				std_logic(l1_trigger_count(0)) & std_logic(l1_trigger_count(1)) &
				std_logic(l1_trigger_count(2)) & std_logic(l1_trigger_count(3)) & "0";	
			serial_data_out_high(31 downto 0)  <= "0000000000000" &"000001" &
				std_logic(bc_count(0)) & std_logic(bc_count(1)) &
				std_logic(bc_count(2)) & std_logic(bc_count(3)) &
				std_logic(bc_count(4)) & std_logic(bc_count(5)) &
				std_logic(bc_count(6)) & std_logic(bc_count(7)) &			
				std_logic(l1_trigger_count(0)) & std_logic(l1_trigger_count(1)) &
				std_logic(l1_trigger_count(2)) & std_logic(l1_trigger_count(3)) & "0";	
--			 else
--				serial_data_out_low(31 downto 0)  <= "000001" & config_i(31 downto 24) &
--					config_i(23 downto 20) & "0";
--				serial_data_out_high(31 downto 0) <= "000001" & config_i(31 downto 24) &
--					config_i(23 downto 20) & "0";
--			 end if;

		  end if;
		  
		when send_empty_event =>
			if (std_logic_vector(bit_count) = "000000" ) then
--        send the data block for an empty event
		    serial_data_out_low(31 downto 0) <= "0000000000000" & "0000000000000000100";
		    serial_data_out_high(31 downto 0) <= "0000000000000" & "0000000000000000100";	
			end if;

      when send_error =>
		   if (std_logic_vector(bit_count) = "000000" ) then
-- The error code can be 001: no data
--                       010: buffer overflow
--                       100: buffer error
--
            serial_data_out_low(31 downto 0) <= "0000000000000" & "00000000" & "1" 
                   & "010"   --  serial_data_in(5 downto 3) -- error code: buffer overflow
				           & std_logic(chipnum(0)) & std_logic(chipnum(1))
			             & std_logic(chipnum(2)) & std_logic(chipnum(3))
									 & "000";
--                                                               
            serial_data_out_high(31 downto 0) <= "0000000000000" & "00000000" & "1" 
                   & "001"  -- serial_data_in(2 downto 0)  -- error code: buffer error
				           & std_logic(chipnum_high(0)) & std_logic(chipnum_high(1))
			             & std_logic(chipnum_high(2)) & std_logic(chipnum_high(3))
									 & "000";
         end if;
		  
      when send_new_cluster =>
        if (std_logic_vector(bit_count) = "000000") then
--  This gives one hit per chip for both sides:
		  if (config_i(2) = '1') then  -- Put the l1id and BC into the chip and channel number
			  serial_data_out_low(31 downto 0) <= "0000000000000" &"000000" &
					std_logic(nhits(0)) & std_logic(nhits(1)) &
					std_logic(nhits(2)) & std_logic(nhits(3)) &
					std_logic(nhits(4)) & std_logic(nhits(5)) &
					std_logic(nhits(6)) &			
					std_logic(l1_trigger_count(0)) & std_logic(l1_trigger_count(1)) &
					std_logic(l1_trigger_count(2)) & std_logic(l1_trigger_count(3)) & "10";					
			  serial_data_out_high(31 downto 0) <= "0000000000000" &"000000" &
					std_logic(bc_count(0)) & std_logic(bc_count(1)) &
					std_logic(bc_count(2)) & std_logic(bc_count(3)) & "000" &
					std_logic(bc_count(4)) & std_logic(bc_count(5)) &
					std_logic(bc_count(6)) & std_logic(bc_count(7)) & "10"; 
		  else  -- a normal event with random strips
			  serial_data_out_low(31 downto 0) <= "0000000000000" &"000000" & serial_data_in(6 downto 0)
									& std_logic(chipnum(0)) & std_logic(chipnum(1))
			                  & std_logic(chipnum(2)) & std_logic(chipnum(3))& "10";
		     serial_data_out_high(31 downto 0) <= "0000000000000" &"000000" & not(serial_data_in(6 downto 0)) 
									& std_logic(chipnum_high(0)) & std_logic(chipnum_high(1))
			                  & std_logic(chipnum_high(2)) & std_logic(chipnum_high(3)) & "10";
        end if;
	end if;
--		  when send_noise =>
--			if (std_logic_vector(bit_count) = "000000") then
--			 serial_data_out_low(31 downto 0) <= "000000" & serial_data_in(6 downto 0) & std_logic(chipnum(0)) & std_logic(chipnum(1))
--			                                   & std_logic(chipnum(2)) & std_logic(chipnum(3))& "10";
--		    serial_data_out_high(31 downto 0) <= "000000" & serial_data_in(13 downto 7) & std_logic(chipnum_high(0)) & std_logic(chipnum_high(1))
--			                                   & std_logic(chipnum_high(2)) &  std_logic(chipnum_high(3)) & "10";
--			end if;
--        Now send the L1ID on the side 1 channels instead of random...
--                                                       channel                    chip
--   	    serial_data_out_low(31 downto 0) <= "000000" & serial_data_in(6 downto 0) & std_logic(chipnum(0)) & std_logic(chipnum(1))
--			                                   & std_logic(chipnum(2)) & std_logic(chipnum(3))& "10";
--		    serial_data_out_high(31 downto 0) <= "000000" & std_logic(chipnum_high(0))
--														  & std_logic(chipnum_high(1)) & std_logic(chipnum_high(2))
--														  & std_logic(l1_trigger_count(0)) & std_logic(l1_trigger_count(1))
--					                             & std_logic(l1_trigger_count(2)) & std_logic(l1_trigger_count(3))
--                                          & std_logic(chipnum_high(0)) &  std_logic(chipnum_high(1))
--			                                   & std_logic(chipnum_high(2)) &  std_logic(chipnum_high(3)) & "10";

--  This gives one random hit looping over chips:
--                                                       channel                    chip
--		    serial_data_out_low(31 downto 0)  <= "000000" & serial_data_in(0 to 6)  & serial_data_in(7 to 9)   & "0" & "10";
--		    serial_data_out_high(31 downto 0) <= "000000" & serial_data_in(9 to 15) & serial_data_in(16 to 18) & "1" & "10";

      when send_hit =>
        if (std_logic_vector(bit_count) = "000000") then
				case (config_i(25 downto 23)) is
					when "000" =>  -- fixed at 011
						serial_data_out_low(31 downto 0) <= "0000000000000" & "000000000000000" & "110" & "1";
						serial_data_out_high(31 downto 0)<= "0000000000000" & "000000000000000" & "110" & "1";
					when "001" =>  -- fixed at 010
						serial_data_out_low(31 downto 0) <= "0000000000000" & "000000000000000" & "010" & "1";
						serial_data_out_high(31 downto 0)<= "0000000000000" & "000000000000000" & "010" & "1";
					when "010" =>  -- detector alignment 1xx
						serial_data_out_low(31 downto 0) <= "0000000000000" & "000000000000000" & serial_data_in(0) & serial_data_in(1) & "1" & "1";
						serial_data_out_high(31 downto 0)<= "0000000000000" & "000000000000000" & serial_data_in(0) & serial_data_in(1) & "1" & "1";
					when "011" =>  -- detector alignment x1x
						serial_data_out_low(31 downto 0) <= "0000000000000" & "000000000000000" & serial_data_in(0) & "1" & serial_data_in(1) & "1";
						serial_data_out_high(31 downto 0)<= "0000000000000" & "000000000000000" & serial_data_in(0) & "1" & serial_data_in(1) & "1";
					when "100" =>  -- detector alignment xx1
						serial_data_out_low(31 downto 0) <= "0000000000000" & "000000000000000" & "1" & serial_data_in(0) & serial_data_in(1) & "1";
						serial_data_out_high(31 downto 0)<= "0000000000000" & "000000000000000" & "1" & serial_data_in(0) & serial_data_in(1) & "1";
					when "101" =>  -- Level mode x1x
						serial_data_out_low(31 downto 0) <= "0000000000000" & "000000000000000" & serial_data_in(0) & "1" & serial_data_in(1) & "1";
						serial_data_out_high(31 downto 0)<= "0000000000000" & "000000000000000" & serial_data_in(0) & "1" & serial_data_in(1) & "1";
					when "110" =>  -- Edge mode 01x
						serial_data_out_low(31 downto 0) <= "0000000000000" & "000000000000000" & serial_data_in(0) & "10" & "1";
						serial_data_out_high(31 downto 0)<= "0000000000000" & "000000000000000" & serial_data_in(0) & "10" & "1";
					when "111" =>  -- Test mode xxx
						serial_data_out_low(31 downto 0) <= "0000000000000" & "000000000000000" & serial_data_in(0) & serial_data_in(1) & serial_data_in(2) & "1";
						serial_data_out_high(31 downto 0)<= "0000000000000" & "000000000000000" & serial_data_in(0) & serial_data_in(1) & serial_data_in(2) & "1";
					when others => -- no hitmap, this shouldn't be possible!
						serial_data_out_low(31 downto 0) <= "0000000000000" & "000000000000000" & "000" & "1";
						serial_data_out_high(31 downto 0)<= "0000000000000" & "000000000000000" & "000" & "1";
				end case;

--        Check to see if user has specified a hitmap, if not, use "011"
--		    if (std_logic_vector(config_i(3 downto 1)) = "000") then
--				serial_data_out_low(31 downto 0)  <= "0000000000000" & "000000000000000" & "110" & "1"; -- hitmap & "1";
--				serial_data_out_high(31 downto 0) <= "0000000000000" & "000000000000000" & "110" & "1"; -- hitmap & "1";
--			 else
--				serial_data_out_low(31 downto 0) <= "0000000000000" & "000000000000000" & config_i(3 downto 1) & "1"; -- hitmap & "1";
--				serial_data_out_high(31 downto 0) <= "0000000000000" & "000000000000000" & config_i(3 downto 1) & "1"; -- hitmap & "1";
--			 end if;
			 
		  end if;

-- Also add the possiblity for returning config readback mode
-- <000> <aaaa> < 111><cccc,cccc><1><cccc,cccc><1>
		when send_config_readback =>
			if (std_logic_vector(bit_count) = "000000") then
		    serial_data_out_low(31 downto 0) <= "0000" & "1" &
									serial_data_in(18 downto 11) & "1" &
									serial_data_in(10 downto 3) & "111" &
									std_logic(chipnum(0)) & std_logic(chipnum(1)) &
			                  std_logic(chipnum(2)) & std_logic(chipnum(3)) & "000";
		    serial_data_out_high(31 downto 0) <= "0000" & "1" &
									serial_data_in(15 downto 8) & "1" &
									serial_data_in(7 downto 0) & "111" &
									std_logic(chipnum_high(0)) & std_logic(chipnum_high(1)) &
			                  std_logic(chipnum_high(2)) & std_logic(chipnum_high(3)) & "000";
			end if;

		when send_clock_by_two =>
			if (std_logic_vector(bit_count) = "000000") then		
				serial_data_out_low(31 downto 0) <= "0000000000000" & "0000000000000010101";
				serial_data_out_high(31 downto 0) <= "0000000000000" & "0000000000000010101";			
			end if;
			
      when send_trailer =>
        if (std_logic_vector(bit_count) = "000000") then
		    serial_data_out_low(31 downto 0) <= "0000000000000" & "0000000000000000001";
		    serial_data_out_high(31 downto 0) <= "0000000000000" & "0000000000000000001";
		  end if;
      when others =>
    end case; -- present state
	 
-- perhaps try to introduce noise/bit flips into here??
-- The user defined noise level for link 0
	case (config_i(29 downto 28)) is
		when "11" =>
			if (std_logic_vector(serial_data_in(18 downto 0)) = "0000000000000000000") then
				serial_data_out_low <= serial_data_out_low(31 downto 1) & not(serial_data_out_low(0));
			end if;		
		when "01" =>
			if (std_logic_vector(serial_data_in(8 downto 0)) = "000000000") then
				serial_data_out_low <= serial_data_out_low(31 downto 1) & not(serial_data_out_low(0));
			end if;		
		when "10" =>
			if (std_logic_vector(serial_data_in(1 downto 0)) = "00") then
				serial_data_out_low <= serial_data_out_low(31 downto 1) & not(serial_data_out_low(0));
			end if;		
		when others =>  -- default "00" no noise
	end case;
-- The user defined noise level for link 1	
	case (config_i(31 downto 30)) is
		when "11" =>
			if (std_logic_vector(serial_data_in(18 downto 0)) = "0000000000000000000") then
				serial_data_out_high <= serial_data_out_high(31 downto 1) & not(serial_data_out_high(0));
			end if;		
		when "01" =>
			if (std_logic_vector(serial_data_in(17 downto 9)) = "000000000") then
				serial_data_out_high <= serial_data_out_high(31 downto 1) & not(serial_data_out_high(0));
			end if;		
		when "10" =>
			if (std_logic_vector(serial_data_in(7 downto 6)) = "00") then
				serial_data_out_high <= serial_data_out_high(31 downto 1) & not(serial_data_out_high(0));
			end if;		
		when others =>  -- default "00" no noise
	end case;	
 
  end if; -- if we're not resetting

end process;

config_transfer : process (
  rst_n_in,
  clk_in,
  pres_state
  )
begin
  if (rst_n_in = '0') then 
    config_i <= "00000000000000000000000000000000";
  elsif (clk_in'event AND clk_in = '1') then
    case pres_state is
	   when idle =>
		  config_i <= config_in;
	   when others =>
		  config_i <= config_i;
	  end case;
  end if;
end process;

end rtl;

