----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    link_formatter.vhd
-- Description:
--    ROD Front End Decoder
--       Single Channel
--       SCT Version
--
-- This file contains the serial decoding state machine for the SCT front
-- end. You should refer to three documents to better understand how the
-- vhdl for the decoder works.
--
--    1) "Format of Data from the SCT Front End Decoders" describes what
--    the output data words look like after the serial bitstreams have
--    been decoded and converted from serial to parallel.
--    /home/edafiles/atlas/rod/docs/sct_fe_format.txt(or .ps)
--
--    2) "ABC Event Structure" presents a state transition diagram
--    that shows how to decode the ABC serial data. The VHDL logic
--    for the SCT decoder is closely based on this diagram.
--    /home/edafiles/atlas/rod/docs/abc_format.ps
--
--    3) "SCT and Pixel Interfaces to the ROD" describes the details
--    about the data format of the serial link from the SCT front end.
--    /home/edafiles/atlas/rod/docs/abc_mcc.pdf
--
-- Note: There should also be numerous other documents describing the SCT
-- decoder and its output data stream format.
--
-- An important aspect of designing the SCT decoder is to try to
-- minimize the logic. Since there will be 96 of these state
-- machines in the ROD, small reductions in logic here are
-- important. 
-- 
-- In general, the SCT decoder must be robust to errors (i.e. cannot
-- go into an undefined state regardless of input data). Therefore,
-- error handling is also important and this adds to the
-- complexity/amount of logic in the decoder.
--

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.std_logic_arith.all;

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

entity link_formatter is
  port(
    clk_in             : in  STD_LOGIC; -- 40 MHz clock
    rst_n_in           : in  STD_LOGIC; -- Powerup global reset pulse
    ecr_in             : in  STD_LOGIC;
    almost_full        : in  STD_LOGIC; -- 1 = header/trailer only mode
                                        -- 0 = normal
    enable             : in  STD_LOGIC; -- enable formatter activity
    expanded_mode      : in  STD_LOGIC; -- 1 = uncondensed data mode,
                                        -- 0 = condensed data mode
    compressed_mode    : in  STD_LOGIC; -- 1 = compressed, 0 = bypass
    config_mode        : in  STD_LOGIC; -- 1 = config mode, 0 = normal
    edge_mode          : in  STD_LOGIC; -- 1 = edge detect "01X" for hit pattern
                                        -- 0 = level detect "X1X" for hit pattern
    link_in            : in  STD_LOGIC; -- front end serial data
    normal_not_raw     : out STD_LOGIC; -- decoder is in 1 = normal data mode, 
                                        --               0 = raw data mode 
    data_strobe        : out STD_LOGIC; -- Indicates a new output data
                                        -- word is available
    decoder_wr_trailer : out STD_LOGIC; -- trailer being written
    decoded_data       : out STD_LOGIC_VECTOR(31 DOWNTO 0); -- Output data word
    
--    l1a1_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a2_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a3_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a4_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a5_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a6_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a7_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);
--    l1a8_pipeline_in   : in  std_logic_vector( 3 DOWNTO 0);

--    link_pipeline_mask : in  std_logic_vector(31 DOWNTO 0);
    masked_chip_id_in      : in  std_logic_vector( 7 downto 0);

    data_overflow_limit_in : in  unsigned(11 downto 0); --
    hit_count_rst_in       : in  std_logic;
    hit_count_out          : out std_logic_vector(11 DOWNTO 0);
    decoder_active_out     : out std_logic;
    dovr_error_out         : out std_logic;
    link_data_tx_out       : out std_logic
    );
end link_formatter ; 

architecture rtl of link_formatter is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  type   sct_state_typedef IS (
    idle, 
    event_id, 
    new_cluster, 
    null_state,
    flagged_error, 
    hit, 
    trailer, 
    wait_for_trailer, 
    raw_data
    );
  
  signal almost_full_flag     : STD_LOGIC;
  signal almost_full_i        : STD_LOGIC;
  signal bit_count            : UNSIGNED (3 downto 0);
  signal bit_count_reset      : STD_LOGIC;
  signal condensed_mode_i     : STD_LOGIC;
  signal config_mode_i        : STD_LOGIC;
  signal data_out             : STD_LOGIC_VECTOR (15 DOWNTO 0);
  signal data_out_temp        : STD_LOGIC_VECTOR (15 DOWNTO 0);
  signal data_out_write       : STD_LOGIC; 
  signal data_strobe_lower    : STD_LOGIC;
  signal decoder_wr_trailer_i : STD_LOGIC;
  signal edge_mode_i          : STD_LOGIC;
  signal header_7bit          : STD_LOGIC;
  signal header_detect        : STD_LOGIC;
  signal header_error         : STD_LOGIC;
  signal header_error_del1    : STD_LOGIC;
  signal header_sh_reg        : STD_LOGIC_VECTOR (0 to 6);
  signal hit_location         : STD_LOGIC_VECTOR (10 downto 0);
  signal hit_pattern_invalid  : STD_LOGIC;
  signal link_in_i            : STD_LOGIC;
  signal new_cluster_flag     : STD_LOGIC; 
  signal next_bit_count       : UNSIGNED (3 downto 0);
  signal next_state           : sct_state_typedef; -- SCT Decoder FSM states
  signal pack_first_hit       : STD_LOGIC;
  signal pack_all_hits        : STD_LOGIC;
  signal pres_state           : sct_state_typedef; -- SCT Decoder FSM states 
  signal serial_data_in       : STD_LOGIC_VECTOR (0 to 18);
  signal serial_link_in       : STD_LOGIC;
  signal trailer_counter      : UNSIGNED (3 DOWNTO 0);
  signal trailer_detect       : STD_LOGIC;
  signal trailer_error        : STD_LOGIC;
  signal trailer_error_del1   : STD_LOGIC;
  signal trailer_error_flag   : STD_LOGIC;
  signal trailer_zeroes       : STD_LOGIC;
  signal pause_bit_count      : STD_LOGIC;
  signal resume_bit_count     : STD_LOGIC;
  signal wait_count           : UNSIGNED (4 DOWNTO 0);
  signal raw_data_count       : UNSIGNED (1 DOWNTO 0);
  signal wr_raw_data_cluster  : STD_LOGIC;
  signal pres_l1id            : STD_LOGIC_VECTOR(2 downto 0);
  signal pres_chip            : STD_LOGIC_VECTOR(3 downto 0);
  signal pres_chan            : STD_LOGIC_VECTOR(6 downto 0);
  signal l1a_pipeline_i       : STD_LOGIC_VECTOR(3 downto 0);
  signal en_header_detect     : STD_LOGIC;
  signal set_header_detect    : STD_LOGIC;
  signal header_counter       : UNSIGNED (3 DOWNTO 0);
  signal hit_counter          : UNSIGNED (11 DOWNTO 0);
  signal hit_count_max        : UNSIGNED (11 DOWNTO 0);
  signal overflow_trailer     : STD_LOGIC;
  signal decoder_active       : STD_LOGIC;
  signal link_data_tx_i       : STD_LOGIC;
  signal event_truncated      : STD_LOGIC;
  signal packed_hit_count     : UNSIGNED (3 DOWNTO 0);
  
  
----------------------------------------------------------------------------
--Constant DECLARATION
----------------------------------------------------------------------------
  
----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------
  
BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

--serial_link_in <= link_in_i and enable;
link_data_tx_out   <= link_data_tx_i;
decoder_active_out <= decoder_active;
dovr_error_out     <= overflow_trailer;
hit_count_out      <= STD_LOGIC_VECTOR(hit_count_max);
  
input_register : process (
  clk_in, 
  rst_n_in
  )
begin
  if (rst_n_in = '0') then
    serial_data_in (0 to 18) <= "0000000000000000000";
  elsif (clk_in'event and clk_in = '1') then
    almost_full_i    <= almost_full;
    condensed_mode_i <= NOT expanded_mode;
    config_mode_i    <= config_mode;
    edge_mode_i      <= edge_mode;
--    link_in_i        <= link_in;
    serial_link_in   <= link_in AND enable;
    serial_data_in   <= serial_data_in (1 to 18) & serial_link_in; 
    if (serial_data_in /= "0000000000000000000") then
	   link_data_tx_i <= '1';
    else
	   link_data_tx_i <= '0';
    end if;  
  end if;  
end process input_register;

find_header : process (
  clk_in, 
  rst_n_in
  )
begin -- find_header
  if (rst_n_in = '0') then
    header_sh_reg (0 to 6) <= "0000000";
  elsif (clk_in'event and clk_in = '1') then
    -- Correct header pattern is 111010. A header can also have up
    -- to one bit error in any location. There are two special hit
    -- patterns (with errors) that are also possible.
    header_7bit <= '0';
    header_sh_reg (0 to 6) <= header_sh_reg (1 to 6) & serial_data_in(2);
    if (en_header_detect = '1' AND enable = '1') then
--    if (en_header_detect = '1') then
      case header_sh_reg(1 TO 6) is
        when "111010" => -- "1111010" is a possible hit pattern with an error
          header_detect <= '1';
          header_error  <= header_sh_reg(0);
          header_7bit   <= header_sh_reg(0);
        when "011010" =>
          header_detect <= '1';
          header_error  <= '1';
  --    when "101010" =>   -- This case removed because it is the Clock/2 state 
  --      header_detect <= '1';
  --      header_error  <= '1';
        when "110010" =>
          header_detect <= '1';
          header_error  <= '1';
        when "111110" =>
          header_detect <= '1';
          header_error  <= '1';
        when "111000" =>
          header_detect <= '1';
          header_error  <= '1';
        when "111011" =>
          header_detect <= '1';
          header_error  <= '1';
        when "110100" =>
          -- "1110100" is a possible hit pattern with an error
          header_detect <= header_sh_reg(0);
          header_error  <= header_sh_reg(0);
          header_7bit   <= header_sh_reg(0);
        when others =>
          header_detect <= '0';
          header_error  <= '0';
      end case;
    else
      header_detect <= '0';
      header_error  <= '0';
    end if;  
  end if;  
end process find_header;


find_trailer : process (
  clk_in, 
  rst_n_in
  )    
begin -- find_trailer
  if (rst_n_in = '0') then
    trailer_counter <= "0000";
    header_counter <= "0000";
    en_header_detect <= '1';
    set_header_detect <= '0';
    event_truncated <= '0';
  elsif (clk_in'event and clk_in = '1') then
    set_header_detect <= '0';
    -- If 1 shifts in and 0 shifts out, decrement num 1's
    -- If 0 shifts in and 1 shifts out, increment num 1's
    if (serial_data_in(18) = '1' and serial_data_in(3) = '0') then
      trailer_counter <= trailer_counter + 1;
    elsif (serial_data_in(18) = '0' and serial_data_in(3) = '1') then
      trailer_counter <= trailer_counter - 1;
    end if;

-- Count the link input bits at the start of an event
    if (serial_data_in(2) = '1' AND header_counter < 8) then
      header_counter <= header_counter + 1 ;
    elsif (set_header_detect = '1') then
      header_counter <= (others => '0');
    end if;

-- If no header is detected after 6 '1's are through the decoder, 
--  mask the header detect
    if (trailer_counter /= 0 AND header_counter >= 5 ) then
      en_header_detect <= '0';
    elsif (set_header_detect = '1') then
      en_header_detect <= '1';
    end if;

    if (std_logic_vector(trailer_counter) = "0000") then
      -- if 0 or 1 followed by 15 0's, a trailer is detected
      trailer_detect <= '1';
      -- Zero followed by 15 zeroes is a trailer w/ bit error
      trailer_error <= not serial_data_in(2);
      -- trailer with 1 or 0 followed by 15 zeroes
      trailer_zeroes <= '1';
      set_header_detect <= '1';
      event_truncated <= '0';
    elsif (std_logic_vector(trailer_counter) = "0001") then
      -- 1 followed by single 1 in next 15 bits is trailer w/bit error
      trailer_detect <= serial_data_in(2);
      trailer_error  <= serial_data_in(2);
      trailer_zeroes <= '0';
      event_truncated <= '0';
      set_header_detect <= '1';
-- ADDED in Ver 3.9 to address TX laser issue 6/22/2010
-- NOT THOROUGHLY TESTED OR IMPLEMENTED
-- STILL NEED TO DETERMINE HOW TO FLAG IN THE DATA STREAM
--    elsif (serial_data_in(0 to 18) = "1111111111111111111" OR    -- RX Laser off
--           serial_data_in(0 to 18) = "1010101010101010101") then -- CLK/2 mode
    elsif (overflow_trailer = '1') then  -- added in ver s3Cf 7/1/2010
      trailer_detect <= '0';  --'1';
      trailer_error  <= '0';
      trailer_zeroes <= '0';
      event_truncated <= '1';
      -- set Truncated Event Flag here
    else
      trailer_detect <= '0';
      trailer_error  <= '0';
      trailer_zeroes <= '0';
      event_truncated <= '0';
   end if;
  end if;  
end process find_trailer;

-------------------------------------------------

-- The next process is all combinatorial logic and performs the
-- task of determining which state trantitions to take.

sct_comb_fsm : process (
  pres_state, 
  bit_count, 
  header_detect,
  serial_data_in, 
  trailer_error, 
  trailer_detect,
  data_out_write,
  trailer_zeroes, 
  header_7bit, 
  config_mode_i,
  resume_bit_count,
  data_out
  )
begin
  -- default FSM values
  next_state <= pres_state;
  next_bit_count <= bit_count + "0001";
  -- check for state transition conditions
-- **** Removed as a test in s4D-4f  JMJ 201209 @ CERN
--  if (trailer_detect = '1' AND trailer_error = '0' AND 
--       pres_state /= idle AND pres_state /= event_id AND pres_state /= trailer) then
--    next_state <= trailer;
--    next_bit_count <= "0000";
--  end if;
  if (enable = '0') then
    next_state <= idle;
  end if;

  case pres_state is
    when idle =>
      pause_bit_count <= '0';
      trailer_error_flag <= '0';
      if (std_logic_vector(bit_count) = "0110") then
        next_bit_count <= "0110";
      end if;
--      if ((header_detect = '1' and 
      if ((header_detect = '1' and enable = '1' AND
          ((std_logic_vector(bit_count) = "0101") or
           (std_logic_vector(bit_count) = "0110"))) or
          (header_7bit = '1' and std_logic_vector(bit_count) = "0110")) then
        next_state <= event_id;
        next_bit_count <= "0000";
      end if;

    when event_id =>
      if (std_logic_vector(bit_count) = "1011" and
          (serial_data_in(1) = '0' or config_mode_i = '1')) then
         next_bit_count <= "0000";
         next_state <= raw_data;
      elsif (std_logic_vector(bit_count) = "1100") then
        next_bit_count <= "0000";
        if (trailer_detect = '1') then
          next_state <= trailer;
--          trailer_error_flag <= '1';
        elsif (event_truncated = '1') then
          next_bit_count <= "0000";
          next_state <= wait_for_trailer;
        elsif (serial_data_in (1) = '1') then
          next_state <= raw_data;
        elsif (serial_data_in (1 to 2) = "01") then
          next_state <= new_cluster;
        elsif (serial_data_in (1 to 3) = "001") then
          next_state <= null_state;
        else
          next_state <= flagged_error;
        end if;
      end if;

    when new_cluster =>
      if (std_logic_vector(bit_count) = "1100") then
        next_bit_count <= "0000";
--        if (trailer_detect = '1' and trailer_error = '0') then
        if (trailer_detect = '1') then
          next_state <= trailer;
          trailer_error_flag <= trailer_error;   --'1';
        elsif (event_truncated = '1') then
          next_bit_count <= "0000";
          next_state <= wait_for_trailer;
        elsif (serial_data_in (1) = '1') then
          next_state <= hit;
        else
          next_state <= raw_data;
        end if;
      end if;

    when hit =>
      if (std_logic_vector(bit_count) = "0011") then
        next_bit_count <= "0000";
        if (trailer_detect = '1') then
          next_state <= trailer;
          trailer_error_flag <= trailer_error;
        elsif (event_truncated = '1') then
          next_bit_count <= "0000";
          next_state <= wait_for_trailer;
        elsif (serial_data_in (1) = '1') then
          next_state <= hit;
        elsif (serial_data_in (1 to 2) = "01") then
          next_state <= new_cluster;
        elsif (serial_data_in (1 to 3) = "001") then
          next_state <= null_state;
        else
          next_state <= flagged_error;
        end if;
      end if;

    when flagged_error =>
      if (std_logic_vector(bit_count) = "1000" AND data_out(2 downto 0) = "111") then
        next_bit_count <= "0000";
        next_state <= wait_for_trailer;
      elsif (std_logic_vector(bit_count) = "1001" AND serial_data_in(1) = '0') then
        next_bit_count <= "0000";
        next_state <= raw_data;
      elsif (std_logic_vector(bit_count) = "1010") then
        next_bit_count <= "0000";
        if (trailer_detect = '1') then
          next_state <= trailer;
          trailer_error_flag <= trailer_error;
        elsif (event_truncated = '1') then
          next_bit_count <= "0000";
          next_state <= wait_for_trailer;
        elsif (serial_data_in(1) = '1') then
          next_state <= raw_data;
        elsif (serial_data_in (1 to 2) = "01") then
          next_state <= new_cluster;
        elsif (serial_data_in (1 to 3) = "001") then
          next_state <= null_state;
        else
          next_state <= flagged_error;
        end if;
      end if;

    when null_state =>
      if (std_logic_vector(bit_count) = "0010") then
        next_bit_count <= "0000";
        if (trailer_detect = '1') then
          next_state <= trailer;
          trailer_error_flag <= trailer_error;
        elsif (event_truncated = '1') then
          next_bit_count <= "0000";
          next_state <= wait_for_trailer;
        elsif (serial_data_in (1) = '1') then
          next_state <= raw_data;
        elsif (serial_data_in (1 to 2) = "01") then
          next_state <= new_cluster;
        elsif (serial_data_in (1 to 3) = "001") then
          next_state <= null_state;
        else
          next_state <= flagged_error;
        end if;
      end if;

    when wait_for_trailer =>
      if (trailer_detect = '1') then
--        next_state <= trailer;
        next_state <= idle;  
      end if;
      
    when trailer =>
      if (std_logic_vector(bit_count) = "1111") then
        next_bit_count <= "0000";
        next_state <= idle;  
      end if;

    when raw_data =>
      if (ecr_in = '1') then  -- link disabled OR link masked
        next_state <= idle;  
      elsif (std_logic_vector(bit_count) = "1111") then
        next_bit_count <= "0000";
        if (trailer_detect = '1') then
          next_state <= idle;  
        end if;
      end if;
--      if (trailer_detect = '1') then
----      if ((trailer_detect = '1' and trailer_error = '0') or
----          trailer_zeroes = '1') then
----***Added to fix decode bug if header detected and no data follows.
--        next_bit_count <= "0000";  -- JMJ 10/30/2007 
-----------------------------------------------------------
--        next_state <= trailer;
--        trailer_error_flag <= trailer_error;
--      elsif (std_logic_vector(bit_count) = "0111") then
--        next_bit_count <= "0000";
--        next_state <= raw_data;
--    end if;
  end case;
end process sct_comb_fsm;

-- Some logic to check if the hit pattern is valid
-- Used in Condensed Mode only
process (edge_mode_i, serial_data_in)
begin
  if (edge_mode_i = '1') then
    if (serial_data_in (1 to 2) = "01") then
      hit_pattern_invalid <= '0';
    else
      hit_pattern_invalid <= '1';
    end if;
  else
    if (serial_data_in (2) = '1') then
      hit_pattern_invalid <= '0';
    else
      hit_pattern_invalid <= '1';
    end if;
  end if;
end process;

-- This next process looks at the current state and outputs
-- the data words depending on which state and which bit
-- is being shifted in.
process (clk_in, rst_n_in)
begin
  if (rst_n_in = '0') then
    wr_raw_data_cluster <= '0';
    pres_l1id <= (others => '0');
    pres_chip <= (others => '0');
    pres_chan <= (others => '0');
    hit_counter <= (others => '0');
    overflow_trailer <= '0';
    hit_count_max <= (others => '0');
    normal_not_raw <= '1';
    decoder_active <= '0';
  elsif (clk_in'event and clk_in = '1') then
    header_error_del1 <= header_error;
    trailer_error_del1 <= trailer_error;
    data_out_write <= '0';
    decoder_wr_trailer_i <= '0';
--    normal_not_raw <= '1';
    if (almost_full_i = '1') then
      almost_full_flag <= '1';
    end if;
    if (hit_count_rst_in = '1') then
      hit_count_max <= (others => '0');
    end if;
    if (hit_counter > hit_count_max AND trailer_detect = '1') then
      hit_count_max <= hit_counter;
    end if;
--    case pres_l1id is
--      when "001"  => l1a_pipeline_i <= l1a1_pipeline_in;
--      when "010"  => l1a_pipeline_i <= l1a2_pipeline_in;
--      when "011"  => l1a_pipeline_i <= l1a3_pipeline_in;
--      when "100"  => l1a_pipeline_i <= l1a4_pipeline_in;
--      when "101"  => l1a_pipeline_i <= l1a5_pipeline_in;
--      when "110"  => l1a_pipeline_i <= l1a6_pipeline_in;
--      when "111"  => l1a_pipeline_i <= l1a7_pipeline_in;
--      when others => l1a_pipeline_i <= l1a8_pipeline_in;
--    end case;
--    if (pres_state = idle AND trailer_counter = 0) then
--    if (pres_state = idle OR pres_state = raw_data) then
    if (pres_state = idle OR pres_state = raw_data OR pres_state = wait_for_trailer) then
      decoder_active <= '0';
    else
      decoder_active <= '1';
    end if;
    case pres_state is
      when idle =>
        almost_full_flag <= '0';
        raw_data_count <= (others => '0');
        hit_counter <= (others => '0');
        overflow_trailer <= '0';
        normal_not_raw <= '1';
    when event_id =>
        hit_counter <= (others => '0');
        overflow_trailer <= '0';
        if (std_logic_vector(next_bit_count) = "0000") then
          data_out_write <= '1';
        end if;
        if (bit_count_reset = '1') then
          data_out <= "001" & header_error_del1 & serial_data_in (0 to 11);
          pres_l1id <= serial_data_in(0 to 2);
        end if;
      when new_cluster =>
--        if (pres_chip /= serial_data_in(2 to 5) AND bit_count_reset = '1') then
--          hit_counter <= (others => '0');  -- Test for new chip here JMJ 062012
--        end if;
        overflow_trailer <= '0';
        packed_hit_count <= (others => '0');
        if (bit_count_reset = '1') then
          hit_location <= serial_data_in (2 to 12);
          data_out (15 downto 4) <= '1' & serial_data_in (2 to 12);
          pres_chip <= serial_data_in(2 to 5);
          pres_chan <= serial_data_in(6 to 12);
          if (compressed_mode = '1') then
            pack_all_hits <= '1';
          elsif (condensed_mode_i = '1') then
            pack_first_hit <= '1';
          else
            pack_first_hit <= '0';
            new_cluster_flag <= '1';
          end if;
        end if;
        if (next_state = raw_data) then
          data_out <= X"6C0" & serial_data_in (2 to 5);
          data_out_write <= '1';
          wr_raw_data_cluster <= '1';
        end if;

      when hit =>
        if (data_overflow_limit_in /= 0 AND hit_counter >= data_overflow_limit_in) then
          overflow_trailer <= '1';
        end if;

        if (std_logic_vector(next_bit_count) = "0000" and almost_full_flag = '0') then
          pack_first_hit <= not pack_first_hit;          
          hit_counter <= hit_counter + 1 ;
          packed_hit_count <= packed_hit_count + 1 ;

          if (compressed_mode = '0') then
            if (pack_first_hit = '1') then
              if (next_state /= hit) then
                data_out_write <= '1';
       --   Check for Pipeline Mask here
--                if ((link_pipeline_mask(31) = '1'                  AND 
--                     link_pipeline_mask(30 downto 24) = pres_chan  AND
--                     link_pipeline_mask(23 downto 20) = pres_chip  AND
--                     link_pipeline_mask(19 downto 16) = l1a_pipeline_i)
--                    OR
--                    (link_pipeline_mask(15) = '1'                  AND 
--                     link_pipeline_mask(14 downto  8) = pres_chan  AND
--                     link_pipeline_mask( 7 downto  4) = pres_chip  AND
--                     link_pipeline_mask( 3 downto  0) = l1a_pipeline_i)) then
--                else              
--                  data_out_write <= '1';
--                end if;
              end if;
            else
              data_out_write <= '1';
            end if;
          else 
            if (packed_hit_count = 15 OR next_state /= hit) then
              data_out_write <= '1';
              packed_hit_count <= (others => '0');
            else
              data_out_write <= '0';
            end if; 
          end if;
        end if;              

        if (compressed_mode = '0') then
          if (bit_count_reset = '1') then
            if (condensed_mode_i = '1') then
              if (pack_first_hit = '1') then
                data_out(3 downto 0) <= '0' & hit_pattern_invalid & "00";
              else
                data_out(3 downto 0) <=  hit_pattern_invalid & data_out(2) & "01";
              end if;
            else
              if (pack_first_hit = '1') then
                data_out <= "1000000000001" & serial_data_in(1 to 3);
              else
                if (new_cluster_flag = '1') then             
                  data_out(3 downto 0) <= '0' & serial_data_in(1 to 3);
                  new_cluster_flag <= '0';
                else
                  data_out <= "100000001" & data_out(2 downto 0) & '1' & serial_data_in(1 to 3);                
                end if;
              end if;
            end if;
          end if;
        else
          data_out(3 downto 0) <= std_logic_vector(packed_hit_count);
        end if;

      when flagged_error =>
        if (data_overflow_limit_in /= 0 AND
          hit_counter >= data_overflow_limit_in) then
          overflow_trailer <= '1';
        end if;

        if (std_logic_vector(next_bit_count) = "0000" and
            almost_full_flag = '0') then
          data_out_write <= '1';
          hit_counter <= hit_counter + 1 ;
        end if;        
        if (bit_count_reset = '1') then
          data_out <= "000000000" & serial_data_in (3 to 9);
        end if;

      when null_state =>
        overflow_trailer <= '0';

      when wait_for_trailer => --NULL;
        if (std_logic_vector(next_bit_count) = "0000") then
          if (raw_data_count = 0) then
            data_out_write <= '1';
            decoder_wr_trailer_i <= '1';
            raw_data_count <= raw_data_count + '1';
          end if;
        end if;
        if (bit_count_reset = '1') then
          data_out <= "010" & trailer_error_del1 & almost_full_flag & event_truncated &"0000000000";
        end if;

      when trailer =>
        hit_counter <= (others => '0');
        if (std_logic_vector(next_bit_count) = "0000") then
          if (raw_data_count = 0) then
            data_out_write <= '1';
          end if;
          decoder_wr_trailer_i <= '1';
        end if;
        if (bit_count_reset = '1') then
--          data_out <= "010" & trailer_error_del1 & almost_full_flag & event_truncated &"0000000000";
          data_out <= "010" & trailer_error_del1 & almost_full_flag & event_truncated &"00" & masked_chip_id_in;
        end if;

      when raw_data =>
        normal_not_raw <= '0';
        if (config_mode_i = '1') then  -- added 3/5/2008 JMJ
--        normal_not_raw <= config_mode_i;
          data_out <= "011" & std_logic_vector(bit_count(2 downto 0)) &
                       "00" & data_out(6 downto 0) & serial_data_in (0);
          if (((next_state /= raw_data) or
              (std_logic_vector(next_bit_count) = "0000")) and
              almost_full_flag = '0') then
            data_out_write <= '1';
          end if;
        else
          if (wr_raw_data_cluster = '1') then
            data_out <= X"78" & '0' & hit_location(10 downto 4);
            data_out_write <= '1';
            wr_raw_data_cluster <= '0';
          elsif (raw_data_count = 0) then
            data_out <= "011" & std_logic_vector(bit_count(2 downto 0)) &
                        "00" & data_out(6 downto 0) & serial_data_in (0);
            if (((next_state /= raw_data) or
                (std_logic_vector(next_bit_count) = "0000")) and
                almost_full_flag = '0') then
              data_out_write <= '1';
              raw_data_count <= raw_data_count + '1';
            end if;
          elsif (raw_data_count = 1) then
            if (data_strobe_lower = '0') then
              data_out <= "011" & std_logic_vector(bit_count(2 downto 0)) &
                          "00" & data_out(6 downto 0) & serial_data_in (0);
            else 
--              data_out <= X"4400";
              data_out <= X"4600";
            end if;            
            if (((next_state /= raw_data) or
                (std_logic_vector(next_bit_count) = "0000")) and
                almost_full_flag = '0') then
              data_out_write <= '1';
              raw_data_count <= raw_data_count + '1';
            end if;
          elsif (raw_data_count = 2) then
            if (data_strobe_lower = '0') then
              data_out <= X"0000";
            else 
--              data_out <= X"4400";
              data_out <= X"4600";
            end if;            
            if (((next_state /= raw_data) or
                (std_logic_vector(next_bit_count) = "0000")) and
                almost_full_flag = '0') then
              data_out_write <= '1';
              raw_data_count <= raw_data_count + '1';
            end if;
          end if;            
        end if;
    end case;
  end if;
end process;

process (clk_in, rst_n_in)
begin
  if (rst_n_in = '0') then
    data_strobe_lower <= '0';
  elsif (clk_in'event and clk_in = '1') then
    if (data_out_write = '1') then
      if (decoder_wr_trailer_i = '1') then
        data_strobe_lower <= '0';
      else
        data_strobe_lower <= not data_strobe_lower;
      end if;
      data_out_temp <= data_out;
    end if;
  end if;
end process;

-- generate the output write signals
process (
  data_out_write, 
  data_strobe_lower,
  decoder_wr_trailer_i,
  data_out_temp, 
  data_out
  )
begin
  data_strobe  <= '0';
  decoded_data <= (others => '0');
  if (data_out_write = '1') then
    if (data_strobe_lower = '1') then
      data_strobe  <= '1';
      decoded_data <= (data_out_temp & data_out);
    elsif (decoder_wr_trailer_i = '1') then
      data_strobe  <= '1';
      decoded_data <= (data_out & "0000000000000000");
    end if;
  end if;
end process;

decoder_wr_trailer <= decoder_wr_trailer_i;

-- Flip flops holding the current state value are
-- set here 
set_clocked_fsm : process (clk_in, rst_n_in)
begin
  if (rst_n_in = '0') then
    pres_state <= idle;
    bit_count <= "0000";
    resume_bit_count <= '0';
    wait_count <= "00000";
  elsif (clk_in'event and clk_in = '1') then
    pres_state <= next_state;
    bit_count <= next_bit_count;
--********************************** JMJ Ver s2.2f
--    if (pause_bit_count = '0') then
--      bit_count <= next_bit_count;
--    elsif (pause_bit_count = '1') then  -- Wait for the Configuration Bits to clear
--      wait_count <= wait_count + 1 ;
--      if (std_logic_vector(wait_count) = "10000") then
--        resume_bit_count <= '1';
--      end if;
--    end if;
--********************************** JMJ Ver s2.2f
    if (std_logic_vector(next_bit_count(3 downto 0)) = "0000") then
      bit_count_reset <= '1';
      resume_bit_count <= '0';
      wait_count <= "00000";
    else
      bit_count_reset <= '0';
    end if;
  end if;
end process set_clocked_fsm;

end rtl;