--------------------------------------------------------------------------
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 1999
-- ReadOutDriver Electronics
--
--------------------------------------------------------------------------
-- Filename: register_block.vhd
-- Description:
--      Register file for frm Register bus. Examine the comments
--	    specifically to discover the exact register definition.
--
--------------------------------------------------------------------------
-- Structure:
--------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk40
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
--------------------------------------------------------------------------
-- Author: Johm M. Joseph
-- Board Engineer: John M. Joseph
-- History:
--          15 March 2000 JMJ Adapted control registers from 
--          ROD resources: register_block.vhd 
--          
--          07 March 2001 JMJ Added new registers to allow more function
--          16 April 2002 JMJ Revised per new DSP Algorithms
--
--------------------------------------------------------------------------
-- Notes:
--------------------------------------------------------------------------
--  C:/work/Atlas/ROD/fpgaDesigns/rod_revF/fsct/vhdl/current/register_block.vhd 
--  Address
--
--------------------------------------------------------------------------
-- LIBRARY INCLUDES
--------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations
use IEEE.std_logic_unsigned.all; -- needed for unsigned operations
--------------------------------------------------------------------------
-- PORT DECLARATION
--------------------------------------------------------------------------
entity register_block is
  port (
    clk40_in               : in  std_logic; -- clk40 input
    rst_n_in               : in  std_logic; -- async global reset
    frm_bus_strb_in        : in  std_logic; --
    frm_bus_rnw_in         : in  std_logic; --
    frm_bus_hwsb_in        : in  std_logic; --
    frm_bus_ds_in          : in  std_logic; --
    frm_bus_addr_in        : in  std_logic_vector( 7 downto 0); --
    frm_bus_data_in        : in  std_logic_vector(15 downto 0); --
    frm_bus_data_out       : out std_logic_vector(15 downto 0); --
    frm_bus_ack_out        : out std_logic; --
    normal_not_raw         : in  std_logic_vector(11 downto 0); --
    header_trailer_error   : in  std_logic_vector(11 downto 0); --
    rod_busy_error         : in  std_logic_vector(11 downto 0); --
    time_out_error         : in  std_logic_vector(11 downto 0); --
    data_overflow_error    : in  std_logic_vector(11 downto 0); --
    occupancy_count_link0  : in  std_logic_vector( 8 downto 0); --
    occupancy_count_link1  : in  std_logic_vector( 8 downto 0); --
    occupancy_count_link2  : in  std_logic_vector( 8 downto 0); --
    occupancy_count_link3  : in  std_logic_vector( 8 downto 0); --
    occupancy_count_link4  : in  std_logic_vector( 8 downto 0); --
    occupancy_count_link5  : in  std_logic_vector( 8 downto 0); --
    occupancy_count_link6  : in  std_logic_vector( 8 downto 0); --
    occupancy_count_link7  : in  std_logic_vector( 8 downto 0); --
    occupancy_count_link8  : in  std_logic_vector( 8 downto 0); --
    occupancy_count_link9  : in  std_logic_vector( 8 downto 0); --
    occupancy_count_link10 : in  std_logic_vector( 8 downto 0); --
    occupancy_count_link11 : in  std_logic_vector( 8 downto 0); --
    mode_bits              : in  std_logic_vector(23 downto 0); --
    trigger_count          : in  std_logic_vector( 9 downto 0); --
    hold_output            : in  std_logic; --
    frm_dll_locked         : in  std_logic; --
    master_not_slave       : in  std_logic; --
    timeout_limit          : out unsigned( 7 downto 0); --
    overflow_limit         : out unsigned(11 downto 0); --
    header_trailer_limit   : out unsigned( 8 downto 0); --
    rod_busy_limit         : out unsigned( 8 downto 0); --
    link_enabled           : out std_logic_vector(11 downto 0); --
    expanded_mode          : out std_logic_vector(11 downto 0); --
    config_mode            : out std_logic_vector(11 downto 0); --
    compressed_mode        : out std_logic_vector(11 downto 0); --
    edge_mode              : out std_logic_vector(11 downto 0); --
    mb_fifo_rd             : out std_logic;
    mb_fifo_ef             : in  std_logic; --
    mb_fifo_ff             : in  std_logic; --
    mb_occ_count_in        : in  std_logic_vector( 8 downto 0);
    mb_rst_n_in            : in  std_logic;
    active_link_status     : in  std_logic_vector( 3 downto 0); --
    data_link_out_mux_sel  : out std_logic_vector( 3 downto 0); --
    chip_has_token_in      : in  std_logic;
    readout_status_in      : in  std_logic_vector( 3 downto 0);
    ht_limit_en_n_out      : out std_logic;
    rb_limit_en_n_out      : out std_logic;
    dovr_error_n_out       : out std_logic;
    fmt_type_in            : in  std_logic_vector( 1 downto 0);
    rod_type_out           : out std_logic;
    spare_pin_reg_in       : in  std_logic_vector( 7 downto 0);
    link_fifo_select       : out std_logic_vector( 3 downto 0);
    link_fifo_data_in      : in  std_logic_vector(31 downto 0);
    link_fifo_data_out     : out std_logic_vector(31 downto 0); 
    link_fifo_rd           : out std_logic;
    link_fifo_wr           : out std_logic;
    boc_scan_data_in       : in  std_logic_vector(11 downto 0);
    config_simu            : out std_logic_vector(31 downto 0);
    l1a_in                 : in  std_logic;
    bcr_in                 : in  std_logic;
    ecr_in                 : in  std_logic;
    link_mask_en_out       : out std_logic_vector(11 downto 0);
    cal_mode_sel_out       : out std_logic;
    
    masked_chip_id_out     : out std_logic_vector(95 downto 0);

--    link0_pipeline_mask    : buffer std_logic_vector(31 downto 0);
--    link1_pipeline_mask    : buffer std_logic_vector(31 downto 0);
--    link2_pipeline_mask    : buffer std_logic_vector(31 downto 0);
--    link3_pipeline_mask    : buffer std_logic_vector(31 downto 0);
--    link4_pipeline_mask    : buffer std_logic_vector(31 downto 0);
--    link5_pipeline_mask    : buffer std_logic_vector(31 downto 0);
--    link6_pipeline_mask    : buffer std_logic_vector(31 downto 0);
--    link7_pipeline_mask    : buffer std_logic_vector(31 downto 0);
--    link8_pipeline_mask    : buffer std_logic_vector(31 downto 0);
--    link9_pipeline_mask    : buffer std_logic_vector(31 downto 0);
--    linkA_pipeline_mask    : buffer std_logic_vector(31 downto 0);
--    linkB_pipeline_mask    : buffer std_logic_vector(31 downto 0);
    
    diag_fifo_flush        : out std_logic;
    diag_fifo_wr_en        : out std_logic;
    diag_fifo_rd_en        : out std_logic;
    diag_fifo_rd_pulse     : out std_logic;
    diag_fifo_trig         : out std_logic;
    diag_fifo_sw_trig_en   : out std_logic;
    diag_fifo_ef_in        : in  std_logic;
    diag_fifo_ff_in        : in  std_logic;
    diag_fifo_occ_in       : in  std_logic_vector(11 downto 0);
    diag_data_fifo_in      : in  std_logic_vector(11 downto 0);
	 
    hit_count_rst_out      : out std_logic_vector(11 DOWNTO 0);
    link01_hit_count_in    : in  std_logic_vector(23 DOWNTO 0); 
    link23_hit_count_in    : in  std_logic_vector(23 DOWNTO 0); 
    link45_hit_count_in    : in  std_logic_vector(23 DOWNTO 0); 
    link67_hit_count_in    : in  std_logic_vector(23 DOWNTO 0); 
    link89_hit_count_in    : in  std_logic_vector(23 DOWNTO 0); 
    linkAB_hit_count_in    : in  std_logic_vector(23 DOWNTO 0);
	 
    decoder_active_in      : in  std_logic_vector(11 downto 0);
    link_data_tx_in        : in  std_logic_vector(11 downto 0)
    );   
end register_block;

architecture rtl of register_block is
--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------

signal edge_mode_i        : std_logic_vector(11 downto 0); --
signal config_mode_i      : std_logic_vector(11 downto 0); --
signal expanded_mode_i    : std_logic_vector(11 downto 0); --
signal compressed_mode_i  : std_logic_vector(11 downto 0); -- 
signal link_enabled_i     : std_logic_vector(11 downto 0); --
signal timeout_limit_i    : std_logic_vector( 7 downto 0); --
signal overflow_limit_i   : std_logic_vector(11 downto 0); --
signal ht_limit_i         : std_logic_vector( 8 downto 0); --
signal rod_busy_limit_i   : std_logic_vector( 8 downto 0); --
signal dlink_out_mux_i    : std_logic_vector( 3 downto 0); --
signal time_out_error_i   : std_logic_vector(11 downto 0); --
signal clr_to_status_i    : std_logic;
signal clr_to_reg_i       : std_logic;
signal dovr_error_i       : std_logic_vector(11 downto 0); --
signal clr_dovr_status_i  : std_logic;
signal clr_dovr_reg_i     : std_logic;
signal normal_not_raw_i   : std_logic_vector(11 downto 0); --
signal clr_nnr_status_i   : std_logic;
signal clr_nnr_reg_i      : std_logic;
signal rbl_error_i        : std_logic_vector(11 downto 0); --
signal clr_rbl_status_i   : std_logic;
signal clr_rbl_reg_i      : std_logic;
signal htl_error_i        : std_logic_vector(11 downto 0); --
signal clr_htl_status_i   : std_logic;
signal clr_htl_reg_i      : std_logic;
signal en_mb_fifo_rd_i    : std_logic;
signal en_link_fifo_rd_i  : std_logic;
signal en_link_fifo_wr_i  : std_logic;
signal link_fifo_select_i : std_logic_vector( 3 downto 0);
signal cal_mode_sel_i     : std_logic;
--signal mb_rst_n_i         : std_logic;
  
type states is (
  idle, 
  acknowledge, 
  wait_cycle
  );
signal acknowledge_cycle_state : states;

signal boc_scan_crtl     : std_logic_vector( 1 downto 0);
signal boc_scan_stat     : std_logic_vector( 1 downto 0);
signal boc_scan_gcount   : std_logic_vector(15 downto 0); --
signal boc_scan_count    : unsigned(15 downto 0); --
type   boc_scan_mcount_array is array (0 to 11) of unsigned(15 downto 0); --
signal boc_scan_mcount : boc_scan_mcount_array; --

signal config_simu_i : std_logic_vector(31 downto 0);

signal l1id_i : std_logic_vector(23 downto 0);
signal bcr_count : std_logic_vector(15 downto 0);
signal ecr_count : std_logic_vector(15 downto 0);

signal link_mask_en_i    : std_logic_vector(11 downto 0);
signal clr_link_mask     : std_logic;
signal clr_hit_count_rst : std_logic;
signal hit_count_rst_o   : std_logic_vector(11 downto 0);
signal masked_chip_id    : std_logic_vector(95 downto 0);

type link_mask_states is (
  idle, 
  wait_for_L1A, 
  done
  );
signal link_mask_state : link_mask_states;

signal diagnostics_i          : std_logic_vector(31 downto 0);

signal diag_fifo_trig_o       : std_logic;
signal diag_fifo_sw_trig_en_o : std_logic;
signal diag_fifo_flush_o      : std_logic;
signal diag_fifo_wr_en_o      : std_logic;
signal diag_fifo_rd_en_o      : std_logic;
signal pulse_width            : unsigned(1 downto 0);

-------------------------------------------------------------------------------
-- PROCESS DECLARATION
-------------------------------------------------------------------------------
begin

-------------------------------------------------------------------------------
-- Process to Enable Read/Write to Router Registers
-------------------------------------------------------------------------------
rw_registers : process (
  clk40_in, 
  rst_n_in,
  frm_bus_addr_in,
  frm_bus_rnw_in,
  frm_bus_strb_in
  )
constant code_version : std_logic_vector(7 downto 0) := X"55";   -- V55.7
constant code_ver2    : std_logic_vector(3 downto 0) := X"8";
constant board_rev    : std_logic_vector(4 downto 0) := "01111"; -- RevF
constant rod_type     : std_logic := '0';                        -- SCT
begin
  if (rst_n_in = '0') then
    frm_bus_data_out   <= (others => '0');
----*********************************************************
--            Synth/Build values
    link_enabled_i         <= (others => '0');
    link_mask_en_i         <= (others => '0');
    diagnostics_i          <= (others => '0');
    timeout_limit_i        <= X"00";  -- 0 us + 75 ns
    config_simu_i          <= (others => '0');
    config_mode_i          <= (others => '0');
    edge_mode_i            <= (others => '0');
    expanded_mode_i        <= (others => '0');
    compressed_mode_i      <= (others => '0');
    overflow_limit_i       <= X"640";  -- (others => '0'); -- 1600 hits
    clr_link_mask          <= '0';
    diag_fifo_flush_o      <= '0';
    diag_fifo_wr_en_o      <= '0';
    diag_fifo_rd_en_o      <= '0';
    diag_fifo_rd_pulse     <= '0';
    diag_fifo_trig_o       <= '0';
    diag_fifo_sw_trig_en_o <= '0';
    pulse_width            <= "00";
    masked_chip_id         <= (others => '0');

----*********************************************************
--            Simulation values
-- pragma translate_off
    link_enabled_i     <= (others => '1');
    link_mask_en_i     <= X"000";
    diagnostics_i      <= (others => '0');
    timeout_limit_i    <= X"00";  -- 0 us + 75 ns
    config_simu_i      <= (others => '0'); --X"0C885001"; -- ON, 6 Chips, All hits, No Error
    config_mode_i      <= (others => '0');
    edge_mode_i        <= (others => '0');
    expanded_mode_i    <= (others => '0');
    compressed_mode_i  <= (others => '1');
    overflow_limit_i   <= X"020";  --X"180";
    clr_link_mask      <= '1';
    diag_fifo_flush_o  <= '0';
    diag_fifo_wr_en_o  <= '1';
    diag_fifo_rd_en_o  <= '0';
    diag_fifo_rd_pulse <= '0';
    masked_chip_id     <= (others => '0');
-- pragma translate_on
----*********************************************************
    masked_chip_id_out <= (others => '0');

    cal_mode_sel_i     <= '0';
    ht_limit_i         <= "111000000";  -- 0x1C0 == 448 words  
    rod_busy_limit_i   <= "111110000";  -- 0x1F0 == 496 words (4B-1 >> "110100000";  -- 0x1A0 == 416 words
    en_mb_fifo_rd_i    <= '0';
    clr_to_status_i    <= '0';
    clr_dovr_status_i  <= '0';
    clr_nnr_status_i   <= '0';
    clr_htl_status_i   <= '0';
    clr_rbl_status_i   <= '0';
    rod_type_out       <= '0';
    link_fifo_data_out <= (others => '0');
    en_link_fifo_rd_i  <= '0';
    en_link_fifo_wr_i  <= '0';
    link_fifo_select_i <= (others => '1');
--    mb_rst_n_i         <= '1';
    boc_scan_crtl      <= (others => '0');
    boc_scan_gcount    <= (others => '0');
    link_mask_en_out   <= (others => '0');
    link_mask_state    <= idle;
    hit_count_rst_o    <= (others => '0');
    
--    link0_pipeline_mask <= (others => '0');
--    link1_pipeline_mask <= (others => '0');
--    link2_pipeline_mask <= (others => '0');
--    link3_pipeline_mask <= (others => '0');
--    link4_pipeline_mask <= (others => '0');
--    link5_pipeline_mask <= (others => '0');
--    link6_pipeline_mask <= (others => '0');
--    link7_pipeline_mask <= (others => '0');
--    link8_pipeline_mask <= (others => '0');
--    link9_pipeline_mask <= (others => '0');
--    linkA_pipeline_mask <= (others => '0');
--    linkB_pipeline_mask <= (others => '0');
    
  elsif (clk40_in'event AND clk40_in = '1') then
 --Read registers
    frm_bus_data_out  <= (others => '0');
    if (clr_hit_count_rst = '1') then
      hit_count_rst_o <= (others => '0');
    end if;
    if (frm_bus_rnw_in = '1' AND frm_bus_strb_in = '1') then
      case frm_bus_addr_in is
        when X"00" =>  -- 0x004nn00
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(11 downto 0) <= link_enabled_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"01" =>  -- 0x004nn04
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(11 downto 0) <= expanded_mode_i;
          else
            frm_bus_data_out(11 downto 0) <= compressed_mode_i;
          end if;
        when X"02" =>  -- 0x004nn08
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(11 downto 0) <= config_mode_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"03" =>  -- 0x004nn0C
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(11 downto 0) <= edge_mode_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"04" =>  -- 0x004nn10
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= cal_mode_sel_i & "0000000" & timeout_limit_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"05" =>  -- 0x004nn14
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(11 downto 0) <= overflow_limit_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"06" =>  -- 0x004nn18
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= ht_limit_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"07" =>  -- 0x004nn1C
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= rod_busy_limit_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        -- X"08" >> 0x004nn20
        -- X"09" >> 0x004nn24
        when X"0A" =>  -- 0x0040028
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(11 downto 0) <= link_mask_en_i;
          else
            frm_bus_data_out <= clr_link_mask & "000" & X"000";
          end if;
        -- X"0B" >> 0x004nn2C
        when X"0C" => -- >> 0x004nn30  L1A count  
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= l1id_i(15 downto 0);
          else
            frm_bus_data_out(15 downto 12) <= bcr_in & ecr_in & l1a_in & mb_rst_n_in;
            frm_bus_data_out(11 downto  8) <= X"0";
            frm_bus_data_out( 7 downto  0) <= l1id_i(23 downto 16);
          end if;
        when X"0D" => -- ECR & BCR count
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= ecr_count;
          else
            frm_bus_data_out <= bcr_count;
          end if;
        when X"0E" =>
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(3 downto 0) <= dlink_out_mux_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"0F" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out( 0)          <= en_mb_fifo_rd_i;
            frm_bus_data_out( 1)          <= en_link_fifo_rd_i;
            frm_bus_data_out( 2)          <= en_link_fifo_wr_i;
            frm_bus_data_out( 7 downto 4) <= link_fifo_select_i;
            frm_bus_data_out(15 downto 8) <= spare_pin_reg_in;  -- place holder for spare pins
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"10" =>
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link0;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"11" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link1;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"12" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link2;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"13" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link3;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"14" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link4;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"15" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link5;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"16" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link6;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"17" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link7;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"18" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link8;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"19" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link9;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"1A" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link10;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"1B" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(8 downto 0) <= occupancy_count_link11;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"1C" =>
          clr_to_status_i <= '1';
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(11 downto 0) <= time_out_error_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"1D" =>
          clr_dovr_status_i <= '1';
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(11 downto 0) <= dovr_error_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"1E" =>
          clr_htl_status_i <= '1';
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(11 downto 0) <= htl_error_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"1F" =>
          clr_rbl_status_i <= '1';
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(11 downto 0) <= rbl_error_i;   --rod_busy_error;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"20" =>
          clr_nnr_status_i <= '1';
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out(11 downto 0) <= normal_not_raw_i;
          else
            frm_bus_data_out <= (others => '0');
          end if;
        when X"21" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= frm_dll_locked     &
                                master_not_slave   &
                                hold_output        &
                                chip_has_token_in  &
                                active_link_status &
                                mb_fifo_ff         &
                                mb_fifo_ef         &
                                '1'                &  --mb_rst_n_i         &
                                trigger_count(4 downto 0);
	       elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <=  "000000" & trigger_count;
          end if;
        when X"22" => 
	       if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= rod_type       & 
                                fmt_type_in(1) & 
                                fmt_type_in(0) &
                                board_rev      & 
                                code_version;
	        elsif (frm_bus_hwsb_in = '1') then
             frm_bus_data_out <=  mb_occ_count_in & "000" & code_ver2;
           end if;
        when X"23" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= mode_bits(15 downto  0);
          elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= readout_status_in & "0000" & mode_bits(23 downto 16);
          end if;
        when X"24" => 
     	    if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out ( 1 downto 0) <= boc_scan_crtl(1 downto 0);
            frm_bus_data_out (15 downto 2) <= (others => '0');
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out  <= (others => '0');
          end if;
        when X"25" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= boc_scan_gcount;
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= std_logic_vector(boc_scan_count);
          end if;
        when X"26" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(0));
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(1));
          end if;
        when X"27" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(2));
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(3));
          end if;
        when X"28" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(4));
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(5));
          end if;
        when X"29" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(6));
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(7));
          end if;
        when X"2A" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(8));
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(9));
          end if;
        when X"2B" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(10));
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= std_logic_vector(boc_scan_mcount(11));
          end if;
        when X"2C" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= config_simu_i(15 downto  0);
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= config_simu_i(31 downto 16);
          end if;

        when X"2D" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= X"00" & "000"            & 
                                  diag_fifo_trig_o       &
                                  diag_fifo_sw_trig_en_o & 
                                  diag_fifo_rd_en_o      & 
                                  diag_fifo_wr_en_o      & 
                                  diag_fifo_flush_o;
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= X"0000";
          end if;
        when X"2E" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <=  X"0" & diag_fifo_occ_in;
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= X"000" & "00" & diag_fifo_ff_in & diag_fifo_ef_in;
          end if;
        when X"2F" => 
          if (pulse_width = 1) then
            diag_fifo_rd_pulse <= '1';
            pulse_width <= pulse_width + 1 ;
          elsif (pulse_width = 3) then
            diag_fifo_rd_pulse <= '0';
          else
            diag_fifo_rd_pulse <= '0';
            pulse_width <= pulse_width + 1 ;
          end if;               
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= X"0" & diag_data_fifo_in;
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= X"0000";
            pulse_width      <= "00";
           end if;

        when X"30" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= link0_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(7 downto 0);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= link0_pipeline_mask(31 downto 16);
          end if;
        when X"31" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= link1_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(15 downto 8);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= link1_pipeline_mask(31 downto 16);
          end if;
        when X"32" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= link2_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(23 downto 16);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= link2_pipeline_mask(31 downto 16);
          end if;
        when X"33" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= link3_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(31 downto 24);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= link3_pipeline_mask(31 downto 16);
          end if;
        when X"34" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= link4_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(39 downto 32);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= link4_pipeline_mask(31 downto 16);
          end if;
        when X"35" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= link5_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(47 downto 40);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= link5_pipeline_mask(31 downto 16);
          end if;
        when X"36" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= link6_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(55 downto 48);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= link6_pipeline_mask(31 downto 16);
          end if;
        when X"37" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= link7_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(63 downto 56);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= link7_pipeline_mask(31 downto 16);
          end if;
        when X"38" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= link8_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(71 downto 64);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= link8_pipeline_mask(31 downto 16);
          end if;
        when X"39" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= link9_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(79 downto 72);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= link9_pipeline_mask(31 downto 16);
          end if;
        when X"3A" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= linkA_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(87 downto 80);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= linkA_pipeline_mask(31 downto 16);
          end if;
        when X"3B" => 
	        if (frm_bus_hwsb_in = '0') then
--            frm_bus_data_out <= linkB_pipeline_mask(15 downto  0);
            frm_bus_data_out(7 downto 0) <= masked_chip_id(95 downto 88);
--	        elsif (frm_bus_hwsb_in = '1') then
--            frm_bus_data_out <= linkB_pipeline_mask(31 downto 16);
          end if;
        when X"3F" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= link_fifo_data_in(15 downto  0);
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= link_fifo_data_in(31 downto 16);
          end if;
        when X"40" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= X"0" & link01_hit_count_in(11 downto  0);
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= X"0" & link01_hit_count_in(23 downto 12);
          end if;
          hit_count_rst_o <= X"003";
        when X"41" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= X"0" & link23_hit_count_in(11 downto  0);
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= X"0" & link23_hit_count_in(23 downto 12);
          end if;
          hit_count_rst_o <= X"00C";
        when X"42" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= X"0" & link45_hit_count_in(11 downto  0);
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= X"0" & link45_hit_count_in(23 downto 12);
          end if;
          hit_count_rst_o <= X"030";
        when X"43" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= X"0" & link67_hit_count_in(11 downto  0);
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= X"0" & link67_hit_count_in(23 downto 12);
          end if;
          hit_count_rst_o <= X"0C0";
        when X"44" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= X"0" & link89_hit_count_in(11 downto  0);
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= X"0" & link89_hit_count_in(23 downto 12);
          end if;
          hit_count_rst_o <= X"300";
        when X"45" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= X"0" & linkAB_hit_count_in(11 downto  0);
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= X"0" & linkAB_hit_count_in(23 downto 12);
          end if;
          hit_count_rst_o <= X"C00";
--      when X"46" => 
--      when X"47" => 
        when X"48" => 
	        if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= X"0" & decoder_active_in;
	        elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= X"0" & link_data_tx_in;
          end if;
        when X"80" => 
          if (frm_bus_hwsb_in = '0') then
            frm_bus_data_out <= diagnostics_i(15 downto  0);
          elsif (frm_bus_hwsb_in = '1') then
            frm_bus_data_out <= diagnostics_i(31 downto 16);
          end if;

        when others => frm_bus_data_out <= (others => '0');
      end case;
 --Write registers
    elsif (frm_bus_rnw_in = '0' AND frm_bus_strb_in = '1') then
      case frm_bus_addr_in is
        when X"00" =>
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            link_enabled_i <= frm_bus_data_in(11 downto 0);
--        elsif (frm_bus_hwsb_in = '1' AND frm_bus_ds_in = '1') then
          end if;
        when X"01" =>
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            expanded_mode_i <= frm_bus_data_in(11 downto 0);
          elsif (frm_bus_hwsb_in = '1' AND frm_bus_ds_in = '1') then
            compressed_mode_i <= frm_bus_data_in(11 downto 0);
          end if;
        when X"02" =>
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            config_mode_i <= frm_bus_data_in(11 downto 0);
          end if;
        when X"03" =>
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            edge_mode_i <= frm_bus_data_in(11 downto 0);
          end if;
        when X"04" =>
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            cal_mode_sel_i  <= frm_bus_data_in(15); 
            timeout_limit_i <= frm_bus_data_in(7 downto 0);
          end if;
        when X"05" =>
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            overflow_limit_i <= frm_bus_data_in(11 downto 0);
          end if;
        when X"06" =>
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            ht_limit_i <= frm_bus_data_in( 8 downto 0);
          end if;
        when X"07" =>
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            rod_busy_limit_i <= frm_bus_data_in( 8 downto 0);
          end if;
        when X"0A" =>  -- 0x0040028
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            link_mask_en_i <= frm_bus_data_in(11 downto 0);
          elsif (frm_bus_hwsb_in = '1' AND frm_bus_ds_in = '1') then
            clr_link_mask <= frm_bus_data_in(15);
          end if;

        when X"0E" =>
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            dlink_out_mux_i <= frm_bus_data_in( 3 downto 0);
          end if;
        when X"0F" =>
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
             en_mb_fifo_rd_i    <= frm_bus_data_in( 0);
             en_link_fifo_rd_i  <= frm_bus_data_in( 1);
             en_link_fifo_wr_i  <= frm_bus_data_in( 2);
             link_fifo_select_i <= frm_bus_data_in( 7 downto 4);
          end if;
        when X"24" => 
	        if (frm_bus_hwsb_in = '0') then
            boc_scan_crtl(1 downto 0) <= frm_bus_data_in(1 downto 0);
          end if;
        when X"25" => 
	        if (frm_bus_hwsb_in = '0') then
            boc_scan_gcount <= frm_bus_data_in;
          end if;
        when X"2C" => 
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            config_simu_i(15 downto  0) <= frm_bus_data_in;
          elsif (frm_bus_hwsb_in = '1' AND frm_bus_ds_in = '1') then
            config_simu_i(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"2D" => 
	        if (frm_bus_hwsb_in = '0') then
            diag_fifo_trig_o       <= frm_bus_data_in(4);
            diag_fifo_sw_trig_en_o <= frm_bus_data_in(3);
            diag_fifo_rd_en_o      <= frm_bus_data_in(2);
            diag_fifo_wr_en_o      <= frm_bus_data_in(1);
            diag_fifo_flush_o      <= frm_bus_data_in(0);
           end if;
--        when X"2E" =>  -- Diag FIFO Status 
--        when X"2F" =>  -- FIFO Data Mail Box
        when X"30" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id( 7 downto  0) <= frm_bus_data_in(7 downto 0);
--            link0_pipeline_mask(15 downto  0) <= frm_bus_data_in;
--	        elsif (frm_bus_hwsb_in = '1') then
--            link0_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"31" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id( 15 downto  8) <= frm_bus_data_in(7 downto 0);
--            link1_pipeline_mask(15 downto  0) <= frm_bus_data_in;
--	        elsif (frm_bus_hwsb_in = '1') then
--            link1_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"32" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id(23 downto  16) <= frm_bus_data_in(7 downto 0);
--            link2_pipeline_mask(15 downto  0) <= frm_bus_data_in;
--	        elsif (frm_bus_hwsb_in = '1') then
--            link2_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"33" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id(31 downto 24) <= frm_bus_data_in(7 downto 0);
--            link3_pipeline_mask(15 downto  0) <= frm_bus_data_in;
--	        elsif (frm_bus_hwsb_in = '1') then
--            link3_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"34" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id(39 downto 32) <= frm_bus_data_in(7 downto 0);
--            link4_pipeline_mask(15 downto  0) <= frm_bus_data_in;
--	        elsif (frm_bus_hwsb_in = '1') then
--            link4_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"35" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id(47 downto 40) <= frm_bus_data_in(7 downto 0);
--            link5_pipeline_mask(15 downto  0) <= frm_bus_data_in;
--	        elsif (frm_bus_hwsb_in = '1') then
--            link5_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"36" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id(55 downto 48) <= frm_bus_data_in(7 downto 0);
--            link6_pipeline_mask(15 downto  0) <= frm_bus_data_in;
--	        elsif (frm_bus_hwsb_in = '1') then
--            link6_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"37" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id(63 downto 56) <= frm_bus_data_in(7 downto 0);
--            link7_pipeline_mask(15 downto  0) <= frm_bus_data_in;
--	        elsif (frm_bus_hwsb_in = '1') then
--            link7_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"38" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id(71 downto 64) <= frm_bus_data_in(7 downto 0);
--            link8_pipeline_mask(15 downto  0) <= frm_bus_data_in;
--	        elsif (frm_bus_hwsb_in = '1') then
--            link8_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"39" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id(79 downto 72) <= frm_bus_data_in(7 downto 0);
--            link9_pipeline_mask(15 downto  0) <= frm_bus_data_in;
--	        elsif (frm_bus_hwsb_in = '1') then
--            link9_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"3A" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id(87 downto 80) <= frm_bus_data_in(7 downto 0);
--            linkA_pipeline_mask(15 downto  0) <= frm_bus_data_in;
--	        elsif (frm_bus_hwsb_in = '1') then
--            linkA_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
        when X"3B" => 
	        if (frm_bus_hwsb_in = '0') then
            masked_chip_id(95 downto 88) <= frm_bus_data_in(7 downto 0);
--            linkB_pipeline_mask(15 downto  0) <= frm_bus_data_in;-
--	        elsif (frm_bus_hwsb_in = '1') then
--            linkB_pipeline_mask(31 downto 16) <= frm_bus_data_in;
          end if;
--        when X"3F" => 
--          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
--            link_fifo_data_out(15 downto  0) <= frm_bus_data_in;
--          elsif (frm_bus_hwsb_in = '1' AND frm_bus_ds_in = '1') then
--            link_fifo_data_out(31 downto 16) <= frm_bus_data_in;
--          end if;
        when X"80" => 
          if (frm_bus_hwsb_in = '0' AND frm_bus_ds_in = '1') then
            diagnostics_i(15 downto  0) <= frm_bus_data_in;
          elsif (frm_bus_hwsb_in = '1' AND frm_bus_ds_in = '1') then
            diagnostics_i(31 downto 16) <= frm_bus_data_in;
          end if;
        when others     => -- NOP
      end case;
    else
      expanded_mode         <= expanded_mode_i;
      compressed_mode       <= compressed_mode_i;
      config_mode           <= config_mode_i;
      edge_mode             <= edge_mode_i;
      timeout_limit         <= unsigned(timeout_limit_i);
      overflow_limit        <= unsigned(overflow_limit_i);
      header_trailer_limit  <= unsigned(ht_limit_i);
      rod_busy_limit        <= unsigned(rod_busy_limit_i);
      en_mb_fifo_rd_i       <= '0';
      data_link_out_mux_sel <= dlink_out_mux_i;
      clr_to_status_i       <= '0';
      clr_dovr_status_i     <= '0';
      clr_nnr_status_i      <= '0';
      clr_htl_status_i      <= '0';
      clr_rbl_status_i      <= '0';
      rod_type_out          <= rod_type;
      link_fifo_select      <= link_fifo_select_i;
      config_simu           <= config_simu_i;
      cal_mode_sel_out      <= cal_mode_sel_i;
      diag_fifo_flush       <= diag_fifo_flush_o;
      diag_fifo_wr_en       <= diag_fifo_wr_en_o;
      diag_fifo_rd_en       <= diag_fifo_rd_en_o;
      diag_fifo_rd_pulse    <= '0';
      diag_fifo_trig        <= diag_fifo_trig_o;
      diag_fifo_sw_trig_en  <= diag_fifo_sw_trig_en_o;

      if (l1a_in = '1') then
        masked_chip_id_out <= masked_chip_id;
      end if;              

      if (diagnostics_i(0) = '0') then  -- Link Enable & Mask Logic per v54-0
        link_enabled  <= link_enabled_i;
		  for i in 0 to 11 loop
          if(link_mask_en_i(i) = '1') then
            if (l1a_in = '1' OR rbl_error_i(i) = '1') then  -- comment out to return to 50-B
              link_mask_en_out(i) <= '1';
            end if;                                         -- comment out to return to 50-B
          else
            case link_mask_state is
              when idle =>
                if (ecr_in = '1' OR clr_link_mask = '1') then
                  link_mask_en_out(i) <= '0';  --link_mask_en_i(i);
                  link_mask_state <= done;
                end if;
              when others => 
                link_mask_state <= idle;
            end case;
          end if;
        end loop;
		elsif (diagnostics_i(0) = '1') then  -- Link Enable & Mask Logic per v54-8
        for i in 0 to 11 loop
          if (link_enabled_i(i) = '1') then
            if (config_mode_i(i) = '1' OR l1a_in = '1') then
              link_enabled(i) <= '1';  --link_enabled_i(i);
            end if;

            if(link_mask_en_i(i) = '1') then
              if (decoder_active_in(i) = '0' AND link_data_tx_in(i) = '0') then
                link_mask_en_out(i) <= '1';
              end if;                                         -- comment out to return to 50-B
            else
              case link_mask_state is
                when idle =>
                  if (l1a_in = '1' OR clr_link_mask = '1' OR link_enabled_i(i) = '0') then
                    link_mask_en_out(i) <= link_mask_en_i(i);
                    link_mask_state <= done;
                  end if;
                  when others => 
                  link_mask_state <= idle;
              end case;
            end if;
          else
            link_enabled(i) <= link_enabled_i(i);
            link_mask_en_out(i) <= link_mask_en_i(i);
          end if;
        end loop;
		end if;
    end if;
--    mb_rst_n_i <= mb_rst_n_in;
  end if;
end process rw_registers;


event_id_counters : process (
  rst_n_in, 
  clk40_in
  )
begin                      
  if (rst_n_in = '0') then
    l1id_i <= (others => '1');
    bcr_count <= (others => '0');
    ecr_count <= (others => '0');
  elsif (clk40_in'event and clk40_in = '1') then
    if (ecr_in = '1') then
      l1id_i <= (others => '1');
      ecr_count <= ecr_count + '1';
    elsif (l1a_in = '1') then
      l1id_i <= l1id_i + '1';
    end if;

    if (bcr_in = '1') then
      bcr_count <= bcr_count + '1';
    end if;
  end if;  
end process;     


error_status : process (
  rst_n_in, 
  clk40_in
  )
begin                      
  if (rst_n_in = '0') then
    time_out_error_i  <= (others => '0');
    dovr_error_i      <= (others => '0');
    htl_error_i       <= (others => '0');
    rbl_error_i       <= (others => '0');
    normal_not_raw_i  <= (others => '1');
    ht_limit_en_n_out <= '1';
    rb_limit_en_n_out <= '1';
  elsif (clk40_in'event and clk40_in = '1') then
    if (clr_to_reg_i = '1') then
      time_out_error_i <= time_out_error;
    else
      time_out_error_i <= time_out_error_i OR time_out_error;
    end if;

    if (clr_dovr_reg_i = '1') then
      dovr_error_i <= data_overflow_error;
    else
      dovr_error_i <= dovr_error_i OR data_overflow_error;
    end if;

    if (dovr_error_i = "000000000000") then
      dovr_error_n_out <= '1';
    else
      dovr_error_n_out <= '0';
    end if;

    if (clr_nnr_reg_i = '1') then
      normal_not_raw_i <= normal_not_raw;
    else
      normal_not_raw_i <= NOT((NOT normal_not_raw_i) OR (NOT normal_not_raw));
	--      normal_not_raw_i <= normal_not_raw_i AND normal_not_raw;
    end if;

    if (clr_htl_reg_i = '1') then
      htl_error_i <= header_trailer_error;
    else
      htl_error_i <= htl_error_i OR header_trailer_error;
    end if;

    if (header_trailer_error = "000000000000") then
      ht_limit_en_n_out <= '1';
    else
      ht_limit_en_n_out <= '0';
    end if;

    if (clr_rbl_reg_i = '1') then
      rbl_error_i <= rod_busy_error AND link_enabled_i;
    else
      rbl_error_i <= (rbl_error_i OR rod_busy_error) AND link_enabled_i;
    end if;

    if (rod_busy_error = "000000000000") then
      rb_limit_en_n_out <= '1';
    else
      rb_limit_en_n_out <= '0';
    end if;
  end if;  
end process error_status;     

boc_scan : process( 
  rst_n_in, 
  clk40_in
  )
begin                      
  if (rst_n_in = '0') then
    boc_scan_count <= (others => '0');
    for i in 0 to 11 loop 
      boc_scan_mcount(i) <= (others => '0');
    end loop;
  elsif (clk40_in'event and clk40_in = '1') then
    if (boc_scan_crtl(0) = '0') then
      for i in 0 to 11 loop 
        boc_scan_mcount(i) <= (others => '0');
      end loop;
      boc_scan_count <= unsigned(boc_scan_gcount);  -- load value only in reset
    elsif (boc_scan_crtl(0) = '1' AND boc_scan_count > 0) then
      boc_scan_count <= boc_scan_count - 1 ;
      for i in 0 to 11 loop 
        boc_scan_mcount(i) <= boc_scan_mcount(i) + boc_scan_data_in(i);
      end loop;
    end if;
  end if;  
end process boc_scan; 

-------------------------------------------------------------------------------
-- Logic to generate Access Acknowledge
-------------------------------------------------------------------------------
access_acknowledge : process (
  rst_n_in, 
  clk40_in, 
  frm_bus_strb_in
  )
begin                      
  if (rst_n_in = '0') then
    frm_bus_ack_out   <= '0';
    clr_to_reg_i      <= '0';
    clr_dovr_reg_i    <= '0';
    clr_nnr_reg_i     <= '0';
    clr_htl_reg_i     <= '0';
    clr_rbl_reg_i     <= '0';
    mb_fifo_rd        <= '0';
    link_fifo_rd      <= '0';
    link_fifo_wr      <= '0';
	  clr_hit_count_rst <= '0';
    hit_count_rst_out <= (others => '0');
    acknowledge_cycle_state <= idle;
  elsif (clk40_in'event and clk40_in = '1') then
    case acknowledge_cycle_state is
      when idle => 
        frm_bus_ack_out   <= '0';
        clr_to_reg_i      <= '0';
        clr_dovr_reg_i    <= '0';
        clr_nnr_reg_i     <= '0';
        clr_htl_reg_i     <= '0';
        clr_rbl_reg_i     <= '0';
        mb_fifo_rd        <= '0';
        link_fifo_rd      <= '0';
        link_fifo_wr      <= '0';
        clr_hit_count_rst <= '0';
        hit_count_rst_out <= (others => '0');
        if (frm_bus_strb_in = '1') then 
          mb_fifo_rd   <= en_mb_fifo_rd_i;
          link_fifo_rd <= en_link_fifo_rd_i;
          acknowledge_cycle_state <= acknowledge;
        end if;
        
      when acknowledge => 
        frm_bus_ack_out <= '1';
        mb_fifo_rd      <= '0';
        link_fifo_rd    <= '0';
        link_fifo_wr    <= '0';
        if (frm_bus_strb_in = '0') then 
          acknowledge_cycle_state <= idle;
          link_fifo_wr   <= en_link_fifo_wr_i;
          clr_to_reg_i   <= clr_to_status_i;
          clr_dovr_reg_i <= clr_dovr_status_i;
          clr_nnr_reg_i  <= clr_nnr_status_i;
          clr_htl_reg_i  <= clr_htl_status_i;
          clr_rbl_reg_i  <= clr_rbl_status_i;
          if (hit_count_rst_o /= "000000000000") then
            hit_count_rst_out <= hit_count_rst_o;
            clr_hit_count_rst <= '1';
	      end if;
        end if;

      when others =>
        frm_bus_ack_out <= '0';
        mb_fifo_rd      <= '0';
        link_fifo_rd    <= '0';
        link_fifo_wr    <= '0';
        acknowledge_cycle_state <= idle;
    end case;
  end if;  
end process access_acknowledge;     

end rtl; -- code of register_block
