--------------------------------------------------------------------------
--
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 2000
-- ReadOutDriver Electronics
--
--------------------------------------------------------------------------
--
-- Filename: 1024x16fifo.vhd                  
-- Description:
--  Input registered and Output registered fifo.     
--
--------------------------------------------------------------------------
-- Structure: 
--  Multiple state machines in a flat topology/hierarchy which communicate
--   with single wire signals.  
--------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
--------------------------------------------------------------------------
-- Author: Mark L. Nagel
-- Board Engineer: John M. Joseph
-- History:
--    23 February 2000 - MLN first version : requirements/functionality finally
--                       understood and sufficiently documented to proceed 
--                       with coding.
--    13 March 2000 - MLN : 1536 word dynamic mask fifo not required.
--                         Must update to a 512 word fifo.
--
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- LIBRARY INCLUDES
--------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations

--------------------------------------------------------------------------
-- PORT DECLARATION
--------------------------------------------------------------------------
entity fifo_1024x16 is
  port(
    clk_in          : in    std_logic; -- clk40 input
    rst_n_in        : in    std_logic; -- negative logic asynchronous global reset
    flush_fifo_in   : in    std_logic; --
    wren_in         : in    std_logic; -- enable input register
    rden_in         : in    std_logic; -- enable output register
    data_in         : in    std_logic_vector(15 downto 0); --
    data_out        : out   std_logic_vector(15 downto 0); --
    empty_flag_out  : out   std_logic; --
    full_flag_out   : out   std_logic; --
    occupancy_count : out   std_logic_vector(9 downto 0)
    );
end fifo_1024x16;

architecture rtl of fifo_1024x16 is
--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------

signal rst_in_i : std_logic;
signal vcc_i    : std_logic;
signal gnd_i    : std_logic;

signal ingoing_data_i    : std_logic_vector(15 downto 0);
signal outcoming_data_i  : std_logic_vector(15 downto 0);
signal write_en_i        : std_logic;
signal writein_adr_i     : unsigned(9 downto 0);
signal readout_adr_i     : unsigned(9 downto 0);
signal occupancy_count_i : unsigned(9 downto 0);

--------------------------------------------------------------------------
-- COMPONENT DECLARATION
--------------------------------------------------------------------------

component dpram_1024x4
  port(
    clkA_in               : in  std_logic; -- clk40 input
    clkB_in               : in  std_logic; -- 
    rst_in                : in  std_logic; -- 
    dprmA_writein_bus_in  : in  std_logic_vector(3 downto 0);
    dprmB_writein_bus_in  : in  std_logic_vector(3 downto 0);
    writein_adr_in        : in  unsigned(9 downto 0); -- 
    readout_adr_in        : in  unsigned(9 downto 0); --
    dprmA_write_strobe_in : in  std_logic; --
    dprmB_write_strobe_in : in  std_logic; --
    dprmA_read_strobe_in  : in  std_logic; --
    dprmB_read_strobe_in  : in  std_logic; --
    dprmA_readout_bus_out : out std_logic_vector(3 downto 0);
    dprmB_readout_bus_out : out std_logic_vector(3 downto 0)
    );
end component;

--------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
--------------------------------------------------------------------------
begin

rst_in_i <= NOT rst_n_in;
vcc_i    <= '1';
gnd_i    <= '0';

occupancy_count <= std_logic_vector(occupancy_count_i);

MB0 : dpram_1024x4
  port map(
    clkA_in               => clk_in,
    clkB_in               => clk_in,
    rst_in                => rst_in_i,
    dprmA_writein_bus_in  => ingoing_data_i(3 downto 0),
    dprmB_writein_bus_in  => ingoing_data_i(3 downto 0),
    writein_adr_in        => writein_adr_i,
    readout_adr_in        => readout_adr_i,
    dprmA_write_strobe_in => wren_in,
    dprmB_write_strobe_in => gnd_i,
    dprmA_read_strobe_in  => vcc_i,
    dprmB_read_strobe_in  => vcc_i,
    dprmA_readout_bus_out => open,
    dprmB_readout_bus_out => outcoming_data_i(3 downto 0)
    );

MB1 : dpram_1024x4
  port map(
    clkA_in               => clk_in,
    clkB_in               => clk_in,
    rst_in                => rst_in_i,
    dprmA_writein_bus_in  => ingoing_data_i(7 downto 4),
    dprmB_writein_bus_in  => ingoing_data_i(7 downto 4),
    writein_adr_in        => writein_adr_i,
    readout_adr_in        => readout_adr_i,
    dprmA_write_strobe_in => wren_in,
    dprmB_write_strobe_in => gnd_i,
    dprmA_read_strobe_in  => vcc_i,
    dprmB_read_strobe_in  => vcc_i,
    dprmA_readout_bus_out => open,
    dprmB_readout_bus_out => outcoming_data_i(7 downto 4)
    );

MB2 : dpram_1024x4
  port map(
    clkA_in               => clk_in,
    clkB_in               => clk_in,
    rst_in                => rst_in_i,
    dprmA_writein_bus_in  => ingoing_data_i(11 downto 8),
    dprmB_writein_bus_in  => ingoing_data_i(11 downto 8),
    writein_adr_in        => writein_adr_i,
    readout_adr_in        => readout_adr_i,
    dprmA_write_strobe_in => wren_in,
    dprmB_write_strobe_in => gnd_i,
    dprmA_read_strobe_in  => vcc_i,
    dprmB_read_strobe_in  => vcc_i,
    dprmA_readout_bus_out => open,
    dprmB_readout_bus_out => outcoming_data_i(11 downto 8)
    );

MB3 : dpram_1024x4
  port map(
    clkA_in               => clk_in,
    clkB_in               => clk_in,
    rst_in                => rst_in_i,
    dprmA_writein_bus_in  => ingoing_data_i(15 downto 12),
    dprmB_writein_bus_in  => ingoing_data_i(15 downto 12),
    writein_adr_in        => writein_adr_i,
    readout_adr_in        => readout_adr_i,
    dprmA_write_strobe_in => wren_in,
    dprmB_write_strobe_in => gnd_i,
    dprmA_read_strobe_in  => vcc_i,
    dprmB_read_strobe_in  => vcc_i,
    dprmA_readout_bus_out => open,
    dprmB_readout_bus_out => outcoming_data_i(15 downto 12)
    );


--------------------------------------------------------------------------
-- PROCESS DECLARATION
--------------------------------------------------------------------------

-- input  address counter
input_counter : process (
  rst_n_in, 
  clk_in, 
  flush_fifo_in
  )
begin
  if (rst_n_in = '0') then
    writein_adr_i  <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
    if (flush_fifo_in = '1') then
      writein_adr_i  <= (others => '0'); 
    elsif (write_en_i = '1' AND occupancy_count_i < 1023) then
      writein_adr_i <= writein_adr_i + 1;
    end if;
  end if;
end process input_counter;

-- output  address counter
output_counter : process (
  rst_n_in, 
  clk_in, 
  flush_fifo_in
  )
begin
  if (rst_n_in = '0') then
    readout_adr_i  <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
    if (flush_fifo_in = '1') then
      readout_adr_i  <= (others => '0'); 
    elsif (rden_in = '1' AND occupancy_count_i > 0) then
      readout_adr_i <= readout_adr_i + 1;
    end if;
  end if;
end process output_counter;


-- fifo 16bit word occupancy counter
occupancy_counter : process (
  rst_n_in, 
  clk_in, 
  flush_fifo_in
  )
begin
  if (rst_n_in = '0') then
    occupancy_count_i  <= (others => '0');
    empty_flag_out <= '1';
    full_flag_out  <= '0';
  elsif (clk_in'event and clk_in = '1') then

    if (occupancy_count_i = 1023) then 
      full_flag_out <= '1';
    else
      full_flag_out <= '0'; 
    end if;

    if (occupancy_count_i = 0   ) then 
      empty_flag_out <= '1';
    else
      empty_flag_out <= '0';
    end if;
    
    if (flush_fifo_in = '1') then
      occupancy_count_i  <= (others => '0');
    else
      occupancy_count_i <= writein_adr_i - readout_adr_i;
    end if;
  end if;
end process occupancy_counter;


-- delay wren to ram block by 1 clk
-- delay rden to ram block by 1 clk
-- read/write enable delay register
enable_delay : process (
  rst_n_in, 
  clk_in,
  wren_in
  )
begin
  if (rst_n_in = '0') then
    write_en_i <= '0';
  elsif (clk_in'event and clk_in = '1') then
    write_en_i <= wren_in;
  end if;
end process enable_delay;


-- input  register enabled by delayed wren
input_register : process (
  rst_n_in, 
  clk_in, 
  data_in
  )
begin
  if (rst_n_in = '0') then
    ingoing_data_i <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
    ingoing_data_i <= data_in;
  end if;
end process input_register;

-- memory output   enabled by delayed rden
output_register : process (
  rst_n_in, 
  clk_in
  )
begin
  if (rst_n_in = '0') then
    data_out <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
    data_out <= outcoming_data_i;
  end if;
end process output_register;

end rtl; -- end code for fotxstfifo 
