--------------------------------------------------------------------------
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 1999
-- ReadOutDriver Electronics
--  
--------------------------------------------------------------------------
-- Filename: 1024x4dpramblock.vhd
-- Description:
--  FifoBuffer for serial link data stream decoder. 
--  Current implementation is XILINX specific and uses built in 
--  block rams for fifo memories. 
--
--  This block is completely Xilinx specific code,
--  and comes directly from the "xapp130.pdf" file available from 
--  the Xilinx appnote web page.
--  Although there is copious documentation there, it is not all very 
--  clear. Specifically, there are no timing diagrams or examples of
--  how the memories need to be accessed.
--  
--------------------------------------------------------------------------
-- Structure:
--------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk40
--    *The timing structure of all engine code will be:
--      1.  Latch inputs on positive edge of clk40
--      2.  Perform all logic operations on positive edge of clk40
--      3.  Perform all memory operations on positive edge of clk40
--      4.  Read out of the RAM blocks at 80MHz (readout__clk)
--------------------------------------------------------------------------
-- Author:         Mark L. Nagel
-- Board Engineer: 
-- History:
--    08/09/99 MLN First version
--    03/08/01 JMJ Modified 256x16 to 512x8
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- LIBRARY INCLUDES
--------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations

--------------------------------------------------------------------------
-- SIMULATION LIBRARY INCLUDES
--------------------------------------------------------------------------
-- pragma translate_off
library UNISIM;
use unisim.all;
-- pragma translate_on

--------------------------------------------------------------------------
-- PORT DECLARATION
--------------------------------------------------------------------------

entity dpram_1024x4 is
  port(
    clkA_in                : in  std_logic; -- clk40 input
    clkB_in                : in  std_logic; -- 80 MHz sbsram clk from the dsp chip
    rst_in                 : in  std_logic; -- asynchronous global reset
    dprmA_writein_bus_in   : in  std_logic_vector(3 downto 0);
    dprmB_writein_bus_in   : in  std_logic_vector(3 downto 0);
    writein_adr_in         : in  unsigned(9 downto 0); -- 
    readout_adr_in         : in  unsigned(9 downto 0); --
    dprmA_write_strobe_in  : in  std_logic; --
    dprmB_write_strobe_in  : in  std_logic; --
    dprmA_read_strobe_in   : in  std_logic; --
    dprmB_read_strobe_in   : in  std_logic; --
    dprmA_readout_bus_out  : out std_logic_vector(3 downto 0);
    dprmB_readout_bus_out  : out std_logic_vector(3 downto 0)
    );
end dpram_1024x4;


architecture behave of dpram_1024x4 is
--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------

signal dprmA_addr_i : std_logic_vector(9 downto 0); -- int write in addr
signal dprmB_addr_i : std_logic_vector(9 downto 0); -- int read out addr

--------------------------------------------------------------------------
-- COMPONENT DECLARATION
--------------------------------------------------------------------------

component RAMB4_S4_S4
  port (
    DIA   : in  STD_LOGIC_VECTOR (3 downto 0);
    DIB   : in  STD_LOGIC_VECTOR (3 downto 0);
    ENA   : in  STD_ULOGIC;
    ENB   : in  STD_ULOGIC;
    WEA   : in  STD_ULOGIC;                
    WEB   : in  STD_ULOGIC;
    RSTA  : in  STD_ULOGIC;
    RSTB  : in  STD_ULOGIC;
    CLKA  : in  STD_ULOGIC;
    CLKB  : in  STD_ULOGIC;
    ADDRA : in  STD_LOGIC_VECTOR (9 downto 0);
    ADDRB : in  STD_LOGIC_VECTOR (9 downto 0);
    DOA   : out STD_LOGIC_VECTOR (3 downto 0);
    DOB   : out STD_LOGIC_VECTOR (3 downto 0)
    ); 
end component;

attribute INIT_00: string;
attribute INIT_01: string;
attribute INIT_02: string;
attribute INIT_03: string;
attribute INIT_04: string;
attribute INIT_05: string;
attribute INIT_06: string;
attribute INIT_07: string;
attribute INIT_08: string;
attribute INIT_09: string;
attribute INIT_0A: string;
attribute INIT_0B: string;
attribute INIT_0C: string;
attribute INIT_0D: string;
attribute INIT_0E: string;
attribute INIT_0F: string;


attribute INIT_00 of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_01 of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_02 of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_03 of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_04 of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_05 of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_06 of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_07 of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_08 of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_09 of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0A of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0B of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0C of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0D of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0E of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";
attribute INIT_0F of MB: label is
"0000000000000000000000000000000000000000000000000000000000000000";

--------------------------------------------------------------------------
--  Truth Table for RAMB4_Sn_Sm  (all column valid for A/B)
--
--  --------------------------------------------------------------------
--  |         Inputs                    ||      Outputs                |
--  --------------------------------------------------------------------
--  | EN | RST | WE | CLK | ADDR |  DI  ||    DO     |  RAM Contents   |
--  --------------------------------------------------------------------
--  |  0 |  X  |  X |  X  |  X   |  X   ||   NoChg   |     NoChng      |
--  --------------------------------------------------------------------
--  |  1 |  1  |  0 |  ^  |  X   |  X   ||    0      |     NoChng      |
--  --------------------------------------------------------------------
--  |  1 |  1  |  1 |  ^  | addr | data ||    0      | RAM(addr)<=data |
--  --------------------------------------------------------------------
--  |  1 |  0  |  0 |  ^  | addr |  X   || RAM(addr) |     NoChng      |
--  --------------------------------------------------------------------
--  |  1 |  0  |  1 |  ^  | addr | data ||   data    | RAM(addr)<=data |
--  --------------------------------------------------------------------

--------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
--------------------------------------------------------------------------
begin

dprmA_addr_i <= std_logic_vector(writein_adr_in);
dprmB_addr_i <= std_logic_vector(readout_adr_in);


MB : RAMB4_S4_S4 -- Block[3..0] of 4bits
  port map(
    DIA   => dprmA_writein_bus_in,
    DIB   => dprmB_writein_bus_in,
    ENA   => dprmA_read_strobe_in,
    ENB   => dprmB_read_strobe_in,
    WEA   => dprmA_write_strobe_in,
    WEB   => dprmB_write_strobe_in,
    RSTA  => rst_in,
    RSTB  => rst_in,
    CLKA  => clkA_in,
    CLKB  => clkB_in,
    ADDRA => dprmA_addr_i,
    ADDRB => dprmB_addr_i,
    DOA   => dprmA_readout_bus_out,
    DOB   => dprmB_readout_bus_out
    ); 

end behave; -- Code for dpram_1024x4 
