-------------------------------------------------------------------------------
--
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 2000
-- ReadOutDriver Electronics
--
-------------------------------------------------------------------------------
--
-- Filename: 4096x8fifo.vhd                  
-- Description:
--  Input registered and Output registered fifo.     
--
--  The following state machines are all 1 stage glitch free MOORE MACHINES.
--   as per the LBL template  "vhdl/sm1_sample.vhd" code.
--
--                                                                   
--            +----------+   +---------------+      +-----------+    
-- data_in ---| Input    |---| Dual Port     |------| Output    |--- data_out
--            | Register |   | SRAM          |      | State     |     
--            |          |   |               |      | Logic     |     
--            |          |   |               |      |           |     
--            |          |   |               |      |           |     
--       +---->   en     |--->               |------>   en      |     
--       |    +----------+   |               |      +-----------+     
-- clk --+        |          +---------------+          |
-- wren ----------+            |                        | 
--                |            |      +-----------+     |                   
--                |            +------| Control   |-----+
--                |                   | Logic     |                        
--                +-------------------|           |                       
-- rden ------------------------------|           |                        
--                                    |           |                      
--                                    +-----------+                       
--                                                      
-------------------------------------------------------------------------------
-- Structure: 
--  Multiple state machines in a flat topology/hierarchy which communicate
--   with single wire signals.  
-------------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
-------------------------------------------------------------------------------
--
-- Author: Mark L. Nagel
-- Board Engineer: John M. Joseph
-- History:
--    23 February 2000 - MLN first version : requirements/functionality finally
--                       understood and sufficiently documented to proceed 
--                       with coding.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- LIBRARY INCLUDES
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations

-------------------------------------------------------------------------------
-- PORT DECLARATION
-------------------------------------------------------------------------------
entity fifo_4096x8 is  -- 4096x8 FIFO
  port (
    clk_in          : in    std_logic; -- clk40 input
    rst_n_in        : in    std_logic; -- asynchronous global reset
    flush_fifo_in   : in    std_logic; --
    wren_in         : in    std_logic; -- enable input register
    rden_in         : in    std_logic; -- enable output register
    rten_in         : in    std_logic; -- enable re-transmit
    data_in         : in    std_logic_vector(7 downto 0); --
    data_out        : out   std_logic_vector(7 downto 0); --
    full_flag_out   : out   std_logic; --
    empty_flag_out  : out   std_logic;
    occupancy_count : out   std_logic_vector(11 downto 0)
    );
end fifo_4096x8;

architecture rtl of fifo_4096x8 is
-------------------------------------------------------------------------------
-- SIGNAL DECLARATION
-------------------------------------------------------------------------------

signal rst_i : std_logic;

signal ingoing_data_i    : std_logic_vector(7 downto 0);
signal gogoing_data_i    : std_logic_vector(7 downto 0);
signal outcoming_data_i  : std_logic_vector(7 downto 0);

signal write_en_i        : std_logic;
signal wren_tmp_i        : std_logic;
signal writein_adr_i     : unsigned(11 downto 0);

signal read_en_i         : std_logic;
signal readout_adr_i     : unsigned(11 downto 0);
signal occupancy_count_i : unsigned(11 downto 0);

-------------------------------------------------------------------------------
-- COMPONENT DECLARATION
-------------------------------------------------------------------------------

component dpram_4096x1 -- 4096x1 DPRAM Block
  port (
    clk              : in  std_logic; 
    rst              : in  std_logic; 
    dprm_writein_bus : in  std_logic_vector(0 downto 0);
    writein_adr      : in  unsigned(11 downto 0); --
    readout_adr      : in  unsigned(11 downto 0); --
    write_strobe     : in  std_logic; --
    dprm_readout_bus : out std_logic_vector(0 downto 0)
    );
end component;

-------------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
-------------------------------------------------------------------------------
begin

MBn : for i in 0 to 7 generate 
  MB : dpram_4096x1
    port map (
      clk              => clk_in,
      rst              => rst_i,
      dprm_writein_bus => ingoing_data_i(i downto i),
      writein_adr      => writein_adr_i,
      readout_adr      => readout_adr_i,
      write_strobe     => wren_tmp_i, 
      dprm_readout_bus => outcoming_data_i(i downto i)
      );
end generate;

-------------------------------------------------------------------------------
-- PROCESS DECLARATION
-------------------------------------------------------------------------------

rst_i <= NOT rst_n_in;

wren_tmp_i     <= wren_in;
ingoing_data_i <= data_in;

occupancy_count <= std_logic_vector(occupancy_count_i);


-- input  address counter
input_counter : process (
  rst_n_in, 
  clk_in, 
  flush_fifo_in,
  write_en_i, 
  writein_adr_i, 
  occupancy_count_i
  )
begin
  if (rst_n_in = '0') then
    writein_adr_i  <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
    if (flush_fifo_in = '1') then  -- Set the WriteIn pointer to 0
      writein_adr_i  <= (others => '0'); 
    elsif (write_en_i = '1' AND occupancy_count_i < 4095) then
      writein_adr_i <= writein_adr_i + 1;
    end if;
  end if;
end process input_counter;


-- output  address counter
output_counter : process (
  rst_n_in, 
  clk_in, 
  flush_fifo_in,
  read_en_i, 
  readout_adr_i, 
  occupancy_count_i
  )
begin
  if (rst_n_in = '0') then
    readout_adr_i  <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
    if (flush_fifo_in = '1' OR rten_in = '1') then -- Set the ReadOut pointer to 0
      readout_adr_i  <= (others => '0'); 
    elsif (read_en_i = '1' AND occupancy_count_i > 0) then
      readout_adr_i <= readout_adr_i + 1;
    end if;
  end if;
end process output_counter;


-- fifo 12bit word occupancy counter
occupancy_counter : process (
  rst_n_in, 
  clk_in, 
  flush_fifo_in,
  readout_adr_i, 
  writein_adr_i,
  occupancy_count_i
  )
begin
  if (rst_n_in = '0') then
    occupancy_count_i  <= (others => '0');
    empty_flag_out <= '1';
    full_flag_out  <= '0';
--elsif (clk_in'event and clk_in = '1') then
  else

    if (occupancy_count_i = 4095) then 
      full_flag_out <= '1';
    else
      full_flag_out <= '0'; 
    end if;

    if (occupancy_count_i = 0) then 
      empty_flag_out <= '1';
    else
      empty_flag_out <= '0';
    end if;

    if (flush_fifo_in = '1') then
      occupancy_count_i  <= (others => '0');
    else
      occupancy_count_i <= writein_adr_i - readout_adr_i;
    end if;
  end if;
end process occupancy_counter;


-- delay wren to ram block by 1 clk
-- delay rden to ram block by 1 clk
-- read/write enable delay register
enable_delay : process (
  rst_n_in, 
  clk_in, 
  rden_in, 
  wren_in
  )
begin
  if (rst_n_in = '0') then
    read_en_i  <= '0';
    write_en_i <= '0';
  elsif (clk_in'event and clk_in = '1') then
    read_en_i  <= rden_in;--
    write_en_i <= wren_in;--
  end if;
end process enable_delay;


signal_register : process (
  rst_n_in, 
  clk_in, 
  outcoming_data_i, 
  rden_in
  )
begin
  if (rst_n_in = '0') then
    data_out <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
--  if (rden_in = '1') then  -- Latch Data out of the FIFO
      data_out <= outcoming_data_i;
--  end if;
  end if;
end process signal_register;

end rtl; -- end code for fifo_4096x1
