--------------------------------------------------------------------------
--
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 2000
-- ReadOutDriver Electronics
--------------------------------------------------------------------------
--
-- Filename: fe_short_command_generator.vhd
-- Description:
--  In as compact a form as possible, implements the 
-- SCT Command bit streams to front end modules.
--  l1accept:  "110" 
--  bcr:       "1010010" 
--  ecr:       "1010100"
-- Pixel Command bit streams to front end modules.
--  l1accept:  "11101" 
--  bcr:       "101100001" 
--  ecr:       "101100010"
--  cal:       "101100100"
--  sync:      "101101000"
--
--------------------------------------------------------------------------
-- Structure: 
--  Multiple state machines in a flat topology/hierarchy which communicate
--   with single wire signals.  
--------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
--------------------------------------------------------------------------
-- Author: Mark L. Nagel
-- Board Engineer: John M. Joseph
-- History:
--    23 February 2000 - MLN first version : requirements/functionality finally
--                       understood and sufficiently documented to proceed 
--                       with coding.
--     6 January 2003 - JMJ Added Pixel Commands
--
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- LIBRARY INCLUDES
--------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations

--------------------------------------------------------------------------
-- PORT DECLARATION
--------------------------------------------------------------------------
entity fe_short_command_generator is
  port (
    clk_in            : in  std_logic; -- clk40 input
    rst_n_in          : in  std_logic; -- negative logic asynchronous global reset
    rod_type_in       : in  std_logic; -- ROD Type 0=>SCT 1 =>Pixel
    l1accept_in       : in  std_logic; -- l1trigger accept pulse, 1 clk wide. 
    bcr_in            : in  std_logic; -- bunch counter reset                 
    ecr_in            : in  std_logic; -- event counter reset 
    ecr_flush_en_in   : in  std_logic_vector( 1 downto 0);
    command_bits_out  : out std_logic_vector(47 downto 0);  -- 
    cmd_proc_busy_out : out std_logic
    );
end fe_short_command_generator;

architecture rtl of fe_short_command_generator is
--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------

type states is (
  reset, 
  idle,
  l1a, 
  bcr, 
  ecr
  );
signal cmd_state : states;

signal cmd_bit_count : unsigned(4 downto 0);

--------------------------------------------------------------------------
-- COMPONENT DECLARATION
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
--------------------------------------------------------------------------
begin

--------------------------------------------------------------------------
-- PROCESS DECLARATION
--------------------------------------------------------------------------
-- SCT Command bit streams to front end modules.
--  l1accept:  "110" 
--  bcr:       "1010010" 
--  ecr:       "1010100"
-- Pixel Command bit streams to front end modules.
--  l1accept:  "11101" 
--  bcr:       "101100001" 
--  ecr:       "101100010"
--  cal:       "101100100"
--  sync:      "101101000"

fe_cmd_out : process (
  clk_in, 
  rst_n_in, 
  rod_type_in,
  l1accept_in, 
  bcr_in, 
  ecr_in
  )
begin
  if (rst_n_in = '0') then
    cmd_proc_busy_out <= '0';   
    command_bits_out  <= (others => '0'); 
    cmd_bit_count     <= (others => '0');
    cmd_state         <= reset;
  elsif (clk_in'event and clk_in = '1') then
    case cmd_state is
      when reset =>
        cmd_proc_busy_out <= '0';   
        command_bits_out  <= (others => '0');                 
        cmd_bit_count     <= (others => '0');                
        cmd_state         <= idle; 

-- Choose which command bitstream will be sent
      when idle =>
        cmd_proc_busy_out <= '0';   
        command_bits_out  <= (others => '0');                 
        cmd_bit_count     <= (others => '0');                
        if (l1accept_in = '1') then 
          cmd_proc_busy_out <= '1';   
          command_bits_out  <= (others => '1');                 
          cmd_state         <= l1a; 
        elsif (bcr_in = '1')      then
          cmd_proc_busy_out <= '1';   
          command_bits_out  <= (others => '1');                 
          cmd_state         <= bcr;
        elsif (ecr_in = '1')      then
          cmd_proc_busy_out <= '1';   
          command_bits_out  <= (others => '1');                 
          cmd_state         <= ecr;
        end if; 

      when l1a =>
        cmd_bit_count <= cmd_bit_count + '1';
        if (rod_type_in = '0') then --  l1accept: "110" 
          case cmd_bit_count(2 downto 0) is
--                      => command_bits_out  <= (others => '1');                 
            when "000"  => command_bits_out  <= (others => '1');                 
            when "001"  => command_bits_out  <= (others => '0');                 
                           cmd_proc_busy_out <= '0';   
                           cmd_state         <= idle; 
            when others => command_bits_out  <= (others => '0');
                           cmd_proc_busy_out <= '0';   
                           cmd_state         <= idle; 
          end case;
        else  --  l1accept: "11101" 
          case cmd_bit_count(2 downto 0) is
--                      => command_bits_out  <= (others => '1');                 
            when "000"  => command_bits_out  <= (others => '1');                 
            when "001"  => command_bits_out  <= (others => '1');                 
            when "010"  => command_bits_out  <= (others => '0');                 
            when "011"  => command_bits_out  <= (others => '1');                 
                           cmd_proc_busy_out <= '0';   
                           cmd_state         <= idle; 
            when others => command_bits_out  <= (others => '0');
                           cmd_proc_busy_out <= '0';   
                           cmd_state         <= idle; 
          end case;
        end if;

      when bcr =>
        cmd_bit_count <= cmd_bit_count + '1';
        if (rod_type_in = '0') then   --  bcr: "1010010" 
          case cmd_bit_count(2 downto 0) is
--                      => command_bits_out  <= (others => '1');                 
            when "000"  => command_bits_out  <= (others => '0');                 
            when "001"  => command_bits_out  <= (others => '1');                 
            when "010"  => command_bits_out  <= (others => '0');                 
            when "011"  => command_bits_out  <= (others => '0');                 
            when "100"  => command_bits_out  <= (others => '1');                 
            when "101"  => command_bits_out  <= (others => '0');                 
                           cmd_proc_busy_out <= '0';   
                           cmd_state         <= idle; 
            when others => command_bits_out  <= (others => '0');
                           cmd_proc_busy_out <= '0';   
                           cmd_state         <= idle; 
          end case;
        else --  bcr: "101100001" 
          case cmd_bit_count(2 downto 0) is
--                      => command_bits_out  <= (others => '1');                 
            when "000"  => command_bits_out  <= (others => '0');                 
            when "001"  => command_bits_out  <= (others => '1');                 
            when "010"  => command_bits_out  <= (others => '1');                 
            when "011"  => command_bits_out  <= (others => '0');                 
            when "100"  => command_bits_out  <= (others => '0');                 
            when "101"  => command_bits_out  <= (others => '0');                 
            when "110"  => command_bits_out  <= (others => '0');                 
            when "111"  => command_bits_out  <= (others => '1');                 
                           cmd_proc_busy_out <= '0';   
                           cmd_state         <= idle; 
            when others => command_bits_out  <= (others => '0');
                           cmd_proc_busy_out <= '0';   
                           cmd_state         <= idle; 
          end case;
        end if;

      when ecr =>
        cmd_bit_count <= cmd_bit_count + '1';
        if (rod_type_in = '0') then   --  ecr: "1010100"
          case cmd_bit_count(2 downto 0) is
--                      => command_bits_out  <= (others => '1');                 
            when "000"  => command_bits_out  <= (others => '0');                 
            when "001"  => command_bits_out  <= (others => '1');                 
            when "010"  => command_bits_out  <= (others => '0');                 
            when "011"  => command_bits_out  <= (others => '1');                 
            when "100"  => command_bits_out  <= (others => '0');                 
            when "101"  => command_bits_out  <= (others => '0');                 
                           cmd_proc_busy_out <= '0';   
                           cmd_state         <= idle; 
            when others => command_bits_out  <= (others => '0');
                           cmd_proc_busy_out <= '0';   
                           cmd_state         <= idle; 
          end case;
        else  --  ecr: "101100010"
--          if (ecr_flush_en_in(1) = '0' OR ecr_flush_en_in = "10") then  --Send ECR Only
          if (ecr_flush_en_in(0) = '0') then  --Send ECR Only
            case cmd_bit_count(2 downto 0) is
  --                      => command_bits_out  <= (others => '1');                 
              when "000"  => command_bits_out  <= (others => '0');                 
              when "001"  => command_bits_out  <= (others => '1');                 
              when "010"  => command_bits_out  <= (others => '1');                 
              when "011"  => command_bits_out  <= (others => '0');                 
              when "100"  => command_bits_out  <= (others => '0');                 
              when "101"  => command_bits_out  <= (others => '0');                 
              when "110"  => command_bits_out  <= (others => '1');                 
              when others => command_bits_out  <= (others => '0');
                             cmd_proc_busy_out <= '0';   
                             cmd_state         <= idle; 
            end case;
          else  -- Send Sync FE then ECR
            case cmd_bit_count is
  --  SyncFE: "101101000"
  --                        => command_bits_out  <= (others => '1');                 
              when "00000"  => command_bits_out  <= (others => '0');                 
              when "00001"  => command_bits_out  <= (others => '1');                 
              when "00010"  => command_bits_out  <= (others => '1');                 
              when "00011"  => command_bits_out  <= (others => '0');                 
              when "00100"  => command_bits_out  <= (others => '1');                 
              when "00101"  => command_bits_out  <= (others => '0');                 
              when "00110"  => command_bits_out  <= (others => '0');                 
              when "00111"  => command_bits_out  <= (others => '0');                 
  -- Wait before sending ECR
              when "01000"  => command_bits_out  <= (others => '0');                 
              when "01001"  => command_bits_out  <= (others => '0');                 
              when "01010"  => command_bits_out  <= (others => '0');                 
              when "01011"  => command_bits_out  <= (others => '0');                 
              when "01100"  => command_bits_out  <= (others => '0');                 
              when "01101"  => command_bits_out  <= (others => '0');                 
              when "01110"  => command_bits_out  <= (others => '0');                 
              when "01111"  => command_bits_out  <= (others => '0');                 
              when "10000"  => command_bits_out  <= (others => '0');                 
              when "10001"  => command_bits_out  <= (others => '0');                 
              when "10010"  => command_bits_out  <= (others => '0');                 
              when "10011"  => command_bits_out  <= (others => '0');                 
              when "10100"  => command_bits_out  <= (others => '0');                 
  --  ECR: "101100010"
              when "10101"  => command_bits_out  <= (others => '1');                 
              when "10110"  => command_bits_out  <= (others => '0');                 
              when "10111"  => command_bits_out  <= (others => '1');                 
              when "11000"  => command_bits_out  <= (others => '1');                 
              when "11001"  => command_bits_out  <= (others => '0');                 
              when "11010"  => command_bits_out  <= (others => '0');                 
              when "11011"  => command_bits_out  <= (others => '0');                 
              when "11100"  => command_bits_out  <= (others => '1');                 
              when "11101"  => command_bits_out  <= (others => '0');                 
              when others   => command_bits_out  <= (others => '0');
                               cmd_proc_busy_out <= '0';   
                               cmd_state         <= idle; 
            end case;
          end if;
        end if;
      when others =>
        cmd_proc_busy_out <= '0';   
        command_bits_out  <= (others => '0');                 
        cmd_bit_count     <= (others => '0');                
        cmd_state         <= idle;
    end case;
  end if;  
end process fe_cmd_out;

end rtl; -- end code for fe_short_command_generator
