-------------------------------------------------------------------------------
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 1999
-- ReadOutDriver Electronics
--
-------------------------------------------------------------------------------
-- Filename: register_block.vhd
-- Description:
--   Register Block used inthe RCF FPGA.  See Users Manual Appendix A for
--   detaled definitions of each register
--------------------------------------------------------------------------
-- Structure:
--
--------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk40
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
--------------------------------------------------------------------------
-- Author: John M. Joseph
-- Board Engineer: John M. Joseph
-- History:
--          12 Feb 2002 : JMJ Map Update to clean up code
--------------------------------------------------------------------------
--
--------------------------------------------------------------------------
-- LIBRARY INCLUDES
--------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations

--------------------------------------------------------------------------
-- PORT DECLARATION
--------------------------------------------------------------------------

entity register_block is
  port (
    clk_in                    : in  std_logic; -- clk40 input
    rst_n_in                  : in  std_logic; -- global reset

    regbus_en_n_in            : in  std_logic;
    regbus_rnw_in             : in  std_logic;
    regbus_a_in               : in  std_logic_vector(11 downto 0);
    regbus_d_in               : in  std_logic_vector(31 downto 0);
    regbus_d_out              : out std_logic_vector(31 downto 0);
    regbus_ack_out            : out std_logic;   

    fcmbus_en_n_in            : in  std_logic;
    fcmbus_d_out              : out std_logic_vector(31 downto 0);

    rrif_ctrl_reg_out         : out std_logic_vector(31 downto 0);
    spare_ctrl_reg_out        : out std_logic_vector(31 downto 0);
    rrif_stts_reg_in          : in  std_logic_vector(31 downto 0);
    spare_stts_reg_in         : in  std_logic_vector(31 downto 0);
    dsp_stts_reg_in           : in  std_logic_vector(31 downto 0);
    test_stts_reg_in          : in  std_logic_vector(31 downto 0);
    pin_hold_reg_in           : in  std_logic_vector(31 downto 0);

    fecmd_link_mask0_out      : out std_logic_vector(47 downto 0);
    fecmd_link_mask1_out      : out std_logic_vector(47 downto 0);

    calstrobe_delay_reg_out   : out std_logic_vector( 7 downto 0);
    cal_command_reg_out       : out std_logic_vector(26 downto 0);

    modebits_fifo_stts_reg_in : in  std_logic_vector(23 downto 0);
    dynmmask_fifo_stts_reg_in : in  std_logic_vector(41 downto 0);
    inmem_wordcount_reg_out   : out std_logic_vector(31 downto 0);
    dbgmem_wordcount_reg_out  : out std_logic_vector(31 downto 0);
    cnfg_rdback_count_reg_out : out std_logic_vector(31 downto 0);
    ide_debugmem_stts_reg_in  : in  std_logic_vector( 7 downto 0);
    ide_debugmem_cnfg_reg_out : out std_logic_vector(17 downto 0);
    debug_fifo_stts_in        : in  std_logic_vector( 7 downto 0);
    rol_test_block_reg_out    : out std_logic_vector( 3 downto 0);  
    rol_test_block_count_out  : out std_logic_vector( 5 downto 0);  

    int_tim_fifo_outdata_out  : out std_logic_vector( 7 downto 0);
    int_tim_fifo_re_in        : in  std_logic;
    int_tim_fifo_rt_in        : in  std_logic;

    slave_dsp_ints_out        : out std_logic_vector( 3 downto 0);
    slave_dsp_ints_in         : in  std_logic_vector( 7 downto 0);
    sdsp_inttohost_signal_out : out std_logic;
     -- or'd interrupt of the sdsp's to one of the RCDSP's ints

    feocc_reset_reg_out       : out std_logic_vector( 95 downto 0);
    feocc_num_l1_reg_out      : out std_logic_vector(103 downto 0);

    cal_event_type0_out       : out std_logic_vector( 9 downto 0);
    cal_event_type1_out       : out std_logic_vector( 9 downto 0);
    cal_l1id0_in              : in  std_logic_vector(23 downto 0);
    cal_l1id1_in              : in  std_logic_vector(23 downto 0);
    static_l1id0_out          : out std_logic_vector(23 downto 0);
    static_l1id1_out          : out std_logic_vector(23 downto 0);
    cal_bcid0_in              : in  std_logic_vector(11 downto 0);
    cal_bcid1_in              : in  std_logic_vector(11 downto 0);
    static_bcid0_out          : out std_logic_vector(11 downto 0);
    static_bcid1_out          : out std_logic_vector(11 downto 0);
    rod_mode_out              : out std_logic_vector( 3 downto 0);
    test_mux_out              : out std_logic_vector( 3 downto 0);
    fe_cmd_mask_sel_out       : out std_logic_vector( 2 downto 0);
    fe_cmd_mask_updated_in    : in  std_logic;
    corrective_mask_sent_in   : in  std_logic;
    inc_hrdy_toc_in           : in  std_logic;
    send_int_test_out         : out std_logic;
    test_irq_out              : out std_logic_vector( 2 downto 0);
    ecr_count_in              : in  std_logic_vector( 7 downto 0);
    ecr_count_out             : out std_logic_vector( 7 downto 0);
    nevent_out                : out std_logic_vector(15 downto 0);  -- NEvent
    precal_delay_out          : out std_logic_vector( 7 downto 0);  -- Pre CAL Delay
    trig_delay0_out           : out std_logic_vector(15 downto 0);  -- TrigDelay1
    trig_delay1_out           : out std_logic_vector(15 downto 0);  -- TrigDelay2
    interval_out              : out std_logic_vector(31 downto 0);  -- Interval
    trig_mode_out             : out std_logic_vector(31 downto 0);  -- TrigMode
    mask_en_in                : in  std_logic_vector( 2 downto 0);  --
    new_mask_in               : in  std_logic;
    load_event_type_out       : out std_logic;
    int_nevent_done_in        : in  std_logic;
    l1id_bits_in              : in  std_logic_vector(23 downto 0);
    fftv_cmnd_out             : out std_logic_vector( 7 downto 0);
    fftv_busy_stat_in         : in  std_logic_vector(28 downto 0);
    fftv_ID_stat_in           : in  std_logic_vector( 8 downto 0);
    fftv_match_level_in       : in  std_logic_vector( 3 downto 0);
    set_rod_busy_out          : out std_logic;
    ecr_busy_count_out        : out std_logic_vector(15 downto 0);
    dropped_l1a_count         : in  unsigned(31 downto 0);
    dropped_ecr_count         : in  unsigned(31 downto 0);
    dropped_bcr_count         : in  unsigned(31 downto 0)
    );   
end register_block;

architecture rtl of register_block is
--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------

type states is (
  reset,
  wait_0_clk,
  wait_1_clk,
  acknowledge
  );
signal ackn_cycle_state : states;               

type mask_fetch_states is (
  idle,
  fetch_mask_data,
  done
  );
signal mask_fetch : mask_fetch_states;               

signal rst_i                   : std_logic;
signal rrif_ctrl_reg_i         : std_logic_vector(31 downto 0);
signal spare_ctrl_reg_i        : std_logic_vector(31 downto 0);
signal rod_mode_i              : std_logic_vector( 3 downto 0);
signal test_mux_i              : std_logic_vector( 3 downto 0);
signal fecmd_link_mask0_i      : std_logic_vector(47 downto 0);
signal fecmd_link_mask1_i      : std_logic_vector(47 downto 0);
signal calstrobe_delay_reg_i   : std_logic_vector( 7 downto 0);
signal cal_command_reg_i       : std_logic_vector(26 downto 0);
signal inmem_wordcount_reg_i   : std_logic_vector(31 downto 0);
signal dbgmem_wordcount_reg_i  : std_logic_vector(31 downto 0);
signal cnfg_rdback_count_reg_i : std_logic_vector(31 downto 0);
signal ide_debugmem_cnfg_reg_i : std_logic_vector(17 downto 0);
signal rol_test_block_reg_i    : std_logic_vector( 3 downto 0);
signal rol_test_block_count_i  : std_logic_vector( 5 downto 0);  
signal sdsp_inttohost_signal_i : std_logic;
signal feocc_reset_i           : std_logic_vector( 95 downto 0);
signal feocc_num_l1_i          : std_logic_vector(103 downto 0);
signal ints_to_sdsps_i         : std_logic_vector(  3 downto 0);
signal cal_event_type0_i       : std_logic_vector(  9 downto 0);
signal cal_event_type1_i       : std_logic_vector(  9 downto 0);
signal static_bcid0_i          : std_logic_vector( 11 downto 0);
signal static_bcid1_i          : std_logic_vector( 11 downto 0);
signal static_l1id0_i          : std_logic_vector( 23 downto 0);
signal static_l1id1_i          : std_logic_vector( 23 downto 0);
signal fe_cmd_mask_count_i     : unsigned( 2 downto 0);
signal fe_cmd_mask_addr_i      : unsigned( 7 downto 0);
signal fe_cmd_mask_data_i      : std_logic_vector(31 downto 0);
signal command_mask_select_i   : std_logic_vector( 2 downto 0);
signal command_mask_ready_i    : std_logic;
signal fetch_fe_mask_i         : std_logic;     
 
signal lut_wr_addr_i           : unsigned(7 downto 0);
signal lut_wr_i                : std_logic;

signal int_tim_fifo_we_i      : std_logic; --
signal int_tim_fifo_wen_i     : std_logic; --
signal int_tim_fifo_re_i      : std_logic; --
signal int_tim_fifo_ren_i     : std_logic; --
signal int_tim_fifo_rt_i      : std_logic; --
signal int_tim_fifo_en_i      : std_logic; --
signal int_tim_fifo_ef_i      : std_logic; --
signal int_tim_fifo_ff_i      : std_logic; --
signal int_tim_fifo_rst_i     : std_logic; --
signal int_tim_fifo_indata_i  : std_logic_vector( 7 downto 0); --
signal int_tim_fifo_outdata_i : std_logic_vector( 7 downto 0); --
signal int_tim_occ_count_i    : std_logic_vector(11 downto 0); --

signal wr_strb_i        : std_logic; --
signal rd_strb_i        : std_logic; --
signal regbus_en_n_in_i : std_logic; -- signal used to detect edges 
                                     -- in regbus_en_n_in
signal hrdy_toc_i       : unsigned(31 downto 0);
signal clr_hrdy_toc_i   : std_logic; --
signal clr_hrdy_toc_ii  : std_logic; --
signal send_int_test_i  : std_logic;
signal test_irq_i       : std_logic_vector( 2 downto 0);
signal ecr_count_i      : std_logic_vector( 7 downto 0);
signal ecr_busy_count_i : std_logic_vector(15 downto 0);

signal nevent_i      : std_logic_vector(15 downto 0);  -- NEvent
signal precal_td_i   : std_logic_vector( 7 downto 0);  -- Pre CAL Delay
signal trig_delay0_i : std_logic_vector(15 downto 0);  -- TrigDelay1
signal trig_delay1_i : std_logic_vector(15 downto 0);  -- TrigDelay2
signal interval_i    : std_logic_vector(31 downto 0);  -- Interval
signal trig_mode_i   : std_logic_vector(31 downto 0);  -- TrigMode
signal mdsp_gpreg_i  : std_logic_vector(31 downto 0);

type rb_access_states is (
  idle,
  write,
  read,
  done
  );
signal rb_access : rb_access_states;

signal fftv_cmnd_i     : std_logic_vector( 7 downto 0);
signal pin_hold_reg_i  : std_logic_vector(11 downto 0);
signal set_rod_busy_i  : std_logic;

--------------------------------------------------------------------------
-- COMPONENT DECLARATION
--------------------------------------------------------------------------
                      
component fifo_4096x8
  port (
    clk_in          : in    std_logic; -- clk40 input
    rst_n_in        : in    std_logic; -- asynchronous global reset
    flush_fifo_in   : in    std_logic; --
    wren_in         : in    std_logic; -- enable input register
    rden_in         : in    std_logic; -- enable output register
    rten_in         : in    std_logic; -- retransmit data
    data_in         : in    std_logic_vector(7 downto 0); --
    data_out        : out   std_logic_vector(7 downto 0); --
    full_flag_out   : out   std_logic; --
    empty_flag_out  : out   std_logic;  --
    occupancy_count : out   std_logic_vector(11 downto 0)
    );
end component;

component dpram_256x16
  port (
    clk              : in  std_logic; -- clk40 input
    rst              : in  std_logic; -- asynchronous global reset
    dprm_writein_bus : in  std_logic_vector(15 downto 0); -- data from the DataRouterManager
    dprm_readiin_bus : out std_logic_vector(15 downto 0); 
    writein_adr      : in  unsigned(7 downto  0); -- 
    readout_adr      : in  unsigned(7 downto  0); --
    write_strobe     : in  std_logic; --
    read_strobe      : in  std_logic; --
    dprm_readout_bus : out std_logic_vector(15 downto 0)  -- data to the output tbuf 
    );
end component;

begin
--------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
--------------------------------------------------------------------------
                              
int_tim_fifo : fifo_4096x8 
  port map (
    clk_in          => clk_in,
    rst_n_in        => rst_n_in,
    flush_fifo_in   => int_tim_fifo_rst_i,
    wren_in         => int_tim_fifo_we_i,
    rden_in         => int_tim_fifo_re_i,
    rten_in         => int_tim_fifo_rt_i,
    data_in         => int_tim_fifo_indata_i,
    data_out        => int_tim_fifo_outdata_i,
    full_flag_out   => int_tim_fifo_ff_i,
    empty_flag_out  => int_tim_fifo_ef_i,
    occupancy_count => int_tim_occ_count_i
    );
                               
int_tim_fifo_re_i  <= int_tim_fifo_re_in OR 
                     (int_tim_fifo_ren_i AND int_tim_fifo_en_i);
int_tim_fifo_we_i  <= int_tim_fifo_wen_i AND int_tim_fifo_en_i;     
int_tim_fifo_rt_i  <= int_tim_fifo_rt_in OR spare_ctrl_reg_i(8);
int_tim_fifo_rst_i <= spare_ctrl_reg_i(6);


FECMD_MBn : for i in 0 to 1 generate 
  fe_cmd_mask_lut : dpram_256x16
    port map (
      clk              => clk_in,
      rst              => rst_i,
      dprm_writein_bus => regbus_d_in((16*i)+15 downto 16*i),
      dprm_readiin_bus => fcmbus_d_out((16*i)+15 downto 16*i),
      writein_adr      => lut_wr_addr_i,
      readout_adr      => fe_cmd_mask_addr_i,
      write_strobe     => lut_wr_i,
      read_strobe      => rst_n_in,  -- (READ Always Active) 
      dprm_readout_bus => fe_cmd_mask_data_i((16*i)+15 downto 16*i)
      );
end generate;

  rst_i         <= NOT rst_n_in;
  lut_wr_addr_i <= unsigned(regbus_a_in(7 downto 0));
  lut_wr_i      <= NOT fcmbus_en_n_in AND NOT regbus_rnw_in;

-------------------------------------------------------------------------------
-- Output Signal Connections
-------------------------------------------------------------------------------

  spare_ctrl_reg_out <= spare_ctrl_reg_i;
--  spare_ctrl_reg_out( 7 downto  0) <= spare_ctrl_reg_i( 7 downto 0);
--  spare_ctrl_reg_out(12 downto  8) <= spare_ctrl_reg_i(12 downto 8);
--  spare_ctrl_reg_out(31 downto 10) <= spare_ctrl_reg_i(31 downto 10);

  rrif_ctrl_reg_out <= rrif_ctrl_reg_i;
  rod_mode_out      <= rod_mode_i;
  test_mux_out      <= test_mux_i;

  fecmd_link_mask0_out <= fecmd_link_mask0_i;
  fecmd_link_mask1_out <= fecmd_link_mask1_i;
  fe_cmd_mask_sel_out  <= command_mask_select_i;
  feocc_reset_reg_out  <= feocc_reset_i;
  feocc_num_l1_reg_out <= feocc_num_l1_i;

  calstrobe_delay_reg_out <= calstrobe_delay_reg_i;      
  cal_command_reg_out     <= cal_command_reg_i;

  cal_event_type0_out <= cal_event_type0_i;
  cal_event_type1_out <= cal_event_type1_i;
  static_l1id0_out    <= static_l1id0_i;
  static_l1id1_out    <= static_l1id1_i;
  static_bcid0_out    <= static_bcid0_i;
  static_bcid1_out    <= static_bcid1_i;
  nevent_out          <= nevent_i;
  precal_delay_out    <= precal_td_i;
  trig_delay0_out     <= trig_delay0_i;
  trig_delay1_out     <= trig_delay1_i;
  interval_out        <= interval_i;
  trig_mode_out       <= trig_mode_i;

  inmem_wordcount_reg_out   <= inmem_wordcount_reg_i;
  dbgmem_wordcount_reg_out  <= dbgmem_wordcount_reg_i;
  cnfg_rdback_count_reg_out <= cnfg_rdback_count_reg_i;
  ide_debugmem_cnfg_reg_out <= ide_debugmem_cnfg_reg_i;
  rol_test_block_reg_out    <= rol_test_block_reg_i;
  rol_test_block_count_out  <= rol_test_block_count_i;

  slave_dsp_ints_out <= ints_to_sdsps_i;
  ecr_busy_count_out <= ecr_busy_count_i;
  ecr_count_out      <= ecr_count_i;
    
  sdsp_inttohost_signal_out <= (    slave_dsp_ints_in(0) AND
                                    slave_dsp_ints_in(1) AND
                                    slave_dsp_ints_in(2) AND
                                    slave_dsp_ints_in(3) AND
                                NOT slave_dsp_ints_in(4) AND
                                NOT slave_dsp_ints_in(5) AND
                                NOT slave_dsp_ints_in(6) AND
                                NOT slave_dsp_ints_in(7)     );
  test_irq_out <= test_irq_i;

  fftv_cmnd_out <= fftv_cmnd_i;
  set_rod_busy_out <= set_rod_busy_i;

-------------------------------------------------------------------------------
-- PROCESS DECLARATION
-------------------------------------------------------------------------------
--  The output of the TIM FIFO is used to drive the FE Command blocks, and the
--  ROD EVID and MB block in stand alone debug mode.  The output must be masked
--  to these blocks when the system is in data-taking mode.
tim_fifo_output_mux : process (
  rst_n_in, 
  clk_in
  )
begin                      
  if (rst_n_in = '0') then
    int_tim_fifo_outdata_out <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
    if (int_tim_fifo_re_in = '1') then
      int_tim_fifo_outdata_out <= int_tim_fifo_outdata_i;
    else
      int_tim_fifo_outdata_out <= (others => '0');
    end if;
  end if;
end process tim_fifo_output_mux;

-------------------------------------------------------------------------------

write_registers : process (
  clk_in, 
  rst_n_in, 
  regbus_a_in, 
  regbus_d_in,
  regbus_en_n_in, 
  regbus_rnw_in
  )
begin
  if (rst_n_in = '0') then
    rrif_ctrl_reg_i         <= (others => '0');
    -- Startup with all functions in reset
    spare_ctrl_reg_i        <= X"000000F8";  --X"000000FF"; 
    -- Startup with all functions in reset
    rod_mode_i              <= (others => '0');
    test_mux_i              <= "0000";
    fecmd_link_mask0_i      <= (others => '0'); -- All Cmd Links Masked
    fecmd_link_mask1_i      <= (others => '0'); -- All Cmd Links Masked
--  calstrobe_delay_reg_i   <= "10011100"; -- Default to 156 clocks (27 + 132 - 3) = 156
    calstrobe_delay_reg_i   <= X"40"; -- Default to 64 clocks
    cal_command_reg_i       <= "101011100001100111111110000"; -- issue Cal Stobe to all Chips
    inmem_wordcount_reg_i   <= (others => '0');
    dbgmem_wordcount_reg_i  <= (others => '0');
    cnfg_rdback_count_reg_i <= (others => '0');
    ide_debugmem_cnfg_reg_i <= (others => '0');
    rol_test_block_reg_i    <= (others => '0');
    feocc_reset_i           <= (others => '1');
    feocc_num_l1_i          <= (others => '0');
    cal_event_type0_i       <= (others => '0');
    cal_event_type1_i       <= (others => '0');
    int_tim_fifo_indata_i   <= (others => '0');
    int_tim_fifo_ren_i      <= '0'; 
    int_tim_fifo_wen_i      <= '0';
    command_mask_ready_i    <= '0';
    command_mask_select_i   <= (others => '0');
    clr_hrdy_toc_i          <= '0';
    send_int_test_i         <= '0';
    send_int_test_out       <= '0';
--  test_irq_out            <= "100";
    test_irq_i              <= "100";
    load_event_type_out     <= '0';
    mdsp_gpreg_i            <= (others => '0');
    nevent_i                <= (others => '0');
    precal_td_i             <= X"05";
    trig_delay0_i           <= X"00F0";
    trig_delay1_i           <= X"00F0";
    interval_i              <= (others => '0');
    trig_mode_i             <= (others => '0');
    pin_hold_reg_i          <= (others => '0');
    ecr_busy_count_i        <= X"1D4C";  --X"7530";
  elsif (clk_in'event AND clk_in = '1') then
    case rb_access is
      when idle => 
        spare_ctrl_reg_i(0) <= '0'; -- Clear SW ECR
--        if (spare_stts_reg_in(31) = '0') then
--        spare_ctrl_reg_i(1) <= '0'; -- Clear ECR Flush
--        end if;
    -- Enable Formatter MB and EFB Dynamic Mask FIFOs after reset
--        spare_ctrl_reg_i(2 downto 0) <= (others => '0');
        spare_ctrl_reg_i(2) <= '0';
    -- Enable the Corrective Trigger FIFO after reset
        spare_ctrl_reg_i(7) <= '0';
    -- Set the RRIF_CMND_1 Register to support the RoL Test selected and mask all FE outputs
        if (rol_test_block_reg_i(0) = '1') then
          fecmd_link_mask0_i <= (others => '0'); -- All Cmd Links Masked
          fecmd_link_mask1_i <= (others => '0'); -- All Cmd Links Masked
          if (rol_test_block_reg_i(2) = '1') then -- TIM Control
            rrif_ctrl_reg_i <= X"00001523"; 
          else  -- RCF Control (SP Trigger Controller)      
            rrif_ctrl_reg_i <= X"00141421"; 
          end if;
        end if;

        if (regbus_en_n_in = '0' AND regbus_en_n_in_i = '1') then 
          if (regbus_rnw_in = '1') then
            rb_access <= read;
          else
            rb_access <= write;
          end if;
        end if;

      when read => 
        if (regbus_a_in(7 downto 0) = "00001011") then
          clr_hrdy_toc_i <= '1';
        elsif (regbus_a_in(7 downto 0) = "00100000") then
          int_tim_fifo_ren_i <= '1';
        end if;
        rb_access <= done;
        
      when write => 
        case regbus_a_in(7 downto 0) is
          when X"00" => mdsp_gpreg_i <= regbus_d_in;
  --      when X"01" => dsp_stts_reg_in       (read only)
  --      when X"02" => test_stts_reg_in      (read only)
  --      when X"03" => NOT USED              (read only)
          when X"04" => rrif_ctrl_reg_i(27 downto  0) <= regbus_d_in(27 downto  0);
                        rrif_ctrl_reg_i(28)           <= regbus_d_in(28) AND spare_stts_reg_in(31);  -- Allow in Pixel ROD only
                        rrif_ctrl_reg_i(31 downto 29) <= regbus_d_in(31 downto 29);
          when X"05" => spare_ctrl_reg_i      <= regbus_d_in;
          when X"06" => rod_mode_i            <= regbus_d_in( 3 downto  0);
                        test_mux_i            <= regbus_d_in(19 downto 16);
          when X"07" => command_mask_select_i <= regbus_d_in( 2 downto  0);
                        command_mask_ready_i  <= '0';
                        fetch_fe_mask_i       <= '1';
  --      when X"08" => rrif_stts_reg_in       (read only)
  --      when X"09" => spare_stts_reg_in      (read only)
          when X"0A" => set_rod_busy_i <= regbus_d_in(24);
          when X"0C" => fecmd_link_mask0_i(31 downto  0) <= regbus_d_in;
                        command_mask_ready_i             <= '0';
          when X"0D" => fecmd_link_mask0_i(47 downto 32) <= regbus_d_in(15 downto 0);
                        command_mask_ready_i             <= '0';
          when X"0E" => fecmd_link_mask1_i(31 downto  0) <= regbus_d_in;
                        command_mask_ready_i             <= '0';
          when X"0F" => fecmd_link_mask1_i(47 downto 32) <= regbus_d_in(15 downto 0);
                        command_mask_ready_i             <= '0';
          when X"10" => calstrobe_delay_reg_i            <= regbus_d_in( 7 downto  0);
          when X"11" => cal_command_reg_i                <= regbus_d_in(26 downto  0);
          when X"12" => ecr_count_i                      <= regbus_d_in( 7 downto  0); 
          when X"13" => ecr_busy_count_i                 <= regbus_d_in(31 downto 16); 
  --      when X"14" => modebits_fifo_stts_reg_in   (read only)
  --      when X"16" => dynmmask_fifo_stts_reg_in   (read only)
  --      when X"17" => dynmmask_wc_stts_reg_in     (read only)
          when X"18" => inmem_wordcount_reg_i   <= regbus_d_in;
          when X"19" => dbgmem_wordcount_reg_i  <= regbus_d_in;
          when X"1A" => cnfg_rdback_count_reg_i <= regbus_d_in;
          when X"1C" => ide_debugmem_cnfg_reg_i <= regbus_d_in(17 downto  0);
                        rol_test_block_reg_i    <= regbus_d_in(23 downto 20);
                        rol_test_block_count_i  <= regbus_d_in(29 downto 24);
  --      when X"1D" => ide_debugmem_stts_reg_in    (read only)
          when X"20" => int_tim_fifo_indata_i    <= regbus_d_in( 7 downto  0);
                        int_tim_fifo_wen_i       <= '1'; 
          when X"24" => ints_to_sdsps_i <= regbus_d_in( 3 downto 0);
  --      when X"25" => slave_dsp_ints_in     (read only)
          when X"26" => send_int_test_i <= regbus_d_in(0);
                        test_irq_i      <= regbus_d_in(6 downto 4);
          when X"28" => feocc_reset_i(31 downto  0) <= regbus_d_in;
          when X"29" => feocc_reset_i(63 downto 32) <= regbus_d_in;
          when X"2A" => feocc_reset_i(95 downto 64) <= regbus_d_in;
          when X"2C" => feocc_num_l1_i( 31 downto  0) <= regbus_d_in;
          when X"2D" => feocc_num_l1_i( 63 downto 32) <= regbus_d_in;
          when X"2E" => feocc_num_l1_i( 95 downto 64) <= regbus_d_in;
          when X"2F" => feocc_num_l1_i(103 downto 96) <= regbus_d_in(7 downto  0);
          when X"30" => nevent_i      <= regbus_d_in(15 downto  0);  -- NEvent
                        precal_td_i   <= regbus_d_in(23 downto 16);  -- PreCAL Delay
          when X"31" => trig_delay0_i <= regbus_d_in(15 downto  0);  -- TrigDelay1
                        trig_delay1_i <= regbus_d_in(31 downto 16);  -- TrigDelay2
          when X"32" => interval_i        <= regbus_d_in;  -- Interval
          when X"33" => trig_mode_i       <= regbus_d_in;  -- TrigMode
          when X"34" => cal_event_type0_i <= regbus_d_in( 9 downto  0);
          when X"35" => cal_event_type1_i <= regbus_d_in( 9 downto  0);
--        when X"36" => Scan Done     (Read Only) 
--        when X"37" =>
          when X"38" => static_l1id0_i <= regbus_d_in(23 downto  0);
          when X"39" => static_l1id1_i <= regbus_d_in(23 downto  0);
          when X"3A" => static_bcid1_i <= regbus_d_in(27 downto 16);
                        static_bcid0_i <= regbus_d_in(11 downto  0);
          when X"3C" => fftv_cmnd_i <= regbus_d_in( 7 downto  0);
          when others =>  -- NOP
        end case;
        rb_access <= done;

      when done => 
        if (ackn_cycle_state = reset) then
          rb_access <= idle;
          int_tim_fifo_ren_i <= '0'; 
          int_tim_fifo_wen_i <= '0'; 
          clr_hrdy_toc_i <= '0';
          spare_ctrl_reg_i( 8) <= '0';  -- Clear TIM FIFO Re-Tx bits
          spare_ctrl_reg_i( 9) <= '0';  -- Clear InMem FIFO Re-Tx bits
          spare_ctrl_reg_i(12) <= '0';  -- Clear ECR mID Reset
          pin_hold_reg_i <= (others => '0'); -- Clear all ROD Busy Latched Status bits
        end if;
      when others => 
    end case;

    if (fe_cmd_mask_updated_in = '1') then
      rrif_ctrl_reg_i(2)   <= '0';
      command_mask_ready_i <= '1';
    end if;
    if (corrective_mask_sent_in = '1') then
      rrif_ctrl_reg_i(23) <= '0';
    end if;

    for i in 0 to 11 loop
      if (pin_hold_reg_in(i) = '1') then
        pin_hold_reg_i(i) <= '1';
      end if;
    end loop;

    if (new_mask_in = '1') then
      command_mask_select_i <= mask_en_in;
      fetch_fe_mask_i <= '1';
    end if;

    case mask_fetch is
      when idle =>
        fe_cmd_mask_addr_i  <= (others => '0');
        fe_cmd_mask_count_i <= (others => '0');
        load_event_type_out <= '0';
        if (fetch_fe_mask_i = '1') then
          mask_fetch          <= fetch_mask_data;
          fe_cmd_mask_addr_i  <= unsigned("100" & command_mask_select_i & "00"); 
          load_event_type_out <= rrif_ctrl_reg_i(29); -- Auto load of ROD Event type if int scan mode enabled
        end if;
      when fetch_mask_data =>
        case fe_cmd_mask_count_i(2 downto 0) is
          when "001"  => fecmd_link_mask0_i(31 downto  0) <= fe_cmd_mask_data_i;
          when "010"  => fecmd_link_mask0_i(47 downto 32) <= fe_cmd_mask_data_i(15 downto  0);
          when "011"  => fecmd_link_mask1_i(31 downto  0) <= fe_cmd_mask_data_i;
          when "100"  => fecmd_link_mask1_i(47 downto 32) <= fe_cmd_mask_data_i(15 downto  0);
                         mask_fetch          <= done;
                         fetch_fe_mask_i     <= '0';
                         rrif_ctrl_reg_i(2)  <= '1';
          when others => null;
        end case;
        fe_cmd_mask_addr_i  <= fe_cmd_mask_addr_i + 1;
        fe_cmd_mask_count_i <= fe_cmd_mask_count_i + 1;
      when done =>
--      load_event_type_out <= rrif_ctrl_reg_i(29); -- Auto load of ROD Event type if int scan mode enabled
        mask_fetch <= idle;
    end case;
     
    if (send_int_test_i = '1') then
      send_int_test_out <= '1';
      send_int_test_i   <= '0';
    else
      send_int_test_out <= '0';
    end if;

  end if;
end process write_registers;

hdry_to_counter : process (
  rst_n_in, 
  clk_in
  )
begin                      
  if (rst_n_in = '0') then
    hrdy_toc_i <= (others => '0');
  elsif (clk_in'event and clk_in = '1') then
    if (inc_hrdy_toc_in = '1') then
      hrdy_toc_i <= hrdy_toc_i + 1 ;
    elsif (clr_hrdy_toc_ii = '1') then
      hrdy_toc_i <= (others => '0');
    end if;
  end if;
end process hdry_to_counter;


-- depending on how fast register routing paths are in the fpga
-- more or less waiting may be required.
-- this provides for an open loop acknowledge system, which is fine
-- for internal registers, since their internal routing is well defined.
-- Must make this ce edge sesetive, otherwise will initial multiple times.

reg_transfer_acknowledge_generator : process (
  rst_n_in, 
  clk_in, 
  regbus_en_n_in
  )
begin                      
  if (rst_n_in = '0') then
    regbus_ack_out    <= '0';
    regbus_en_n_in_i  <= '1';
    int_tim_fifo_en_i <= '0';
    clr_hrdy_toc_ii   <= '0';
    ackn_cycle_state  <= reset;

  elsif (clk_in'event and clk_in = '1') then
    regbus_en_n_in_i <= regbus_en_n_in; -- en_n_in negative edge delay detection register
  
    case ackn_cycle_state is
      when reset =>
        regbus_ack_out    <= '0';
        int_tim_fifo_en_i <= '0';
        clr_hrdy_toc_ii   <= '0';
      --    
        if (regbus_en_n_in = '0' AND regbus_en_n_in_i = '1') then 
          ackn_cycle_state <= wait_0_clk;
        end if;
        
      when wait_0_clk =>   -- Latch Read Data from TIM Fifo
        regbus_ack_out    <= '0';
        ackn_cycle_state  <= wait_1_clk;

      when wait_1_clk =>   -- Write Data to TIM Fifo
        regbus_ack_out    <= '0';
        int_tim_fifo_en_i <= int_tim_fifo_wen_i OR int_tim_fifo_ren_i;
        ackn_cycle_state  <= acknowledge;
  
      when acknowledge =>
        regbus_ack_out    <= '1';
        int_tim_fifo_en_i <= '0';
      --
        if (regbus_en_n_in = '1') then
          clr_hrdy_toc_ii  <= clr_hrdy_toc_i;
          ackn_cycle_state <= reset;
        end if;
    end case;
  end if;  
end process reg_transfer_acknowledge_generator;     

regbus_readout : process (
  rst_n_in, 
  clk_in
  )
constant code_version : std_logic_vector(7 downto 0) := X"5E"; -- v5.E;
constant rod_revision : std_logic_vector(3 downto 0) := X"F";  -- revF;
begin      
  if (rst_n_in = '0') then
  elsif (clk_in'event and clk_in = '1') then
  -- Always clear the regbus_out bus so as left over garbage won't be 
  -- asserted onto the bus turnaround time required to transition into
  -- the "read/decoder" stage of this block.
    regbus_d_out <= (others => '0');
    case regbus_a_in(7 downto 0) is
      when X"00" => regbus_d_out <= mdsp_gpreg_i;
      when X"01" => regbus_d_out <= dsp_stts_reg_in;
      when X"02" => regbus_d_out <= test_stts_reg_in;

  -- Code Version and ROD Revision #
      when X"03" => regbus_d_out(11 downto 0) <= rod_revision & code_version;

  -- Main & Spare Control Registers
      when X"04" => regbus_d_out <= rrif_ctrl_reg_i;
      when X"05" => regbus_d_out <= spare_ctrl_reg_i;

  -- ROD Mode Register
      when X"06" => regbus_d_out( 3 downto  0) <= rod_mode_i;
                    regbus_d_out(19 downto 16) <= test_mux_i;

  -- FE Command Mask Select Register
      when X"07" => regbus_d_out(2 downto 0) <= command_mask_select_i;

  -- Main & Spare Status Registers
      when X"08" => regbus_d_out <= command_mask_ready_i &
                                    rrif_stts_reg_in(30 downto 0);
      when X"09" => regbus_d_out <= spare_stts_reg_in;
      when X"0A" => regbus_d_out(31 downto 25) <= pin_hold_reg_in(31 downto 25); -- ROD Busy Status
                    regbus_d_out(24)           <= set_rod_busy_i;
                    regbus_d_out(23 downto 12) <= pin_hold_reg_in(23 downto 12); -- ROD Busy Status
                    regbus_d_out(11 downto  0) <= pin_hold_reg_i;
     --  HRDY Time Out Count  
      when X"0B" => regbus_d_out   <= std_logic_vector(hrdy_toc_i);
--                  clr_hrdy_toc_i <= '1';

  -- FE Command Link Mask Registers
      when X"0C" => regbus_d_out              <= fecmd_link_mask0_i(31 downto  0);
      when X"0D" => regbus_d_out(15 downto 0) <= fecmd_link_mask0_i(47 downto 32);
      when X"0E" => regbus_d_out              <= fecmd_link_mask1_i(31 downto  0);
      when X"0F" => regbus_d_out(15 downto 0) <= fecmd_link_mask1_i(47 downto 32);
        
  -- Calstrobe Registers       
      when X"10" => regbus_d_out(7 downto 0) <= calstrobe_delay_reg_i;
     -- time to from the start of the cal command to the start of the cal l1 trigger pulse + 1clock
      
  -- Calibrate Command Bit Stream
      when X"11" => regbus_d_out(26 downto 0) <= cal_command_reg_i;

  -- ECR Count Readback Register
      when X"12" => regbus_d_out( 7 downto 0) <= ecr_count_i;
      when X"13" => regbus_d_out(31 downto 0) <= ecr_busy_count_i & X"00" & ecr_count_in;

  -- Formatter Mode Bits, and EFB Dynamic Mask Registers
      when X"14" => 
        regbus_d_out(27 downto 16) <= modebits_fifo_stts_reg_in(23 downto 14) &
                                      modebits_fifo_stts_reg_in(3) &
                                      modebits_fifo_stts_reg_in(1);
          -- bits [25..18]: internalB_fifo_occupancy,        
          -- bits      17: internalB_fifo_ff,        
          --           16: internalB_fifo_ef        
        regbus_d_out(11 downto 0) <= modebits_fifo_stts_reg_in(13 downto 4) &
                                     modebits_fifo_stts_reg_in(2) &
                                     modebits_fifo_stts_reg_in(0);
          -- bits [9..2]: internalA_fifo_occupancy, 
          -- bits      1: internalA_fifo_ff,        
          --           0: internalA_fifo_ef,       
                                            
      when X"16" =>
        regbus_d_out( 4 downto 0) <= dynmmask_fifo_stts_reg_in(4 downto 0);
        regbus_d_out(12 downto 8) <= dynmmask_fifo_stts_reg_in(9 downto 5);
      when X"17" =>
        regbus_d_out( 7 downto  0) <= dynmmask_fifo_stts_reg_in(17 downto 10);
        regbus_d_out(15 downto  8) <= dynmmask_fifo_stts_reg_in(25 downto 18);
        regbus_d_out(23 downto 16) <= dynmmask_fifo_stts_reg_in(33 downto 26);
        regbus_d_out(31 downto 24) <= dynmmask_fifo_stts_reg_in(41 downto 34);

  -- TestBench Debug Word_Counts_Register
      when X"18" => regbus_d_out <= inmem_wordcount_reg_i;
      when X"19" => regbus_d_out <= dbgmem_wordcount_reg_i;
      when X"1A" => regbus_d_out <= cnfg_rdback_count_reg_i;

  -- DebugMode_Configuration_Register[31..0] :
      when X"1C" => 
        regbus_d_out(17 downto  0) <= ide_debugmem_cnfg_reg_i;
        regbus_d_out(19 downto 18) <= "00";
        regbus_d_out(23 downto 20) <= rol_test_block_reg_i;
        regbus_d_out(29 downto 24) <= rol_test_block_count_i;

  -- DebugMode_Status_Register[31..0]
      when X"1D" =>
        regbus_d_out(31 downto 30) <= (others => '0');
        regbus_d_out(29 downto 18) <= int_tim_occ_count_i;
        regbus_d_out(17)           <= int_tim_fifo_ff_i;
        regbus_d_out(16)           <= int_tim_fifo_ef_i;
        regbus_d_out(15 downto  8) <= debug_fifo_stts_in;
        regbus_d_out( 7 downto  0) <= ide_debugmem_stts_reg_in;   -- 

  -- Internal Tim Fifo
      when X"20" => regbus_d_out(7 downto 0) <= int_tim_fifo_outdata_i;
--                  int_tim_fifo_ren_i       <= '1'; 

  -- Debug Counters
      when X"21" => regbus_d_out <= std_logic_vector(dropped_l1a_count);
      when X"22" => regbus_d_out <= std_logic_vector(dropped_ecr_count);
      when X"23" => regbus_d_out <= std_logic_vector(dropped_bcr_count);

  -- The master dsp interrupts the slave dsp's hpi-interrupt edge mode should be selected 
  -- in the slave dsp interrupt configuration register.
      when X"24" => regbus_d_out( 3 downto 0) <= ints_to_sdsps_i;

  -- The master dsp gets an OR'd interrupt from all of the slave dsp on one of it main
  -- interrup lines.  It can then utilizes this register to determine which slave issued
  -- the interrupt.  SET THIS TO LATCH!!!!!!!!!!
      when X"25" => regbus_d_out(7 downto 0) <= slave_dsp_ints_in;
      when X"26" => regbus_d_out(6 downto 4) <= test_irq_i;
                    regbus_d_out(0)          <= send_int_test_i;

  -- Front end occupancy counters reset command pulse 
      when X"28" => regbus_d_out <= feocc_reset_i(31 downto  0);
      when X"29" => regbus_d_out <= feocc_reset_i(63 downto 32);
      when X"2A" => regbus_d_out <= feocc_reset_i(95 downto 64);

  -- Adder value for Pixel FE Occ Counters
      when X"2C" => regbus_d_out <= feocc_num_l1_i( 31 downto  0);
      when X"2D" => regbus_d_out <= feocc_num_l1_i( 63 downto 32);
      when X"2E" => regbus_d_out <= feocc_num_l1_i( 95 downto 64);
      when X"2F" => regbus_d_out(31 downto 8) <= (others => '0');
		                regbus_d_out(7 downto  0) <= feocc_num_l1_i(103 downto 96);
  --RCF Internal Scan Register
      when X"30" => regbus_d_out(15 downto  0) <= nevent_i;      -- NEvent
                    regbus_d_out(23 downto 16) <= precal_td_i;   -- PreCAL Delay
      when X"31" => regbus_d_out(15 downto  0) <= trig_delay0_i; -- TrigDelay1
                    regbus_d_out(31 downto 16) <= trig_delay1_i; -- TrigDelay2
      when X"32" => regbus_d_out <= interval_i;   -- Interval
      when X"33" => regbus_d_out <= trig_mode_i;  -- TrigMode

  -- Calibration Event Type Register 0
      when X"34" => regbus_d_out( 9 downto  0) <= cal_event_type0_i;

  -- Calibration Event Type Register 1
      when X"35" => regbus_d_out( 9 downto  0) <= cal_event_type1_i;

  -- Scan Done Status Register
      when X"36" => regbus_d_out(0) <= int_nevent_done_in;
--        when "00110111" => regbus_d_out <=

  -- Calibration L1ID and BCID Status
      when X"38" => regbus_d_out(23 downto  0) <= cal_l1id0_in;
      when X"39" => regbus_d_out(23 downto  0) <= cal_l1id1_in;
      when X"3A" => regbus_d_out(27 downto 16) <= cal_bcid1_in;
                    regbus_d_out(11 downto  0) <= cal_bcid0_in;
      when X"3B" => regbus_d_out(23 downto  0) <= l1id_bits_in;

  -- FFTV Registers
      when X"3C" => regbus_d_out( 7 downto  0) <= fftv_cmnd_i;
      when X"3D" => regbus_d_out(28 downto  0) <= fftv_busy_stat_in;
      when X"3E" => regbus_d_out( 8 downto  0) <= fftv_ID_stat_in;
                    regbus_d_out(15 downto 12) <= fftv_match_level_in;

      when others => regbus_d_out <= (others => '0');
    end case;
  end if;  
end process;
    
end rtl; -- code of register_block
