-------------------------------------------------------------------------------
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 1999
-- ReadOutDriver Electronics
--                                                                                             
-------------------------------------------------------------------------------
-- ROD Controller FPGA
-- Filename: top.vhd 
-- Description:
--  Glue for all the necessary ROD rescources to talk to the Controller
--  DSP. Implemention of the real-time and quasi-real-time control paths.
--
-------------------------------------------------------------------------------
-- Documentation:
--
-------------------------------------------------------------------------------
--
-------------------------------------------------------------------------------
-- Structure: 
-------------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk40
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
-------------------------------------------------------------------------------
-- Authors: Mark L. Nagel, John Joseph, Thorsten Stezelberger, Matt Warren
-- History:
--
--
--    07/17/01 JMJ First version - adapted from MLN RRIF Version
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- LIBRARY INCLUDES
-------------------------------------------------------------------------------
--
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations
use IEEE.std_logic_unsigned.all;
--
-------------------------------------------------------------------------------
-- SIMULATION LIBRARY INCLUDES
-------------------------------------------------------------------------------
--
-- --#pragma translate_off
library UNISIM;
use unisim.all;
use unisim.vcomponents;
-- --#pragma translate_on
--
-------------------------------------------------------------------------------
-- PORT DECLARATION
-------------------------------------------------------------------------------
--
-- Pin count of this device ==> 447
--

entity top_rcf is           -- RodControllerFPGA
port(
-- main clock and reset for fpga -- 2
  rst_n_in : in  std_logic; -- asynchronous global reset 
  clk40_in : in  std_logic; -- rod_clk40 input

-- rod_resources_interface_fpga LED status outputs -- 8
  led_output_n_out : out   std_logic_vector(7 downto 0); -- LED outputs

-- controller dsp External Memory Interface -- 61
  controller_bus_data_inout : inout std_logic_vector(31 downto 0); -- dsp data lines
  controller_bus_adr_in     : in    std_logic_vector(19 downto 0); -- dsp address lines 
  controller_bus_be_n_in    : in    std_logic_vector( 3 downto 0); -- dsp byte eanbles 
  controller_bus_ce_n_in    : in    std_logic; -- dsp chip enable
  controller_bus_oe_n_in    : in    std_logic; -- dsp output enable
  controller_bus_re_n_in    : in    std_logic; -- dsp read enable
  controller_bus_rnw_in     : in    std_logic; -- dsp write enable
  controller_bus_ack_out    : out   std_logic; -- dsp acknowledge input (ARDY)

-- controller dsp Timer Counter Unit Inteface -- 4
  controller_tinp_out : out   std_logic_vector(1 downto 0); -- dsp timer1,0 input
  controller_tout_in  : in    std_logic_vector(1 downto 0); -- dsp timer1,0 output 

-- controller dsp Syncronous Serial Port Interface  -- 12
  controller_dx_in    : in    std_logic_vector(1 downto 0); -- dsp sspi1,0 transmit serial data 
  controller_clkx_out : out   std_logic_vector(1 downto 0); -- dsp sspi1,0 transmit clock 
  controller_fsx_out  : out   std_logic_vector(1 downto 0); -- dsp sspi1,0 transmit frame syncroniztion
  controller_dr_out   : out   std_logic_vector(1 downto 0); -- dsp sspi1,0 receive serial data (Used for Reset Commands)
  controller_clkr_in  : in    std_logic_vector(1 downto 0); -- dsp sspi1,0 receive clock        
  controller_fsr_in   : in    std_logic_vector(1 downto 0); -- dsp sspi1,0 receive frame syncroniztion

-- controller dsp Interrupt Inputs -- 10
  controller_ext_int_n_out : out   std_logic_vector(4 downto 0); -- mdsp external interrupt
  ext_iacknum_in           : in    std_logic_vector(4 downto 0); -- mdsp external interrupt ack sigs  

-- slave dsp Interrupt Inputs  -- 8
  slave_ext_int_in : in    std_logic_vector(3 downto 0); -- interrupts from the slavedsps
  slave_hint_n_in  : in    std_logic_vector(3 downto 0); -- /HINT from the slavedsps
  
-- slave dsp Interrupt Outputs -- 4
  slave_ext_int_n_out : out   std_logic_vector(3 downto 0); -- interrupts to the slavedsp

-- master dsp Interrupt Input  -- 1
  mdsp_int_n_in       : in    std_logic; -- /HINT for the Master DSP
  
-- vme interrupt I/O           -- 4
  vme_bus_int_n_out  : out   std_logic; -- VME interrupt output 
  vme_irq_out        : out   std_logic_vector(2 downto 0); --
  
-- rodBUS -- 75
  rod_bus_data_inout : inout std_logic_vector(31 downto 0); -- dsp data lines
  rod_bus_adr_out    : out   std_logic_vector(19 downto 0); -- dsp address lines 
  rod_bus_ce_n_out   : out   std_logic_vector(10 downto 0); -- external component/module enables
  rod_bus_rnw_out    : out   std_logic; -- dsp read/not write signal
  fmt_hwob_out       : out   std_logic; -- Formatter Half Word Order
  fmt_ds_out         : out   std_logic; -- Formatter Data Strobe
  sdsp_wr_n_out      : out   std_logic; -- DSP write strobe
  sdsp_rd_n_out      : out   std_logic; -- DSP read strobe
  formA_ack_n_in     : in    std_logic; -- Formatter A acknowledge (active low)
  formB_ack_n_in     : in    std_logic; -- Formatter B acknowledge (active low)
  efb_ack_in         : in    std_logic; -- EFB acknowledge (active high)
  router_ack_in      : in    std_logic; -- Router acknowledge (active high)
  dsp0_hrdy_in       : in    std_logic; -- SDSP0 HRDY
  dsp1_hrdy_in       : in    std_logic; -- SDSP1 HRDY
  dsp2_hrdy_in       : in    std_logic; -- SDSP2 HRDY
  dsp3_hrdy_in       : in    std_logic; -- SDSP3 HRDY
  dsp0_detect_n_in   : in    std_logic;
  dsp1_detect_n_in   : in    std_logic;
  dsp2_detect_n_in   : in    std_logic;
  dsp3_detect_n_in   : in    std_logic;

-- reset control/management -- 4
  reset_commands_in  : in    std_logic_vector(1 downto 0); -- fpga program memory and reset manager reset directives to rodcontrollerDSP 
  reset_commands_out : out   std_logic_vector(1 downto 0); -- fpga program memory and reset manager reset directives from rodcontrollerDSP

-- PRM Host Port shared with MDSP
  mdsp_hd            : inout std_logic_vector(15 downto 0);
  mdsp_hhwil_in      : in    std_logic;  -- HHWIL to the Master DSP
  mdsp_hcntl0_in     : in    std_logic;  -- HCNTL0 to the Master DSP
  mdsp_hcntl1_in     : in    std_logic;  -- HCNTL1 to the Master DSP
  mdsp_hds1_n_in     : in    std_logic;  -- HDS1_N to the Master DSP
  mdsp_hcs_n_in      : in    std_logic;  -- HCS_N to the Master DSP
  mdsp_hrnw_in       : in    std_logic;  -- HRNW to the Master DSP
  rcf_hrdy_n_out     : out   std_logic;  -- HRDY from the Master DSP
  
-- formatters trailer flags messaging interface -- 20
  formatter_writes_trailer_in : in    std_logic_vector(11 downto 0); -- signal to readout formatter/fifoFPGA ic trailer flags 
  show_trailer_flags_out      : out   std_logic_vector( 7 downto 0); -- signal to readout formatter/fifoFPGA ic trailer flags 
  header_trailer_limit_in     : in    std_logic_vector( 1 downto 0);

-- TIM interface -- 8
  serial_eventtype_n_in       : in    std_logic; --
  serial_l1id_n_in            : in    std_logic; --
  cal_strobe_n_in             : in    std_logic; --
  l1_accept_n_in              : in    std_logic; --
  bunch_counter_reset_n_in    : in    std_logic; --
  event_counter_reset_n_in    : in    std_logic; --
  ttc6_n_in                   : in    std_logic; --
  ttc7_n_in                   : in    std_logic; --

-- formatters mode bits interface "mb" -- 26
  formatter_mb_fff_n_in        : in    std_logic_vector( 1 downto 0); -- mode bit fifos full flag
  formatter_mb_fef_n_in        : in    std_logic_vector( 1 downto 0); -- mode bit fifos empty flag
  formatter_mb_data_out        : out   std_logic_vector(11 downto 0); -- mode bit information to be loaded to the formatter mode bit fifos
  formatter_mb_we_n_out        : out   std_logic_vector( 7 downto 0); -- mode bit fifo write enable
  formatter_mb_fifo_rst_n_out  : out   std_logic_vector( 1 downto 0); -- mode bit fifo clear        
  fe_cmdpulse_formattersAB_out : out   std_logic_vector( 1 downto 0); -- l1trigger pulse to formatter master accumulators
  
-- formatter rod busy ios -- 3
  rod_busy_n_in                : in    std_logic_vector(1 downto 0); -- this signal is new
  rod_busy_n_out               : out   std_logic; --                    this signal is new
  
-- event fragment builder event dynamic mask interface "edm" -- 20
  efb_edm_fifo_empty_error_in : in    std_logic; -- event dynamic mask fifos full flag
  efb_edm_fff_n_in            : in    std_logic; -- event dynamic mask fifos full flag
  efb_edm_fef_n_in            : in    std_logic; -- event dynamic mask fifos empty flag
  efb_edm_data_out            : out   std_logic_vector(15 downto 0); -- event dynamic mask information to be loaded to the formatter dynamic mask fifos
  efb_edm_we_n_out            : out   std_logic; -- event dynamic mask fifo write enable 
  efb_edm_fifo_rst_n_out      : out   std_logic; -- EFB Flush/Reset

-- front end command/configuration messaging interface -- 50
  fe_command_link_out         : out   std_logic_vector(47 downto 0); -- front end command link interface output leads
  fe_command_link_dir_out     : out   std_logic;
  fe_command_link_en_out      : out   std_logic;

-- inmem, debugmem and eventmem debug memory buffers control and status signals -- 74
  inmem_ctrl_out              : out   std_logic_vector(24 downto 0); -- InputMem control signals
  inmem_status_in             : in    std_logic_vector( 3 downto 0); -- InputMem Empty/Full Flags
  eventmem_ctrl_out           : out   std_logic_vector(10 downto 0); -- EventMem control signals
  eventmem_status_in          : in    std_logic_vector( 3 downto 0); -- EventMem Empty/Full Flags

-- component boc_interface -- 3
  bocbus_strb_out             : out   std_logic;
  bocbus_rodoe_out            : out   std_logic;
  bocbus_write_out            : out   std_logic;  -- '0' - read, '1' - write
  bocbus_busy_in              : in    std_logic;  -- indicates that the BOC is busy

  boc_clock_ok_in             : in    std_logic;
  tim_clock_ok_in             : in    std_logic;
  slink_we_n_in               : in    std_logic;
  slink_ldown_n_in            : in    std_logic;
  slink_lff_n_in              : in    std_logic;
  rod_type_in                 : in    std_logic;

-- test header inout pins for debug -- 38
  l1_trigger_test_out         : out   std_logic;
  test_out                    : out   std_logic_vector( 7 downto 0);
  spare_pin                   : inout std_logic_vector(31 downto 0)
  );
end top_rcf; 

architecture rtl of top_rcf is
-------------------------------------------------------------------------------
-- SIGNAL DECLARATION
-------------------------------------------------------------------------------

-- pad buffer signals go here
signal rst_n_in_p                     : std_logic; -- asynchronous global reset 
signal rst_in_i                       : std_logic;
signal evt_id_rst_n                   : std_logic;

signal clk40_g                        : std_logic; -- rod_clk40 output of global clock buffer
signal rcf_dll_clk40_o                : std_logic; -- rod_clk40 input to global clock buffer
signal clk80_g                        : std_logic; -- 
signal rcf_dll_clk80_o                : std_logic; -- 
signal nclk40_g                       : std_logic; -- 180 Version of clk40
signal rcf_dll_nclk40_o               : std_logic; -- 
signal rcf_dll_clk40_i                : std_logic; -- rod_clk40 input to clock dll
signal rcf_dll_locked_i               : std_logic;
	 
signal led_output_n_out_p             : std_logic_vector(7 downto 0); -- LED outputs

signal controller_bus_data_in_p       : std_logic_vector(31 downto 0); -- dsp data lines
signal controller_bus_data_out_p      : std_logic_vector(31 downto 0); -- dsp data lines
signal controller_bus_adr_in_p        : std_logic_vector(19 downto 0); -- dsp address lines
signal controller_bus_be_n_in_p       : std_logic_vector( 3 downto 0); -- dsp be_n
signal controller_bus_ce_n_in_p       : std_logic; -- dsp chip enable
signal controller_bus_oe_n_in_p       : std_logic; -- dsp output enable
signal controller_bus_re_n_in_p       : std_logic; -- dsp read enable
signal controller_bus_rnw_in_p        : std_logic; -- dsp write enable
signal controller_bus_tbuf_driver_i   : std_logic; -- dsp bus buffer enable  
signal controller_bus_ack_out_p       : std_logic; -- dsp acknowledge input (ARDY)
signal controller_bus_ack_out_i       : std_logic; -- dsp acknowledge internal

signal controller_tinp_out_p          : std_logic_vector(1 downto 0); -- dsp timer1,0 input
signal controller_tout_in_p           : std_logic_vector(1 downto 0); -- dsp timer1,0 output 
signal controller_dr_out_p            : std_logic_vector(1 downto 0); -- dsp sspi1,0 receive serial data 
signal controller_dx_in_p             : std_logic_vector(1 downto 0); -- dsp sspi1,0 transmit serial data 
signal controller_dx_i                : std_logic_vector(1 downto 0); -- dsp sspi1,0 transmit serial data 
signal controller_clkr_in_p           : std_logic_vector(1 downto 0); -- dsp sspi1,0 receive clock        
signal controller_clkx_out_p          : std_logic_vector(1 downto 0); -- dsp sspi1,0 transmit clock 
signal controller_fsr_in_p            : std_logic_vector(1 downto 0); -- dsp sspi1,0 receive frame syncroniztion
signal controller_fsx_out_p           : std_logic_vector(1 downto 0); -- dsp sspi1,0 transmit frame syncroniztion
signal controller_ext_int_n_out_p     : std_logic_vector(4 downto 0); -- dsp external interrupt

signal ext_iacknum_in_p               : std_logic_vector(4 downto 0); -- mdsp external interrupt ack sigs  
signal slave_ext_int_in_p             : std_logic_vector(3 downto 0); -- slavedsp external interrupt inputs
signal slave_hint_n_in_p              : std_logic_vector(3 downto 0); -- slavedsp /HINT inputs
signal slave_ext_int_n_out_p          : std_logic_vector(3 downto 0); -- slavedsp external interrupt outputs
signal mdsp_int_n_in_p                : std_logic; -- this signal is new from the change in vme bus bridge chips
signal vme_bus_int_n_out_p            : std_logic; -- vme interrupt output 
signal vme_irq_out_p                  : std_logic_vector(2 downto 0); --

signal rod_bus_data_in_p              : std_logic_vector(31 downto 0); -- dsp data lines
signal rod_bus_data_out_p             : std_logic_vector(31 downto 0); -- dsp data lines
signal rod_bus_adr_out_p              : std_logic_vector(19 downto 0); -- dsp address lines 
signal rod_bus_ce_n_out_p             : std_logic_vector(10 downto 0); -- external component/module enables
signal rod_bus_rnw_out_p              : std_logic; -- dsp read/now write signal

signal sdsp_wr_n_out_p                : std_logic;
signal sdsp_rd_n_out_p                : std_logic;
signal sdsp_wr_n_out_i                : std_logic;
signal sdsp_rd_n_out_i                : std_logic;
signal fmt_hwob_p                     : std_logic;
signal fmt_ds_p                       : std_logic;

signal formA_ack_n_in_p               : std_logic; -- Formatter A acknowledge (active low)
signal formB_ack_n_in_p               : std_logic; -- Formatter B acknowledge (active low)
signal efb_ack_in_p                   : std_logic; -- EFB acknowledge (active high)
signal router_ack_in_p                : std_logic; -- Router acknowledge (active high)
signal dsp0_hrdy_in_p                 : std_logic; -- SDSP0 HRDY
signal dsp1_hrdy_in_p                 : std_logic; -- SDSP1 HRDY
signal dsp2_hrdy_in_p                 : std_logic; -- SDSP2 HRDY
signal dsp3_hrdy_in_p                 : std_logic; -- SDSP3 HRDY
signal dsp0_detect_in_n_p             : std_logic;
signal dsp1_detect_in_n_p             : std_logic;
signal dsp2_detect_in_n_p             : std_logic;
signal dsp3_detect_in_n_p             : std_logic;
signal dsp_detect_i                   : std_logic_vector( 3 downto 0);

signal reset_commands_in_p            : std_logic_vector(1 downto 0); -- 
signal reset_commands_out_p           : std_logic_vector(1 downto 0); --

signal formatter_writes_trailer_in_p  : std_logic_vector(11 downto 0); -- signal to readout formatter/fifoFPGA ic trailer flags 
signal show_trailer_flags_out_p       : std_logic_vector( 7 downto 0); -- signal to readout formatter/fifoFPGA ic trailer flags 
signal header_trailer_limit_in_p      : std_logic_vector( 1 downto 0);

signal l1_accept_n_in_p               : std_logic; --
signal event_counter_reset_n_in_p     : std_logic; --
signal bunch_counter_reset_n_in_p     : std_logic; --
signal cal_strobe_n_in_p              : std_logic; --
signal serial_l1id_n_in_p             : std_logic; --
signal serial_eventtype_n_in_p        : std_logic; --
signal ttc6_n_in_p                    : std_logic; --
signal ttc7_n_in_p                    : std_logic; --

signal formatter_mb_fff_n_in_p        : std_logic_vector(1 downto 0); -- mode bit fifos full flag
signal formatter_mb_fef_n_in_p        : std_logic_vector(1 downto 0); -- mode bit fifos empty flag
signal formatter_mb_data_out_p        : std_logic_vector(11 downto 0); -- 
signal formatter_mb_we_n_out_p        : std_logic_vector(7 downto 0); -- mode bit fifo write enable
signal formatter_mb_fifo_rst_n_out_p  : std_logic_vector(1 downto 0); -- mode bit fifo clear        
signal fe_cmdpulse_formattersAB_out_p : std_logic_vector(1 downto 0); -- l1trigger pulse to formatter master accumulators

signal rod_busy_n_in_p                : std_logic_vector(1 downto 0); -- ROD Formatters Error
signal rod_busy_n_out_p               : std_logic;

signal efb_edm_fifo_empty_error_in_p  : std_logic; -- event dynamic mask fifos full flag
signal efb_edm_fff_n_in_p             : std_logic; -- event dynamic mask fifos full flag
signal efb_edm_fef_n_in_p             : std_logic; -- event dynamic mask fifos empty flag
signal efb_edm_data_out_p             : std_logic_vector(15 downto 0); --
signal efb_edm_we_n_out_p             : std_logic; -- event dynamic mask fifo write enable 
signal efb_edm_fifo_rst_n_out_p       : std_logic; -- event dynamic mask fifo write enable 
signal efb_ready_p                    : std_logic;

signal fe_command_link_out_p          : std_logic_vector(47 downto 0); -- front end command link interface output leads
signal fe_command_link_dir_out_p      : std_logic;
signal fe_command_link_en_out_p       : std_logic;

signal inmem_status_in_p              : std_logic_vector(3 downto 0); -- inmem status signals
signal eventmem_status_in_p           : std_logic_vector(3 downto 0); -- eventmem status signals

signal inmem_ctrl_out_i               : std_logic_vector(24 downto 0); -- inmem control signals
signal inmem_ctrl_out_p               : std_logic_vector(24 downto 0); -- inmem control signals
signal eventmem_ctrl_out_p            : std_logic_vector(10 downto 0); -- eventmem control signals

signal bocbus_strb_out_p              : std_logic;
signal bocbus_rodoe_out_p             : std_logic;
signal bocbus_busy_in_p               : std_logic;  -- of these signals
signal bocbus_write_out_p             : std_logic;

signal boc_clock_ok_p                 : std_logic;
signal tim_clock_ok_p                 : std_logic;
signal slink_ldown_n_p                : std_logic;
signal slink_lff_n_p                  : std_logic;
signal slink_we_n_p                   : std_logic;

signal rod_type_p                     : std_logic;
signal test_out_p                     : std_logic_vector( 7 downto 0);
signal spare_i                        : std_logic_vector(31 downto 0);
signal spare_o                        : std_logic_vector(31 downto 0);
signal spare_t                        : std_logic_vector(31 downto 0);

-- interconnection signals go here

signal slave_dsp_ints_out_i       : std_logic_vector(3 downto 0);
signal slave_dsp_ints_in_i        : std_logic_vector(7 downto 0);
signal slv_int_to_mdsp_i          : std_logic;
signal slv_hint_to_mdsp_i         : std_logic;
signal enable_slv_int_out_i       : std_logic;
signal modebits_fifo_stts_reg_i   : std_logic_vector(23 downto 0);
signal dynmmask_fifo_stts_reg_i   : std_logic_vector(41 downto 0);
signal calstrobe_delay_reg_i      : std_logic_vector( 7 downto 0);
signal cal_command_reg_i          : std_logic_vector(26 downto 0);
signal feocc_reset_i              : std_logic_vector(95 downto 0);
signal feocc_num_l1_i             : std_logic_vector(103 downto 0);
signal rrif_ctrl_reg_i            : std_logic_vector(31 downto 0);
signal spare_ctrl_reg_i           : std_logic_vector(31 downto 0);
signal rrif_stts_reg_i            : std_logic_vector(31 downto 0);
signal spare_stts_reg_i           : std_logic_vector(31 downto 0);
signal dsp_stts_reg_i             : std_logic_vector(31 downto 0);
signal test_stts_reg_i            : std_logic_vector(31 downto 0);
signal pin_hold_reg_i             : std_logic_vector(31 downto 0);
signal rol_test_block_i           : std_logic_vector( 3 downto 0);
signal rol_test_block_count_i     : std_logic_vector( 5 downto 0);
signal rol_test_block0_ii         : std_logic;
signal ide_debugmem_cnfg_reg_i    : std_logic_vector(17 downto 0);
signal ide_debugmem_stts_reg_i    : std_logic_vector( 7 downto 0);
signal debug_fifo_stts_i          : std_logic_vector( 7 downto 0);
signal inmem_ctrl_i               : std_logic_vector(15 downto 0);
signal eventmem_ctrl_i            : std_logic_vector(10 downto 0);
signal tb_inmem_ctrl_out_i        : std_logic_vector(11 downto 0);
signal inmem_wordcount_i          : std_logic_vector(31 downto 0);
signal dbgmem_wordcount_i         : std_logic_vector(31 downto 0);
signal cnfg_rdback_count_i        : std_logic_vector(31 downto 0);
signal fe_command_mask0_i         : std_logic_vector(47 downto 0); --
signal fe_command_mask1_i         : std_logic_vector(47 downto 0); --
signal focbus_a_i                 : std_logic_vector(11 downto 0); -- Frontend Occupancy Counter bus
signal focbus_din_i               : std_logic_vector(31 downto 0);
signal focbus_dout_i              : std_logic_vector(31 downto 0); --
signal focbus_en_n_i              : std_logic;
signal focbus_rnw_i               : std_logic;
signal focbus_ack_i               : std_logic;
signal fe_l1_pulse_i              : std_logic_vector(47 downto 0);
signal fe_pulse_i                 : std_logic;
signal corrective_fe_pulse_i      : std_logic;
signal fe_cmd_mask_updated_i      : std_logic;
signal frebus_a_i                 : std_logic_vector(7 downto 0); -- Formatter Readoutmodebits Encoder bus
signal frebus_a_ii                : unsigned(7 downto 0);
signal frebus_din_i               : std_logic_vector(15 downto 0);
signal frebus_dout_i              : std_logic_vector(15 downto 0);
signal frebus_en_n_i              : std_logic;
signal frebus_rnw_i               : std_logic;
signal frebus_ack_i               : std_logic;
signal l1id_bits_i                : std_logic_vector(23 downto 0);
signal bcid_bits_i                : std_logic_vector(11 downto 0);
signal ttype_bits_i               : std_logic_vector( 9 downto 0);
signal new_l1id_valid_i           : std_logic;
signal new_ttype_valid_i          : std_logic;
signal ecr_id                     : unsigned(7 downto 0);
signal ecr_count_i                : std_logic_vector(7 downto 0);
signal cal_l1a0_i                 : std_logic;
signal cal_l1a1_i                 : std_logic;
signal cal_l1a_i                  : std_logic_vector( 1 downto 0);
signal cal_l1id0_i                : std_logic_vector(23 downto 0);
signal cal_l1id1_i                : std_logic_vector(23 downto 0);
signal cal_l1id0_m                : std_logic_vector(23 downto 0);
signal cal_l1id1_m                : std_logic_vector(23 downto 0);
signal static_l1id0_i             : std_logic_vector(23 downto 0);
signal static_l1id1_i             : std_logic_vector(23 downto 0);
signal cal_bcid0_i                : std_logic_vector(11 downto 0);
signal cal_bcid1_i                : std_logic_vector(11 downto 0);
signal static_bcid0_i             : std_logic_vector(11 downto 0);
signal static_bcid1_i             : std_logic_vector(11 downto 0);
signal cal_event_type0_i          : std_logic_vector( 9 downto 0);
signal cal_event_type1_i          : std_logic_vector( 9 downto 0);
signal edebus_a_i                 : std_logic_vector(7 downto 0); -- EFB dynamic mask encoder bus
signal edebus_a_ii                : unsigned(7 downto 0);
signal edebus_din_i               : std_logic_vector(15 downto 0);
signal edebus_dout_i              : std_logic_vector(15 downto 0);
signal edebus_rnw_i               : std_logic;
signal edebus_ack_i               : std_logic;
signal edebus_en_n_i              : std_logic;
signal dfibus_a_i                 : std_logic_vector(6 downto 0); -- Debgumem Fifo Inteface bus
signal dfibus_rnw_i               : std_logic;
signal dfibus_inmem_en_n_i        : std_logic;
signal dfibus_eventmem_en_n_i     : std_logic;
signal dfibus_ack_i               : std_logic;
signal fcmbus_en_n_i              : std_logic;
signal fcmbus_dout_i              : std_logic_vector(31 downto 0);
signal regbus_en_n_i              : std_logic;
signal regbus_rnw_i               : std_logic;
signal regbus_a_i                 : std_logic_vector(11 downto 0);
signal regbus_d_in_i              : std_logic_vector(31 downto 0);
signal regbus_d_out_i             : std_logic_vector(31 downto 0);
signal regbus_ack_i               : std_logic;
signal rbus_d_in_i                : std_logic_vector(7 downto 0);
signal rbus_d_out_i               : std_logic_vector(7 downto 0);
signal trap_indicator_i           : std_logic_vector(1 downto 0);
signal trigger_descriptor_i       : std_logic_vector(1 downto 0);
signal read_trigger_que_i         : std_logic;
signal store_trig2ti_que_i        : std_logic;
signal rodbus_en_n_itr_i          : std_logic_vector(6 downto 0);
signal int_tim_fifo_re_i          : std_logic; --
signal int_tim_fifo_rt_i          : std_logic; --
signal int_tim_fifo_outdata_i     : std_logic_vector(7 downto 0); 
signal tim_signal_ii              : std_logic_vector(5 downto 0);
signal l1_accept_i                : std_logic; --
signal event_counter_reset_i      : std_logic; -- 
signal fe_delayed_ecr_i           : std_logic; -- 
signal bunch_counter_reset_i      : std_logic; --
signal cal_strobe_i               : std_logic; --
signal serial_l1id_i              : std_logic; --
signal serial_eventtype_i         : std_logic; --
signal ttc6_i                     : std_logic; --
signal ttc7_i                     : std_logic; --
signal slink_ldown_i              : std_logic;
signal slink_lff_i                : std_logic;
signal slink_we_i                 : std_logic;
signal emif_re_i			            : std_logic;
signal dsp_timer1                 : std_logic;
signal pause_triggers_i           : std_logic;

signal gnd_i : std_logic;
signal vcc_i : std_logic;

type cnfg_rdbk_states is (
  reset, 
  idle, 
  start_delay, 
  wr_to_fifo,
  set_done_flag, 
  return_to_idle
  );
signal cnfg_rdbk_state : cnfg_rdbk_states;
signal wr_cnfg_data_inmemA_i : std_logic;
signal wr_cnfg_data_inmemB_i : std_logic;
signal rdbk_done_i           : std_logic;

type cal_test_states is (
  reset, 
  idle, 
  start_delay, 
  rd_from_fifo,
  set_rt_bit, 
  return_to_idle
  );
signal cal_test_state : cal_test_states;
signal rd_data_inmemA_i : std_logic;
signal rd_data_inmemB_i : std_logic;
signal rt_data_inmemA_i : std_logic;
signal rt_data_inmemB_i : std_logic;

type trap_data_states is (
  reset, 
  idle,
  start_delay, 
  wr_to_fifo,
  set_done_flag
  );
signal trap_data_state : trap_data_states;
signal wr_link_data_inmemA_i  : std_logic;
signal wr_link_data_inmemB_i  : std_logic;
signal trap_done_i            : std_logic;

signal modebits_sent_i : std_logic;

signal extbus_ack_i : std_logic; --
signal intbus_ack_i : std_logic; --

signal fsx0_count     : unsigned(4 downto 0);
signal fsx1_count     : unsigned(4 downto 0);
signal fsx_en_n_obuft : std_logic;

signal apply_dynmask_mb_i : std_logic_vector(1 downto 0);
signal fe_cmd_mask_sel_i  : std_logic_vector(2 downto 0);

signal mdsp_hpi_data_out_p : std_logic_vector(15 downto 0);
signal mdsp_hpi_data_in_p  : std_logic_vector(15 downto 0);
signal mdsp_hpi_data_dir_t : std_logic;
signal rcf_hrdy_n_p	       : std_logic;
signal mdsp_cen_n_p        : std_logic;
signal mdsp_rnw_p          : std_logic;
signal mdsp_hds1_n_p       : std_logic;
signal mdsp_hhwil_p	       : std_logic;
signal mdsp_hcntl0_p       : std_logic;
signal mdsp_hcntl1_p       : std_logic;

type led_states is (
  idle, 
  led_on, 
  led_off
  );
signal xoff_led_state : led_states;
signal slon_led_state : led_states;

signal test_mux_i         : std_logic_vector(3 downto 0);
signal cmd_cycle_done_i   : std_logic;
signal efb_header_pause_i : std_logic;
signal inc_hrdy_toc_i     : std_logic;

type vme_int_states is (
  idle,
  issue_int,
  clear_int,
  done
  );
signal vme_int_state : vme_int_states;
signal send_int_test_i : std_logic;
signal test_irq_i      : std_logic_vector(2 downto 0);
   
signal nevent_i       : std_logic_vector(15 downto 0);  -- NEvent
signal nevent_o       : std_logic_vector(15 downto 0);  -- NEvent
signal precal_delay_i : std_logic_vector( 7 downto 0);  -- PreCal Delay
signal trig_delay0_i  : std_logic_vector(15 downto 0);  -- TrigDelay1
signal trig_delay1_i  : std_logic_vector(15 downto 0);  -- TrigDelay2
signal interval_i     : std_logic_vector(31 downto 0);  -- Interval
signal trig_mode_i    : std_logic_vector(31 downto 0);  -- TrigMode
signal feocc_zero_i   : std_logic_vector( 3 downto 0); --
signal act_group_i    : std_logic_vector( 3 downto 0); --
signal cal_serial0_i  : std_logic; -- SerPort0
signal cal_serial1_i  : std_logic; -- SerPort1
signal cal_serial0_o  : std_logic; -- SerPort0
signal cal_serial1_o  : std_logic; -- SerPort1
signal mask_dxi_i     : std_logic;

signal int_scan_fecmd0_i : std_logic;
signal int_scan_fecmd1_i : std_logic;
signal int_load_event_i  : std_logic;
signal int_mask_en_i     : std_logic_vector(2 downto 0);
signal int_new_mask_i    : std_logic;
signal int_nevent_done_i : std_logic;

signal rol_l1a_i     : std_logic;
signal cmd_bit_count : unsigned(2 downto 0);
type rol_test_states is (
  idle,
  setup_delay,
  send_l1a,
  wait_for_cycle
  );
signal rol_test         : rol_test_states;
signal rol_led_n_i      : std_logic;
signal rol_test_pause_i : std_logic;
signal rol_test_i       : std_logic;
signal rol_trigs_on_i   : std_logic;

signal sweeper_event_on_rod_i : std_logic;
signal efb_rod_busy_i         : std_logic;
signal rod_tt_wr_i            : std_logic;

signal event_count1 : unsigned(23 downto 0);
signal event_count2 : unsigned(23 downto 0);
signal event_l1id   : std_logic_vector(23 downto 0);

signal fftv_cmnd_i        : std_logic_vector( 7 downto 0);
signal fftv_busy_stat_i   : std_logic_vector(28 downto 0);
signal fftv_ID_stat_i     : std_logic_vector( 8 downto 0);
signal fftv_match_level_i : std_logic_vector( 3 downto 0);
signal fftv_disable_i     : std_logic;

signal fmt_fe_cmd_i     : std_logic;
signal set_rod_busy_i   : std_logic;
signal ecr_rod_busy_i   : std_logic;
signal en_rb_window_cnt : std_logic;
signal rb_window_cnt    : unsigned(17 downto 0);

signal sw_ecr_cmd_i     : std_logic;
signal sw_ecr_cmd_p     : std_logic;
signal ecr_busy_count_i : std_logic_vector(15 downto 0);
signal cmd_proc_busy_i  : std_logic;

signal dropped_l1a_count : unsigned(31 downto 0);
signal dropped_ecr_count : unsigned(31 downto 0);
signal dropped_bcr_count : unsigned(31 downto 0);

-------------------------------------------------------------------------------
-- COMPONENT DECLARATION
-------------------------------------------------------------------------------

component address_decoder_busbridge
  port (
    rst_n_in           : in    std_logic; -- negative logic asynchronous global reset
    clk_in             : in    std_logic; -- rodclk40 input
    dspbus_ea_in       : in    std_logic_vector(21 downto 2); --
    dspbus_ed_in       : in    std_logic_vector(31 downto 0); --
    dspbus_ed_out      : out   std_logic_vector(31 downto 0); --
    dspbus_en_n_in     : in    std_logic; -- MDSP CE_N
    dspbus_be_n_in     : in    std_logic_vector( 3 downto 0); -- MDSP Bytes Enables
    dspbus_rnw_in      : in    std_logic; -- Write Strobe from the MDSP
    dspbus_re_n_in     : in    std_logic; -- Read Enable from the MDSP
    dspbus_oe_n_in     : in    std_logic; -- Output Enable from the MDSP
    dspbus_ack_out     : out   std_logic; --
    rodbus_a_out       : out   std_logic_vector(19 downto 0); --
    rodbus_d_in        : in    std_logic_vector(31 downto 0); --
    rodbus_d_out       : out   std_logic_vector(31 downto 0); --
    rodbus_en_n_out    : out   std_logic_vector(10 downto 0); -- strobe/enable from the addr
    rodbus_en_n_itr    : out   std_logic_vector( 6 downto 0); -- strobe/enable from the addr
    rodbus_rnw_out     : out   std_logic; --
    formA_ack_n_in     : in    std_logic; --
    formB_ack_n_in     : in    std_logic; --
    efb_ack_in         : in    std_logic; --
    router_ack_in      : in    std_logic; --
    bocbus_busy_in     : in    std_logic;
    bocbus_strb_out    : out   std_logic;
    bocbus_rodoe_out   : out   std_logic;
    bocbus_write_out   : out   std_logic;
    regbus_a_out       : out   std_logic_vector(11 downto 0); -- rrif global registers bus
    regbus_d_in        : in    std_logic_vector(31 downto 0); --
    regbus_d_out       : out   std_logic_vector(31 downto 0); --
    regbus_rnw_out     : out   std_logic; --
    regbus_ack_in      : in    std_logic; --
    fcmbus_d_in        : in    std_logic_vector(31 downto 0);
    focbus_a_out       : out   std_logic_vector(11 downto 0); -- front end occupancy counter
    focbus_d_in        : in    std_logic_vector(31 downto 0); --
    focbus_rnw_out     : out   std_logic; --
    focbus_ack_in      : in    std_logic; --
    fmebus_a_out       : out   std_logic_vector( 7 downto 0); -- Formatter ModeBit Encoder
    fmebus_d_in        : in    std_logic_vector(15 downto 0); --
    fmebus_d_out       : out   std_logic_vector(15 downto 0); --
    fmebus_rnw_out     : out   std_logic; --
    fmebus_ack_in      : in    std_logic; --
    edebus_a_out       : out   std_logic_vector( 7 downto 0); -- efb Dynamic Mask Encoder
    edebus_d_in        : in    std_logic_vector(15 downto 0); --
    edebus_d_out       : out   std_logic_vector(15 downto 0); -- 
    edebus_rnw_out     : out   std_logic; --
    edebus_ack_in      : in    std_logic; --
    dfibus_a_out       : out   std_logic_vector( 6 downto 0); -- Debugmem Fifo Interface bus
    dfibus_rnw_out     : out   std_logic;
    dfibus_ack_in      : in    std_logic;
    sdsp0_hrdy_in      : in    std_logic;   -- HRDY directly from SDSP0
    sdsp1_hrdy_in      : in    std_logic;   -- HRDY directly from SDSP1
    sdsp2_hrdy_in      : in    std_logic;   -- HRDY directly from SDSP2
    sdsp3_hrdy_in      : in    std_logic;   -- HRDY directly from SDSP3
    sdsp_wr_strb_n_out : out   std_logic;   -- SDSP WR Strobe
    sdsp_rd_strb_n_out : out   std_logic;   -- SDSP RD Strobe
    fmt_hwob_out       : out   std_logic;
    fmt_ds_out         : out   std_logic;
    efb_hwob_out       : out   std_logic;
    extbus_ack_out     : out   std_logic; --
    intbus_ack_out     : out   std_logic; --
    sdsp_detect_in     : in    std_logic_vector(3 downto 0);
    sdsp_slow_write_in : in    std_logic;
    inc_hrdy_toc_out   : out   std_logic;
    sdsp_address_error : out   std_logic --
    );
end component; 

component register_block
  port (
    clk_in                    : in  std_logic; -- clk40 input
    rst_n_in                  : in  std_logic; -- global reset
    regbus_en_n_in            : in  std_logic;
    regbus_rnw_in             : in  std_logic;
    regbus_a_in               : in  std_logic_vector(11 downto 0);
    regbus_d_in               : in  std_logic_vector(31 downto 0);
    regbus_d_out              : out std_logic_vector(31 downto 0);
    regbus_ack_out            : out std_logic;   
    fcmbus_en_n_in            : in  std_logic;
    fcmbus_d_out              : out std_logic_vector(31 downto 0);
    rrif_ctrl_reg_out         : out std_logic_vector(31 downto 0);
    spare_ctrl_reg_out        : out std_logic_vector(31 downto 0);
    rrif_stts_reg_in          : in  std_logic_vector(31 downto 0);
    spare_stts_reg_in         : in  std_logic_vector(31 downto 0);
    dsp_stts_reg_in           : in  std_logic_vector(31 downto 0);
    test_stts_reg_in          : in  std_logic_vector(31 downto 0);
    pin_hold_reg_in           : in  std_logic_vector(31 downto 0);
    fecmd_link_mask0_out      : out std_logic_vector(47 downto 0);
    fecmd_link_mask1_out      : out std_logic_vector(47 downto 0);
    calstrobe_delay_reg_out   : out std_logic_vector( 7 downto 0);
    cal_command_reg_out       : out std_logic_vector(26 downto 0);
    modebits_fifo_stts_reg_in : in  std_logic_vector(23 downto 0);
    dynmmask_fifo_stts_reg_in : in  std_logic_vector(41 downto 0);
    inmem_wordcount_reg_out   : out std_logic_vector(31 downto 0);
    dbgmem_wordcount_reg_out  : out std_logic_vector(31 downto 0);
    cnfg_rdback_count_reg_out : out std_logic_vector(31 downto 0);
    ide_debugmem_stts_reg_in  : in  std_logic_vector( 7 downto 0);
    ide_debugmem_cnfg_reg_out : out std_logic_vector(17 downto 0);
    debug_fifo_stts_in        : in  std_logic_vector( 7 downto 0);
    rol_test_block_reg_out    : out std_logic_vector( 3 downto 0);  
    rol_test_block_count_out  : out std_logic_vector( 5 downto 0);  
    int_tim_fifo_outdata_out  : out std_logic_vector( 7 downto 0);
    int_tim_fifo_re_in        : in  std_logic;
    int_tim_fifo_rt_in        : in  std_logic;
    slave_dsp_ints_out        : out std_logic_vector( 3 downto 0);
    slave_dsp_ints_in         : in  std_logic_vector( 7 downto 0);
    sdsp_inttohost_signal_out : out std_logic;
    feocc_reset_reg_out       : out std_logic_vector( 95 downto 0);
    feocc_num_l1_reg_out      : out std_logic_vector(103 downto 0);
    cal_event_type0_out       : out std_logic_vector(  9 downto 0);
    cal_event_type1_out       : out std_logic_vector(  9 downto 0);
    cal_l1id0_in              : in  std_logic_vector( 23 downto 0);
    cal_l1id1_in              : in  std_logic_vector( 23 downto 0);
    static_l1id0_out          : out std_logic_vector( 23 downto 0);
    static_l1id1_out          : out std_logic_vector( 23 downto 0);
    cal_bcid0_in              : in  std_logic_vector( 11 downto 0);
    cal_bcid1_in              : in  std_logic_vector( 11 downto 0);
    static_bcid0_out          : out std_logic_vector( 11 downto 0);
    static_bcid1_out          : out std_logic_vector( 11 downto 0);
    rod_mode_out              : out std_logic_vector(  3 downto 0);
    test_mux_out              : out std_logic_vector(  3 downto 0);
    fe_cmd_mask_sel_out       : out std_logic_vector(  2 downto 0);
    fe_cmd_mask_updated_in    : in  std_logic;
    corrective_mask_sent_in   : in  std_logic;
    inc_hrdy_toc_in           : in  std_logic;
    send_int_test_out         : out std_logic;
    test_irq_out              : out std_logic_vector( 2 downto 0);
    ecr_count_in              : in  std_logic_vector( 7 downto 0);
    ecr_count_out             : out std_logic_vector( 7 downto 0);
    nevent_out                : out std_logic_vector(15 downto 0);  -- NEvent
    precal_delay_out          : out std_logic_vector( 7 downto 0);  -- Pre CAL Delay
    trig_delay0_out           : out std_logic_vector(15 downto 0);  -- TrigDelay1
    trig_delay1_out           : out std_logic_vector(15 downto 0);  -- TrigDelay2
    interval_out              : out std_logic_vector(31 downto 0);  -- Interval
    trig_mode_out             : out std_logic_vector(31 downto 0);  -- TrigMode
    mask_en_in                : in  std_logic_vector( 2 downto 0);  --
    new_mask_in               : in  std_logic;
    load_event_type_out       : out std_logic;
    int_nevent_done_in        : in  std_logic;
    l1id_bits_in              : in  std_logic_vector(23 downto 0);
    fftv_cmnd_out             : out std_logic_vector( 7 downto 0);
    fftv_busy_stat_in         : in  std_logic_vector(28 downto 0);
    fftv_ID_stat_in           : in  std_logic_vector( 8 downto 0);
    fftv_match_level_in       : in  std_logic_vector( 3 downto 0);
    set_rod_busy_out          : out std_logic;
    ecr_busy_count_out        : out std_logic_vector(15 downto 0);
    dropped_l1a_count         : in  unsigned(31 downto 0);
    dropped_ecr_count         : in  unsigned(31 downto 0);
    dropped_bcr_count         : in  unsigned(31 downto 0)
    );
end component;

component debugmem_fifo_interface
  port (
    rst_n_in          : in  std_logic; -- negative logic asynchronous global rese
    clk_in            : in  std_logic; -- clk40 input
    control_in        : in  std_logic; -- 1: enable, 0: idle
    address_in        : in  std_logic_vector( 6 downto 0); --
    rnw_in            : in  std_logic; --
    inmem_en_n_in     : in  std_logic; --
    eventmem_en_n_in  : in  std_logic; --
    inmem_ctrl_out    : out std_logic_vector(15 downto 0); --
    eventmem_ctrl_out : out std_logic_vector(10 downto 0); --
    ack_out           : out std_logic  --
    );
end component; 

component testbench_interface
  port (
    rst_n_in                     : in  std_logic; -- global reset
    clk_in                       : in  std_logic; -- clk40 input
    control_in                   : in  std_logic_vector( 1 downto 0); --
--  control_in:  2=>run/stop, 1=>enable, 0=>clear
    debugmode_cnfg_register_in   : in  std_logic_vector(17 downto 0); --
    debugmode_stts_register_out  : out std_logic_vector( 7 downto 0); -- 
    inmem_wordcount_register_in  : in  std_logic_vector(31 downto 0); -- 
    dbgmem_wordcount_register_in : in  std_logic_vector(31 downto 0); -- 
    inmem_ctrl_out               : out std_logic_vector(11 downto 0); --
    int_tim_fifo_re_out          : out std_logic; --
    int_tim_fifo_rt_out          : out std_logic; --
    l1_trigger_in                : in  std_logic;
    pause_trigger_rt             : in  std_logic
    );
end component;

component trigger_signal_decoder
  port(
    clk_in              : in  std_logic; -- clk40 input
    rst_n_in            : in  std_logic; 
    enable_decoding_in  : in  std_logic; 
    serial_id_in        : in  std_logic; --
    serial_tt_in        : in  std_logic; --
    l1id_bits_out       : out std_logic_vector(23 downto 0); --
    bcid_bits_out       : out std_logic_vector(11 downto 0); --
    ttype_bits_out      : out std_logic_vector( 9 downto 0); --
    new_l1id_valid_out  : out std_logic; --
    new_ttype_valid_out : out std_logic  --
    );
end component; 

component efb_headerdynamicmask_encoder
  port(
    clk_in                   : in  std_logic;  -- clk40 input
    rst_n_in                 : in  std_logic;  -- negative logic global reset
    control_in               : in  std_logic;  -- 1:enable transfer, 0:flush all
    rnw_in                   : in  std_logic;
    ce_n_in                  : in  std_logic;
    ack_out                  : out std_logic;
    apply_default_mask_in    : in  std_logic;  -- interrupt to the MDSP to 
-- indicate that a DEFAULT trigger has occurred and that the default dynamic 
-- mask and default readoutmodebits will be written to the efb and formatter 
-- (respectively) on this trigger.
    apply_corrective_mask_in : in  std_logic;  -- interrupt to the master dsp
-- to indicate that a CORRECTIVE trigger has occurred and that the corrective 
-- dynamic mask and corrective readoutmodebits will be written to the efb and
-- formatter (respectively) on this AND ONLY this trigger.  This signal is used
-- to clear the new_mask_ready_in set/reset flip-flop in the 
-- front_end_command_processor control_register.
    record_headerdynamicmask_in    : in  std_logic;  --
    read_trig_indictor_que_out     : out std_logic;  --
    l1id_bits_in                   : in  std_logic_vector(23 downto 0);  --
    bcid_bits_in                   : in  std_logic_vector(11 downto 0);  --
    ttype_bits_in                  : in  std_logic_vector( 9 downto 0);  --
    new_l1id_valid_in              : in  std_logic;                      --
    new_ttype_valid_in             : in  std_logic;                      --
    int_lut_bdata_in               : in  std_logic_vector(15 downto 0);  --
    int_lut_bdata_out              : out std_logic_vector(15 downto 0);  --
    int_lut_adr_in                 : in  unsigned(7 downto 0);           --
    event_headermask_fifo_ef_n_in  : in  std_logic;                      --
    event_headermask_fifo_ff_n_in  : in  std_logic;                      --
    event_headermask_out           : out std_logic_vector(15 downto 0);  --
    event_lbid_count_out           : out std_logic_vector( 7 downto 0);
    event_trgt_count_out           : out std_logic_vector( 7 downto 0);
    event_mask_count_out           : out std_logic_vector( 7 downto 0);  --
    event_header_count_out         : out std_logic_vector( 7 downto 0);  --
    event_headermask_fifo_we_n_out : out std_logic;                      --
    int_fifo_ef_out                : out std_logic_vector( 4 downto 0);  -- 
    int_fifo_ff_out                : out std_logic_vector( 4 downto 0);  -- 
    ecrid_in                       : in  unsigned( 7 downto 0);
    int_cal_mode_en_in             : in  std_logic; -- 1:CalMode 0:Normal
    cal_l1a0_in                    : in  std_logic;
    cal_l1a1_in                    : in  std_logic;
    cal_tt0_in                     : in  std_logic_vector( 9 downto 0);  --
    cal_tt1_in                     : in  std_logic_vector( 9 downto 0);  --
    cal_l1id0_in                   : in  std_logic_vector(23 downto 0);
    cal_l1id1_in                   : in  std_logic_vector(23 downto 0);
    cal_bcid0_in                   : in  std_logic_vector(11 downto 0);
    cal_bcid1_in                   : in  std_logic_vector(11 downto 0);
    pause_triggers_out             : out std_logic;
    event_type_in                  : in  std_logic_vector( 2 downto 0);
    load_event_type_in             : in  std_logic;
    mask_dxi_in                    : in  std_logic;
    nevent_in                      : in  std_logic_vector(15 downto 0);
    boc_ok_in                      : in  std_logic;
    tim_ok_in                      : in  std_logic;
    rol_test_in                    : in  std_logic;
    rol_test_count_in              : in  std_logic_vector(5 downto 0);
    rod_tt_wr_out                  : out std_logic;
    event_l1id_out                 : out std_logic_vector(23 downto 0)
    );
end component;

component formatter_readoutmodebits_encoder
  port(
    clk_in                   : in  std_logic;  -- clk40 input
    rst_n_in                 : in  std_logic;  -- negative logic global reset
    control_in               : in  std_logic;  -- 0:flush all, 1:enable
    int_lut_adr_in           : in  unsigned(7 downto 0);
    int_lut_bdata_in         : in  std_logic_vector(15 downto 0);  --
    int_lut_bdata_out        : out std_logic_vector(15 downto 0);  --
    ce_n_in                  : in  std_logic;
    rnw_in                   : in  std_logic;
    ack_out                  : out std_logic;
-- interrupt to the MDSP to indicate that a DEFAULT trigger has occurred
    apply_default_mode_in    : in  std_logic;
-- interrupt to the MDSP to indicate that a CORRECTIVE trigger has occurred
    apply_corrective_mode_in : in  std_logic;
    modebits_fifo_ff_n_in    : in  std_logic_vector( 1 downto 0);  --
    modebits_out             : out std_logic_vector(11 downto 0);  --
    modebits_fifo_we_n_out   : out std_logic_vector( 7 downto 0);  --
    int_fifo_ef_out          : out std_logic_vector( 1 downto 0);  --
    int_fifo_ff_out          : out std_logic_vector( 1 downto 0);
    mb0_fifo_occ_count       : out std_logic_vector( 9 downto 0);
    mb1_fifo_occ_count       : out std_logic_vector( 9 downto 0);
    modebits_sent_out        : out std_logic;
    fe_cmd_mask_sel_in       : in  std_logic_vector( 2 downto 0)
    );
end component;

component fe_command_pulse_counter
  port(
   clk_in                    : in  std_logic; -- clk40 input
   rst_n_in                  : in  std_logic; -- negative logic asynchronous global reset
   control_in                : in  std_logic_vector(1 downto 0);
   fe_cmdpulse_in            : in  std_logic; --
   corrective_fe_cmdpulse_in : in  std_logic; --
   fe_cmdpulse_count_out     : out std_logic_vector(7 downto 0); --
   fe_cmdpulse_AB_out        : out std_logic_vector(1 downto 0); --
   controller_int_out        : out std_logic_vector(1 downto 0); --
   command_cycle_done_out    : out std_logic; --
   store_current_trigger_out : out std_logic; --
   current_trigger_out       : out std_logic_vector(1 downto 0);  --
   dynamic_mask_sent_in      : in  std_logic;
   modebits_sent_in          : in  std_logic;
   efb_ready_in              : in  std_logic;
   fe_trigger_occupancy_out  : out std_logic_vector(10 downto 0);
   modebits_error_out        : out std_logic
   );
end component; 

component fe_command_processor
  port(
    clk_in                    : in  std_logic; -- clk40 input
    rst_n_in                  : in  std_logic; -- asyncronous global reset
    enable_in                 : in  std_logic; -- 1=enabled, 0=disabled               
    rod_type_in               : in  std_logic; -- ROD Type 0=>SCT 1 =>Pixel
    cnfg_cal_new_mask_load_in : in  std_logic; -- 1=Enable Load, 0=Disable Load
    l1accept_in               : in  std_logic; -- l1trigger accept pulse, 1 clk wide.
    bcr_in                    : in  std_logic; -- bunch counter reset
    ecr_in                    : in  std_logic; -- event counter reset
    calibrate_in              : in  std_logic; -- calibrate command pulse, 1 clk wide
    calibrate_command_in      : in  std_logic_vector(26 downto 0); -- calibrate comma
    cal_trig_delay_in         : in  std_logic_vector( 7 downto 0); -- clks from when the cal
    fast_not_long_commands_in : in  std_logic;
    new_mask_ready_in         : in  std_logic; -- a new mask from the controller is ready a
    corrective_mask_ready_in  : in  std_logic; -- a new Corrective mask from the controller is ready and should be loaded on the next trigger
    fe_command_mask0_in       : in  std_logic_vector(47 downto 0); --
    fe_command_mask1_in       : in  std_logic_vector(47 downto 0); --
    long_command_stream_in    : in  std_logic_vector( 1 downto 0); -- the dsp_sspi_data_out[
    long_command_pulse_in     : in  std_logic_vector( 1 downto 0); -- the dsp_tcu_out[1..0]
    command_stream_out        : out std_logic_vector(47 downto 0); -- command_stream bits o
    pulse_stream_out          : out std_logic_vector(47 downto 0); -- command_pulse bits ou
    fe_pulse_out              : out std_logic; -- interrupt to the master dsp to signify th
    corrective_fe_pulse_out   : out std_logic;
    fe_cmd_mask_updated_out   : out std_logic; -- after a new mask ready signal has been re
    fmt_fe_cmd_out            : out std_logic;
    fmt_fe_cmd_mask_n_in      : in  std_logic;
    ecr_flush_en_in           : in  std_logic_vector( 1 downto 0);
    cmd_proc_busy_out         : out std_logic
    );
end component; 

component front_end_occupancy_counters
  port(
    clk_in                        : in    std_logic; -- clk40 input
    rst_n_in                      : in    std_logic; -- asyncronous global reset
    rod_type_in                   : in    std_logic; -- rod type bit 0 = SCT, 1 = Pixel
    trailerflags_readout_en_in    : in    std_logic; -- enable trailer flags readout/counter
    trailerflags_in               : in    std_logic_vector( 11 downto 0); --
    clear_fe_occupancy_counter_in : in    std_logic_vector( 95 downto 0); --
    num_l1_value_in               : in    std_logic_vector(103 downto 0); --  value 0-15 for Links 0-31 Pixel only
    fe_l1_pulse_in                : in    std_logic_vector( 47 downto 0); --
    fe_occupancy_gl_zero_out      : out   std_logic; --
    group_zero_out                : out   std_logic_vector(  3 downto 0); --
    group_en_out                  : out   std_logic_vector(  3 downto 0); --
    show_trailer_flags_out        : out   std_logic_vector(  7 downto 0); --
    focbus_a_in                   : in    std_logic_vector( 11 downto 0); -- front end occupancy counters bus
    focbus_d_in                   : in    std_logic_vector( 31 downto 0); --
    focbus_d_out                  : out   std_logic_vector( 31 downto 0); --
    focbus_rnw_in                 : in    std_logic; --
    focbus_en_n_in                : in    std_logic; --
    focbus_ack_out                : out   std_logic  --
    );
end component; 

component fifo_2048x2
  port(
    clk_in         : in  std_logic; -- clk40 input
    rst_n_in       : in  std_logic; -- negative logic asynchronous global reset
    flush_fifo_in  : in  std_logic; --
    wren_in        : in  std_logic; -- enable input register
    rden_in        : in  std_logic; -- enable output register
    data_in        : in  std_logic_vector(1 downto 0); --
    data_out       : out std_logic_vector(1 downto 0); --
    full_flag_out  : out std_logic; --
    empty_flag_out : out std_logic  --
    );
end component;

component sp_trig_cmd_detector
  port (
    clk_in         : in  std_logic; -- clk40 input
    rst_n_in       : in  std_logic; -- asynchronous global reset
    enable_in      : in  std_logic; -- 
    rod_type_in    : in  std_logic; -- 0=>SCT, 1=>Pixel
    l1id_select_in : in  std_logic; -- 
    bcid_select_in : in  std_logic; -- 
    cal_serial0_in : in  std_logic; -- SerPort0
    cal_serial1_in : in  std_logic; -- SerPort1
    l1_accept0_out : out std_logic; -- L1 Trig Detected on SerPort0
    l1_accept1_out : out std_logic; -- L1 Trig Detected on SerPort1
    l1id0_reg_in   : in  std_logic_vector(23 downto 0); --
    l1id1_reg_in   : in  std_logic_vector(23 downto 0); --
    l1id0_out      : out std_logic_vector(23 downto 0); --
    l1id1_out      : out std_logic_vector(23 downto 0); --
    bcid0_reg_in   : in  std_logic_vector(11 downto 0); --
    bcid1_reg_in   : in  std_logic_vector(11 downto 0); --
    bcid0_out      : out std_logic_vector(11 downto 0); --
    bcid1_out      : out std_logic_vector(11 downto 0)  --
    );
end component;

component int_scan_engine
  port (
    clk_in            : in  std_logic; -- clk40 input
    rst_n_in          : in  std_logic; -- asynchronous global reset
    enable_in         : in  std_logic; -- 
    clear_nevent_in   : in  std_logic; --
    new_mask_ready_in : in  std_logic; -- 
    rod_tt_wr_in      : in  std_logic;
    rod_type_in       : in  std_logic; -- 
    num_events_in     : in  std_logic_vector(15 downto 0); -- 
    precal_delay_in   : in  std_logic_vector( 7 downto 0);
    trig_delay0_in    : in  std_logic_vector(15 downto 0); --
    trig_delay1_in    : in  std_logic_vector(15 downto 0); --
    evt_interval_in   : in  std_logic_vector(31 downto 0); --
    trig_mode_in      : in  std_logic_vector(31 downto 0); --
    act_groups_in     : in  std_logic_vector( 3 downto 0); --
    feocc_zero_in     : in  std_logic_vector( 3 downto 0); --
    xoff_in           : in  std_logic;
    mask_en_out       : out std_logic_vector( 2 downto 0); --
    new_mask_out      : out std_logic;
    mask_dxi_out      : out std_logic;
    fecmd0_out        : out std_logic;
    fecmd1_out        : out std_logic;
    nevent_out        : out std_logic_vector(15 downto 0);
    nevent_done_out   : out std_logic
    );
end component;

component fftv_main
  generic(
  -- ROD firmware devs: Do not modify these generics - they work as is
    ROD_TRIGGER_MODE  : std_logic := '1';
    COUNT_WIDTH       : integer := 24; -- Considering veto's last for 1ms, this allows
    ID_WIDTH          : integer := 8;  -- best balance of ID's counted vs time/1ms counted (255/419)
    PERIOD_MIN        : integer := 85;
    PERIOD_MAX        : integer := 2666;
    PERIOD_OVERFLOW   : integer := 3076;
    PERIOD_DELTAT     : integer := 40;
    VETO_DURATION     : integer := 40000
    );

  port( 
    rst                   : in  std_logic;
    clk40                 : in  std_logic;
    fftv_match_lvl_i      : in  std_logic_vector (3 downto 0);
    fftv_match_lvl_o      : out std_logic_vector (3 downto 0);
    fftv_count_o          : out std_logic_vector ((COUNT_WIDTH-1) downto 0);
    fftv_count_clr_i      : in  std_logic;
    fftv_count_rollover_o : out std_logic;
    fftv_id_o             : out std_logic_vector ((ID_WIDTH-1) downto 0);
    fftv_id_clr_i         : in  std_logic;
    fftv_id_rollover_o    : out std_logic;
    fftv_emergency_o      : out std_logic;
    fftv_emergency_clr_i  : in  std_logic;
    trigger0_i            : in  std_logic;
    trigger0_o            : out std_logic;
    trigger1_i            : in  std_logic;
    trigger1_o            : out std_logic;
    fftv_busy_o           : out std_logic;
    fftv_disable_i        : in  std_logic
    );
end component;

--------------------------------------------------------------------------
-- PAD DECLARATIONS
--------------------------------------------------------------------------
component IBUFG
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC
    );
end component;

component BUFG
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC
    );
end component;

component IBUF
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC
    );
end component;

component OBUF
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC
    );
end component;

component OBUF_S_24
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC
    );
end component;

component OBUF_F_24
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC
    );
end component;

component OBUFT
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC;
    T : in  STD_LOGIC
    );
end component;

component IOBUF
  port (
    I  : in    STD_LOGIC;
    O  : out   STD_LOGIC;
    IO : inout STD_LOGIC;
    T  : in    STD_LOGIC
    );
end component;

component CLKDLL
  port (
    CLK0   : out STD_LOGIC;
    CLK90  : out STD_LOGIC;
    CLK180 : out STD_LOGIC;
    CLK270 : out STD_LOGIC;
    CLK2X  : out STD_LOGIC;
    CLKDV  : out STD_LOGIC;
    LOCKED : out STD_LOGIC;
    CLKIN  : in  STD_LOGIC;
    CLKFB  : in  STD_LOGIC;
    RST    : in  STD_LOGIC
    );
end component;


begin

--------------------------------------------------------------------------
-- PAD INSTANTIATION
--------------------------------------------------------------------------

rst_buf : IBUF
  port map (I => rst_n_in, O => rst_n_in_p);

rclk_buf    : IBUFG
  port map (I => clk40_in, O => rcf_dll_clk40_i);
  
rclk_bufg   : BUFG
  port map (I => rcf_dll_clk40_o, O => clk40_g);
  
rnclk_bufg   : BUFG
  port map (I => rcf_dll_nclk40_o, O => nclk40_g);

rod_clk_dll : CLKDLL 
  port map (
    CLK0   => rcf_dll_clk40_o,
    CLK90  => open,
    CLK180 => rcf_dll_nclk40_o,
    CLK270 => open,
    CLK2X  => open,
    CLKDV  => open,
    LOCKED => rcf_dll_locked_i,
    CLKIN  => rcf_dll_clk40_i,
    CLKFB  => clk40_g,
    RST    => gnd_i
    );

led_output_block :
  for i in 0 to  7 generate led_obufs : OBUF_S_24
    port map (I => led_output_n_out_p(i), O => led_output_n_out(i));
  end generate;

dsp_data_block :
  for i in 0 to 31 generate dsp_data_iobufs : IOBUF
    port map (
      I  => controller_bus_data_out_p(i),
      O  => controller_bus_data_in_p(i),
      IO => controller_bus_data_inout(i),
      T  => controller_bus_tbuf_driver_i
      );
  end generate;

dsp_addr_block :
  for i in 0 to 19 generate dsp_addr_ibufs : IBUF
    port map (
      I => controller_bus_adr_in(i),
      O => controller_bus_adr_in_p(i)
      );
  end generate;

dsp_be_n_block :
  for i in 0 to 3 generate dsp_be_n_ibuf : IBUF 
    port map (
      I => controller_bus_be_n_in(i),
      O => controller_bus_be_n_in_p(i)
      );
  end generate;

dsp_ce_n_ibuf : IBUF
  port map (I => controller_bus_ce_n_in, O => controller_bus_ce_n_in_p);
  
dsp_oe_n_ibuf : IBUF
  port map (I => controller_bus_oe_n_in, O => controller_bus_oe_n_in_p);
  
dsp_re_n_ibuf : IBUF
  port map (I => controller_bus_re_n_in, O => controller_bus_re_n_in_p);
  
dsp_we_n_ibuf : IBUF
  port map (I => controller_bus_rnw_in, O => controller_bus_rnw_in_p);
  
dsp_ARDY_obuf : OBUF
  port map (I => controller_bus_ack_out_p, O => controller_bus_ack_out);

dsp_McSP_block : for i in 0 to 1 generate
  dsp_tinp_obuf : OBUF
    port map (I => controller_tinp_out_p(i), O => controller_tinp_out(i));

  dsp_tout_ibuf : IBUF
    port map (I => controller_tout_in(i), O => controller_tout_in_p(i));

  dsp_dr_obuf : OBUF
    port map (I => controller_dr_out_p(i), O => controller_dr_out(i));

  dsp_dx_ibuf : IBUF
    port map (I => controller_dx_in(i), O => controller_dx_in_p(i));

  dsp_clkr_ibuf : IBUF
    port map (I => controller_clkr_in(i), O => controller_clkr_in_p(i));

  dsp_clkx_obuft : OBUFT
    port map (
      I => controller_clkx_out_p(i),
      O => controller_clkx_out(i),
      T => fsx_en_n_obuft
      );

  dsp_fsr_ibuf : IBUF
    port map (I => controller_fsr_in(i), O => controller_fsr_in_p(i));

  dsp_fsx_obuft : OBUFT
    port map (
      I => controller_fsx_out_p(i), 
      O => controller_fsx_out(i),
      T => fsx_en_n_obuft
      );
end generate;


dsp_ext_int_block :  for i in 0 to  4 generate
  dsp_ext_int_obuf : OBUF
    port map (
      I => controller_ext_int_n_out_p(i),
      O => controller_ext_int_n_out(i)
      );
end generate;

slave_ext_int_block :  for i in 0 to  3 generate
  slave_ext_int_n_ibuf : IBUF
    port map (I => slave_ext_int_in(i), O => slave_ext_int_in_p(i));

  slave_hint_n_ibuf : IBUF
    port map (I => slave_hint_n_in(i), O => slave_hint_n_in_p(i));

  slave_ext_int_n_obuf : OBUF
    port map(I => slave_ext_int_n_out_p(i), O => slave_ext_int_n_out(i));
end generate;

mdsp_int_n_ibuf : IBUF port map (I => mdsp_int_n_in, O => mdsp_int_n_in_p);

vme_bus_int_n_obuf : OBUF port map (I => vme_bus_int_n_out_p, O => vme_bus_int_n_out);
                             
vme_irq_out_block : for i in 0 to  2 generate
  vme_irq_obuf : OBUF
    port map (I => vme_irq_out_p(i), O => vme_irq_out(i));
end generate;


rodbus_data_block : for i in 0 to 31 generate
  rodbus_data_iobuf : IOBUF
    port map (
      I  => rod_bus_data_out_p(i),
      O  => rod_bus_data_in_p(i),
      IO => rod_bus_data_inout(i),
      T  => rod_bus_rnw_out_p
      );
end generate;

rodbus_addr_block :  for i in 0 to 19 generate
  rodbus_adr_obuf : OBUF
    port map (I => rod_bus_adr_out_p(i), O => rod_bus_adr_out(i));
end generate;

rodbus_cen_block_s12 : for i in 0 to 6 generate  -- Rod Bus peripherals
  rodbus_cen_obuf : OBUF
    port map (I => rod_bus_ce_n_out_p(i), O => rod_bus_ce_n_out(i));
end generate;

rodbus_cen_block_s24 : for i in 7 to 10 generate  -- SDSP Chip Selects
  rodbus_cen_obuf : OBUF_S_24
    port map (I => rod_bus_ce_n_out_p(i), O => rod_bus_ce_n_out(i));
end generate;

rodbus_rnw_obuf : OBUF  port map (I => rod_bus_rnw_out_p, O => rod_bus_rnw_out);

-------------------------------------------------------------------------------
-- DSP I/O And Reset commands
mdsp_hpi_data_inout_block : 
  for i in 0 to 15 generate
    mdsp_hpi_data_iobufs : IOBUF 
      port map (
        I  => mdsp_hpi_data_out_p(i),
        O  => mdsp_hpi_data_in_p(i),
        IO => mdsp_hd(i),
        T  => mdsp_hpi_data_dir_t
        );
end generate;

rcf_hrdy_n_obuf : OBUF
port map (I => rcf_hrdy_n_p, O => rcf_hrdy_n_out);

mdsp_cen_n_ibuf : IBUF 
port map (I => mdsp_hcs_n_in, O => mdsp_cen_n_p);

mdsp_rnw_ibuf : IBUF 
port map (I => mdsp_hrnw_in, O => mdsp_rnw_p);

mdsp_hds1_n_ibuf : IBUF 
port map (I => mdsp_hds1_n_in, O => mdsp_hds1_n_p);

mdsp_hhwil_ibuf : IBUF
port map (I => mdsp_hhwil_in, O => mdsp_hhwil_p);

mdsp_hcntl0_ibuf : IBUF
port map (I => mdsp_hcntl0_in, O => mdsp_hcntl0_p);

mdsp_hcntl1_ibuf : IBUF
port map (I => mdsp_hcntl1_in, O => mdsp_hcntl1_p);

reset_cmds_block : for i in 0 to 1 generate
  reset_commands_ibuf : IBUF
    port map (I => reset_commands_in(i), O => reset_commands_in_p(i));

  reset_commands_obuf : OBUF
    port map (I => reset_commands_out_p(i), O => reset_commands_out(i));
end generate;

fwt_block : for i in 0 to 11 generate 
  formatter_writes_trailer_ibuf : IBUF
    port map (I => formatter_writes_trailer_in(i), O => formatter_writes_trailer_in_p(i));
end generate;
  
stf_block : for i in 0 to  7 generate
  show_trailer_flags_obuf : OBUF
    port map (I => show_trailer_flags_out_p(i), O => show_trailer_flags_out(i));
end generate;

header_trailer_limit_in_block :
  for i in 0 to 1 generate
    header_trailer_limit_ibufs : IBUF
      port map (
        I => header_trailer_limit_in(i), 
        O => header_trailer_limit_in_p(i)
        );
   end generate;


ttc0_n_ibuf : IBUF
  port map (I => l1_accept_n_in, O => l1_accept_n_in_p);

ttc1_n_ibuf : IBUF
  port map (I => event_counter_reset_n_in, O => event_counter_reset_n_in_p);

ttc2_n_ibuf : IBUF
  port map (I => bunch_counter_reset_n_in, O => bunch_counter_reset_n_in_p);

ttc3_n_ibuf : IBUF
  port map (I => cal_strobe_n_in, O => cal_strobe_n_in_p);

ttc4_n_ibuf : IBUF
  port map (I => serial_l1id_n_in, O => serial_l1id_n_in_p);

ttc5_n_ibuf : IBUF
  port map (I => serial_eventtype_n_in, O => serial_eventtype_n_in_p);

ttc6_n_ibuf : IBUF
  port map (I => ttc6_n_in, O => ttc6_n_in_p);

ttc7_n_ibuf : IBUF
  port map (I => ttc7_n_in, O => ttc7_n_in_p);


modebits_flag_reset_block : for i in 0 to 1 generate
  mb_fff_n_ibuf : IBUF
    port map (I => formatter_mb_fff_n_in(i), O => formatter_mb_fff_n_in_p(i));

  mb_fef_n_ibuf : IBUF
    port map (I => formatter_mb_fef_n_in(i), O => formatter_mb_fef_n_in_p(i));   -- EFB to Formatter Pause Output

  mb_fifo_rst_n_obuf : OBUF
    port map (I => formatter_mb_fifo_rst_n_out_p(i), O => formatter_mb_fifo_rst_n_out(i));
end generate;

modebits_data_block : for i in 0 to 11 generate
  mb_data_obuf : OBUF
    port map (I => formatter_mb_data_out_p(i), O => formatter_mb_data_out(i));
end generate;

modebits_wen_block : for i in 0 to 7 generate
  mb_wen_obuf : OBUF
    port map (I => formatter_mb_we_n_out_p(i), O => formatter_mb_we_n_out(i));
end generate;

fe_cmdpulse_block : for i in 0 to 1 generate
  fe_cmdpulse_formattersAB_obuf : OBUF
    port map (I => fe_cmdpulse_formattersAB_out_p(i), O => fe_cmdpulse_formattersAB_out(i));
end generate;

rodbusy_block : for i in 0 to 1 generate
  rodbusy_ibuf : IBUF
    port map (I => rod_busy_n_in(i), O => rod_busy_n_in_p(i));
end generate;

rodbusy_obuf : OBUF
  port map (I => rod_busy_n_out_p, O => rod_busy_n_out);    

efb_edm_fifo_empty_error_ibuf : IBUF
  port map (I => efb_edm_fifo_empty_error_in, O => efb_edm_fifo_empty_error_in_p);


efb_edm_fff_n_ibuf : IBUF
  port map (I => efb_edm_fff_n_in, O => efb_edm_fff_n_in_p);

efb_edm_fef_n_ibuf : IBUF
  port map (I => efb_edm_fef_n_in, O => efb_edm_fef_n_in_p);

efb_edm_data_block : for i in 0 to 15 generate
  efb_edm_data_obuf : OBUF
    port map (I => efb_edm_data_out_p(i), O => efb_edm_data_out(i));
end generate;

efb_edm_wen_obuf : OBUF
  port map (I => efb_edm_we_n_out_p, O => efb_edm_we_n_out);

efb_edm_fifo_rst_n_obuf : OBUF
  port map(I => efb_edm_fifo_rst_n_out_p, O => efb_edm_fifo_rst_n_out);

fe_cmd_block : for i in 0 to 47 generate
  fe_command_link_obuf : OBUF
    port map (I => fe_command_link_out_p(i), O => fe_command_link_out(i));
end generate;
               
fe_command_link_dir_obuf : OBUF
  port map (I => fe_command_link_dir_out_p, O => fe_command_link_dir_out);
  
fe_command_link_en_obuf : OBUF
  port map (I => fe_command_link_en_out_p, O => fe_command_link_en_out);


inmem_ctrl_block : for i in 0 to 24 generate
  inmem_ctrl_obuf : OBUF
    port map (I => inmem_ctrl_out_p(i), O => inmem_ctrl_out(i));
end generate;

eventmem_ctrl_block : for i in 0 to 10 generate
  eventmem_ctrl_obuf : OBUF
    port map (I => eventmem_ctrl_out_p(i), O => eventmem_ctrl_out(i));
end generate;

mem_status_block : for i in 0 to  3 generate
  inmem_status_ibuf    : IBUF port map (I => inmem_status_in(i),    O => inmem_status_in_p(i));
  eventmem_status_ibuf : IBUF port map (I => eventmem_status_in(i), O => eventmem_status_in_p(i));
end generate;

ext_iacknum_block : for i in 0 to 4 generate
  ext_iacknum_ibuf : IBUF
    port map (I => ext_iacknum_in(i), O => ext_iacknum_in_p(i));
end generate;

bocbus_busy_ibuf : IBUF
  port map (I => bocbus_busy_in, O => bocbus_busy_in_p);

bocbus_strb_obuf : OBUF
  port map (I => bocbus_strb_out_p, O => bocbus_strb_out);

bocbus_rodoe_obuf : OBUF
  port map (I => bocbus_rodoe_out_p, O => bocbus_rodoe_out);

bocbus_write_obuf : OBUF
  port map (I => bocbus_write_out_p, O => bocbus_write_out);

boc_clock_error_ibuf : IBUF
  port map (I => boc_clock_ok_in, O => boc_clock_ok_p);

tim_clock_error_ibuf : IBUF
  port map (I => tim_clock_ok_in, O => tim_clock_ok_p);

sdsp_wr_n_obuf : OBUF_S_24
  port map (I => sdsp_wr_n_out_p, O => sdsp_wr_n_out);

sdsp_rd_n_obuf : OBUF_S_24
  port map (I => sdsp_rd_n_out_p, O => sdsp_rd_n_out);

fmt_hwob_obuf : OBUF
  port map (I => fmt_hwob_p, O => fmt_hwob_out);

fmt_ds_obuf : OBUF
  port map (I => fmt_ds_p, O => fmt_ds_out);

formA_ack_n_ibuf : IBUF
  port map (I => formA_ack_n_in, O => formA_ack_n_in_p);
  
formB_ack_n_ibuf : IBUF
  port map (I => formB_ack_n_in, O => formB_ack_n_in_p);

efb_ack_ibuf : IBUF
  port map (I => efb_ack_in, O => efb_ack_in_p);

router_ack_ibuf : IBUF
  port map (I => router_ack_in, O => router_ack_in_p);
  
dsp0_hrdy_ibuf : IBUF
  port map (I => dsp0_hrdy_in, O => dsp0_hrdy_in_p);

dsp1_hrdy_ibuf : IBUF
  port map (I => dsp1_hrdy_in, O => dsp1_hrdy_in_p);

dsp2_hrdy_ibuf : IBUF
  port map (I => dsp2_hrdy_in, O => dsp2_hrdy_in_p);

dsp3_hrdy_ibuf : IBUF
  port map (I => dsp3_hrdy_in, O => dsp3_hrdy_in_p);

l1_trigger_test_obuf : OBUF
  port map (I => tim_signal_ii(0), O => l1_trigger_test_out);
  
slink_we_n_ibuf : IBUF
  port map (I => slink_we_n_in, O => slink_we_n_p);

slink_lff_n_ibuf : IBUF
  port map (I => slink_lff_n_in, O => slink_lff_n_p);

slink_ldown_n_ibuf : IBUF
  port map (I => slink_ldown_n_in, O => slink_ldown_n_p);

dsp0_detect_n_ibuf : IBUF
port map (I => dsp0_detect_n_in, O => dsp0_detect_in_n_p);

dsp1_detect_n_ibuf : IBUF
port map (I => dsp1_detect_n_in, O => dsp1_detect_in_n_p);

dsp2_detect_n_ibuf : IBUF
port map (I => dsp2_detect_n_in, O => dsp2_detect_in_n_p);

dsp3_detect_n_ibuf : IBUF
port map (I => dsp3_detect_n_in, O => dsp3_detect_in_n_p);

-------------------------------------------------------------------------------
--  ROD Type Input
rod_type_ibuf : IBUF
  port map (I => rod_type_in, O => rod_type_p);

-------------------------------------------------------------------------------
--  RCF Spare Inputs
spare_in_block :
  for i in 0 to 31 generate 
    spare_ibufs : IOBUF
      port map (
      I  => spare_o(i),
      O  => spare_i(i),
      IO => spare_pin(i),
      T  => spare_t(i)
      );
  end generate;

-- spare(31) == SDSP3 Bus Request (FSR1)
-- spare(30) == SDSP2 Bus Request (FSR1)
-- spare(29) == SDSP1 Bus Request (FSR1)
-- spare(28) == SDSP0 Bus Request (FSR1)
spare_t(31 downto 28) <= (others => '1');  -- inputs

-- spare(27) == OUTPUT  -> ROL Test Block to Router
--spare_o(27) <=      rol_test_block_i(2) OR rol_test_block_i(1) OR rol_test_block_i(0) ;
--spare_t(27) <= NOT (rol_test_block_i(2) OR rol_test_block_i(1) OR rol_test_block_i(0));
spare_t(27) <= '1';  -- input

spare_t(26) <= '1';  -- intput
-- spare(25) == INPUT  -> EFB Clear ROD Busy
--spare_t(25) <= '1';  -- input
-- spare(25) == OUTPUT -> not delayed ecr as ecr announcement
spare_o(25) <= '0';
spare_t(25) <= NOT (event_counter_reset_i and spare_ctrl_reg_i(1)); --'0';  -- output

spare_t(24) <= '1';  -- input
efb_ready_p <= spare_i(24);

spare_o(23) <= efb_edm_data_out_p(0); -- JMJ 20131027
spare_t(23) <= '0';  -- output

-- spare(22) == OUTPUT -> EFB Half Word Select 
spare_t(22) <= '0';  -- output

-- spare( 5) == INPUT  -> PRM VME Cycle Active
-- spare( 4) == INPUT  -> PRM Clock Select Status
spare_t(21 downto  0) <= (others => '1');  -- inputs

-------------------------------------------------------------------------------
--
test_out_block :
  for i in 0 to 7 generate 
    test_obufs : OBUF
      port map (I => test_out_p(i), O => test_out(i));
  end generate;

--------------------------------------------------------------------------
-- PLACE HOLDER SIGNAL ATTACHMENTS
--------------------------------------------------------------------------

gnd_i <= '0';
vcc_i <= '1';

rst_in_i <= NOT rst_n_in_p;

tim_signal_ii(0) <= (l1_accept_i           OR int_tim_fifo_outdata_i(5)) AND NOT ecr_rod_busy_i;
--tim_signal_ii(1) <=  event_counter_reset_i OR int_tim_fifo_outdata_i(4);
tim_signal_ii(1) <=  fe_delayed_ecr_i      OR int_tim_fifo_outdata_i(4); -- from V58
tim_signal_ii(2) <= (bunch_counter_reset_i OR int_tim_fifo_outdata_i(3)) AND NOT ecr_rod_busy_i;
tim_signal_ii(3) <= (cal_strobe_i          OR int_tim_fifo_outdata_i(2)) AND NOT ecr_rod_busy_i;
tim_signal_ii(4) <= (serial_l1id_i         OR int_tim_fifo_outdata_i(1)) AND NOT ecr_rod_busy_i;
tim_signal_ii(5) <= (serial_eventtype_i    OR int_tim_fifo_outdata_i(0)) AND NOT ecr_rod_busy_i;

-- External Reset for the FIFOs in the Formatters and the EFB
--efb_edm_fifo_rst_n_out_p <= NOT spare_ctrl_reg_i(2);
--efb_edm_fifo_rst_n_out_p <= NOT (spare_ctrl_reg_i(2) OR (event_counter_reset_i AND spare_ctrl_reg_i(1)));

-- Send command streams to the Formatters for the Simulator
-- Redirected in V35f to allow TX of FE commands to Formatters
formatter_mb_fifo_rst_n_out_p(1) <= fmt_fe_cmd_i;
formatter_mb_fifo_rst_n_out_p(0) <= fmt_fe_cmd_i;
--formatter_mb_fifo_rst_n_out_p(1) <= NOT spare_ctrl_reg_i(1);
--formatter_mb_fifo_rst_n_out_p(0) <= NOT spare_ctrl_reg_i(0);

-- ROD Bus peripheral select Signals
regbus_en_n_i          <= rodbus_en_n_itr_i(0); -- Controller Internal Registers
focbus_en_n_i          <= rodbus_en_n_itr_i(1); -- FE Occupancy Counters
fcmbus_en_n_i          <= rodbus_en_n_itr_i(2); -- FE Command Mask LUT
edebus_en_n_i          <= rodbus_en_n_itr_i(3); -- EFB DynamicMask Encoder
frebus_en_n_i          <= rodbus_en_n_itr_i(4); -- Formatter Readout ModeBits Encoder
dfibus_inmem_en_n_i    <= rodbus_en_n_itr_i(5); -- Inmem Control
dfibus_eventmem_en_n_i <= rodbus_en_n_itr_i(6); -- Eventmem Control

controller_clkx_out_p(0) <= clk40_g;
controller_clkx_out_p(1) <= clk40_g;
fsx_en_n_obuft <= NOT spare_ctrl_reg_i(18);

fe_command_link_dir_out_p <= '0'; --NOT rrif_ctrl_reg_i(0); 
fe_command_link_en_out_p  <= '0'; --NOT rrif_ctrl_reg_i(0);

sdsp_wr_n_out_p <= sdsp_wr_n_out_i AND sdsp_rd_n_out_i;  -- jmj 12/11/2003
sdsp_rd_n_out_p <= '1';                                  -- jmj 12/11/2003

--------------------------------------------------------------------------
-- PROCESS DECLARATIONS
--------------------------------------------------------------------------

signal_register : process (
  rst_n_in_p, 
  clk40_g
  )
begin
  if (rst_n_in_p = '0') then
    debug_fifo_stts_i     <= (others => '0');
    dsp_detect_i          <= (others => '0');
    l1_accept_i           <= '0';
    event_counter_reset_i <= '0';
    fe_delayed_ecr_i      <= '0';
    sw_ecr_cmd_i          <= '0';
    sw_ecr_cmd_p          <= '0';
    bunch_counter_reset_i <= '0';
    cal_strobe_i          <= '0';
    serial_l1id_i         <= '0';
    serial_eventtype_i    <= '0';
    ttc6_i                <= '0';
    ttc7_i                <= '0';
    emif_re_i             <= '0';
    dsp_timer1            <= '0';
    slink_lff_i           <= '0';
    slink_ldown_i         <= '0';
    slink_we_i            <= '0';
    pause_triggers_i      <= '0';
    rol_test_pause_i      <= '0';
    rod_busy_n_out_p      <= '1';
    ecr_rod_busy_i        <= '0';
    rb_window_cnt         <= (others => '0');
    en_rb_window_cnt      <= '0';
    dropped_l1a_count     <= (others => '0');
    dropped_ecr_count     <= (others => '0');
    dropped_bcr_count     <= (others => '0');
    efb_edm_fifo_rst_n_out_p <= '1'; -- added rcf_4D
  elsif (clk40_g'event AND clk40_g = '1') then
-- rod_busy_in is Low True
--    rod_busy_n_out_p  <=     rod_busy_n_in_p(1)         AND 
--                             rod_busy_n_in_p(0)         AND 
--								       formatter_mb_fff_n_in_p(1) AND
--								       formatter_mb_fff_n_in_p(0) AND
--                         NOT efb_header_pause_i         AND 
--                         NOT set_rod_busy_i             AND  -- Manual ROD Busy from VME host
--                         NOT rol_test_pause_i           AND  -- Required when TIM issues RoL triggers
--                         NOT ecr_rod_busy_i;


    rod_busy_n_out_p   <= NOT pin_hold_reg_i( 1) AND  -- <= NOT rod_busy_n_in_p(0); -- ROD Busy from Formatters
                          NOT pin_hold_reg_i( 2) AND  -- <= NOT rod_busy_n_in_p(1); -- ROD Busy from Formatters
                          NOT pin_hold_reg_i( 3) AND  -- <=     efb_header_pause_i; -- ROD Busy from Event ID FIFOs
                          --NOT pin_hold_reg_i( 4) AND  -- <= rol_test_pause_i;       -- Required when TIM issues RoL triggers
                          --NOT pin_hold_reg_i( 5) AND  -- <= NOT formatter_mb_fef_n_in_p(0);  -- Formatter A pause from EFB
                          --NOT pin_hold_reg_i( 6) AND  -- <= NOT formatter_mb_fef_n_in_p(1);  -- Formatter B pause from EFB
                          --NOT pin_hold_reg_i( 7) AND  -- <=     pause_triggers_i;
                          NOT pin_hold_reg_i( 8) AND  -- <=     ecr_rod_busy_i;    --sweeper_event_on_rod_i;
                          --NOT pin_hold_reg_i( 9) AND  -- <= NOT header_trailer_limit_in_p(0);
                          --NOT pin_hold_reg_i(10) AND  -- <= NOT header_trailer_limit_in_p(1);
                          NOT pin_hold_reg_i(11) AND  -- <= NOT formatter_mb_fff_n_in_p(0);
                          NOT pin_hold_reg_i(12) AND  -- <= NOT formatter_mb_fff_n_in_p(1);
                          NOT pin_hold_reg_i(13);  -- <=     set_rod_busy_i;

                         --NOT (NOT formatter_mb_fef_n_in_p(0) AND spare_ctrl_reg_i(10)) AND -- Formatter A pause from EFB
                         --NOT (NOT formatter_mb_fef_n_in_p(1) AND spare_ctrl_reg_i(10)) AND -- Formatter B pause from EFB
                         --NOT sweeper_event_on_rod_i

    rol_test_pause_i <= (rol_test_block_i(0) OR rol_test_block_i(2)) AND 
                        (slink_lff_i OR rrif_stts_reg_i(16)); -- {rrif_stts_reg_i(16)==EFB Event ID FIFO Full}

    pause_triggers_i  <=   NOT rod_busy_n_in_p(1)  OR
                           NOT rod_busy_n_in_p(0)  OR
                               efb_header_pause_i  OR 
                               set_rod_busy_i      OR
                               rol_test_pause_i;
                          --(NOT formatter_mb_fef_n_in_p(0) AND spare_ctrl_reg_i(10)) OR  --Formatter A pause from EFB
                          --(NOT formatter_mb_fef_n_in_p(1) AND spare_ctrl_reg_i(10)) OR  --Formatter B pause from EFB

    debug_fifo_stts_i <= rol_trigs_on_i &
                         "000" &
                         NOT inmem_status_in_p(3) &
                         NOT inmem_status_in_p(2) &
                         NOT inmem_status_in_p(1) &
                         NOT inmem_status_in_p(0) ;

-- The TIM signals are active LOW!
    l1_accept_i           <= NOT l1_accept_n_in_p;
    event_counter_reset_i <= NOT event_counter_reset_n_in_p OR sw_ecr_cmd_p;
    bunch_counter_reset_i <= NOT bunch_counter_reset_n_in_p;
    cal_strobe_i          <= NOT cal_strobe_n_in_p;
    serial_l1id_i         <= NOT serial_l1id_n_in_p;
    serial_eventtype_i    <= NOT serial_eventtype_n_in_p;
    ttc6_i                <= NOT ttc6_n_in_p;
    ttc7_i                <= NOT ttc7_n_in_p;
    emif_re_i			        <= NOT controller_bus_re_n_in_p;
    dsp_timer1            <=     controller_tout_in_p(1);  
    slink_lff_i           <= NOT slink_lff_n_p;
    slink_ldown_i         <= NOT slink_ldown_n_p;
    slink_we_i            <= NOT slink_we_n_p;
    dsp_detect_i(0)       <= NOT dsp0_detect_in_n_p;
    dsp_detect_i(1)       <= NOT dsp1_detect_in_n_p;
    dsp_detect_i(2)       <= NOT dsp2_detect_in_n_p;
    dsp_detect_i(3)       <= NOT dsp3_detect_in_n_p;


-- Dropped Command Counters
    if (spare_ctrl_reg_i(21) = '1') then
      dropped_l1a_count <= (others => '0');
      dropped_ecr_count <= (others => '0');
      dropped_bcr_count <= (others => '0');
    elsif (cmd_proc_busy_i = '1') then
      dropped_l1a_count <= dropped_l1a_count + l1_accept_i;
      dropped_ecr_count <= dropped_ecr_count + event_counter_reset_i;
      dropped_bcr_count <= dropped_bcr_count + bunch_counter_reset_i;
    end if;
    
-- Pixel Command bit streams to front end modules.
--  l1accept:  "11101" 
--  bcr:       "101100001" 
--  ecr:       "101100010"
--  cal:       "101100100"
--  sync:      "101101000"
--    if (spare_ctrl_reg_i(23) = '1') then  --   if (spare_stts_reg_i(31) = '1') then  -- 0=>SCT, 1=>Pixel = '1'
    if (spare_ctrl_reg_i(23) = '1' AND spare_stts_reg_i(31) = '1') then  -- For Pixel Only 
      sw_ecr_cmd_i <= spare_ctrl_reg_i(0);
      if (spare_ctrl_reg_i(0) = '1' AND sw_ecr_cmd_i = '0') then
        sw_ecr_cmd_p <= '1';
      else
        sw_ecr_cmd_p <= '0';
      end if;        

      if (en_rb_window_cnt = '1') then
        rb_window_cnt  <= rb_window_cnt + 1;
      else
        rb_window_cnt <= (others => '0');
      end if;
      
      if (event_counter_reset_i = '1' AND rb_window_cnt = 0) then
        if (unsigned(ecr_busy_count_i) /= 0) then
          ecr_rod_busy_i <= '1';
        else
          ecr_rod_busy_i <= '0';
        end if;
      elsif (rb_window_cnt(17 downto 2) > unsigned(ecr_busy_count_i)) then
        ecr_rod_busy_i <= '0';
      end if;
      
      fe_delayed_ecr_i <= '0';
      if (event_counter_reset_i = '1' AND rb_window_cnt = 0) then
        efb_edm_fifo_rst_n_out_p <= '1';
        en_rb_window_cnt <= '1';

      elsif (rb_window_cnt = 19999) then  -- == 500us
        efb_edm_fifo_rst_n_out_p <= '0';
        fe_delayed_ecr_i <= '1';

      elsif (rb_window_cnt = 30000) then  -- == 750us
        efb_edm_fifo_rst_n_out_p <= '1';

--      elsif (rb_window_cnt > 65533 OR (rb_window_cnt > 30000 AND ecr_rod_busy_i = '0')) then
      elsif (rb_window_cnt > 30000 AND ecr_rod_busy_i = '0') then
        en_rb_window_cnt <= '0';
      end if;      
    else
      en_rb_window_cnt <= '0';
      fe_delayed_ecr_i <= event_counter_reset_i;
      efb_edm_fifo_rst_n_out_p <= NOT spare_ctrl_reg_i(2);
    end if;      
  end if;
end process signal_register;

------------------------------------------------------------------------------
sweeper_event_detect : process (
  rst_n_in_p, 
  clk40_g
  )
begin
  if (rst_n_in_p = '0') then
    sweeper_event_on_rod_i <= '0';
    efb_rod_busy_i <= '0';
  elsif (clk40_g'event AND clk40_g = '1') then
    efb_rod_busy_i <= spare_i(25);
  -- Sweeper Event detected by the RCF
    if ((new_ttype_valid_i = '1' AND ttype_bits_i(7 downto 0) = "00000111") OR
  -- EFB handshake to set sweeper event status on the rising edge
        (spare_i(25) = '1' AND efb_rod_busy_i = '0')) then
      sweeper_event_on_rod_i <= '1';
  -- EFB handshake to clear sweeper event status on the falling edge
  --  elsif (spare_i(25) = '0' AND efb_rod_busy_i = '1') then
    elsif (efb_rod_busy_i = '1') then
      sweeper_event_on_rod_i <= '0';
    end if;
  end if;
end process sweeper_event_detect;

------------------------------------------------------------------------------
led_output_test : process (
  rst_n_in_p, 
  clk40_g
  )
variable rst_start    : std_logic;
variable rst_counter  : unsigned(23 downto 0);
variable xoff_counter : unsigned(20 downto 0);
variable slon_counter : unsigned(20 downto 0);
variable bclk_counter : unsigned(26 downto 0);
begin
  if (rst_n_in_p = '0') then
    led_output_n_out_p <= (others => '0');
    xoff_counter       := (others => '0');                       
    slon_counter       := (others => '0');
    xoff_led_state     <= idle;                       
    slon_led_state     <= idle;
    rst_start          := '1';                       
  elsif (clk40_g'event AND clk40_g = '1') then
    if (rst_start = '1') then
      led_output_n_out_p <= (others => '0');
      if (rst_counter(23) = '1') then
        rst_start:= '0';
      else
        rst_counter := rst_counter + 1 ;
      end if;
    else 
      led_output_n_out_p(0) <= NOT controller_tout_in_p(0);     -- GRN
      led_output_n_out_p(1) <= controller_bus_ce_n_in_p;        -- YEL
      led_output_n_out_p(2) <= bocbus_busy_in_p;                -- GRN
      led_output_n_out_p(3) <= NOT controller_clkr_in_p(0) AND  -- YEL
                               NOT pin_hold_reg_i(31);	-- SDSP ADDRESS ERROR

      if (rol_test_block_i /= "0000") then
		    led_output_n_out_p(4) <= rol_led_n_i;
      else
        led_output_n_out_p(4) <= NOT controller_clkr_in_p(1) AND  -- GRN
                                     vme_bus_int_n_out_p;
      end if;

--    if (spare_i(4) = '1' OR boc_clock_ok_p = '0') then
      if (spare_i(4) = '1') then
        led_output_n_out_p(5) <= NOT bclk_counter(25) AND 
                                 NOT controller_fsr_in_p(0);    -- RED
        bclk_counter := bclk_counter + 1 ;
      else
        led_output_n_out_p(5) <= NOT controller_fsr_in_p(0);    -- RED
      end if;                    

      case xoff_led_state is   -- RED (BP from S-Link)
        when idle =>
          led_output_n_out_p(6) <= slink_lff_n_p;            
          xoff_counter          := (others => '0');                       
          if (slink_lff_n_p = '0') then
            xoff_led_state <= led_on;                       
          end if;
  
        when led_on =>
          led_output_n_out_p(6) <= '0';
          xoff_counter := xoff_counter + 1;
          if (xoff_counter(20) = '1') then
            xoff_led_state <= idle;                       
          end if;
        
        when others => led_output_n_out_p(6) <= slink_lff_n_p;
      end case;
     
      case slon_led_state is  -- GRN
        when idle =>
          led_output_n_out_p(7) <= slink_ldown_i;
          slon_counter := (others => '0');                       
          if (slink_ldown_i = '0' AND slink_we_i = '1') then
            slon_led_state <= led_off;                       
          end if;
  
        when led_off =>
          led_output_n_out_p(7) <= '1';                         
          slon_counter := slon_counter + 1;
          if (slon_counter(20) = '1') then
            slon_counter := (others => '0');                       
            slon_led_state <= led_on;                       
          end if;

        when led_on =>
          slon_counter := slon_counter + 1;
          if (std_logic_vector(slon_counter(20 downto 19)) = "11") then
            slon_led_state <= idle;                       
          elsif (std_logic_vector(slon_counter(20 downto 19)) = "01") then
            led_output_n_out_p(7) <= '1';
          else
            led_output_n_out_p(7) <= '0';                         
          end if;
        when others => led_output_n_out_p(7) <= slink_ldown_i;
      end case;
    end if;
  end if;
end process led_output_test;

------------------------------------------------------------------------------
-- Test Outputs
  test_out_p(7) <= fe_cmdpulse_formattersAB_out_p(0) AND fe_cmdpulse_formattersAB_out_p(1);
  test_out_p(6) <= vme_bus_int_n_out_p;
  test_out_p(5) <= cmd_cycle_done_i;
  test_out_p(4) <= controller_dx_in_p(1);
  test_out_p(3) <= controller_dx_in_p(0);
  test_out_p(2) <= efb_edm_we_n_out_p;
  test_out_p(1) <= efb_edm_fef_n_in_p; -- efb_dmask_ready
  test_out_p(0) <= tim_signal_ii(0);

------------------------------------------------------------------------------
rrif_stts_reg_bitmap : process (
  rst_n_in_p, 
  clk40_g
  )
begin
  if (rst_n_in_p = '0') then
  elsif (clk40_g'event AND clk40_g = '1') then
    rrif_stts_reg_i( 0) <= tim_clock_ok_p; 
    rrif_stts_reg_i( 1) <= boc_clock_ok_p;
    rrif_stts_reg_i( 2) <= bocbus_busy_in_p;
--  rrif_stts_reg_i( 3) <= rdbk_done_i OR trap_done_i;
--  rrif_stts_reg_i( 4) == Cal Test Ready
--  rrif_stts_reg_i( 5) == FE Trigger FIFO empty flag, connected to PortMap
--  rrif_stts_reg_i( 6) == FE Trigger FIFO full flag, connected to PortMap
    rrif_stts_reg_i( 7) <= NOT formatter_mb_fef_n_in_p(0);
    rrif_stts_reg_i( 8) <= NOT formatter_mb_fff_n_in_p(0);
    rrif_stts_reg_i( 9) <= NOT formatter_mb_fef_n_in_p(1);
    rrif_stts_reg_i(10) <= NOT formatter_mb_fff_n_in_p(1);
    rrif_stts_reg_i(11) <= NOT header_trailer_limit_in_p(0);
    rrif_stts_reg_i(12) <= NOT header_trailer_limit_in_p(1);
    rrif_stts_reg_i(13) <= NOT rod_busy_n_in_p(0);
    rrif_stts_reg_i(14) <= NOT rod_busy_n_in_p(1);
    rrif_stts_reg_i(15) <= NOT efb_edm_fef_n_in_p;
    rrif_stts_reg_i(16) <= NOT efb_edm_fff_n_in_p;
    rrif_stts_reg_i(17) <= efb_edm_fifo_empty_error_in_p;
    rrif_stts_reg_i(18) <= NOT eventmem_status_in_p(1); --Evtmem FIFO A EF
    rrif_stts_reg_i(19) <= NOT eventmem_status_in_p(0); --Evtmem FIFO A FF
    rrif_stts_reg_i(20) <= NOT eventmem_status_in_p(3); --Evtmem FIFO B EF
    rrif_stts_reg_i(21) <= NOT eventmem_status_in_p(2); --Evtmem FIFO B FF
--  rrif_stts_reg_i(22:29) == FE cmdpulse_count_out
--  rrif_stts_reg_i(30)    == FE Occupancy Count All Zero
--  rrif_stts_reg_i(31)    == Command Mask Ready 
  end if ;
end process;
  rrif_stts_reg_i( 3) <= rdbk_done_i OR trap_done_i;

spare_stts_reg_bitmap : process (
  rst_n_in_p, 
  clk40_g
  )
begin
  if (rst_n_in_p = '0') then
  elsif (clk40_g'event AND clk40_g = '1') then
--  spare_stts_reg_i(18 downto  8) == FE Trigger Occupancy
    spare_stts_reg_i(23 downto 19) <= ext_iacknum_in_p;
    spare_stts_reg_i(31)           <= rod_type_p;      -- 0=> SCT 1=>Pixel
  end if ;
end process;
  spare_stts_reg_i(0)  <= l1_accept_i;
  spare_stts_reg_i(1)  <= event_counter_reset_i;
  spare_stts_reg_i(2)  <= bunch_counter_reset_i;
  spare_stts_reg_i(3)  <= cal_strobe_i;
  spare_stts_reg_i(4)  <= serial_l1id_i;
  spare_stts_reg_i(5)  <= serial_eventtype_i;
  spare_stts_reg_i(6)  <= ttc6_i;
  spare_stts_reg_i(7)  <= ttc7_i;
  spare_stts_reg_i(24) <= rcf_dll_locked_i;
  spare_stts_reg_i(25) <= slink_lff_i;
  spare_stts_reg_i(26) <= slink_ldown_i;
  spare_stts_reg_i(30 downto 27) <= dsp_detect_i;

dsp_stts_reg_bitmap : process (
  rst_n_in_p, 
  clk40_g
  )
begin
  if (rst_n_in_p = '0') then
  elsif (clk40_g'event AND clk40_g = '1') then
    dsp_stts_reg_i( 0) <= dsp0_hrdy_in_p;
    dsp_stts_reg_i( 1) <= dsp1_hrdy_in_p;
    dsp_stts_reg_i( 2) <= dsp2_hrdy_in_p;
    dsp_stts_reg_i( 3) <= dsp3_hrdy_in_p;
    dsp_stts_reg_i( 5) <= mdsp_int_n_in_p;
    dsp_stts_reg_i( 6) <= mdsp_hcntl1_p;
    dsp_stts_reg_i( 7) <= mdsp_hcntl0_p;
    dsp_stts_reg_i( 8) <= mdsp_hhwil_p;
    dsp_stts_reg_i( 9) <= mdsp_hds1_n_p;
    dsp_stts_reg_i(10) <= mdsp_rnw_p AND mdsp_cen_n_p;
    dsp_stts_reg_i(31 downto 16) <= mdsp_hpi_data_in_p;
  end if ;
end process;
  dsp_stts_reg_i( 4) <= emif_re_i;
  dsp_stts_reg_i(11) <= dsp_timer1;

rcf_hrdy_n_reg : process (
  rst_n_in_p, 
  clk40_g
  )
begin
  if (rst_n_in_p = '0') then
    rcf_hrdy_n_p <= '1';
  elsif (clk40_g'event AND clk40_g = '1') then
    rcf_hrdy_n_p <= controller_bus_ack_out_p;
  end if ;
end process;

  mdsp_hpi_data_dir_t <= '1';
  mdsp_hpi_data_out_p <= (others => 'Z');
 
  test_stts_reg_i <= spare_i;

pin_hold_reg_bitmap : process (
  rst_n_in_p, 
  clk40_g
  )
begin
  if (rst_n_in_p = '0') then
  elsif (clk40_g'event AND clk40_g = '1') then
    pin_hold_reg_i( 0) <= NOT rod_busy_n_out_p;
    pin_hold_reg_i( 1) <= NOT rod_busy_n_in_p(0); -- ROD Busy from Formatters
    pin_hold_reg_i( 2) <= NOT rod_busy_n_in_p(1); -- ROD Busy from Formatters
    pin_hold_reg_i( 3) <=     efb_header_pause_i; -- ROD Busy from Event ID FIFOs
    pin_hold_reg_i( 4) <= rol_test_pause_i;       -- Required when TIM issues RoL triggers
    pin_hold_reg_i( 5) <= NOT formatter_mb_fef_n_in_p(0);  -- Formatter A pause from EFB
    pin_hold_reg_i( 6) <= NOT formatter_mb_fef_n_in_p(1);  -- Formatter B pause from EFB
    pin_hold_reg_i( 7) <=     pause_triggers_i;
    pin_hold_reg_i( 8) <=     ecr_rod_busy_i;    --sweeper_event_on_rod_i;
    pin_hold_reg_i( 9) <= NOT header_trailer_limit_in_p(0);
    pin_hold_reg_i(10) <= NOT header_trailer_limit_in_p(1);
    pin_hold_reg_i(11) <= NOT formatter_mb_fff_n_in_p(0);
    pin_hold_reg_i(12) <= NOT formatter_mb_fff_n_in_p(1);
    pin_hold_reg_i(13) <=     set_rod_busy_i;
    pin_hold_reg_i(27 downto 14) <= (others => '0');
    pin_hold_reg_i(28) <= formA_ack_n_in_p;
    pin_hold_reg_i(29) <= formB_ack_n_in_p;
--  pin_hold_reg_i(30) == modebits_error
--  pin_hold_reg_i(31) == sdsp_address_error

  end if ;
end process;
--  pin_hold_reg_i( 3) <= efb_header_pause_i;     -- ROD Busy from Event ID FIFOs
--  pin_hold_reg_i( 4) <= rol_test_pause_i;       -- Required when TIM issues RoL triggers
--  pin_hold_reg_i( 7) <= pause_triggers_i;
--  pin_hold_reg_i( 8) <= sweeper_event_on_rod_i;
--  pin_hold_reg_i(11) <= '0';
--  pin_hold_reg_i(23 downto 12) <= pin_hold_reg_i(11 downto 0);
--  pin_hold_reg_i(27 downto 24) <= (others => '0');
--  pin_hold_reg_i(30) == modebits_error
--  pin_hold_reg_i(31) == sdsp_address_error

------------------------------------------------------------------------------
ecrid_counter : process (
  rst_n_in_p, 
  clk40_g
  )
begin
  if (rst_n_in_p = '0') then
    ecr_id <= (others => '0');
  elsif (clk40_g'event AND clk40_g = '1') then
    if (spare_ctrl_reg_i(12) = '1') then
      ecr_id <= UNSIGNED(ecr_count_i);
    elsif (tim_signal_ii(1) = '1') then
      ecr_id <= ecr_id + 1;
    end if;
  end if;
end process ecrid_counter;

------------------------------------------------------------------------------
-- Process to drive the DSP bus when required
-- '0' : tbuf active  --  '1' : tbuf high Z
-- only drive the tbuf outputs when the OE is active! 
controller_bus_tbuf_driver : process (
  rst_n_in_p, 
  clk40_g
  )
begin
  if (rst_n_in_p = '0') then
    controller_bus_tbuf_driver_i <= '1';
  elsif (clk40_g'event AND clk40_g = '1') then
    controller_bus_tbuf_driver_i <= controller_bus_oe_n_in_p OR
                                    controller_bus_ce_n_in_p; 
  end if;
end process;
                                                                        
------------------------------------------------------------------------------
dsp_emif_ARDY_driver : process (
  rst_n_in_p,
  controller_bus_ack_out_i, 
  controller_bus_ce_n_in_p
  )
begin
  if (rst_n_in_p = '0') then
    controller_bus_ack_out_p <= '1';
  elsif (controller_bus_ce_n_in_p = '0') then
    controller_bus_ack_out_p <= controller_bus_ack_out_i;
  else
    controller_bus_ack_out_p <= '1';
  end if;
end process dsp_emif_ARDY_driver; 
  
--------------------------------------------------------------------------------
--
reset_command_io : process (
  clk40_g, 
  rst_n_in_p
  )
begin
  if (rst_n_in_p = '0') then
    controller_dr_out_p  <= (others => '0');
    reset_commands_out_p <= (others => '0');
  elsif (clk40_g'event AND clk40_g = '1') then
    controller_dr_out_p  <= reset_commands_in_p;
  -- MDSP to PRM Reset functions
    reset_commands_out_p(0) <= controller_fsr_in_p(1);
  -- Command to trigger ROD Busy histogram capture in the PRM    
    reset_commands_out_p(1) <= cal_l1a0_i OR cal_l1a1_i OR tim_signal_ii(0);
  end if;
end process;

-------------------------------------------------------------------------------
-- Clock and Frame Sync for SP0
-- fsx0_count : unsigned(4 downto 0);
msdsp_fsx_out_sp0 : process (
  clk40_g, 
  rst_n_in_p
  )
begin
  if (rst_n_in_p = '0') then
    controller_fsx_out_p(0) <= '0';
    fsx0_count <= (others => '0');
  elsif (clk40_g'event AND clk40_g = '1') then
    if (fsx0_count = 0) then
      controller_fsx_out_p(0) <= '1';
    else
      controller_fsx_out_p(0) <= '0';
    end if;
    fsx0_count <= fsx0_count + 1;
  end if;
end process;

-- Clock and Frame Sync for SP1
-- fsx1_count : unsigned(4 downto 0);
msdsp_fsx_out_sp1 : process (
  clk40_g, 
  rst_n_in_p
  )
begin
  if (rst_n_in_p = '0') then
    controller_fsx_out_p(1) <= '0';
    fsx1_count <= (others => '0');
  elsif (clk40_g'event AND clk40_g = '1') then
 -- Adjustable offset to fsx0
    if (fsx1_count = unsigned(spare_ctrl_reg_i(17 downto 13))) then
      controller_fsx_out_p(1) <= '1';
    else
      controller_fsx_out_p(1) <= '0';
    end if;
    fsx1_count <= fsx1_count + 1;
  end if;
end process;

--------------------------------------------------------------------------------
--
dsp_dx_mask : process (
  clk40_g, 
  rst_n_in_p
  )
begin
  if (rst_n_in_p = '0') then
    controller_dx_i <= (others => '0');
  elsif (clk40_g'event AND clk40_g = '1') then
    controller_dx_i(0) <= controller_dx_in_p(0) AND NOT spare_ctrl_reg_i(19);
    controller_dx_i(1) <= controller_dx_in_p(1) AND NOT spare_ctrl_reg_i(20);
  end if;
end process dsp_dx_mask;

--------------------------------------------------------------------------------
--
--------------------------------------------------------------------------------
--
interrupts_enable : process (
  clk40_g, 
  rst_n_in_p
  )
begin
  if (rst_n_in_p = '0') then
    controller_tinp_out_p         <= (others => '0');
    controller_ext_int_n_out_p    <= (others => '0');
    slave_ext_int_n_out_p         <= (others => '0');
    slave_dsp_ints_in_i           <= (others => '0');
  elsif (clk40_g'event AND clk40_g = '1') then
    controller_tinp_out_p(0)      <=                              spare_ctrl_reg_i(25);
    controller_tinp_out_p(1)      <=                              spare_ctrl_reg_i(25);
    controller_ext_int_n_out_p(0) <= apply_dynmask_mb_i(0)    AND spare_ctrl_reg_i(26);
-- interrupt to the master dsp to signify that a DEFAULT trigger has occurred
    controller_ext_int_n_out_p(1) <= apply_dynmask_mb_i(1)    AND spare_ctrl_reg_i(27);
-- interrupt to the master dsp to signify that a CORRECTIVE trigger has occurred
    controller_ext_int_n_out_p(2) <=                              spare_ctrl_reg_i(28);
    controller_ext_int_n_out_p(3) <=                              spare_ctrl_reg_i(29);
    controller_ext_int_n_out_p(4) <= slv_int_to_mdsp_i        AND spare_ctrl_reg_i(30);
    slave_ext_int_n_out_p(0)      <= slave_dsp_ints_out_i(0)  AND spare_ctrl_reg_i(30);
    slave_ext_int_n_out_p(1)      <= slave_dsp_ints_out_i(1)  AND spare_ctrl_reg_i(30);
    slave_ext_int_n_out_p(2)      <= slave_dsp_ints_out_i(2)  AND spare_ctrl_reg_i(30);
    slave_ext_int_n_out_p(3)      <= slave_dsp_ints_out_i(3)  AND spare_ctrl_reg_i(30);
    slave_dsp_ints_in_i(0)        <= slave_ext_int_in_p(0)    AND spare_ctrl_reg_i(30);
    slave_dsp_ints_in_i(1)        <= slave_ext_int_in_p(1)    AND spare_ctrl_reg_i(30);
    slave_dsp_ints_in_i(2)        <= slave_ext_int_in_p(2)    AND spare_ctrl_reg_i(30);
    slave_dsp_ints_in_i(3)        <= slave_ext_int_in_p(3)    AND spare_ctrl_reg_i(30);
    slave_dsp_ints_in_i(4)        <= NOT slave_hint_n_in_p(0) AND spare_ctrl_reg_i(30);
    slave_dsp_ints_in_i(5)        <= NOT slave_hint_n_in_p(1) AND spare_ctrl_reg_i(30);
    slave_dsp_ints_in_i(6)        <= NOT slave_hint_n_in_p(2) AND spare_ctrl_reg_i(30);
    slave_dsp_ints_in_i(7)        <= NOT slave_hint_n_in_p(3) AND spare_ctrl_reg_i(30);
  end if;
end process interrupts_enable;

--------------------------------------------------------------------------------
--  This is the first cut of the RCF Interrupt handler, and needs improvements
--------------------------------------------------------------------------------
vme_irq_vector : process (
  clk40_g, 
  rst_n_in_p
  )
begin
  if (rst_n_in_p = '0') then
    vme_bus_int_n_out_p <= '1';
    vme_irq_out_p       <= (others => '0');
  elsif (clk40_g'event AND clk40_g = '1') then
    if (spare_ctrl_reg_i(31) = '1') then
      case vme_int_state is
        when idle =>
          vme_bus_int_n_out_p <= '1';
          vme_irq_out_p <= (others => '0');
          if (send_int_test_i = '1') then
            vme_irq_out_p <= test_irq_i;
            vme_int_state <= issue_int;
          end if;
        when issue_int => 
          vme_bus_int_n_out_p <= '0';
          if (spare_i(5) = '1') then
            vme_int_state <= clear_int;
          end if;
        when clear_int  =>
          if (spare_i(5) = '0') then
            vme_bus_int_n_out_p <= '1';
            vme_int_state <= done;
          end if;
        when done =>
          vme_int_state <= idle;
          vme_bus_int_n_out_p <= '1';
          vme_irq_out_p <= (others => '0');
        when others     =>
          vme_bus_int_n_out_p <= '1';
          vme_int_state <= idle;
      end case;
    else
      vme_bus_int_n_out_p <= '1';
      vme_irq_out_p <= (others => '0');
    end if;
  end if;
end process vme_irq_vector;

------------------------------------------------------------------------------
--
rol_test_l1a_source : process (
  clk40_g, 
  rst_n_in_p
  )
variable wait_count : unsigned(10 downto 0);
variable trig_count : unsigned(23 downto 0);
begin
  if (rst_n_in_p = '0') then
    rol_l1a_i <= '0';
    rol_test <= idle;
    rol_led_n_i <= '1';
    rol_test_i <= '0';
    rol_trigs_on_i <= '0';
    cmd_bit_count <= (others => '0');
    wait_count := (others => '0');
    trig_count := (others => '0');
  elsif (clk40_g'event AND clk40_g = '1') then
    rol_test_block0_ii <= rol_test_block_i(0);
    case rol_test is
      when idle =>
        rol_l1a_i <= '0';
        rol_test_i <= rol_test_block_i(2);  -- Enable ROL Test Block for TIM Triggers
        cmd_bit_count <= (others => '0');
        wait_count := (others => '0');
        trig_count := (others => '0');
        if (
            (rol_test_block_i = "0001") OR   -- Send Multiple triggers
            (rol_test_block_i = "0011" AND rol_test_block0_ii  = '0') OR -- Send 1 trigger
            (rol_test_block_i = "1001" AND rol_test_block0_ii  = '0')    -- Send n triggers
             ) then
          rol_led_n_i <= '0';
          rol_trigs_on_i <= '1';
          rol_test <= setup_delay;
        end if;
 -- Delay is required to allow proper configuration and to keep the trigger 
 -- rate close to 100kHz
 -- rrif_stts_reg_i(16) == NOT efb_edm_fff_n_in_p
 -- slink_lff_i == Xoff from the Slink
 -- slink_ldown_i == LSlink Down
      when setup_delay =>
        rol_test_i <= '1';
        rol_led_n_i <= '0';
        if (pause_triggers_i = '0') then
          wait_count := wait_count + 1 ;
          if (wait_count = 220) then
            rol_test <= send_l1a;
            wait_count := (others => '0');
          end if;
        end if;
      when send_l1a =>
        if (spare_stts_reg_i(31) = '0') then --  l1accept: "110"  SCT
          case cmd_bit_count is
            when "000"  => rol_l1a_i <= '1';                 
            when "001"  => rol_l1a_i <= '1';                 
            when "010"  => rol_l1a_i <= '0';                 
            when others => rol_l1a_i <= '0';
                           cmd_bit_count <= (others => '0');
                           rol_test <= wait_for_cycle;
          end case;
          cmd_bit_count <= cmd_bit_count + '1';
        elsif (spare_stts_reg_i(31) = '1') then --  l1accept: "11101"  Pixel
          case cmd_bit_count is
            when "000"  => rol_l1a_i <= '1';                 
            when "001"  => rol_l1a_i <= '1';                 
            when "010"  => rol_l1a_i <= '1';                 
            when "011"  => rol_l1a_i <= '0';                 
            when "100"  => rol_l1a_i <= '1';                 
            when "101"  => rol_l1a_i <= '0';
            when others => rol_l1a_i <= '0';
                           cmd_bit_count <= (others => '0');
                           rol_test <= wait_for_cycle;
          end case;
          cmd_bit_count <= cmd_bit_count + '1';
        end if;
      when wait_for_cycle =>
        if (cmd_cycle_done_i = '1') then
          rol_led_n_i <= '1';
          if (rol_test_block_i(3) = '1') then
            if (trig_count = unsigned(dbgmem_wordcount_i(23 downto 0))) then
              rol_trigs_on_i <= '0';
              rol_test <= idle;
            else
              cmd_bit_count <= (others => '0');
              trig_count := trig_count + 1 ;
              wait_count := (others => '0');
              rol_test <= setup_delay;
            end if;  
          else
            rol_trigs_on_i <= '0';
            rol_test <= idle;
          end if; 
        end if; 
      when others =>
        rol_l1a_i <= '0';
        rol_test <= idle;
        rol_led_n_i <= '1';
        cmd_bit_count <= (others => '0');
    end case;
  end if;
end process;

------------------------------------------------------------------------------
--
-- Switch control over the debug memories when instructed to do so by the 
-- rrif_control_register signals below.
--      25: debugmem_fifo_interface control(1) : 1=enable, 0=disable/clear
--      26: testbench_interface                : 1=enable, 0=disable/clear
--      27: normal data path                   : 1=enable, 0=disable/clear
-- So: Whenever the normal datapath is enabled, this process excludes control 
-- signals from the debugmem_fifo_interface and the testbench_interface from 
-- reaching the debug fifo memories
--
-- In the debugmem_fifo_interface uses a modified internal representation of 
-- the control leads and are defined and used as follows:
-- 
-- inmem_ctrl_i(15 downto 0)
-- inmem_ctrl_i( 0)      : inmem_to_fifo_a_sel_n
-- inmem_ctrl_i( 1)      : inmem_to_fifo_a_ldstrb
-- inmem_ctrl_i( 2)      : inmem_to_fifo_a_udstrb
-- inmem_ctrl_i( 3)      : inmem_to_fifo_b_sel_n
-- inmem_ctrl_i( 4)      : inmem_to_fifo_b_ldstrb
-- inmem_ctrl_i( 5)      : inmem_to_fifo_b_udstrb
-- inmem_ctrl_i( 6)      : fifo_a_to_inmem_ldsel_n
-- inmem_ctrl_i( 7)      : fifo_a_to_inmem_udsel_n
-- inmem_ctrl_i( 8)      : fifo_a_to_inmem_strb
-- inmem_ctrl_i( 9)      : fifo_b_to_inmem_ldsel_n
-- inmem_ctrl_i(10)      : fifo_b_to_inmem_udsel_n
-- inmem_ctrl_i(11)      : fifo_b_to_inmem_strb
-- inmem_ctrl_i(12)      : inmem_fifo_a_we_n
-- inmem_ctrl_i(13)      : inmem_fifo_a_re_n
-- inmem_ctrl_i(14)      : inmem_fifo_b_we_n
-- inmem_ctrl_i(15)      : inmem_fifo_b_re_n
--
-- eventmem_ctrl_i(10 downto 0)
-- eventmem_ctrl_out_i( 0) == efb_to_eventmem_strb
-- eventmem_ctrl_out_i( 1) == efb_to_eventmem_ldsel_n
-- eventmem_ctrl_out_i( 2) == efb_to_eventmem_udsel_n
-- eventmem_ctrl_out_i( 3) == efb_a_re_n
-- eventmem_ctrl_out_i( 4) == efb_a_oe_n
-- eventmem_ctrl_out_i( 5) == efb_a_en_n
-- eventmem_ctrl_out_i( 6) == efb_b_re_n
-- eventmem_ctrl_out_i( 7) == efb_b_oe_n
-- eventmem_ctrl_out_i( 8) == efb_b_en_n
-- eventmem_ctrl_out_i( 9) == efb_htf_re_n
-- eventmem_ctrl_out_i(10) == efb_htf_en_n
--
--
--  THEY MUST be remapped here to the ROD board specification: 
-- Actual ROD Output MAP
-- inmem_ctrl_out_p( 0) == detector_to_formA_sel_n
-- inmem_ctrl_out_p(24) == detector_to_formB_sel_n
-- inmem_ctrl_out_p( 1) == detector_to_fifo_a_sel_n 
-- inmem_ctrl_out_p( 2) == detector_to_fifo_b_sel_n
-- inmem_ctrl_out_p( 3) == fifo_a_to_detector_sel_n
-- inmem_ctrl_out_p( 4) == fifo_b_to_detector_sel_n
-- inmem_ctrl_out_p( 5) == inmem_to_fifo_a_sel_n 
-- inmem_ctrl_out_p( 6) == inmem_to_fifo_a_ldstrb
-- inmem_ctrl_out_p( 7) == inmem_to_fifo_a_udstrb
-- inmem_ctrl_out_p( 8) == inmem_to_fifo_b_sel_n 
-- inmem_ctrl_out_p( 9) == inmem_to_fifo_a_ldstrb
-- inmem_ctrl_out_p(10) == inmem_to_fifo_a_udstrb
-- inmem_ctrl_out_p(11) == fifo_a_to_inmem_ldsel_n 
-- inmem_ctrl_out_p(12) == fifo_a_to_inmem_udsel_n 
-- inmem_ctrl_out_p(13) == fifo_a_to_inmem_strb    
-- inmem_ctrl_out_p(14) == fifo_b_to_inmem_ldsel_n 
-- inmem_ctrl_out_p(15) == fifo_b_to_inmem_udsel_n 
-- inmem_ctrl_out_p(16) == fifo_b_to_inmem_strb    
-- inmem_ctrl_out_p(17) == inmem_fifo_a_wen_n 
-- inmem_ctrl_out_p(18) == inmem_fifo_a_ren_n
-- inmem_ctrl_out_p(19) == inmem_fifo_b_wen_n 
-- inmem_ctrl_out_p(20) == inmem_fifo_b_ren_n 
-- inmem_ctrl_out_p(21) == inmem_rst_n
-- inmem_ctrl_out_p(22) == inmem_a_rt 
-- inmem_ctrl_out_p(23) == inmem_b_rt 
--
-- eventmem_ctrl_out_p(10 downto  0) <= eventmem_ctrl_i(10 downto  0);
-- eventmem_ctrl_out_p( 0) == efb_to_eventmem_strb
-- eventmem_ctrl_out_p( 1) == efb_to_eventmem_ldsel_n
-- eventmem_ctrl_out_p( 2) == efb_to_eventmem_udsel_n
-- eventmem_ctrl_out_p( 3) == efb_a_re_n
-- eventmem_ctrl_out_p( 4) == efb_a_oe_n
-- eventmem_ctrl_out_p( 5) == efb_a_en_n
-- eventmem_ctrl_out_p( 6) == efb_b_re_n
-- eventmem_ctrl_out_p( 7) == efb_b_oe_n
-- eventmem_ctrl_out_p( 8) == efb_b_en_n
-- eventmem_ctrl_out_p( 9) == efb_htf_re_n
-- eventmem_ctrl_out_p(10) == efb_htf_en_n
--
-- Test Bench Interface Mapping
-- TB inmem_ctrl_out( 0) == detector_to_form_sel_n
-- TB inmem_ctrl_out( 1) == detector_to_form_sel_n
-- TB inmem_ctrl_out( 2) == detector_to_fifo_a_sel_n
-- TB inmem_ctrl_out( 3) == detector_to_fifo_b_sel_n
-- TB inmem_ctrl_out( 4) == fifo_a_to_detector_sel_n
-- TB inmem_ctrl_out( 5) == fifo_b_to_detector_sel_n)
-- TB inmem_ctrl_out( 6) == inmem_fifo_a_wen_n
-- TB inmem_ctrl_out( 7) == inmem_fifo_a_ren_n
-- TB inmem_ctrl_out( 8) == inmem_fifo_b_wen_n
-- TB inmem_ctrl_out( 9) == inmem_fifo_b_ren_n
-- TB inmem_ctrl_out(10) == inmem_a_rt
-- TB inmem_ctrl_out(11) == inmem_b_rt
--
-- When switching between debug modes, the reset lines and other lines 
-- glitch from the testbench_interface.  It's not clear why this is so, since 
-- the signals are driven high in all of the idle cases, and the default 
-- "other" cases.
-- These register bits are reset to '1' in the register_block.vhd
inmem_ctrl_out_p(21)   <= NOT spare_ctrl_reg_i(3);
--
-- Map inmem control lines
  inmem_ctrl_out_p(24)           <= inmem_ctrl_out_i(24);
  inmem_ctrl_out_p(23 downto 22) <= inmem_ctrl_out_i(23 downto 22);
--inmem_ctrl_out_p(21) <==> FIFO Reset Register
--inmem_ctrl_out_p(20) <==> inmem_rd_shift process
  inmem_ctrl_out_p(19)           <= inmem_ctrl_out_i(19);
--inmem_ctrl_out_p(18) <==> inmem_rd_shift process
  inmem_ctrl_out_p(17 downto  1) <= inmem_ctrl_out_i(17 downto  1);
  inmem_ctrl_out_p( 0)           <= inmem_ctrl_out_i(0);

inmem_rd_shift : process ( -- The RD En strobes to the Input FIFOs must be 
  nclk40_g,                -- shifted by 12.5ns to guarantee that the FIFO
  rst_n_in_p               -- RCLK falls near the start of the REN STRB
  )
begin
  if (rst_n_in_p = '0') then
    inmem_ctrl_out_p(20) <= '1';   -- inmem_fifo_b_ren_n 
    inmem_ctrl_out_p(18) <= '1';   -- inmem_fifo_a_ren_n
  elsif (nclk40_g'event AND nclk40_g = '1') then
    inmem_ctrl_out_p(20) <= inmem_ctrl_out_i(20);
    inmem_ctrl_out_p(18) <= inmem_ctrl_out_i(18);
  end if;
end process;

control_signal_mux : process (
  clk40_g,
  rst_n_in_p
  )
begin
  if (rst_n_in_p = '0') then
    inmem_ctrl_out_i( 0)              <= '1';  -- detector_to_formA_sel_n
    inmem_ctrl_out_i(24)              <= '1';  -- detector_to_formB_sel_n 
    inmem_ctrl_out_i(20 downto 1)     <= (others => '1');
    inmem_ctrl_out_i(23 downto 22)    <= (others => '0');
    eventmem_ctrl_out_p(10 downto  0) <= (others => '1');
  elsif (clk40_g'event AND clk40_g = '1') then
    case rrif_ctrl_reg_i(27 downto 25) is
-- Connect the Debugmem Fifo Interface Control Signals to the Debug FIFOs
-- This mode is used to Load and Read the FIFOs with DSP control
      when "001" =>
        inmem_ctrl_out_i( 4 downto  0)    <= (others => '1');
        inmem_ctrl_out_i(24)              <= '1';  
        inmem_ctrl_out_i(20 downto  5)    <= inmem_ctrl_i(15 downto 0);
        inmem_ctrl_out_i(23 downto 22)    <= (others => '0');
        eventmem_ctrl_out_p               <= eventmem_ctrl_i;

--  Connect to the Test Bench Interface Control Signals
      when "010" =>
        inmem_ctrl_out_i( 0)              <= tb_inmem_ctrl_out_i( 0);
        inmem_ctrl_out_i(24)              <= tb_inmem_ctrl_out_i( 1);
        inmem_ctrl_out_i( 4 downto  1)    <= tb_inmem_ctrl_out_i( 5 downto  2);
        inmem_ctrl_out_i(16 downto  5)    <= (others => '1');
        inmem_ctrl_out_i(20 downto 17)    <= tb_inmem_ctrl_out_i( 9 downto  6);
        inmem_ctrl_out_i(22)              <= tb_inmem_ctrl_out_i(10) OR spare_ctrl_reg_i(9);
        inmem_ctrl_out_i(23)              <= tb_inmem_ctrl_out_i(11) OR spare_ctrl_reg_i(9);
        eventmem_ctrl_out_p(10 downto  0) <= (others => '1');

--  Enable the normal data Path
      when "100" =>
        inmem_ctrl_out_i( 0)              <= '0';  -- detector_to_formA_sel_n
        inmem_ctrl_out_i(24)              <= '0';  -- detector_to_formB_sel_n 
  -- fifo_a_to_detector_sel_n (this signal is now detector_to_formA_sel_n)
  -- fifo_b_to_detector_sel_n (this signal is now detector_to_formB_sel_n)
        inmem_ctrl_out_i( 4 downto  3)    <= (others => '0');
        inmem_ctrl_out_i( 2 downto  1)    <= (others => '1');
        inmem_ctrl_out_i(20 downto  5)    <= (others => '1');
        inmem_ctrl_out_i(23 downto 22)    <= (others => '0');
        eventmem_ctrl_out_p(10 downto  0) <= (others => '1');

-- For Configuration Readback Mode
      when "101" =>
        inmem_ctrl_out_i( 0) <= '1'; -- detector_to_formA_sel_n
        inmem_ctrl_out_i(24) <= '1'; -- detector_to_formB_sel_n
        inmem_ctrl_out_i( 1) <= '0'; -- detector_to_fifo_a_sel_n 
        inmem_ctrl_out_i( 2) <= '0'; -- detector_to_fifo_b_sel_n
        inmem_ctrl_out_i( 3) <= '1'; -- fifo_a_to_detector_sel_n
        inmem_ctrl_out_i( 4) <= '1'; -- fifo_b_to_detector_sel_n
      -- Block Write Functions
        inmem_ctrl_out_i(10 downto 5) <= (others => '1');
      -- Allow Readback
        inmem_ctrl_out_i(11) <= inmem_ctrl_i( 6); -- fifo_a_to_inmem_ldsel_n 
        inmem_ctrl_out_i(12) <= inmem_ctrl_i( 7); -- fifo_a_to_inmem_udsel_n 
        inmem_ctrl_out_i(13) <= inmem_ctrl_i( 8); -- fifo_a_to_inmem_strb    
        inmem_ctrl_out_i(14) <= inmem_ctrl_i( 9); -- fifo_b_to_inmem_ldsel_n 
        inmem_ctrl_out_i(15) <= inmem_ctrl_i(10); -- fifo_b_to_inmem_udsel_n 
        inmem_ctrl_out_i(16) <= inmem_ctrl_i(11); -- fifo_b_to_inmem_strb    
      -- WEN/REN Control
        inmem_ctrl_out_i(17) <= NOT wr_cnfg_data_inmemA_i; -- inmem_fifo_a_wen_n 
        inmem_ctrl_out_i(18) <= inmem_ctrl_i(13);          -- inmem_fifo_a_ren_n
        inmem_ctrl_out_i(19) <= NOT wr_cnfg_data_inmemB_i; -- inmem_fifo_b_wen_n 
        inmem_ctrl_out_i(20) <= inmem_ctrl_i(15);          -- inmem_fifo_b_ren_n 
      -- Disable Re-Tx
        inmem_ctrl_out_i(23 downto 22) <= (others => '0');
        eventmem_ctrl_out_p(10 downto  0) <= (others => '1');

-- For Calibration Test Mode
      when "110" =>
   --  Turn off all inputs
        inmem_ctrl_out_i( 0) <= '1'; -- detector_to_formA_sel_n
        inmem_ctrl_out_i(24) <= '1'; -- detector_to_formB_sel_n
        inmem_ctrl_out_i(17 downto 1) <= (others => '1');
        inmem_ctrl_out_i(18) <= NOT rd_data_inmemA_i; -- inmem_fifo_a_ren_n
        inmem_ctrl_out_i(19) <= '1'; 
        inmem_ctrl_out_i(20) <= NOT rd_data_inmemB_i; -- inmem_fifo_b_ren_n 
        inmem_ctrl_out_i(22) <= rt_data_inmemA_i;     -- inmem_fifo_a_rt
        inmem_ctrl_out_i(23) <= rt_data_inmemB_i;     -- inmem_fifo_b_rt
        eventmem_ctrl_out_p(10 downto  0) <= (others => '1');

--  Route the input link data to the InMems and Formatters
--  Route the output of the Formatters to the DebugMems and the EFB
      when "111" =>
      -- Protect Input Drivers when Reading the Inmem FIFOs
      --  Read operation:  inmem_ctrl_i( 8/11): ---  ---
      --                                           __
      --                   inmem_ctrl_i(13/15): --- ----
      --                                           _    
      --
        inmem_ctrl_out_i( 0) <= NOT inmem_ctrl_i( 8); -- detector_to_formA_sel_n
        inmem_ctrl_out_i(24) <= NOT inmem_ctrl_i(11); -- detector_to_formB_sel_n
        inmem_ctrl_out_i( 1) <= '0';                  -- detector_to_fifo_a_sel_n 
        inmem_ctrl_out_i( 2) <= '0';                  -- detector_to_fifo_b_sel_n
        inmem_ctrl_out_i( 3) <= NOT inmem_ctrl_i( 8); -- detector_to_formA_sel_n
        inmem_ctrl_out_i( 4) <= NOT inmem_ctrl_i(11); -- detector_to_formB_sel_n
      -- Block Write Functions
        inmem_ctrl_out_i(10 downto 5) <= (others => '1');
      -- Allow Readback
        inmem_ctrl_out_i(16 downto 11) <= inmem_ctrl_i(11 downto 6); 
      -- WEN/REN Control
        inmem_ctrl_out_i(17) <= NOT wr_link_data_inmemA_i; -- inmem_fifo_a_wen_n 
        inmem_ctrl_out_i(18) <=     inmem_ctrl_i(13);      -- inmem_fifo_a_ren_n
        inmem_ctrl_out_i(19) <= NOT wr_link_data_inmemB_i; -- inmem_fifo_b_wen_n 
        inmem_ctrl_out_i(20) <=     inmem_ctrl_i(15);      -- inmem_fifo_b_ren_n 
      -- Disable Re-Tx
        inmem_ctrl_out_i(23 downto 22) <= (others => '0');
      -- Allow ReadBack
        eventmem_ctrl_out_p <= eventmem_ctrl_i;

-- Disable the entire datapath
      when others =>
        inmem_ctrl_out_i(24)              <= '1';
        inmem_ctrl_out_i(23 downto 22)    <= (others => '0');
        inmem_ctrl_out_i(20 downto  0)    <= (others => '1');
        eventmem_ctrl_out_p(10 downto  0) <= (others => '1');
    end case;
  end if;
end process control_signal_mux;

--------------------------------------------------------------------------
-- Process to allow Configuration Readback of Modules.
-- To setup a read back:
-- Set rrif_ctrl_reg_i(27 downto 25) = "101"
-- Write a delay time value to Register Location 0x00404468 Bits[23:16]
-- Write a Fill time value to Register Location 0x00404468 Bits[15:0]
-- Set rrif_ctrl_reg_i(19) = 1
-- Send the Commands to the Module (tranSerialData)
-- When the readBackDone Flag is set Read Data from FIFOs
-- Clear rrif_ctrl_reg_i(19) = 0
--
config_readback : process (
  clk40_g, 
  rst_n_in_p, 
  cnfg_rdbk_state,
  cnfg_rdback_count_i, 
  rrif_ctrl_reg_i
  )
  variable start_delay_count : unsigned(15 downto 0);
  variable start_write_count : unsigned(15 downto 0);
begin
  if (rst_n_in_p = '0') then
    wr_cnfg_data_inmemA_i <= '0';
    wr_cnfg_data_inmemB_i <= '0';
    rdbk_done_i           <= '0';
    start_delay_count     := (others => '0');
    start_write_count     := (others => '0');
    cnfg_rdbk_state       <= reset;
  elsif (clk40_g'event AND clk40_g = '1') then
    if (rrif_ctrl_reg_i(27 downto 25) = "101") then
      case cnfg_rdbk_state is
        when reset =>
          wr_cnfg_data_inmemA_i <= '0';
          wr_cnfg_data_inmemB_i <= '0';
          rdbk_done_i           <= '0';
          start_delay_count     := (others => '0');
          start_write_count     := (others => '0');
          cnfg_rdbk_state       <= idle;
        when idle  =>
          wr_cnfg_data_inmemA_i <= '0';
          wr_cnfg_data_inmemB_i <= '0';
          rdbk_done_i           <= '0';
          start_delay_count     := (others => '0');
          start_write_count     := (others => '0');
          if (rrif_ctrl_reg_i(19) = '1') then
            cnfg_rdbk_state <= start_delay;
          end if;
        when start_delay =>
          wr_cnfg_data_inmemA_i <= '0';
          wr_cnfg_data_inmemB_i <= '0';
          rdbk_done_i           <= '0';
          if (start_delay_count = unsigned(cnfg_rdback_count_i(31 downto 16))) then
            cnfg_rdbk_state <= wr_to_fifo;
          else
            start_delay_count := start_delay_count + 1;
          end if;
        when wr_to_fifo =>
          wr_cnfg_data_inmemA_i <= '1';
          wr_cnfg_data_inmemB_i <= '1';
          rdbk_done_i           <= '0';
          if (start_write_count = unsigned(cnfg_rdback_count_i(15 downto 0))) then
            wr_cnfg_data_inmemA_i <= '0';
            wr_cnfg_data_inmemB_i <= '0';
            cnfg_rdbk_state       <= set_done_flag;
          else
            start_write_count := start_write_count + 1;
          end if;
        when set_done_flag =>
          wr_cnfg_data_inmemA_i <= '0';
          wr_cnfg_data_inmemB_i <= '0';
          rdbk_done_i           <= '1';
          start_delay_count     := (others => '0');
          start_write_count     := (others => '0');
          cnfg_rdbk_state       <= return_to_idle;
        when return_to_idle =>
          wr_cnfg_data_inmemA_i <= '0';
          wr_cnfg_data_inmemB_i <= '0';
          rdbk_done_i           <= '1';
          start_delay_count     := (others => '0');
          start_write_count     := (others => '0');
          if (rrif_ctrl_reg_i(19) = '0') then
            cnfg_rdbk_state <= idle;
          end if;
        when others => 
          wr_cnfg_data_inmemA_i <= '0';
          wr_cnfg_data_inmemB_i <= '0';
          rdbk_done_i           <= '0';
          start_delay_count     := (others => '0');
          start_write_count     := (others => '0');
          cnfg_rdbk_state       <= idle;
      end case;
    else
      wr_cnfg_data_inmemA_i <= '0';
      wr_cnfg_data_inmemB_i <= '0';
      rdbk_done_i           <= '0';
      start_delay_count     := (others => '0');
      start_write_count     := (others => '0');
      cnfg_rdbk_state       <= reset;
    end if;
  end if;
end process config_readback;


--------------------------------------------------------------------------
-- Process to allow Calibration Test using Input Memory FIFOs.
-- To setup a Cal Test:
-- Set rrif_ctrl_reg_i(27 downto 25) = "110"
-- Write a delay time value to Register Location 0x00404468 Bits[31:16]
-- Write a play time value to Register Location 0x00404468 Bits[15:0]
-- Send L1T on SP0 or SP1
cal_data_test : process (
  clk40_g, 
  rst_n_in_p
  )
variable cal_test_count : unsigned(16 downto 0);
begin
  if (rst_n_in_p = '0') then
    rd_data_inmemA_i   <= '0';
    rd_data_inmemB_i   <= '0';
    rt_data_inmemA_i   <= '0';
    rt_data_inmemB_i   <= '0';
    rrif_stts_reg_i(4) <= '0';
    cal_test_count := (others => '0');
    cal_test_state <= reset;
  elsif (clk40_g'event AND clk40_g = '1') then
    if (rrif_ctrl_reg_i(27 downto 25) = "110") then
      case cal_test_state is
        when reset =>
          rd_data_inmemA_i   <= '0';
          rd_data_inmemB_i   <= '0';
          rt_data_inmemA_i   <= '0';
          rt_data_inmemB_i   <= '0';
          rrif_stts_reg_i(4) <= '0';
          cal_test_state     <= idle;
        when idle  =>
          rd_data_inmemA_i   <= '0';
          rd_data_inmemB_i   <= '0';
          rt_data_inmemA_i   <= '0';
          rt_data_inmemB_i   <= '0';
          rrif_stts_reg_i(4) <= NOT rrif_ctrl_reg_i(24);
          if (rrif_ctrl_reg_i(24) = '0' AND (cal_l1a0_i = '1' OR cal_l1a1_i = '1')) then
            cal_test_count := '0' & unsigned(cnfg_rdback_count_i(31 downto 16)); -- Load delay count
            cal_test_state <= start_delay;
          end if;
        when start_delay =>
          rd_data_inmemA_i   <= '0';
          rd_data_inmemB_i   <= '0';
          rt_data_inmemA_i   <= '0';
          rt_data_inmemB_i   <= '0';
          rrif_stts_reg_i(4) <= '0';
          if (cal_test_count(16) = '1') then
            cal_test_count := '0' & unsigned(cnfg_rdback_count_i(15 downto 0)); -- Load delay count
            cal_test_state <= rd_from_fifo;
          else
            cal_test_count := cal_test_count - 1;
          end if;
        when rd_from_fifo =>
          rd_data_inmemA_i   <= '1';
          rd_data_inmemB_i   <= '1';
          rt_data_inmemA_i   <= '0';
          rt_data_inmemB_i   <= '0';
          rrif_stts_reg_i(4) <= '0';
          if (cal_test_count(16) = '1') then
            rd_data_inmemA_i <= '0';
            rd_data_inmemB_i <= '0';
            cal_test_state   <= set_rt_bit;
          else
            cal_test_count := cal_test_count - 1;
          end if;
        when set_rt_bit =>
          rd_data_inmemA_i   <= '0';
          rd_data_inmemB_i   <= '0';
          rt_data_inmemA_i   <= '1';
          rt_data_inmemB_i   <= '1';
          rrif_stts_reg_i(4) <= '0';
          cal_test_state     <= return_to_idle;
        when return_to_idle =>
          rd_data_inmemA_i   <= '0';
          rd_data_inmemB_i   <= '0';
          rt_data_inmemA_i   <= '0';
          rt_data_inmemB_i   <= '0';
          rrif_stts_reg_i(4) <= '0';
          cal_test_state       <= idle;
        when others => 
          rd_data_inmemA_i   <= '0';
          rd_data_inmemB_i   <= '0';
          rt_data_inmemA_i   <= '0';
          rt_data_inmemB_i   <= '0';
          rrif_stts_reg_i(4) <= '0';
          cal_test_state       <= idle;
      end case;
    else
      rd_data_inmemA_i   <= '0';
      rd_data_inmemB_i   <= '0';
      rt_data_inmemA_i   <= '0';
      rt_data_inmemB_i   <= '0';
      rrif_stts_reg_i(4) <= '0';
      cal_test_state     <= idle;
    end if;
  end if;
end process cal_data_test;


--------------------------------------------------------------------------
-- Process to allow Trapping of Link Data int the Inmems.
-- To setup a Trap:
-- Set rrif_ctrl_reg_i(27 downto 25) = "111"
-- Write a delay time value to Register Location 0x00404468 Bits[23:16]
-- Write a Fill time value to Register Location 0x00404468 Bits[15:0]
-- When a trigger is detected the counters start, write to Inmem FIFOs
-- When fe_cmdpulse_formattersAB_out is sent, write to the Debug FIFOs
-- When the TrapDone Flag is set Read Data from FIFOs
--  
trap_link_data : process (
  clk40_g, 
  rst_n_in_p, 
  trap_data_state,
  cnfg_rdback_count_i, 
  rrif_ctrl_reg_i,
  cal_l1a0_i, 
  cal_l1a1_i, 
  tim_signal_ii,
  fe_cmdpulse_formattersAB_out_p
  )
variable start_trapdelay_count  : unsigned(15 downto 0);
variable start_trapinmem_count  : unsigned(15 downto 0);
begin
  if (rst_n_in_p = '0') then
    wr_link_data_inmemA_i  <= '0';
    wr_link_data_inmemB_i  <= '0';
    trap_done_i            <= '0';
    start_trapdelay_count  := (others => '0');
    start_trapinmem_count  := (others => '0');
    trap_data_state        <= reset;
  elsif (clk40_g'event AND clk40_g = '1') then
    if (rrif_ctrl_reg_i(27 downto 25) = "111") then
      case trap_data_state is
        when reset =>
          wr_link_data_inmemA_i  <= '0';
          wr_link_data_inmemB_i  <= '0';
          trap_done_i            <= '0';
          start_trapdelay_count  := (others => '0');
          start_trapinmem_count  := (others => '0');
          trap_data_state        <= idle;
        when idle  =>
          wr_link_data_inmemA_i  <= '0';
          wr_link_data_inmemB_i  <= '0';
          trap_done_i            <=    inmem_status_in_p(2)
                                    OR inmem_status_in_p(0);
          start_trapdelay_count  := (others => '0');
          start_trapinmem_count  := (others => '0');
          if (cal_l1a0_i = '1' OR cal_l1a1_i = '1' OR tim_signal_ii(0) = '1') then
            trap_data_state <= start_delay;
          end if;
        when start_delay =>
          wr_link_data_inmemA_i  <= '0';
          wr_link_data_inmemB_i  <= '0';
          trap_done_i            <= '0';
          if (start_trapdelay_count = unsigned(cnfg_rdback_count_i(31 downto 16))) then
            trap_data_state <= wr_to_fifo;
            wr_link_data_inmemA_i  <= '1';
            wr_link_data_inmemB_i  <= '1';
          else
            start_trapdelay_count  := start_trapdelay_count + 1;
          end if;
        when wr_to_fifo =>
          trap_done_i <= '0';
          if (start_trapinmem_count = unsigned(cnfg_rdback_count_i(15 downto 0))) then
            wr_link_data_inmemA_i <= '0';
            wr_link_data_inmemB_i <= '0';
            trap_data_state       <= set_done_flag;
          else
            wr_link_data_inmemA_i <= '1';
            wr_link_data_inmemB_i <= '1';
            start_trapinmem_count := start_trapinmem_count + 1;
          end if;

        when set_done_flag =>
          wr_link_data_inmemA_i  <= '0';
          wr_link_data_inmemB_i  <= '0';
          trap_done_i            <=    inmem_status_in_p(2)
                                    OR inmem_status_in_p(0);
          start_trapdelay_count  := (others => '0');
          start_trapinmem_count  := (others => '0');
          trap_data_state        <= idle;
        when others => 
          wr_link_data_inmemA_i  <= '0';
          wr_link_data_inmemB_i  <= '0';
          trap_done_i            <= '0';
          start_trapdelay_count  := (others => '0');
          start_trapinmem_count  := (others => '0');
          trap_data_state <= idle;
      end case;
    else
      wr_link_data_inmemA_i  <= '0';
      wr_link_data_inmemB_i  <= '0';
      trap_done_i            <= '0';
      start_trapdelay_count  := (others => '0');
      start_trapinmem_count  := (others => '0');
      trap_data_state <= reset;
    end if;
  end if;
end process trap_link_data;

cal_l1id_mux : process (
  clk40_g, 
  rst_n_in_p
  )
begin
  if (rst_n_in_p = '0') then
    cal_l1id0_m <= (others => '0');
    cal_l1id1_m <= (others => '0');
  elsif (clk40_g'event AND clk40_g = '1') then
    if (mask_dxi_i = '1') then
      cal_l1id0_m <= "00000000" & nevent_o;
      cal_l1id1_m <= "00000000" & nevent_o;
    elsif (spare_ctrl_reg_i(11) = '1') then
      cal_l1id0_m <= std_logic_vector(event_count1);
      cal_l1id1_m <= std_logic_vector(event_count2);
    else
      cal_l1id0_m <= cal_l1id0_i;
      cal_l1id1_m <= cal_l1id1_i;
    end if;
--  cal_l1id1_m <= cal_l1id1_i;
  end if;
end process cal_l1id_mux;

diag_evt_count : process (
  clk40_g, 
  rst_n_in_p
  )
begin
  if (rst_n_in_p = '0') then
    event_count1 <= (others => '0');
    event_count2 <= (others => '0');
  elsif (clk40_g'event AND clk40_g = '1') then
    if (spare_ctrl_reg_i(11) = '0') then
      event_count1 <= (others => '0');
      event_count2 <= (others => '0');
    else
      if (tim_signal_ii(0) = '1') then
        event_count1 <= event_count1 + '1';
      end if;
      if (tim_signal_ii(0) = '1' AND rod_busy_n_out_p = '0') then
        event_count2 <= event_count2 + '1';
      end if;
    end if;
  end if;
end process diag_evt_count;

--------------------------------------------------------------------------
-- COMPONENT INSTATIATION
--------------------------------------------------------------------------

U2 : address_decoder_busbridge
  port map (
    rst_n_in           => rst_n_in_p,
    clk_in             => clk40_g,
    dspbus_ea_in       => controller_bus_adr_in_p(19 downto 0),
    dspbus_ed_in       => controller_bus_data_in_p(31 downto 0),
    dspbus_ed_out      => controller_bus_data_out_p(31 downto 0),
    dspbus_en_n_in     => controller_bus_ce_n_in_p,
    dspbus_be_n_in     => controller_bus_be_n_in_p,
    dspbus_rnw_in      => controller_bus_rnw_in_p,
    dspbus_re_n_in     => controller_bus_re_n_in_p,
    dspbus_oe_n_in     => controller_bus_oe_n_in_p,
    dspbus_ack_out     => controller_bus_ack_out_i,
    rodbus_a_out       => rod_bus_adr_out_p(19 downto 0),
    rodbus_d_in        => rod_bus_data_in_p(31 downto 0),
    rodbus_d_out       => rod_bus_data_out_p(31 downto 0),
    rodbus_en_n_out    => rod_bus_ce_n_out_p(10 downto 0),
    rodbus_en_n_itr    => rodbus_en_n_itr_i(6 downto 0),
    rodbus_rnw_out     => rod_bus_rnw_out_p,
    formA_ack_n_in     => formA_ack_n_in_p,
    formB_ack_n_in     => formB_ack_n_in_p,
    efb_ack_in         => efb_ack_in_p,
    router_ack_in      => router_ack_in_p,
    bocbus_busy_in     => bocbus_busy_in_p,
    bocbus_strb_out    => bocbus_strb_out_p,
    bocbus_rodoe_out   => bocbus_rodoe_out_p,
    bocbus_write_out   => bocbus_write_out_p,
    regbus_a_out       => regbus_a_i(11 downto 0),
    regbus_d_in        => regbus_d_in_i(31 downto 0),
    regbus_d_out       => regbus_d_out_i(31 downto 0),
    regbus_rnw_out     => regbus_rnw_i, 
    regbus_ack_in      => regbus_ack_i,
    fcmbus_d_in        => fcmbus_dout_i,
    focbus_a_out       => focbus_a_i(11 downto 0),
    focbus_d_in        => focbus_din_i,
    focbus_rnw_out     => focbus_rnw_i,
    focbus_ack_in      => focbus_ack_i,
    fmebus_a_out       => frebus_a_i,
    fmebus_d_in        => frebus_din_i(15 downto 0),
    fmebus_d_out       => frebus_dout_i(15 downto 0),
    fmebus_rnw_out     => frebus_rnw_i,
    fmebus_ack_in      => frebus_ack_i,
    edebus_a_out       => edebus_a_i,
    edebus_d_in        => edebus_din_i(15 downto 0),
    edebus_d_out       => edebus_dout_i(15 downto 0),
    edebus_rnw_out     => edebus_rnw_i,
    edebus_ack_in      => edebus_ack_i,
    dfibus_a_out       => dfibus_a_i,
    dfibus_rnw_out     => dfibus_rnw_i,
    dfibus_ack_in      => dfibus_ack_i,
    sdsp0_hrdy_in      => dsp0_hrdy_in_p,
    sdsp1_hrdy_in      => dsp1_hrdy_in_p,
    sdsp2_hrdy_in      => dsp2_hrdy_in_p,
    sdsp3_hrdy_in      => dsp3_hrdy_in_p,
    sdsp_wr_strb_n_out => sdsp_wr_n_out_i,   --p,
    sdsp_rd_strb_n_out => sdsp_rd_n_out_i,   --p,
    fmt_hwob_out       => fmt_hwob_p,
    fmt_ds_out         => fmt_ds_p,
    efb_hwob_out       => spare_o(22),
    extbus_ack_out     => extbus_ack_i,
    intbus_ack_out     => intbus_ack_i,
    sdsp_detect_in     => dsp_detect_i,
    sdsp_slow_write_in => spare_ctrl_reg_i(24),
    inc_hrdy_toc_out   => inc_hrdy_toc_i,
    sdsp_address_error => pin_hold_reg_i(31)
    );

U3 : register_block
  port map (
    clk_in                     => clk40_g,
    rst_n_in                   => rst_n_in_p,
    regbus_en_n_in             => regbus_en_n_i,
    regbus_rnw_in              => regbus_rnw_i,
    regbus_a_in                => regbus_a_i(11 downto 0),
    regbus_d_in                => regbus_d_out_i,
    regbus_d_out               => regbus_d_in_i,
    regbus_ack_out             => regbus_ack_i,
    fcmbus_en_n_in             => fcmbus_en_n_i,
    fcmbus_d_out               => fcmbus_dout_i,
    rrif_ctrl_reg_out          => rrif_ctrl_reg_i,
    spare_ctrl_reg_out         => spare_ctrl_reg_i,
    rrif_stts_reg_in           => rrif_stts_reg_i,
    spare_stts_reg_in          => spare_stts_reg_i,
    dsp_stts_reg_in            => dsp_stts_reg_i,
    test_stts_reg_in           => test_stts_reg_i,
    pin_hold_reg_in            => pin_hold_reg_i,
    fecmd_link_mask0_out       => fe_command_mask0_i,
    fecmd_link_mask1_out       => fe_command_mask1_i,
    calstrobe_delay_reg_out    => calstrobe_delay_reg_i,
    cal_command_reg_out        => cal_command_reg_i,
    modebits_fifo_stts_reg_in  => modebits_fifo_stts_reg_i,
    dynmmask_fifo_stts_reg_in  => dynmmask_fifo_stts_reg_i,
    inmem_wordcount_reg_out    => inmem_wordcount_i,
    dbgmem_wordcount_reg_out   => dbgmem_wordcount_i,
    cnfg_rdback_count_reg_out  => cnfg_rdback_count_i,
    ide_debugmem_stts_reg_in   => ide_debugmem_stts_reg_i,
    ide_debugmem_cnfg_reg_out  => ide_debugmem_cnfg_reg_i,
    debug_fifo_stts_in         => debug_fifo_stts_i,
    rol_test_block_reg_out     => rol_test_block_i,
    rol_test_block_count_out   => rol_test_block_count_i,  
    int_tim_fifo_outdata_out   => int_tim_fifo_outdata_i,
    int_tim_fifo_re_in         => int_tim_fifo_re_i,
    int_tim_fifo_rt_in         => int_tim_fifo_rt_i,
    slave_dsp_ints_out         => slave_dsp_ints_out_i,
    slave_dsp_ints_in          => slave_dsp_ints_in_i,
    sdsp_inttohost_signal_out  => slv_int_to_mdsp_i,
    feocc_reset_reg_out        => feocc_reset_i,
    feocc_num_l1_reg_out       => feocc_num_l1_i,
    cal_event_type0_out        => cal_event_type0_i,
    cal_event_type1_out        => cal_event_type1_i,
    cal_l1id0_in               => cal_l1id0_m,   --cal_l1id0_i,
    cal_l1id1_in               => cal_l1id1_m,   --cal_l1id1_i,
    static_l1id0_out           => static_l1id0_i,
    static_l1id1_out           => static_l1id1_i,
    cal_bcid0_in               => cal_bcid0_i,
    cal_bcid1_in               => cal_bcid1_i,
    static_bcid0_out           => static_bcid0_i,
    static_bcid1_out           => static_bcid1_i,
    rod_mode_out               => open,
    test_mux_out               => test_mux_i,      
    fe_cmd_mask_sel_out        => fe_cmd_mask_sel_i,
    fe_cmd_mask_updated_in     => fe_cmd_mask_updated_i,
    corrective_mask_sent_in    => corrective_fe_pulse_i,
    inc_hrdy_toc_in            => inc_hrdy_toc_i,
    send_int_test_out          => send_int_test_i,
    test_irq_out               => test_irq_i,
    ecr_count_in               => std_logic_vector(ecr_id),
    ecr_count_out              => ecr_count_i,
    nevent_out                 => nevent_i,
    precal_delay_out           => precal_delay_i,
    trig_delay0_out            => trig_delay0_i,
    trig_delay1_out            => trig_delay1_i,
    interval_out               => interval_i,
    trig_mode_out              => trig_mode_i,
    mask_en_in                 => int_mask_en_i,  -- Select Mask LUT
    new_mask_in                => int_new_mask_i, -- WR new mask to Register
    load_event_type_out        => int_load_event_i,
    int_nevent_done_in         => int_nevent_done_i,
    l1id_bits_in               => event_l1id,
    fftv_cmnd_out              => fftv_cmnd_i,
    fftv_busy_stat_in          => fftv_busy_stat_i,
    fftv_ID_stat_in            => fftv_ID_stat_i,
    fftv_match_level_in        => fftv_match_level_i,
    set_rod_busy_out           => set_rod_busy_i,
    ecr_busy_count_out         => ecr_busy_count_i,
    dropped_l1a_count          => dropped_l1a_count,
    dropped_ecr_count          => dropped_ecr_count,
    dropped_bcr_count          => dropped_bcr_count
    );

U4 : debugmem_fifo_interface
  port map (
    rst_n_in          => rst_n_in_p,
    clk_in            => clk40_g,
    control_in        => rrif_ctrl_reg_i(25),
    address_in        => dfibus_a_i(6 downto 0),
    rnw_in            => dfibus_rnw_i,
    inmem_en_n_in     => dfibus_inmem_en_n_i,
    eventmem_en_n_in  => dfibus_eventmem_en_n_i,
    inmem_ctrl_out    => inmem_ctrl_i(15 downto 0),
    eventmem_ctrl_out => eventmem_ctrl_i(10 downto 0),
    ack_out           => dfibus_ack_i
    );

evt_id_rst_n <= efb_edm_fifo_rst_n_out_p AND rst_n_in_p;

U5 : trigger_signal_decoder
  port map (
    clk_in              => clk40_g,
    rst_n_in            => rst_n_in_p,
    enable_decoding_in  => rrif_ctrl_reg_i(8),
    serial_id_in        => tim_signal_ii(4),
    serial_tt_in        => tim_signal_ii(5),
    l1id_bits_out       => l1id_bits_i,
    bcid_bits_out       => bcid_bits_i,
    ttype_bits_out      => ttype_bits_i,
    new_l1id_valid_out  => new_l1id_valid_i,
    new_ttype_valid_out => new_ttype_valid_i
    );

U6 : efb_headerdynamicmask_encoder
  port map  (
    clk_in                         => clk40_g,
    rst_n_in                       => evt_id_rst_n,  -- rst_n_in_p,  --
    control_in                     => rrif_ctrl_reg_i(12),
    rnw_in                         => edebus_rnw_i,
    ce_n_in                        => edebus_en_n_i,
    ack_out                        => edebus_ack_i,
    apply_default_mask_in          => apply_dynmask_mb_i(0),
    apply_corrective_mask_in       => apply_dynmask_mb_i(1),
    record_headerdynamicmask_in    => trap_indicator_i(1),
    read_trig_indictor_que_out     => read_trigger_que_i,
    l1id_bits_in                   => l1id_bits_i,
    bcid_bits_in                   => bcid_bits_i,
    ttype_bits_in                  => ttype_bits_i,
    new_l1id_valid_in              => new_l1id_valid_i,
    new_ttype_valid_in             => new_ttype_valid_i,
    int_lut_bdata_in               => edebus_dout_i,
    int_lut_bdata_out              => edebus_din_i,
    int_lut_adr_in                 => edebus_a_ii,
    event_headermask_fifo_ef_n_in  => efb_edm_fef_n_in_p,
    event_headermask_fifo_ff_n_in  => efb_edm_fff_n_in_p,
    event_headermask_out           => efb_edm_data_out_p,
    event_lbid_count_out           => dynmmask_fifo_stts_reg_i(17 downto 10),
    event_trgt_count_out           => dynmmask_fifo_stts_reg_i(25 downto 18),
    event_mask_count_out           => dynmmask_fifo_stts_reg_i(33 downto 26),
    event_header_count_out         => dynmmask_fifo_stts_reg_i(41 downto 34),
    event_headermask_fifo_we_n_out => efb_edm_we_n_out_p,
    int_fifo_ef_out                => dynmmask_fifo_stts_reg_i(4 downto 0),
    int_fifo_ff_out                => dynmmask_fifo_stts_reg_i(9 downto 5),
    ecrid_in                       => ecr_id,
    int_cal_mode_en_in             => rrif_ctrl_reg_i(18),
    cal_l1a0_in                    => cal_l1a0_i,
    cal_l1a1_in                    => cal_l1a1_i,
    cal_tt0_in                     => cal_event_type0_i,
    cal_tt1_in                     => cal_event_type1_i,
    cal_l1id0_in                   => cal_l1id0_m,   --cal_l1id0_i,  -- nevent_o
    cal_l1id1_in                   => cal_l1id1_m,   --cal_l1id1_i,  -- nevent_o
    cal_bcid0_in                   => cal_bcid0_i,
    cal_bcid1_in                   => cal_bcid1_i,
    pause_triggers_out             => efb_header_pause_i,
    event_type_in                  => fe_cmd_mask_sel_i,
    load_event_type_in             => int_load_event_i,
    mask_dxi_in                    => mask_dxi_i,
    nevent_in                      => nevent_o,
    boc_ok_in                      => boc_clock_ok_p,
    tim_ok_in                      => tim_clock_ok_p,
    rol_test_in                    => rol_test_i,
    rol_test_count_in              => rol_test_block_count_i,
    rod_tt_wr_out                  => rod_tt_wr_i,
    event_l1id_out                 => event_l1id
    );
edebus_a_ii <= unsigned(edebus_a_i);
--
-- Signal from the Trigger Indicator Que (which traps which l1 mask updates occur 
-- and therefore when the corrective mask is used) to record "this trigger's" ID 
-- information. The Whole id/header/dynamic mask is trapped here instead of just 
-- the id information, as it saved us some additional complexity in the vhdl.

U7: formatter_readoutmodebits_encoder
  port map(
    clk_in                   => clk40_g,
    rst_n_in                 => rst_n_in_p,
    control_in               => rrif_ctrl_reg_i(10),
    int_lut_adr_in           => frebus_a_ii,
    int_lut_bdata_in         => frebus_dout_i,
    int_lut_bdata_out        => frebus_din_i,
    ce_n_in                  => frebus_en_n_i,
    rnw_in                   => frebus_rnw_i,
    ack_out                  => frebus_ack_i,
    apply_default_mode_in    => apply_dynmask_mb_i(0),
    apply_corrective_mode_in => apply_dynmask_mb_i(1),
    modebits_fifo_ff_n_in    => formatter_mb_fff_n_in_p(1 downto 0),
    modebits_out             => formatter_mb_data_out_p(11 downto 0),
    modebits_fifo_we_n_out   => formatter_mb_we_n_out_p(7 downto 0),
    int_fifo_ef_out          => modebits_fifo_stts_reg_i(1 downto 0),
    int_fifo_ff_out          => modebits_fifo_stts_reg_i(3 downto 2),
    mb0_fifo_occ_count       => modebits_fifo_stts_reg_i(13 downto  4),
    mb1_fifo_occ_count       => modebits_fifo_stts_reg_i(23 downto 14),
    modebits_sent_out        => modebits_sent_i,
    fe_cmd_mask_sel_in       => fe_cmd_mask_sel_i
    );
frebus_a_ii <= unsigned(frebus_a_i);

U8: fe_command_pulse_counter
  port map (
    clk_in                    => clk40_g,
    rst_n_in                  => evt_id_rst_n,  -- rst_n_in_p,  -- 
    control_in                => rrif_ctrl_reg_i(6 downto 5),
    fe_cmdpulse_in            => fe_pulse_i,
    corrective_fe_cmdpulse_in => corrective_fe_pulse_i,
    fe_cmdpulse_count_out     => rrif_stts_reg_i(29 downto 22),
    fe_cmdpulse_AB_out        => fe_cmdpulse_formattersAB_out_p,
    controller_int_out        => apply_dynmask_mb_i(1 downto 0),
    command_cycle_done_out    => cmd_cycle_done_i,    --test_out_p(5),
    store_current_trigger_out => store_trig2ti_que_i,
    current_trigger_out       => trigger_descriptor_i,
    dynamic_mask_sent_in      => read_trigger_que_i,
    modebits_sent_in          => modebits_sent_i,
    efb_ready_in              => efb_edm_fef_n_in_p,
    fe_trigger_occupancy_out  => spare_stts_reg_i(18 downto 8),
    modebits_error_out        => pin_hold_reg_i(30)
    );

U9 : fifo_2048x2  -- FE Trigger FIFO
  port map (
    clk_in                    => clk40_g,
    rst_n_in                  => evt_id_rst_n,  --  rst_n_in_p,  -- 
    flush_fifo_in             => spare_ctrl_reg_i(7),
    wren_in                   => store_trig2ti_que_i,
    rden_in                   => read_trigger_que_i,
    data_in                   => trigger_descriptor_i,
    data_out                  => trap_indicator_i,
    full_flag_out             => rrif_stts_reg_i(6),
    empty_flag_out            => rrif_stts_reg_i(5)
    );

U10: fe_command_processor
  port map (
    clk_in                    => clk40_g,
    rst_n_in                  => rst_n_in_p,
    enable_in                 => rrif_ctrl_reg_i(0),
    rod_type_in               => spare_stts_reg_i(31),
    cnfg_cal_new_mask_load_in => rrif_ctrl_reg_i(20),
    l1accept_in               => tim_signal_ii(0),
    bcr_in                    => tim_signal_ii(2),
    ecr_in                    => tim_signal_ii(1),
    calibrate_in              => tim_signal_ii(3),
    calibrate_command_in      => cal_command_reg_i,
    cal_trig_delay_in         => calstrobe_delay_reg_i,
    fast_not_long_commands_in => rrif_ctrl_reg_i(1),
    new_mask_ready_in         => rrif_ctrl_reg_i(2),
    corrective_mask_ready_in  => rrif_ctrl_reg_i(23),
    fe_command_mask0_in       => fe_command_mask0_i,
    fe_command_mask1_in       => fe_command_mask1_i,
    long_command_stream_in    => controller_dx_i,       -- controller_dx_in_p,
    long_command_pulse_in     => cal_l1a_i,             -- controller_tout_in_p,
    command_stream_out        => fe_command_link_out_p,
    pulse_stream_out          => fe_l1_pulse_i,
    fe_pulse_out              => fe_pulse_i,
    corrective_fe_pulse_out   => corrective_fe_pulse_i,
    fe_cmd_mask_updated_out   => fe_cmd_mask_updated_i,
    fmt_fe_cmd_out            => fmt_fe_cmd_i,
    fmt_fe_cmd_mask_n_in      => rrif_ctrl_reg_i(18),
    ecr_flush_en_in           => spare_ctrl_reg_i(23 downto 22),
    cmd_proc_busy_out         => cmd_proc_busy_i
    );
-- Add rrif_ctrl_reg_i(18) to U10 to mask DSP serial ports off of fmt_fe_cmd_out

U11 : front_end_occupancy_counters
  port map  (
    clk_in                        => clk40_g,
    rst_n_in                      => rst_n_in_p,
    rod_type_in                   => spare_stts_reg_i(31),   -- 0=>SCT, 1=>Pixel
    trailerflags_readout_en_in    => rrif_ctrl_reg_i(3),
    trailerflags_in               => formatter_writes_trailer_in_p,
    clear_fe_occupancy_counter_in => feocc_reset_i,
    num_l1_value_in               => feocc_num_l1_i,
    fe_l1_pulse_in                => fe_l1_pulse_i,
    fe_occupancy_gl_zero_out      => rrif_stts_reg_i(30),
    group_zero_out                => feocc_zero_i,
    group_en_out                  => act_group_i,
    show_trailer_flags_out        => show_trailer_flags_out_p,
    focbus_a_in                   => focbus_a_i,
    focbus_d_in                   => regbus_d_out_i,
    focbus_d_out                  => focbus_din_i,
    focbus_rnw_in                 => focbus_rnw_i,
    focbus_en_n_in                => focbus_en_n_i,
    focbus_ack_out                => focbus_ack_i
    );

U12 : testbench_interface
  port map(
    rst_n_in                     => rst_n_in_p,
    clk_in                       => clk40_g,
    control_in                   => rrif_ctrl_reg_i(17 downto 16),
    debugmode_cnfg_register_in   => ide_debugmem_cnfg_reg_i,
    debugmode_stts_register_out  => ide_debugmem_stts_reg_i,
    inmem_wordcount_register_in  => inmem_wordcount_i,
    dbgmem_wordcount_register_in => dbgmem_wordcount_i,
    inmem_ctrl_out               => tb_inmem_ctrl_out_i,
    int_tim_fifo_re_out          => int_tim_fifo_re_i,
    int_tim_fifo_rt_out          => int_tim_fifo_rt_i,
    l1_trigger_in                => tim_signal_ii(0),
    pause_trigger_rt             => pause_triggers_i
    );

U13 : sp_trig_cmd_detector
  port map (
    clk_in         => clk40_g,
    rst_n_in       => rst_n_in_p,
    enable_in      => rrif_ctrl_reg_i(18),
    rod_type_in    => spare_stts_reg_i(31), -- 0=>SCT, 1=>Pixel
    l1id_select_in => rrif_ctrl_reg_i(22),  --if (rrif_ctrl_reg_i(22) = 1, use reg value)
    bcid_select_in => rrif_ctrl_reg_i(21),  --if (rrif_ctrl_reg_i(21) = 1, use reg value)
-- **** FFTV in the circuit ****
--    cal_serial0_in => cal_serial0_o,
--    cal_serial1_in => cal_serial1_o,
-- **** FFTV bypassed ****
    cal_serial0_in => cal_serial0_i,   -- controller_dx_in_p(0),
    cal_serial1_in => cal_serial1_i,   -- controller_dx_in_p(1),
    l1_accept0_out => cal_l1a0_i,
    l1_accept1_out => cal_l1a1_i,
    l1id0_reg_in   => static_l1id0_i,
    l1id1_reg_in   => static_l1id1_i,
    l1id0_out      => cal_l1id0_i,
    l1id1_out      => cal_l1id1_i,
    bcid0_reg_in   => static_bcid0_i,
    bcid1_reg_in   => static_bcid1_i,
    bcid0_out      => cal_bcid0_i,
    bcid1_out      => cal_bcid1_i
    );
cal_l1a_i <= cal_l1a1_i & cal_l1a0_i;
cal_serial0_i <= (controller_dx_i(0) AND NOT mask_dxi_i) OR 
                 (rol_l1a_i          AND NOT rol_test_block_i(2)) OR
                 (int_scan_fecmd0_i  AND     mask_dxi_i);
cal_serial1_i <= (controller_dx_i(1) AND NOT mask_dxi_i) OR 
                 (int_scan_fecmd1_i  AND     mask_dxi_i);

U14 : int_scan_engine
  port map(
    clk_in            => clk40_g,
    rst_n_in          => rst_n_in_p,
    enable_in         => rrif_ctrl_reg_i(29),
    clear_nevent_in   => rrif_ctrl_reg_i(30),
    new_mask_ready_in => rrif_ctrl_reg_i(2),
    rod_tt_wr_in      => rod_tt_wr_i,
    rod_type_in       => spare_stts_reg_i(31),   -- 0=>SCT, 1=>Pixel
    num_events_in     => nevent_i,
    precal_delay_in   => precal_delay_i,
    trig_delay0_in    => trig_delay0_i,
    trig_delay1_in    => trig_delay1_i,
    evt_interval_in   => interval_i,
    act_groups_in     => act_group_i,
    trig_mode_in      => trig_mode_i,
    feocc_zero_in     => feocc_zero_i,
    xoff_in           => pause_triggers_i,
    mask_en_out       => int_mask_en_i,     -- Select Mask LUT
    new_mask_out      => int_new_mask_i,    -- WR new mask to Register
    mask_dxi_out      => mask_dxi_i,
    fecmd0_out        => int_scan_fecmd0_i,
    fecmd1_out        => int_scan_fecmd1_i,
    nevent_out        => nevent_o,    -- L1ID
    nevent_done_out   => int_nevent_done_i
    );

U15 : fftv_main
  port map( 
    rst                   => rst_in_i,
    clk40                 => clk40_g,
    fftv_match_lvl_i      => fftv_cmnd_i(7 downto 4),
    fftv_match_lvl_o      => fftv_match_level_i,
    fftv_count_o          => fftv_busy_stat_i(23 downto 0),
    fftv_count_clr_i      => fftv_cmnd_i(1),
    fftv_count_rollover_o => fftv_busy_stat_i(24),
    fftv_id_o             => fftv_id_stat_i(7 downto 0),
    fftv_id_clr_i         => fftv_cmnd_i(2),
    fftv_id_rollover_o    => fftv_id_stat_i(8),
    fftv_emergency_o      => fftv_busy_stat_i(27),
    fftv_emergency_clr_i  => fftv_cmnd_i(0),
    trigger0_i            => cal_serial0_i,
    trigger0_o            => cal_serial0_o,
    trigger1_i            => cal_serial1_i,
    trigger1_o            => cal_serial1_o,
    fftv_busy_o           => fftv_busy_stat_i(25),
    fftv_disable_i        => fftv_disable_i
    );

-- Disable if Pixel ROD or RoL Test Block Tx
fftv_disable_i <= '1';
--fftv_disable_i <= spare_stts_reg_i(31) OR rol_trigs_on_i;

end rtl; -- code for top_rcf
