----------------------------------------------------------------------------
-- LBL - WISCONSIN
-- ATLAS ROD ELECTRONICS
----------------------------------------------------------------------------
-- Filename:
--    efb_source.vhd
-- Description:
--    ROD EFB output simulator
--       SCT Version
--
-- Reads a test data file and produces a data stream to help simulate
-- the EFB output into the Router. 
--
-- The code works by first reading in the data file and storing the
-- entire string of bits into a large array of characters. As the
-- data file is being read, whitespace and comments are filtered out
-- before being written into the array.
--
----------------------------------------------------------------------------

library IEEE ;
use IEEE.STD_LOGIC_1164.all ;
use IEEE.std_logic_arith.all;

----------------------------------------------------------------------------
--PORT DECLARATION
----------------------------------------------------------------------------

entity efb_source is
port (
  clk40_out : out  STD_LOGIC ; -- Distributed 40 Mhz clock 
  clk80_out : out  STD_LOGIC ; -- Distributed 40 Mhz clock 
  rst_out : out  STD_LOGIC ; -- Powerup Reset

  -- Event Data Bus
  event_data_out  : out  STD_LOGIC_VECTOR(45 downto 0);
  data_valid_out  : out  STD_LOGIC;
  tristate_n_out  : out  STD_LOGIC_VECTOR(2 downto 0);
  -- tristate_n(0) -> efb_a output
  -- tristate_n(1) -> efb_b output
  -- tristate_n(2) -> header/trailer output

  router_bus_done_in : in STD_LOGIC ;

  vcc_out : out  STD_LOGIC ; -- connects to VCC to drive it.
  gnd_out : out  STD_LOGIC   -- connects to GND to drive it.
  ) ;
end  efb_source;

architecture rtl of efb_source is

----------------------------------------------------------------------------
--SIGNAL DECLARATION
----------------------------------------------------------------------------

  signal tsf_ctr: integer; -- index into data bit stream character array
  signal rst : std_logic := '0'; -- internal version of powerup reset signal
  signal clk : std_logic := '0'; -- internal version of clock signal
  signal clk80 : std_logic := '1'; -- internal version of 80MHz clock signal
  signal data_output : STD_LOGIC_VECTOR(49 downto 0); -- internal
     -- version of efb output signals
  constant clk_period : time := 25 ns; -- set system clock period here
  constant max_file_length : integer := 60000000; -- max file size.
     -- Change here for bigger files (keep in mind it eats up memory)
 
----------------------------------------------------------------------------
--COMPONENT DECLARATION
----------------------------------------------------------------------------
  
BEGIN  --  Main Body of VHDL code

----------------------------------------------------------------------------
--COMPONENT INSTANTIATION
----------------------------------------------------------------------------

  -- since you can't read an output signal, I used internal versions
  -- that get reassigned here before going to the output.
  clk40_out <= clk;
  clk80_out <= clk80;
  rst_out <= rst;
  event_data_out <= data_output(45 downto 0);
  tristate_n_out <= data_output(48 downto 46);
  data_valid_out <= data_output(49);

------------------------------------------------------------------------------
-- Test Stand: File to Bitstream  
------------------------------------------------------------------------------
  vcc_out <= '1';
  gnd_out <= '0';

  -- Generate reset pulse and clock signals
  rst <= '1' after ((clk_period * 4) + 1 ns);
  clk <= not clk after (clk_period/2);
  clk80 <= not clk80 after (clk_period/4);
    
  eventdata: process  (clk, rst, router_bus_done_in)
    type load_file_type is file of character;
    -- data file path/filename
  
  file load_file: load_file_type is "/mnt/edafiles/atlas/rod/fpga_code/router/simulation_input_files/efb_out.rod";

    type storage_array is array(0 to max_file_length) of character;
    -- character array containing the data bitstream
    variable storage: storage_array;
    variable index: integer;
    -- location of last character
    variable max_tsf_ctr: integer;
    variable offset: integer;
    variable tsf_ctr: integer;
    variable i: integer;
    variable outpnibble : std_logic_vector (49 downto 0) := (others =>'0');
  begin
    if (rst = '0') then
      -- Main loop to read data file, filter comments/white space and
      -- write data to character array
      tsf_ctr := 0;
      index:=0;
      data_output <= (others => '0');
      while not endfile(load_file) loop
	read(load_file, storage(index));
	if (storage(index) /= lf and storage(index) /= ' ') then
          assert (index < max_file_length)
            report "input file is too long"
            severity failure;
          if (storage(index) = '#') then
            while (not endfile(load_file) and storage(index) /= lf) loop
              if (not endfile(load_file)) then
                 read(load_file, storage(index));
              end if;
            end loop;
            index:= index - 1;
          end if;  
	  index := index+1;
        else
          storage(index):='0';
	end if;
        max_tsf_ctr := index;
      end loop;
    elsif (clk'event and clk = '1') then
      -- Simple clocked loop that read character array and outputs 50
      -- bit data words. Any character not a '1' or '0' gets 'Z' value
        if (router_bus_done_in = '1') then
          for i in 49 downto 0 loop                
	    if (storage(tsf_ctr) = '0') then
              outpnibble(i) := '0';
            elsif (storage(tsf_ctr) = '1') then
              outpnibble(i) := '1';
            else       
              outpnibble(i) := 'Z';
            end if;
            data_output <= outpnibble after 5 ns;
         -- Stops when the last character is read in and outputs all
         -- zeroes after that. The last word in the data file should
         -- contain dummy data of all zeroes to make sure it doesn't cut off
         --  or lose the last few data bits in the file.
            if (tsf_ctr < max_tsf_ctr-1) then
              tsf_ctr := tsf_ctr + 1;
            else
              data_output(48 downto 0) <= "1110000000000000000000000000000000000000000000000";
              data_output(49) <= '0';
            end if;
          end loop;
        else
          data_output(48 downto 0) <= (others => '1');
          data_output(49) <= '0';
        end if;
    end if;
end process;

END rtl;

