-------------------------------------------------------------------------------
--
-- Particle Physics Detector Electronics Support 
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 2000
-- ReadOutDriver Electronics
--
-------------------------------------------------------------------------------
--
-- Filename: dsp_data_fifo.vhd                  
-- Description:
--  Data fifo for DSP interface, uses 8 1024x4 dpram blocks.    
--
--  The following state machines are all 1 stage glitch free MOORE 
--  MACHINES, as per the LBL template  "vhdl/sm1_sample.vhd" code.
--
--                                                                   
--            +----------+   +---------------+      +---------+   
-- data_in ---| Input    |---| Dual Port     |------| Output  |--- data_out
--            | Register |   | SRAM          |      | State   |     
--            |          |   |       |       |      | Logic   |     
--            |          |   |       |       |      |         |     
--            |          |   |   A   |  B    |      |         |     
--       +---->   en     |--->       |       <----+-->   en   |     
--       |    +----------+   |       |       |   |  +---------+     
-- clk --+        |          +---------------+   |      |
-- wren ----------+            |         clk80---+      | 
--                |            |      +-----------+     |                   
--                |            +------| Control   |-----+
--                |                   | Logic     |                        
--                +-------------------|           |                       
-- rden ------------------------------|           |                        
--                                    |           |                      
--                                    +-----------+                       
--                                                      
--
-------------------------------------------------------------------------------
-- Structure: 
--  Multiple state machines in a flat topology/hierarchy which communicate
--   with single wire signals.  
-------------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
-------------------------------------------------------------------------------
-- Author: Mark L. Nagel
-- Board Engineer: John M. Joseph
-- History:
--    23 February 2000 - JMJ first version : requirements/functionality finally
--                       understood and sufficiently documented to proceed 
--                       with coding.
--    24 January 2002  - JMJ changed the logic for the transfer counter.
--
-------------------------------------------------------------------------------
-- NOTES
-------------------------------------------------------------------------------
-- 1.  FIFO Reads:  
--     The output of the Router is connected to the Bakend DSP SBSRAM ports
--     of the EMIF.
-------------------------------------------------------------------------------
-- LIBRARY INCLUDES
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations
-------------------------------------------------------------------------------
-- PORT DECLARATION
-------------------------------------------------------------------------------

entity dsp_data_fifo is
  port (
    clk40_in              : in  std_logic; -- clk40 input
    clk80_in              : in  std_logic; -- sbsram clk 80MHz
    rst_n_in              : in  std_logic; -- async global reset
    mask_dfr_in           : in  std_logic;
    int_mask_delay_in     : in  std_logic_vector( 5 downto 0);
    cen_in                : in  std_logic; -- CE from DSP
    rden_in               : in  std_logic; -- OE from DSP
    wren_in               : in  std_logic; -- enable writes
    wr_evt_word_count_in  : in  std_logic; -- write event word count
    data_in               : in  std_logic_vector(31 downto 0);
    data_out              : out std_logic_vector(31 downto 0);
    occupancy_count_out   : out std_logic_vector(12 downto 0); 
    full_flag_out         : out std_logic; --  fifo full flag
    afull_flag_out        : out std_logic; --  fifo almost full flag
    empty_flag_out        : out std_logic; --  fifo empty flag
    data_frame_rdy_out    : out std_logic; --  data frame ready,
    rd_pointer_out        : out std_logic_vector( 9 downto 0);
    data_frame_rdy_stat   : out std_logic_vector( 3 downto 0);
    fifo_data_fill_mode   : in  std_logic; -- 1 = Data Driven
    mask_fifo_int4_in     : in  std_logic
    );
end dsp_data_fifo;

architecture rtl of dsp_data_fifo is
-------------------------------------------------------------------------------
-- SIGNAL DECLARATION
-------------------------------------------------------------------------------

signal gnd   : std_logic;
signal vcc   : std_logic;
signal rst_i : std_logic;

signal ingoing_data_i   : std_logic_vector(31 downto 0);
signal outcoming_data_i : std_logic_vector(31 downto 0);

signal write_en_i    : std_logic;
signal writein_adr_i : unsigned(9 downto 0);

signal cen_i           : std_logic;
signal read_en_i       : std_logic;
signal rden_i          : std_logic;
signal readout_adr_i   : unsigned(9 downto 0);

signal occupancy_count_i : unsigned(9 downto 0);
signal block_count_i     : unsigned(2 downto 0);
signal inc_block_count_i : std_logic;
signal dec_block_count_i : std_logic;

signal data_frame_ready_i : std_logic;
signal afull_flag_out_i   : std_logic;
signal full_flag_out_i    : std_logic;
signal empty_flag_out_i   : std_logic;

type fifo_readout_states is (
  idle,    -- 0x01
  setup,   -- 0x02
  readout, -- 0x04
  done     -- 0x08
  );
signal fifo_ro_state : fifo_readout_states;

type data_frame_rdy_states is (
  reset,               -- 0x0
  idle,                -- 0x1
  retest_occupancy,    -- 0x2
  issue_frame_rdy_int, -- 0x3
  wait_for_isr,        -- 0x4
  isr_done,            -- 0x5
  dec_block_count,     -- 0x6
  mask                 -- 0x7
  );
signal df_rdy_state : data_frame_rdy_states;

signal dec_read_pointer_i  : std_logic;
signal clr_data_fifo_i     : std_logic;
signal rst_wr_pointer_i    : std_logic;

-------------------------------------------------------------------------------
-- COMPONENT DECLARATION
-------------------------------------------------------------------------------

component dpram_1024x4
  port (
    clkA_in                : in  std_logic; -- clk40 input
    clkB_in                : in  std_logic; -- 80 MHz sbsram clk from the dsp chip
    rst_in                 : in  std_logic; -- asynchronous global reset
    dprmA_writein_bus_in   : in  std_logic_vector(3 downto 0);
    dprmB_writein_bus_in   : in  std_logic_vector(3 downto 0);
    writein_adr_in         : in  unsigned(9 downto 0); -- 
    readout_adr_in         : in  unsigned(9 downto 0); --
    dprmA_write_strobe_in  : in  std_logic; --
    dprmB_write_strobe_in  : in  std_logic; --
    dprmA_read_strobe_in   : in  std_logic; --
    dprmB_read_strobe_in   : in  std_logic; --
    dprmA_readout_bus_out  : out std_logic_vector(3 downto 0);
    dprmB_readout_bus_out  : out std_logic_vector(3 downto 0)
    );
end component;

-------------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
-------------------------------------------------------------------------------
begin

M1A : dpram_1024x4
  port map(
    clkA_in                => clk40_in,
    clkB_in                => clk80_in,
    rst_in                 => rst_i,
    dprmA_writein_bus_in   => ingoing_data_i(3 downto 0),
    dprmB_writein_bus_in   => ingoing_data_i(3 downto 0),
    writein_adr_in         => writein_adr_i,
    readout_adr_in         => readout_adr_i,
    dprmA_write_strobe_in  => write_en_i,
    dprmB_write_strobe_in  => gnd,
    dprmA_read_strobe_in   => gnd,
    dprmB_read_strobe_in   => vcc,
    dprmA_readout_bus_out  => open,
    dprmB_readout_bus_out  => outcoming_data_i(3 downto 0)
    );

M1B : dpram_1024x4
  port map(
    clkA_in                => clk40_in,
    clkB_in                => clk80_in,
    rst_in                 => rst_i,
    dprmA_writein_bus_in   => ingoing_data_i(7 downto 4),
    dprmB_writein_bus_in   => ingoing_data_i(7 downto 4),
    writein_adr_in         => writein_adr_i,
    readout_adr_in         => readout_adr_i,
    dprmA_write_strobe_in  => write_en_i,
    dprmB_write_strobe_in  => gnd,
    dprmA_read_strobe_in   => gnd,
    dprmB_read_strobe_in   => vcc,
    dprmA_readout_bus_out  => open,
    dprmB_readout_bus_out  => outcoming_data_i(7 downto 4)
    );

M1C : dpram_1024x4
  port map(
    clkA_in                => clk40_in,
    clkB_in                => clk80_in,
    rst_in                 => rst_i,
    dprmA_writein_bus_in   => ingoing_data_i(11 downto 8),
    dprmB_writein_bus_in   => ingoing_data_i(11 downto 8),
    writein_adr_in         => writein_adr_i,
    readout_adr_in         => readout_adr_i,
    dprmA_write_strobe_in  => write_en_i,
    dprmB_write_strobe_in  => gnd,
    dprmA_read_strobe_in   => gnd,
    dprmB_read_strobe_in   => vcc,
    dprmA_readout_bus_out  => open,
    dprmB_readout_bus_out  => outcoming_data_i(11 downto 8)
    );

M1D : dpram_1024x4
  port map(
    clkA_in                => clk40_in,
    clkB_in                => clk80_in,
    rst_in                 => rst_i,
    dprmA_writein_bus_in   => ingoing_data_i(15 downto 12),
    dprmB_writein_bus_in   => ingoing_data_i(15 downto 12),
    writein_adr_in         => writein_adr_i,
    readout_adr_in         => readout_adr_i,
    dprmA_write_strobe_in  => write_en_i,
    dprmB_write_strobe_in  => gnd,
    dprmA_read_strobe_in   => gnd,
    dprmB_read_strobe_in   => vcc,
    dprmA_readout_bus_out  => open,
    dprmB_readout_bus_out  => outcoming_data_i(15 downto 12)
    );

M1E : dpram_1024x4
  port map(
    clkA_in                => clk40_in,
    clkB_in                => clk80_in,
    rst_in                 => rst_i,
    dprmA_writein_bus_in   => ingoing_data_i(19 downto 16),
    dprmB_writein_bus_in   => ingoing_data_i(19 downto 16),
    writein_adr_in         => writein_adr_i,
    readout_adr_in         => readout_adr_i,
    dprmA_write_strobe_in  => write_en_i,
    dprmB_write_strobe_in  => gnd,
    dprmA_read_strobe_in   => gnd,
    dprmB_read_strobe_in   => vcc,
    dprmA_readout_bus_out  => open,
    dprmB_readout_bus_out  => outcoming_data_i(19 downto 16)
    );

M1F : dpram_1024x4
  port map(
    clkA_in                => clk40_in,
    clkB_in                => clk80_in,
    rst_in                 => rst_i,
    dprmA_writein_bus_in   => ingoing_data_i(23 downto 20),
    dprmB_writein_bus_in   => ingoing_data_i(23 downto 20),
    writein_adr_in         => writein_adr_i,
    readout_adr_in         => readout_adr_i,
    dprmA_write_strobe_in  => write_en_i,
    dprmB_write_strobe_in  => gnd,
    dprmA_read_strobe_in   => gnd,
    dprmB_read_strobe_in   => vcc,
    dprmA_readout_bus_out  => open,
    dprmB_readout_bus_out  => outcoming_data_i(23 downto 20)
    );

M1G : dpram_1024x4
  port map(
    clkA_in                => clk40_in,
    clkB_in                => clk80_in,
    rst_in                 => rst_i,
    dprmA_writein_bus_in   => ingoing_data_i(27 downto 24),
    dprmB_writein_bus_in   => ingoing_data_i(27 downto 24),
    writein_adr_in         => writein_adr_i,
    readout_adr_in         => readout_adr_i,
    dprmA_write_strobe_in  => write_en_i,
    dprmB_write_strobe_in  => gnd,
    dprmA_read_strobe_in   => gnd,
    dprmB_read_strobe_in   => vcc,
    dprmA_readout_bus_out  => open,
    dprmB_readout_bus_out  => outcoming_data_i(27 downto 24)
    );

M1H : dpram_1024x4
  port map(
    clkA_in                => clk40_in,
    clkB_in                => clk80_in,
    rst_in                 => rst_i,
    dprmA_writein_bus_in   => ingoing_data_i(31 downto 28),
    dprmB_writein_bus_in   => ingoing_data_i(31 downto 28),
    writein_adr_in         => writein_adr_i,
    readout_adr_in         => readout_adr_i,
    dprmA_write_strobe_in  => write_en_i,
    dprmB_write_strobe_in  => gnd,
    dprmA_read_strobe_in   => gnd,
    dprmB_read_strobe_in   => vcc,
    dprmA_readout_bus_out  => open,
    dprmB_readout_bus_out  => outcoming_data_i(31 downto 28)
    );

-------------------------------------------------------------------------------
-- SIGNAL ASSIGNMENTS
-------------------------------------------------------------------------------
gnd <= '0';
vcc <= '1';

rst_i <= NOT rst_n_in;

afull_flag_out <= afull_flag_out_i;
full_flag_out  <= full_flag_out_i;
empty_flag_out <= empty_flag_out_i;
data_out       <= outcoming_data_i;

data_frame_rdy_out  <= data_frame_ready_i;
--occupancy_count_out <= std_logic_vector(block_count_i) &
--                       std_logic_vector(writein_adr_i);
occupancy_count_out <= std_logic_vector(block_count_i) &
                       std_logic_vector(occupancy_count_i);
rd_pointer_out      <= std_logic_vector(readout_adr_i);

-------------------------------------------------------------------------------
-- PROCESS DECLARATION
-------------------------------------------------------------------------------
-- The ReadOut State Machine controls the FIFO outputs to interact with the
-- SBSRAM port of the DSP.  The transfer from Router to DSP IDRAM is handled
-- using DMA channel 0 of the DSP.  The maximum block of data that can be 
-- transfered in one operation is 256 words.  If SBSRAM_CE_N  becomes inactive
-- for any reason in the middle of a transfer, when the DMA continues, 2 clocks
-- will be added to the width of the transfer.  See pg. 101 of the TMS320C6713
-- data sheet (TI Doc#: SRPS186C) for the exact timing specification.

readout_state_machine : process (
  rst_n_in, 
  clk80_in, 
  cen_in, 
  rden_in
  )
begin
  if (rst_n_in = '0') then
    rden_i             <= '0';
    cen_i              <= '0';
    fifo_ro_state      <= idle;
    dec_read_pointer_i <= '0';
  elsif (clk80_in'event AND clk80_in = '1') then
    rden_i <= rden_in;
    cen_i  <= cen_in;
    case fifo_ro_state is
      when idle =>
        dec_read_pointer_i <= '0';
-- If SBSRAM_CE_N has become active and the FIFO has data, start a transfer.
        if (cen_i = '1') then
          fifo_ro_state <= setup;
        end if;
      
      when setup =>
-- Test to be sure that SBSRAM_OE_N has become active before proceeding
        if (rden_i = '1') then
          fifo_ro_state <= readout;
        end if;
        
      when readout =>           
        if (rden_i = '0') then
          dec_read_pointer_i <= '1';
          fifo_ro_state <= done;
        end if;
        
      when done =>
        dec_read_pointer_i <= '0';
-- Return to idle only after the SBSRAM CE_N signal is cleared       
        if (cen_i = '0') then
          fifo_ro_state <= idle;
        end if;
      
      when others => 
    end case;
  end if;
end process readout_state_machine;

-------------------------------------------------------------------------------
-- output address counter
-------------------------------------------------------------------------------
output_counter : process (
  rst_n_in, 
  clk80_in
  )
begin
  if (rst_n_in = '0') then
    readout_adr_i <= (others => '0');
    read_en_i           <= '0';
  elsif (clk80_in'event AND clk80_in = '1') then
    read_en_i <= cen_in;
    if (dec_read_pointer_i = '1') then
      readout_adr_i <= readout_adr_i - 1;
    elsif (read_en_i = '1') then
      readout_adr_i <= readout_adr_i + 1;
    end if;
    -- Occupancy Counter
    occupancy_count_i <= writein_adr_i - readout_adr_i;
  end if;
end process output_counter;

-------------------------------------------------------------------------------
-- Write Strobe Decoder
-------------------------------------------------------------------------------
fifo_wr_strobe : process (
  rst_n_in, 
  clk40_in, 
  wr_evt_word_count_in
  )
begin
  if (rst_n_in = '0') then
    write_en_i <= '0';
  elsif (clk40_in'event AND clk40_in = '1') then
    write_en_i <= (NOT full_flag_out_i AND wren_in) OR 
                   wr_evt_word_count_in OR
                   clr_data_fifo_i;
  end if;
end process fifo_wr_strobe;

-------------------------------------------------------------------------------
-- Register the data into the DPRAM for stability and to keep it aligned with
-- the Write Strobes
-------------------------------------------------------------------------------
input_register : process (
  rst_n_in, 
  clk40_in, 
  data_in
  )
begin
  if (rst_n_in = '0') then
    ingoing_data_i <= (others => '0');
  elsif (clk40_in'event AND clk40_in = '1') then
    if (clr_data_fifo_i = '1') then
      ingoing_data_i <= (others => '0');
    else
      ingoing_data_i <= data_in;
    end if;
  end if;
end process input_register;

-------------------------------------------------------------------------------
-- input address counter
-------------------------------------------------------------------------------
input_counter : process (
  rst_n_in, 
  clk40_in, 
  write_en_i, 
  writein_adr_i
  )
begin
  if (rst_n_in = '0') then
    writein_adr_i  <= (others => '0');
  elsif (clk40_in'event AND clk40_in = '1') then
    if (clr_data_fifo_i = '1') then
      writein_adr_i <= writein_adr_i + 1;
    elsif (rst_wr_pointer_i = '1') then
      writein_adr_i <= (others => '0');
    else
      if (wr_evt_word_count_in = '1' AND (writein_adr_i < 255 OR writein_adr_i = 1023)) then
        writein_adr_i <= "0011111111";
      elsif (wr_evt_word_count_in = '1' AND writein_adr_i < 511) then
        writein_adr_i <= "0111111111";
      elsif (wr_evt_word_count_in = '1' AND writein_adr_i < 767) then
        writein_adr_i <= "1011111111";
      elsif (wr_evt_word_count_in = '1' AND writein_adr_i < 1023) then
        writein_adr_i <= "1111111111";
      elsif (write_en_i = '1') then
        writein_adr_i <= writein_adr_i + 1;
      end if;
    end if;
  end if;
end process input_counter;

-------------------------------------------------------------------------------
-- FIFO 1024 word occupancy counter
-------------------------------------------------------------------------------
occupancy_counter : process (
  rst_n_in, 
  clk40_in
  )
begin
-- Clear the Occupancy Counter at Chip Reset
  if (rst_n_in = '0') then
    block_count_i <= (others => '0');
  elsif (clk40_in'event AND clk40_in = '1') then
    if (clr_data_fifo_i = '1') then
      block_count_i <= (others => '0');
    else
      if (writein_adr_i(7 downto 0) = 255 AND write_en_i = '1' AND dec_block_count_i = '1') then
        block_count_i <= block_count_i;
      elsif (writein_adr_i(7 downto 0) = 255 AND write_en_i = '1' AND block_count_i < 4) then
        block_count_i <= block_count_i + 1;
      elsif (dec_block_count_i = '1' AND block_count_i > 0) then
        block_count_i <= block_count_i - 1;
      end if;
    end if;
  end if;
end process occupancy_counter;

occupancy_flags : process (
  rst_n_in, 
  block_count_i,
  writein_adr_i,
  clr_data_fifo_i
  )
begin
  if (rst_n_in = '0') then
    empty_flag_out_i <= '1';
    full_flag_out_i  <= '0';
    afull_flag_out_i <= '0';
  else
    if (clr_data_fifo_i = '1') then
      full_flag_out_i <= '0';
      afull_flag_out_i <= '0';
      empty_flag_out_i <= '1';
    else      
      if (block_count_i = 4 OR 
         (block_count_i = 3 AND writein_adr_i(7 downto 0) >= 252)) then 
        full_flag_out_i <= '1';
      else
        full_flag_out_i <= '0';
      end if;

      if (fifo_data_fill_mode = '1') then -- If the Router is in Data Drive fill mode
        if (block_count_i = 4 OR 
           (block_count_i = 3 AND writein_adr_i(7 downto 0) >= 240)) then 
          afull_flag_out_i <= '1';
        else
          afull_flag_out_i <= '0';
        end if;
      else  -- If Router is set in default Event fill mode
        if (block_count_i >= 2) then -- Was 3 in previous versions 2 is fine for Cal Mode
          afull_flag_out_i <= '1';
        else
          afull_flag_out_i <= '0';
        end if;
      end if;
 
      if (block_count_i = 0 AND writein_adr_i(7 downto 0) = 0) then 
        empty_flag_out_i <= '1';
      else 
        empty_flag_out_i <= '0';
      end if;
    end if;
  end if;
end process occupancy_flags;

-------------------------------------------------------------------------------
-- State Machine to generate the Data Frame Ready Signal
-- This process insures that only 1 interrupt will be issued per transfer.
-- The HPI priority in the slave must be set to the lowest priority for this
-- algorithm to function correctly.  This process includes the interrupt
-- masking required by the DSP algorithm
-------------------------------------------------------------------------------
gen_data_frame_ready : process (
  rst_n_in, 
  clk40_in, 
  mask_dfr_in, 
  int_mask_delay_in
  )
variable dfr_mask_counter : unsigned(5 downto 0);
begin
  if (rst_n_in = '0') then
    data_frame_ready_i  <= '0';
    dfr_mask_counter    := (others => '0');
    dec_block_count_i   <= '0';
    df_rdy_state        <= reset;
    data_frame_rdy_stat <= (others => '0');
  elsif (clk40_in'event AND clk40_in = '1') then
    if (clr_data_fifo_i = '0') then
      case df_rdy_state is
        when reset => 
          data_frame_ready_i <= '0';
          dec_block_count_i  <= '0';
          dfr_mask_counter := (others => '0');
          df_rdy_state <= idle;
          data_frame_rdy_stat <= "0000";

        when idle =>
          data_frame_ready_i <= '0';
          dec_block_count_i  <= '0';
          if (block_count_i > 0 AND mask_fifo_int4_in = '0') then
            df_rdy_state <= retest_occupancy;
          end if;
          data_frame_rdy_stat <= "0001";

        when retest_occupancy =>
          data_frame_ready_i <= '0';
          if (block_count_i > 0 AND mask_fifo_int4_in = '0') then
            df_rdy_state <= issue_frame_rdy_int;
          else
            df_rdy_state <= reset;
          end if;
          data_frame_rdy_stat <= "0010";

        when issue_frame_rdy_int =>
          data_frame_ready_i <= '1';
          if (cen_i = '1') then
            df_rdy_state <= wait_for_isr;
          end if;
          data_frame_rdy_stat <= "0011";

        when wait_for_isr =>
          data_frame_ready_i <= '0';
          if (mask_dfr_in = '1') then
            df_rdy_state <= isr_done;
          end if;
          data_frame_rdy_stat <= "0100";
            
        when isr_done =>
          data_frame_ready_i <= '0';
          if (mask_dfr_in = '0') then
            df_rdy_state <= dec_block_count; 
          end if;
          data_frame_rdy_stat <= "0101";

        when dec_block_count =>
          dec_block_count_i <= '1';
          df_rdy_state      <= mask; 
          data_frame_rdy_stat <= "0110";

        when mask =>
          data_frame_ready_i <= '0';
          dec_block_count_i  <= '0';
          if (std_logic_vector(dfr_mask_counter) = int_mask_delay_in) then
            df_rdy_state <= reset; 
          else
            dfr_mask_counter := dfr_mask_counter + 1;
          end if;
          data_frame_rdy_stat <= "0111";
  
        when others => 
          data_frame_ready_i <= '0';
          dec_block_count_i  <= '0';
          dfr_mask_counter := (others => '0');
          df_rdy_state <= idle;
          data_frame_rdy_stat <= "1000";
      end case;
    else
      data_frame_ready_i  <= '0';
      dfr_mask_counter    := (others => '0');
      dec_block_count_i   <= '0';
      df_rdy_state        <= reset;
      data_frame_rdy_stat <= (others => '0');
    end if;
  end if;
end process gen_data_frame_ready;

-------------------------------------------------------------------------------
-- State Machine to clear FIFOs after reset
-------------------------------------------------------------------------------
clear_data_fifo : process (
  rst_n_in, 
  clk40_in
  )
variable fifo_clr_counter : unsigned(10 downto 0);
begin
  if (rst_n_in = '0') then
    clr_data_fifo_i  <= '1';
    rst_wr_pointer_i <= '0';
    fifo_clr_counter := (others => '0');
  elsif (clk40_in'event AND clk40_in = '1') then
    if (clr_data_fifo_i = '1') then
      if (fifo_clr_counter <= 1024) then
        fifo_clr_counter := fifo_clr_counter + 1;
      else
        clr_data_fifo_i  <= '0';
        rst_wr_pointer_i <= '1';
      end if;
    else
      fifo_clr_counter := (others => '0');
      clr_data_fifo_i  <= '0';
      rst_wr_pointer_i <= '0';
    end if;
  end if;
end process clear_data_fifo;

end rtl; -- end code for dsp_data_fifo


