--------------------------------------------------------------------------
-- Particle Physics Detector Electronics Support
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 2000
-- ReadOutDriver Electronics
--
--------------------------------------------------------------------------
-- Filename: dump data event.vhd
-- Description:
--  Searches the event headers for specific event types, and dumps data
--  as required
--
--------------------------------------------------------------------------
-- Structure:
--------------------------------------------------------------------------
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk40
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
--
--   
--------------------------------------------------------------------------
-- Author:         John Joseph
-- Board Engineer:
-- History:
--    3/17/00 JMJ First version
--    2/22/05 JMJ Modified Data dump state machine for mor clarity and 
--                smaller slice utilization
--    
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- LIBRARY INCLUDES
--------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations 
--------------------------------------------------------------------------
-- SIMULATION LIBRARY INCLUDES
--------------------------------------------------------------------------

entity dump_event_data is
port (
    clk40_in           : in  std_logic;
    rst_n_in           : in  std_logic;
    slink_cmd_reg_in   : in  std_logic_vector(5 downto 0);
           -- Bit 0: Dump Atlas Specific Event Type
           -- Bit 1: Dump TIM Specific Event Type
           -- Bit 2: Dump ROD Specific Event Type
           -- Bit 3: Inhibit S-Link Write Enable
           -- Bit 4: Reset S-Link
           -- Bit 5: Set S-Link Test Bit
--  slink_status_in    : in  std_logic_vector(1 downto 0);
           -- Bit 0: S-Link Xoff Status
           -- Bit 1: S-Link Bad Link Status
    dump_match_reg_in  : in  std_logic_vector(15 downto 0);
           -- [ 7: 0] Atlas Specific Events
           -- [ 9: 8] TIM Specific Events
           -- [15:10] ROD Specific Events
    L1TT_hw_in         : in  std_logic_vector(9 downto 0);
    DET_hw_in          : in  std_logic_vector(5 downto 0);
    hw_capture_in      : in  std_logic;
    dump_event_out     : out std_logic
    );
end dump_event_data;

architecture rtl of dump_event_data is
--------------------------------------------------------------------------
-- SIGNAL DECLARATION
--------------------------------------------------------------------------

type dump_states is (
  dump_event, 
  keep_event
  );
signal dump_state  : dump_states;

begin
--------------------------------------------------------------------------
-- PROCESS DECLARATION
--------------------------------------------------------------------------

dump_data_event: process (
  rst_n_in, 
  clk40_in
  )
begin
  if (rst_n_in = '0') then
    dump_state <= keep_event;
    dump_event_out <= '0';
  elsif (clk40_in'event and clk40_in = '1') then
    if (hw_capture_in = '1') then
      case slink_cmd_reg_in(2 downto 0) is
    -- Match bits [7:0] of Event Header Word 7 to dump_match_reg_in[7:0]
        when "001" =>
          if (dump_match_reg_in(7 downto 0) = L1TT_hw_in(7 downto 0)) then
            dump_state <= dump_event;
          else
            dump_state <= keep_event;
          end if;

    -- Match bits [17:16] of Event Header Word 8 to dump_match_reg_in[9:8]
        when "010" =>
          if (dump_match_reg_in(9 downto 8) = L1TT_hw_in(9 downto 8)) then
            dump_state <= dump_event;
          else
            dump_state <= keep_event;
          end if;

    -- Match bits [5:0] of Event Header Word 8 to dump_match_reg_in[15:0]
        when "100" =>
          if (dump_match_reg_in(15 downto 10) = DET_hw_in) then
            dump_state <= dump_event;
          else
            dump_state <= keep_event;
          end if;

        when others => dump_state <= keep_event;
      end case;
    end if;

    if (dump_state = keep_event) then
      dump_event_out <= '0';
    else
      dump_event_out <= '1';
    end if;
  end if;
end process; -- 
end rtl; -- code for dump data event