-------------------------------------------------------------------------------
-- Particle Physics Detector Electronics Support
-- University of Wisconsin
-- Lawrence Berkeley National Laboratory (c) 2000
-- ReadOutDriver Electronics
--
-------------------------------------------------------------------------------
-- Filename: top.vhd alias "RouterFPGA.vhd"
-- Description:
--  The Router FPGA directs the event data to the slink and to the 
--  Slave DSPs.
--
--  I. BACKGROUND:
--
--   The S-Link and the monitoring/global-error-scaling DSP have a common data
--  format which consists of a 9 word event header, a 5 word event trailer
--  with the link data sandwiched in between.
--
--  event fragment header
--    h0: 0xB0F00000 + UCTL (beginning of fragment)
--    h1: 0xEE1234EE        (start of header)
--    h2: 0x9               (header size)
--    h3: tbd               (format version number)
--    h4: tbd               (source identifier)
--    h5: tbd               (run number)
--    h6: L1ID              (level 1 ID)
--    h7: BCID              (bunch crossing ID)
--    h8: L1TT              (level 1 trigger type)
--    h9: DET               (detector event type - ROD/TIM)
--  
--  event fragment trailer
--    t0: error flags       (status 1: bitted error type occurrence word)
--    t1: error count       (status 2: count of words with errors)
--    t2: 0x2               (number of status words)
--    t3: nData             (number of data words
--    t4: 0x1               (status block position: 0/1 = before/after data)
--    t5: 0xE0F00000 + UCTL (end of fragment)
--  
--  link data format
--  
--   The 'out of router' column is valid for the SLink and the monitoring DSP.
--  The labels in this case, e.g. d1, d2, do not enumerate data words as in
--  the case of h0, ... and t0, ..., rather they label types of data words.
--  
--                                    into router            out of router
--   d1: header              001pLLLLBBBBBBBB  blKwMMMMMMM  001pwlbKxMMMMMMM
--   d2: trailer             010thvxxxxxxxxxx  xxxxxxxxxxx   same as input
--       condensed format
--   d3:  1 hit              1CCCCSSSSSSSxex0  xxxxxxxxxxx   same as input
--   d4:  2 hits             1CCCCSSSSSSSeex1  xxxxxxxxxxx   same as input
--       full format
--   d5:  1st hit            1CCCCSSSSSSS0DDD  xxxxxxxxxxx   same as input
--   d6:  hit pair           1xxxxxxx1DDD1DDD  xxxxxxxxxxx   same as input
--   d7:  last hit (if even) 1xxxxxxx0xxx1DDD  xxxxxxxxxxx   same as input
--   d8: FE flagged error    000xxxxxxCCCCEEE  xxxxxxxxxxx   same as input
--   d9: raw data            011NNNxxDDDDDDDD  xxxxxxxxxxx   same as input
--  
--   The link data is organized...by link, and they arrive in a fixed order
--  from event to event.  Each link block begins with a link header.  The link
--  trailer is only present if there are error bits set in it; these are the
--  'v', 't', and the 'h' in d2 and will be explained below.  The link header 
--  itself is surpressed in the event that it would be the only word in the 
--  event for that link AND there are no error flags set in it, the 'p', 'w', 
--  'l' and 'b'.

--  Data arrives at the router from the event fragment builder on a 43 bit
--  (plus a couple of spares) bus.  The first 32 bits are two of the 16 bit
--  fields from the left column below 'into router' above.  Bits 32-42 are the
--  field in the right column below 'into router'.  Note that this field is a
--  don't care except for link headers.  (It is also don't care for event
--  fragment header and event fragment trailer words.)  Because of this, if
--  the first 16 bits are the last field for link n-1 and the next 16 bits are
--  the first field for link n, the last 11 bits correspond to link n.
--  
--    1) Error bits added by the ROD are in lower case letters.
--    2) Bits extracted directly from the incoming data are in upper case.
--    3) Bits which are fixed are written explicitly as 0 or 1.
--    4) Don't care bits are indicated by 'x'.
--    5) Hit words always begin with 1.
--      a) Each link formatter instantiated in the formatter FPGAs is in
--        condensed or full mode. This determines how it formats the incoming
--        link data.  The mode will only change in response to an instruction
--        from the crate processor.  The router also knows what mode each of
--        the formatters is in via up to 3 32 bit registers (the maximum
--        number of links is 96) which are written at the same time the
--        formatters are configured.
--      b) During normal running we are in condensed mode.  In this case, a 3
--        bit data field which accompanies every hit coming from the modules
--        is dropped.  This allows packing clusters of 1 or 2 hits into a
--        single 16 bit field.  Because about 20% of clusters have 2 hits and
--        very few (< 1%) have 3 or more hits, this results in a significant
--        reduction in bandwidth out of the ROD.  In condensed mode, before
--        dropping the data field, the formatter checks it for errors.  Errors
--        are flagged by the 'e' bits in d3 and d4, one 'e' for the case of 1
--        hit encoded in a word (b15=0) and two 'e's for the case of 2 hits
--        encoded in a word (b15=1).  The base address of the cluster, chip
--        number and strip number, is encoded in CCCCSSSSSSS at b1-b11.  The
--        same chip and strip number are repeated for all hits in the cluster.
--        e.g. a 4 hit cluster in chip 4 starting at strip 32 and with an
--        error on hit 3 would be encoded in the following 2 16 bit fields:
--           10100010000000x1  hits 1 and 2 (strips 32,33)
--           10100010000001x1  hits 3 and 4 (strips 34,35) and error flag
--      c) Any link can be put in full mode, in which case the 3 data bits
--        which accompany each hit are kept.  These are the DDD fields in
--        d5-d7.  In this case, the first hit in any cluster (hit with lowest
--        strip number) is indicated by b12=0.  The chip number (CCCC) and
--        base address (SSSSSSS, number of lowest strip in cluster) are
--        encoded in b1-b11.  b12=1 indicates a continuation of a cluster. 
--        b8=0 indicates the continuation is 1 hit.  b8=1 indicates the
--        continuation is at least 2 hits.  e.g. a 4 hit cluster on chip 4
--        starting at strip 32 is described by the following 3 16 bit fields:
--            1010001000000DDD  1st hit (strip 32)
--            1xxxxxxx1DDD1DDD  hit 2 (strip 33, MSBs) and 3 (strip 34, LSBs)
--            1xxxxxxx0xxx1DDD  hit 4 (strip 35)
--    6) The 4 other types of word begin with '0' and are distinguished by b1
--      and b2.
--    7) Link headers begin with 001.  A single bit error in the event
--      preamble coming from the detector module is indicated by p=1 at b3.
--      The L1ID (4 bits) and BCID (8 bits) from the detector module are
--      encoded as LLLLBBBBBBBB in b4-b15.  The link number (up to 96) is
--      encoded as MMMMMMM in the high order 11 bit field.  Also in that field
--      are:  w=1 in case of a timeout
--            l=1 indicates a L1ID error (LLLL from detector did not match the
--               lower order bits of the L1ID sent from the T/DAQ system.
--               The latter is sent back to T/DAQ via the SLink in h5.
--            b=1 indicates a BCID error (BBBBBBBB from the detector did not
--               match the lower order bits of the BCID sent from the T/DAQ
--               system. The latter is sent back to T/DAQ via the SLink in h6.
--            K=1 indicates the link is formatted in condensed mode.
--    8) Link trailer words begin with 010.  Link trailer words are inserted
--      when the formatter FPGA detects a link trailer pattern coming up the
--      link.  (This pattern is 1000000000000000 and should never occur in the
--      data.)  The event fragment builder FPGA, after counting event errors
--      and filling in the error summary words, t0 and t2, removes trailer 
--      words (which contain no information) before they reach the router.
--      The exception is the case when one of the error bits is set.
--       t = trailer with error (If the trailer pattern above arrives, but
--          with a single bit error, and it arrives at a time when we are
--          expecting it OR the pattern above arrives with no bit errors but
--          at a time when we are not expecting it, we call it a trailer with
--          an error...this philosophy might change.)
--       h = header only mode.  This indicates that the derandomizing buffer
--          at the output of the formatter is full and so data is being thrown
--          away; only headers and trailers are kept.
--       v = data overflow error.  This indicates that data arrived on the
--          link for a given event, but just kept on coming...the link
--          basically went nuts and a at a certain point the readout had to
--          advance to the next link formatter.
--    9) Front end chips can send 3 bit error codes up the data link to the
--      ROD.  When one of these error codes arrives it is formatted in a word
--      beginning with 000.  The error code is EEE and the number of the chip
--     which had the error is, again, CCCC.
--   10) If the formatter FPGA gets confused, e.g. due to a bit error on a
--      link, and can no longer decode data for an event, it goes into raw
--      data mode.  All raw data words begin with 011.  The field DDDDDDDD is
--      the raw data.  The formatter will generate raw data until it detects
--      an event trailer.  (It can still do that even if the state machine is
--      confused.)  Because the last raw data word may have less than 8 bits,
--      the count of relevant bits is encoded in the NNN field (0-7 mapping to
--      1-8.  This may not be needed.  It should be trivial offline to find
--      the end of the data, in which case 13 rather than 8 raw data bits
--      could be packed in a word).  Raw data mode can also be selected as the
--      default via a configuration register.

--  II. FORMAT FOR THE ERROR DIAGNOSTIC DSP

--     One of the slave DSPs is concerned with diagnosing errors.  This DSP
--    will need more than just the error summaries.  It will need all error
--    flags as well as the module number and the L1ID and BCID fields from the
--    detectors.  Now we have the problem of cramming a 43 bit bus into the 32
--    bit bus to the DSP w/o increasing the bandwidth too much.  We did solve
--    this problem once, but it seems that the detailed solution was not
--    written down.

--     We ended up taking advantage of the fact that not all of the event
--    fragment header and trailer words (hn and tn) are needed by the DSP.
--    Some of them are just needed by the LVL2 trigger or offline analysis.
--    So, here is something which works...(taking the 96 link case)

--   1) The condensed mode bits will no longer be used in the router, but they
--      will still be inserted into the Event Header in words H2 -> H4.  The
--      information will be loaded directly into the BackEnd DSP.  This will
--      allow the router to compact the Event Header by 3 words H2 -> H4, and 
--      extend the Event Trailer by 3 words T5->T7.  This can be accomplished
--      easily using a tapped pipeline for the DSP data.

--   2) As the data comes into the router, it builds up six 32 bit words
--     with 2 bits per link.  The two bits encode the following cases:
--      00 no data (the router did not get a header for this link, there were
--                  no hits and no errors)
--      01 good data on the link, no errors
--      10 bcid or l1id error (Note that this is included so that the DSP does
--         not need to scan all of the data to find the error.  This will
--         point to a link with a problem.  The DSP can then go to that link
--         and determine whether the error was a BCID or L1ID and exactly what
--         the error was.)
--      11 timeout error
--     Three of these words overwrite the event fragment trailer words
--     t2-t4.  The other three are additional words which would be inserted
--     before the E0F00000 word in the event fragment trailer.  (I seem to
--     recall that the solution did have 3 additional words in the event
--     fragment trailer...I may be wrong.)

--   3) This takes care of the bits wKlb, and also, because the 00 marking no
--     data allows the DSP to keep track of module numbers (module number = 
--     number of link header words read + number of intervening 00 codes), the
--     MMMMMMM can also be dropped. 
--
--  event fragment header
--    h0: 0xB0F00000 + UCTL (beginning of fragment)
--    h1: 0xEE1234EE        (start of header)
--    h2: 0x00000006        (Number of Header Words)
--    h3: L1ID              (level 1 ID)           
--    h4: BCID              (bunch crossing ID)
--    h5: L1TT              (level 1 trigger type)
--    h6: DET               (detector event type)
--
--  event fragment trailer
--    t0: error count       (status 1: count of words with errors)
--    t1: error flags       (status 2: bitted error type occurrence word)
--    t2: error status for links  1 - 16
--    t3: error status for links 17 - 32
--    t4: error status for links 33 - 48
--    t5: error status for links 49 - 64
--    t6: error status for links 65 - 80
--    t7: error status for links 81 - 96
--    t8: 0xE0F00000 + UCTL (end of fragment)
--
-------------------------------------------------------------------------------
-- Structure:
-------------------------------------------------------------------------------
--
-- Timing:
--    *The Engine FPGA runs at 40MHz => clk40
--    *The timing structure of all engine code will be:
--      1.  Perform all logic operations on positive edge of clk40
--      2.  Perform all memory operations on positive edge of clk40
--
--  The readout portion of the counters will run at nominally 80MHz
--   and be sync'd to the SBSRAM clock of their respective DSP.
--
-------------------------------------------------------------------------------
-- Author:         John M. Joseph
-- Board Engineer: John M. Joseph
-- History:
--
-------------------------------------------------------------------------------
-- LIBRARY INCLUDES
-------------------------------------------------------------------------------
--
library IEEE;
use IEEE.std_logic_1164.all;  -- needed for logic operations
use IEEE.std_logic_arith.all; -- needed for +/- operations
--
-------------------------------------------------------------------------------
-- SIMULATION LIBRARY INCLUDES
-------------------------------------------------------------------------------
-- pragma translate_off
library UNISIM;
use unisim.all;
-- pragma translate_on
-------------------------------------------------------------------------------
-- ENTITY DECLARATION
-------------------------------------------------------------------------------
--
entity router_top is
PORT (
-- main clock and reset signals for fpga --2
  clk40_in                : in    std_logic; -- 40MHz Clock input
  clk80_in                : in    std_logic; -- 80MHz Clock input
  rst_n_in                : in    std_logic; -- Asynchronous active low reset
-- event input from the efb --50
  event_info_in           : in    std_logic_vector(45 downto 0); -- event data
  tristate_n_in           : in    std_logic_vector( 2 downto 0);  -- indicates header/trailer or EFB engine
  data_valid_in	        : in    std_logic; -- indicates valid data from the EFB
-- backpressure to the efb --1 
  stop_output_out         : out   std_logic; --  signal to stop output from the EFB
-- slink interface --38 
  slink_data_out          : out   std_logic_vector(31 downto 0); -- SLink data bus
  slink_test_n_out        : out   std_logic; -- SLink test signal, active low
  slink_rst_n_out         : out   std_logic; -- SLink reset signal, active low
  slink_wen_n_out         : out   std_logic; -- SLink write enable signal, active low
  slink_uctrl_n_out       : out   std_logic; -- SLink begin/end of event signal, active low
  slink_xoff_n_in         : in    std_logic; -- SLink fifo full signal, active low
  slink_down_n_in         : in    std_logic; -- SLink bad signal, active low
-- router configuration register bus --28     
  router_bus_data_inout   : inout std_logic_vector(15 downto 0); -- Router bus data
  router_bus_addr_in      : in    std_logic_vector( 7 downto 0); -- Router bus address
  router_bus_rnw_in       : in    std_logic; -- Router bus read not write (direction control)
  router_bus_strb_n_in    : in    std_logic; -- Router bus strobe input
  router_bus_ack_out      : out   std_logic; -- Router bus acknowledge 
-- slave DSP0 interface --38 
  dsp0_sbsram_me_n_in     : in    std_logic; -- DSP to Router Module Enable
  dsp0_send_data_n_in     : in    std_logic; -- DSP ready to receive data
  dsp0_trap_enable_in     : in    std_logic; -- DSP to Router Trap Enable in
  dsp0_dfr_int_mask_in    : in    std_logic; -- Signal to Mask DataFrameReady Interrupt to SDSP
  dsp0_dataframeready_out : out   std_logic; -- Router to DSP Data Frame ready for transfer
  dsp0_event_data_out     : out   std_logic_vector(31 downto 0);
-- slave DSP1 interface --38    
  dsp1_send_data_n_in     : in    std_logic; -- DSP ready to receive data
  dsp1_sbsram_me_n_in     : in    std_logic; -- DSP to Router Module Enable
  dsp1_trap_enable_in     : in    std_logic; -- DSP to Router Trap Enable in
  dsp1_dfr_int_mask_in    : in    std_logic; -- Signal to Mask DataFrameReady Interrupt to SDSP
  dsp1_dataframeready_out : out   std_logic; -- Router to DSP Data Frame ready for transfer
  dsp1_event_data_out     : out   std_logic_vector(31 downto 0);
-- slave DSP2 interface --38    
  dsp2_send_data_n_in     : in    std_logic; -- DSP ready to receive data
  dsp2_sbsram_me_n_in     : in    std_logic; -- DSP to Router Module Enable
  dsp2_trap_enable_in     : in    std_logic; -- DSP to Router Trap Enable in
  dsp2_dfr_int_mask_in    : in    std_logic; -- Signal to Mask DataFrameReady Interrupt to SDSP
  dsp2_dataframeready_out : out   std_logic; -- Router to DSP Data Frame ready for transfer
  dsp2_event_data_out     : out   std_logic_vector(31 downto 0);
-- slave DSP3 interface --38   
  dsp3_send_data_n_in     : in    std_logic; -- DSP ready to receive data
  dsp3_sbsram_me_n_in     : in    std_logic; -- DSP to Router Module Enable
  dsp3_trap_enable_in     : in    std_logic; -- DSP to Router Trap Enable in
  dsp3_dfr_int_mask_in    : in    std_logic; -- Signal to Mask DataFrameReady Interrupt to SDSP
  dsp3_dataframeready_out : out   std_logic; -- Router to DSP Data Frame ready for transfer
  dsp3_event_data_out     : out   std_logic_vector(31 downto 0);
-- ROD Type
  rod_type_in             : in    std_logic;
-- test outputs --8
  test_out                : out   std_logic_vector( 7 downto 0); -- test outputs for debug 
-- spare outputs --3
  spare_out               : out   std_logic_vector( 3 downto 0); -- spare outputs
  spare_pin               : in    std_logic_vector(12 downto 0);  -- spare inputs
  hold_next_event_out     : out   std_logic  -- RTR-SP(13)
  ) ;                                        -- RTR-SP(13:10)==>EFB-SP(23:20)
end router_top ;

architecture rtl of router_top is
-------------------------------------------------------------------------------
-- SIGNAL DECLARATION
-------------------------------------------------------------------------------
  signal gnd : std_logic; -- internal version of ground signal

  signal rst           : std_logic; -- inverted version of reset signal
  signal rst_n         : std_logic; -- internal version of reset signal
  signal clk40         : std_logic; -- internal version of clock signal
  signal clk40_dll_in  : std_logic;
  signal clk40_dll_out : std_logic;
  signal clk80         : std_logic; -- internal version of clock signal
  signal clk80_dll_in  : std_logic;
  signal clk80_dll_out : std_logic;

-- Internal event data signals:
  signal event_info_p  : std_logic_vector(45 downto 0); 
  signal event_info_i  : std_logic_vector(45 downto 0); 
  signal tristate_n_p  : std_logic_vector(2 downto 0);
  signal tristate_n_i  : std_logic_vector(2 downto 0);
  signal data_valid_p  : std_logic;
  signal data_valid_i  : std_logic;
  signal stop_output_i : std_logic;
  signal hold_efb_i    : std_logic;

-- Internal SLink signals:
  signal slink_data_p    : std_logic_vector(31 downto 0); 
  signal slink_data_i    : std_logic_vector(31 downto 0); 
  signal slink_xoff_n_i  : std_logic; 
  signal slink_xoff_i    : std_logic; 
  signal slink_down_n_i  : std_logic; 
  signal slink_down_i    : std_logic; 
  signal slink_rst_n_i   : std_logic; 
  signal slink_test_n_i  : std_logic; 
  signal slink_wen_n_i   : std_logic; 
  signal slink_wen_i     : std_logic; 
  signal slink_uctrl_n_i : std_logic; 

  signal dspe_data_i     : std_logic_vector(31 downto 0);
  signal dspe_wen_i      : std_logic;
  signal dsps_data_i     : std_logic_vector(31 downto 0);
  signal dsps_wen_i      : std_logic;

-- Internal DSP0 Data FIFO signals:  (df = data FIFO)
  signal dsp0_send_data_n_i       : std_logic;
  signal dsp0_sbsram_me_n_i       : std_logic;
  signal dsp0_df_wren_i           : std_logic;
  signal dsp0_df_rden_i           : std_logic;
  signal dsp0_df_dfr_int_mask_i   : std_logic;
  signal dsp0_df_dfr_int_mask_p   : std_logic;
  signal dsp0_df_data_out_i       : std_logic_vector(31 downto 0);
  signal dsp0_data_in_i           : std_logic_vector(31 downto 0);
  signal dsp0_df_empty_flag_out_i : std_logic;
  signal dsp0_df_afull_flag_out_i : std_logic;
  signal dsp0_df_full_flag_out_i  : std_logic;
  signal dsp0_dfrdy_i             : std_logic;
  signal dsp0_dfrdy_p             : std_logic;
  signal dsp0_int4_i              : std_logic;
  signal dsp0_df_bus_en_i         : std_logic;

-- Internal DSP1 Data FIFO signals:  (df = data FIFO)
  signal dsp1_send_data_n_i       : std_logic;
  signal dsp1_sbsram_me_n_i       : std_logic;
  signal dsp1_df_wren_i           : std_logic;
  signal dsp1_df_rden_i           : std_logic;
  signal dsp1_df_dfr_int_mask_i   : std_logic;
  signal dsp1_df_dfr_int_mask_p   : std_logic;
  signal dsp1_df_data_out_i       : std_logic_vector(31 downto 0);
  signal dsp1_data_in_i           : std_logic_vector(31 downto 0);
  signal dsp1_df_full_flag_out_i  : std_logic;
  signal dsp1_df_afull_flag_out_i : std_logic;
  signal dsp1_df_empty_flag_out_i : std_logic;
  signal dsp1_dfrdy_i             : std_logic;
  signal dsp1_dfrdy_p             : std_logic;
  signal dsp1_int4_i              : std_logic;
  signal dsp1_df_bus_en_i         : std_logic;

-- Internal DSP2 Data FIFO signals:  (df = data FIFO)
  signal dsp2_send_data_n_i       : std_logic;
  signal dsp2_sbsram_me_n_i       : std_logic;
  signal dsp2_df_wren_i           : std_logic;
  signal dsp2_df_rden_i           : std_logic;
  signal dsp2_df_dfr_int_mask_i   : std_logic;
  signal dsp2_df_dfr_int_mask_p   : std_logic;
  signal dsp2_df_data_out_i       : std_logic_vector(31 downto 0);
  signal dsp2_data_in_i           : std_logic_vector(31 downto 0);
  signal dsp2_df_full_flag_out_i  : std_logic;
  signal dsp2_df_afull_flag_out_i : std_logic;
  signal dsp2_df_empty_flag_out_i : std_logic;
  signal dsp2_dfrdy_i             : std_logic;
  signal dsp2_dfrdy_p             : std_logic;
  signal dsp2_int4_i              : std_logic;
  signal dsp2_df_bus_en_i         : std_logic;

-- Internal DSP3 Data FIFO signals:  (df = data FIFO)
  signal dsp3_send_data_n_i       : std_logic;
  signal dsp3_sbsram_me_n_i       : std_logic;
  signal dsp3_df_wren_i           : std_logic;
  signal dsp3_df_rden_i           : std_logic;
  signal dsp3_df_dfr_int_mask_i   : std_logic;
  signal dsp3_df_dfr_int_mask_p   : std_logic;
  signal dsp3_df_data_out_i       : std_logic_vector(31 downto 0);
  signal dsp3_data_out_i          : std_logic_vector(31 downto 0);
  signal dsp3_data_in_i           : std_logic_vector(31 downto 0);
  signal dsp3_df_full_flag_out_i  : std_logic;
  signal dsp3_df_afull_flag_out_i : std_logic;
  signal dsp3_df_empty_flag_out_i : std_logic;
  signal dsp3_dfrdy_i             : std_logic;
  signal dsp3_dfrdy_p             : std_logic;
  signal dsp3_int4_i              : std_logic;
  signal dsp3_df_bus_en_i         : std_logic;

-- Internal Router bus signals:
  signal router_bus_data_in_i   : std_logic_vector(15 downto 0);
  signal router_bus_data_out_i  : std_logic_vector(15 downto 0);
  signal router_bus_addr_i      : std_logic_vector( 7 downto 0);
  signal router_bus_rnw_i       : std_logic;
  signal router_bus_strb_n_i    : std_logic;
  signal router_bus_strb_n_ii   : std_logic;
  signal router_bus_ack_i       : std_logic;
  signal router_bus_en_i        : std_logic;

  signal slink_cmd_reg_i        : std_logic_vector( 7 downto 0);
  signal slink_status_reg_i     : std_logic_vector( 2 downto 0);
  signal slink_dump_match_reg_i : std_logic_vector(15 downto 0);
  signal dump_event_i           : std_logic;
  signal L1TT_hw_i              : std_logic_vector( 9 downto 0);
  signal DET_hw_i               : std_logic_vector( 5 downto 0);
  signal hw_capture_i           : std_logic;

  signal dsp0_cmd_reg0_i        : std_logic_vector( 8 downto 0);
  signal dsp0_cmd_reg1_i        : std_logic_vector( 4 downto 0);
  signal dsp0_status_reg_i      : std_logic_vector( 5 downto 0);
  signal dsp0_trap_match0_i     : std_logic_vector( 7 downto 0);
  signal dsp0_prescale0_i       : std_logic_vector(15 downto 0);
  signal dsp0_trap_match1_i     : std_logic_vector( 7 downto 0);
  signal dsp0_prescale1_i       : std_logic_vector(15 downto 0);
  signal dsp0_fifo_word_count_i : std_logic_vector(12 downto 0);
  signal dsp0_int_mask_delay_i  : std_logic_vector( 5 downto 0);
  signal dsp0_trap_rst_i        : std_logic;
  signal dsp0_trap_en_i         : std_logic;
  signal dsp0_trap_en_p         : std_logic;
  signal dsp0_cen_i             : std_logic;
  signal dsp0_trap0_load_i      : std_logic;
  signal dsp0_trap1_load_i      : std_logic;
  signal dsp0_trap0_stat_i      : std_logic;
  signal dsp0_trap1_stat_i      : std_logic;
  signal dsp0_fifo_flush_i      : std_logic;
  
  signal dsp1_cmd_reg0_i        : std_logic_vector( 8 downto 0);
  signal dsp1_cmd_reg1_i        : std_logic_vector( 4 downto 0);
  signal dsp1_status_reg_i      : std_logic_vector( 5 downto 0);
  signal dsp1_trap_match0_i     : std_logic_vector( 7 downto 0);
  signal dsp1_prescale0_i       : std_logic_vector(15 downto 0);
  signal dsp1_trap_match1_i     : std_logic_vector( 7 downto 0);
  signal dsp1_prescale1_i       : std_logic_vector(15 downto 0);
  signal dsp1_fifo_word_count_i : std_logic_vector(12 downto 0);
  signal dsp1_int_mask_delay_i  : std_logic_vector( 5 downto 0);
  signal dsp1_trap_rst_i        : std_logic;
  signal dsp1_trap_en_i         : std_logic;
  signal dsp1_trap_en_p         : std_logic;
  signal dsp1_cen_i             : std_logic;
  signal dsp1_trap0_load_i      : std_logic;
  signal dsp1_trap1_load_i      : std_logic;
  signal dsp1_trap0_stat_i      : std_logic;
  signal dsp1_trap1_stat_i      : std_logic;
  signal dsp1_fifo_flush_i      : std_logic;

  signal dsp2_cmd_reg0_i        : std_logic_vector( 8 downto 0);
  signal dsp2_cmd_reg1_i        : std_logic_vector( 4 downto 0);
  signal dsp2_status_reg_i      : std_logic_vector( 5 downto 0);
  signal dsp2_trap_match0_i     : std_logic_vector( 7 downto 0);
  signal dsp2_prescale0_i       : std_logic_vector(15 downto 0);
  signal dsp2_trap_match1_i     : std_logic_vector( 7 downto 0);
  signal dsp2_prescale1_i       : std_logic_vector(15 downto 0);
  signal dsp2_fifo_word_count_i : std_logic_vector(12 downto 0);
  signal dsp2_int_mask_delay_i  : std_logic_vector( 5 downto 0);
  signal dsp2_trap_rst_i        : std_logic;
  signal dsp2_trap_en_i         : std_logic;
  signal dsp2_trap_en_p         : std_logic;
  signal dsp2_cen_i             : std_logic;
  signal dsp2_trap0_load_i      : std_logic;
  signal dsp2_trap1_load_i      : std_logic;
  signal dsp2_trap0_stat_i      : std_logic;
  signal dsp2_trap1_stat_i      : std_logic;
  signal dsp2_fifo_flush_i      : std_logic;

  signal dsp3_cmd_reg0_i        : std_logic_vector( 8 downto 0);
  signal dsp3_cmd_reg1_i        : std_logic_vector( 4 downto 0);
  signal dsp3_status_reg_i      : std_logic_vector( 5 downto 0);
  signal dsp3_trap_match0_i     : std_logic_vector( 7 downto 0);
  signal dsp3_prescale0_i       : std_logic_vector(15 downto 0);
  signal dsp3_trap_match1_i     : std_logic_vector( 7 downto 0);
  signal dsp3_prescale1_i       : std_logic_vector(15 downto 0);
  signal dsp3_fifo_word_count_i : std_logic_vector(12 downto 0);
  signal dsp3_int_mask_delay_i  : std_logic_vector( 5 downto 0);
  signal dsp3_trap_rst_i        : std_logic;
  signal dsp3_trap_en_i         : std_logic;
  signal dsp3_trap_en_p         : std_logic;
  signal dsp3_cen_i             : std_logic;
  signal dsp3_trap0_load_i      : std_logic;
  signal dsp3_trap1_load_i      : std_logic;
  signal dsp3_trap0_stat_i      : std_logic;
  signal dsp3_trap1_stat_i      : std_logic;
  signal dsp3_fifo_flush_i      : std_logic;

  signal dsp_afull_i : std_logic;
  signal no_errors_i : std_logic;

  signal top_test_out_i  : std_logic_vector(7 downto 0);
  signal top_spare_i     : std_logic_vector(2 downto 0);
  signal top_spare_pin_i : std_logic;
  signal rtr_spare_in_i  : std_logic_vector(13 downto 0);
  signal test_mux_i      : std_logic_vector( 3 downto 0);

  signal router_bus_read  : std_logic;
  signal router_bus_write : std_logic;
  
  signal clk40_locked_i      : std_logic;
  signal clk80_locked_i      : std_logic;
  signal dll_status_i        : std_logic_vector( 1 downto 0);
  signal test_i              : std_logic_vector( 7 downto 0);
  signal rod_type_i          : std_logic;

  signal dsp0_rd_pointer_i   : std_logic_vector( 9 downto 0);
  signal dsp1_rd_pointer_i   : std_logic_vector( 9 downto 0);
  signal dsp2_rd_pointer_i   : std_logic_vector( 9 downto 0);
  signal dsp3_rd_pointer_i   : std_logic_vector( 9 downto 0);
  signal dsp0_dfrdy_stat_i   : std_logic_vector( 3 downto 0);
  signal dsp1_dfrdy_stat_i   : std_logic_vector( 3 downto 0);
  signal dsp2_dfrdy_stat_i   : std_logic_vector( 3 downto 0);
  signal dsp3_dfrdy_stat_i   : std_logic_vector( 3 downto 0);

  signal dsp0_cen_n_i        : std_logic;  
  signal dsp1_cen_n_i        : std_logic;  
  signal dsp2_cen_n_i        : std_logic;  
  signal dsp3_cen_n_i        : std_logic;  

  type sl_reset_states is (
    idle,             -- 0x01
    issue_reset,      -- 0x02
    wait_for_sldown,  -- 0x04
    done              -- 0x08
    );
  signal sl_reset_state  : sl_reset_states;
  signal hold_reset_i    : std_logic;
  signal clr_sl_reset_i  : std_logic;
  signal endian_select_i : std_logic;
  signal force_xoff_i    : std_logic;
  signal sl_fifo_stat_i  : std_logic_vector(2 downto 0);

  signal err_evt_i       : std_logic;
  signal evt_data_wd_i   : std_logic;
  signal slink_dword_i   : std_logic;
  signal rol_test_en_i   : std_logic;  -- Spare Pin(9) from RCF SP(27)

  type hold_next_event_states is (
    idle,            -- 0x01
    hold_off_event,  -- 0x02
    done             -- 0x04
    );
  signal hold_next_event : hold_next_event_states;

  signal hold_next_event_i  : std_logic;
  signal start_hold_delay_i : std_logic;
  signal hold_next_delay_i  : std_logic_vector(15 downto 0);
  signal hold_next_count_i  : unsigned(15 downto 0);
  signal mask_fifo_int4_i   : std_logic_vector(3 downto 0);
  
  signal event_over_threshold : std_logic;
  
  signal fragment_count_i     : std_logic_vector(31 downto 0);
  signal hit_count_i          : std_logic_vector(31 downto 0);
  signal to_error_count_i     : std_logic_vector(31 downto 0);
  signal eoc_over_count_i     : std_logic_vector(31 downto 0);
  signal afull_count_i        : std_logic_vector(31 downto 0);
  signal rc_error_count_i     : std_logic_vector(31 downto 0);
  signal sync_err_count_i     : std_logic_vector(31 downto 0);
  signal bcid_chkfail_count_i : std_logic_vector(31 downto 0);
  signal l1_chkfail_count_i   : std_logic_vector(31 downto 0);
  signal hitoverflow_count_i  : std_logic_vector(31 downto 0);
  signal regparity_count_i    : std_logic_vector(31 downto 0);
  signal bitflip_count_i      : std_logic_vector(31 downto 0);
  signal hammingcode_count_i  : std_logic_vector(31 downto 0);
  signal total_error_count_i  : std_logic_vector(31 downto 0);
  signal eot_count_i          : std_logic_vector(31 downto 0);

  signal stat_count_crtl_i    : std_logic_vector( 1 downto 0);


-------------------------------------------------------------------------------
-- COMPONENT DECLARATION
-------------------------------------------------------------------------------

-- The write over link header block looks at the event_data from the EFB,
-- identifies the link headers, and writes the link and error information
-- into specific bits of the link header.  The data output is input into 
-- a large array of pipeline registers that delays the S-Link data until 
-- the event header is read into the Router and analyzed to determine if 
-- the Event Fragment should be transmitted to the S-Link bus or dumped 
-- on the floor.  It is also used in the same way for the DSP data

component write_over_link_header
  port (
    clk40_in               : in  std_logic; -- 40MHz clock input
    rst_n_in               : in  std_logic; -- active low reset
    event_data_in          : in  std_logic_vector(45 downto 0);
    data_source_n_in       : in  std_logic_vector( 2 downto 0);
    data_valid_in          : in  std_logic; -- data valid from EFB
    slink_wen_out          : out std_logic; --
    slink_data_out         : out std_logic_vector(31 downto 0);
    slink_uctrl_n_out      : out std_logic;
    slink_dword_out        : out std_logic;
    dsps_data_out          : out std_logic_vector(31 downto 0);
    dsps_wen_out           : out std_logic; --
    dspe_data_out          : out std_logic_vector(31 downto 0);
    dspe_wen_out           : out std_logic; --
    L1TT_hw_out            : out std_logic_vector( 9 downto 0);
    DET_hw_out             : out std_logic_vector( 5 downto 0);
    ERR_hw_out             : out std_logic;
    EOT_hw_out             : out std_logic;
    hw_capture_out         : out std_logic;
    event_ok               : out std_logic;
    slink_xoff_in          : in  std_logic;
    hold_efb_out           : out std_logic;
    sl_fifo_stat_out       : out std_logic_vector(2 downto 0);
    rol_test_en_in         : in  std_logic;
    fragment_count_out     : out std_logic_vector(31 downto 0);
    hit_count_out          : out std_logic_vector(31 downto 0);
    to_error_count_out     : out std_logic_vector(31 downto 0);
    eoc_over_count_out     : out std_logic_vector(31 downto 0);
    afull_count_out        : out std_logic_vector(31 downto 0);
    rc_error_count_out     : out std_logic_vector(31 downto 0);
    sync_err_count_out     : out std_logic_vector(31 downto 0); -- (Bit 19 in the Error Flags Status Word 0)
    bcid_chkfail_count_out : out std_logic_vector(31 downto 0); -- (Bit 30)
    l1_chkfail_count_out   : out std_logic_vector(31 downto 0); -- (Bit 31)
    hitoverflow_count_out  : out std_logic_vector(31 downto 0); -- (Bit 27)
    regparity_count_out    : out std_logic_vector(31 downto 0); -- (Bit 24)
    bitflip_count_out      : out std_logic_vector(31 downto 0); -- (Bit 26)
    hammingcode_count_out  : out std_logic_vector(31 downto 0); -- (Bit 23)
    total_error_count_out  : out std_logic_vector(31 downto 0); -- (Bit 23)
    eot_count_out          : out std_logic_vector(31 downto 0); -- (Bit 18)
    stat_count_crtl_in     : in  std_logic_vector( 1 downto 0)
    );
end component;

-- The dump data event block reads the event header and determines if the 
-- Event Fragmentshould be transmitted to the S-Link bus or dumped on the 
-- floor. This block only controls "slink_wen_n_out" and does not actually 
-- affect the data in any way.
component dump_event_data
  port (
    clk40_in          : in  std_logic;
    rst_n_in          : in  std_logic;
    slink_cmd_reg_in  : in  std_logic_vector( 5 downto 0);
--  slink_status_in   : in  std_logic_vector( 1 downto 0);
    dump_match_reg_in : in  std_logic_vector(15 downto 0);
    L1TT_hw_in        : in  std_logic_vector( 9 downto 0);
    DET_hw_in         : in  std_logic_vector( 5 downto 0);
    hw_capture_in     : in  std_logic;
    dump_event_out    : out std_logic
    );
end component;

-- The trap data event block compares the trap event type registers of 
-- the router to the incoming data to determine if an event should be 
-- trapped in the Data FIFOs.
component trap_event_data
  port (
    clk40_in            : in  std_logic;
    clk80_in            : in  std_logic;  -- sbsram clk 80MHz
    rst_n_in            : in  std_logic;
    fifo_flush_in       : in  std_logic;
    trap_reset_in       : in  std_logic;
    trap0_load_in       : in  std_logic;
    trap1_load_in       : in  std_logic;
    trap_enable_in      : in  std_logic;
    mask_dfr_in         : in  std_logic;
    hw_capture_in       : in  std_logic;  -- flag to indicate that the header words have been captured
    dsp_cmd_reg0_in     : in  std_logic_vector( 8 downto 0);
    dsp_cmd_reg1_in     : in  std_logic_vector( 4 downto 0);
    dsp_trap_match0_in  : in  std_logic_vector( 7 downto 0);
    dsp_prescale0_in    : in  std_logic_vector(15 downto 0);
    dsp_trap_match1_in  : in  std_logic_vector( 7 downto 0);
    dsp_prescale1_in    : in  std_logic_vector(15 downto 0);
    int_mask_delay_in   : in  std_logic_vector( 5 downto 0);
    atlas_tt_in         : in  std_logic_vector( 7 downto 0);
    tim_tt_in           : in  std_logic_vector( 1 downto 0);
    rod_tt_in           : in  std_logic_vector( 5 downto 0);
    err_evt_in          : in  std_logic;
    eot_evt_in          : in  std_logic;
    fifo_cen_in         : in  std_logic; -- CE from DSP
    fifo_rden_in        : in  std_logic; -- OE from DSP
    fifo_wren_in        : in  std_logic; -- enable writes to FIFO
    data_in             : in  std_logic_vector(31 downto 0);
    data_out            : out std_logic_vector(31 downto 0);
    word_count_out      : out std_logic_vector(12 downto 0);
    full_flag_out       : out std_logic; --  fifo full flag
    afull_flag_out      : out std_logic; --  fifo almost full flag
    empty_flag_out      : out std_logic; --  fifo empty flag
    data_frame_rdy_out  : out std_logic; --  data frame ready
    trap0_status_out    : out std_logic; --  Status of Trap0 1=Trap On
    trap1_status_out    : out std_logic; --  Status of Trap1 1=Trap On
    event_ok_in         : in  std_logic;
    fifo_rd_pointer_out : out std_logic_vector( 9 downto 0);
    data_frame_rdy_stat : out std_logic_vector( 3 downto 0); 
    mask_fifo_int4_in   : in  std_logic
    );
end component;

--
-- The register block is the register map of the Router
component register_block
  port (
    clk40_in                     : in     std_logic; -- clk40 input
    rst_n_in                     : in     std_logic; -- async global reset
    router_bus_strb_n_in         : in     std_logic;
    router_bus_rnw_in            : in     std_logic;
    router_bus_adr_in            : in     std_logic_vector( 7 downto 0);
    router_bus_data_in           : in     std_logic_vector(15 downto 0);
    router_bus_data_out          : out    std_logic_vector(15 downto 0);
    router_bus_ack_out           : out    std_logic;
    global_cmd_reg_out           : out    std_logic_vector( 7 downto 0);
    slink_status_reg_in          : in     std_logic_vector( 2 downto 0);
    slink_AT_dump_match_reg_out  : out    std_logic_vector( 9 downto 0);
    slink_ROD_dump_match_reg_out : out    std_logic_vector( 5 downto 0);
    dsp0_cmd_reg0_out            : buffer std_logic_vector( 8 downto 0);
    dsp0_cmd_reg1_out            : buffer std_logic_vector( 4 downto 0);
    dsp0_fifo_flush_out          : out    std_logic;
    dsp0_trap_reset_out          : out    std_logic;
    dsp0_trap0_load_out          : out    std_logic;
    dsp0_trap1_load_out          : out    std_logic;
    dsp0_status_reg_in           : in     std_logic_vector( 5 downto 0);
    dsp0_trap_match0_out         : buffer std_logic_vector( 7 downto 0); 
    dsp0_prescale0_out           : buffer std_logic_vector(15 downto 0); 
    dsp0_trap_match1_out         : buffer std_logic_vector( 7 downto 0); 
    dsp0_prescale1_out           : buffer std_logic_vector(15 downto 0); 
    dsp0_fifo_word_count_reg_in  : in     std_logic_vector(12 downto 0);
    dsp0_int_mask_delay_out      : out    std_logic_vector( 5 downto 0); 
    dsp1_cmd_reg0_out            : buffer std_logic_vector( 8 downto 0);
    dsp1_cmd_reg1_out            : buffer std_logic_vector( 4 downto 0);
    dsp1_fifo_flush_out          : out    std_logic;
    dsp1_trap_reset_out          : out    std_logic;
    dsp1_trap0_load_out          : out    std_logic;
    dsp1_trap1_load_out          : out    std_logic;
    dsp1_status_reg_in           : in     std_logic_vector( 5 downto 0);
    dsp1_trap_match0_out         : buffer std_logic_vector( 7 downto 0); 
    dsp1_prescale0_out           : buffer std_logic_vector(15 downto 0); 
    dsp1_trap_match1_out         : buffer std_logic_vector( 7 downto 0); 
    dsp1_prescale1_out           : buffer std_logic_vector(15 downto 0); 
    dsp1_fifo_word_count_reg_in  : in     std_logic_vector(12 downto 0);
    dsp1_int_mask_delay_out      : out    std_logic_vector( 5 downto 0); 
    dsp2_cmd_reg0_out            : buffer std_logic_vector( 8 downto 0);
    dsp2_cmd_reg1_out            : buffer std_logic_vector( 4 downto 0);
    dsp2_fifo_flush_out          : out    std_logic;
    dsp2_trap_reset_out          : out    std_logic;
    dsp2_trap0_load_out          : out    std_logic;
    dsp2_trap1_load_out          : out    std_logic;
    dsp2_status_reg_in           : in     std_logic_vector( 5 downto 0);
    dsp2_trap_match0_out         : buffer std_logic_vector( 7 downto 0); 
    dsp2_prescale0_out           : buffer std_logic_vector(15 downto 0); 
    dsp2_trap_match1_out         : buffer std_logic_vector( 7 downto 0); 
    dsp2_prescale1_out           : buffer std_logic_vector(15 downto 0); 
    dsp2_fifo_word_count_reg_in  : in     std_logic_vector(12 downto 0);
    dsp2_int_mask_delay_out      : out    std_logic_vector( 5 downto 0); 
    dsp3_cmd_reg0_out            : buffer std_logic_vector( 8 downto 0);
    dsp3_cmd_reg1_out            : buffer std_logic_vector( 4 downto 0);
    dsp3_fifo_flush_out          : out    std_logic;
    dsp3_trap_reset_out          : out    std_logic;
    dsp3_trap0_load_out          : out    std_logic;
    dsp3_trap1_load_out          : out    std_logic;
    dsp3_status_reg_in           : in     std_logic_vector( 5 downto 0);
    dsp3_trap_match0_out         : buffer std_logic_vector( 7 downto 0); 
    dsp3_prescale0_out           : buffer std_logic_vector(15 downto 0); 
    dsp3_trap_match1_out         : buffer std_logic_vector( 7 downto 0); 
    dsp3_prescale1_out           : buffer std_logic_vector(15 downto 0); 
    dsp3_fifo_word_count_reg_in  : in     std_logic_vector(12 downto 0);
    dsp3_int_mask_delay_out      : out    std_logic_vector( 5 downto 0);
    dll_status_in                : in     std_logic_vector( 1 downto 0);
    rtr_dsp_spares_in            : in     std_logic_vector(13 downto 0);
    rod_type_in                  : in     std_logic;
    test_mux_out                 : out    std_logic_vector( 3 downto 0);
    dsp0_fifo_rd_pointer_in      : in     std_logic_vector( 9 downto 0);
    dsp1_fifo_rd_pointer_in      : in     std_logic_vector( 9 downto 0);
    dsp2_fifo_rd_pointer_in      : in     std_logic_vector( 9 downto 0);
    dsp3_fifo_rd_pointer_in      : in     std_logic_vector( 9 downto 0);
    dsp0_dfrdy_stat_in           : in     std_logic_vector( 3 downto 0);
    dsp1_dfrdy_stat_in           : in     std_logic_vector( 3 downto 0);
    dsp2_dfrdy_stat_in           : in     std_logic_vector( 3 downto 0);
    dsp3_dfrdy_stat_in           : in     std_logic_vector( 3 downto 0);
    clr_sl_reset_in              : in     std_logic;
    dsp0_int4_out                : out    std_logic;
    dsp1_int4_out                : out    std_logic;
    dsp2_int4_out                : out    std_logic;
    dsp3_int4_out                : out    std_logic;
    mask_fifo_int4_out           : out    std_logic_vector( 3 downto 0);
    endian_select_out            : out    std_logic;
    force_xoff_out               : out    std_logic;
    sl_fifo_stat_in              : in     std_logic_vector( 2 downto 0);
    hold_next_delay_out          : out    std_logic_vector(15 downto 0);
    slink_data_valid_in          : in     std_logic;
    fragment_count_in            : in     std_logic_vector(31 downto 0);
    hit_count_in                 : in     std_logic_vector(31 downto 0);
    to_error_count_in            : in     std_logic_vector(31 downto 0);
    eoc_over_count_in            : in     std_logic_vector(31 downto 0);
    afull_count_in               : in     std_logic_vector(31 downto 0);
    rc_error_count_in            : in     std_logic_vector(31 downto 0);
    sync_err_count_in            : in     std_logic_vector(31 downto 0);
    bcid_chkfail_count_in        : in     std_logic_vector(31 downto 0);
    l1_chkfail_count_in          : in     std_logic_vector(31 downto 0);
    hitoverflow_count_in         : in     std_logic_vector(31 downto 0);
    regparity_count_in           : in     std_logic_vector(31 downto 0);
    bitflip_count_in             : in     std_logic_vector(31 downto 0);
    hammingcode_count_in         : in     std_logic_vector(31 downto 0);
    total_error_count_in         : in     std_logic_vector(31 downto 0);
    eot_count_in                 : in     std_logic_vector(31 downto 0);
    status_counter_crtl_out      : out    std_logic_vector( 1 downto 0)
    );   
end component;

-------------------------------------------------------------------------------
-- PAD DECLARATIONS (Xilinx specific I/O pads)
-------------------------------------------------------------------------------

component IBUFG  -- Global clock input pad
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC
    );
end component;

component BUFG  -- Global clock buffer
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC
    );
end component;

component IBUF  -- Input pad
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC
    );
end component;

component OBUF  -- Output pad
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC
    );
end component;

component OBUF_F_12  -- Output pad
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC
    );
end component;

component OBUFT  -- Tristate output pad
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC;
    T : in  STD_LOGIC
    );
end component;

component OBUFT_F_12  -- Tristate output pad, fast slew, 12mA drive
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC;
    T : in  STD_LOGIC
    );
end component;

component OBUFT_F_24  -- Tristate output pad, fast slew, 24mA drive
  port (
    I : in  STD_LOGIC;
    O : out STD_LOGIC;
    T : in  STD_LOGIC
    );
end component;

component IOBUF  -- Tristate I/O pad
  port (
    I  : in    STD_LOGIC;
    O  : out   STD_LOGIC;
    IO : inout STD_LOGIC;
    T  : in    STD_LOGIC
    );
end component;

component CLKDLL  -- Clock delay locked loop
  port (
    CLK0   : out STD_LOGIC;
    CLK90  : out STD_LOGIC;
    CLK180 : out STD_LOGIC;
    CLK270 : out STD_LOGIC;
    CLK2X  : out STD_LOGIC;
    CLKDV  : out STD_LOGIC;
    LOCKED : out STD_LOGIC;
    CLKIN  : in  STD_LOGIC;
    CLKFB  : in  STD_LOGIC;
    RST    : in  STD_LOGIC
    );
end component;

-------------------------------------------------------------------------------
-- XILINX PARTS SPECIFIC COMPONENTS/SIGNALS
-------------------------------------------------------------------------------

component STARTUP_VIRTEX
  port (
    GSR : in std_logic;
    GTS : in std_logic;
    CLK : in std_logic
    );
end component;

begin
--------------------------------------------------------------------------
-- COMPONENT INSTANTIATION
--------------------------------------------------------------------------

U1 : write_over_link_header
  port map (
    clk40_in               => clk40,
    rst_n_in               => rst_n,
    event_data_in          => event_info_i,
    data_source_n_in       => tristate_n_i,
    data_valid_in          => data_valid_i,
    slink_wen_out          => slink_wen_i,
    slink_data_out         => slink_data_i,
    slink_uctrl_n_out      => slink_uctrl_n_i,
    slink_dword_out        => slink_dword_i,
    dsps_data_out          => dsps_data_i,
    dsps_wen_out           => dsps_wen_i,
    dspe_data_out          => dspe_data_i,
    dspe_wen_out           => dspe_wen_i,
    L1TT_hw_out            => L1TT_hw_i,
    DET_hw_out             => DET_hw_i,
    ERR_hw_out             => err_evt_i,
    EOT_hw_out             => event_over_threshold,
    hw_capture_out         => hw_capture_i,
    event_ok               => no_errors_i,
    slink_xoff_in          => slink_xoff_i,
    hold_efb_out           => hold_efb_i,
    sl_fifo_stat_out       => sl_fifo_stat_i,
    rol_test_en_in         => rol_test_en_i,
    fragment_count_out     => fragment_count_i,
    hit_count_out          => hit_count_i,
    to_error_count_out     => to_error_count_i,
    eoc_over_count_out     => eoc_over_count_i,
    afull_count_out        => afull_count_i,
    rc_error_count_out     => rc_error_count_i,
    sync_err_count_out     => sync_err_count_i,
    bcid_chkfail_count_out => bcid_chkfail_count_i,
    l1_chkfail_count_out   => l1_chkfail_count_i,
    hitoverflow_count_out  => hitoverflow_count_i,
    regparity_count_out    => regparity_count_i,
    bitflip_count_out      => bitflip_count_i,
    hammingcode_count_out  => hammingcode_count_i,
    total_error_count_out  => total_error_count_i,
    eot_count_out          => eot_count_i,
    stat_count_crtl_in     => stat_count_crtl_i
    );

U2 : dump_event_data
  port map (
    clk40_in          => clk40,
    rst_n_in          => rst_n,
    slink_cmd_reg_in  => slink_cmd_reg_i(5 downto 0),
--  slink_status_in   => slink_status_reg_i(1 downto 0),
    dump_match_reg_in => slink_dump_match_reg_i,
    L1TT_hw_in        => L1TT_hw_i,
    DET_hw_in         => DET_hw_i,
    hw_capture_in     => hw_capture_i,
    dump_event_out    => dump_event_i
    );

U3 : trap_event_data  -- SDSP0
  port map (
    clk40_in            => clk40,
    clk80_in            => clk80,
    rst_n_in            => rst_n,
    fifo_flush_in       => dsp0_fifo_flush_i,
    trap_reset_in       => dsp0_trap_rst_i,
    trap0_load_in       => dsp0_trap0_load_i,
    trap1_load_in       => dsp0_trap1_load_i,
    trap_enable_in      => dsp0_trap_en_i,
    mask_dfr_in         => dsp0_df_dfr_int_mask_i,
    hw_capture_in       => hw_capture_i,
    dsp_cmd_reg0_in     => dsp0_cmd_reg0_i,
    dsp_cmd_reg1_in     => dsp0_cmd_reg1_i,
    dsp_trap_match0_in  => dsp0_trap_match0_i,
    dsp_prescale0_in    => dsp0_prescale0_i,
    dsp_trap_match1_in  => dsp0_trap_match1_i,
    dsp_prescale1_in    => dsp0_prescale1_i,
    int_mask_delay_in   => dsp0_int_mask_delay_i,
    atlas_tt_in         => L1TT_hw_i(7 downto 0),
    tim_tt_in           => L1TT_hw_i(9 downto 8),
    rod_tt_in           => DET_hw_i,
    err_evt_in          => err_evt_i,
    eot_evt_in          => event_over_threshold,
    fifo_cen_in         => dsp0_cen_i,
    fifo_rden_in        => dsp0_df_rden_i,
    fifo_wren_in        => dsp0_df_wren_i,
    data_in             => dsp0_data_in_i,
    data_out            => dsp0_df_data_out_i,
    word_count_out      => dsp0_fifo_word_count_i,
    full_flag_out       => dsp0_df_full_flag_out_i,
    afull_flag_out      => dsp0_df_afull_flag_out_i,
    empty_flag_out      => dsp0_df_empty_flag_out_i,
    data_frame_rdy_out  => dsp0_dfrdy_i,
    trap0_status_out    => dsp0_trap0_stat_i,
    trap1_status_out    => dsp0_trap1_stat_i,
    event_ok_in         => no_errors_i,
    fifo_rd_pointer_out => dsp0_rd_pointer_i,
    data_frame_rdy_stat => dsp0_dfrdy_stat_i,
    mask_fifo_int4_in   => mask_fifo_int4_i(0)
    );

U4 : trap_event_data  -- SDSP1
  port map (
    clk40_in            => clk40,
    clk80_in            => clk80,
    rst_n_in            => rst_n,
    fifo_flush_in       => dsp1_fifo_flush_i,
    trap_reset_in       => dsp1_trap_rst_i,
    trap0_load_in       => dsp1_trap0_load_i,
    trap1_load_in       => dsp1_trap1_load_i,
    trap_enable_in      => dsp1_trap_en_i,
    mask_dfr_in         => dsp1_df_dfr_int_mask_i,
    hw_capture_in       => hw_capture_i,
    dsp_cmd_reg0_in     => dsp1_cmd_reg0_i,
    dsp_cmd_reg1_in     => dsp1_cmd_reg1_i,
    dsp_trap_match0_in  => dsp1_trap_match0_i,
    dsp_prescale0_in    => dsp1_prescale0_i,
    dsp_trap_match1_in  => dsp1_trap_match1_i,
    dsp_prescale1_in    => dsp1_prescale1_i,
    int_mask_delay_in   => dsp1_int_mask_delay_i,
    atlas_tt_in         => L1TT_hw_i(7 downto 0),
    tim_tt_in           => L1TT_hw_i(9 downto 8),
    rod_tt_in           => DET_hw_i,
    err_evt_in          => err_evt_i,
    eot_evt_in          => event_over_threshold,
    fifo_cen_in         => dsp1_cen_i,
    fifo_rden_in        => dsp1_df_rden_i,
    fifo_wren_in        => dsp1_df_wren_i,
    data_in             => dsp1_data_in_i,
    data_out            => dsp1_df_data_out_i,
    word_count_out      => dsp1_fifo_word_count_i,
    full_flag_out       => dsp1_df_full_flag_out_i,
    afull_flag_out      => dsp1_df_afull_flag_out_i,
    empty_flag_out      => dsp1_df_empty_flag_out_i,
    data_frame_rdy_out  => dsp1_dfrdy_i,
    trap0_status_out    => dsp1_trap0_stat_i,
    trap1_status_out    => dsp1_trap1_stat_i,
    event_ok_in         => no_errors_i,
    fifo_rd_pointer_out => dsp1_rd_pointer_i,
    data_frame_rdy_stat => dsp1_dfrdy_stat_i,
    mask_fifo_int4_in   => mask_fifo_int4_i(1)
    );

U5 : trap_event_data  -- SDSP2
  port map (
    clk40_in            => clk40,
    clk80_in            => clk80,
    rst_n_in            => rst_n,
    fifo_flush_in       => dsp2_fifo_flush_i,
    trap_reset_in       => dsp2_trap_rst_i,
    trap0_load_in       => dsp2_trap0_load_i,
    trap1_load_in       => dsp2_trap1_load_i,
    trap_enable_in      => dsp2_trap_en_i,
    mask_dfr_in         => dsp2_df_dfr_int_mask_i,
    hw_capture_in       => hw_capture_i,
    dsp_cmd_reg0_in     => dsp2_cmd_reg0_i,
    dsp_cmd_reg1_in     => dsp2_cmd_reg1_i,
    dsp_trap_match0_in  => dsp2_trap_match0_i,
    dsp_prescale0_in    => dsp2_prescale0_i,
    dsp_trap_match1_in  => dsp2_trap_match1_i,
    dsp_prescale1_in    => dsp2_prescale1_i,
    int_mask_delay_in   => dsp2_int_mask_delay_i,
    atlas_tt_in         => L1TT_hw_i(7 downto 0),
    tim_tt_in           => L1TT_hw_i(9 downto 8),
    rod_tt_in           => DET_hw_i,
    err_evt_in          => err_evt_i,
    eot_evt_in          => event_over_threshold,
    fifo_cen_in         => dsp2_cen_i,
    fifo_rden_in        => dsp2_df_rden_i,
    fifo_wren_in        => dsp2_df_wren_i,
    data_in             => dsp2_data_in_i,
    data_out            => dsp2_df_data_out_i,
    word_count_out      => dsp2_fifo_word_count_i,
    full_flag_out       => dsp2_df_full_flag_out_i,
    afull_flag_out      => dsp2_df_afull_flag_out_i,
    empty_flag_out      => dsp2_df_empty_flag_out_i,
    data_frame_rdy_out  => dsp2_dfrdy_i,
    trap0_status_out    => dsp2_trap0_stat_i,
    trap1_status_out    => dsp2_trap1_stat_i,
    event_ok_in         => no_errors_i,
    fifo_rd_pointer_out => dsp2_rd_pointer_i,
    data_frame_rdy_stat => dsp2_dfrdy_stat_i,
    mask_fifo_int4_in   => mask_fifo_int4_i(2)
    );

U6 : trap_event_data  -- SDSP3
  port map (
    clk40_in            => clk40,
    clk80_in            => clk80,
    rst_n_in            => rst_n,
    fifo_flush_in       => dsp3_fifo_flush_i,
    trap_reset_in       => dsp3_trap_rst_i,
    trap0_load_in       => dsp3_trap0_load_i,
    trap1_load_in       => dsp3_trap1_load_i,
    trap_enable_in      => dsp3_trap_en_i,
    mask_dfr_in         => dsp3_df_dfr_int_mask_i,
    hw_capture_in       => hw_capture_i,
    dsp_cmd_reg0_in     => dsp3_cmd_reg0_i,
    dsp_cmd_reg1_in     => dsp3_cmd_reg1_i,
    dsp_trap_match0_in  => dsp3_trap_match0_i,
    dsp_prescale0_in    => dsp3_prescale0_i,
    dsp_trap_match1_in  => dsp3_trap_match1_i,
    dsp_prescale1_in    => dsp3_prescale1_i,
    int_mask_delay_in   => dsp3_int_mask_delay_i,
    atlas_tt_in         => L1TT_hw_i(7 downto 0),
    tim_tt_in           => L1TT_hw_i(9 downto 8),
    rod_tt_in           => DET_hw_i,
    err_evt_in          => err_evt_i,
    eot_evt_in          => event_over_threshold,
    fifo_cen_in         => dsp3_cen_i,
    fifo_rden_in        => dsp3_df_rden_i,
    fifo_wren_in        => dsp3_df_wren_i,
    data_in             => dsp3_data_in_i,
    data_out            => dsp3_df_data_out_i,
    word_count_out      => dsp3_fifo_word_count_i,
    full_flag_out       => dsp3_df_full_flag_out_i,
    afull_flag_out      => dsp3_df_afull_flag_out_i,
    empty_flag_out      => dsp3_df_empty_flag_out_i,
    data_frame_rdy_out  => dsp3_dfrdy_i,
    trap0_status_out    => dsp3_trap0_stat_i,
    trap1_status_out    => dsp3_trap1_stat_i,
    event_ok_in         => no_errors_i,
    fifo_rd_pointer_out => dsp3_rd_pointer_i,
    data_frame_rdy_stat => dsp3_dfrdy_stat_i,
    mask_fifo_int4_in   => mask_fifo_int4_i(3)
    );

U7 : register_block
  port map (
    clk40_in                     => clk40,
    rst_n_in                     => rst_n,
    router_bus_strb_n_in         => router_bus_strb_n_ii,
    router_bus_rnw_in            => router_bus_rnw_i, 
    router_bus_adr_in            => router_bus_addr_i,
    router_bus_data_in           => router_bus_data_in_i,
    router_bus_data_out          => router_bus_data_out_i,
    router_bus_ack_out           => router_bus_ack_i,
    global_cmd_reg_out           => slink_cmd_reg_i,
    slink_status_reg_in          => slink_status_reg_i,
    slink_AT_dump_match_reg_out  => slink_dump_match_reg_i( 9 downto 0),
    slink_ROD_dump_match_reg_out => slink_dump_match_reg_i(15 downto 10),
    dsp0_cmd_reg0_out            => dsp0_cmd_reg0_i,
    dsp0_cmd_reg1_out            => dsp0_cmd_reg1_i,
    dsp0_fifo_flush_out          => dsp0_fifo_flush_i,
    dsp0_trap_reset_out          => dsp0_trap_rst_i,
    dsp0_trap0_load_out          => dsp0_trap0_load_i,
    dsp0_trap1_load_out          => dsp0_trap1_load_i,
    dsp0_status_reg_in           => dsp0_status_reg_i,
    dsp0_trap_match0_out         => dsp0_trap_match0_i,
    dsp0_prescale0_out           => dsp0_prescale0_i,
    dsp0_trap_match1_out         => dsp0_trap_match1_i,
    dsp0_prescale1_out           => dsp0_prescale1_i,
    dsp0_fifo_word_count_reg_in  => dsp0_fifo_word_count_i,
    dsp0_int_mask_delay_out      => dsp0_int_mask_delay_i,
    dsp1_cmd_reg0_out            => dsp1_cmd_reg0_i,
    dsp1_cmd_reg1_out            => dsp1_cmd_reg1_i,
    dsp1_fifo_flush_out          => dsp1_fifo_flush_i,
    dsp1_trap_reset_out          => dsp1_trap_rst_i,
    dsp1_trap0_load_out          => dsp1_trap0_load_i,
    dsp1_trap1_load_out          => dsp1_trap1_load_i,
    dsp1_status_reg_in           => dsp1_status_reg_i,
    dsp1_trap_match0_out         => dsp1_trap_match0_i,
    dsp1_prescale0_out           => dsp1_prescale0_i,
    dsp1_trap_match1_out         => dsp1_trap_match1_i,
    dsp1_prescale1_out           => dsp1_prescale1_i,
    dsp1_fifo_word_count_reg_in  => dsp1_fifo_word_count_i,
    dsp1_int_mask_delay_out      => dsp1_int_mask_delay_i,
    dsp2_cmd_reg0_out            => dsp2_cmd_reg0_i,
    dsp2_cmd_reg1_out            => dsp2_cmd_reg1_i,
    dsp2_fifo_flush_out          => dsp2_fifo_flush_i,
    dsp2_trap_reset_out          => dsp2_trap_rst_i,
    dsp2_trap0_load_out          => dsp2_trap0_load_i,
    dsp2_trap1_load_out          => dsp2_trap1_load_i,
    dsp2_status_reg_in           => dsp2_status_reg_i,
    dsp2_trap_match0_out         => dsp2_trap_match0_i,
    dsp2_prescale0_out           => dsp2_prescale0_i,
    dsp2_trap_match1_out         => dsp2_trap_match1_i,
    dsp2_prescale1_out           => dsp2_prescale1_i,
    dsp2_fifo_word_count_reg_in  => dsp2_fifo_word_count_i,
    dsp2_int_mask_delay_out      => dsp2_int_mask_delay_i,
    dsp3_cmd_reg0_out            => dsp3_cmd_reg0_i,
    dsp3_cmd_reg1_out            => dsp3_cmd_reg1_i,
    dsp3_fifo_flush_out          => dsp3_fifo_flush_i,
    dsp3_trap_reset_out          => dsp3_trap_rst_i,
    dsp3_trap0_load_out          => dsp3_trap0_load_i,
    dsp3_trap1_load_out          => dsp3_trap1_load_i,
    dsp3_status_reg_in           => dsp3_status_reg_i,
    dsp3_trap_match0_out         => dsp3_trap_match0_i,
    dsp3_prescale0_out           => dsp3_prescale0_i,
    dsp3_trap_match1_out         => dsp3_trap_match1_i,
    dsp3_prescale1_out           => dsp3_prescale1_i,
    dsp3_fifo_word_count_reg_in  => dsp3_fifo_word_count_i,
    dsp3_int_mask_delay_out      => dsp3_int_mask_delay_i,
    dll_status_in                => dll_status_i,
    rtr_dsp_spares_in            => rtr_spare_in_i,
    rod_type_in                  => rod_type_i,
    test_mux_out                 => test_mux_i,
    dsp0_fifo_rd_pointer_in      => dsp0_rd_pointer_i,
    dsp1_fifo_rd_pointer_in      => dsp1_rd_pointer_i,
    dsp2_fifo_rd_pointer_in      => dsp2_rd_pointer_i,
    dsp3_fifo_rd_pointer_in      => dsp3_rd_pointer_i,
    dsp0_dfrdy_stat_in           => dsp0_dfrdy_stat_i,
    dsp1_dfrdy_stat_in           => dsp1_dfrdy_stat_i,
    dsp2_dfrdy_stat_in           => dsp2_dfrdy_stat_i,
    dsp3_dfrdy_stat_in           => dsp3_dfrdy_stat_i,
    clr_sl_reset_in              => clr_sl_reset_i,
    dsp0_int4_out                => dsp0_int4_i,
    dsp1_int4_out                => dsp1_int4_i,
    dsp2_int4_out                => dsp2_int4_i,
    dsp3_int4_out                => dsp3_int4_i,
    mask_fifo_int4_out           => mask_fifo_int4_i,
    endian_select_out            => endian_select_i,
    force_xoff_out               => force_xoff_i,
    sl_fifo_stat_in              => sl_fifo_stat_i,
    hold_next_delay_out          => hold_next_delay_i,
    slink_data_valid_in          => slink_wen_i,
    fragment_count_in            => fragment_count_i,
    hit_count_in                 => hit_count_i,
    to_error_count_in            => to_error_count_i,
    eoc_over_count_in            => eoc_over_count_i,
    afull_count_in               => afull_count_i,
    rc_error_count_in            => rc_error_count_i,
    sync_err_count_in            => sync_err_count_i,
    bcid_chkfail_count_in        => bcid_chkfail_count_i,
    l1_chkfail_count_in          => l1_chkfail_count_i,
    hitoverflow_count_in         => hitoverflow_count_i,
    regparity_count_in           => regparity_count_i,
    bitflip_count_in             => bitflip_count_i,
    hammingcode_count_in         => hammingcode_count_i,
    total_error_count_in         => total_error_count_i,
    eot_count_in                 => eot_count_i,
    status_counter_crtl_out      => stat_count_crtl_i
    );

dsp0_dfrdy_p <= dsp0_dfrdy_i OR dsp0_int4_i;
dsp1_dfrdy_p <= dsp1_dfrdy_i OR dsp1_int4_i;
dsp2_dfrdy_p <= dsp2_dfrdy_i OR dsp2_int4_i;
dsp3_dfrdy_p <= dsp3_dfrdy_i OR dsp3_int4_i;

-------------------------------------------------------------------------------
-- XILINX PARTS SPECIFIC COMPONENTS/SIGNALS
-------------------------------------------------------------------------------
U8 : STARTUP_VIRTEX
  port map (
    GSR => rst,
    GTS => rst,
    CLK => clk40
    );

-------------------------------------------------------------------------------
-- I/0 Buffer / Clock DLL Instantiation
-------------------------------------------------------------------------------
reset_ibuf : IBUF 
  port map (I => rst_n_in, O => rst_n);

-- Main Clock (40MHz)
clk40_ibufg : IBUFG
  port map (I => clk40_in, O => clk40_dll_in);

clk40_bufg : BUFG
  port map (I => clk40_dll_out, O => clk40);

clk40_dll : CLKDLL 
  port map (
    CLK0   => clk40_dll_out,
    CLK90  => open,
    CLK180 => open,
    CLK270 => open,
    CLK2X  => open,
    CLKDV  => open,
    LOCKED => clk40_locked_i,
    CLKIN  => clk40_dll_in,
    CLKFB  => clk40,
    RST    => gnd
    );

-- FIFO Readout/DSP Interface Clock (80MHz)
clk80_ibufg : IBUFG
  port map (I => clk80_in, O => clk80_dll_in);

clk80_bufg : BUFG
  port map (I => clk80_dll_out, O => clk80);

clk80_dll : CLKDLL 
  port map (
    CLK0   => clk80_dll_out,
    CLK90  => open,
    CLK180 => open,
    CLK270 => open,
    CLK2X  => open,
    CLKDV  => open,
    LOCKED => clk80_locked_i,
    CLKIN  => clk80_dll_in,
    CLKFB  => clk80,
    RST    => gnd
    );

-- event I/O

event_info_in_block : 
  for i in 0 to 45 generate
    event_info_in_ibufs : IBUF 
      port map (I => event_info_in(i), O => event_info_p(i));
  end generate;

tristate_n_in_block : 
  for i in 0 to 2 generate 
    tristate_n_in_ibufs : IBUF
      port map (I => tristate_n_in(i), O => tristate_n_p(i));
  end generate;

data_valid_in_ibuf : IBUF
port map (I => data_valid_in, O => data_valid_p);

stop_output_obuf : OBUF
port map (I => stop_output_i, O => stop_output_out);

-- SLINK I/O

slink_xoff_n_ibuf : IBUF
port map (I => slink_xoff_n_in, O => slink_xoff_n_i);

slink_down_n_ibuf : IBUF
port map (I => slink_down_n_in, O => slink_down_n_i);

slink_test_n_obuf : OBUF
port map (I => slink_test_n_i, O => slink_test_n_out);

slink_rst_n_obuf : OBUF 
port map (I => slink_rst_n_i, O => slink_rst_n_out);

slink_wen_n_obuf : OBUF
port map (I => slink_wen_n_i, O => slink_wen_n_out);

slink_uctrl_n_obuf : OBUF
port map (I => slink_uctrl_n_i, O => slink_uctrl_n_out);
 
slink_data_out_block : 
  for i in 0 to 31 generate 
    slink_data_obufs : OBUF
      port map (I => slink_data_p(i), O => slink_data_out(i));
  end generate;

-- DSP I/O
dsp0_send_data_n_ibuf : IBUF
port map (I => dsp0_send_data_n_in, O => dsp0_send_data_n_i);

dsp0_sbsram_me_n_ibuf : IBUF
port map (I => dsp0_sbsram_me_n_in, O => dsp0_sbsram_me_n_i);

dsp0_dfr_int_mask_ibuf : IBUF
port map (I => dsp0_dfr_int_mask_in, O => dsp0_df_dfr_int_mask_p);

dsp0_data_frame_ready_obuf : OBUF
port map (I => dsp0_dfrdy_p, O => dsp0_dataframeready_out);

dsp0_trap_enable_ibuf : IBUF
port map (I => dsp0_trap_enable_in, O => dsp0_trap_en_p);

dsp0_event_data_out_block : 
  for i in 0 to 31 generate 
    dsp0_event_data_obufs : OBUFT_F_12
      port map (
        I => dsp0_df_data_out_i(i),
        O => dsp0_event_data_out(i),
        T => dsp0_df_bus_en_i
        );
  end generate;

dsp1_send_data_n_ibuf : IBUF 
port map (I => dsp1_send_data_n_in, O => dsp1_send_data_n_i);

dsp1_sbsram_me_n_ibuf : IBUF 
port map (I => dsp1_sbsram_me_n_in, O => dsp1_sbsram_me_n_i);

dsp1_dfr_int_mask_ibuf : IBUF
port map (I => dsp1_dfr_int_mask_in, O => dsp1_df_dfr_int_mask_p);

dsp1_data_frame_ready_obuf : OBUF 
port map (I => dsp1_dfrdy_p, O => dsp1_dataframeready_out);

dsp1_trap_enable_ibuf : IBUF
port map (I => dsp1_trap_enable_in, O => dsp1_trap_en_p);
	
dsp1_event_data_out_block : 
  for i in 0 to 31 generate 
    dsp1_event_data_obufs : OBUFT_F_12 
      port map (
        I => dsp1_df_data_out_i(i),
        O => dsp1_event_data_out(i),
        T => dsp1_df_bus_en_i
        );
  end generate;

dsp2_send_data_n_ibuf : IBUF 
port map (I => dsp2_send_data_n_in, O => dsp2_send_data_n_i);

dsp2_sbsram_me_n_ibuf : IBUF 
port map (I => dsp2_sbsram_me_n_in, O => dsp2_sbsram_me_n_i);

dsp2_trap_enable_ibuf : IBUF
port map (I => dsp2_trap_enable_in, O => dsp2_trap_en_p);
	
dsp2_dfr_int_mask_ibuf : IBUF 
port map (I => dsp2_dfr_int_mask_in, O => dsp2_df_dfr_int_mask_p);
                          
dsp2_data_frame_ready_obuf : OBUF 
port map (I => dsp2_dfrdy_p, O => dsp2_dataframeready_out);

dsp2_event_data_out_block : 
  for i in 0 to 31 generate 
    dsp2_event_data_obufs : OBUFT_F_12
      port map (
        I => dsp2_df_data_out_i(i),
        O => dsp2_event_data_out(i),
        T => dsp2_df_bus_en_i
        );
  end generate;

dsp3_send_data_n_ibuf : IBUF
port map (I => dsp3_send_data_n_in, O => dsp3_send_data_n_i);

dsp3_sbsram_me_n_ibuf : IBUF 
port map (I => dsp3_sbsram_me_n_in, O => dsp3_sbsram_me_n_i);

dsp3_trap_enable_ibuf : IBUF
port map (I => dsp3_trap_enable_in, O => dsp3_trap_en_p);
	
dsp3_dfr_int_mask_ibuf : IBUF 
port map (I => dsp3_dfr_int_mask_in, O => dsp3_df_dfr_int_mask_p);

dsp3_data_frame_ready_obuf : OBUF 
port map (I => dsp3_dfrdy_p, O => dsp3_dataframeready_out);

dsp3_event_data_out_block : 
  for i in 0 to 31 generate 
    dsp3_event_data_obufs : OBUFT_F_24 
      port map (
        I => dsp3_df_data_out_i(i),
        O => dsp3_event_data_out(i),
        T => dsp3_df_bus_en_i
        );
  end generate;

-- router configuration register bus I/O --28
router_bus_addr_in_block : 
  for i in 0 to 7 generate 
    router_bus_addr_in_ibufs : IBUF 
      port map (
        I => router_bus_addr_in(i),
        O => router_bus_addr_i(i)
        );
  end generate;

router_bus_rnw_ibuf : IBUF 
port map (I => router_bus_rnw_in, O => router_bus_rnw_i);

router_bus_strb_n_ibuf : IBUF 
port map (I => router_bus_strb_n_in, O => router_bus_strb_n_i);

router_bus_ack_obuf : OBUF 
port map (I => router_bus_ack_i, O => router_bus_ack_out);

router_bus_data_inout_block : 
  for i in 0 to 15 generate
    router_bus_data_iobuf : IOBUF
      port map (
        I  => router_bus_data_out_i(i),
        O  => router_bus_data_in_i(i),
        IO => router_bus_data_inout(i),
        T  => router_bus_en_i
        );
  end generate;

-- ROD Type Select
rod_type_ibuf : IBUF
  port map (I => rod_type_in, O => rod_type_i);

-- test output I/O
test_port_out_block : 
  for i in 0 to 7 generate 
    test_port_obufs : OBUF 
      port map (
        I => test_i(i),
        O => test_out(i)
        );
  end generate;

--spare output I/O
spare0 : OBUF
port map (I => event_info_i(43), O => spare_out(0));

spare1 : OBUF
port map (I => event_info_i(44), O => spare_out(1));

spare2 : OBUF
port map (I => event_info_i(45), O => spare_out(2));

spare3 : OBUF
port map (I => clk40, O => spare_out(3));

router_spare_pin :  
  for i in 0 to 12 generate
    router_spare_ibuf : IBUF
      port map (
        I => spare_pin(i),
        O => rtr_spare_in_i(i)
        );
  end generate;

hold_next_event_obuf : OBUF
port map (I => hold_next_event_i, O => hold_next_event_out); -- I => rtr_spare_in_i(13)

--------------------------------------------------------------------------
-- PROCESS DECLARATION
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- Global and Misc. Signals
--------------------------------------------------------------------------
gnd <= '0';
rst <= NOT rst_n;

dll_status_i <= clk40_locked_i & clk80_locked_i;
  
-------------------------------------------------------------------------------
-- S-Link Signals
-------------------------------------------------------------------------------
slink_test_n_i     <= NOT slink_cmd_reg_i(5);
slink_status_reg_i <= stop_output_i & slink_down_i & slink_xoff_i;
slink_wen_n_i      <= NOT slink_wen_i OR slink_cmd_reg_i(3) OR
                      (dump_event_i AND slink_dword_i);

slink_reset : process (
  rst_n,
  clk40
  )
variable wait_count : unsigned(3 downto 0);
begin
  if (rst_n = '0') then
    slink_rst_n_i  <= '1';
    clr_sl_reset_i <= '0';
    hold_reset_i   <= '0';
    wait_count     :=  (others => '0');
    sl_reset_state <= idle;
  elsif (clk40'event and clk40 = '1') then
    case sl_reset_state is
      when idle =>
        slink_rst_n_i  <= '1';
        clr_sl_reset_i <= '0';
        hold_reset_i   <= '0';
        wait_count     :=  (others => '0');
        if (slink_cmd_reg_i(4) = '1') then
          sl_reset_state <= issue_reset;
        end if;

      when issue_reset =>
        slink_rst_n_i  <= '0';
        clr_sl_reset_i <= '0';
        hold_reset_i   <= '1';
        if (wait_count = 8) then
          sl_reset_state <= wait_for_sldown;
        end if;      
        wait_count := wait_count + 1 ;

      when wait_for_sldown =>
        slink_rst_n_i  <= '0';
        clr_sl_reset_i <= '0';
        hold_reset_i   <= '1';
        wait_count     :=  (others => '0');
        if (slink_down_i = '0') then
          sl_reset_state <= done;
        end if;

      when done =>
        slink_rst_n_i  <= '1';
        clr_sl_reset_i <= '1';
        hold_reset_i   <= '1';
        if (wait_count = 4) then
          sl_reset_state <= idle;
        end if;      
        wait_count := wait_count + 1 ;

      when others =>
        slink_rst_n_i  <= '1';
        clr_sl_reset_i <= '0';
        hold_reset_i   <= '0';
        wait_count     :=  (others => '0');
        sl_reset_state <= idle;
    end case;
    
    if (slink_cmd_reg_i(4) = '0') then
      sl_reset_state <= idle;
    end if;
  end if;
end process;

-------------------------------------------------------------------------------
--  SDSP0 DMA Control & Status Signals
-------------------------------------------------------------------------------
dsp0_cen_i        <= NOT dsp0_sbsram_me_n_i;
dsp0_df_rden_i    <= NOT dsp0_send_data_n_i;
dsp0_df_bus_en_i  <=     dsp0_send_data_n_i OR dsp0_cen_n_i;

dsp0_status_reg_i <=   dsp0_trap1_stat_i
                     & dsp0_trap0_stat_i
                     & dsp0_df_full_flag_out_i 
                     & dsp0_df_empty_flag_out_i
                     & dsp0_df_dfr_int_mask_i
                     & dsp0_trap_en_i;
-------------------------------------------------------------------------------
--  SDSP1 DMA Control & Status Signals
-------------------------------------------------------------------------------
dsp1_cen_i        <= NOT dsp1_sbsram_me_n_i;
dsp1_df_rden_i    <= NOT dsp1_send_data_n_i;
dsp1_df_bus_en_i  <=     dsp1_send_data_n_i OR dsp1_cen_n_i;

dsp1_status_reg_i <=   dsp1_trap1_stat_i
                     & dsp1_trap0_stat_i
                     & dsp1_df_full_flag_out_i 
                     & dsp1_df_empty_flag_out_i
                     & dsp1_df_dfr_int_mask_i
                     & dsp1_trap_en_i;
-------------------------------------------------------------------------------
--  SDSP2 DMA Control & Status Signals
-------------------------------------------------------------------------------
dsp2_cen_i        <= NOT dsp2_sbsram_me_n_i;
dsp2_df_rden_i    <= NOT dsp2_send_data_n_i;
dsp2_df_bus_en_i  <=     dsp2_send_data_n_i OR dsp2_cen_n_i;

dsp2_status_reg_i <=   dsp2_trap1_stat_i
                     & dsp2_trap0_stat_i
                     & dsp2_df_full_flag_out_i 
                     & dsp2_df_empty_flag_out_i
                     & dsp2_df_dfr_int_mask_i
                     & dsp2_trap_en_i;
-------------------------------------------------------------------------------
--  SDSP3 DMA Control & Status Signals
-------------------------------------------------------------------------------
dsp3_cen_i        <= NOT dsp3_sbsram_me_n_i;
dsp3_df_rden_i    <= NOT dsp3_send_data_n_i;
dsp3_df_bus_en_i  <=     dsp3_send_data_n_i OR dsp3_cen_n_i;

dsp3_status_reg_i <=   dsp3_trap1_stat_i
                     & dsp3_trap0_stat_i
                     & dsp3_df_full_flag_out_i 
                     & dsp3_df_empty_flag_out_i
                     & dsp3_df_dfr_int_mask_i
                     & dsp3_trap_en_i;


-------------------------------------------------------------------------------
--  Clk80 Signal Register
-------------------------------------------------------------------------------
clk80_signal_register : process (
  rst_n, 
  clk80
  )
begin
  if (rst_n = '0') then
    dsp0_cen_n_i           <= '1';
    dsp1_cen_n_i           <= '1';
    dsp2_cen_n_i           <= '1';
    dsp3_cen_n_i           <= '1';
    dsp0_trap_en_i         <= '0';
    dsp1_trap_en_i         <= '0';
    dsp2_trap_en_i         <= '0';
    dsp3_trap_en_i         <= '0';
    dsp0_df_dfr_int_mask_i <= '0';
    dsp1_df_dfr_int_mask_i <= '0';
    dsp2_df_dfr_int_mask_i <= '0';
    dsp3_df_dfr_int_mask_i <= '0';
  elsif (clk80'event and clk80 = '1') then
    dsp0_cen_n_i           <= dsp0_sbsram_me_n_i;
    dsp1_cen_n_i           <= dsp1_sbsram_me_n_i;
    dsp2_cen_n_i           <= dsp2_sbsram_me_n_i;
    dsp3_cen_n_i           <= dsp3_sbsram_me_n_i;
    dsp0_trap_en_i         <= dsp0_trap_en_p;
    dsp1_trap_en_i         <= dsp1_trap_en_p;
    dsp2_trap_en_i         <= dsp2_trap_en_p;
    dsp3_trap_en_i         <= dsp3_trap_en_p;
    dsp0_df_dfr_int_mask_i <= dsp0_df_dfr_int_mask_p;
    dsp1_df_dfr_int_mask_i <= dsp1_df_dfr_int_mask_p;
    dsp2_df_dfr_int_mask_i <= dsp2_df_dfr_int_mask_p;
    dsp3_df_dfr_int_mask_i <= dsp3_df_dfr_int_mask_p;
  end if;
end process;

-------------------------------------------------------------------------------
--  Clk40 Signal Register
-------------------------------------------------------------------------------
clk40_signal_register : process (
  rst_n, 
  clk40
  )
begin
  if (rst_n = '0') then
    event_info_i <= (others => '0');
    tristate_n_i <= (others => '1');
    data_valid_i <= '0';
    rol_test_en_i <= '0';
  elsif (clk40'event and clk40 = '1') then
    event_info_i <= event_info_p;
    tristate_n_i <= tristate_n_p;
    data_valid_i <= data_valid_p;
	 -- Verify that this is not causing trouble  20200302
    if (event_info_p(31 downto 17) = "101100001111001" AND data_valid_p ='1') then
      rol_test_en_i <= '1';
    elsif (event_info_p(31 downto 17) = "101100001111000" AND data_valid_p ='1') then
      rol_test_en_i <= '0';
    end if;           
  end if;
end process;

-------------------------------------------------------------------------------
--  TestPort Signals
-------------------------------------------------------------------------------
testport_outputs : process (
  test_mux_i,
  rtr_spare_in_i,
  dsp0_dfrdy_i,
  dsp0_df_dfr_int_mask_i,
  dsp0_trap_en_i,
  dsp0_df_bus_en_i,
  dsp0_sbsram_me_n_i,
  dsp0_send_data_n_i,
  dsp1_dfrdy_i,
  dsp1_df_dfr_int_mask_i,
  dsp1_trap_en_i,
  dsp1_df_bus_en_i,
  dsp1_sbsram_me_n_i,
  dsp1_send_data_n_i,
  dsp2_dfrdy_i,
  dsp2_df_dfr_int_mask_i,
  dsp2_trap_en_i,
  dsp2_df_bus_en_i,
  dsp2_sbsram_me_n_i,
  dsp2_send_data_n_i,
  dsp3_dfrdy_i,
  dsp3_df_dfr_int_mask_i,
  dsp3_trap_en_i,
  dsp3_df_bus_en_i,
  dsp3_sbsram_me_n_i,
  dsp3_send_data_n_i 
  )
begin
  case test_mux_i is
    when "0000" => 
      test_i(0) <= rtr_spare_in_i(0);
      test_i(1) <= rtr_spare_in_i(1);
      test_i(2) <= dsp0_dfrdy_i;
      test_i(3) <= dsp0_df_dfr_int_mask_i;
      test_i(4) <= dsp0_trap_en_i;
      test_i(5) <= dsp0_df_bus_en_i;
      test_i(6) <= dsp0_sbsram_me_n_i;
      test_i(7) <= dsp0_send_data_n_i;
    when "0001" => 
      test_i(0) <= rtr_spare_in_i(2);
      test_i(1) <= rtr_spare_in_i(3);
      test_i(2) <= dsp1_dfrdy_i;
      test_i(3) <= dsp1_df_dfr_int_mask_i;
      test_i(4) <= dsp1_trap_en_i;
      test_i(5) <= dsp1_df_bus_en_i;
      test_i(6) <= dsp1_sbsram_me_n_i;
      test_i(7) <= dsp1_send_data_n_i;
    when "0010" => 
      test_i(0) <= rtr_spare_in_i(4);
      test_i(1) <= rtr_spare_in_i(5);
      test_i(2) <= dsp2_dfrdy_i;
      test_i(3) <= dsp2_df_dfr_int_mask_i;
      test_i(4) <= dsp2_trap_en_i;
      test_i(5) <= dsp2_df_bus_en_i;
      test_i(6) <= dsp2_sbsram_me_n_i;
      test_i(7) <= dsp2_send_data_n_i;
    when "0011" => 
      test_i(0) <= rtr_spare_in_i(6);
      test_i(1) <= rtr_spare_in_i(7);
      test_i(2) <= dsp3_dfrdy_i;
      test_i(3) <= dsp3_df_dfr_int_mask_i;
      test_i(4) <= dsp3_trap_en_i;
      test_i(5) <= dsp3_df_bus_en_i;
      test_i(6) <= dsp3_sbsram_me_n_i;
      test_i(7) <= dsp3_send_data_n_i;
    when others => 
      test_i <= rtr_spare_in_i(7 downto 0);
  end case;
end process;


-------------------------------------------------------------------------------
--  Logic to Apply Back Pressure as required
-------------------------------------------------------------------------------
stop_output : process (
  clk40, 
  rst_n
  )
begin
  if(rst_n = '0') then
    dsp_afull_i   <= '0';
    stop_output_i <= '0';
    dsp_afull_i   <= '0';
    slink_xoff_i  <= '0';
    slink_down_i  <= '0';
    hold_next_event_i  <= '0';
    start_hold_delay_i <= '0';
    hold_next_count_i  <= (others => '0') ;
  elsif (clk40'event AND clk40 = '1') then
    dsp_afull_i   <= dsp0_df_afull_flag_out_i OR
                     dsp1_df_afull_flag_out_i OR
                     dsp2_df_afull_flag_out_i OR
                     dsp3_df_afull_flag_out_i;
    slink_down_i  <= NOT (slink_down_n_i     OR 
                          slink_cmd_reg_i(3) OR   --  WR Inhibit
                          slink_cmd_reg_i(7) );   --  Mask LDOWN

-- If WR Inhibit = '1', then the S-Link Xoff signal is masked
-- slink_cmd_reg(6) must be set to 1 to enable the DSP backpressure
-- during physics data-taking, slink_cmd_reg(6) must be set to 0 
    slink_xoff_i  <= NOT (slink_xoff_n_i OR slink_cmd_reg_i(3)) OR -- Slink Xoff
                         (dsp_afull_i AND slink_cmd_reg_i(6))   OR -- DSP Xoff
                          slink_down_i OR force_xoff_i;

    stop_output_i <= hold_efb_i OR hold_reset_i;  -- OR force_xoff_i;
    
    start_hold_delay_i <= sl_fifo_stat_i(0);
    case hold_next_event is
      when idle =>
        hold_next_event_i  <= slink_cmd_reg_i(6) AND NOT sl_fifo_stat_i(0);
        hold_next_count_i  <= (others => '0') ;
        if (slink_cmd_reg_i(6) = '1' AND sl_fifo_stat_i(0) = '1' AND start_hold_delay_i = '0') then
          hold_next_event <= hold_off_event;
          hold_next_event_i <= '1';
        end if;
      when hold_off_event =>
        hold_next_count_i <= hold_next_count_i + 1;
        hold_next_event_i <= '1';
        if (std_logic_vector(hold_next_count_i) = hold_next_delay_i) then
          hold_next_event <= done;
        end if;
      when done =>
        hold_next_event_i  <= slink_cmd_reg_i(6) AND NOT sl_fifo_stat_i(0);
        hold_next_count_i  <= (others => '0') ;
        hold_next_event <= idle;
      when others =>
        hold_next_event_i  <= slink_cmd_reg_i(6) AND NOT sl_fifo_stat_i(0);
        hold_next_event <= idle;
    end case;
  end if;
end process stop_output;

-------------------------------------------------------------------------------
--  Event Data Output Endianess Select
-------------------------------------------------------------------------------
endianess_select : process (
  endian_select_i,
  slink_data_i,
  slink_dword_i
  )
begin
  if (endian_select_i = '0' AND slink_dword_i = '1') then
    -- use BE format for data half words
    slink_data_p(31 downto 16) <= slink_data_i(15 downto  0);
    slink_data_p(15 downto  0) <= slink_data_i(31 downto 16);
  else
    slink_data_p <= slink_data_i;
  end if;
end process endianess_select; 

-------------------------------------------------------------------------------
--  Event Data Signal Mux
-------------------------------------------------------------------------------
post_pipeline_data_mux : process (
  rst_n, 
  clk40
  )
begin
  if (rst_n = '0') then
    dsp0_data_in_i <= (others => '0');
    dsp0_df_wren_i <= '0';
    dsp1_data_in_i <= (others => '0');
    dsp1_df_wren_i <= '0';
    dsp2_data_in_i <= (others => '0');
    dsp2_df_wren_i <= '0';
    dsp3_data_in_i <= (others => '0');
    dsp3_df_wren_i <= '0';
  elsif (clk40'event and clk40 = '1') then
    case dsp0_cmd_reg0_i(4 downto 3) is
      when "01"   => dsp0_data_in_i <= dspe_data_i;
                     dsp0_df_wren_i <= dspe_wen_i;
      when "10"   => dsp0_data_in_i <= dsps_data_i;
                     dsp0_df_wren_i <= dsps_wen_i;
      when others => dsp0_data_in_i <= slink_data_i;
                     dsp0_df_wren_i <= slink_wen_i;
    end case;
  
    case dsp1_cmd_reg0_i(4 downto 3) is
      when "01"   => dsp1_data_in_i <= dspe_data_i;
                     dsp1_df_wren_i <= dspe_wen_i;
      when "10"   => dsp1_data_in_i <= dsps_data_i;
                     dsp1_df_wren_i <= dsps_wen_i;
      when others => dsp1_data_in_i <= slink_data_i;
                     dsp1_df_wren_i <= slink_wen_i;
    end case;

    case dsp2_cmd_reg0_i(4 downto 3) is
      when "01"   => dsp2_data_in_i <= dspe_data_i;
                     dsp2_df_wren_i <= dspe_wen_i;
      when "10"   => dsp2_data_in_i <= dsps_data_i;
                     dsp2_df_wren_i <= dsps_wen_i;
      when others => dsp2_data_in_i <= slink_data_i;
                     dsp2_df_wren_i <= slink_wen_i;
    end case;
  
    case dsp3_cmd_reg0_i(4 downto 3) is
      when "01"   => dsp3_data_in_i <= dspe_data_i;
                     dsp3_df_wren_i <= dspe_wen_i;
      when "10"   => dsp3_data_in_i <= dsps_data_i;
                     dsp3_df_wren_i <= dsps_wen_i;
      when others => dsp3_data_in_i <= slink_data_i;
                     dsp3_df_wren_i <= slink_wen_i;
    end case;
  end if;
end process post_pipeline_data_mux; 

-------------------------------------------------------------------------------
--  Router Bus Tristate Logic
-------------------------------------------------------------------------------
router_bus_tristate : process (
  rst_n, 
  clk40
  )
begin
  if (rst_n = '0') then
    router_bus_en_i      <= '1';
    router_bus_strb_n_ii <= '1';
  elsif (clk40'event and clk40 = '1') then
    router_bus_strb_n_ii <= router_bus_strb_n_i;
    if (router_bus_strb_n_ii = '0' AND router_bus_rnw_i = '1') then 
      router_bus_en_i <= '0';
    else
      router_bus_en_i <= '1';
    end if;
  end if;
end process router_bus_tristate;

end rtl; -- code for top
