-------------------------------------------------------------------------------
--  Lawrence Berkeley National Laboratory
-------------------------------------------------------------------------------
--  Filename: top.vhd
--  Title: FPGA Program Reset Manager
--  Description:    This FPGA controls configuration of all FPGA parts, and
--                  controls the reset signals to all FPGAs and DSPs on the 
--                  Read Out Driver Board used in Atlas. 
--		
--  Constraints required to that define the location of the FPGA PRM global
--  clocks and DLLs, and are used to properly synthesize the proto-type ver:
--
--  INST main_clk_dll LOC = DLL0;     -> Location of the main clk dll
--  INST clk40_bufg LOC = GCLKBUF0;   -> Location of the main clk global bus
--  INST cclk_bufg LOC = GCLKBUF1;    -> Location of the cclk global bus
--  INST reset_bufg LOC = GCLKBUF2;   -> Location of the reset global bus
--  INST vme_clk_bufg LOC = GCLKBUF3; -> Location of the vme clk global bus
--  
--  NET clk40_in PERIOD = 25ns;       -> Define period of system clock input
--  NET clk40_i PERIOD = 25ns;        -> Define period of internal system clock
--  NET vme_clk_in PERIOD = 16ns;     -> Define period of vme clock input
--  NET vme_clk_in PERIOD = 16ns;     -> Define period of internal vme clock
--
-------------------------------------------------------------------------------
-- Author:   John Joseph
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-- Revision History
-------------------------------------------------------------------------------
--
--  June 8, 2000  JMJ defined top ports to match schematic requirements
--  Aug 21, 2000  JMJ began to revise code for first prototype
--
-------------------------------------------------------------------------------
--  Library Declaration
-------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;          --needed for logical operations
use IEEE.std_logic_arith.all;         --needed for +/- operations

--  comment out the 2 UNISIM lines for synthesis using Foundation 3.1i
--library UNISIM;
--use unisim.all;

-------------------------------------------------------------------------------
--  PORT DECLARATION
-------------------------------------------------------------------------------

entity fpga_prm_top is 
port (
-- clock and reset ports
  clk40_in    : in  std_logic;		
  vme_clk_in  : in  std_logic;    
  rst_n_in    : in	std_logic;

-- FLASH memory I/O and control ports
  memory_address_out : out   std_logic_vector(20 downto 0);
  memory_data        : inout std_logic_vector( 7 downto 0);
  memory_busy_n_in   : in    std_logic;
  memory_rst_n_out   : out   std_logic;
  CE0_n_out          : out   std_logic;
  OE0_n_out          : out   std_logic;
  WE0_n_out          : out   std_logic;
         
-- dsp I/O and Reset command ports
  mdsp_hd            : inout std_logic_vector(15 downto 0);
  mdsp_hrdy_n_in     : in    std_logic;  -- HRDY from the Master DSP
  mdsp_hhwil_out     : out   std_logic;  -- HHWIL to the Master DSP
  mdsp_hcntl0_out    : out   std_logic;  -- HCNTL0 to the Master DSP
  mdsp_hcntl1_out    : out   std_logic;  -- HCNTL1 to the Master DSP
  mdsp_hds1_n_out    : out   std_logic;  -- HDS1_N to the Master DSP
  mdsp_hcs_n_out     : out   std_logic;  -- HCS_N to the Master DSP
  mdsp_hrnw_out      : out   std_logic;  -- HRNW to the Master DSP
  rcf_hrdy_n_in      : in    std_logic;  -- HRDY from the RCF
  rcf_hcs_n_out      : out   std_logic;  -- HCS_N to the RCF
  
  mdsp_boot0_out     : out   std_logic;
  mdsp_boot1_out     : out   std_logic;
  mdsp_boot2_out     : out   std_logic;
  mdsp_boot3_out     : out   std_logic;
  mdsp_boot4_out     : out   std_logic;

-- RCF Reset command ports
  reset_cmds_in      : in    std_logic_vector(1 downto 0);  -- Reset Cmd from the RRIF
  reset_cmds_out     : out   std_logic_vector(1 downto 0);  -- Ack and clear of Reset Cmd to RRIF

-- vme interface ports
  vme_data_inout     : inout std_logic_vector(31 downto 0);                
  vme_de_n_out       : out   std_logic;
  vme_ddir_out       : out   std_logic;
  vme_addr_in        : in    std_logic_vector(31 downto 0);
  vme_am_in          : in    std_logic_vector( 5 downto 0);
  vme_as_in          : in    std_logic;
  vme_ds0_in         : in    std_logic;
  vme_ds1_in         : in    std_logic;
  vme_wr_in          : in    std_logic;
  vme_iack_in        : in    std_logic;
  vme_iackin_n_in    : in    std_logic;
  vme_iackout_n_out  : out   std_logic;
  vme_sysfail_n_in   : in    std_logic;
  vme_dtack_n_out    : out   std_logic;
  dtack_en_n_out     : out   std_logic;
  vme_berr_n_out     : out   std_logic;
  berr_en_n_out      : out   std_logic;
  lirq_n_in          : in    std_logic;
  lirq_in            : in    std_logic_vector(2 downto 0);
  board_addr_n_in    : in    std_logic_vector(4 downto 0);
  vme_gap_n_in       : in    std_logic;  -- parity bit for Geographical Addr

-- ROD Status
  rod_busy_in        : in  std_logic;
  rod_busy_n_out     : out std_logic;
  rod_type_in        : in  std_logic;
  fmt_type_in        : in  std_logic_vector(1 downto 0);
  slink_lff_n_in     : in  std_logic;
  slink_ldown_n_in   : in  std_logic;
  dsp0_detect_n_in   : in  std_logic;
  dsp1_detect_n_in   : in  std_logic;
  dsp2_detect_n_in   : in  std_logic;
  dsp3_detect_n_in   : in  std_logic;
                    -- fpga programming and reset control/handshaking ports
  formA_prgbits_out    : out   std_logic;
  formA_cclk_out       : out   std_logic;
  formA_prgm_n_out     : out   std_logic;
  formA_done_in        : in    std_logic;
  formA_init_n_in      : in    std_logic;
  formA_rst_n_out      : out   std_logic;

  formB_prgbits_out    : out   std_logic;
  formB_cclk_out       : out   std_logic;
  formB_prgm_n_out     : out   std_logic;
  formB_done_in        : in    std_logic;
  formB_init_n_in      : in    std_logic;
  formB_rst_n_out      : out   std_logic;

  efb_prgbits_out      : out   std_logic;
  efb_cclk_out         : out   std_logic;
  efb_prgm_n_out       : out   std_logic;
  efb_done_in          : in    std_logic;
  efb_init_n_in        : in    std_logic;
  efb_rst_n_out        : out   std_logic;
                        
  router_prgbits_out   : out   std_logic;
  router_cclk_out      : out   std_logic;
  router_prgm_n_out    : out   std_logic;
  router_done_in       : in    std_logic;
  router_init_n_in     : in    std_logic;
  router_rst_n_out     : out   std_logic;

  rodresio_prgbits_out : out   std_logic;
  rodresio_cclk_out    : out   std_logic;
  rodresio_prgm_n_out  : out   std_logic;
  rodresio_done_in     : in    std_logic;
  rodresio_init_n_in   : in    std_logic;
  rodresio_rst_n_out   : out   std_logic;

  spare_prgbits_out    : out   std_logic;
  spare_cclk_out       : out   std_logic;
  spare_prgm_n_out     : out   std_logic;
  spare_done_in        : in    std_logic_vector(3 downto 0);
  spare_init_n_in      : in    std_logic;

  fpga_pgrm_done_out   : out   std_logic;               
    
-- dsp reset ports
  dsp_rst_n_out        : out   std_logic_vector(4 downto 0);

-- ROD serial number inputs
  serial_number_n_in   : in    std_logic_vector(9 downto 0);
  rod_clock_select_in  : in    std_logic;
                        
-- test signal & spare output ports
  test_in    : in    std_logic_vector(7 downto 0);
--test_out   : out   std_logic_vector(7 downto 0);
  spare_pin  : in    std_logic_vector(7 downto 0);
  spare_pout : out   std_logic_vector(1 downto 0)
  );
end fpga_prm_top;

architecture rtl of fpga_prm_top is

-------------------------------------------------------------------------------
--SIGNAL DECLARATION
-------------------------------------------------------------------------------
-- internal version of power and ground signals
signal gnd : std_logic;

-- internal clock signals
signal clk40_g   : std_logic;
signal clk05_g   : std_logic;
signal vme_clk_g : std_logic;
signal cclk_g    : std_logic;

--  ROD clock select Status
signal rod_clk_sel_p : std_logic;

--  clock signals attached to the main_clk_dll module
signal dll_clk40_in     : std_logic;
signal dll_clk40_out    : std_logic;
signal dll_clk05_out    : std_logic;
signal clk40_locked_i   : std_logic;
signal vme_clk_ibufg_o  : std_logic;
signal dll_vme_clk_out  : std_logic;
signal dll_cclk_out     : std_logic;
signal vme_clk_locked_i : std_logic;

attribute CLKDV_DIVIDE : string;

--  divide attribute for main_clk_dll / divide by 8 (5.0 MHz | 200ns)
attribute CLKDV_DIVIDE of main_clk_dll : label is "8";

--  divide attribute for vme_clk_dll / divide by 16 (5 MHz | 200ns)
--attribute CLKDV_DIVIDE of vme_clk_dll : label is "16";
--  divide attribute for vme_clk_dll / divide by  8 (10.0 MHz | 100ns)
--attribute CLKDV_DIVIDE of vme_clk_dll : label is "8";
--  divide attribute for vme_clk_dll / divide by  4 (20.0 MHz | 50ns)
  attribute CLKDV_DIVIDE of vme_clk_dll : label is "4";

--  internal async reset signal
signal rst_n_i      : std_logic;
signal rst_i        : std_logic;

--  internal flash memory signals
signal flash_mem_addr_out_i : std_logic_vector(20 downto 0);
signal flash_mem_data_out_i : std_logic_vector( 7 downto 0);
signal flash_mem_data_in_i  : std_logic_vector( 7 downto 0);
signal memory_data_dir_en_i : std_logic;
signal memory_busy_n_p      : std_logic;
signal memory_rst_n_p       : std_logic;
signal WE_n_out_i           : std_logic;
signal OE_n_out_i           : std_logic;
signal CE0_n_out_i          : std_logic;

signal flash_addr_i       : std_logic_vector(23 downto 0);
signal flash_data_out_i   : std_logic_vector( 7 downto 0);
signal flash_data_in_i    : std_logic_vector( 7 downto 0);
signal flash_cmd_reg_i    : std_logic_vector( 2 downto 0);
signal flash_status_reg_i : std_logic_vector( 3 downto 0);

--  internal reset cmd interface signals
signal mdsp_hpi_data_in_i  : std_logic_vector(15 downto 0);
signal mdsp_hpi_data_out_i : std_logic_vector(15 downto 0);
signal mdsp_hpi_data_dir_i : std_logic;
signal mdsp_cen_n_i        : std_logic;
signal mdsp_cen_i          : std_logic;
signal mdsp_rnw_i          : std_logic;
signal mdsp_hrdy_n_i       : std_logic;
signal mdsp_hhwil_i        : std_logic;
signal mdsp_hcntl0_i       : std_logic;
signal mdsp_hcntl1_i       : std_logic;
signal mdsp_hds1_n_i       : std_logic;
signal mdsp_boot0_p        : std_logic;
signal mdsp_boot1_p        : std_logic;
signal mdsp_boot2_p        : std_logic;
signal mdsp_boot3_p        : std_logic;
signal mdsp_boot4_p        : std_logic;
signal rcf_hrdy_n_i        : std_logic;
signal rcf_cen_i           : std_logic;
signal rcf_cen_n_i         : std_logic;
    
signal reset_cmds_in_i     : std_logic_vector(1 downto 0);
signal reset_cmds_out_i    : std_logic_vector(1 downto 0);

signal data_strb_i         : std_logic;
   
--  internal vmebus signals
signal vme_data_out_p   : std_logic_vector(31 downto 0);
signal vme_data_in_p    : std_logic_vector(31 downto 0);
signal vme_prm_out_i    : std_logic_vector(31 downto 0);
signal vme_hpi_do_i     : std_logic_vector(31 downto 0);
signal vme_addr_i       : std_logic_vector(31 downto 0);
signal vme_data_dir_i   : std_logic;
signal board_addr_n_i   : std_logic_vector( 4 downto 0);
signal ga_n_i           : std_logic_vector( 4 downto 0);
signal vme_timeout_i    : std_logic_vector(31 downto 0);
signal vme_rnw_i        : std_logic;
signal vme_am_code_i    : std_logic_vector( 5 downto 0);
signal vme_iack_i       : std_logic;
signal lvme_iack_i      : std_logic;

signal vme_de_n_p       : std_logic;
signal vme_ddir_p       : std_logic;
signal vme_am_p         : std_logic_vector( 5 downto 0);
signal vme_as_p         : std_logic;
signal vme_as_i         : std_logic;
signal vme_ds_i         : std_logic;
signal vme_ds0_p        : std_logic;
signal vme_ds1_p        : std_logic;
signal vme_wr_p         : std_logic;
signal vme_iack_p       : std_logic;
signal vme_iackin_n_p   : std_logic;
signal vme_iackout_n_p  : std_logic;
signal vme_sysfail_n_p  : std_logic;
signal vme_sysfail_i    : std_logic;
signal vme_dtack_n_p    : std_logic;
signal dtack_en_n_p     : std_logic;
signal dtack_en_n_d1    : std_logic;
signal vme_berr_n_p     : std_logic;
signal berr_en_n_p      : std_logic;
signal berr_en_n_d1     : std_logic;
signal lirq_n_p         : std_logic;
signal lirq_i           : std_logic;
signal lirq_val_p       : std_logic_vector( 2 downto 0);
signal vme_gap_n_p      : std_logic;  -- parity bit for Geographical Addr
signal vme_gap_n_i      : std_logic;  -- parity bit for Geographical Addr
                       
--  program reset manager bus signals
signal prm_cen_i  : std_logic;
                         
--  internal fpga configuration signals
signal efb_prgbits_out_i    : std_logic;
signal efb_cclk_out_i       : std_logic;
signal efb_prgm_n_out_i     : std_logic;
signal efb_done_in_i        : std_logic;
signal efb_init_n_in_i      : std_logic;
signal efb_rst_n_out_i      : std_logic;
signal efb_rst_n_out_p      : std_logic;
signal efb_halt_i           : std_logic;
                        
signal formA_prgbits_out_i  : std_logic;
signal formA_cclk_out_i     : std_logic;
signal formA_prgm_n_out_i   : std_logic;
signal formA_done_in_i      : std_logic;
signal formA_init_n_in_i    : std_logic;
signal formA_rst_n_out_i    : std_logic;
signal formA_rst_n_out_p    : std_logic;
signal formA_halt_i         : std_logic;

signal formB_prgbits_out_i  : std_logic;
signal formB_cclk_out_i     : std_logic;
signal formB_prgm_n_out_i   : std_logic;
signal formB_done_in_i      : std_logic;
signal formB_init_n_in_i    : std_logic;
signal formB_rst_n_out_i    : std_logic;
signal formB_rst_n_out_p    : std_logic;
signal formB_halt_i         : std_logic;

signal router_prgbits_out_i : std_logic;
signal router_cclk_out_i    : std_logic;
signal router_prgm_n_out_i  : std_logic;
signal router_done_in_i     : std_logic;
signal router_init_n_in_i   : std_logic;
signal router_rst_n_out_i   : std_logic;
signal router_rst_n_out_p   : std_logic;
signal router_halt_i        : std_logic;

signal rodresio_prgbits_out_i : std_logic;
signal rodresio_cclk_out_i    : std_logic;
signal rodresio_prgm_n_out_i  : std_logic;
signal rodresio_done_in_i     : std_logic;
signal rodresio_init_n_in_i   : std_logic;
signal rodresio_rst_n_out_i   : std_logic;
signal rodresio_rst_n_out_p   : std_logic;
signal rodresio_halt_i        : std_logic;

signal spare_prgbits_out_i : std_logic;
signal spare_cclk_out_i    : std_logic;
signal spare_prgm_n_out_i  : std_logic;
signal spare_done_in_i     : std_logic_vector(3 downto 0);
signal spare_init_n_in_i   : std_logic;

--  internal dsp
signal dsp_rst_n_out_i : std_logic_vector(4 downto 0);

--  internal configuration and status signals
signal fpga_pgrm_done_out_i  : std_logic; --  indicates that config is complete
signal fpga_sel_i            : std_logic_vector(2 downto 0); 
signal fpga_rst_status_reg_i : std_logic_vector(4 downto 0);
signal dsp_rst_status_reg_i  : std_logic_vector(4 downto 0);

--  internal shift register control, status, and data bits
signal shift_reg0_load_i   : std_logic;
signal shift_reg1_load_i   : std_logic;
signal sr_cnfg_data_i      : std_logic;
signal sr_cnfg_data0_i     : std_logic;
signal sr_cnfg_data1_i     : std_logic;
signal shift_reg_en_i      : std_logic_vector(1 downto 0);
signal shift_reg_en_d1     : std_logic_vector(1 downto 0);
signal shift_reg_load_en_i : std_logic_vector(1 downto 0);

--  internal control signals (the names should define)
signal cnfg_enable_i     : std_logic;  -- flag to indicate that prm is in cnfg mode
signal cnfg_addr_i       : std_logic_vector(23 downto 0); -- cnfg data addr location
signal cnfg_data_i       : std_logic_vector( 7 downto 0); -- internal cnfg data path
signal clr_cnfg_reg_i    : std_logic;
signal fpga_cnfg_reg_i   : std_logic_vector( 6 downto 0);
signal fpga_rst_reg_i    : std_logic_vector( 6 downto 0);
signal dsp_rst_reg_i     : std_logic_vector( 6 downto 0);
signal reset_dsp_i       : std_logic_vector( 5 downto 0);
signal dsp_reset_cmd_i   : std_logic_vector( 3 downto 0);
signal clr_rst_reg_i     : std_logic;

signal rod_busy_n_p : std_logic;
signal rod_busy_i   : std_logic;
signal rod_busy_n_i : std_logic;

signal fmt_type_p      : std_logic_vector(1 downto 0);
signal rod_type_p      : std_logic;
signal slink_lff_n_p   : std_logic;
signal slink_lff_i     : std_logic;
signal slink_ldown_n_p : std_logic;
signal slink_ldown_i   : std_logic;
signal dsp0_detect_n_p : std_logic;
signal dsp1_detect_n_p : std_logic;
signal dsp2_detect_n_p : std_logic;
signal dsp3_detect_n_p : std_logic;
signal dsp_detect_i    : std_logic_vector(3 downto 0);

signal serial_number_n_i : std_logic_vector(9 downto 0);
signal serial_number_i   : std_logic_vector(9 downto 0);

signal spare_pin_i : std_logic_vector(7 downto 0);

signal ill_match_i   : std_logic;  -- illegal decode region
signal hpi_dtack_n_i : std_logic;
signal hpi_berr_i    : std_logic;
signal prb_bus_ack_i : std_logic;
signal test_i        : std_logic_vector(7 downto 0);

signal trigger_i     : std_logic;

-------------------------------------------------------------------------------
--COMPONENT DECLARATION
-------------------------------------------------------------------------------

--  The shift register converts the flash data from an 8-bit word to a serial
--  stream with the LSB out first.  Two shift registers are used in the design
--  to ensure a continuous stream of configuration bits to the FPGAs.
component shift_register
  port (
    clk_in      : in  std_logic;  -- cclk input
    rst_n_in    : in  std_logic;  -- reset input
    en_load_in  : in  std_logic;  -- enable load
    en_shift_in : in  std_logic;  -- enable shift
    data_in     : in  std_logic_vector(7 downto 0); -- 8-bit word in
    data_out    : out std_logic   -- single bit out
    );
end component;

--  The register block preforms all register function required by the FPGA PRM
component progman_register_block
  port (
    clk_in                     : in  std_logic; -- clk input
    cclk_in                    : in  std_logic; -- cclk input
    rst_n_in                   : in  std_logic; -- async global reset
    prb_bus_strb_in            : in  std_logic;
    prb_bus_ce_in              : in  std_logic;
    prb_bus_rnw_in             : in  std_logic;
    prb_bus_addr_in            : in  std_logic_vector(31 downto 0);
    prb_bus_data_in            : in  std_logic_vector(31 downto 0);
    prb_bus_data_out           : out std_logic_vector(31 downto 0);
    prb_bus_ack_out            : out std_logic;
    prb_rrif_done_in           : in  std_logic;
    prb_rrif_init_n_in         : in  std_logic;
    prb_rrif_halt_in           : in  std_logic;
    prb_formA_done_in          : in  std_logic;
    prb_formA_init_n_in        : in  std_logic;
    prb_formA_halt_in          : in  std_logic;
    prb_formB_done_in          : in  std_logic;
    prb_formB_init_n_in        : in  std_logic;
    prb_formB_halt_in          : in  std_logic;
    prb_efb_done_in            : in  std_logic;
    prb_efb_init_n_in          : in  std_logic;
    prb_efb_halt_in            : in  std_logic;
    prb_router_done_in         : in  std_logic;
    prb_router_init_n_in       : in  std_logic;
    prb_router_halt_in         : in  std_logic;
    prb_spare_done_in          : in  std_logic_vector( 3 downto 0);
    prb_spare_init_n_in        : in  std_logic;
    prb_fpga_rst_status_reg_in : in  std_logic_vector( 4 downto 0);
    prb_dsp_rst_status_reg_in  : in  std_logic_vector( 4 downto 0);
    fpga_cnfg_reg_out          : out std_logic_vector( 6 downto 0);
    fpga_rst_reg_out           : out std_logic_vector( 6 downto 0);
    dsp_rst_reg_out            : out std_logic_vector( 6 downto 0);
    flash_cmd_reg_out          : out std_logic_vector( 2 downto 0);
    flash_status_reg_in        : in  std_logic_vector( 3 downto 0);
    flash_addr_reg_out         : out std_logic_vector(23 downto 0);
    flash_data_reg_in          : in  std_logic_vector( 7 downto 0);
    flash_data_reg_out         : out std_logic_vector( 7 downto 0);
    clr_cnfg_reg_in            : in  std_logic;
    clr_rst_reg_in             : in  std_logic;
    prb_cnfg_en_in             : in  std_logic;
    serial_number_in           : in  std_logic_vector( 9 downto 0);
    rod_busy_in                : in  std_logic;
    rod_clk_status_in          : in  std_logic;
    vme_clk_status_in          : in  std_logic;
    rod_clk_select_in          : in  std_logic;
    rod_type_in                : in  std_logic;
    fmt_type_in                : in  std_logic_vector(1 downto 0);
    slink_lff_in               : in  std_logic;
    slink_ldown_in             : in  std_logic;
    dsp_detect_in              : in  std_logic_vector( 3 downto 0);
    prm_spares_in              : in  std_logic_vector( 7 downto 0);
    mdsp_hrdy_n_in             : in  std_logic;
    cnfg_done_in               : in  std_logic;         
    vme_timeout_out            : out std_logic_vector(31 downto 0);
    vme_am_code_in             : in  std_logic_vector( 5 downto 0);  -- vme AM code
    ga_n_in                    : in  std_logic_vector( 4 downto 0);
    gap_n_in                   : in  std_logic;
    trigger_in                 : in  std_logic
    );
end component;

--  The flash decoder block interfaces the FPGA PRM with the Flash Memory ICs
component flash_decoder_block
  port (
    clk_in            : in  std_logic;
    cclk_in           : in  std_logic;
    rst_n_in          : in  std_logic;
    fdb_addr_in       : in  std_logic_vector(23 downto 0);
    fdb_data_in       : in  std_logic_vector( 7 downto 0);
    fdb_data_out      : out std_logic_vector( 7 downto 0);
    fdb_cr_in         : in  std_logic_vector( 2 downto 0);
    fdb_sr_out        : out std_logic_vector( 2 downto 0);
    cnfg_enable_in    : in  std_logic;
    cnfg_addr_in      : in  std_logic_vector(23 downto 0);
    cnfg_data_out     : out std_logic_vector( 7 downto 0);
    memory_addr_out   : out std_logic_vector(20 downto 0);
    memory_data_in    : in  std_logic_vector( 7 downto 0);
    memory_data_out   : out std_logic_vector( 7 downto 0);
    write_en_n_out    : out std_logic;
    output_en_n_out   : out std_logic;
    CE0_n_out         : out std_logic;
    shift_reg_load_in : in  std_logic_vector( 1 downto 0)
    );
end component;

component fpga_cnfg_reset_controller_block
  port (
    clk_in                     : in  std_logic;
    cclk_in                    : in  std_logic;
    rst_n_in                   : in  std_logic;
    cb_formA_done_in           : in  std_logic;
    cb_formA_init_n_in         : in  std_logic;
    cb_formA_halt_out          : out std_logic;
    cb_formB_done_in           : in  std_logic;
    cb_formB_init_n_in         : in  std_logic;
    cb_formB_halt_out          : out std_logic;
    cb_efb_done_in             : in  std_logic;
    cb_efb_init_n_in           : in  std_logic;
    cb_efb_halt_out            : out std_logic;
    cb_router_done_in          : in  std_logic;
    cb_router_init_n_in        : in  std_logic;
    cb_router_halt_out         : out std_logic;
    cb_rrif_done_in            : in  std_logic;
    cb_rrif_init_n_in          : in  std_logic;
    cb_rrif_halt_out           : out std_logic;
    cb_fpga_cnfg_reg_in        : in  std_logic_vector(5 downto 0);
    cb_clr_cnfg_reg_out        : out std_logic;
    cb_fpga_rst_reg_in         : in  std_logic_vector(5 downto 0);
    cb_dsp_rst_reg_in          : in  std_logic_vector(5 downto 0);
    cb_clr_rst_reg_out         : out std_logic;
    cb_fpga_rst_status_reg_out : out std_logic_vector(4 downto 0);
    cb_dsp_rst_status_reg_out  : out std_logic_vector(4 downto 0);
    cnfg_enable_out            : out std_logic;
    cnfg_done_out              : out std_logic;
    cnfg_addr_out              : out std_logic_vector(23 downto 0);
    cb_rrif_prgm_n_out         : out std_logic;
    cb_formA_prgm_n_out        : out std_logic;
    cb_formB_prgm_n_out        : out std_logic;
    cb_efb_prgm_n_out          : out std_logic;
    cb_router_prgm_n_out       : out std_logic;
    cb_spare_prgm_n_out        : out std_logic;
    cb_rrif_rst_n_out          : out std_logic;
    cb_formA_rst_n_out         : out std_logic;
    cb_formB_rst_n_out         : out std_logic;
    cb_efb_rst_n_out           : out std_logic;
    cb_router_rst_n_out        : out std_logic;
    fpga_sel_out               : out std_logic_vector(2 downto 0);
    shift_reg_en_out           : out std_logic_vector(1 downto 0);
    shift_reg_load_out         : out std_logic_vector(1 downto 0);
    cb_mdsp_rst_n_out          : out std_logic;
    cb_sdsp0_rst_n_out         : out std_logic;
    cb_sdsp1_rst_n_out         : out std_logic;
    cb_sdsp2_rst_n_out         : out std_logic;
    cb_sdsp3_rst_n_out         : out std_logic;
    cnfg_override_in           : in  std_logic
    );
end component;

component hpi_data_controller
  port (
    clk_in            : in  std_logic; -- vme_clk66 input
    rst_n_in          : in  std_logic; -- asynchronous global reset 
    mdsp_cen_in       : in  std_logic; -- MDSP CE_N signal in
    mdsp_ds_in        : in  std_logic; -- MDSP Data Strobe (VME BE)
    mdsp_hrdy_n_in    : in  std_logic; -- Master DSP HRDY_N signal
    mdsp_hpi_data_in  : in  std_logic_vector(15 downto 0); --
    mdsp_hpi_data_out : out std_logic_vector(15 downto 0); --
    mdsp_cen_n_out    : out std_logic; -- MDSP CE_N signal out
    mdsp_rnw_out      : out std_logic; -- MDSP RNW signal out
    mdsp_hhwil_out    : out std_logic; -- MDSP HHWIL signal
    mdsp_hds1_n_out   : out std_logic; -- MDSP HHDS1 signal
    vme_bus_rnw_in    : in  std_logic; -- VMEBus read/not_write signal
    mdsp_vme_data_in  : in  std_logic_vector(31 downto 0); --
    mdsp_vme_data_out : out std_logic_vector(31 downto 0); --
    hpi_dtack_n_out   : out std_logic; -- VMEBus LACK_N 
    hpi_berr_out      : out std_logic; -- VMEBus BERR
    rcf_cen_in        : in  std_logic;
    rcf_hrdy_n_in     : in  std_logic;
    rcf_cen_n_out     : out std_logic;
    vme_timeout_in    : in  std_logic_vector(31 downto 0)
    );
end component; 

component vme_address_decoder
  port (
    clk_in          : in  std_logic;  -- VMEBus clk (60MHz)
    rst_n_in        : in  std_logic;  -- asynchronous reset (active low)
    rod_addr_n_in   : in  std_logic_vector( 4 downto 0); -- GA bits must be inv
    gaddr_par_n_in  : in  std_logic;  --vme_gap_n_p
    vme_addr_in     : in  std_logic_vector(31 downto 0); -- VMEBus Addr (31:20)
    vme_am_in       : in  std_logic_vector( 5 downto 0); -- LVME Byte Enables
    vme_as_in       : in  std_logic;
    vme_ds0_in      : in  std_logic;
    vme_ds1_in      : in  std_logic;
    prm_cen_out     : out std_logic;  -- PRM Register Block CS
    mdsp_cen_out    : out std_logic;  -- MDSP HPI State Machine Enable
    rcf_cen_out     : out std_logic;  -- RCF HPI State Machine Enable
    ill_match_out   : out std_logic;  -- Address Match is true, but the decode region is false
    data_strb_out   : out std_logic;   -- VME DS signal
    vme_am_out      : out std_logic_vector( 5 downto 0)
    );
end component;

component reset_command_controller
  port (
    clk40_in           : in  std_logic;
    clk05_in           : in  std_logic;
    rst_n_in           : in  std_logic;
    clr_reset_cmd_in   : in  std_logic;
    rod_busy_in        : in  std_logic;
    reset_commands_in  : in  std_logic_vector(1 downto 0);
    reset_commands_out : out std_logic_vector(1 downto 0);
    reset_sdsp_out     : out std_logic_vector(3 downto 0);
    trigger_out        : out std_logic
    );
end component;

-------------------------------------------------------------------------------
-- PAD DECLARATIONS
-------------------------------------------------------------------------------

component IBUFG
port (I : in  STD_LOGIC; O : out STD_LOGIC);
end component;

component BUFG
port (I : in  STD_LOGIC; O : out STD_LOGIC);
end component;

component IBUF
port (I : in  STD_LOGIC; O : out STD_LOGIC);
end component;

component OBUF
port (I : in  STD_LOGIC; O : out STD_LOGIC);
end component;

component OBUF_S_24
port (I : in  STD_LOGIC; O : out STD_LOGIC);
end component;

component IOBUF
port (
  I  : in    STD_LOGIC;
  O  : out   STD_LOGIC;
  IO : inout STD_LOGIC;
  T  : in    STD_LOGIC 
  );
end component;

component OBUFT
port (
  I : in  STD_LOGIC;
  O : out STD_LOGIC;
  T : in  STD_LOGIC
  );
end component;

-------------------------------------------------------------------------------
-- XILINX PARTS SPECIFIC COMPONENTS/SIGNALS
-------------------------------------------------------------------------------

component STARTUP_VIRTEX
port (
  GSR : in std_logic;
  GTS : in std_logic;
  CLK : in std_logic
  );
end component;

component CLKDLL
port (
  CLK0   : out STD_LOGIC;
  CLK90  : out STD_LOGIC;
  CLK180 : out STD_LOGIC;
  CLK270 : out STD_LOGIC;
  CLK2X  : out STD_LOGIC;
  CLKDV  : out STD_LOGIC;
  LOCKED : out STD_LOGIC;
  CLKIN  : in  STD_LOGIC;
  CLKFB  : in  STD_LOGIC;
  RST    : in  STD_LOGIC
  );
end component;

-------------------------------------------------------------------------------
-- SIGNAL ASSIGNMENTS
-------------------------------------------------------------------------------

begin

gnd   <= '0';
rst_i <= NOT rst_n_i;

memory_data_dir_en_i <= WE_n_out_i;

shift_reg0_load_i <= shift_reg_load_en_i(0);
shift_reg1_load_i <= shift_reg_load_en_i(1);

reset_dsp_i <= dsp_rst_reg_i(6 downto 1) OR '0' & dsp_reset_cmd_i & '0';

rodresio_rst_n_out_p <= rodresio_rst_n_out_i AND  NOT fpga_rst_reg_i(6);
formA_rst_n_out_p    <= formA_rst_n_out_i    AND  NOT fpga_rst_reg_i(6);
formB_rst_n_out_p    <= formB_rst_n_out_i    AND  NOT fpga_rst_reg_i(6);
efb_rst_n_out_p      <= efb_rst_n_out_i      AND  NOT fpga_rst_reg_i(6);
router_rst_n_out_p   <= router_rst_n_out_i   AND  NOT fpga_rst_reg_i(6);

-------------------------------------------------------------------------------
-- XILINX PARTS SPECIFIC COMPONENTS/SIGNALS
-------------------------------------------------------------------------------

U1 : STARTUP_VIRTEX
  port map (
    GSR => rst_i,
    GTS => rst_i,
    CLK => cclk_g
    );

--------------------------------------------------------------------------------
--COMPONENT INSTANTIATION
--------------------------------------------------------------------------------

U2 : fpga_cnfg_reset_controller_block
  port map (
    clk_in                     => vme_clk_g,
    cclk_in                    => cclk_g,
    rst_n_in                   => rst_n_i,
    cb_formA_done_in           => formA_done_in_i,
    cb_formA_init_n_in         => formA_init_n_in_i,
    cb_formA_halt_out          => formA_halt_i,
    cb_formB_done_in           => formB_done_in_i,
    cb_formB_init_n_in         => formB_init_n_in_i,
    cb_formB_halt_out          => formB_halt_i,
    cb_efb_done_in             => efb_done_in_i,
    cb_efb_init_n_in           => efb_init_n_in_i,
    cb_efb_halt_out            => efb_halt_i,
    cb_router_done_in          => router_done_in_i,
    cb_router_init_n_in        => router_init_n_in_i,
    cb_router_halt_out         => router_halt_i,
    cb_rrif_done_in            => rodresio_done_in_i,
    cb_rrif_init_n_in          => rodresio_init_n_in_i,
    cb_rrif_halt_out           => rodresio_halt_i,
    cb_fpga_cnfg_reg_in        => fpga_cnfg_reg_i(5 downto 0),
    cb_clr_cnfg_reg_out        => clr_cnfg_reg_i,
    cb_fpga_rst_reg_in         => fpga_rst_reg_i(5 downto 0),
    cb_dsp_rst_reg_in          => reset_dsp_i,    --dsp_rst_reg_i,
    cb_clr_rst_reg_out         => clr_rst_reg_i,
    cb_fpga_rst_status_reg_out => fpga_rst_status_reg_i,
    cb_dsp_rst_status_reg_out  => dsp_rst_status_reg_i,
    cnfg_enable_out            => cnfg_enable_i,
    cnfg_done_out              => fpga_pgrm_done_out_i,
    cnfg_addr_out              => cnfg_addr_i,
    cb_rrif_prgm_n_out         => rodresio_prgm_n_out_i,
    cb_formA_prgm_n_out        => formA_prgm_n_out_i,
    cb_formB_prgm_n_out        => formB_prgm_n_out_i,
    cb_efb_prgm_n_out          => efb_prgm_n_out_i,
    cb_router_prgm_n_out       => router_prgm_n_out_i,
    cb_spare_prgm_n_out        => spare_prgm_n_out_i,
    cb_rrif_rst_n_out          => rodresio_rst_n_out_i,
    cb_formA_rst_n_out         => formA_rst_n_out_i,
    cb_formB_rst_n_out         => formB_rst_n_out_i,
    cb_efb_rst_n_out           => efb_rst_n_out_i,
    cb_router_rst_n_out        => router_rst_n_out_i,
    fpga_sel_out               => fpga_sel_i,
    shift_reg_en_out           => shift_reg_en_i,
    shift_reg_load_out         => shift_reg_load_en_i,
    cb_mdsp_rst_n_out          => dsp_rst_n_out_i(4),
    cb_sdsp0_rst_n_out         => dsp_rst_n_out_i(0),
    cb_sdsp1_rst_n_out         => dsp_rst_n_out_i(1),
    cb_sdsp2_rst_n_out         => dsp_rst_n_out_i(2),
    cb_sdsp3_rst_n_out         => dsp_rst_n_out_i(3),
    cnfg_override_in           => fpga_cnfg_reg_i(6)
    );

U3 : progman_register_block
  port map (
    clk_in                     => vme_clk_g,
    cclk_in                    => cclk_g,
    rst_n_in                   => rst_n_i,
    prb_bus_strb_in            => data_strb_i,
    prb_bus_ce_in              => prm_cen_i,
    prb_bus_rnw_in             => vme_rnw_i,
    prb_bus_addr_in            => vme_addr_i,
    prb_bus_data_in            => vme_data_in_p,
    prb_bus_data_out           => vme_prm_out_i,
    prb_bus_ack_out            => prb_bus_ack_i,
    prb_rrif_done_in           => rodresio_done_in_i,
    prb_rrif_init_n_in         => rodresio_init_n_in_i,
    prb_rrif_halt_in           => rodresio_halt_i,
    prb_formA_done_in          => formA_done_in_i,
    prb_formA_init_n_in        => formA_init_n_in_i,
    prb_formA_halt_in          => formA_halt_i,
    prb_formB_done_in          => formB_done_in_i,
    prb_formB_init_n_in        => formB_init_n_in_i,
    prb_formB_halt_in          => formB_halt_i,
    prb_efb_done_in            => efb_done_in_i,
    prb_efb_init_n_in          => efb_init_n_in_i,
    prb_efb_halt_in            => efb_halt_i,
    prb_router_done_in         => router_done_in_i,
    prb_router_init_n_in       => router_init_n_in_i,
    prb_router_halt_in         => router_halt_i,
    prb_spare_done_in          => spare_done_in_i,
    prb_spare_init_n_in        => spare_init_n_in_i,
    prb_fpga_rst_status_reg_in => fpga_rst_status_reg_i,
    prb_dsp_rst_status_reg_in  => dsp_rst_status_reg_i,
    fpga_cnfg_reg_out          => fpga_cnfg_reg_i,
    fpga_rst_reg_out           => fpga_rst_reg_i,
    dsp_rst_reg_out            => dsp_rst_reg_i,
    flash_cmd_reg_out          => flash_cmd_reg_i,
    flash_status_reg_in        => flash_status_reg_i,
    flash_addr_reg_out         => flash_addr_i,
    flash_data_reg_in          => flash_data_in_i,
    flash_data_reg_out         => flash_data_out_i,
    clr_cnfg_reg_in            => clr_cnfg_reg_i,
    clr_rst_reg_in             => clr_rst_reg_i,
    prb_cnfg_en_in             => cnfg_enable_i,
    serial_number_in           => serial_number_i,
    rod_busy_in                => rod_busy_i,
    rod_clk_status_in          => clk40_locked_i,
    vme_clk_status_in          => vme_clk_locked_i,
    rod_clk_select_in          => rod_clk_sel_p,
    rod_type_in                => rod_type_p,
    fmt_type_in                => fmt_type_p,
    slink_lff_in               => slink_lff_i,
    slink_ldown_in             => slink_ldown_i,
    dsp_detect_in              => dsp_detect_i,
    prm_spares_in              => spare_pin_i,
    mdsp_hrdy_n_in             => mdsp_hrdy_n_i,
    cnfg_done_in               => fpga_pgrm_done_out_i,
    vme_timeout_out            => vme_timeout_i,
    vme_am_code_in             => vme_am_code_i,
    ga_n_in                    => ga_n_i,  --board_addr_n_i,
    gap_n_in                   => vme_gap_n_i,
    trigger_in                 => trigger_i
    );

U4 : flash_decoder_block
  port map (
    clk_in            => vme_clk_g,
    cclk_in           => cclk_g,
    rst_n_in          => rst_n_i,
    fdb_addr_in       => flash_addr_i,
    fdb_data_in       => flash_data_out_i,
    fdb_data_out      => flash_data_in_i,
    fdb_cr_in         => flash_cmd_reg_i,
    fdb_sr_out        => flash_status_reg_i(2 downto 0),
    cnfg_enable_in    => cnfg_enable_i,
    cnfg_addr_in      => cnfg_addr_i,
    cnfg_data_out     => cnfg_data_i,
    memory_addr_out   => flash_mem_addr_out_i,
    memory_data_in    => flash_mem_data_in_i,
    memory_data_out   => flash_mem_data_out_i,
    write_en_n_out    => WE_n_out_i,
    output_en_n_out   => OE_n_out_i,
    CE0_n_out         => CE0_n_out_i,
    shift_reg_load_in => shift_reg_load_en_i
    );

U5 : shift_register
  port map (
    clk_in       => cclk_g,
    rst_n_in     => rst_n_i,
    en_load_in   => shift_reg0_load_i,
    en_shift_in  => shift_reg_en_i(0),
    data_in      => cnfg_data_i,
    data_out     => sr_cnfg_data0_i
    );

U6 : shift_register
  port map (
    clk_in       => cclk_g,
    rst_n_in     => rst_n_i,
    en_load_in   => shift_reg1_load_i,
    en_shift_in  => shift_reg_en_i(1),
    data_in      => cnfg_data_i,
    data_out     => sr_cnfg_data1_i
    );

U7 : hpi_data_controller
  port map (
    clk_in            => vme_clk_g,
    rst_n_in          => rst_n_i,
    mdsp_cen_in       => mdsp_cen_i,
    mdsp_ds_in        => data_strb_i,
    mdsp_hrdy_n_in    => mdsp_hrdy_n_i,
    mdsp_hpi_data_in  => mdsp_hpi_data_in_i,
    mdsp_hpi_data_out => mdsp_hpi_data_out_i,
    mdsp_cen_n_out    => mdsp_cen_n_i,
    mdsp_rnw_out      => mdsp_rnw_i,
    mdsp_hhwil_out    => mdsp_hhwil_i,
    mdsp_hds1_n_out   => mdsp_hds1_n_i,
    vme_bus_rnw_in    => vme_rnw_i,
    mdsp_vme_data_in  => vme_data_in_p,
    mdsp_vme_data_out => vme_hpi_do_i,
    hpi_dtack_n_out   => hpi_dtack_n_i,
    hpi_berr_out      => hpi_berr_i,
    rcf_cen_in        => rcf_cen_i,
    rcf_hrdy_n_in     => rcf_hrdy_n_i,
    rcf_cen_n_out     => rcf_cen_n_i,
    vme_timeout_in    => vme_timeout_i
    );

U8 :  vme_address_decoder
  port map (
    clk_in          => vme_clk_g,
    rst_n_in        => rst_n_i,
    rod_addr_n_in   => board_addr_n_i,
    gaddr_par_n_in  => vme_gap_n_i,
    vme_addr_in     => vme_addr_i,
    vme_am_in       => vme_am_p,
    vme_as_in       => vme_as_p,
    vme_ds0_in      => vme_ds0_p,
    vme_ds1_in      => vme_ds1_p,
    prm_cen_out     => prm_cen_i,
    mdsp_cen_out    => mdsp_cen_i,
    rcf_cen_out     => rcf_cen_i,
    ill_match_out   => ill_match_i,
    data_strb_out   => data_strb_i,
    vme_am_out      => vme_am_code_i
    );

U9 : reset_command_controller
port map (
    clk40_in           => clk40_g,
    clk05_in           => clk05_g,
    rst_n_in           => rst_n_i,
    clr_reset_cmd_in   => clr_rst_reg_i,
    rod_busy_in        => rod_busy_i,
    reset_commands_in  => reset_cmds_in_i,
    reset_commands_out => reset_cmds_out_i,
    reset_sdsp_out     => dsp_reset_cmd_i,
    trigger_out        => trigger_i
    );

-------------------------------------------------------------------------------
-- I/O Buffer / Clock DLL Instantiation
-------------------------------------------------------------------------------
-- 40MHz Clock Circuit 
clk40_ibufg : IBUFG 
port map (I => clk40_in, O => dll_clk40_in);

main_clk_dll : CLKDLL
port map (
  CLK0   => dll_clk40_out,
  CLK90  => open,
  CLK180 => open,
  CLK270 => open,
  CLK2X  => open,
  CLKDV  => dll_clk05_out,
  LOCKED => clk40_locked_i,
  CLKIN  => dll_clk40_in,
  CLKFB  => clk40_g,
  RST    => gnd
  ); 

clk40_bufg : BUFG
port map (I => dll_clk40_out, O => clk40_g);

clk05_bufg : BUFG
port map (I => dll_clk05_out, O => clk05_g);

-- 60MHz Clock Circuit 
vme_clk_ibufg : IBUFG 
port map (I => vme_clk_in, O => vme_clk_ibufg_o);

vme_clk_dll : CLKDLL
port map (
  CLK0   => open,  -- dll_vme_clk_out,
  CLK90  => open,
  CLK180 => open,
  CLK270 => open,
  CLK2X  => dll_vme_clk_out,  -- open;
  CLKDV  => dll_cclk_out,
  LOCKED => vme_clk_locked_i,
  CLKIN  => vme_clk_ibufg_o,
  CLKFB  => vme_clk_g,
  RST    => gnd
  ); 

vme_clk_bufg : BUFG
port map (I => dll_vme_clk_out, O => vme_clk_g);

cclk_bufg : BUFG
port map (I => dll_cclk_out, O => cclk_g);

reset_ibuf : IBUF 
port map (I => rst_n_in, O => rst_n_i);

-------------------------------------------------------------------------------
-- FLASH Memory I/O
memory_address_out_block : 
  for i in 0 to 20 generate
    memory_address_out_obufs : OBUF 
      port map (
        I => flash_mem_addr_out_i(i), 
        O => memory_address_out(i)
        );
  end generate;

memory_data_inout_block : 
  for i in 0 to 7 generate
    memory_data_inout_iobufs : IOBUF 
      port map (
        I =>  flash_mem_data_out_i(i),
        O =>  flash_mem_data_in_i(i),
        IO => memory_data(i),
        T =>  memory_data_dir_en_i
        );
end generate;

memory_busy_n_ibuf : IBUF
port map (I => memory_busy_n_in, O => memory_busy_n_p);

memory_rst_n_obuf : OBUFT
port map (I => gnd, O => memory_rst_n_out, T => memory_rst_n_p);

WE0_n_out_obuf : OBUF
port map (I => WE_n_out_i, O => WE0_n_out);

OE0_n_out_obuf : OBUF
port map (I => OE_n_out_i, O => OE0_n_out);

CE0_n_out_obuf : OBUF
port map (I => CE0_n_out_i, O => CE0_n_out);

-------------------------------------------------------------------------------
-- DSP I/O And Reset commands
mdsp_hpi_data_inout_block : 
  for i in 0 to 15 generate
    mdsp_hpi_data_iobufs : IOBUF 
      port map (
        I  => mdsp_hpi_data_out_i(i),
        O  => mdsp_hpi_data_in_i(i),
        IO => mdsp_hd(i),
        T  => mdsp_hpi_data_dir_i
        );
end generate;

mdsp_hrdy_n_ibuf : IBUF
port map (I => mdsp_hrdy_n_in, O => mdsp_hrdy_n_i);

mdsp_cen_n_obuf : OBUF 
port map (I => mdsp_cen_n_i, O => mdsp_hcs_n_out);

mdsp_rnw_obuf : OBUF 
port map (I => mdsp_rnw_i, O => mdsp_hrnw_out);

mdsp_hds1_n_obuf : OBUF 
port map (I => mdsp_hds1_n_i, O => mdsp_hds1_n_out);

mdsp_hhwil_obuf : OBUF
port map (I => mdsp_hhwil_i, O => mdsp_hhwil_out);

mdsp_hcntl0_obuf : OBUF
port map (I => mdsp_hcntl0_i, O => mdsp_hcntl0_out);

mdsp_hcntl1_obuf : OBUF
port map (I => mdsp_hcntl1_i, O => mdsp_hcntl1_out);

mdsp_boot_mode0_obuf : OBUF 
  port map (I => mdsp_boot0_p, O => mdsp_boot0_out);
mdsp_boot_mode1_obuf : OBUF 
  port map (I => mdsp_boot1_p, O => mdsp_boot1_out);
mdsp_boot_mode2_obuf : OBUF 
  port map (I => mdsp_boot2_p, O => mdsp_boot2_out);
mdsp_boot_mode3_obuf : OBUF 
  port map (I => mdsp_boot3_p, O => mdsp_boot3_out);
mdsp_boot_mode4_obuf : OBUF  
  port map (I => mdsp_boot4_p, O => mdsp_boot4_out);
					 
rcf_cen_n_obuf : OBUF 
port map (I => rcf_cen_n_i, O => rcf_hcs_n_out);

rcf_hrdy_n_ibuf : IBUF
port map (I => rcf_hrdy_n_in, O => rcf_hrdy_n_i);

reset_cmds_out_block : 
  for i in 0 to 1 generate
    reset_cmds_out_obufs : OBUF 
      port map (I => reset_cmds_out_i(i), O => reset_cmds_out(i));
  end generate;

reset_cmds_in_block : 
  for i in 0 to 1 generate
    reset_cmds_in_obufs : IBUF 
      port map (I => reset_cmds_in(i), O => reset_cmds_in_i(i));
  end generate;
  
-------------------------------------------------------------------------------
--  FPGA configuration and reset bits
rodresio_prgbits_out_obuf : OBUF 
port map (I => rodresio_prgbits_out_i, O => rodresio_prgbits_out);

rodresio_cclk_out_obuf : OBUF 
port map (I => rodresio_cclk_out_i, O => rodresio_cclk_out);

rodresio_prgm_n_out_obuf : OBUFT
port map (I => gnd, O => rodresio_prgm_n_out, T => rodresio_prgm_n_out_i);

rodresio_done_in_ibuf : IBUF 
port map (I => rodresio_done_in, O => rodresio_done_in_i);

rodresio_init_n_in_ibuf : IBUF
port map (I =>  rodresio_init_n_in, O => rodresio_init_n_in_i);

rodresio_rst_n_out_obuf : OBUF
port map (I => rodresio_rst_n_out_p, O => rodresio_rst_n_out);

router_prgbits_out_obuf : OBUF 
port map (I => router_prgbits_out_i, O => router_prgbits_out);

router_cclk_out_obuf : OBUF 
port map (I => router_cclk_out_i, O => router_cclk_out);

router_prgm_n_out_obuf : OBUFT
port map (I => gnd, O => router_prgm_n_out, T => router_prgm_n_out_i);

router_done_in_ibuf : IBUF 
port map (I => router_done_in, O => router_done_in_i);

router_init_n_in_ibuf : IBUF
port map (I =>  router_init_n_in, O =>  router_init_n_in_i);

router_rst_n_out_obuf : OBUF
port map (I => router_rst_n_out_p, O => router_rst_n_out);

efb_prgbits_out_obuf : OBUF 
port map (I => efb_prgbits_out_i, O => efb_prgbits_out);

efb_cclk_out_obuf : OBUF
port map (I => efb_cclk_out_i, O => efb_cclk_out);

efb_prgm_n_out_obuf : OBUFT
port map (I => gnd, O => efb_prgm_n_out, T => efb_prgm_n_out_i);

efb_done_in_ibuf : IBUF 
port map (I => efb_done_in, O => efb_done_in_i);

efb_init_n_in_ibuf : IBUF 
port map (I =>  efb_init_n_in, O =>  efb_init_n_in_i);

efb_rst_n_out_obuf : OBUF
port map (I => efb_rst_n_out_p, O => efb_rst_n_out);

formA_prgbits_out_obuf : OBUF 
port map (I => formA_prgbits_out_i, O => formA_prgbits_out);

formA_cclk_out_obuf : OBUF
port map (I => formA_cclk_out_i, O => formA_cclk_out);

formA_prgm_n_out_obuf : OBUFT
port map (I => gnd, O => formA_prgm_n_out, T => formA_prgm_n_out_i);

formA_done_in_ibuf : IBUF 
port map (I => formA_done_in, O => formA_done_in_i);

formA_init_n_in_ibuf : IBUF   
port map (I => formA_init_n_in, O => formA_init_n_in_i);

formA_rst_n_out_obuf : OBUF
port map (I => formA_rst_n_out_p, O => formA_rst_n_out);

formB_prgbits_out_obuf : OBUF 
port map (I => formB_prgbits_out_i, O => formB_prgbits_out);

formB_cclk_out_obuf : OBUF
port map (I => formB_cclk_out_i, O => formB_cclk_out);

formB_prgm_n_out_obuf : OBUFT
port map (I => gnd, O => formB_prgm_n_out, T => formB_prgm_n_out_i);

formB_done_in_ibuf : IBUF 
port map (I => formB_done_in, O => formB_done_in_i);

formB_init_n_in_ibuf : IBUF 
port map (I =>  formB_init_n_in, O =>  formB_init_n_in_i);

formB_rst_n_out_obuf : OBUF
port map (I => formB_rst_n_out_p, O => formB_rst_n_out);

spare_prgbits_out_obuf : OBUF 
port map (I => spare_prgbits_out_i, O => spare_prgbits_out);

spare_cclk_out_obuf : OBUF
port map (I => spare_cclk_out_i, O => spare_cclk_out);

spare_prgm_n_out_obuf : OBUFT
port map (I => gnd, O => spare_prgm_n_out, T => spare_prgm_n_out_i);

sp_done_input_block : 
  for i in 0 to 3 generate
    spare_done_in_ibuf : IBUF 
      port map (
	   I => spare_done_in(i),
	   O => spare_done_in_i(i)
	   );
end generate;

spare_init_n_in_ibuf : IBUF 
port map (I =>  spare_init_n_in, O =>  spare_init_n_in_i);

dsp_rst_n_out_block : 
  for i in 0 to 4 generate
    dsp_rst_n_out_obufts : OBUFT
      port map (
        I => gnd,
        O => dsp_rst_n_out(i),
        T => dsp_rst_n_out_i(i)
        );
end generate;

fpga_pgrm_done_out_obuf : OBUF_S_24
  port map (I => rod_busy_n_i, O => fpga_pgrm_done_out);

-------------------------------------------------------------------------------
-- ROD Status
rod_busy_ibuf : IBUF 
port map (I =>  rod_busy_in, O =>  rod_busy_n_p);  -- Low true signal

rod_busy_n_obuf : OBUF
port map (I => rod_busy_n_i, O => rod_busy_n_out);

rod_type_ibuf : IBUF
port map (I => rod_type_in, O => rod_type_p);

fmt_type_block : for i in 0 to 1 generate
  fmt_type_ibuf : IBUF
    port map (
      I => fmt_type_in(i), 
      O => fmt_type_p(i)
      );
end generate;

slink_lff_n_ibuf : IBUF
port map (I => slink_lff_n_in, O => slink_lff_n_p);

slink_down_n_ibuf : IBUF
port map (I => slink_ldown_n_in, O => slink_ldown_n_p);

rod_clock_select_ibuf : IBUF
port map (I => rod_clock_select_in, O => rod_clk_sel_p);

dsp0_detect_n_ibuf : IBUF
port map (I => dsp0_detect_n_in, O => dsp0_detect_n_p);

dsp1_detect_n_ibuf : IBUF
port map (I => dsp1_detect_n_in, O => dsp1_detect_n_p);

dsp2_detect_n_ibuf : IBUF
port map (I => dsp2_detect_n_in, O => dsp2_detect_n_p);

dsp3_detect_n_ibuf : IBUF
port map (I => dsp3_detect_n_in, O => dsp3_detect_n_p);

-------------------------------------------------------------------------------
-- Serial Number Inputs
serial_number_in_block :
  for i in 0 to 9 generate
    serial_number_ibufs : IBUF
      port map (I => serial_number_n_in(i), O => serial_number_n_i(i));
        
    serial_number_i(i) <= NOT serial_number_n_i(i);
  end generate;

-------------------------------------------------------------------------------
-- VMEBus I/O 
vme_data_inout_block : 
  for i in 0 to 31 generate
    vme_data_inout_iobufs : IOBUF 
      port map (
        I =>  vme_data_out_p(i),
        O =>  vme_data_in_p(i),
        IO => vme_data_inout(i),
        T =>  vme_data_dir_i
        );
  end generate;

vme_de_n_obuf : OBUF 
  port map (I => vme_de_n_p, O => vme_de_n_out);

vme_ddir_obuf : OBUF 
  port map (I => vme_ddir_p, O => vme_ddir_out);

vme_addr_in_block : 
  for i in 0 to 31 generate 
    vme_addr_in_ibufs : IBUF 
      port map (I => vme_addr_in(i), O => vme_addr_i(i));
  end generate;

vme_am_in_block : 
  for i in 0 to 5 generate 
    vme_am_in_ibufs : IBUF 
      port map (I => vme_am_in(i), O => vme_am_p(i));
  end generate;

vme_as_ibuf : IBUF 
  port map (I => vme_as_in, O => vme_as_p);

vme_ds0_ibuf : IBUF 
  port map (I => vme_ds0_in, O => vme_ds0_p);

vme_ds1_ibuf : IBUF 
  port map (I => vme_ds1_in, O => vme_ds1_p);

vme_wr_ibuf : IBUF 
  port map (I => vme_wr_in, O => vme_wr_p);

lirq_n_ibuf : IBUF 
  port map (I => lirq_n_in, O => lirq_n_p);

lirq_in_block : 
  for i in 0 to 2 generate 
    lirq_in_ibufs : IBUF 
      port map (I => lirq_in(i), O => lirq_val_p(i));
  end generate;

vme_iack_ibuf : IBUF 
  port map (I => vme_iack_in, O => vme_iack_p);

vme_iackin_n_ibuf : IBUF 
  port map (I => vme_iackin_n_in, O => vme_iackin_n_p);

vme_iackout_n_obuf : OBUF  --
  port map (I => vme_iackout_n_p, O => vme_iackout_n_out);

vme_int_enable_obuf : OBUF  --  signal to RCF that interrupt cycle is active
  port map (I => lvme_iack_i, O => spare_pout(0));

vme_sysfail_n_ibuf : IBUF 
  port map (I => vme_sysfail_n_in, O => vme_sysfail_n_p);

vme_dtack_n_obuf : OBUF 
  port map (I => vme_dtack_n_p, O => vme_dtack_n_out);

dtack_en_n_obuf : OBUF 
  port map (I => dtack_en_n_p, O => dtack_en_n_out);

vme_berr_n_obuf : OBUF 
  port map (I => vme_berr_n_p, O => vme_berr_n_out);

berr_en_n_obuf : OBUF 
  port map (I => berr_en_n_p, O => berr_en_n_out);

vme_gap_n_ibuf : IBUF 
  port map (I => vme_gap_n_in, O => vme_gap_n_p);

vme_ga_n_block : 
  for i in 0 to 4 generate 
    vme_ga_n_ibufs : IBUF 
      port map (I => board_addr_n_in(i), O => ga_n_i(i));  -- board_addr_n_i(i));
  end generate;

-------------------------------------------------------------------------------
-- Test Output I/O
test_ibuf_block :
  for i in 0 to 7 generate 
    test_in_ibufs : IBUF 
      port map (I => test_in(i), O => test_i(i));
  end generate;
--test0_obuf : OBUF  --  
--port map (I => prm_cen_i, O => test_out(0));
--
--test1_obuf : OBUF  -- 
--port map (I => vme_clk_g, O => test_out(1));
--
--test2_obuf : OBUF  -- 
--port map (I => vme_dtack_n_p, O => test_out(2)); 
--
--test3_obuf : OBUF  -- 
--port map (I => vme_data_dir_i, O => test_out(3)); 
--
--test4_obuf : OBUF  --  
--port map (I => data_strb_i, O => test_out(4));
--
--test5_obuf : OBUF  --  
--port map (I => mdsp_cen_i, O => test_out(5));
--
--test6_obuf : OBUF  --   
--port map (I => vme_de_n_p, O => test_out(6));
--
--test7_obuf : OBUF  --  
--port map (I => vme_wr_p, O => test_out(7));

spare_pin_block : 
  for i in 0 to 7 generate 
    spare_pin_ibufs : IBUF 
      port map (I => spare_pin(i), O => spare_pin_i(i));
  end generate;
-- spare_pin(1 downto 0) == Revision LSB Indicator spare_pin(0) grounded on the revE1 board

clock_select_obuf : OBUF  --  signal to RCF that interrupt cycle is active
port map (I => rod_clk_sel_p, O => spare_pout(1));


-------------------------------------------------------------------------------
-- PROCESS DECLARATION
-------------------------------------------------------------------------------

sr_cnfg_data_mux : process (
  cclk_g, 
  rst_n_i, 
  shift_reg_en_i, 
  shift_reg_en_d1,
  sr_cnfg_data0_i, 
  sr_cnfg_data1_i
  )
begin
  if (rst_n_i = '0') then
    sr_cnfg_data_i <= '0';
  elsif (cclk_g'event and cclk_g = '1') then
    shift_reg_en_d1 <= shift_reg_en_i;
    if (shift_reg_en_d1 = "01") then
      sr_cnfg_data_i <= sr_cnfg_data0_i;
    elsif (shift_reg_en_d1 = "10") then
      sr_cnfg_data_i <= sr_cnfg_data1_i;
    else
      sr_cnfg_data_i <= '0';
    end if;
  end if;
end process sr_cnfg_data_mux;

-------------------------------------------------------------------------------
fpga_pgrm_mux : process (
  vme_clk_g, 
  cclk_g, 
  rst_n_i, 
  fpga_sel_i, 
  sr_cnfg_data_i
  )
begin
  if (rst_n_i = '0') then
    efb_prgbits_out_i      <= '0';
    efb_cclk_out_i         <= '0';
    formA_prgbits_out_i    <= '0';
    formA_cclk_out_i       <= '0';
    formB_prgbits_out_i    <= '0';
    formB_cclk_out_i       <= '0';
    rodresio_prgbits_out_i <= '0';
    rodresio_cclk_out_i    <= '0';
    router_prgbits_out_i   <= '0';
    router_cclk_out_i      <= '0';
    spare_prgbits_out_i    <= '0';
    spare_cclk_out_i       <= '0';
  elsif (vme_clk_g'event and vme_clk_g = '1') then
    case fpga_sel_i is
      when "000" =>        -- select rrif
        efb_prgbits_out_i      <= '0';
        efb_cclk_out_i         <= '0';
        formA_prgbits_out_i    <= '0';
        formA_cclk_out_i       <= '0';
        formB_prgbits_out_i    <= '0';
        formB_cclk_out_i       <= '0';
        rodresio_prgbits_out_i <= sr_cnfg_data_i;
        rodresio_cclk_out_i    <= NOT cclk_g;
        router_prgbits_out_i   <= '0';
        router_cclk_out_i      <= '0'; 
      when "001" =>        -- select both formatter banks
        efb_prgbits_out_i      <= '0';
        efb_cclk_out_i         <= '0';
        formA_prgbits_out_i    <= sr_cnfg_data_i;
        formA_cclk_out_i       <= NOT cclk_g;
        formB_prgbits_out_i    <= sr_cnfg_data_i;
        formB_cclk_out_i       <= NOT cclk_g;
        rodresio_prgbits_out_i <= '0';
        rodresio_cclk_out_i    <= '0';
        router_prgbits_out_i   <= '0';
        router_cclk_out_i      <= '0'; 
      when "010" =>        -- select formatter A bank
        efb_prgbits_out_i      <= '0';
        efb_cclk_out_i         <= '0';
        formA_prgbits_out_i    <= sr_cnfg_data_i;
        formA_cclk_out_i       <= NOT cclk_g;
        formB_prgbits_out_i    <= '0';
        formB_cclk_out_i       <= '0';
        rodresio_prgbits_out_i <= '0';
        rodresio_cclk_out_i    <= '0';
        router_prgbits_out_i   <= '0';
        router_cclk_out_i      <= '0'; 
      when "011" =>        -- select formatter B bank
        efb_prgbits_out_i      <= '0';
        efb_cclk_out_i         <= '0';
        formA_prgbits_out_i    <= '0';
        formA_cclk_out_i       <= '0';
        formB_prgbits_out_i    <= sr_cnfg_data_i;
        formB_cclk_out_i       <= NOT cclk_g;
        rodresio_prgbits_out_i <= '0';
        rodresio_cclk_out_i    <= '0';
        router_prgbits_out_i   <= '0';
        router_cclk_out_i      <= '0'; 
      when "100" =>        -- select efb
        efb_prgbits_out_i      <= sr_cnfg_data_i;
        efb_cclk_out_i         <= NOT cclk_g;
        formA_prgbits_out_i    <= '0';
        formA_cclk_out_i       <= '0';
        formB_prgbits_out_i    <= '0';
        formB_cclk_out_i       <= '0';
        rodresio_prgbits_out_i <= '0';
        rodresio_cclk_out_i    <= '0';
        router_prgbits_out_i   <= '0';
        router_cclk_out_i      <= '0'; 
      when "101" =>        -- select router
        efb_prgbits_out_i      <= '0';
        efb_cclk_out_i         <= '0';
        formA_prgbits_out_i    <= '0';
        formA_cclk_out_i       <= '0';
        formB_prgbits_out_i    <= '0';
        formB_cclk_out_i       <= '0';
        rodresio_prgbits_out_i <= '0';
        rodresio_cclk_out_i    <= '0';
        router_prgbits_out_i   <= sr_cnfg_data_i;
        router_cclk_out_i      <= NOT cclk_g; 
      when "110" =>        -- select Spares
        efb_prgbits_out_i      <= '0';
        efb_cclk_out_i         <= '0';
        formA_prgbits_out_i    <= '0';
        formA_cclk_out_i       <= '0';
        formB_prgbits_out_i    <= '0';
        formB_cclk_out_i       <= '0';
        rodresio_prgbits_out_i <= '0';
        rodresio_cclk_out_i    <= '0';
        router_prgbits_out_i   <= '0';
        router_cclk_out_i      <= '0'; 
        spare_prgbits_out_i    <= sr_cnfg_data_i;
        spare_cclk_out_i       <= NOT cclk_g; 
      when others =>
        efb_prgbits_out_i      <= '0';
        efb_cclk_out_i         <= '0';
        formA_prgbits_out_i    <= '0';
        formA_cclk_out_i       <= '0';
        formB_prgbits_out_i    <= '0';
        formB_cclk_out_i       <= '0';
        rodresio_prgbits_out_i <= '0';
        rodresio_cclk_out_i    <= '0';
        router_prgbits_out_i   <= '0';
        router_cclk_out_i      <= '0';
        spare_prgbits_out_i    <= '0';
        spare_cclk_out_i       <= '0';
    end case;
  end if;
end process fpga_pgrm_mux;


-------------------------------------------------------------------------------
rod_busy : process (
  clk40_g, 
  rst_n_i
  )
begin
  if (rst_n_i = '0') then
    rod_busy_i   <= '1';
    rod_busy_n_i <= '0';
  elsif (clk40_g'event AND clk40_g = '1') then
    rod_busy_i   <= NOT fpga_pgrm_done_out_i OR NOT rod_busy_n_p;
    rod_busy_n_i <= NOT rod_busy_i;
  end if;
end process rod_busy;

-------------------------------------------------------------------------------
vme_data_out_mux : process (
  vme_clk_g, 
  rst_n_i
  )
begin
  if (rst_n_i = '0') then
    vme_data_out_p <= (others => '1');
  elsif (vme_clk_g'event AND vme_clk_g = '1') then
    if (prm_cen_i = '1') then
      vme_data_out_p <= vme_prm_out_i;
    elsif (mdsp_cen_i = '1') then
      vme_data_out_p <= vme_hpi_do_i;
    elsif (lvme_iack_i = '1') then
      vme_data_out_p <= "111111111111111111111111000" & NOT board_addr_n_i;  -- interrupt vector
    else
      vme_data_out_p <= (others => '1');
    end if;
  end if; 
end process vme_data_out_mux;

vme_data_output_en : process (
  vme_clk_g, 
  rst_n_i
  )
begin
  if (rst_n_i = '0') then
    vme_data_dir_i <= '1';  -- set FPGA data port to receive
    vme_ddir_p     <= '0';  -- lvt162245 set to receive
    vme_de_n_p     <= '1';  -- tristate lvt162245
  elsif (vme_clk_g'event AND vme_clk_g = '1') then
    vme_data_dir_i <= NOT (vme_as_i AND (prm_cen_i OR mdsp_cen_i OR lvme_iack_i) AND vme_rnw_i);
    vme_ddir_p     <=      vme_as_i AND (prm_cen_i OR mdsp_cen_i OR lvme_iack_i) AND vme_rnw_i;
    if (vme_sysfail_i = '1') then
      vme_de_n_p <= '1';  -- tristate lvt162245
    else
      vme_de_n_p <= NOT (vme_as_i AND (prm_cen_i OR mdsp_cen_i OR lvme_iack_i) AND vme_ds_i);
    end if;
  end if;
end process vme_data_output_en;

vme_dtack_output : process (
  vme_clk_g, 
  rst_n_i
  )
begin
  if (rst_n_i = '0') then
    vme_dtack_n_p <= '1';
    dtack_en_n_p  <= '1';
    dtack_en_n_d1 <= '1';
  elsif (vme_clk_g'event AND vme_clk_g = '1') then
    dtack_en_n_p  <= dtack_en_n_d1;
    if (vme_sysfail_i = '1') then
      vme_dtack_n_p <= '1';
      dtack_en_n_d1 <= '1';
    elsif (prm_cen_i = '1' AND vme_as_i = '1') then
      dtack_en_n_d1 <= NOT prb_bus_ack_i;
      vme_dtack_n_p <= NOT prb_bus_ack_i;
    elsif (mdsp_cen_i = '1' AND vme_as_i = '1') then
      dtack_en_n_d1 <= hpi_dtack_n_i;
      vme_dtack_n_p <= hpi_dtack_n_i;
    elsif (lvme_iack_i = '1' AND vme_as_i = '1') then
      dtack_en_n_d1 <= NOT vme_ds_i;
      vme_dtack_n_p <= NOT vme_ds_i;
    else
      vme_dtack_n_p <= '1';
      dtack_en_n_d1 <= '1';
    end if;
  end if;
end process ;

vme_berr_output : process (
  vme_clk_g, 
  rst_n_i
  )
begin
  if (rst_n_i = '0') then
    berr_en_n_d1 <= '1';
    berr_en_n_p  <= '1';
    vme_berr_n_p <= '1';
  elsif (vme_clk_g'event AND vme_clk_g = '1') then
    if (vme_sysfail_i = '1') then
      berr_en_n_d1 <= '1';
      berr_en_n_p  <= '1';
      vme_berr_n_p <= '1';
    else
      berr_en_n_d1 <= NOT (ill_match_i OR hpi_berr_i);
      berr_en_n_p  <= berr_en_n_d1;
      vme_berr_n_p <= NOT (ill_match_i OR hpi_berr_i);
    end if;
  end if;
end process ;

vme_int_decode : process (
  vme_clk_g, 
  rst_n_i
  )
begin
  if (rst_n_i = '0') then
    lvme_iack_i <= '0';
  elsif (vme_clk_g'event AND vme_clk_g = '1') then
    if (vme_iack_i = '1' AND vme_iackout_n_p = '1' AND vme_as_i = '1') then
      if (vme_addr_i(3 downto 1) = lirq_val_p AND lirq_i = '1') then
        lvme_iack_i <= '1';
      else
        lvme_iack_i <= '0';
      end if;      
    else
      lvme_iack_i <= '0';
    end if;
  end if;
end process ;

hpi_data_output_en : process (
  vme_clk_g, 
  rst_n_i
  )
begin
  if (rst_n_i = '0') then
    mdsp_hpi_data_dir_i <= '1';
  elsif (vme_clk_g'event AND vme_clk_g = '1') then
    if (mdsp_cen_i = '1' AND vme_rnw_i = '0') then
      mdsp_hpi_data_dir_i <= '0';  -- output to MDSP
    else
      mdsp_hpi_data_dir_i <= '1';  -- input from MDSP
    end if;
  end if;
end process hpi_data_output_en;

vme_to_mdsp_register : process (
  vme_clk_g, 
  rst_n_i
  )
begin
  if (rst_n_i = '0') then
    mdsp_hcntl1_i <= '0';
    mdsp_hcntl0_i <= '0';
  elsif (vme_clk_g'event AND vme_clk_g = '1') then
    mdsp_hcntl1_i <= NOT vme_addr_i(23) AND vme_addr_i(22);
    mdsp_hcntl0_i <= NOT vme_addr_i(23) AND vme_addr_i(21);
  end if;
end process vme_to_mdsp_register;

vme_signal_register : process (
  vme_clk_g, 
  rst_n_i
  )
begin
  if (rst_n_i = '0') then
    vme_rnw_i       <= '0';
    vme_as_i        <= '0';
    vme_ds_i        <= '0';
    vme_iackout_n_p <= '1';
    vme_iack_i      <= '0';
    lirq_i          <= '0';
    vme_sysfail_i   <= '0';
  elsif (vme_clk_g'event AND vme_clk_g = '1') then
    vme_rnw_i       <= NOT vme_wr_p;
    vme_as_i        <= vme_as_p;
    vme_ds_i        <= vme_ds0_p AND vme_ds1_p;
    vme_iackout_n_p <= vme_iackin_n_p;
    vme_iack_i      <= vme_iack_p;
    lirq_i          <= NOT lirq_n_p;
    vme_sysfail_i   <= NOT vme_sysfail_n_p;
  end if;
end process vme_signal_register;

clk40_signal_register : process (
  clk40_g, 
  rst_n_i
  )
begin
  if (rst_n_i = '0') then
    slink_lff_i           <= '0';
    slink_ldown_i         <= '0';
    flash_status_reg_i(3) <= '0';
    dsp_detect_i          <= (others => '0');
    memory_rst_n_p        <= '0';
  elsif (clk40_g'event AND clk40_g = '1') then
    slink_lff_i           <= NOT slink_lff_n_p;
    slink_ldown_i         <= NOT slink_ldown_n_p;
    flash_status_reg_i(3) <= NOT memory_busy_n_p;
    dsp_detect_i(0)       <= NOT dsp0_detect_n_p;
    dsp_detect_i(1)       <= NOT dsp1_detect_n_p;
    dsp_detect_i(2)       <= NOT dsp2_detect_n_p;
    dsp_detect_i(3)       <= NOT dsp3_detect_n_p;
    memory_rst_n_p        <= '1';
  end if;
end process clk40_signal_register;

mdsp_boot_mode_select : process (
  clk40_g, 
  rst_n_i
  )
begin
  if (rst_n_i = '0') then
    mdsp_boot4_p <= '0';   -- "01101" ROM, Map 1, Internal Memory at Loc 0x0h 
    mdsp_boot3_p <= '1';
    mdsp_boot2_p <= '1';
    mdsp_boot1_p <= '0';
    mdsp_boot0_p <= '1';
  elsif (clk40_g'event AND clk40_g = '1') then
    if (dsp_rst_reg_i(0) = '0') then
      mdsp_boot4_p <= '0'; -- "01101" ROM, Map 1, Internal Memory at Loc 0x0h 
      mdsp_boot3_p <= '1';
      mdsp_boot2_p <= '1';
      mdsp_boot1_p <= '0';
      mdsp_boot0_p <= '1';
    elsif (dsp_rst_reg_i(0) = '1') then
      mdsp_boot4_p <= '0'; -- "00111" HPI, Map 1, Internal Memory at Loc 0x0h 
      mdsp_boot3_p <= '0';
      mdsp_boot2_p <= '1';
      mdsp_boot1_p <= '1';
      mdsp_boot0_p <= '1';
    end if;
  end if;
end process mdsp_boot_mode_select;

board_addr : process (
   ga_n_i,
   test_i
   )
begin
  if (test_i(7 downto 6) = "00") then
    board_addr_n_i <= test_i(4 downto 0);
    vme_gap_n_p    <= test_i(5);
  else
    board_addr_n_i <= ga_n_i;
    vme_gap_n_p    <= vme_gap_n_p;
  end if;
end process;

-------------------------------------------------------------------------------

end rtl;
