--  Lawrence Berkeley National Laboratory
--------------------------------------------------------------------------------
--  Filename: cnfg_reset_controller.vhd
--  Title: FPGA Program Reset Manager
--  Description:    This FPGA controls configuration of all FPGA parts, and
--                  controls the reset signals to all FPGAs and DSPs on the
--                  Read Out Driver Board used in Atlas.
--
--------------------------------------------------------------------------------
-- Author: John Joseph
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
-- Revision History
--------------------------------------------------------------------------------
--
--  June 8, 2000  JMJ defined top ports to match schematic requirements
--  Aug 21, 2000  JMJ began to revise code for first prototype

--------------------------------------------------------------------------------
--  Library Declaration
--------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;          --needed for logical operations
use IEEE.std_logic_arith.all;         --needed for +/- operations

--------------------------------------------------------------------------------
--  PORT DECLARATION
--------------------------------------------------------------------------------

entity fpga_cnfg_reset_controller_block is
port (
  clk_in   : in  std_logic;
  cclk_in  : in  std_logic;
  rst_n_in : in  std_logic;

  cb_formA_done_in    : in  std_logic;
  cb_formA_init_n_in  : in  std_logic;
  cb_formA_halt_out   : out std_logic;
  cb_formB_done_in    : in  std_logic;
  cb_formB_init_n_in  : in  std_logic;
  cb_formB_halt_out   : out std_logic;
  cb_efb_done_in      : in  std_logic;
  cb_efb_init_n_in    : in  std_logic;
  cb_efb_halt_out     : out std_logic;
  cb_router_done_in   : in  std_logic;
  cb_router_init_n_in : in  std_logic;
  cb_router_halt_out  : out std_logic;
  cb_rrif_done_in     : in  std_logic;
  cb_rrif_init_n_in   : in  std_logic;
  cb_rrif_halt_out    : out std_logic;
 
  cb_fpga_cnfg_reg_in : in  std_logic_vector( 5 downto 0);
  cb_clr_cnfg_reg_out : out std_logic;
  cb_fpga_rst_reg_in  : in  std_logic_vector( 5 downto 0);
  cb_dsp_rst_reg_in   : in  std_logic_vector( 5 downto 0);
  cb_clr_rst_reg_out  : out std_logic;

  cb_fpga_rst_status_reg_out : out std_logic_vector(4 downto 0);
  cb_dsp_rst_status_reg_out  : out std_logic_vector(4 downto 0);

  cnfg_enable_out : out std_logic;
  cnfg_done_out   : out std_logic;
  cnfg_addr_out   : out std_logic_vector(23 downto 0);

  cb_rrif_prgm_n_out   : out std_logic;
  cb_formA_prgm_n_out  : out std_logic;
  cb_formB_prgm_n_out  : out std_logic;
  cb_efb_prgm_n_out    : out std_logic;
  cb_router_prgm_n_out : out std_logic;
  cb_spare_prgm_n_out  : out std_logic;

  cb_rrif_rst_n_out    : out std_logic;
  cb_formA_rst_n_out   : out std_logic;
  cb_formB_rst_n_out   : out std_logic;
  cb_efb_rst_n_out     : out std_logic;
  cb_router_rst_n_out  : out std_logic;

  fpga_sel_out : out std_logic_vector(2 downto 0);
        --  "000" -> RRIF Select
        --  "001" -> FormAB Select
        --  "010" -> FormA Select
        --  "011" -> FormB Select
        --  "100" -> EFB Select
        --  "101" -> Router Select

  shift_reg_en_out   : out std_logic_vector(1 downto 0);
  shift_reg_load_out : out std_logic_vector(1 downto 0);

  cb_mdsp_rst_n_out  : out std_logic;
  cb_sdsp0_rst_n_out : out std_logic;
  cb_sdsp1_rst_n_out : out std_logic;
  cb_sdsp2_rst_n_out : out std_logic;
  cb_sdsp3_rst_n_out : out std_logic;
    
  cnfg_override_in    : in  std_logic
  );
end fpga_cnfg_reset_controller_block;

architecture rtl of fpga_cnfg_reset_controller_block is

--------------------------------------------------------------------------------
--SIGNAL DECLARATION
--------------------------------------------------------------------------------

type cnfg_states is (
  startup, 
  cnfg_all,
  init_spare, 
  init_rrif, 
  cnfg_rrif, 
  init_form, 
  cnfg_form,
  init_efb, 
  cnfg_efb, 
  init_router, 
  cnfg_router,
  halt_at_rrif, 
  halt_at_form, 
  halt_at_efb, 
  halt_at_router,
  wait_for_rrif, 
  wait_for_form, 
  wait_for_efb,
  wait_for_router, 
  cnfg_done, 
  clear_cnfg_all_bit,
  load_rrif_counter, 
  load_router_counter, 
  load_efb_counter,
  load_form_counter, 
  wait_for_rrif_counter_load,
  wait_for_router_counter_load, 
  wait_for_efb_counter_load,
  wait_for_form_counter_load, 
  fpga_load_error
  );
signal cnfg_state : cnfg_states;

type mode_states is (
  ms_cnfg_all, 
  ms_cnfg_rrif, 
  ms_cnfg_form, 
  ms_cnfg_efb,
  ms_cnfg_router, 
  ms_no_cnfg
  );
signal cnfg_mode  : mode_states;     --to identify current configuration state

-- internal registers to hold start and stop location for all FPGA data
constant efb_cnfg_start_addr_i    : std_logic_vector(23 downto 0) := "111100000000000000000000"; -- 0xF00000h
constant efb_cnfg_stop_addr_i     : std_logic_vector(23 downto 0) := "111101010010001100100111"; -- 0xF52327h
constant form_cnfg_start_addr_i   : std_logic_vector(23 downto 0) := "111010000000000000000000"; -- 0xE80000h
constant form_cnfg_stop_addr_i    : std_logic_vector(23 downto 0) := "111011111000111001101111"; -- 0xEF8E6Fh
constant router_cnfg_start_addr_i : std_logic_vector(23 downto 0) := "111110000000000000000000"; -- 0xF80000h
constant router_cnfg_stop_addr_i  : std_logic_vector(23 downto 0) := "111111010010001100100111"; -- 0xFD2327h
constant rrif_cnfg_start_addr_i   : std_logic_vector(23 downto 0) := "111000000000000000000000"; -- 0xE00000h
constant rrif_cnfg_stop_addr_i    : std_logic_vector(23 downto 0) := "111001111000111001100011"; -- 0xE78E63h

signal cnfg_start_addr_i : std_logic_vector(23 downto 0); 
signal cnfg_stop_addr_i  : std_logic_vector(23 downto 0);

--  internal control signals for the Flash memories
signal enable_counter_i     : std_logic;  
signal load_counter_i       : std_logic;
signal flash_counter_init_i : std_logic;
signal flash_count_i        : unsigned(23 downto 0);

--  internal counter enable used during configuration process
signal enable_cnfg_counter_i  : std_logic;

--  internal status flags
signal first_load_i          : std_logic;
signal last_cnfg_word_i      : std_logic;
signal load_last_cnfg_word_i : std_logic;
signal sr0_empty_i           : std_logic;
signal sr1_empty_i           : std_logic;

--  internal signals that indicate the status of the cnfg counter
signal rrif_cntr_init_i   : std_logic;
signal form_cntr_init_i   : std_logic;
signal efb_cntr_init_i    : std_logic;
signal router_cntr_init_i : std_logic;
signal fpga_init_n_i      : std_logic;

--  internal signals that indicate the status of each FPGA
signal rrif_ready_i   : std_logic;
signal form_ready_i   : std_logic;
signal efb_ready_i    : std_logic;
signal router_ready_i : std_logic;

-- internal reset signals
signal rc_rrif_rst_n_i   : std_logic;
signal rc_formA_rst_n_i  : std_logic;
signal rc_formB_rst_n_i  : std_logic;
signal rc_efb_rst_n_i    : std_logic;
signal rc_router_rst_n_i : std_logic;
signal rc_mdsp_rst_n_i   : std_logic;
signal rc_sdsp0_rst_n_i  : std_logic;
signal rc_sdsp1_rst_n_i  : std_logic;
signal rc_sdsp2_rst_n_i  : std_logic;
signal rc_sdsp3_rst_n_i  : std_logic;

signal csm_rrif_rst_n_i   : std_logic;
signal csm_formA_rst_n_i  : std_logic;
signal csm_formB_rst_n_i  : std_logic;
signal csm_efb_rst_n_i    : std_logic;
signal csm_router_rst_n_i : std_logic;
signal csm_sdsp0_rst_n_i  : std_logic;
signal csm_sdsp1_rst_n_i  : std_logic;
signal csm_sdsp2_rst_n_i  : std_logic;
signal csm_sdsp3_rst_n_i  : std_logic;

--  internal signals that indicate the status of the confg process
signal halt_at_rrif_i   : std_logic;
signal halt_at_formA_i  : std_logic;
signal halt_at_formB_i  : std_logic;
signal halt_at_efb_i    : std_logic;
signal halt_at_router_i : std_logic;

--  shift register status and control
signal shift_reg0_count_i : unsigned(2 downto 0);
signal shift_reg1_count_i : unsigned(2 downto 0);
signal shift_reg_en_out_i : std_logic_vector(1 downto 0);
signal shift_reg_en_d1    : std_logic_vector(1 downto 0);
signal shift_reg_load_i   : std_logic_vector(1 downto 0);
signal flush_shift_reg_i  : std_logic;

--  internal signals for process counters
signal wait_count_i      : unsigned(5 downto 0);
signal reset_count_i     : unsigned(18 downto 0);
signal init_count_i      : unsigned(11 downto 0);
signal timeout_count_i   : unsigned(7 downto 0);

signal en_reset_control_i : std_logic;

--------------------------------------------------------------------------------
--COMPONENT DECLARATION
--------------------------------------------------------------------------------

begin
--------------------------------------------------------------------------------
-- Signal Connections
--------------------------------------------------------------------------------

cb_rrif_rst_n_out   <= csm_rrif_rst_n_i   AND rc_rrif_rst_n_i;
cb_formA_rst_n_out  <= csm_formA_rst_n_i  AND rc_formA_rst_n_i;
cb_formB_rst_n_out  <= csm_formB_rst_n_i  AND rc_formB_rst_n_i;
cb_efb_rst_n_out    <= csm_efb_rst_n_i    AND rc_efb_rst_n_i;
cb_router_rst_n_out <= csm_router_rst_n_i AND rc_router_rst_n_i;
cb_mdsp_rst_n_out   <= rc_mdsp_rst_n_i;
cb_sdsp0_rst_n_out  <= csm_sdsp0_rst_n_i  AND rc_sdsp0_rst_n_i;
cb_sdsp1_rst_n_out  <= csm_sdsp1_rst_n_i  AND rc_sdsp1_rst_n_i;
cb_sdsp2_rst_n_out  <= csm_sdsp2_rst_n_i  AND rc_sdsp2_rst_n_i;
cb_sdsp3_rst_n_out  <= csm_sdsp3_rst_n_i  AND rc_sdsp3_rst_n_i;
 
flash_counter_init_i <= rrif_cntr_init_i    OR 
                        form_cntr_init_i    OR 
                        efb_cntr_init_i     OR 
                        router_cntr_init_i;

enable_counter_i <= enable_cnfg_counter_i;

--------------------------------------------------------------------------------
--PROCESS DECLARATION
--------------------------------------------------------------------------------
--
--  The process cnfg_state_machine is the main controller of the FPGA PRM
--  circuit.  The configuration state are run in a specific order in this block, 
--  and all functions are performed or setup here.
--
cnfg_state_machine : process (
  rst_n_in, 
  cclk_in, 
  cnfg_state, 
  cnfg_start_addr_i, 
  cnfg_stop_addr_i,
  rrif_cntr_init_i,
  form_cntr_init_i, 
  efb_cntr_init_i,
  router_cntr_init_i, 
  load_counter_i, 
  first_load_i,
  rrif_ready_i, 
  form_ready_i, 
  efb_ready_i,
  router_ready_i, 
  fpga_init_n_i, 
  flash_count_i, 
  wait_count_i, 
  timeout_count_i,
  cnfg_override_in
  )
begin
  if (rst_n_in = '0') then
    cnfg_state             <= startup;
    cnfg_start_addr_i      <= (others => '0');
    cnfg_stop_addr_i       <= (others => '0');
    wait_count_i           <= (others => '0');
    timeout_count_i        <= (others => '0');
    cb_clr_cnfg_reg_out    <= '0';
    fpga_sel_out           <= "111";
    cb_rrif_prgm_n_out     <= '0';
    cb_formA_prgm_n_out    <= '0';
    cb_formB_prgm_n_out    <= '0';
    cb_efb_prgm_n_out      <= '0';
    cb_router_prgm_n_out   <= '0';
    cb_spare_prgm_n_out    <= '0';
    rrif_cntr_init_i       <= '0';
    form_cntr_init_i       <= '0';
    efb_cntr_init_i        <= '0';
    router_cntr_init_i     <= '0';
    rrif_ready_i           <= '0';
    form_ready_i           <= '0';
    efb_ready_i            <= '0';
    router_ready_i         <= '0';
    csm_rrif_rst_n_i       <= '0';
    csm_formA_rst_n_i      <= '0';
    csm_formB_rst_n_i      <= '0';
    csm_efb_rst_n_i        <= '0';
    csm_router_rst_n_i     <= '0';
    csm_sdsp0_rst_n_i      <= '0';
    csm_sdsp1_rst_n_i      <= '0';
    csm_sdsp2_rst_n_i      <= '0';
    csm_sdsp3_rst_n_i      <= '0';
    halt_at_rrif_i         <= '0';
    halt_at_formA_i        <= '0';
    halt_at_formB_i        <= '0';
    halt_at_efb_i          <= '0';
    halt_at_router_i       <= '0';
    cnfg_done_out          <= '0';
    cnfg_enable_out        <= '0';
    load_counter_i         <= '0';  -- disable flash memory counter load enable
    first_load_i           <= '0';  -- clear first load bit
    fpga_init_n_i          <= '0';
    flush_shift_reg_i      <= '1';
    en_reset_control_i     <= '0';
  elsif (cclk_in'event and cclk_in = '1') then
    if (cnfg_override_in = '1') then
      cnfg_state         <= fpga_load_error;  -- Set State machine into NOP state
      cnfg_enable_out    <= '0';   -- Clear the Config Enable bit to allow Flash Access
      halt_at_rrif_i     <= '1';
      halt_at_formA_i    <= '1';
      halt_at_formB_i    <= '1';
      halt_at_efb_i      <= '1';
      halt_at_router_i   <= '1';
      fpga_sel_out       <= "111";
      fpga_init_n_i      <= '0';    -- 
      flush_shift_reg_i  <= '1';    -- flush the config data shift registers
      cnfg_done_out      <= '0';    -- config is not done and ROD is busy
      csm_sdsp0_rst_n_i  <= '0';    -- Hold all Slave DSPs in Reset
      csm_sdsp1_rst_n_i  <= '0';
      csm_sdsp2_rst_n_i  <= '0';
      csm_sdsp3_rst_n_i  <= '0';
    end if;

--  Following power up or reset, "startup' is the first state of the controller
    case cnfg_state is
      when startup => 
        cnfg_enable_out    <= '1';
        en_reset_control_i <= '0';
        cnfg_state         <= cnfg_all;

--  After the location registers have been loaded, the configuration process can begin.
--  The Location registers hold their initial values until power down of reset.

      when cnfg_all => 
        fpga_sel_out         <= "111";
        cb_rrif_prgm_n_out   <= '0';
        cb_formA_prgm_n_out  <= '0';
        cb_formB_prgm_n_out  <= '0';
        cb_efb_prgm_n_out    <= '0';
        cb_router_prgm_n_out <= '0';
        cb_spare_prgm_n_out  <= '0';
        load_counter_i       <= '0';
        cnfg_state           <= init_spare;
--      cnfg_state           <= init_rrif;
        cnfg_enable_out      <= '1';  -- flag to set flash controller in read only mode
        cnfg_done_out        <= '0';  -- flag to indicate if the cnfg is done
        rrif_ready_i         <= '0';  -- flag to indicate if the RRIF is ready for cnfg
        form_ready_i         <= '0';  -- flag to indicate if the FormA is ready for cnfg
        efb_ready_i          <= '0';  -- flag to indicate if the EFB is ready for cnfg
        router_ready_i       <= '0';  -- flag to indicate if the Router is ready for cnfg
        first_load_i         <= '1';  -- flag to indicate first load of any config cycle
        wait_count_i         <= (others => '0');

      when init_spare =>
        cnfg_state           <= init_rrif;
        cb_spare_prgm_n_out  <= '1';

--  A loop to allow the pgrm_n bit of the RRIF to cycle, and begin configuration
      when init_rrif =>
        cb_rrif_prgm_n_out <= '0';
        fpga_sel_out <= "111";
        flush_shift_reg_i <= '1';
        first_load_i    <= '1';     -- flag to indicate first load of any config cycle
        load_counter_i <= '0';
        cnfg_start_addr_i <= rrif_cnfg_start_addr_i; -- 0xE00000h
        cnfg_stop_addr_i <= rrif_cnfg_stop_addr_i;
        csm_rrif_rst_n_i <= '0';
        if (std_logic_vector(wait_count_i) = "100111") then
          cnfg_state <= load_rrif_counter;
          wait_count_i <= (others => '0');
        else
          wait_count_i <= wait_count_i + 1;
        end if;
                
      when load_rrif_counter =>
        cnfg_enable_out <= '1';
        load_counter_i <= '1';
        cnfg_state <= wait_for_rrif_counter_load;
        first_load_i <= '1';
        flush_shift_reg_i <= '0';
        rrif_cntr_init_i <= '1';
        cb_rrif_prgm_n_out <= '0';
        fpga_sel_out <= "111";

      when wait_for_rrif_counter_load =>
        cnfg_enable_out <= '1';
        load_counter_i <= '0';
        cnfg_state <= cnfg_rrif;
        first_load_i <= '1';
        flush_shift_reg_i <= '0';
        rrif_cntr_init_i <= '1';
        cb_rrif_prgm_n_out <= '0';
        fpga_sel_out <= "111";

      when cnfg_rrif =>
        if (cb_rrif_done_in = '0' AND 
            first_load_i = '1'    AND
            (sr0_empty_i = '1' OR sr1_empty_i = '1')) then
          first_load_i <= '1';
          rrif_ready_i <= '0';
        elsif (cb_rrif_done_in = '0' AND 
               first_load_i = '1'    AND
              (sr0_empty_i = '0' AND sr1_empty_i = '0')) then
          first_load_i <= '0';
          rrif_ready_i <= '1';
        elsif (cb_rrif_done_in = '0' AND 
               first_load_i = '0'    AND 
               rrif_ready_i = '1') then
          cb_rrif_prgm_n_out <= '1';
          fpga_sel_out <= "000";
          fpga_init_n_i <= cb_rrif_init_n_in;
          if (flash_count_i > unsigned(cnfg_stop_addr_i) AND
              std_logic_vector(timeout_count_i) = "00011111") then
            cb_clr_cnfg_reg_out <= '1';
            cnfg_state <= halt_at_rrif;
          elsif (flash_count_i > unsigned(cnfg_stop_addr_i) AND
                 std_logic_vector(timeout_count_i) < "00011111") then
            timeout_count_i <= timeout_count_i + 1;
          end if;
        elsif (cb_rrif_done_in = '1') then
          cnfg_state <= wait_for_rrif;
          rrif_cntr_init_i <= '0';
          halt_at_rrif_i <= '0';
          flush_shift_reg_i <= '1';
          fpga_init_n_i <= '0';
          timeout_count_i <= (others => '0');
        end if;
                
      when halt_at_rrif =>
        cb_clr_cnfg_reg_out <= '0';
        cnfg_state <= cnfg_done;
        halt_at_rrif_i <= '1';
        rrif_ready_i <= '0';
        cb_rrif_prgm_n_out <= '0';
        fpga_sel_out <= "111";
        flush_shift_reg_i <= '1';
        fpga_init_n_i <= '0';
        cb_clr_cnfg_reg_out <= '1';  -- if cnfg_all register bit is set, clear it here
        timeout_count_i <= (others => '0');

      when wait_for_rrif =>
        flush_shift_reg_i <= '0';
        if (std_logic_vector(wait_count_i) = "100111") then
          cb_clr_cnfg_reg_out <= '0';
          rrif_ready_i <= '0';
          csm_rrif_rst_n_i <= '1';
          fpga_sel_out <= "111";
          wait_count_i <= (others => '0');
          if (cnfg_mode = ms_cnfg_all) then
            cnfg_state <= init_form;
          elsif (cnfg_mode = ms_cnfg_rrif) then
            cnfg_state <= cnfg_done;
          end if;
        else
          wait_count_i <= wait_count_i + 1;
          cb_clr_cnfg_reg_out <= '1';
        end if;

      when init_router =>
        cb_router_prgm_n_out <= '0';
        fpga_sel_out <= "111";
        load_counter_i <= '0';
        first_load_i <= '1';     -- flag to indicate first load of any config cycle
        flush_shift_reg_i <= '1';
        cnfg_start_addr_i <= router_cnfg_start_addr_i;
        cnfg_stop_addr_i <= router_cnfg_stop_addr_i;
        csm_router_rst_n_i <= '0';
        if (std_logic_vector(wait_count_i) = "100111") then
          cnfg_state <= load_router_counter;
          wait_count_i <= (others => '0');
        else
          wait_count_i <= wait_count_i + 1;
        end if;

      when load_router_counter =>
        cnfg_enable_out <= '1';
        load_counter_i <= '1';
        cnfg_state <= wait_for_router_counter_load;
        first_load_i <= '1';
        flush_shift_reg_i <= '0';
        router_cntr_init_i <= '1';
        cb_router_prgm_n_out <= '0';
        fpga_sel_out <= "111";

      when wait_for_router_counter_load =>
        cnfg_enable_out <= '1';
        load_counter_i <= '0';
        cnfg_state <= cnfg_router;
        first_load_i <= '1';
        flush_shift_reg_i <= '0';
        router_cntr_init_i <= '1';
        cb_router_prgm_n_out <= '0';
        fpga_sel_out <= "111";

      when cnfg_router =>
        flush_shift_reg_i <= '0';
        if (cb_router_done_in = '0' AND 
            first_load_i = '1' AND 
           (sr0_empty_i = '1' OR sr1_empty_i = '1')) then
          first_load_i <= '1';
          router_ready_i <= '0';
        elsif (cb_router_done_in = '0' AND 
               first_load_i = '1' AND
              (sr0_empty_i = '0' AND sr1_empty_i = '0')) then
          first_load_i <= '0';
          router_ready_i <= '1';
        elsif (cb_router_done_in = '0' AND 
               first_load_i = '0' AND 
               router_ready_i = '1') then
          cb_router_prgm_n_out <= '1';
          fpga_sel_out <= "101";
          fpga_init_n_i <= cb_router_init_n_in;
          if (flash_count_i > unsigned(cnfg_stop_addr_i) AND
              std_logic_vector(timeout_count_i) = "11111110") then
            cb_clr_cnfg_reg_out <= '1';  -- clear router cnfg bit
            cnfg_state <= halt_at_router;
          elsif (flash_count_i > unsigned(cnfg_stop_addr_i) AND
                 std_logic_vector(timeout_count_i) < "11111110") then
            timeout_count_i <= timeout_count_i + 1;
          end if;
        elsif (cb_router_done_in = '1') then
          cnfg_state <= wait_for_router;
          router_cntr_init_i <= '0';
          halt_at_router_i <= '0';
          flush_shift_reg_i <= '1';
          fpga_init_n_i <= '0';
          timeout_count_i <= (others => '0');
        end if;

      when halt_at_router =>
        cnfg_state <= cnfg_done;
        cb_clr_cnfg_reg_out <= '0';
        router_ready_i <= '0';
        halt_at_router_i <= '1';
        cb_router_prgm_n_out <= '0';
        fpga_sel_out <= "111";
        flush_shift_reg_i <= '1';
        fpga_init_n_i <= '0';
        cb_clr_cnfg_reg_out <= '1';  -- if cnfg_all register bit is set, clear it here
        timeout_count_i <= (others => '0');

      when wait_for_router =>
        if (std_logic_vector(wait_count_i) = "100111") then
          cb_clr_cnfg_reg_out <= '0';
          router_ready_i <= '0';
          flush_shift_reg_i <= '0';
          fpga_sel_out <= "111";
          wait_count_i <= (others => '0');
          if (cnfg_mode = ms_cnfg_all) then
            cnfg_state <= clear_cnfg_all_bit;
--          cnfg_state <= init_efb;
          elsif (cnfg_mode = ms_cnfg_router) then
            cnfg_state <= cnfg_done;
          end if;
        else
          wait_count_i <= wait_count_i + 1;
          cb_clr_cnfg_reg_out <= '1';
        end if;

      when init_efb =>
        cb_efb_prgm_n_out <= '0';
        fpga_sel_out <= "111";
        load_counter_i <= '0';
        first_load_i    <= '1';     -- flag to indicate first load of any config cycle
        flush_shift_reg_i <= '1';
        cnfg_start_addr_i <= efb_cnfg_start_addr_i;
        cnfg_stop_addr_i <= efb_cnfg_stop_addr_i;
        csm_efb_rst_n_i <= '0';
        if (std_logic_vector(wait_count_i) = "100111") then
          cnfg_state <= load_efb_counter;
          wait_count_i <= (others => '0');
        else
          wait_count_i <= wait_count_i + 1;
        end if;

      when load_efb_counter =>
        cnfg_enable_out <= '1';
        load_counter_i <= '1';
        cnfg_state <= wait_for_efb_counter_load;
        first_load_i <= '1';
        flush_shift_reg_i <= '0';
        efb_cntr_init_i <= '1';
        cb_efb_prgm_n_out <= '0';
        fpga_sel_out <= "111";

      when wait_for_efb_counter_load =>
        cnfg_enable_out <= '1';
        load_counter_i <= '0';
        cnfg_state <= cnfg_efb;
        first_load_i <= '1';
        efb_cntr_init_i <= '1';
        flush_shift_reg_i <= '0';
        cb_efb_prgm_n_out <= '0';
        fpga_sel_out <= "111";

      when cnfg_efb =>
        flush_shift_reg_i <= '0';
        if (cb_efb_done_in = '0' AND 
            first_load_i = '1' AND
           (sr0_empty_i = '1' OR sr1_empty_i = '1')) then
          first_load_i <= '1';
          efb_ready_i <= '0';
        elsif (cb_efb_done_in = '0' AND 
               first_load_i = '1' AND
              (sr0_empty_i = '0' AND sr1_empty_i = '0')) then
          first_load_i <= '0';
          efb_ready_i <= '1';
        elsif (cb_efb_done_in = '0' AND 
               first_load_i = '0' AND 
               efb_ready_i = '1') then
          cb_efb_prgm_n_out <= '1';
          fpga_sel_out <= "100";
          fpga_init_n_i <= cb_efb_init_n_in;
          if (flash_count_i > unsigned(cnfg_stop_addr_i) AND
              std_logic_vector(timeout_count_i) = "11111110") then
            cb_clr_cnfg_reg_out <= '1';
            cnfg_state <= halt_at_efb;
          elsif (flash_count_i > unsigned(cnfg_stop_addr_i) AND
                 std_logic_vector(timeout_count_i) < "11111110") then
            timeout_count_i <= timeout_count_i + 1;
          end if;
        elsif (cb_efb_done_in = '1') then
          cnfg_state <= wait_for_efb;
          efb_cntr_init_i <= '0';
          halt_at_efb_i <= '0';
          flush_shift_reg_i <= '1';
          fpga_init_n_i <= '0';
          timeout_count_i <= (others => '0');
        end if;

      when halt_at_efb =>
        cnfg_state <= cnfg_done;
        cb_clr_cnfg_reg_out <= '0';
        efb_ready_i <= '0';
        halt_at_efb_i <= '1';
        cb_efb_prgm_n_out <= '0';
        fpga_sel_out <= "111";
        flush_shift_reg_i <= '1';
        fpga_init_n_i <= '0';
        cb_clr_cnfg_reg_out <= '1';  -- if cnfg_all register bit is set, clear it here
        timeout_count_i <= (others => '0');

      when wait_for_efb =>
        if (std_logic_vector(wait_count_i) = "100111") then
          cb_clr_cnfg_reg_out <= '0';
          efb_ready_i <= '0';
          flush_shift_reg_i <= '0';
          fpga_sel_out <= "111";
          wait_count_i <= (others => '0');
          if (cnfg_mode = ms_cnfg_all) then
            cnfg_state <= init_router;
--          cnfg_state <= init_form;
          elsif (cnfg_mode = ms_cnfg_efb) then
            cnfg_state <= cnfg_done;
          end if;
        else
          wait_count_i <= wait_count_i + 1;
          cb_clr_cnfg_reg_out <= '1';
        end if;

      when init_form =>
        cb_formA_prgm_n_out <= '0';
        cb_formB_prgm_n_out <= '0';
        fpga_sel_out <= "111";
        load_counter_i <= '0';
        first_load_i    <= '1';     -- flag to indicate first load of any config cycle
        flush_shift_reg_i <= '1';
        cnfg_start_addr_i <= form_cnfg_start_addr_i;
        cnfg_stop_addr_i <= form_cnfg_stop_addr_i;
        csm_formA_rst_n_i <= '0';
        csm_formB_rst_n_i <= '0';
        if (std_logic_vector(wait_count_i) = "100111") then
          cnfg_state <= load_form_counter;
          wait_count_i <= (others => '0');
        else
          wait_count_i <= wait_count_i + 1;
        end if;

      when load_form_counter =>
        cnfg_enable_out <= '1';
        load_counter_i <= '1';
        cnfg_state <= wait_for_form_counter_load;
        first_load_i <= '1';
        flush_shift_reg_i <= '1';
        form_cntr_init_i <= '1';
        cb_formA_prgm_n_out <= '0';
        cb_formB_prgm_n_out <= '0';
        fpga_sel_out <= "111";

      when wait_for_form_counter_load =>
        cnfg_enable_out <= '1';
        load_counter_i <= '0';
        cnfg_state <= cnfg_form;
        first_load_i <= '1';
        flush_shift_reg_i <= '0';
        form_cntr_init_i <= '1';
        cb_formA_prgm_n_out <= '0';
        cb_formB_prgm_n_out <= '0';
        fpga_sel_out <= "111";

      when cnfg_form =>
        flush_shift_reg_i <= '0';
        load_counter_i <= '0';

        if (cb_formA_done_in = '0' AND 
            cb_formB_done_in = '0' AND 
            first_load_i = '1' AND
           (sr0_empty_i = '1' OR sr1_empty_i = '1')) then
          first_load_i <= '1';
          form_ready_i <= '0';
        elsif (cb_formA_done_in = '0' AND 
               cb_formB_done_in = '0' AND 
               first_load_i = '1' AND
              (sr0_empty_i = '0' AND sr1_empty_i = '0')) then
          first_load_i <= '0';
          form_ready_i <= '1';
        elsif (cb_formA_done_in = '0' AND 
               cb_formB_done_in = '0' AND 
               first_load_i = '0' AND
               form_ready_i = '1') then
          cb_formA_prgm_n_out <= '1';
          cb_formB_prgm_n_out <= '1';
          fpga_sel_out <= "001";
          fpga_init_n_i <= cb_formA_init_n_in AND cb_formB_init_n_in;

          if (flash_count_i > unsigned(cnfg_stop_addr_i) AND 
              std_logic_vector(timeout_count_i) = "11111110") then
            cb_clr_cnfg_reg_out <= '1';
            cnfg_state <= halt_at_form;
          elsif (flash_count_i > unsigned(cnfg_stop_addr_i) AND
                 load_counter_i = '0' AND
                 std_logic_vector(timeout_count_i) = "00000000") then
            cnfg_start_addr_i <= form_cnfg_start_addr_i;
            cnfg_stop_addr_i <= form_cnfg_stop_addr_i;
            load_counter_i <= '1';
            timeout_count_i <= timeout_count_i + 1;
          elsif (flash_count_i > unsigned(cnfg_stop_addr_i) AND
                 load_counter_i = '0'  AND
                 std_logic_vector(timeout_count_i) = "00000001") then
            cnfg_start_addr_i <= form_cnfg_start_addr_i;
            cnfg_stop_addr_i <= form_cnfg_stop_addr_i;
            load_counter_i <= '1';
            timeout_count_i <= timeout_count_i + 1;
          elsif (flash_count_i > unsigned(cnfg_stop_addr_i) AND
                 load_counter_i = '0'  AND
                 std_logic_vector(timeout_count_i) = "00000010") then
            cnfg_start_addr_i <= form_cnfg_start_addr_i;
            cnfg_stop_addr_i <= form_cnfg_stop_addr_i;
            load_counter_i <= '1';
            timeout_count_i <= timeout_count_i + 1;
          elsif (flash_count_i > unsigned(cnfg_stop_addr_i) AND
                 load_counter_i = '0'  AND
                 std_logic_vector(timeout_count_i) >= "00000011" AND
                 std_logic_vector(timeout_count_i) < "11111110") then
            timeout_count_i <= timeout_count_i + 1;
          end if;
        elsif (cb_formA_done_in = '1' AND cb_formB_done_in = '1') then
          cnfg_state <= wait_for_form;
          form_cntr_init_i <= '0';
          halt_at_formA_i <= '0';
          halt_at_formB_i <= '0';
          flush_shift_reg_i <= '1';
          fpga_init_n_i <= '0';
          timeout_count_i <= (others => '0');
        end if;

      when halt_at_form =>
        if (cb_formA_done_in = '0' AND cb_formB_done_in = '0') then
          halt_at_formA_i <= '1';
          halt_at_formB_i <= '1';
          cb_formA_prgm_n_out <= '0';
          cb_formB_prgm_n_out <= '0';
        elsif (cb_formA_done_in = '0' AND cb_formB_done_in = '1') then
          halt_at_formA_i <= '1';
          cb_formA_prgm_n_out <= '0';
        elsif (cb_formA_done_in = '1' AND cb_formB_done_in = '0') then
          halt_at_formB_i <= '1';
          cb_formB_prgm_n_out <= '0';
        end if;
                    
        cb_clr_cnfg_reg_out <= '0';
        cnfg_state <= cnfg_done;
        form_ready_i <= '0';
        fpga_sel_out <= "111";
        flush_shift_reg_i <= '1';
        fpga_init_n_i <= '0';
        cb_clr_cnfg_reg_out <= '1';  -- if cnfg_all register bit is set, clear it here
        timeout_count_i <= (others => '0');

      when wait_for_form =>
        if (std_logic_vector(wait_count_i) = "100111") then
          cb_clr_cnfg_reg_out <= '0';
          form_ready_i <= '0';
          flush_shift_reg_i <= '0';
          fpga_sel_out <= "111";
          wait_count_i <= (others => '0');
          if (cnfg_mode = ms_cnfg_all) then
            cnfg_state <= init_efb;
--          cnfg_state <= clear_cnfg_all_bit;
          elsif (cnfg_mode = ms_cnfg_form) then
            cnfg_state <= cnfg_done;
          end if;
        else
          wait_count_i <= wait_count_i + 1;
          cb_clr_cnfg_reg_out <= '1';
        end if;

      when clear_cnfg_all_bit =>
        if (std_logic_vector(wait_count_i) = "100111") then
          cnfg_state <= cnfg_done;
          cb_clr_cnfg_reg_out <= '0';
          wait_count_i <= (others => '0');
        else
          wait_count_i <= wait_count_i + 1;
          cb_clr_cnfg_reg_out <= '1';
        end if;

      when cnfg_done => 
        cnfg_enable_out    <= '0';
        flush_shift_reg_i  <= '0';
        csm_formA_rst_n_i  <= '1';
        csm_formB_rst_n_i  <= '1';
        csm_efb_rst_n_i    <= '1';
        csm_router_rst_n_i <= '1';
  -- Release Slave DSP resets here
        csm_sdsp0_rst_n_i  <= '1';
        csm_sdsp1_rst_n_i  <= '1';
        csm_sdsp2_rst_n_i  <= '1';
        csm_sdsp3_rst_n_i  <= '1';
        en_reset_control_i <= '1';
        
        cb_clr_cnfg_reg_out <= '0';
        wait_count_i <= (others => '0');

        if (halt_at_rrif_i = '1'   OR 
            halt_at_router_i = '1' OR 
            halt_at_efb_i = '1'    OR
            halt_at_formA_i = '1'  OR 
            halt_at_formB_i = '1') then
          cnfg_done_out <= '0';
        else
          cnfg_done_out <= '1';
        end if;

        case cnfg_mode is
          when ms_cnfg_all => cnfg_state <= startup;
          when ms_cnfg_rrif => 
            cnfg_state <= init_rrif;
            cnfg_enable_out <= '1';
            cnfg_done_out <= '0';
            rrif_ready_i <= '0';
            first_load_i <= '1';         -- first load of any config cycle
          when ms_cnfg_form => 
            cnfg_state <= init_form;
            cnfg_enable_out <= '1';
            cnfg_done_out <= '0';
            form_ready_i <= '0';
            first_load_i <= '1';         -- first load of any config cycle
          when ms_cnfg_efb => 
            cnfg_state <= init_efb;
            cnfg_enable_out <= '1';
            cnfg_done_out <= '0';
            efb_ready_i <= '0';
            first_load_i <= '1';         -- first load of any config cycle
          when ms_cnfg_router => 
            cnfg_state <= init_router;
            cnfg_enable_out <= '1';
            cnfg_done_out <= '0';
            router_ready_i <= '0';
            first_load_i <= '1';         -- first load of any config cycle
--        when ms_no_cnfg =>
--          cnfg_state <= cnfg_done;
--          cnfg_enable_out <= '0';
          when others =>
            cnfg_state <= cnfg_done;
            cnfg_enable_out <= '0';
        end case;

      when others => null;
    end case;
    cb_formA_halt_out   <= halt_at_formA_i;
    cb_formB_halt_out   <= halt_at_formB_i;
    cb_efb_halt_out     <= halt_at_efb_i;
    cb_router_halt_out  <= halt_at_router_i;
    cb_rrif_halt_out    <= halt_at_rrif_i;
  end if;
end process cnfg_state_machine;

--------------------------------------------------------------------------------        
flash_address_counter : process (
  rst_n_in, 
  cclk_in, 
  enable_counter_i,
  load_counter_i, 
  flash_count_i,
  cnfg_start_addr_i, 
  cnfg_stop_addr_i,
  sr0_empty_i, 
  sr1_empty_i
  )
begin
  if (rst_n_in = '0') then
    cnfg_addr_out <= (others => '0');  -- if reset is true,
    flash_count_i <= (others => '0');
    load_last_cnfg_word_i <= '0';
  elsif (cclk_in'event and cclk_in = '1') then
    if (load_counter_i = '1') then
      flash_count_i <= unsigned(cnfg_start_addr_i);
      load_last_cnfg_word_i <= '0';
    elsif (enable_counter_i = '1' AND 
           flash_count_i <= unsigned(cnfg_stop_addr_i)) then
      flash_count_i <= flash_count_i + 1;
      if (flash_count_i = unsigned(cnfg_stop_addr_i)) then
        load_last_cnfg_word_i <= '1';
      end if;
    elsif (enable_counter_i = '0' AND
           flash_count_i > unsigned(cnfg_stop_addr_i) AND
           (sr0_empty_i = '1' AND sr1_empty_i = '1')) then
      load_last_cnfg_word_i <= '0';
    end if;
    cnfg_addr_out <= std_logic_vector(flash_count_i);
  end if;
end process flash_address_counter;

--------------------------------------------------------------------------------        
shift_register_load_controller : process (
  rst_n_in, 
  cclk_in, 
  last_cnfg_word_i,
  first_load_i, 
  flash_counter_init_i,
  shift_reg_load_i, 
  shift_reg0_count_i,
  shift_reg1_count_i, 
  shift_reg_en_out_i,
  flush_shift_reg_i
  )
begin
  if (rst_n_in = '0') then
    shift_reg_load_out <= (others => '0');
    shift_reg_load_i <= (others => '0');
    sr0_empty_i  <= '1';   -- Shift Register 0 is empty
    sr1_empty_i  <= '1';   -- Shift Register 1 is empty
    enable_cnfg_counter_i <= '0';
  elsif (cclk_in'event and cclk_in = '1') then
    if (cnfg_state = startup OR 
        cnfg_state = cnfg_all) then
          null;
    elsif (cnfg_state = init_rrif OR 
           cnfg_state = init_router OR
           cnfg_state = init_efb OR 
           cnfg_state = init_form) then
      sr0_empty_i  <= '1';   -- Shift Register 0 is empty
      sr1_empty_i  <= '1';   -- Shift Register 1 is empty
    else
      if (flush_shift_reg_i = '1') then
        sr0_empty_i  <= '1';   -- Shift Register 0 is empty
        sr1_empty_i  <= '1';   -- Shift Register 1 is empty
      elsif (last_cnfg_word_i = '0' AND first_load_i = '1' 
          AND flash_counter_init_i = '1' AND sr0_empty_i = '1' 
          AND sr1_empty_i = '1') then
        enable_cnfg_counter_i <= '1';
        sr0_empty_i <= '0';
        shift_reg_load_i <= "01";
      elsif (last_cnfg_word_i = '0' AND 
             first_load_i = '1' AND
             flash_counter_init_i = '1' AND 
             sr0_empty_i = '0' AND
             sr1_empty_i = '1') then
        enable_cnfg_counter_i <= '1';
        sr1_empty_i <= '0';
        shift_reg_load_i <= "10";
      elsif (last_cnfg_word_i = '0' AND 
             first_load_i = '0' AND
             flash_counter_init_i = '1' AND 
             sr0_empty_i = '1' AND
             sr1_empty_i = '0') then
        enable_cnfg_counter_i <= '1';
        sr0_empty_i <= '0';
        shift_reg_load_i <= "01";
      elsif (last_cnfg_word_i = '0' AND 
             first_load_i = '0' AND
             flash_counter_init_i = '1' AND 
             sr0_empty_i = '0' AND
             sr1_empty_i = '1') then
        enable_cnfg_counter_i <= '1';
        sr1_empty_i <= '0';
        shift_reg_load_i <= "10";
      elsif (first_load_i = '0' AND 
             flash_counter_init_i = '1' AND
             std_logic_vector(shift_reg0_count_i) = "111") then
        sr0_empty_i <= '1';
      elsif (first_load_i = '0' AND 
             flash_counter_init_i = '1' AND
             std_logic_vector(shift_reg1_count_i) = "111") then
        sr1_empty_i <= '1';
      elsif (last_cnfg_word_i = '1' AND
             std_logic_vector(shift_reg0_count_i) = "000" AND
             std_logic_vector(shift_reg1_count_i) = "000" AND
             shift_reg_en_out_i = "00") then
        sr0_empty_i <= '1';
        sr1_empty_i <= '1';
      else
        enable_cnfg_counter_i <= '0';
        shift_reg_load_i <= "00";
      end if;
    end if;
    shift_reg_load_out <= shift_reg_load_i;
  end if;
end process shift_register_load_controller;

-------------------------------------------------------------------------------
shift_register_controller : process (
  rst_n_in,  
  cclk_in, 
  cnfg_state,
  fpga_init_n_i, 
  last_cnfg_word_i,
  load_last_cnfg_word_i,
  sr0_empty_i, 
  sr1_empty_i,
  shift_reg0_count_i, 
  shift_reg1_count_i,
  shift_reg_en_out_i, 
  shift_reg_en_d1
  )
begin
  if (rst_n_in = '0') then
    shift_reg_en_out_i <= (others => '0');
    shift_reg_en_d1 <= (others => '0');
    shift_reg0_count_i <= (others => '0');
    shift_reg1_count_i <= (others => '0');
    last_cnfg_word_i <= '0';
  elsif (cclk_in'event and cclk_in = '1') then
    if (cnfg_state = startup OR cnfg_state = cnfg_all) then
          null;
    else
      if (last_cnfg_word_i = '0' AND sr0_empty_i = '1' AND sr1_empty_i = '1') then 
        null;
      elsif (fpga_init_n_i = '1' AND last_cnfg_word_i = '0') then
        if   (std_logic_vector(shift_reg0_count_i) = "000" AND 
              std_logic_vector(shift_reg1_count_i) = "000" AND 
              shift_reg_en_out_i = "00") then
          shift_reg_en_out_i <= "01";
        elsif (std_logic_vector(shift_reg0_count_i) < "111" AND
               std_logic_vector(shift_reg1_count_i) = "000" AND
               shift_reg_en_out_i = "01") then
          shift_reg0_count_i <= shift_reg0_count_i + 1;
        elsif (std_logic_vector(shift_reg0_count_i) = "111" AND
               std_logic_vector(shift_reg1_count_i) = "000" AND
               shift_reg_en_out_i = "01") then
          shift_reg_en_out_i <= "10";
          shift_reg0_count_i <= (others => '0');
          if (load_last_cnfg_word_i = '1') then
            last_cnfg_word_i <= '1';
          else
            last_cnfg_word_i <= '0';
          end if;
        elsif (std_logic_vector(shift_reg0_count_i) = "000" AND
               std_logic_vector(shift_reg1_count_i) < "111" AND 
               shift_reg_en_out_i = "10") then
          shift_reg1_count_i <= shift_reg1_count_i + 1;
        elsif (std_logic_vector(shift_reg0_count_i) = "000" AND
               std_logic_vector(shift_reg1_count_i) = "111" AND
               shift_reg_en_out_i = "10") then
          shift_reg_en_out_i <= "01";
          shift_reg1_count_i <= (others => '0');
          if (load_last_cnfg_word_i = '1') then
            last_cnfg_word_i <= '1';
          else
            last_cnfg_word_i <= '0';
          end if;
        end if;

      elsif (fpga_init_n_i = '1' AND last_cnfg_word_i = '1') then
        if (std_logic_vector(shift_reg0_count_i) < "111" AND
            sr1_empty_i = '1') then
          shift_reg_en_out_i <= "01";
          shift_reg0_count_i <= shift_reg0_count_i + 1;
        elsif (std_logic_vector(shift_reg1_count_i) < "111" AND
               sr0_empty_i = '1') then
          shift_reg_en_out_i <= "10";
          shift_reg1_count_i <= shift_reg1_count_i + 1;
        else
          shift_reg_en_out_i <= "00";
          last_cnfg_word_i <= '0';
          shift_reg0_count_i <= "000";
          shift_reg1_count_i <= "000";
        end if;
      end if;
      shift_reg_en_d1  <= shift_reg_en_out_i;
      shift_reg_en_out <= shift_reg_en_d1;
    end if;
  end if;
end process shift_register_controller;
--
-------------------------------------------------------------------------------
--
--  The process set_cnfg_mode reads the bits from the cnfg register blocks
--  and sets the cnfg_mode state.  Only one bit is allowed to be set at a time.
--  The process only reads from the configuration register when FPGA prm is
--  done withall configuration processes.
--
set_cnfg_mode : process (
  clk_in, 
  rst_n_in, 
  cnfg_state, 
  cb_fpga_cnfg_reg_in
  )
subtype select_device_to_cnfg is std_logic_vector(5 downto 0);
begin
  if (rst_n_in = '0') then
    cnfg_mode <= ms_cnfg_all;
  elsif (clk_in'event AND clk_in = '1') then
    if (cnfg_state = cnfg_done) then
      case select_device_to_cnfg(cb_fpga_cnfg_reg_in) is
        when "000001" => cnfg_mode <= ms_cnfg_rrif;
        when "000010" => cnfg_mode <= ms_cnfg_form;
        when "000100" => cnfg_mode <= ms_cnfg_form;
        when "001000" => cnfg_mode <= ms_cnfg_efb;
        when "010000" => cnfg_mode <= ms_cnfg_router;
        when "100000" => cnfg_mode <= ms_cnfg_all;
        when others   => cnfg_mode <= ms_no_cnfg;
      end case;
    end if;
  end if;
end process set_cnfg_mode;

-------------------------------------------------------------------------------
--
--  The process reset_device_controller resets the devices on the ROD board by
--  reading the reset register located in the register block.  If a reset
--  command is sent, the reset controller will clear the appropriate register
--  when the reset is complete.
--
device_reset_controller : process (
  cclk_in, 
  rst_n_in,
  cb_fpga_rst_reg_in, 
  cb_dsp_rst_reg_in
  )
subtype select_device_to_reset is std_logic_vector(11 downto 0);
begin
  if (rst_n_in = '0') then
    reset_count_i      <= (others => '0');
    cb_clr_rst_reg_out <= '0';
    rc_rrif_rst_n_i    <= '1';
    rc_formA_rst_n_i   <= '1';
    rc_formB_rst_n_i   <= '1';
    rc_efb_rst_n_i     <= '1';
    rc_router_rst_n_i  <= '1';
    rc_mdsp_rst_n_i    <= '1';
    rc_sdsp0_rst_n_i   <= '1';
    rc_sdsp1_rst_n_i   <= '1';
    rc_sdsp2_rst_n_i   <= '1';
    rc_sdsp3_rst_n_i   <= '1';
  elsif (cclk_in'event AND cclk_in = '1') then
    if (en_reset_control_i = '1') then
      case select_device_to_reset(cb_dsp_rst_reg_in & cb_fpga_rst_reg_in) is
-- Only 1 CCLK required to successfully reset an FPGA
        when "000000000001" =>
          cb_clr_rst_reg_out <= '1';
          rc_rrif_rst_n_i    <= '0';

        when "000000000010" =>
          cb_clr_rst_reg_out <= '1';
          rc_formA_rst_n_i   <= '0';

        when "000000000100" =>
          cb_clr_rst_reg_out <= '1';
          rc_formB_rst_n_i   <= '0';

        when "000000001000" => 
          cb_clr_rst_reg_out <= '1';
          rc_efb_rst_n_i     <= '0';

        when "000000010000" => 
          cb_clr_rst_reg_out <= '1';
          rc_router_rst_n_i <= '0';
  
        when "000000100000" =>
          cb_clr_rst_reg_out <= '1';
          rc_rrif_rst_n_i   <= '0';
          rc_formA_rst_n_i  <= '0';
          rc_formB_rst_n_i  <= '0';
          rc_efb_rst_n_i    <= '0';
          rc_router_rst_n_i <= '0';
-- The   DSPs require a long reset pulse
        when "000001000000" =>
          if (reset_count_i > 524000) then   -- 140ms
            rc_mdsp_rst_n_i    <= '1';
            cb_clr_rst_reg_out <= '1';
          else
            reset_count_i <= reset_count_i + 1;
            rc_mdsp_rst_n_i <= '0';
          end if;

        when "000010000000" =>
          if (reset_count_i > 524000) then   -- 140ms
            rc_sdsp0_rst_n_i   <= '1';
            cb_clr_rst_reg_out <= '1';
          else
            reset_count_i <= reset_count_i + 1;
            rc_sdsp0_rst_n_i <= '0';
          end if;

        when "000100000000" =>
          if (reset_count_i > 524000) then   -- 140ms
            rc_sdsp1_rst_n_i  <= '1';
            cb_clr_rst_reg_out <= '1';
          else
            reset_count_i <= reset_count_i + 1;
            rc_sdsp1_rst_n_i <= '0';
          end if;

        when "001000000000" => 
          if (reset_count_i > 524000) then   -- 140ms
            rc_sdsp2_rst_n_i   <= '1';
            cb_clr_rst_reg_out <= '1';
          else
            reset_count_i <= reset_count_i + 1;
            rc_sdsp2_rst_n_i <= '0';
          end if;

        when "010000000000" => 
          if (reset_count_i > 524000) then   -- 140ms
            rc_sdsp3_rst_n_i   <= '1';
            cb_clr_rst_reg_out <= '1';
          else
            reset_count_i <= reset_count_i + 1;
            rc_sdsp3_rst_n_i <= '0';
          end if;
                
        when "100000000000" =>
          if (reset_count_i > 524000) then   -- 140ms
            rc_rrif_rst_n_i    <= '1';
            rc_formA_rst_n_i   <= '1';
            rc_formB_rst_n_i   <= '1';
            rc_efb_rst_n_i     <= '1';
            rc_router_rst_n_i  <= '1';
            rc_mdsp_rst_n_i    <= '1';
            rc_sdsp0_rst_n_i   <= '1';
            rc_sdsp1_rst_n_i   <= '1';
            rc_sdsp2_rst_n_i   <= '1';
            rc_sdsp3_rst_n_i   <= '1';
            cb_clr_rst_reg_out <= '1';
          else
            reset_count_i <= reset_count_i + 1;
            rc_rrif_rst_n_i   <= '0';
            rc_formA_rst_n_i  <= '0';
            rc_formB_rst_n_i  <= '0';
            rc_efb_rst_n_i    <= '0';
            rc_router_rst_n_i <= '0';
            rc_mdsp_rst_n_i   <= '0';
            rc_sdsp0_rst_n_i  <= '0';
            rc_sdsp1_rst_n_i  <= '0';
            rc_sdsp2_rst_n_i  <= '0';
            rc_sdsp3_rst_n_i  <= '0';
          end if;
                  
        when others => 
          rc_rrif_rst_n_i    <= '1';
          rc_formA_rst_n_i   <= '1';
          rc_formB_rst_n_i   <= '1';
          rc_efb_rst_n_i     <= '1';
          rc_router_rst_n_i  <= '1';
          rc_mdsp_rst_n_i    <= '1';
          rc_sdsp0_rst_n_i   <= '1';
          rc_sdsp1_rst_n_i   <= '1';
          rc_sdsp2_rst_n_i   <= '1';
          rc_sdsp3_rst_n_i   <= '1';
          cb_clr_rst_reg_out <= '0';
          reset_count_i      <= (others => '0');
      end case;
    else
      rc_rrif_rst_n_i    <= '1';
      rc_formA_rst_n_i   <= '1';
      rc_formB_rst_n_i   <= '1';
      rc_efb_rst_n_i     <= '1';
      rc_router_rst_n_i  <= '1';
      rc_mdsp_rst_n_i    <= '1';
      rc_sdsp0_rst_n_i   <= '1';
      rc_sdsp1_rst_n_i   <= '1';
      rc_sdsp2_rst_n_i   <= '1';
      rc_sdsp3_rst_n_i   <= '1';
      cb_clr_rst_reg_out <= '0';
      reset_count_i      <= (others => '0');
    end if;
  end if;
end process device_reset_controller;
------------------------------------------------------------------------------------------
--
--
register_status_out : process (
  cclk_in,
  rst_n_in
  )
begin
  if (rst_n_in = '0') then
    cb_fpga_rst_status_reg_out <= (others => '0');
    cb_dsp_rst_status_reg_out  <= (others => '0');
  elsif (cclk_in'event AND cclk_in = '1') then
    cb_fpga_rst_status_reg_out(0) <= rc_rrif_rst_n_i   AND csm_rrif_rst_n_i;
    cb_fpga_rst_status_reg_out(1) <= rc_formA_rst_n_i  AND csm_formA_rst_n_i;
    cb_fpga_rst_status_reg_out(2) <= rc_formB_rst_n_i  AND csm_formB_rst_n_i;
    cb_fpga_rst_status_reg_out(3) <= rc_efb_rst_n_i    AND csm_efb_rst_n_i;
    cb_fpga_rst_status_reg_out(4) <= rc_router_rst_n_i AND csm_router_rst_n_i;

    cb_dsp_rst_status_reg_out(0)  <= rc_mdsp_rst_n_i;
    cb_dsp_rst_status_reg_out(1)  <= rc_sdsp0_rst_n_i AND csm_sdsp0_rst_n_i ;
    cb_dsp_rst_status_reg_out(2)  <= rc_sdsp1_rst_n_i AND csm_sdsp1_rst_n_i;
    cb_dsp_rst_status_reg_out(3)  <= rc_sdsp2_rst_n_i AND csm_sdsp2_rst_n_i;
    cb_dsp_rst_status_reg_out(4)  <= rc_sdsp3_rst_n_i AND csm_sdsp3_rst_n_i;
  end if;
end process register_status_out;

------------------------------------------------------------------------------------------

end rtl;


